#!/usr/bin/perl
#
# boost_install.pl
#
# ABSTRACT:
# ============
# A perl script to compile the boost libraries and build bjam.
# 
# Use "boost_install.pl --help" to get more details on usage and command line options.
# 
use File::Basename;
use File::Copy;
use File::Spec;
use File::Path;
use DirHandle;
use FileHandle;
use Getopt::Long;
use English;

use strict 'vars';

#
# Please align it with the boost build rule set; i.e. the boost version the rule set
# was taken from.
#
#
# BOOST      BOOST RULE SET     COMMENT
# ===============================================================         
# 1.40.0     1.40.O             perfect alignment
# 1.39.0     1.40.0             undefined and disabled
# >1.40.0    1.40.0             undefined and disabled
#

# Notice: the trailing -<n> indicates the current revision of the BoostBuild release.
# For the time being, a version spec like '1.40.0-x' is supposed to be compatible with 
# boost 1.40.0. 
my $boost_build_version = '1.41.0-0';

my $bListUsage = 0;
my $bDryRun = 0;
my $builtin_test = '';

my $boost_root = '';
my $boost_root_fnd = 0;
my $boost_bin_dir;
my $boost_lib_dir;

my $bBuildBjam = -1;
my $bBuildLibs = -1;
my $bStaticLibs = 1;
my $bSharedLibs = 1;
my $bBuildLib32 = -1;
my $bBuildLib64 = -1;
my $bRebuildAll = -1;

# a string consisting of colon speparated system attributes
my $system_info;

my $bjam_bin_dir = '';
my $bjam_jobs = 0;

my $boost_toolset = '';

my $bQuitOnError = 1;
my $bKeepTmpFiles = 0;

my $bStaticLibs32 = 0;
my $bSharedLibs32 = 0;

my $bStaticLibs64 = 0;
my $bSharedLibs64 = 0;

my $bIsPlatform64Bit = 0;

my $boost_build_dir;
my $tool_dir;
my $boost_build_srcdir;
my $boost_build_destdir;
my $boost_link_options = 'link=';
my $boost_toolset_dir_suffix = '';
my $bjam_options = '';

my $bjam_prog;
my $bjam_prog_quoted;
my $bjam_prog_built;
my $b_bjam_prog_updated = 0;
my $b_bjam_bin_dir_on_path = 0;

my $boost_lib_exclude_list_args = '';
my @boost_lib_exclude_list = ( 'python' );

my $boost_lib_include_list_args = '';
my @boost_lib_include_list;


# the jamroot file to be searched for "path-constant BOOST_SDK_ROOT : ../Boost_1_40_0 ;" etc ...
my $Jamroot;

# The boost version string in dotted decimal format
my $boost_version;

# splitted in separate numbers; $boost_version_elems[ 0 ] => major ID, $boost_version_elems[ 1 ] => minor ID, ...
my @boost_version_elems;


my $build_log_msg;
my @build_log_summary;

my @todo_list;

# the directory to store boost patches
my $boost_patch_dir;
my @boost_patch_file_list;

my %optctl = (
  'help'           => \$bListUsage,
  'dry-run'        => \$bDryRun,
  'quit-on-error'  => \$bQuitOnError,
  'bjam-dir'       => \$bjam_bin_dir,
  'build-bjam'     => \$bBuildBjam,
  'rebuild'        => \$bRebuildAll,
  'build-libs'     => \$bBuildLibs,
  'toolset'        => \$boost_toolset,
  'static'         => \$bStaticLibs,
  'shared'         => \$bSharedLibs,
  '64'             => \$bBuildLib64,
  '32'             => \$bBuildLib32,
  'jobs'           => \$bjam_jobs,
  'keep-tmp-files' => \$bKeepTmpFiles,
  'boost-exclude'  => \$boost_lib_exclude_list_args,
  'boost-include'  => \$boost_lib_include_list_args,
  
  'test-case'      => \$builtin_test
);

# a hash to store a mapping between header files and package names; e.g. if zlib.h is missing, tell the user
# which package needs to be installed.
my %missing_pkg_info;



# parse the command line 
if( GetOptions( \%optctl, 
                'help!', 
                'dry-run!', 
                'quit-on-error!', 
                'build-bjam!', 
                'build-libs!', 
                'toolset=s', 
                'static!', 
                'shared!', 
                '64!', 
                '32!', 
                'jobs=i', 
                'keep-tmp-files!',
                'bjam-dir=s',
                'rebuild!',
                'boost-exclude=s',
                'boost-include=s',
                'test-case=s'    
                  ) != 1 ) {
  fatal_error( __LINE__, "command line not understood, check the syntax\n" );
}

if( $bListUsage != 0 ) {
  usage();
  exit 0;
}

if( scalar( @ARGV ) > 0 ) {
  usage();
  exit 1;
}

# find "BoostBuild" given the script name
$boost_build_dir = find_boost_build_dir( $0 );

# get the directory of boost_install.pl, which is needed to look up other helper scripts. 
$tool_dir = File::Spec->catdir( $boost_build_dir, 'bin' );

# some platforms require patches to compile the boost libraries and these patches are 
# stored in $boost_patch_dir.
$boost_patch_dir = File::Spec->catdir( $boost_build_dir, 'src', 'boost_patches' );

$system_info = get_system_info( $tool_dir );
#print "system_info: \"$system_info\"\n";

#detect 64bit platforms
$bIsPlatform64Bit = is_platform_64bit( $system_info );
#exit 0;


if( $boost_toolset eq '' ) {
  $boost_toolset = get_default_toolset( $tool_dir );
}

if( $bBuildBjam == -1 ) {
  # The user did not tell us whether to build bjam or not, build it by default.
  $bBuildBjam = 1;
}

if( $bRebuildAll == -1 ) {
  # The user did not tell us whether to do a complete rebuild, do a smart build by default.
  $bRebuildAll = 0;
}

if( $bBuildLibs == -1 ) {
  # The user did not tell us whether to build boost libraries or not, enable building them by default. 
  # N.B. we check the jamroot later on to see whether the boost libraries are needed at all.
  $bBuildLibs = 1;
}


# add package install hints
add_pkg_name_info( \%missing_pkg_info );


if( $OSNAME eq 'linux' ) {
  if( $boost_toolset ne 'gcc' ) {
    fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not yet supported.\nPlease use --toolset=gcc or remove the command line option.\n" );
  }
}

if( $bjam_bin_dir eq '' ) {
  $bjam_bin_dir = get_default_bjam_dir( $tool_dir, $bIsPlatform64Bit );
}

# auto detection of boost root and version
$Jamroot = File::Spec->rel2abs( File::Spec->catfile( $boost_build_dir, '..', 'Jamroot' ) );

if( ! -f $Jamroot ) {
  fatal_error( __LINE__, "Jamroot \"$Jamroot\" not found\n" );
}

add_build_log_entry( \@build_log_summary, "boost_build version:", "$boost_build_version\n" );

if( ! get_boost_root_from_jamroot( \$boost_root, $Jamroot ) ) {
  
  print "\n";
  print "The Jamroot file does not define \"BOOST_SDK_ROOT\" => no need to build any boost libraries.\n";
  print "\n";
  
  # the boost libraries are not needed, disable building the libraries.
  $boost_root_fnd = 0;
  $bBuildLibs = 0;
}
else {
  #my $boost_lib_suffix;
  
  $boost_root_fnd = 1;
  
  $boost_version = get_boost_version( $boost_root );
  if( $boost_version =~ /^([0-9]+)[.]([0-9]+)[.]([0-9])/) {
    push @boost_version_elems, $1;
    push @boost_version_elems, $2;
    push @boost_version_elems, $3;
  }
  else {
    die "boost_install.pl: internal error: boost version ID \"$boost_version\" not understood.";
  }
}

if( $bBuildLibs != 0 ) {
 
  print "\n";
  print "bjam_bin_dir:     $bjam_bin_dir\n";
  print "boost_root:       $boost_root\n";
  print "boost version:    $boost_version\n";
  print "\n";
  
  # process library build options (32 bit, 64 bit)
  process_lib_build_options( $boost_toolset, $boost_version, $system_info, $bIsPlatform64Bit, \$bBuildLib32, \$bBuildLib64, \@boost_lib_exclude_list );
  
  
  # handle the optional include and exclude library lists
  if( $boost_lib_exclude_list_args ne '' ) {
    push @boost_lib_exclude_list, ( split /,/, $boost_lib_exclude_list_args );
  }
  
  if( $boost_lib_include_list_args ne '' ) {
    push @boost_lib_include_list, ( split /,/, $boost_lib_include_list_args );
  }
  
  check_include_exclude_lists( \@boost_lib_include_list, \@boost_lib_exclude_list );
  
  #print "include list: @boost_lib_include_list\n";
  #print "exclude list: @boost_lib_exclude_list\n";
  #exit 0;
  
   
  if( $bStaticLibs ) {
    $bStaticLibs32 = 1 if( $bBuildLib32 );
    $bStaticLibs64 = 1 if( $bBuildLib64 );     
  }
 
  if( $bSharedLibs ) {
    $bSharedLibs32 = 1 if( $bBuildLib32 );
    $bSharedLibs64 = 1 if( $bBuildLib64 );     
  } 
 
  compose_patch_list( $boost_toolset, $boost_version,  \@boost_patch_file_list );
  if( scalar( @boost_patch_file_list  ) > 0  ) {
     install_boost_patches( $boost_patch_dir, $boost_root, $boost_version, \@boost_patch_file_list );
  }
  
  #check boost_version and compare it with our version
  compare_boost_and_boost_build_version( $boost_version, $boost_build_version );
  
  if( ( $OSNAME eq 'linux' ) || ( $OSNAME eq 'solaris' ) ) {
    #check for developer packages / header files
    check_for_dev_packages( \%missing_pkg_info, $system_info, $bIsPlatform64Bit, $bBuildLib32, $bBuildLib64 );
  }
  
  if( $OSNAME eq 'linux' ) {
    if(   ( ( $bIsPlatform64Bit == 0 ) && ( ( $bStaticLibs64 != 0 ) || ( $bSharedLibs64 ) ) )
       || ( ( $bIsPlatform64Bit != 0 ) && ( ( $bSharedLibs32 != 0 ) || ( $bSharedLibs32 ) ) ) ) {
      # LINUX: cross compiling requires boost 1.40.x or higher 
      if(      ( $boost_version_elems[ 0 ] < 1 )
          || ( ( $boost_version_elems[ 0 ] == 1 ) && ( $boost_version_elems[ 1 ] < 40 ) ) ) {
           
        # disabling cross compiling on pre 1.40.0 version
        if( $bIsPlatform64Bit != 0 ) {
          $bStaticLibs32 = 0;
          $bSharedLibs32 = 0;
        }
        else {
          $bStaticLibs64 = 0;
          $bSharedLibs64 = 0;
        }
      }  
    }
  }
}

check_boost_build_root( File::Spec->catfile( $boost_build_dir, '..', 'boost-build.jam' ), \@todo_list );
if( $boost_root_fnd ) {
  check_jamroot( $Jamroot, \@todo_list );
}

$boost_build_srcdir = File::Spec->catfile( $boost_build_dir, 'boost-build' );

if( ! -d $boost_build_srcdir ) {
  fatal_error( __LINE__, "boost build source tree \"$boost_build_srcdir\" not found\n" );
}

$bjam_bin_dir = File::Spec->rel2abs( $bjam_bin_dir );
if( ! -d $bjam_bin_dir ) {
  mkpath( $bjam_bin_dir, 1, 0755 );
}

if( $OSNAME eq 'MSWin32') {
  $bjam_prog = File::Spec->catfile( $bjam_bin_dir, 'bjam.exe' );
}
else {
  $bjam_prog = File::Spec->catfile( $bjam_bin_dir, 'bjam' );
}

if( $bBuildBjam ) {
  build_bjam( \$bjam_prog_built, $boost_build_dir, $boost_toolset, $bIsPlatform64Bit );
  
  if( ! -f $bjam_prog_built ) {
    fatal_error( __LINE__, "bjam executable \"$bjam_prog_built\" not found." );
  }
  
  if( check_bjam_version( $bjam_prog, $bjam_prog_built ) ) {
    $b_bjam_prog_updated = 1;
    
    print "\n";
    print "Installing bjam executable to $bjam_bin_dir\n";
    print "\n";
    
    # copy does not preserve the unix mode!
    copy( $bjam_prog_built, $bjam_bin_dir );
    
    if( $OSNAME ne 'MSWin32') {
      $bjam_prog_quoted = quote_executable_path( $bjam_prog );
      system( "chmod 0755 $bjam_prog_quoted" );
    }
  } 
  unlink( $bjam_prog_built );
  rmdir( dirname( $bjam_prog_built ) );
}

if( ! -f "$bjam_prog" ) {
  fatal_error( __LINE__, "bjam executable \"$bjam_prog\" not found; consider building it from source (--build-bjam)\n" );
}
else {
  # compose summary log entry 
  my $bjam_version = get_bjam_version( $bjam_prog );
  
  add_build_log_entry( \@build_log_summary, "bjam:", "$bjam_prog\n" ); 
  if( $b_bjam_prog_updated ) {
    add_build_log_entry( \@build_log_summary, "bjam version:", "$bjam_version, updated\n" );
  } 
  else {
    add_build_log_entry( \@build_log_summary, "bjam version:", "$bjam_version\n" );
  }
}

if( $boost_root_fnd ) {
  add_build_log_entry( \@build_log_summary, "boost_root:", "$boost_root\n" );
  add_build_log_entry( \@build_log_summary, "boost version:", "$boost_version\n" );
}
else {
  add_build_log_entry( \@build_log_summary, "boost_root:", "<none>\n" );
}  
  


if( $bBuildLibs != 0 ) {
  # analyzing phase
  $boost_toolset_dir_suffix = get_toolset_dir_suffix( $tool_dir, $boost_toolset );
  $boost_bin_dir = File::Spec->catfile( $boost_root, 'bin' );
  $boost_lib_dir = File::Spec->catfile( $boost_root, 'lib' );

  
  #auto_lib_migration( $boost_bin_dir, $boost_lib_dir, $boost_toolset_dir_suffix, $bIsPlatform64Bit );

  if( ! $bRebuildAll ) {
    
    inspect_boost_lib_dir ( $boost_version, 
                            $boost_toolset, 
                            File::Spec->catfile( $boost_lib_dir, $boost_toolset_dir_suffix ), 
                            \@boost_lib_exclude_list,
                            \@boost_lib_include_list,
                            \$bStaticLibs32, 
                            \$bStaticLibs64, 
                            \$bSharedLibs32, 
                            \$bSharedLibs64 );
    
    if(   ( $bStaticLibs32 == 0 ) 
       && ( $bSharedLibs32 == 0 )
       && ( $bStaticLibs64 == 0 )
       && ( $bSharedLibs64 == 0 ) ) {
       $bBuildLibs = 0;   
    }
    
    #print "build static libs, 32 bit: $bStaticLibs32\n";
    #print "build shared libs, 32 bit: $bSharedLibs32\n";
    #print "build static libs, 64 bit: $bStaticLibs64\n";
    #print "build shared libs, 64 bit: $bSharedLibs64\n";
    #exit 0;
    
    $bBuildLib32 = 0 if ( ( $bStaticLibs32 == 0 ) &&  ( $bSharedLibs32 == 0 ) );
    $bBuildLib64 = 0 if ( ( $bStaticLibs64 == 0 ) &&  ( $bSharedLibs64 == 0 ) );   
    
    if( ! $bBuildLibs ) {
      print "\n* All libraries seem to be present, you may force a complete rebuild using the command line option \"--rebuild\"\n\n";
    }
  }
}


# build phase
if( $bBuildLibs != 0 ) {
  
  my @bjam_libdir_options;
  my @bjam_layout_options;
  
  my $bjam_layout;
  my $bjam_libdir;
  my $bjam_variant_option;
  my $bjam_variant_option_key;
  my $bjam_cmdline;
  my $system_cmdline;
  my $current_dir;
  my $boost_lib;
  
  check_and_set_bjam_jobs( $system_info, \$bjam_jobs );
  
  # build bjam command line
  if( $bQuitOnError != 0 ) {
    $bjam_options = '-q';
  }
  
  if( $bjam_jobs > 1 ) {
    $bjam_options .= " -j$bjam_jobs";
  }
  
  if( ( $bStaticLibs32 != 0 ) || ( $bStaticLibs64 != 0 ) ) {
    $boost_link_options .= 'static';
  }
  
  if( ( $bSharedLibs32 != 0 ) || ( $bSharedLibs64 != 0 ) ) {
    if( $boost_link_options eq 'link=' ) {
      $boost_link_options .= 'shared';
    }
    else {
      $boost_link_options .= ',shared';
    }
  }
  
  $bjam_prog_quoted = quote_executable_path( $bjam_prog );
  $bjam_cmdline = "$bjam_prog_quoted $bjam_options --layout=versioned --prefix=tmp/dist"; 
  
  foreach $boost_lib ( @boost_lib_exclude_list ) {
    $bjam_cmdline .= ' ' . "--without-$boost_lib";
  }

  foreach $boost_lib ( @boost_lib_include_list ) {
    $bjam_cmdline .= ' ' . "--with-$boost_lib";
  }
  
  
  if( $boost_toolset ne '' ) {
    $bjam_cmdline .= " toolset=$boost_toolset";
  }
    
  if( $OSNAME eq 'MSWin32' ) {
    my $libdir_option;
    my $libdir;
    
    $libdir_option = "--libdir=lib/$boost_toolset_dir_suffix";
    
    if( $bBuildLib32 != 0 ) {
      $libdir = $libdir_option . '/x86 variant=release,debug';
      push @bjam_libdir_options, $libdir;
    }

    if( $bBuildLib64 != 0 ) {
      $libdir = $libdir_option . '/x86_64 address-model=64 variant=release,debug';
      push @bjam_libdir_options, $libdir;
    }
  }
  elsif( $OSNAME eq 'linux' ) {
    my $libdir;
    
    #compile all libraries with -fPIC to support shared libraries linked to static boost libraries
    #$bjam_cmdline .= ' cflags="-v -fPIC"';
    $bjam_cmdline .= ' cflags=-fPIC';
    
    if( $bBuildLib32 != 0 ) {
      $libdir = "--libdir=lib/$boost_toolset_dir_suffix/x86";
      
      if( $bIsPlatform64Bit ) {
        # cross compiling mode
        $libdir .= ' address-model=32';
      }
      
      $libdir .= ' variant=release,debug';
      
      push @bjam_libdir_options, $libdir;
    }

    if( $bBuildLib64 != 0 ) {
      $libdir = "--libdir=lib/$boost_toolset_dir_suffix/x86_64";
      
      if( ! $bIsPlatform64Bit ) {
        #cross compiling mode
        $libdir .= ' address-model=64';
      }
      
      $libdir .= ' variant=release,debug';
            
      push @bjam_libdir_options, $libdir;
    }
  }
  elsif( $OSNAME eq 'darwin' ) {
    my $libdir;
    
    #compile all libraries with -fPIC to support shared libraries linked to static boost libraries
    $bjam_cmdline .= ' cflags=-fPIC';
    
    #check architecture
    my $uname_m;
    $uname_m = `uname -m`;
    chomp $uname_m;
    if( ($uname_m eq 'i386') || ($uname_m eq 'x86_64') ) {
      # always build 32/64bit universal binaries on intel macs.
      # don't build with ppc support (architecture=combined is the default)
      # since ppc is discontinued by Apple and TR1 math lib won't compile
      $bjam_cmdline .= ' architecture=x86 address-model=32_64';
    }
    else {
      # could be arm for iphone
      fatal_error( __LINE__, "architecture \"$uname_m\" currently not supported for os darwin\n" );
    }
    
    $libdir = "--libdir=lib/$boost_toolset_dir_suffix/x86 variant=release,debug";
    push @bjam_libdir_options, $libdir;
  }
  elsif( $OSNAME eq 'solaris' ) {
    my $libdir;

    #compile all libraries with -PIC to support shared libraries linked to static boost libraries
 
    if( $boost_toolset eq 'sun' ) {
      $bjam_cmdline .= ' stdlib=sun-stlport cflags=-PIC';
    }
    else {
      $bjam_cmdline .= ' cflags=-PIC';
    }
    
    if( $bBuildLib64 != 0 ) {
      # On 64 bit solaris, just build the 64 bit libraries and not the 32 bit variants.
      $libdir = "--libdir=lib/$boost_toolset_dir_suffix/x86_64 variant=release,debug address-model=64";
      push @bjam_libdir_options, $libdir;
    }
    
    if( $bBuildLib32 != 0 ) {
      $libdir = "--libdir=lib/$boost_toolset_dir_suffix/x86 variant=release,debug";
      push @bjam_libdir_options, $libdir;
    }
  }
  
  push @build_log_summary, "bjam command lines:\n\n";
  
  foreach $bjam_libdir (@bjam_libdir_options) {
    
    $system_cmdline = $bjam_cmdline;
    $system_cmdline .= " $bjam_libdir $boost_link_options install";

    #remember the current working directory
    $current_dir = File::Spec->rel2abs( File::Spec->curdir() );
    
    print "\n";
    print "========================================================\n";
    print "Start building the boost libraries in $boost_root using\n";
    print "\n";
    print "$system_cmdline\n\n" ;
    print "========================================================\n";
    print "\n";
    
    push @build_log_summary, "$system_cmdline\n\n";
    
    chdir( $boost_root );
    
    if( $bDryRun == 0 ) {
      system( $system_cmdline );

      if( ( $bQuitOnError != 0 ) && ( $? != 0 ) ) {
        fatal_error( __LINE__, "some boost libraries failed to compile or install, please fix it\n" );
      }
    }
    
    # restore the current working directory
    chdir( $current_dir );
  }
  
  if( ( $OSNAME eq 'MSWin32' ) && ( $bSharedLibs != 0 ) ) {
    my @boost_bin_dir_list;
    my @boost_lib_dir_list;
    
    if( $bBuildLib32 != 0 ) {
      # copy all DLLs to the bin /win32 directory 
      $boost_lib_dir = File::Spec->catfile( $boost_root, 'lib', $boost_toolset_dir_suffix, 'x86' );
      push @boost_lib_dir_list, $boost_lib_dir;
      
      #$boost_lib_dir = File::Spec->catfile( $boost_root, 'lib', $boost_toolset_dir_suffix, 'x86', 'release' );
      #push @boost_lib_dir_list, $boost_lib_dir;
      
      #$boost_lib_dir = File::Spec->catfile( $boost_root, 'lib', $boost_toolset_dir_suffix, 'x86', 'debug' );
      #push @boost_lib_dir_list, $boost_lib_dir;
    
      $boost_bin_dir = File::Spec->catfile( $boost_root, 'bin', $boost_toolset_dir_suffix, 'x86' );    
      push @boost_bin_dir_list, $boost_bin_dir;
      
      #$boost_bin_dir = File::Spec->catfile( $boost_root, 'bin', $boost_toolset_dir_suffix, 'x86', 'release' );    
      #push @boost_bin_dir_list, $boost_bin_dir;
 
      #$boost_bin_dir = File::Spec->catfile( $boost_root, 'bin', $boost_toolset_dir_suffix, 'x86', 'debug' );    
      #push @boost_bin_dir_list, $boost_bin_dir;
 
      foreach $boost_bin_dir (@boost_bin_dir_list) {
        $boost_lib_dir = shift( @boost_lib_dir_list );
        
        if( $bDryRun == 0 ) {
          if( -d $boost_lib_dir ) {
            &copy_boost_dlls( $boost_bin_dir, $boost_lib_dir );
          }
        }
        else {
          print "copy_boost_dlls: $boost_lib_dir => $boost_bin_dir\n";
        }
      }
    }
    
    if( $bBuildLib64 != 0 ) {
      # copy all DLLs to the bin /x64 directory
      
      @boost_bin_dir_list = ();
      @boost_lib_dir_list = ();
      
      $boost_lib_dir = File::Spec->catfile( $boost_root, 'lib', $boost_toolset_dir_suffix, 'x86_64' );
      push @boost_lib_dir_list, $boost_lib_dir;
      
      #$boost_lib_dir = File::Spec->catfile( $boost_root, 'lib', $boost_toolset_dir_suffix, 'x86_64', 'release' );
      #push @boost_lib_dir_list, $boost_lib_dir;
      
      #$boost_lib_dir = File::Spec->catfile( $boost_root, 'lib', $boost_toolset_dir_suffix, 'x86_64', 'debug' );
      #push @boost_lib_dir_list, $boost_lib_dir;
    
      $boost_bin_dir = File::Spec->catfile( $boost_root, 'bin', $boost_toolset_dir_suffix, 'x86_64' );    
      push @boost_bin_dir_list, $boost_bin_dir;
      
      #$boost_bin_dir = File::Spec->catfile( $boost_root, 'bin', $boost_toolset_dir_suffix, 'x86_64', 'release' );    
      #push @boost_bin_dir_list, $boost_bin_dir;
 
      #$boost_bin_dir = File::Spec->catfile( $boost_root, 'bin', $boost_toolset_dir_suffix, 'x86_64', 'debug' );    
      #push @boost_bin_dir_list, $boost_bin_dir;
 
      foreach $boost_bin_dir (@boost_bin_dir_list) {
        $boost_lib_dir = shift( @boost_lib_dir_list );
        
        if( $bDryRun == 0 ) {
          if( -d $boost_lib_dir ) {
            &copy_boost_dlls( $boost_bin_dir, $boost_lib_dir );
          }
        }
        else {
          print "copy_boost_dlls: $boost_lib_dir => $boost_bin_dir\n";
        }
      }
    }
  }
  
  if( $bKeepTmpFiles != 1 ) {
    #get rid of all temporary files generated while compiling the boost libraries
    my $boost_build_dir = File::Spec->catfile( $boost_root, 'bin.v2' );
    my $boost_tmp_dir = File::Spec->catfile( $boost_root, 'tmp' );
  
    print "\n";
    if( -d $boost_build_dir ) {
      print "Removing boost build tree: $boost_build_dir ...\n";
      if( $bDryRun == 0 ) {
        rmtree( $boost_build_dir, { verbose => 0 } );
      }
    }
    
    if( -d $boost_tmp_dir ) {
      print "Removing $boost_tmp_dir ...\n";
      if( $bDryRun == 0 ) {
        rmtree( $boost_tmp_dir, { verbose => 0 } );
      }
    }
    print "\n";
  }
}

print "\n";

if( ( $bBuildLibs != 0 ) || ( $bBuildBjam != 0 ) ) {
  $b_bjam_bin_dir_on_path = check_bjam_path( $bjam_bin_dir, \@todo_list );
  
  if( ( ! $b_bjam_bin_dir_on_path ) && ( $OSNAME eq 'linux' ) || ( $OSNAME eq 'solaris' ) ) {
    
    &create_default_bjamrc_file( $bjam_bin_dir );
    
    print "\n";
    print "Saving the current bjam path to ~/.bjamrc.\n";
    print "You may need to update your shell startup file ~/.profile with lines like\n";
    print "\n";
    print "#---------------------------------------------\n";
    print "RC_FILES=\".bjamrc .cudadevrc\"\n";
    print "\n";
    print "for f in \$RC_FILES\n";
    print "do\n";
    print "  if [ -f \$HOME/\$f ]; then\n";
    print "    . \$HOME/\$f\n";
    print "  fi\n";
    print "done\n";
    print "#---------------------------------------------\n";
    print "\n";
  }
}

print "\n";
print "                      SUMMARY                           \n";
print "                      =======                           \n";
print "\n";

foreach $build_log_msg (@build_log_summary) {
  print "$build_log_msg"; 
}

if( scalar( @todo_list ) > 0 ) {
  print "\nTODO:\n";
  foreach $build_log_msg (@todo_list) {
    print " - $build_log_msg\n"; 
  }
}


print "\n";
print "boost_install.pl finished successfully.\n";
print "\n";

sub get_system_info {
  my ( $tool_dir ) = @_;
  my $cmd;
  my $system_info;
  my @system_info_components;
  
  if( $OSNAME eq 'MSWin32' ) {
    $cmd = File::Spec->catfile( $tool_dir, 'windows', 'get_system_info_win.pl' );
  }
  else {
    $cmd = File::Spec->catfile( $tool_dir, 'unix', 'get_system_info_unix.pl' );
  }

  $system_info = invoke_external_cmd( $cmd );
  chomp( $system_info );
  
  # check $system_info (number of required attributes)
  @system_info_components = split /:/, $system_info; 
  if( scalar( @system_info_components ) < 5 ) {
    fatal_error( __LINE__, "get_system_info() has returned an unexpected string: \"$system_info\"\n" );
  }

  return $system_info;
}

sub get_default_toolset {
  my ( $tool_dir ) = @_;
  my $toolset;
  
  if( $OSNAME eq 'MSWin32' ) {
    my $toolset_version = find_latest_msvc_version( $tool_dir );
 
    $toolset = 'msvc-' . $toolset_version;
  }
  elsif( $OSNAME eq 'linux' ) {
    $toolset = 'gcc';
  } 
  elsif( $OSNAME eq 'darwin' ) {
    $toolset = 'darwin';
  }
  elsif( $OSNAME eq 'solaris' ) {
    #g++ and gcc are usually not installed on solaris systems.
    $toolset = 'sun';
  }
  else {
    die "unsupported platform";    
  }
  
  return $toolset;   
} 

sub process_lib_build_options {
  my ( $boost_toolset, $boost_version, $system_info, $bIsPlatform64Bit, $r_bBuildLib32, $r_bBuildLib64, $rl_boost_lib_exclude_list ) = @_;
  my @system_info_components =  split /:/, $system_info;
  my $b_cross_compile_32_bit_default = 0;
  
  if( $bIsPlatform64Bit ) {
    if( ( $OSNAME eq 'MSWin32' ) || ( $OSNAME eq 'solaris' ) ) {
      $b_cross_compile_32_bit_default = 1;
    }
    elsif( $OSNAME eq 'linux' ) {
      # linux/64 bit: 32 bit development is not working on all LINUX distros
      
      if( $system_info_components[ 3 ] eq 'ubuntu' ) {
        # I tested fedora 11 and opensuse 11.1 and neither was able to compile the 32 bit boost libraries.
        $b_cross_compile_32_bit_default = 1;
      }
    }
  }
  
  print "\n";
  
  #exlude python by default on all platforms
  push @$rl_boost_lib_exclude_list, 'python';
  
  if( $OSNAME eq 'MSWin32' ) {
    
    if( $$r_bBuildLib32 == -1 ) {
      $$r_bBuildLib32 = 1;
    }
    
    if( $bIsPlatform64Bit == 0 ) {
  
      print "32 bit windows detected; ";
      
      if( $$r_bBuildLib64 == -1 ) {
        $$r_bBuildLib64 = 0;
      }
      
      if( $$r_bBuildLib64 == 0 ) {
        print "disabling 64 bit cross-compiling\n";     
      }
      else {
        print "enabling 64 bit cross-compiling\n";      
      }
    }
    else {
      print "64 bit windows detected; ";
      
      if( $$r_bBuildLib64 == -1 ) {
        $$r_bBuildLib64 = 1;
      }
      
      if( $$r_bBuildLib32 == 0 ) {
        print "disabling 32 bit cross-compiling\n";
      }
      else {
        print "enabling 32 bit cross-compiling\n";
      }    
    }
    
    if( $$r_bBuildLib64 != 0 ) {
      print "Please make sure you've installed the amd64 compiler options,\n";
      print "otherwise the 64 bit libraries won't compile.\n";  
    }
  }
  elsif( ( $OSNAME eq 'linux' ) ||  ( $OSNAME eq 'solaris' ) )  {
  
    if( $bIsPlatform64Bit == 0 )
    {
      if( $$r_bBuildLib32 == -1 ) {
        $$r_bBuildLib32 = 1;
      }
    }
    else { 
      # linux/64 bit, check if 32 bit development is possible by default
      if( $b_cross_compile_32_bit_default != 0 ) {
        if( $$r_bBuildLib32 == -1 ) {
          $$r_bBuildLib32 = 1;
        }
      }
      else {
        if( $$r_bBuildLib32 == -1 ) {
          $$r_bBuildLib32 = 0;
        }        
      }      
    }
    
    if( $bIsPlatform64Bit == 0 ) {
      print "32 bit $OSNAME detected; ";
      
      if( $$r_bBuildLib64 == -1 ) {
        # disabling cross compiling to amd64 by default
        $$r_bBuildLib64 = 0;
      }
      
      if( $$r_bBuildLib64 != 0 ) {
        print "enabling 64 bit cross-compiling\n";   
      }
      else {
        print "disabling 64 bit cross-compiling\n";   
      } 
    }
    else {
      print "64 bit $OSNAME detected; ";
      
      if( $$r_bBuildLib64 == -1 ) {
        $$r_bBuildLib64 = 1;
      }
  
      if( $$r_bBuildLib32 != 0 ) {
        print "enabling 32 bit cross-compiling\n";      
      }
      else {
        print "disabling 32 bit cross-compiling\n";
      } 
    }
    
    if( ( $OSNAME eq 'solaris' ) && ( $boost_toolset eq 'sun' ) ) {
      # the following libraries don't compile on solaris yet
      if( $boost_version eq '1.40.0' ) {
        push @$rl_boost_lib_exclude_list, 'graph', 'wave';
      }
      elsif( $boost_version eq '1.41.0' ) {
        push @$rl_boost_lib_exclude_list, 'wave';
      }
    }
    
    if( ( $system_info_components[ 3 ] eq 'fedora' ) && ( $system_info_components[ 4 ] eq 'leonidas' ) ) {
      if( $boost_version eq '1.41.0' ) {
        push @$rl_boost_lib_exclude_list, 'wave';
      }
    }
  } 
  elsif( $OSNAME eq 'darwin' ) {
    fatal_error( __LINE__, "please add darwin support to function process_lib_build_options()\n" ); 
  }
  else {
    die "unsupported platform detected: $OSNAME" ;
  }
  print "\n";  
}

sub compose_patch_list {
   my ( $boost_toolset, $boost_version,  $rl_boost_patch_file_list ) = @_;
   my $file;
   
   if( $OSNAME eq 'solaris' ) {
      if( ( $boost_toolset eq 'sun' ) && ( $boost_version eq '1.40.0' ) ) {
         $file = File::Spec->catfile( 'boost', 'thread', 'detail', 'thread.hpp' );
         push @$rl_boost_patch_file_list, $file;
      }
   }
} 

sub install_boost_patches{
   my ( $boost_patch_dir, $boost_root, $boost_version, $rl_boost_patch_file_list ) = @_;
   my $file;
   my $src_file;
   my $dst_file;
   
   print "\n";
   
   foreach $file ( @$rl_boost_patch_file_list ) {
      $src_file =  File::Spec->catfile( $boost_patch_dir, $boost_version, $file );
      $dst_file = File::Spec->catfile( $boost_root, $file );
      
      print "Installing patch $src_file ...\n";
      copy( $src_file, $dst_file ) or die "Copy( $src_file, $dst_file ) failed: $!";       
   }
   
    print "\n";
}

sub add_build_log_entry {
  my ( $rl_build_log, $entry, $entry_msg ) = @_;
  my $line;
  
  $line = sprintf( '%-25s%s', $entry, $entry_msg );
  
  push @$rl_build_log, $line;
}


sub compare_boost_and_boost_build_version {
  my ( $boost_version, $boost_build_version ) = @_;
  my $result = version_compare( $boost_build_version, $boost_version );
  my $msg;
  my $boost_build_version_normalized = $boost_build_version;
  
  if( $boost_build_version_normalized =~ /^([0-9.]+)[-][0-9]+$/ ) {
    $boost_build_version_normalized = $1;
  }
  
  $result = version_compare( $boost_build_version_normalized, $boost_version );
  
  
  # the boost build version must be equal or higher than the boost version
  if( $result != 0 ) {
    $msg = "boost build version \"$boost_build_version\" and boost version \"$boost_version\" are incompatible.\n";
    $msg .= "In general, the boost build version and boost version must match on the first 3 version number components.\n";
    $msg .= "Please check in your Jamroot \"BOOST_SDK_ROOT\" and the svn external reference to BoostBuild.\n"; 
    fatal_error( __LINE__, $msg );
  }
}

sub quote_executable_path {
  my ( $prog ) = @_;
  
  if( $prog =~ /^"/ ) {
    # already quoted
    return $prog;
  }
  
  if( $prog =~ /[ ]+/ ) {
    $prog = "\"$prog\"";
  }
  
  return $prog;
}

sub check_and_set_bjam_jobs {
  my ( $system_info, $r_bjam_jobs ) = @_;
  my @system_info_components;
  
  if( $$r_bjam_jobs > 0 ) {
    # accept the value provided on the command line
    return;
  }
  
  @system_info_components = split /:/, $system_info;
   
  # initialize the number of parallel bjam jobs to be equal to the number of processors
  $$r_bjam_jobs = $system_info_components[ 2 ];
  
  if( $$r_bjam_jobs > 0 ) {
    print "\n";
    print "Found $$r_bjam_jobs processors on your system, this will be the default for\n";
    print "the number of parallel bjam jobs (use --jobs=n to override).\n";
    print "\n";
  }
  else {
    #last resort, should be fixed sometime.
    $$r_bjam_jobs = 1;
  }
}

sub auto_lib_migration {
  my ( $boost_bin_dir, $boost_lib_dir, $boost_toolset_dir_suffix, $bIsPlatform64Bit ) = @_;
  my $boost_lib_dir_32 = File::Spec->catdir( $boost_lib_dir, $boost_toolset_dir_suffix, 'x86' );
  my $boost_lib_dir_64 = File::Spec->catdir( $boost_lib_dir, $boost_toolset_dir_suffix, 'x86_64' );
  my $dir;
  my $dir2;
  
  if( $OSNAME eq 'linux' ) {
    if( ( ! -d "$boost_lib_dir_32" ) && ( ! -d "$boost_lib_dir_64" ) ) {
      # assume this is a native build
      if( $bIsPlatform64Bit ) {
        $dir = File::Spec->catdir( $boost_lib_dir, $boost_toolset_dir_suffix );
        
        if( -d "$dir" ) {
          print "* Migrating $dir to $dir/x86_64\n";
          
          $dir2 = File::Spec->catdir( $boost_lib_dir, 'x86_64' );
          
          move( $dir, $dir2 );
          
          #create toolset directory
          mkpath( $dir, {verbose => 1} );
          
          move( $dir2, $boost_lib_dir_64 );
        }
      }
      else {
        $dir = File::Spec->catdir( $boost_lib_dir, $boost_toolset_dir_suffix );
        
        if( -d "$dir" ) {
          print "* Migrating $dir to $dir/x86\n";
          
          $dir2 = File::Spec->catdir( $boost_lib_dir, 'x86' );
 
          move( $dir, $dir2 );
          
          #create toolset directory
          mkpath( $dir, {verbose => 1} );
          
          move( $dir2, $boost_lib_dir_32 );    
        }    
      }
    }
  }
  elsif( $OSNAME eq 'MSWin32' ) {
    if( ( ! -d "$boost_lib_dir_32" ) && ( ! -d "$boost_lib_dir_64" ) ) {
    
      $dir = File::Spec->catdir( $boost_lib_dir, $boost_toolset_dir_suffix, 'win32' );
      if( -d "$dir" ) {
        print "* Migrating $dir to $dir/x86\n";
        
        $dir2 = File::Spec->catdir( $boost_lib_dir, $boost_toolset_dir_suffix, 'x86' );
        move( $dir, $dir2 );
      }
      
      $dir = File::Spec->catdir( $boost_lib_dir, $boost_toolset_dir_suffix, 'x64' );
      if( -d "$dir" ) {
        print "* Migrating $dir to $dir/x86_64\n";
        
        $dir2 = File::Spec->catdir( $boost_lib_dir, $boost_toolset_dir_suffix, 'x86_64' );
        move( $dir, $dir2 );
      }
      
      $dir = File::Spec->catdir( $boost_bin_dir, $boost_toolset_dir_suffix, 'win32' );
      if( -d "$dir" ) {
        print "* Migrating $dir to $dir/x86\n";
        $dir2 = File::Spec->catdir( $boost_bin_dir, $boost_toolset_dir_suffix, 'x86' );
        move( $dir, $dir2 );
      }
      
      $dir = File::Spec->catdir( $boost_bin_dir, $boost_toolset_dir_suffix, 'x64' );
      if( -d "$dir" ) {
        print "* Migrating $dir to $dir/x86_64\n";
        $dir2 = File::Spec->catdir( $boost_bin_dir, $boost_toolset_dir_suffix, 'x86_64' );
        move( $dir, $dir2 );
      }
    }
  }
}

sub get_boost_core_lib_list {
  my ( $rl_boost_lib_exclude_list, $rl_boost_lib_include_list, $rl_boost_core_lib_list ) = @_;
  my $boost_lib;
  my %boost_core_lib_default_map = (
    'date_time' => '1',
    'filesystem' => '1',
    'graph' => '1',
    'iostreams' => '1',
    'math_c99f' => '1',
    'math_c99'  => '1',
    'math_c99l'  => '1',
    'math_tr1f'  => '1',
    'math_tr1'  => '1',
    'math_tr1l'  => '1',
    'prg_exec_monitor'  => '1',
    'program_options' => '1',
    'regex' => '1',
    'serialization' => '1',
    'signals' => '1',
    'system' => '1',
    'thread'  => '1',
    'unit_test_framework'  => '1',
    'wave'  => '1',
    'wserialization' => '1'
  );
  
  # process the optional include list
  foreach $boost_lib ( @$rl_boost_lib_include_list ) {
    if( ! exists( $boost_core_lib_default_map{ $boost_lib } ) ) {
      $boost_core_lib_default_map{ $boost_lib } = '1';
    }
  }
  
  # process the optional exclude list
  foreach $boost_lib ( @$rl_boost_lib_exclude_list ) {
    if( exists( $boost_core_lib_default_map{ $boost_lib } ) ) {
      delete $boost_core_lib_default_map{ $boost_lib };
    }
  }  
  
  push @$rl_boost_core_lib_list, sort( keys( %boost_core_lib_default_map ) );
}


sub inspect_boost_lib_dir {
  my ( $boost_version, $boost_toolset, $boost_lib_dir, $rl_boost_lib_exclude_list, $rl_boost_lib_include_list, $r_bStaticLibs32, $r_bStaticLibs64, $r_bSharedLibs32, $r_bSharedLibs64 ) = @_;
  my $boost_lib_release_suffix =  get_boost_lib_suffix ( $boost_version, $boost_toolset, 0 );
  my $boost_lib_debug_suffix = get_boost_lib_suffix ( $boost_version, $boost_toolset, 1 );
  my @boost_core_lib_list;
  my %boost_lib_map;
  my $dir;
  my $lib;
  my $boost_lib;
  my $boost_lib_release_static;
  my $boost_lib_release_shared;
  my $boost_lib_debug_static;
  my $boost_lib_debug_shared;
  my $boost_libs_ok;
  
  my @boost_lib_shared_list;
  my @boost_lib_static_list;
  
  
  get_boost_core_lib_list( $rl_boost_lib_exclude_list, $rl_boost_lib_include_list, \@boost_core_lib_list );
  
  #print "==========================================================\n";
  #print "@boost_core_lib_list\n";
  #print "==========================================================\n";
  #exit( 0 );
  
  if( $OSNAME eq 'MSWin32' ) {
    for $lib ( @boost_core_lib_list ) {
      $boost_lib_release_static = 'libboost_' . $lib . $boost_lib_release_suffix . '.lib';
      $boost_lib_release_shared = 'boost_' . $lib . $boost_lib_release_suffix . '.lib';
      
      push @boost_lib_static_list, $boost_lib_release_static;
      push @boost_lib_shared_list, $boost_lib_release_shared;  
      
      $boost_lib_debug_static = 'libboost_' . $lib . $boost_lib_debug_suffix . '.lib';
      $boost_lib_debug_shared = 'boost_' . $lib . $boost_lib_debug_suffix . '.lib';
      
      push @boost_lib_static_list, $boost_lib_debug_static;
      push @boost_lib_shared_list, $boost_lib_debug_shared;    
    } 
  }
  elsif( ( $OSNAME eq 'linux' ) || ( $OSNAME eq 'solaris' ) ) {
    
    for $lib ( @boost_core_lib_list ) {
      $boost_lib = 'libboost_' . $lib . $boost_lib_release_suffix;
  
      $boost_lib_release_static = $boost_lib . '.a';
      $boost_lib_release_shared = $boost_lib . '.so.' . $boost_version;
      
      push @boost_lib_static_list, $boost_lib_release_static;
      push @boost_lib_shared_list, $boost_lib_release_shared;  
      
      $boost_lib = 'libboost_' . $lib . $boost_lib_debug_suffix;
      $boost_lib_debug_static = $boost_lib . '.a';
      $boost_lib_debug_shared = $boost_lib . '.so.' . $boost_version;
      
      push @boost_lib_static_list, $boost_lib_debug_static;
      push @boost_lib_shared_list, $boost_lib_debug_shared;   
    }     
  }
  else {
    die "inspect_boost_lib_dir2: unsupported platform encountered."
  }
  
  print "\n";
  
  if( ( $$r_bStaticLibs32 != 0 ) || ( $$r_bSharedLibs32 != 0 ) ) {
    $dir = File::Spec->catdir( $boost_lib_dir, 'x86' );
    
    if( -d "$dir" ) {
      get_boost_lib_dir_map( $dir, \%boost_lib_map );
      
      print "Searching for existing boost libraries in $dir ... \n";
      
      if( $$r_bStaticLibs32 != 0 ) {
        $boost_libs_ok = 1;
        for $boost_lib ( @boost_lib_static_list ) {
          if( ! exists( $boost_lib_map{ $boost_lib } ) ) {
            #print "library $boost_lib not found\n";
            $boost_libs_ok = 0;
          } 
        }
        $$r_bStaticLibs32 = 0 if( $boost_libs_ok == 1 );
      }
      
      if( $$r_bSharedLibs32 != 0 ) {
        $boost_libs_ok = 1;
        for $boost_lib ( @boost_lib_shared_list ) {
          if( ! exists( $boost_lib_map{ $boost_lib } ) ) {
            #print "library $boost_lib not found\n";
            $boost_libs_ok = 0;
          }           
        }
        $$r_bSharedLibs32 = 0 if( $boost_libs_ok == 1 );        
      }            
    }
    else {
      print "Searching for existing boost libraries in $dir => the directory does not exist!\n";
    }
  }
 
  
  if( ( $$r_bStaticLibs64 != 0 ) || ( $$r_bSharedLibs64 != 0 ) ) {
    $dir = File::Spec->catdir( $boost_lib_dir, 'x86_64' );
    
    if( -d "$dir" ) {
      %boost_lib_map = ();
      get_boost_lib_dir_map( $dir, \%boost_lib_map );
      
      print "Searching for existing boost libraries in $dir ... \n";
      
      if( $$r_bStaticLibs64 != 0 ) {
        $boost_libs_ok = 1;
        for $boost_lib ( @boost_lib_static_list ) {
          if( ! exists( $boost_lib_map{ $boost_lib } ) ) {
            $boost_libs_ok = 0;
          } 
        }
        $$r_bStaticLibs64 = 0 if( $boost_libs_ok == 1 );
      }
      
      if( $$r_bSharedLibs64 != 0 ) {
        $boost_libs_ok = 1;
        for $boost_lib ( @boost_lib_shared_list ) {
          if( ! exists( $boost_lib_map{ $boost_lib } ) ) {
            $boost_libs_ok = 0;
          }           
        }
        $$r_bSharedLibs64 = 0 if( $boost_libs_ok == 1 );        
      }            
    }
    else {
      print "Searching for existing boost libraries in $dir => the directory does not exist!\n";
    }    
  }
}
    

sub get_boost_lib_dir_map {
  my ( $dir, $rh_boost_lib_map ) = @_;
  my $dh;
  my $filesys_comp;
  my $f;
  my $filename;
  my @lib_list;
  
  $dh = new DirHandle $dir;
  if( ! defined $dh ) {
    fatal_error( __LINE__, "cannot search \"$dir\"\n" );
  }
  
  while( defined( $filesys_comp = $dh->read ) ) { 
    #need to save the file system component temporarily and perform the file type check later on because
    #some platforms donnot allow to check the file type while the directory is being searched.
    push @lib_list, $filesys_comp;
  }
  undef $dh;
  
  
  #filter out files and hidden directories
  foreach $f (@lib_list) {
    if( $f !~ /^\./ ) {
      # the file system component does not start with a dot 
      
      if( $OSNAME eq 'MSWin32' ) {
        if( $f =~ /\.lib$/i ) {
          $$rh_boost_lib_map{ $f } = 1;
        }
      }
      elsif( $OSNAME eq 'darwin' ) {
        $filename = File::Spec->catfile( $dir, $f ) ;
        
        if( ( ( -f $filename ) && ( ! -l $filename ) ) 
           && (  ( $f =~ /\.a$/ ) || ( $f =~ /\.dylib/ ) ) ) {
          $$rh_boost_lib_map{ $f } = 1;
        }
      }
      else {
        $filename = File::Spec->catfile( $dir, $f ) ;
        
        if( ( ( -f $filename ) && ( ! -l $filename ) ) 
           && (  ( $f =~ /\.a$/ ) || ( $f =~ /\.so\./ ) ) ) {
          $$rh_boost_lib_map{ $f } = 1;
        }
      }
    }
  }
}


sub get_boost_root_from_jamroot {
  my ( $r_boost_root_dir, $Jamroot ) = @_;
  my $boost_root_dir;
  my $boost_root_found = 0;
  my $fh;
  my $line;
  my $jamroot_dir;
  
  if( ! -f $Jamroot ) {
    fatal_error( __LINE__, "the file \"$Jamroot\" does not exist\n" );
  }

  $fh = new FileHandle $Jamroot, "r";
  if( ! defined( $fh ) ) {
    fatal_error( __LINE__, "cannot open \"$Jamroot\" for reading\n" );
  }

  while( defined( $fh ) && ( ! $fh->eof() ) ) {
    $line = $fh->getline;
    chomp( $line );
  
    # path-constant BOOST_SDK_ROOT : ../Boost_1_40_0 ;
    if( $line =~ /^\s*path-constant\s*BOOST_SDK_ROOT\s*:\s*([^;]+)/ ) {
      $boost_root_dir = $1;
      
      $boost_root_dir =~ s/[ \t]*//g;
      if( $OSNAME eq 'MSWin32' ) {
        $boost_root_dir =~ s/\//\\/g;
      }      
      $boost_root_found = 1;
      undef $fh;
    }
  } 

  if( defined( $fh ) ) {
    # close the file handle if the version expression was not found.
    $fh->close();
  }

  return 0 if( ! $boost_root_found );
 
  if( ! File::Spec->file_name_is_absolute( $boost_root_dir ) ) {
    $jamroot_dir = dirname( $Jamroot );
    $boost_root_dir = File::Spec->catdir( $jamroot_dir, $boost_root_dir );
  }
  
  if( ! -d $boost_root_dir ) {
    fatal_error( __LINE__, "BOOST_SDK_ROOT \"$boost_root_dir\" does not exist.\n" );  
  }

  $$r_boost_root_dir = File::Spec->rel2abs( $boost_root_dir );
  
  return 1;
}
  
# find the boost build directory given the script name 
sub find_boost_build_dir {
  my ( $script ) = @_;
  my $boost_build_dir;
  my $filename;
  my $suffix;
  my $script_dir;
  my $no_file = 1;
  my $volume;
  my $directories;
  my @dir_list;
  
  ( $filename, $script_dir, $suffix ) = fileparse( $script );
  
  $script_dir = File::Spec->rel2abs( $script_dir );
  
  ( $volume, $directories, $filename ) = File::Spec->splitpath( $script_dir, $no_file );

  @dir_list = File::Spec->splitdir( $directories );
  
  #drop the last path component
  pop @dir_list;
  
  # reassemble the path to get the boost build directory
  $directories = File::Spec->catdir( @dir_list );

  $boost_build_dir = File::Spec->catpath( $volume, $directories, '' );
  
  return $boost_build_dir;
}

sub check_for_dev_packages {
  my ( $rh_missing_pkg_info, $system_info, $bIsLinux64Bit, $bBuildLib32, $bBuildLib64 ) = @_;
  my $b_pkg_missing = 0;
  my $ret_value;
  my @system_info_components = split /:/, $system_info;
  
  print "Searching for required developer packages ...                       ";
  
  if( ! -f '/usr/include/zlib.h' ) {
    print "\n" if( $b_pkg_missing == 0 );
    print "\tboost_install.pl: error: /usr/include/zlib.h not found"; 
    if( exists( $$rh_missing_pkg_info{ 'zlib.h' } ) ) {
      print ", did you install $missing_pkg_info{ 'zlib.h' } ?";
    }
    print "\n";
    $b_pkg_missing = 1;
  }

   if( ! -f '/usr/include/bzlib.h' ) {
     print "\n" if( $b_pkg_missing == 0 );
     print "\tboost_install.pl: error: /usr/include/bzlib.h not found";

     if( exists( $$rh_missing_pkg_info{ 'bzlib.h' } ) ) {
       print ", did you install $missing_pkg_info{ 'bzlib.h' } ?";
     }
     print "\n";
     $b_pkg_missing = 1;
   }
   
   if( $system_info_components[ 3 ] eq 'ubuntu' ) {
     # ubuntu distribution, check for cross compiling support
     if(  ( ( $bIsLinux64Bit == 0 ) && ( $bBuildLib64 != 0 ) )
       || ( ( $bIsLinux64Bit != 0 ) && ( $bBuildLib32 != 0 ) ) )
     {
       my $pkg;
       my @lib_pkg_list;
       
       push @lib_pkg_list, 'gcc-multilib';
       push @lib_pkg_list, 'g++-multilib';
        
       
       if( $bIsLinux64Bit != 0 ) {
         # cross compiling on 64 bit ubuntu: target 32 bit
         push @lib_pkg_list, 'ia32-libs';
         push @lib_pkg_list, 'lib32z1';         
         push @lib_pkg_list, 'lib32z1-dev';
         
         if( $system_info_components[ 4 ] eq 'karmic' ) {
           push @lib_pkg_list, 'lib32bz2-1.0';
           push @lib_pkg_list, 'lib32bz2-dev';
         }
       }
       else {
         # cross compiling on 32 bit ubuntu: target 64 bit
         push @lib_pkg_list, 'ia64-libs';
         push @lib_pkg_list, 'lib64z1';
         push @lib_pkg_list, 'lib64z1-dev';
         
         if( $system_info_components[ 4 ] eq 'karmic' ) {
           push @lib_pkg_list, 'lib64bz2-1.0';
           push @lib_pkg_list, 'lib64bz2-dev';
         }
       }
       
       foreach $pkg (@lib_pkg_list) {
         $ret_value = is_debian_package_installed( $pkg );
         if( ! $ret_value ) {
           print "\n" if( $b_pkg_missing == 0 );
           print "\tboost_install.pl: error: cross compiling support requires package $pkg.\n\tPlease check the command line or add the missing packages.\n";
           $b_pkg_missing = 1;
         } 
       }
     }
   } 
   print "[done]\n" if ( $b_pkg_missing == 0 );
   
   exit( 1 ) if ( $b_pkg_missing != 0 );
}

sub is_debian_package_installed {
  my ( $package_name ) = @_;
  my $is_installed = 0;
  my $ret_value;
  my @lines;
  my $line;
  my $package_status;
  my @package_status_components;
  
  $ret_value = `dpkg -l $package_name 2>/dev/null`;
  if( $? != 0 ) {
    #fatal_error( __LINE__, "the package \"$package_name\" is unknown, please check the script logic.\n" );
    return 0;
  }
  
  @lines = split /\n/, $ret_value;
  
  #print "\n";
  #foreach $line (@lines) {
  #  print "$line\n"
  #}
  
  $package_status = pop @lines;
  
  #print "package_status: $package_status\n";
  
  if( $package_status =~ /^ii / ) {
    $is_installed = 1;
  }  
  
  return $is_installed;
} 

sub check_bjam_path {
  my ( $bjam_bin_dir, $rl_todo_list ) = @_;
  my $b_bjam_din_dir_on_path = 0;
  my $bjam_prog_name; 
  my $bjam_prog;
  my @search_path;
  my $dir;
  my @multiple_bjam_prog_list;
  my $msg;
  
  if( $OSNAME eq 'MSWin32' ) {
    $bjam_prog_name = 'bjam.exe';
  }
  else {
    $bjam_prog_name = 'bjam';
  }
  
  @search_path = File::Spec->path();
  foreach $dir (@search_path) {
    if( $dir eq $bjam_bin_dir ) {
        $b_bjam_din_dir_on_path = 1;
    }
  }  
    
  #search for multiple bjam executables on the path
  foreach $dir (@search_path) {
    $bjam_prog = File::Spec->catfile( $dir, $bjam_prog_name );
    if( -f $bjam_prog ) {
      push @multiple_bjam_prog_list, $dir;
    }
  }  
  
  if( ! $b_bjam_din_dir_on_path ) {
    $msg = "the bjam executable is currently not on your search path, consider updating the path.\n";
    
    if( scalar( @multiple_bjam_prog_list ) > 0 ) {
      $msg =    "Warning: found other bjam executables in the following directories\n";
      foreach $dir (@multiple_bjam_prog_list) {
        $msg .= "            $dir\n";
      }
    }
    push @$rl_todo_list, $msg;
  }
  elsif( scalar( @multiple_bjam_prog_list ) > 1 ) {    
    $msg =    "Warning: found multiple bjam executables in the following directories, please check the path carefully.\n";
    foreach $dir (@multiple_bjam_prog_list) {
      $msg .= "            $dir\n";
    }
    
    push @$rl_todo_list, $msg;
  }
  
  return $b_bjam_din_dir_on_path;
}


sub build_bjam {
  my ( $r_bjam_prog, $boost_build_dir, $boost_toolset, $bIsPlatform64Bit ) = @_;
  my $platform_suffix;
  my $arch;
  my $cur_dir = File::Spec->curdir();
  my $bjam_src_dir;
  my @tmp_dirs;
  my $dir;
  
  
  $bjam_src_dir = File::Spec->catdir( $boost_build_dir, 'boost-build', 'jam', 'src' );
  
  if( ! -d $bjam_src_dir ) {
    die "bjam source directory \"$bjam_src_dir\" does not exist.";
  }
 
  if( $OSNAME eq 'MSWin32' ) {
    if( $bIsPlatform64Bit ) {
      $platform_suffix = 'ntx86_64';
    }
    else {
      $platform_suffix = 'ntx86';
    }
  }
  elsif( $OSNAME eq 'linux' ) {
    if( $bIsPlatform64Bit != 0 ) {
      $platform_suffix = 'linuxx86_64';
    }
    else {
      $platform_suffix = 'linuxx86';
    }
  }
  elsif( $OSNAME eq 'darwin' ) {
    $arch = `arch`;
    chomp $arch;
    
    if( $arch eq 'x86_64' ) {
      $platform_suffix = 'macosxx86_64';
    }
    else {
      $platform_suffix = 'macosxx86';
    }
  }
  elsif( $OSNAME eq 'solaris' ) {
    $arch = `arch`;
    chomp $arch;
    
    if( $arch eq 'i86pc' ) {
      $platform_suffix = 'solarisx86';
    }
    else {
      fatal_error( __LINE__, "the solaris architecture \"$arch\" is not yet supported.\n" );
    }
  }
  else {
    fatal_error( __LINE__, "this platform is not yet supported.\n" );
  }
  
  if( $OSNAME eq 'MSWin32' ) {
    $$r_bjam_prog = File::Spec->catfile( $bjam_src_dir, "bin.$platform_suffix", 'bjam.exe' );
  }
  else {
    $$r_bjam_prog = File::Spec->catfile( $bjam_src_dir, "bin.$platform_suffix", 'bjam' );
  }
  
  if( -f $$r_bjam_prog ) {
    unlink( $$r_bjam_prog );
  }
  
    
  print "\n";
  print "========================================================\n";
  print "Start building bjam in $bjam_src_dir ...\n";
  print "========================================================\n";
  print "\n";
    
    
  chdir( $bjam_src_dir );
  
  push @tmp_dirs, 'bootstrap';
  
  if( -d 'bootstrap') {
    rmtree( 'bootstrap', { verbose => 1 } );
  }
    
    
  if(   ( $OSNAME eq 'linux' ) 
     || ( $OSNAME eq 'solaris' ) 
     || ( $OSNAME eq 'darwin' ) ) {
      
      system( "./build.sh" );
      
      if( $? != 0 ) {
        fatal_error( __LINE__, "bjam failed to build, please check the messages.\n" );
      }
      
      
  } 
  elsif( $OSNAME eq 'MSWin32' ) {
    my %toolset_map = ( 
      'msvc-8.0'  => 'vc8',
      'msvc-9.0'  => 'vc9',
      'msvc-10.0' => 'vc10'
    );
   
    if( ! exists( $toolset_map{$boost_toolset} ) )  {
      fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not support or unknown (bjam build environment.\n" );
    }
    
    if( $bIsPlatform64Bit ) {
      #build the 64 bit bjam executable.
      system( "build.bat $toolset_map{$boost_toolset} x86_amd64" );
    }
    else {
      system( "build.bat" );
    }
    
    if( $? != 0 ) {
      fatal_error( __LINE__, "bjam failed to build, please check the messages.\n" );
    }
  }
  else {
    fatal_error( __LINE__, "internal error: don't know how to build bjam on this platform.\n" );
  }
    
  if( -d 'bootstrap') {
    rmtree( 'bootstrap', { verbose => 1 } );
  }  
  
  chdir( $cur_dir );
}

sub parse_bjam_verbose_response {
  my ( $bjam_verbose_response ) = @_;
  my @lines;
  my $line;
  my @token_list;
  my $version_string;
  
  @lines = split( '\n', $bjam_verbose_response );
 
  $line = shift( @lines );
  
  # "Boost.Jam  Version 3.1.17. OS=LINUX." 
  @token_list = split /\s+/, $line;
  
  if( scalar( @token_list ) != 4 ) {
    die "bjam -v response not understood.";
  }
  
  $version_string = $token_list[ 2 ];
  
  $version_string =~ s/\.$//;
  
  return $version_string;
}

sub get_bjam_version {
  my ( $bjam_prog ) = @_;
  my $bjam_verbose_response;
  my $bjam_version;
  
  $bjam_prog = quote_executable_path( $bjam_prog );
  
  $bjam_verbose_response = `$bjam_prog -v`;
  if( $? != 0 ) {
    fatal_error( __LINE__, "internal error: bjam does not understand -v\n" );
  }
  
  $bjam_version = parse_bjam_verbose_response( $bjam_verbose_response );
  
  return $bjam_version;
}

sub check_bjam_version {
  my ( $bjam_prog, $bjam_prog_built ) = @_;
  my $bjam_version;
  my $bjam_version_built;
  my $result;
  
  if( ! -f $bjam_prog ) {
    return 1;
  }
  
  $bjam_version_built = get_bjam_version( $bjam_prog_built );
  $bjam_version = get_bjam_version( $bjam_prog );
  
  $result = version_compare( $bjam_version_built, $bjam_version );
  
  return ( $result == 1 ) ? 1 : 0; 
}


#
# return 1 if $bjam_version_built is greater than $bjam_version
#
sub version_compare {
  my ( $version_left, $version_right ) = @_;
  my @version_ids_1;
  my @version_ids_2;
  my $ver_1;
  my $ver_2;
  
  # make sure the version numbers are numeric
  if( $version_left !~ /^([0-9.\-]+)$/ ) {
    fatal_error( __LINE__, "version \"$version_left\" not understood.\n" );
  }
  
  if( $version_right !~ /^([0-9.\-]+)$/ ) {
    fatal_error( __LINE__, "version \"$version_right\" not understood.\n" );
  }
  
  
  @version_ids_1 = split /[.\-]+/, $version_left;
  @version_ids_2 = split /[.\-]+/, $version_right;
  
  
  while( scalar( @version_ids_1 ) > scalar( @version_ids_2 ) ) {
    push @version_ids_2, '0';
  }
  
  while( scalar( @version_ids_2 ) > scalar( @version_ids_1 ) ) {
    push @version_ids_1, '0';
  }
  
  while( scalar( @version_ids_1 ) > 0 ) {
    $ver_1 = shift( @version_ids_1 );
    $ver_2 = shift( @version_ids_2 );
    
    return 1  if( $ver_1 > $ver_2 );
    return -1 if( $ver_1 < $ver_2 );
  }
  
  return 0;
}

sub check_include_exclude_lists {
  my ( $rl_include_list, $rl_exclude_list ) = @_;
  my %include_element_hash;
  my $element;
  my $b_exclude_list_modified = 0;
  my @exclude_list_tmp;
  
  # remove any duplicate elements from the list
  remove_duplicate_list_elements( $rl_include_list );
  remove_duplicate_list_elements( $rl_exclude_list );
  
  # elements on the include list take precedence over elements on the exclude list;
  # e.g. python is on the exclude list by default and a user wishing to have python has
  # to put it on the include list.
  
  foreach $element ( @$rl_include_list ) {
    $include_element_hash{ $element } = '1';
  }

  # walk the exclude list and remove elements that are on the include list as well
  foreach $element ( @$rl_exclude_list ) {
    if( exists( $include_element_hash{ $element } ) ) {
      $b_exclude_list_modified = 1;
    }
    else {
      push @exclude_list_tmp, $element;
    }    
  }  
  
  if( $b_exclude_list_modified ) {
    @$rl_exclude_list = @exclude_list_tmp;
  }  
}

sub remove_duplicate_list_elements {
  my ( $rl_list ) = @_;
  my %element_hash;
  my @tmp_list;
  my $element;
  
  # a boolean flag to indicate the input list has duplicate elements
  my $b_list_modified = 0;
  
  foreach $element ( @$rl_list ) {
    if( exists( $element_hash{ $element } ) ) {
      $b_list_modified = 1;
    }
    else {
      $element_hash{ $element } = '1';
      push @tmp_list, $element;
    }
  }
  
  if( $b_list_modified ) {
    @$rl_list = ();
    push @$rl_list, @tmp_list;
  }
}

sub check_jamroot {
  my ( $Jamroot, $rl_todo_list ) = @_;
  my $line;
  my $line_cnt = 0;
  my $fh;
  my $boost_lib_defs;
  my $script_dir;
  my $msg;
  
  $fh = new FileHandle $Jamroot, "r";
  
  if( ! defined( $fh ) ) {
    fatal_error(  __LINE__, "cannot open \"$Jamroot\" for reading." );
  }
  
  while( defined( $fh ) && ( ! $fh->eof() ) ) {
    $line = $fh->getline;
    chomp( $line );
    $line_cnt++;
    
    # path-constant BOOST_LIB_DEFS : ./BoostAddon/src/lib/boost ;
    if( $line =~ /^\s*path-constant\s*BOOST_LIB_DEFS\s*:\s*([^;]+)/ ) {
      $boost_lib_defs = $1;
      
      $boost_lib_defs =~ s/[ \t]*//g;
      
      if( ( $boost_lib_defs ne './BoostBuild/src/lib/boost' ) 
         && ( $boost_lib_defs ne 'BoostBuild/src/lib/boost' ) 
         && ( $boost_lib_defs ne '$(TOP)/BoostBuild/src/lib/boost') ) {
        
        $msg = "Jamroot: line $line_cnt: please change to \"path-constant BOOST_LIB_DEFS : ./BoostBuild/src/lib/boost ;\"\n";
        push @$rl_todo_list, $msg;
      }
    }   
    elsif( $line =~ /^\s*constant\s*BOOST_SDK_VERSION\s*:\s*([^;]+)/ ) {
      $msg = "Jamroot: line $line_cnt: \"constant BOOST_SDK_VERSION ...\" is now obsolete, consider to delete it.\n";
      push @$rl_todo_list, $msg;
    }  
    elsif( $line =~ /^\s*constant\s*BOOST_SDK_MAJOR_VERSION\s*:\s*([^;]+)/ ) {
      $msg = "Jamroot: line $line_cnt: \"constant BOOST_SDK_MAJOR_VERSION ...\" is now obsolete, consider to delete it.\n";
      push @$rl_todo_list, $msg;      
    }
    elsif( $line =~ /^\s*constant\s*BOOST_SDK_MINOR_VERSION\s*:\s*([^;]+)/ ) {
      $msg = "Jamroot: line $line_cnt: \"constant BOOST_SDK_MINOR_VERSION ...\" is now obsolete, consider to delete it.\n";
      push @$rl_todo_list, $msg;         
    }
    elsif( $line =~ /^\s*constant\s*BOOST_SDK_BUILD_VERSION\s*:\s*([^;]+)/ ) {
      $msg = "Jamroot: line $line_cnt: \"constant BOOST_SDK_BUILD_VERSION ...\" is now obsolete, consider to delete it.\n";
      push @$rl_todo_list, $msg;      
    }
  }
  
  undef $fh;
  
} 

sub check_boost_build_root {
  my ( $boost_build_file, $rl_todo_list ) = @_;
  my $fh;
  my $line;
  my $line_cnt = 0;
  my $boost_build_root_found = 0;
  my $boost_build_root;
  my $boost_build_filename;
  my $msg;
  
  $fh = new FileHandle $boost_build_file, "r";
  
  if( ! defined( $fh ) ) {
    fatal_error(  __LINE__, "cannot open \"$boost_build_file\" for reading." );
  }

  while( defined( $fh ) && ( ! $fh->eof() ) ) {
    $line = $fh->getline;
    chomp( $line );
    $line_cnt++;
    
    # boost-build BoostBuild/boost-build ;    
    if( $line =~ /^\s*boost-build\s+([^ \t;]+)/ ) {
      $boost_build_root = $1;
      
      $boost_build_root_found = 1;
      undef $fh;
    }
  } 

  if( defined( $fh ) ) {
    # close the file handle if the version expression was not found.
    $fh->close();
  }

  if( ! $boost_build_root_found ) {
    return;
  }
  else {
    # check path to boost-build
    $boost_build_filename = basename( $boost_build_file );
    if( ( $boost_build_root ne 'BoostBuild/boost-build' ) && ( $boost_build_root ne './BoostBuild/boost-build' ) ) {
      $boost_build_filename = basename( $boost_build_file );
      $msg = "$boost_build_filename: line $line_cnt: consider to replace the line with \"boost-build BoostBuild/boost-build ;\"\n";
      
      push @$rl_todo_list, $msg;
    }
  }
}

sub create_default_bjamrc_file {
  my ( $bjam_bin_dir ) = @_;
  my $rc_file_content;
  my $fh;
  my $rc_file;
  
  
  $rc_file = File::Spec->catfile( $ENV{ 'HOME' }, '.bjamrc'  );

  $fh = new FileHandle $rc_file, "w";
  if( ! defined( $fh ) ) {
    fatal_error(  __LINE__, "cannot open \"$rc_file\" for writing." );
  }
  
  $rc_file_content = <<"__END_OF_CONTENT__";
#
# \$HOME/.bjamrc
#

if [ -d "$bjam_bin_dir" ]; then
  PATH="$bjam_bin_dir:\$PATH"
fi

__END_OF_CONTENT__

  print $fh $rc_file_content;
 
  undef $fh;
}

sub is_platform_64bit {
  my ( $system_info ) = @_;
  my $b_platform_64bit = 0;
  my @system_info_components = split /:/, $system_info;

  # The supported hw platform is intel, hence it is sufficient to check for x86_64.
  # N.B. hw platform intel includes solaris and macosx.
  if( $system_info_components[ 1 ] eq 'x86_64' ) {
    $b_platform_64bit = 1;
  }
  
  return $b_platform_64bit;
}


sub find_latest_msvc_version {
  my ( $tool_dir ) = @_;
  my $external_cmd;

  $external_cmd = File::Spec->catfile( $tool_dir, 'windows', 'find_latest_msvc_version.pl' );

  return invoke_external_cmd( $external_cmd );
}

sub invoke_external_cmd {
  my ( $external_cmd ) = @_;
  my $return_value;
  my $external_cmd_quoted = quote_executable_path( $external_cmd );
  
  if( ! -f "$external_cmd" ) {
    fatal_error( __LINE__, "the external command \"$external_cmd\" does not exist, check your installation\n" );
  }
  
  $return_value = `$external_cmd_quoted`;
  if( $? != 0 ) {
    die "error: the external command \"$external_cmd\" failed unexpectedly, please contact technical support";
  }
  
  return $return_value;
}


sub get_boost_lib_suffix {
  my ( $boost_version, $boost_toolset, $b_debug_build ) = @_;
  my $boost_lib_suffix = '-';
  my $boost_toolset_version;
  my @boost_version_elems;
  my @compiler_version_elems;
  
  @boost_version_elems = split /\./, $boost_version;
  
  if( $OSNAME eq 'MSWin32' ) {
    if( $boost_toolset =~ /^msvc-([0-9.]+)/ ) {
      $boost_toolset_version = $1;
    }
    else {
      fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not supported yet (version missing), check spelling\n" );
    }
    @compiler_version_elems = split /\./, $boost_toolset_version;
    $boost_lib_suffix .= 'vc' . $compiler_version_elems[ 0 ] . $compiler_version_elems[ 1 ];
    
    if( $b_debug_build ) {
      $boost_lib_suffix .= '-mt-gd-' . $boost_version_elems[ 0 ] . '_' . $boost_version_elems[ 1 ];
    }
    else {
      $boost_lib_suffix .= '-mt-' . $boost_version_elems[ 0 ] . '_' . $boost_version_elems[ 1 ];
    }       
  }
  elsif( $OSNAME eq 'linux' ) {
    if( $boost_toolset eq 'gcc' ) {
      $boost_toolset_version = `gcc -dumpversion`;
      chomp $boost_toolset_version;

    }
    else {
      if( $boost_toolset =~ /^gcc-([0-9.]+)/ ) {
        $boost_toolset_version = $1;
      }    
      else {
        fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not supported yet (version missing), check spelling\n" );
      }
    }
    
    @compiler_version_elems = split /\./, $boost_toolset_version;  
    $boost_lib_suffix .= 'gcc' . $compiler_version_elems[ 0 ] . $compiler_version_elems[ 1 ];
    
    if( $b_debug_build ) {
      $boost_lib_suffix .= '-mt-d-' . $boost_version_elems[ 0 ] . '_' . $boost_version_elems[ 1 ];
    }
    else {
      $boost_lib_suffix .= '-mt-' . $boost_version_elems[ 0 ] . '_' . $boost_version_elems[ 1 ];
    }    
  }
  elsif( $OSNAME eq 'darwin' ) {
  
    if( $boost_toolset !~ /^darwin.*/ ) {
      fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not supported yet, check spelling (must be \"darwin[-optionalversion]\")\n" );
    }

    if( $boost_toolset eq 'darwin' ) {
      $boost_toolset_version = `gcc -dumpversion`;
      chomp $boost_toolset_version;
    }
    else {
      # the eclipse perl checker does not like the \1 and recommends $1 instead.
      #if( $boost_toolset =~ s/^darwin-([0-9.]+).*/\1/ ) { 
       
      if( $boost_toolset =~ s/^darwin-([0-9.]+).*/$1/ ) {
        $boost_toolset_version = $boost_toolset;
      }    
      else {
        fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not supported yet (version missing), check spelling\n" );
      }
    }
    @compiler_version_elems = split /\./, $boost_toolset_version;  
    $boost_lib_suffix .= 'gcc' . $compiler_version_elems[ 0 ] . $compiler_version_elems[ 1 ];
    if( $b_debug_build ) {
      $boost_lib_suffix .= '-mt-d-' . $boost_version_elems[ 0 ] . '_' . $boost_version_elems[ 1 ];
    }
    else {
      $boost_lib_suffix .= '-mt-' . $boost_version_elems[ 0 ] . '_' . $boost_version_elems[ 1 ];
    }    
  }
  elsif( $OSNAME eq 'solaris' ) {
    if( $b_debug_build ) {
      $boost_lib_suffix .= 'sw-mt-d-' . $boost_version_elems[ 0 ] . '_' . $boost_version_elems[ 1 ];
    }
    else {
      $boost_lib_suffix .= 'sw-mt-' . $boost_version_elems[ 0 ] . '_' . $boost_version_elems[ 1 ];
    }   
  }
  else {
    fatal_error( __LINE__, "unsupported platform encountered." );
  }
  
  return $boost_lib_suffix;
}

sub get_toolset_dir_suffix {
  my ( $tool_dir, $boost_toolset ) = @_;
  my $boost_toolset_dir_suffix = '';
  my $boost_toolset_version;
  
  if( $OSNAME eq 'MSWin32' ) {
    if( $boost_toolset !~ /^msvc.*/ ) {
      fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not supported yet, check spelling\n" );
    }
  
    if( $boost_toolset =~ /^msvc-([0-9.]+).*/ ) {
      $boost_toolset_dir_suffix = $boost_toolset;
    }    
    else {
      fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not supported yet (version missing), check spelling\n" );
    }
  }
  elsif( $OSNAME eq 'linux' ) {
    if( $boost_toolset eq 'gcc' ) {
      $boost_toolset_version = `gcc -dumpversion`;
      chomp $boost_toolset_version;

      $boost_toolset_dir_suffix = 'gcc-' . $boost_toolset_version;
    }
    else {
      if( $boost_toolset =~ /^gcc-([0-9.]+).*/ ) {
        $boost_toolset_dir_suffix = $boost_toolset;
      }    
      else {
        fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not supported yet (version missing), check spelling\n" );
      }
    }
  }
  elsif( $OSNAME eq 'darwin' ) {
  
    if( $boost_toolset !~ /^darwin.*/ ) {
      fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not supported yet, check spelling (must be \"darwin[-optionalversion]\")\n" );
    }

    if( $boost_toolset eq 'darwin' ) {
      $boost_toolset_version = `gcc -dumpversion`;
      chomp $boost_toolset_version;
      $boost_toolset_dir_suffix = 'gcc-' . $boost_toolset_version;
    }
    else {
      # the eclipse perl checker does not like the \1 and recommends $1 instead.
      #if( $boost_toolset =~ s/^darwin-([0-9.]+).*/\1/ ) { 
       
      if( $boost_toolset =~ s/^darwin-([0-9.]+).*/$1/ ) {
        $boost_toolset_dir_suffix = "gcc-" . $boost_toolset;
      }    
      else {
        fatal_error( __LINE__, "the toolset \"$boost_toolset\" is not supported yet (version missing), check spelling\n" );
      }
    }
  }
  elsif(  $OSNAME eq 'solaris' ) {
    $boost_toolset_dir_suffix = 'sw'
  }
  
  return $boost_toolset_dir_suffix;
}

# 
# On windows it is more convenient to store the DLLs in a different directory.
#
sub copy_boost_dlls {
  my ( $boost_bin_dir, $boost_lib_dir ) = @_;
  my @dll_list;
  my @dir_list_tmp;
  my $dirh;
  my $filesys_comp;
  my $f;
  my $filename;
  
  #search lib_dir
  $dirh = new DirHandle $boost_lib_dir;
  if( ! defined $dirh ) {
    fatal_error( __LINE__, "cannot search \"$boost_lib_dir\"\n" );
  }
  
  while( defined( $filesys_comp = $dirh->read ) ) { 
    #need to save the file system component temporarily and perform the file type check later on because
    #some platforms donnot allow to check the file type while the directory is being searched.
    push @dir_list_tmp, $filesys_comp;
    #print "filesys_comp => $filesys_comp\n";
  }
  undef $dirh;
  
  #filter out files and hidden directories
  foreach $f (@dir_list_tmp) {
    if( $f !~ /^\./ ) {
      # the file system component does not start with a dot 
      $filename = File::Spec->catfile( $boost_lib_dir, $f ) ;
      
      if( ( -f $filename ) && ( $f =~ /\.dll$/ ) ) {
        push @dll_list, $filename;
      }
    }
  }
  
  if( scalar( @dll_list ) > 0 ) {
    #create destination directory if needed
    if( ! -e $boost_bin_dir ) {
      mkpath( $boost_bin_dir, {verbose => 1} );
    }    
    
    foreach $f (@dll_list) {      
      print "copy $f $boost_bin_dir \n";
      copy( $f, $boost_bin_dir ) or die "Copy( $f $boost_bin_dir ) failed: $!"; 
      
      #remove the DLL from the library directory
      unlink( $f );
    }
  }
}

sub add_pkg_name_info {
  my ( $rh_missing_pkg_info ) = @_;

  if( $OSNAME eq 'linux' ) {
    if( -f '/etc/fedora-release' ) {
      #fedora system
      $rh_missing_pkg_info->{ 'zlib.h' } = 'zlib-devel';
      $rh_missing_pkg_info->{ 'bzlib.h' } = 'bzip2-devel';
    }
    elsif( -f '/etc/SuSE-release' )
    {
      $rh_missing_pkg_info->{ 'zlib.h' } = 'zlib-devel';
      $rh_missing_pkg_info->{ 'bzlib.h' } = 'libbz2-devel';
    }
    elsif( ( -f '/etc/lsb-release' ) && ( -f '/etc/debian_version' ) ) {
      #ubuntu/debian system
      $rh_missing_pkg_info->{ 'zlib.h' } = 'zlib1g-dev';
      $rh_missing_pkg_info->{ 'bzlib.h' } = 'libbz2-dev';
    }
  }
}

sub get_boost_version {
  my( $boost_root ) = @_;
  my $boost_jamroot;
  my $line;
  my $fd_jamroot;
  my $boost_version;
  my $boost_version_found = 0;
  
  $boost_jamroot = File::Spec->catfile( $boost_root, 'Jamroot' );

  if( ! -f $boost_jamroot ) {
    fatal_error( __LINE__, "the file \"$boost_jamroot\" does not exist\n" );
  }

  $fd_jamroot = new FileHandle $boost_jamroot, "r";
  if( ! defined( $fd_jamroot ) ) {
    fatal_error( __LINE__, "cannot open \"$boost_jamroot\" for reading.\n" );
  }

  while( defined( $fd_jamroot ) && ( ! $fd_jamroot->eof() ) ) {
    $line = $fd_jamroot->getline;
    chomp( $line );
  
    if( $line =~ /^constant\s*BOOST_VERSION\s*:\s*([0-9.]+)/ ) {
      $boost_version = $1;
      
      $boost_version_found = 1;
      undef $fd_jamroot;
    }
  }

  if( defined( $fd_jamroot ) ) {
    # close the file handle if the version expression was not found.
    $fd_jamroot->close();
  }

  if( ! $boost_version_found ) {
    fatal_error( __LINE__, "failed to parse boost version from Jamroot\n" );  
  }

  return $boost_version;
}

sub get_home_dir {
  my $home_dir;

  if( $OSNAME eq 'MSWin32' ) {
    if( ! exists( $ENV{ 'USERPROFILE' } ) ) {
      fatal_error( __LINE__, "failed to determine your home directory; make sure \%USERPROFILE\% is defined.\n" );  
    }
    $home_dir = $ENV{ 'USERPROFILE' };
  }
  else {
    if( ! exists( $ENV{ 'HOME' } ) ) {
      fatal_error( __LINE__, "failed to determine your home directory; make sure \$HOME is defined.\n" );  
    }
    $home_dir = $ENV{ 'HOME' };
  }  
  
  if( ! -d "$home_dir" ) {
    fatal_error( __LINE__, "the home directory \"$home_dir\" does not exist, check your environment.\n" );  
  }
  
  return $home_dir;
}

sub get_default_bjam_dir {
  my ( $tool_dir, $bIsPlatform64Bit ) = @_;
  my $home_dir;
  my $bjam_bin_dir;
  
  $home_dir = get_home_dir();
  
  #if( ( $OSNAME eq 'MSWin32' ) || ( $OSNAME eq 'darwin' ) ) {
  #  $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin' );
  #}
  if(  $OSNAME eq 'darwin' ) {
    $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin' );
  }
  if(  $OSNAME eq 'MSWin32' ) {
    if( $bIsPlatform64Bit ) {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'x86_64' );
    }
    else {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'x86' );
    }
  }    
  elsif( $OSNAME eq 'linux' )  {
    # I only support intel x86/amd64 platforms
    if( $bIsPlatform64Bit ) {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'linux', 'x86_64' );
    }
    else {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'linux', 'x86' );
    }
  }
  elsif( $OSNAME eq 'solaris' ) {
    # I only support intel x86/amd64 platforms
    if( $bIsPlatform64Bit ) {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'solaris', 'x86_64' );
    }
    else {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'solaris', 'x86' );
    }
  } 
  else {
    die "get_default_bjam_bin_dir: unsupported platform encountered.";
  }
  
  return $bjam_bin_dir;
}

sub usage {

  my $how_to_call = <<"__END_OF_USAGE__";
Usage: boost_install.pl [options]
       
  The script boost_install.pl compiles bjam and most likely also the BOOST libraries 
  referred to by your Jamroot file. The later depends on the content of your Jamroot file, which 
  need to refer to the suitable BOOST_SDK by means of a statement similar to the following
  
  path-constant BOOST_SDK_ROOT : ../Boost_1_41_0 ;
  
  This statement tells boost_install.pl that the suitable BOOST_SDK can be found in ../Boost_1_41_0 
  relatively to the location of the Jamroot file. If the Jamroot file does not include such a statement, 
  boost_install.pl assumes the build environment does not require any BOOST libraries and will just
  build the bjam executable.
  
  
  options:=
     --help              show the usage information
     
     --dry-run           show install steps but don't modify anything
     
     --rebuild           force a rebuild of all boost libraries [default: off]
     
     --bjam-dir          specify a non-standard bjam directory. 
                         
                         Default directories:
                        
                         WINDOWS 32 bit:   %USERPROFILE%\\bin\\x86
                         WINDOWS 64 bit:   %USERPROFILE%\\bin\\x86_64
                         
                         MACOSX:           \$HOME/bin
                         
                         LINUX 32 bit:     \$HOME/bin/linux/x86
                         LINUX 64 bit:     \$HOME/bin/linux/x86_64
                         
     --[no-]build-bjam   build bjam from source [default: on]
     
     --[no-]build-libs   build boost libraries [default: on]
                         example:
                         use "--no-build-libs" if you just need the header 
                         only boost libraries.
                         
     --boost-exclude=<lib1,lib2>
                         specifies a comma separated list of boost libraries not to
                         be built.
                         
                         example:  
                         
                            use "--boost-exclude=graph" to disable building the boost library graph
                            
     --boost-include=<lib1,lib2>
                         specifies a comma separated list of optional boost libraries to
                         be built.
                               
     --toolset=<s>       use the non-default toolset <s> to build the libraries. The default 
                         toolset is platform and sometimes installation dependent.  
                         
                         WINDOWS:                    
                         
                         use "--toolset=msvc-8.0" if msvc-9.0 and msvc-8.0 are installed
                         and you would like to build the msvc-8.0 compliant libraries.
                         
                         LINUX:
                         
                         This option is of limited use at the moment because --toolset=gcc is
                         the only possible option, which is also the default toolset on LINUX.
                         
                         * Don't try something like --toolset=gcc-4.2, it will not work unless
                         * gcc-4.2 is also the default gcc version.
                         
                         OpenSolaris:
                           
                           default: sun (Sun Studio Express) 
                           
                                                         
     --[no-]quit-on-error 
                         terminate build process on first error [default: on]
     
     --[no-]64           build 64 bit libraries [default: on if the platform is able to run 64 bit code]
                         e.g. --64 is required on 32 bit platforms to force building 64 bit libraries.
     
     --[no-]32           build 32 bit libraries [default: on if the platform is able to run 32 bit code]
     
     --[no-]static       build static libraries [default: on]
     
     --[no-]shared       build shared libraries [default: on]
     
     --jobs=<n>          number of parallel bjam jobs [default: <number of processors>]

Default Library Build Options:

   WINDOWS 32 bit: 
     - versioned layout of 32 bit libraries
   
   WINDOWS 64 bit: 
     - versioned layout of 32 bit and 64 bit libraries
   
   LINUX 32 bit:
     - versioned layout of 32 bit libraries  
   
   LINUX 64 bit:
     ubuntu:
       - versioned layout of 32 bit and 64 bit libraries
     other distros:
       - versioned layout of 64 bit libraries
   
   OpenSolaris 32 bit:
     - versioned layout of 32 bit
     
   OpenSolaris 64 bit:
     - versioned layout of 32 bit and 64 bit libraries
     
example 1:

   cd <project>
   WINDOWS: BoostBuild\\bin\\boost_install.pl
   LINUX:  ./BoostBuild/bin/boost_install.pl
   
example 2:

   # This example builds only the static BOOST libraries

   cd <project>
   WINDOWS: BoostBuild\\bin\\boost_install.pl --no-shared
   LINUX:  ./BoostBuild/bin/boost_install.pl --no-shared
   
example 3:

   # How to disable cross compiling on LINUX 64 bit or WINDOWS 64 bit:
   
   cd <project>
   WINDOWS: BoostBuild\\bin\\boost_install.pl --no-32
   LINUX:  ./BoostBuild/bin/boost_install.pl --no-32
   
example 4: 

  # This example builds bjam from source and creates a simple rc file to 
  # be sourced by ~/.profile.

  cd <project>
  WINDOWS: BoostBuild\\bin\\boost_install.pl --build-bjam --no-build-libs
  LINUX: ./BoostBuild/bin/boost_install.pl --build-bjam --no-build-libs
   
   
__END_OF_USAGE__

  print STDERR $how_to_call;
}

#
# issue an error message and terminate the program
#
sub fatal_error {
  my ($line, $msg) = @_;

  #if( defined( $::g_fh_error ) ) {
  #  print $::g_fh_error  "error: $msg";
  #}
  #else {
    print STDERR  "boost_install.pl: error at line $line:  $msg";
  #}

  exit( 1 ) ;
}

#############################################################################################################
#                                         builtin tests
#############################################################################################################

sub test_version_compare {
  my $ver_cmp_result;
  
  print "===============================================================================================\n";
  print "testing function version_compare\n\n";
  
  $ver_cmp_result = version_compare ( '3.1.17', '3.1.17' );
  print "version_compare ( '3.1.17', '3.1.17' ): $ver_cmp_result\n";
  
  $ver_cmp_result = version_compare ( '3.1.18', '3.1.17' );
  print "version_compare ( '3.1.18', '3.1.17' ): $ver_cmp_result\n";
  
  $ver_cmp_result = version_compare ( '3.1.17', '3.1.18' );
  print "version_compare ( '3.1.17', '3.1.18' ): $ver_cmp_result\n";
  
  $ver_cmp_result = version_compare ( '3.1.17', '3.1' );
  print "bjam_version_compare ( '3.1.17', '3.1' ): $ver_cmp_result\n";
  
  $ver_cmp_result = version_compare ( '3.1.17', '3.2' );
  print "version_compare ( '3.1.17', '3.2' ): $ver_cmp_result\n";
 
  $ver_cmp_result = version_compare ( '3.1.17', '3.1.17-1' );
  print "version_compare ( '3.1.17', '3.1.17-1' ): $ver_cmp_result\n";
  print "===============================================================================================\n";
}
