#!/usr/bin/perl
#
# cuda_install.pl
#
use File::Basename;
use File::Copy;
use File::Spec;
use File::Path;
use FileHandle;

use Getopt::Long;
use English;

use strict 'vars';

my $nvsdkcuda_root = '';
my $cuda_root = '';
my $bListUsage = 0;
my $bjam_jobs = 1;
my $bjam_toolset = '';
my $bjam_dir = '';
my $bjam_prog;
my $bBuildSdk = 1;
my $cuda_toolkit_version = '0.0';
my $BoostBuildDir = '';
my $tool_dir;
my $bIsPlatform64Bit = 0;

my %optctl = (
  'help'           => \$bListUsage,
  'jobs'           => \$bjam_jobs,
  'toolset'        => \$bjam_toolset,
  'cudasdk-dir'    => \$nvsdkcuda_root,
  'cuda-dir'       => \$cuda_root,
  'bjam-dir'       => \$bjam_dir
);


# parse the command line 
if( GetOptions( \%optctl, 'help!', 'toolset=s', 'jobs=i', 'cudasdk-dir=s', 'cuda-dir=s', 'bjam-dir' ) != 1 ) {
  fatal_error( __LINE__, "command line not understood, check the syntax\n" );
}

if( $bListUsage != 0 ) {
  usage();
  exit 0;
}

if( scalar(@ARGV) != 0 ) {
  usage();
  exit 1;
}

if( $bjam_dir ne '' ) {
  if( -d "$bjam_dir" ) {
    fatal_error( __LINE__, "the bjam directory \"$bjam_dir\" does not exist." );
  }
}


$bIsPlatform64Bit = is_platform_64bit();

# find BoostBuild given the script name
$BoostBuildDir = find_boost_build_dir( $0 );

$tool_dir = File::Spec->catdir( $BoostBuildDir, 'bin' );


if( $cuda_root eq '' ) {
  $cuda_root = get_default_cuda_root();
}

if ( ! -d "$cuda_root" ) {
  fatal_error(  __LINE__, "the CUDA toolkit root directory \"$cuda_root\" does not exist.\n" );
  exit 1;
}

# make directories absolute
$cuda_root = File::Spec->rel2abs( $cuda_root );
print "cuda root:             $cuda_root\n";


if( $OSNAME ne 'MSWin32' ) {
  # Modify the cuda toolkit environment to point to the current installation directory.
  # Since the user may not have updated his environment yet or the settings are outdated, it is
  # mandatory to override the environment before starting to build the NVSDK.
  # 
  # get_cuda_toolkit_version() needs CUDA_INC_PATH to find the version header.
  #
  $ENV{ 'CUDA_BIN_PATH' } = File::Spec->catdir( $cuda_root, 'bin' );
  $ENV{ 'CUDA_INC_PATH' } = File::Spec->catdir( $cuda_root, 'include' );
  $ENV{ 'CUDA_LIB_PATH' } = File::Spec->catdir( $cuda_root, 'lib' );
}

$cuda_toolkit_version = get_cuda_toolkit_version( $BoostBuildDir );
print "cuda toolkit version:  $cuda_toolkit_version\n";

if( $nvsdkcuda_root eq 'none' ) {
  $bBuildSdk = 0;
  $nvsdkcuda_root = '';
}
else {
  if( $nvsdkcuda_root eq '' ) {
    $nvsdkcuda_root = get_default_cudasdk_root( $cuda_toolkit_version );
  }

  if ( ! -d "$nvsdkcuda_root" ) {
    fatal_error(  __LINE__, "the CUDA SDK root directory \"$nvsdkcuda_root\" does not exist.\n" );
  }
}


if( ( $bBuildSdk != 0 ) && ( $nvsdkcuda_root eq '' ) ) {
  fatal_error(  __LINE__, "the CUDA SDK root directory is unknown.\n\n" );
  
  usage();
  exit 1;
}


if( $bBuildSdk ) {
  my $curdir;
  my $bjam_prog_quoted;
  my $bjam_common_options;
  my $bjam_address_model = '';
  my $bjam_targets;
  my $bjam_build_cmd;
  my @bjam_build_cmds;
  my $build_dir;
  
  $nvsdkcuda_root = File::Spec->rel2abs( $nvsdkcuda_root );
  print "nvsdkcuda_root:        $nvsdkcuda_root\n";

  $build_dir = File::Spec->catdir( $nvsdkcuda_root, 'build' );
  
  if( -d "$build_dir" ) {
    # remove the previous build tree 
    rmtree( $build_dir, { verbose => 0 } );
  }
  
  #find bjam as it may not be on user's search path
  &find_bjam( \$bjam_prog, $bjam_dir, $tool_dir, $bIsPlatform64Bit );
  print "bjam:                  $bjam_prog\n";

  &create_nvsdk_bjam_build_env( $nvsdkcuda_root, $cuda_root, $cuda_toolkit_version, $BoostBuildDir );
  
  $bjam_common_options = '-q';
  if( $bjam_jobs > 1 ) {
    $bjam_common_options .= " -j$bjam_jobs"; 
  }
  if( $bjam_toolset ne '' ) {
    $bjam_common_options .= " toolset=$bjam_toolset";
  }
  
  $bjam_targets = 'variant=release,debug'; 
  if( $OSNAME eq 'MSWin32' ) {
    $bjam_targets .= ' link=shared';
  }
  elsif( $OSNAME eq 'linux' ) {
    $bjam_targets .= ' link=static';
  }
  
  $bjam_prog_quoted = quote_executable_path( $bjam_prog );
 
  if( $OSNAME eq 'MSWin32' ) {
    if( $bIsPlatform64Bit ) {
      $bjam_build_cmd = "$bjam_prog_quoted $bjam_common_options address-model=64 $bjam_targets install";
      
      push @bjam_build_cmds, $bjam_build_cmd;
      
      if( is_cuda_toolkit_version_2_3( $cuda_toolkit_version ) ) {
        $bjam_build_cmd = "$bjam_prog $bjam_common_options $bjam_targets install";

        push @bjam_build_cmds, $bjam_build_cmd;
      }
    }
    else {
      $bjam_build_cmd = "$bjam_prog_quoted $bjam_common_options $bjam_targets install";
      
      push @bjam_build_cmds, $bjam_build_cmd;
    }
  }
  elsif( $OSNAME eq 'linux' )  {
    $bjam_build_cmd = "$bjam_prog_quoted $bjam_common_options $bjam_targets install";
    push @bjam_build_cmds, $bjam_build_cmd;
  }
  else {
    fatal_error( __LINE__, "the platform $OSNAME is unsupported." );
  }
  
  
  #remember where we are
  $curdir = File::Spec->curdir();

  chdir( $nvsdkcuda_root );
  
  foreach $bjam_build_cmd (@bjam_build_cmds) {
    print "=======================================================================================\n";
    print "NVSDK build command: $bjam_build_cmd\n";
    print "=======================================================================================\n";
    system( $bjam_build_cmd );
    if( $? != 0 ) {
      fatal_error( __LINE__, "failed to build the NVSDK.\n" );
    }
  }

  if( -d "$build_dir" ) {
    rmtree( $build_dir, { verbose => 0 } );
  }
  
  # get back to the intial cwd
  chdir( $curdir );
  
  print "\n";
  print "==================================\n";
  print "Finished building NVSDK libraries.\n";
  print "==================================\n";
  print "\n";
}

# 
if( $OSNAME eq 'linux' ) {
  # assume a unix system and create a new cudadevrc
  &create_cudadevrc_file( $cuda_root, $nvsdkcuda_root, $cuda_toolkit_version, $bIsPlatform64Bit );
}

sub quote_executable_path {
  my ( $prog ) = @_;
  
  if( $prog =~ /^"/ ) {
    # already quoted
    return $prog;
  }
  
  if( $prog =~ /[ ]+/ ) {
    $prog = "\"$prog\"";
  }
  
  return $prog;
}

sub get_cuda_toolkit_version {
  my ( $boost_build_dir ) = @_;
  my $cuda_toolkit_version;
  my $version_script = File::Spec->catfile( $boost_build_dir, 'bin', 'get_cuda_toolkit_version.pl' ); 

  if( ! -f $version_script ) {
    fatal_error(  __LINE__, "cannot find version script \"$version_script\"." );
  }
  
  $cuda_toolkit_version = `$version_script`;
  if( $? != 0 ) {
    fatal_error(  __LINE__, "internal error: version script \"$version_script\" failed." );
  }
  
  return $cuda_toolkit_version;
}

sub is_cuda_toolkit_version_2_3 {
 my ( $cuda_toolkit_version ) = @_;
 
 return ( $cuda_toolkit_version =~ /(2\.3b)|(2\.3)/ ) ? 1 : 0;
}

sub create_cudadevrc_file {
  my ( $cuda_root, $nvsdkcuda_root, $cuda_toolkit_version, $bIsPlatform64Bit ) = @_;
  my $cuda_bin_path;
  my $cuda_inc_path;
  my $cuda_lib_path;
  my $cuda_lib64_path;
  my $cuda_lib32_path;
  my $fh;
  my $rc_file;
  my $rc_file_content;


  &configure_cuda_manpath( $cuda_root );

  $cuda_bin_path = File::Spec->catfile( $cuda_root, 'bin' );
  $cuda_inc_path = File::Spec->catfile( $cuda_root, 'include' );
  $cuda_lib_path = File::Spec->catfile( $cuda_root, 'lib' );
  $cuda_lib64_path = File::Spec->catfile( $cuda_root, 'lib64' );
  $cuda_lib32_path = File::Spec->catfile( $cuda_root, 'lib' );
  
  if( ( $OSNAME eq 'linux' ) && ( 1 == $bIsPlatform64Bit ) ) {
    # starting with cuda 2.3 (not 2.3b !!) the 64 bit libraries are kept in a separate 
    # folder
    
    if( -d "$cuda_lib64_path" ) {
      $cuda_lib_path = $cuda_lib64_path;
    }    
  }
 
  
  $rc_file = File::Spec->catfile( $ENV{ 'HOME' }, '.cudadevrc'  );

  $fh = new FileHandle $rc_file, "w";
  if( ! defined( $fh ) ) {
    fatal_error(  __LINE__, "cannot open \"$rc_file\" for writing." );
  }
  
  $rc_file_content = <<"__END_OF_CONTENT__";
#
# \$HOME/.cudadevrc
#

if [ -d "$cuda_bin_path" ]; then
  CUDA_BIN_PATH="$cuda_bin_path"
  export CUDA_BIN_PATH

  PATH="$cuda_bin_path:\${PATH}"
  
fi

if [ -d "$cuda_lib_path" ]; then
  CUDA_LIB_PATH="$cuda_lib_path"
  export CUDA_LIB_PATH
fi

if [ -d "$cuda_inc_path" ]; then
  CUDA_INC_PATH="$cuda_inc_path"
  export CUDA_INC_PATH
fi

__END_OF_CONTENT__

  print $fh $rc_file_content;
  
  if( $nvsdkcuda_root ne '' ) {
    my $nvsdkcompute_root = dirname( $nvsdkcuda_root );

    if( $cuda_toolkit_version eq '2.2' ) {
      $nvsdkcompute_root = $nvsdkcuda_root;
    }
    else {
      $nvsdkcompute_root = $nvsdkcuda_root;
      $nvsdkcuda_root = File::Spec->catdir( $nvsdkcompute_root, 'C' );
    }

    print $fh "if [ -d \"$nvsdkcuda_root\" ]; then\n";
    print $fh "  NVSDKCUDA_ROOT=\"$nvsdkcuda_root\"\n";
    print $fh "  export NVSDKCUDA_ROOT\n";
    print $fh "fi\n";
    print $fh "\n";
    
    #if( $cuda_toolkit_version ne '2.2' ) {
      print $fh "if [ -d \"$nvsdkcompute_root\" ]; then\n";
      print $fh "  NVSDKCOMPUTE_ROOT=\"$nvsdkcompute_root\"\n";
      print $fh "  export NVSDKCOMPUTE_ROOT\n";
      print $fh "fi\n";
    #}

  }
  
  undef $fh;
  
  print "\n";
  print "Saving the new CUDA environment to \"$rc_file\"\n";
  print "You may need to update your shell startup file ~/.profile with lines like\n";
  print "\n";
  print "#---------------------------------------------\n";
  print "RC_FILES=\".bjamrc .cudadevrc\"\n";
  print "\n";
  print "for f in \$RC_FILES\n";
  print "do\n";
  print "  if [ -f \$HOME/\$f ]; then\n";
  print "    . \$HOME/\$f\n";
  print "  fi\n";
  print "done\n";
  print "#---------------------------------------------\n";
  print "\n";  

}

sub configure_cuda_manpath {
  my ( $cuda_root ) = @_;
  my $cuda_manpath = File::Spec->catfile( $cuda_root, 'man' );
  my $manpath_config;
  my @manpath_config_lines_orig;
  my @manpath_config_lines;
  my $line;
  my $fh;

  print "Configuring ~/.manpath ...\n";

  if( ! -d "$cuda_manpath" ) {
    fatal_error( __LINE__, "the CUDA manual pages are expected to be found in \"$cuda_manpath\".");
  }

  $manpath_config = File::Spec->catfile( $ENV{ 'HOME' }, '.manpath' );

  if( -f $manpath_config ) {
    my $manpath_config_tag;
    my $manpath_config_value;
    my $remainder;

    &read_cuda_manpath_config( $manpath_config,\@manpath_config_lines_orig ); 

    #remove lines ending with cuda/man
    foreach $line (@manpath_config_lines_orig) {
      if( $line =~ /^MANDATORY_MANPATH/ ) {
        ( $manpath_config_tag, $manpath_config_value, $remainder ) = split /\s+/, $line, 3;
        if( -d $manpath_config_value ) {
          if( $manpath_config_value !~ /\/cuda/ ) {
            push @manpath_config_lines, $line;
          }
          #else {
          #  print "found existing cuda manpath: $manpath_config_value\n";
          #}
        }
        else {
          print "Removing non-existing manpath \"$manpath_config_value\" from \"$manpath_config\"\n";
        }
      }
      else {
        push @manpath_config_lines, $line;
      }
    }

    # add cuda manpath
    unshift @manpath_config_lines, "MANDATORY_MANPATH $cuda_manpath"; 
  }

  if( scalar( @manpath_config_lines ) == 0 ) { 
    push @manpath_config_lines, "MANDATORY_MANPATH $cuda_manpath"; 
  }

  $fh = new FileHandle $manpath_config, "w";
  if( ! defined( $fh ) ) {
    fatal_error( __LINE__, "cannot open \"$manpath_config\" for writing." );
  }

  foreach $line (@manpath_config_lines) {
    print $fh "$line\n";
  }
  undef $fh;
}

sub read_cuda_manpath_config {
  my ( $manpath_config, $ra_manpath_config_lines ) = @_;
  my $fh;
  my $line;

  $fh = new FileHandle $manpath_config, "r";
  if( ! defined( $fh ) ) {
    fatal_error( __LINE__, "cannot open \"$manpath_config\" for reading." );
  }

  while( ! $fh->eof() ) {
    $line = $fh->getline();
    chomp( $line );
    
    push @$ra_manpath_config_lines, $line;
  }

  undef $fh;
}


sub get_default_cudasdk_root {
  my ( $cuda_toolkit_version ) = @_;
  my $nvsdkcuda_root = '';
  

  if( $OSNAME eq 'linux' ) {
    if( $cuda_toolkit_version eq '2.2' ) {
      $nvsdkcuda_root = File::Spec->catdir( $ENV{ 'HOME' } , 'NVIDIA_CUDA_SDK' );
    }
    elsif( is_cuda_toolkit_version_2_3( $cuda_toolkit_version ) ) {
      $nvsdkcuda_root = File::Spec->catdir( $ENV{ 'HOME' } , 'NVIDIA_GPU_Computing_SDK' );
    }

    if( ! -d "$nvsdkcuda_root" ) {
      $nvsdkcuda_root = '';
    }
  }
  elsif( $OSNAME eq 'MSWin32' ) {
    if( $cuda_toolkit_version eq '2.2' ) {
      if( exists( $ENV{ 'NVSDKCUDA_ROOT' } ) ) {
        $nvsdkcuda_root = $ENV{ 'NVSDKCUDA_ROOT' };
      
        if( -d "$nvsdkcuda_root" ) {
          return $nvsdkcuda_root;
        }
        else {
          $nvsdkcuda_root = '';
        }
      }
    }
    elsif( is_cuda_toolkit_version_2_3( $cuda_toolkit_version ) ) {
      if( exists( $ENV{ 'NVSDKCOMPUTE_ROOT' } ) ) {
        $nvsdkcuda_root = $ENV{ 'NVSDKCOMPUTE_ROOT' };
      
        if( -d "$nvsdkcuda_root" ) {
          return $nvsdkcuda_root;
        }
        else {
          $nvsdkcuda_root = '';
        }
      }
    }      
  }
  return $nvsdkcuda_root;
}

sub get_default_cuda_root {
  my $cuda_root = '';
  my $cuda_bin_path;
  
  if( $OSNAME eq 'linux' ) {
    if( -d '/usr/local/cuda' ) {
      $cuda_root = '/usr/local/cuda';
    }
  }
  elsif( $OSNAME eq 'MSWin32' ) {
    if( exists( $ENV{ 'CUDA_BIN_PATH' } ) ) {
      $cuda_bin_path = $ENV{ 'CUDA_BIN_PATH' };
    
      $cuda_root = dirname( $cuda_bin_path );
    }
  }

  return $cuda_root;
}

sub create_nvsdk_bjam_build_env {
  my ( $nvsdkcuda_root, $cuda_root, $cuda_toolkit_version, $BoostBuildDir ) = @_;
  my $nvsdk_common_src_dir;
  my $nvsdk_shared_src_dir;
  my $nvsdk_bjam_template_dir;
  my $nvsdk_jamroot;
  my $nvsdk_jamroot_final;
  
  my $nvsdk_jamfile_cutil;
  my $nvsdk_jamfile_cutil_dest;
  
  my $nvsdk_jamfile_shrUtils;
  my $nvsdk_jamfile_shrUtils_dest;
  
  my $nvsdk_boost_build_jam;
  my $nvsdk_boost_build;
  my $fh;
  my $fh_jamroot;
  my $BoostBuildDirNormalized;
  my $line;
  
  $nvsdk_bjam_template_dir = File::Spec->catdir( $BoostBuildDir, 'src', 'templates', 'BuildCudaSdkLib' );
  
  $nvsdk_jamroot = File::Spec->catfile( $nvsdk_bjam_template_dir, 'Jamroot.tmpl' );
  if( ! -f "$nvsdk_jamroot" ) {
    fatal_error( __LINE__, "the file \"$nvsdk_jamroot\" does not exist\n" );
  }
  
  $nvsdk_jamfile_cutil = File::Spec->catfile( $nvsdk_bjam_template_dir, 'cutil', 'Jamfile.v2.tmpl' );
  if( ! -f "$nvsdk_jamfile_cutil" ) {
    fatal_error( __LINE__, "the file \"$nvsdk_jamfile_cutil\" does not exist\n" );
  }  
  
  $nvsdk_boost_build_jam = File::Spec->catfile( $nvsdkcuda_root, 'boost-build.jam' );

  if( $cuda_toolkit_version eq '2.2' )
  {
    $nvsdk_common_src_dir = File::Spec->catdir( $nvsdkcuda_root, 'common', 'src' );
  }
  else
  {
    $nvsdk_common_src_dir = File::Spec->catdir( $nvsdkcuda_root, 'C', 'common', 'src' );
    $nvsdk_shared_src_dir = File::Spec->catdir( $nvsdkcuda_root, 'shared', 'src' );
    
    $nvsdk_jamfile_shrUtils = File::Spec->catfile( $nvsdk_bjam_template_dir, 'shrUtils', 'Jamfile.v2.tmpl' );
    if( ! -f "$nvsdk_jamfile_shrUtils" ) {
      fatal_error( __LINE__, "the file \"$nvsdk_jamfile_shrUtils\" does not exist\n" );
    }  
    
    $nvsdk_jamfile_shrUtils_dest = File::Spec->catfile( $nvsdk_shared_src_dir, 'Jamfile.v2' );
    
  }
  if( ! -d "$nvsdk_common_src_dir" ) {
    fatal_error( __LINE__, "the directory \"$nvsdk_common_src_dir\" does not exist." );
  }
  
  $nvsdk_jamfile_cutil_dest = File::Spec->catfile( $nvsdk_common_src_dir, 'Jamfile.v2' );
  
  
  #$nvsdk_boost_build
  $BoostBuildDirNormalized = $BoostBuildDir;
  if( $OSNAME eq 'MSWin32' ) {
    $BoostBuildDirNormalized =~ s/\\/\//g ; 
  }
  $BoostBuildDirNormalized .= '/boost-build';
  
  #
  $fh = new FileHandle $nvsdk_boost_build_jam, "w";
  if( ! defined( $fh ) ) {
    fatal_error( __LINE__, "cannot open \"$nvsdk_boost_build_jam\" for writing\n" );
  }
  
  print $fh "boost-build \"$BoostBuildDirNormalized\" ;";
  
  #close the file
  undef $fh;
  copy( $nvsdk_jamfile_cutil, $nvsdk_jamfile_cutil_dest ) or die "Copy( $nvsdk_jamfile_cutil, $nvsdk_jamfile_cutil_dest ) failed: $!";

  if( is_cuda_toolkit_version_2_3( $cuda_toolkit_version ) ) {
    copy( $nvsdk_jamfile_shrUtils, $nvsdk_jamfile_shrUtils_dest ) or die "Copy( $nvsdk_jamfile_shrUtils, $nvsdk_jamfile_shrUtils_dest ) failed: $!";
  }
  
  $nvsdk_jamroot_final = File::Spec->catfile( $nvsdkcuda_root, 'Jamroot' );
  
  $fh_jamroot = new FileHandle $nvsdk_jamroot_final, "w";
  if( ! defined( $fh_jamroot ) ) {
    fatal_error( __LINE__, "cannot open \"$nvsdk_jamroot_final\" for writing\n" );
  }
 
  $fh = new FileHandle $nvsdk_jamroot, "r";
  if( ! defined( $fh ) ) {
    fatal_error( __LINE__, "cannot open \"$nvsdk_jamroot\" for reading\n" );
  }

  while( ! $fh->eof() ) {
    $line = $fh->getline();
    chomp( $line );
    if( $line =~ /^constant CUDA_VERSION/ ) {
      print $fh_jamroot "constant CUDA_VERSION : $cuda_toolkit_version ;\n";
    }
    else {
      print $fh_jamroot "$line\n";
    }
  }

  undef $fh;
  undef $fh_jamroot;
  


  print "\n";
  print "==============================================================\n";
  print "Finished setting up bjam environment to build NVSDK libraries.\n";
  print "==============================================================\n";
  print "\n"
}

sub get_home_dir {
  my ( $tool_dir ) = @_;
  my $home_dir;

  if( $OSNAME eq 'MSWin32' ) {
    if( ! exists( $ENV{ 'USERPROFILE' } ) ) {
      fatal_error( __LINE__, "failed to determine your home directory; make sure \%USERPROFILE\% is defined.\n" );  
    }
    $home_dir = $ENV{ 'USERPROFILE' };
  }
  else {
    if( ! exists( $ENV{ 'HOME' } ) ) {
      fatal_error( __LINE__, "failed to determine your home directory; make sure \$HOME is defined.\n" );  
    }
    $home_dir = $ENV{ 'HOME' };
  }  
  
  if( ! -d "$home_dir" ) {
    fatal_error( __LINE__, "the home directory \"$home_dir\" does not exist, check your environment.\n" );  
  }
  
  return $home_dir;
}

sub get_default_bjam_dir {
  my ( $tool_dir, $bIsPlatform64Bit ) = @_;
  my $home_dir;
  my $bjam_bin_dir;
  
  $home_dir = get_home_dir( $tool_dir );
  
#  if( ( $OSNAME eq 'MSWin32' ) || ( $OSNAME eq 'darwin' ) ) {
#    $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin' );
#  }

  if( $OSNAME eq 'darwin'  ) {
    $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin' );
  }
  elsif( $OSNAME eq 'MSWin32' ) {
    if( $bIsPlatform64Bit ) {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'x86_64' );
    }
    else {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'x86' );
    }    
  }
  elsif( $OSNAME eq 'linux' ) {
    # I only support intel x86/amd64 platforms
    if( $bIsPlatform64Bit ) {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'linux', 'x86_64' );
    }
    else {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'linux', 'x86' );
    }
  }
  elsif( $OSNAME eq 'solaris' ) {
    # I only support intel x86/amd64 platforms
    if( $bIsPlatform64Bit ) {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'solaris', 'x86_64' );
    }
    else {
      $bjam_bin_dir = File::Spec->catdir( $home_dir, 'bin', 'solaris', 'x86' );
    }
  } 
  else {
    die "get_default_bjam_dir: unsupported platform encountered.";
  }
  
  return $bjam_bin_dir;
}

#  find  the boost build directory given the script name 
sub find_boost_build_dir {
  my ( $script ) = @_;
  my $boost_build_dir;
  my $filename;
  my $suffix;
  my $script_dir;
  my $no_file = 1;
  my $volume;
  my $directories;
  my @dir_list;
  
  ( $filename, $script_dir, $suffix ) = fileparse( $script );
  
  $script_dir = File::Spec->rel2abs( $script_dir );
  
  ( $volume, $directories, $filename ) = File::Spec->splitpath( $script_dir, $no_file );

  @dir_list = File::Spec->splitdir( $directories );
  
  #drop the last path component
  pop @dir_list;
  
  # reassemble the path to get the boost build directory
  $directories = File::Spec->catdir( @dir_list );

  $boost_build_dir = File::Spec->catpath( $volume, $directories, '' );
  
  return $boost_build_dir;
}

sub find_bjam {
  my ( $r_bjam_prog, $bjam_dir, $tool_dir, $bIsPlatform64Bit ) = @_;
  my $bjam_prog;
  my @search_path;
  my $dir;
  
  if( $bjam_dir eq '' ) {
    $bjam_dir = get_default_bjam_dir( $tool_dir, $bIsPlatform64Bit );
    
    if( $OSNAME eq 'MSWin32' ) {
      $bjam_prog = File::Spec->catfile( $bjam_dir, 'bjam.exe' );  
    }
    else {
      $bjam_prog = File::Spec->catfile( $bjam_dir, 'bjam' );  
    }
    
    if( -f "$bjam_prog" ) {
      # found bjam in the default directory.
      $$r_bjam_prog = $bjam_prog;
      return;
    }    
  }
  else {
    # the user has provided a bjam path that has to be used (no search)
    if( $OSNAME eq 'MSWin32' ) {
      $$r_bjam_prog = File::Spec->catfile( $bjam_dir, 'bjam.exe' );  
    }
    else {
      $$r_bjam_prog = File::Spec->catfile( $bjam_dir, 'bjam' );  
    }
    
    if( ! -f "$$r_bjam_prog" ) {
      fatal_error( __LINE__, "the bjam executable \"$$r_bjam_prog\" does not exist.\n" );
    }
    return;
  }
  
  # second approach: search the path
  @search_path = File::Spec->path();
  
  foreach $dir (@search_path) {
    #assemble bjam
    if( $OSNAME eq 'MSWin32' ) {
      $bjam_prog = File::Spec->catfile( $dir, 'bjam.exe' );  
    }
    else {
      $bjam_prog = File::Spec->catfile( $dir, 'bjam' );  
    }
    
    if( -f "$bjam_prog" ) {
      $$r_bjam_prog = $bjam_prog;
      return;
    }
  }
  
  fatal_error( __LINE__, "bjam executable not found on the search path, consider to use --bjam-dir or\nyou may need to build it from source (=> boost_install.pl)\n" );
}

sub is_platform_64bit {
  my $b_platform_64bit = 0;

  if( $OSNAME eq 'linux' ) {
    $b_platform_64bit = is_linux_64bit();
  }
  elsif( $OSNAME eq 'MSWin32' ) {
    $b_platform_64bit = is_windows_64bit();
  }
  else {
    fatal_error( __LINE__, "is_platform_64bit() does not support $OSNAME yet.\n" );
  }
  
  return $b_platform_64bit;
}


sub is_linux_64bit {
  my $linux_x64 = 0;

  if( $OSNAME eq 'linux' ) {
    my $uname_m;
    
    $uname_m = `uname -m`;

    if( $uname_m =~ /^x86_64/ ) { 
        $linux_x64 = 1;
    }
  }

  return $linux_x64;
}

sub is_windows_64bit {
  my $windows_64bit = 0;
  my $proc_arch;
  
  if( $OSNAME ne 'MSWin32' ) {
    die "is_windows_64bit(): called on a non-windows platform";
  }  

  if( ! exists( $ENV{'PROCESSOR_ARCHITECTURE'} ) ) {
    die "is_windows_64bit(): the environment variable PROCESSOR_ARCHITECTURE does not exist\n";
  }
  
  $proc_arch = $ENV{'PROCESSOR_ARCHITECTURE'};
  if( $proc_arch =~ /AMD64/ ) {
    $windows_64bit = 1;
  }
  
  return $windows_64bit;
}

sub usage {

  my $how_to_call = <<"__END_OF_USAGE__";
Usage: cuda_install.pl [options]

  options:=
     --help              show the usage information
     
    --cudasdk-dir=<s>   The installation directory of the CUDA SDK. The user needs write permission to be 
                        able to build the SDK libraries.
                        Defaults: 
                            LINUX:    
                              CUDA 2.2:   \$HOME/NVIDIA_CUDA_SDK 
                              CUDA 2.3:   \$HOME/NVIDIA_CPU_Computing_SDK

                            WINDOWS:  
                              CUDA 2.2:   \%NVSDKCUDA_ROOT\%       (created by the NVIDIA installer!)
                              CUDA 2.3:   \%NVSDKCOMPUTE_ROOT\%    (created by the NVIDIA installer!)
                         
                         The special value none can be used to disable building the SDK libraries.
                         
     --cuda-dir=<s>      The installation directory of the CUDA toolkit.
                         Defaults: 
                            LINUX:    /usr/local/cuda
                            WINDOWS:  \%CUDA_BIN_PATH\%\..       (created by the NVIDIA installer!)
                             
     --toolset=<s>       use the non-default toolset <s> to build the libraries. The default 
                         toolset is platform and sometimes installation dependent. For the time
                         being it is highly recommended to use msvc-9.0 on WINDOWS as NVIDIDA use the
                         same toolset to develop the CUDA toolkit and SDK.
                                             
                         example:
                         use "--toolset=msvc-8.0" if msvc-9.0 and msvc-8.0 are installed
                         and you would like to build the msvc-8.0 compliant libraries.
                                                            
     --jobs=<n>          number of parallel bjam jobs to compile the SDK [default: 1]

     --bjam-dir=<s>      The bjam installation directory. 
     
                         Defaults:
                         
                         WINDOWS 32 bit:   %USERPROFILE%\\bin\\x86
                         WINDOWS 64 bit:   %USERPROFILE%\\bin\\x86_64
                         
                         LINUX 32 bit:     \$HOME/bin/linux/x86
                         LINUX 64 bit:     \$HOME/bin/linux/x86_64
                         
                         All plaforms:  The search path if the default directory does not include a bjam
                                        executable.
 
__END_OF_USAGE__

  print STDERR $how_to_call;
}

#
# issue an error message and terminate the program
#
sub fatal_error {
  my ($line, $msg) = @_;

  #if( defined( $::g_fh_error ) ) {
  #  print $::g_fh_error  "error: $msg";
  #}
  #else {
    print STDERR  "cuda_install.pl: error at line $line: $msg";
  #}

  die ;
}

sub test_configure_cuda_manpath {
  my ( $cuda_root ) = @_;

  if( $OSNAME ne 'MSWin32' ) {
    &configure_cuda_manpath( $cuda_root );
  }
  else {
    print "test_configure_cuda_manpath is not supported on windows\n";
  }
}

