#!/usr/bin/perl
#
# get_boost_version.pl
#

use FileHandle;

use File::Basename;
use File::Spec;
use File::Path;
use English;
use Getopt::Long;

use strict 'vars';

my $boost_root;
my $boost_jamroot;
my $fd_jamroot;
my $line;
my $boost_version = '';
my $boost_version_major;
my $boost_version_minor;
my $boost_version_build;

my $boost_version_found = 0;
my $bListUsage = 0;
my $bListMajorVersion = 0;
my $bListMinorVersion = 0;
my $bListBuildNumber = 0;

my %optctl = (
  'help'     => \$bListUsage,
  'major'   => \$bListMajorVersion,
  'minor'   => \$bListMinorVersion,
  'build'   => \$bListBuildNumber
);

# parse the command line 
if( GetOptions( \%optctl, 'help!', 'major!', 'minor!', 'build!' ) != 1 ) {
  print "error: command line not understood, check the syntax\n";
  die;
}

if( $bListUsage ) {
  print "get_boost_version [options] [boost_dir]\n";
  print "  --help\n";
  print "  --major     list major version number\n";
  print "  --minor     list minor version number\n";
  print "  --build     list build number\n";
  print "\n";
  exit 0;
}


if( scalar(@ARGV) > 1 ) {
  usage();
  exit( 1 );
}
elsif( scalar( @ARGV ) == 1 ) {
  # boost root specified
  $boost_root = $ARGV[ 0 ];
}
else {
  # autodection of boost_root via the default Jamroot given the script name
  my $top_dir = get_top_dir( $0 );
  my $jamroot = File::Spec->catfile( $top_dir, 'Jamroot' );
  
  $boost_root = get_boost_root_from_jamroot( $jamroot );
}


if( ! -d $boost_root ) {
  die "error: boost source tree \"$boost_root\" not found";
}

$boost_jamroot = File::Spec->catfile( $boost_root, 'Jamroot' );

if( ! -f $boost_jamroot ) {
  die "error: the file \"$boost_jamroot\" does not exist.";
}

$fd_jamroot = new FileHandle $boost_jamroot, "r";
if( ! defined( $fd_jamroot ) ) {
  die "error: cannot open \"$boost_jamroot\" for reading.";
}

while( defined( $fd_jamroot ) && ( ! $fd_jamroot->eof() ) ) {
  $line = $fd_jamroot->getline;
  chomp( $line );
  
  # constant BOOST_VERSION : 1.39.0
  
  if( $line =~ /^constant\s*BOOST_VERSION\s*:\s*([0-9.]+)/ ) {
    $boost_version = $1;
    #print "line: $line   \"$boost_version\" \n";
    
    $boost_version_found = 1;
    undef $fd_jamroot;
  }
}

if( defined( $fd_jamroot ) ) {
  # close the file handle if the version expression was not found.
  $fd_jamroot->close();
}

if( ! $boost_version_found ) {
  die "error: failed to parse boost version from Jamroot";  
}


if( ( $bListMajorVersion != 0 ) || ( $bListMinorVersion != 0 ) || ( $bListBuildNumber != 0 ) ) {
  my $boost_version_major;
  my $boost_version_minor;
  my $boost_version_build;

  if( $boost_version =~ /([0-9]+)[.]([0-9]+)[.]([0-9]+)/ ) {
    $boost_version_major = $1;
    $boost_version_minor = $2;
    $boost_version_build = $3;
  
    print "$boost_version_major" if $bListMajorVersion;
    print "$boost_version_minor" if $bListMinorVersion;
    print "$boost_version_build" if $bListBuildNumber;
  }
  else {
    print "error: boost version \"$boost_version\" not understood\n";
    die;
  }
}
else {
  print "$boost_version";
}

sub get_top_dir {
  my ( $script ) = @_;
  my $top_dir;
  my $filename;
  my $suffix;
  my $script_dir;
  my $no_file = 1;
  my $volume;
  my $directories;
  my @dir_list;
  
  ( $filename, $script_dir, $suffix ) = fileparse( $script );
  
  $script_dir = File::Spec->rel2abs( $script_dir );
  
  ( $volume, $directories, $filename ) = File::Spec->splitpath( $script_dir, $no_file );

  @dir_list = File::Spec->splitdir( $directories );
  
  #e.g. dir_list = BoostBuild/bin/linux 
  #drop the last two path components
  if( scalar( @dir_list ) < 3 ) {
    die "illegal script directory encountered, please check \"$script_dir\" and its position within the project directory.";
  }
  
  pop @dir_list;
  pop @dir_list;
  
  # reassemble the path to the top directory; i.e. the default jamroot directory
  $directories = File::Spec->catdir( @dir_list );

  $top_dir = File::Spec->catpath( $volume, $directories, '' );
  
  return $top_dir;
}

sub get_boost_root_from_jamroot {
  my ( $Jamroot ) = @_;
  my $boost_root_dir;
  my $boost_root_found = 0;
  my $fh;
  my $line;
  my $jamroot_dir;
  
  if( ! -f $Jamroot ) {
    die "the file \"$Jamroot\" does not exist";
  }

  $fh = new FileHandle $Jamroot, "r";
  if( ! defined( $fh ) ) {
    die "cannot open \"$Jamroot\" for reading.";
  }

  while( defined( $fh ) && ( ! $fh->eof() ) ) {
    $line = $fh->getline;
    chomp( $line );
  
    # path-constant BOOST_SDK_ROOT : ../Boost_1_40_0 ;
    if( $line =~ /^\s*path-constant\s*BOOST_SDK_ROOT\s*:\s*([^;]+)/ ) {
      $boost_root_dir = $1;
      
      $boost_root_dir =~ s/[ \t]*//g;
      if( $OSNAME eq 'MSWin32' ) {
        $boost_root_dir =~ s/\//\\/g;
      }      
      $boost_root_found = 1;
      undef $fh;
    }
  } 

  if( defined( $fh ) ) {
    # close the file handle if the version expression was not found.
    $fh->close();
  }

  die "the file \"$Jamroot\" does not define path to boost" if( ! $boost_root_found );
 
  if( ! File::Spec->file_name_is_absolute( $boost_root_dir ) ) {
    $jamroot_dir = dirname( $Jamroot );
    $boost_root_dir = File::Spec->catdir( $jamroot_dir, $boost_root_dir );
  }
  
  return File::Spec->rel2abs( $boost_root_dir );
}
  
