#!/usr/bin/perl
#
# get_version.pl
# 
# A perl script to extract the version ID out of
# the version header file found in <repository>/include/<libname>/version.h
# 
use File::Basename;
use File::Copy;
use File::Spec;
use FileHandle;
use DirHandle;
use Getopt::Long;
use English;

use strict 'vars';

# use  $bUpdateRcFile = 0 if you want to disable updating the rc file on windows. 
my $bUpdateRcFile = 1;
my $bListUsage = 0;
my $bVerbatim = 0;
my $bListVersionId = 1;
my $bIsWindows = ( $OSNAME eq 'MSWin32' ) ? 1 : 0;
my $bStripBuildId = 1;
my $rc_file = "";
my $bVerbose = 0;
my $top_dir;
my $default_version_header_file;
my $version_header_file = "";
my $libname;
my $lib_version;
my @dir_list;

#print "args => @ARGV\n";

my %optctl = (
  'help'     => \$bListUsage,
  'verbatim' => \$bVerbatim,
  'rc-update' => \$bUpdateRcFile,
  'list-version' => \$bListVersionId,
  'rc-file' => \$rc_file,
  'version-header' => \$version_header_file
);

# parse the command line 
if( GetOptions( \%optctl, 'help!', 'verbatim!', 'rc-update!', 'list-version!', 'rc-file=s', 'version-header=s' ) != 1 ) {
  fatal_error( "command line not understood, check the syntax\n" );
}

if( $bListUsage != 0 ) {
  &show_help;
  exit 0;
}

if( $bIsWindows == 0 ) {
  #never update rc files on non windows systems
  $bUpdateRcFile = 0;
  $rc_file = "";
}

# check command line arguments

if( scalar(@ARGV) == 0 ) {
  #msvc resource file update triggered by pre build event 

  if( $rc_file ne '' ) {
    my $rc_file_dir;
    my $suffix;
    my $filename;
    
    if( ! -f $rc_file ) {
      fatal_error( "The specified resource file $rc_file does not exist\n" );
    }

    if( $version_header_file eq '' ) {
      #derive a default version header file given the resource file
      ( $filename, $rc_file_dir, $suffix ) = fileparse( $rc_file );

      #get the last path component
      @dir_list = File::Spec->splitdir( $rc_file_dir );

      #fileparse returns rc_file_dir with a trailing path separator, hence we get one more empty list element and have to 
      #index the second element from the end of the list to get the library name.
      $libname = $dir_list[ $#dir_list - 1 ];
      
      #print "rc_dir => $rc_file_dir dir_list => $#dir_list  @dir_list $dir_list[ 3 ] \n";
      $version_header_file = File::Spec->catfile( $rc_file_dir, '..', '..', '..', 'include', $libname, 'version.h' );
      
      if( ( ! -f $version_header_file ) && ( $#dir_list >= 4 ) ) {
        # try the default version header file
        $libname = $dir_list[ $#dir_list - 4 ];
        #print "libname => $libname\n";
        $version_header_file = File::Spec->catfile( $rc_file_dir, '..', '..', '..', 'include', $libname, 'version.h' );
      }
    }
    
    if( -f $version_header_file ) {
      &get_lib_version( \$lib_version, $version_header_file, $bVerbatim, $rc_file, 0 );
    }
  }
  elsif( $version_header_file ne '' ) {
    #special support for installer scripts to get a specific version ID
    # examples.  
    # get_version.pl --verbatim --version-header=MSysLib/include/MSysLib/version.h => version ID with dots
    # get_version.pl --version-header=MSysLib/include/MSysLib/version.h => version ID with underscores
    &get_lib_version( \$lib_version, $version_header_file, $bVerbatim, $rc_file, 1 );
    print $lib_version;
  }
  else {
    usage();
    exit 1;
  }
}
else {
  my @top_dir_list = @ARGV;
  my @libdir_list;
  my $lib_dir;
  my $repository;
  my $library_version_info = '';

  foreach $top_dir (@top_dir_list) {
    # top_dir = MSysLib                    =>  default_version_header_file MSysLib/include/MSysLib/version.h
    #                                          libdir_list => MSysLib/src/lib/MSysLib MSysLib/src/lib/ToolLib MSysLib/src/lib/CpuInfoLib
    #
    # top_dir = MSysLib/src/lib/ToolLib    => default_version_header_file MSysLib/include/MSysLib/version.h
    #                                         libdir_list => MSysLib/src/lib/ToolLib
    # 
    $lib_dir = File::Spec->catfile( $top_dir, 'src', 'lib' );
    if( -d $lib_dir ) {
      # $top_dir refers to a repository folder with a standard project folder layout, scan any folder in $top_dir/src/lib and 
      # assume it's a library with version information.
      $default_version_header_file = File::Spec->catfile( $top_dir, 'include', $top_dir, 'version.h' );
      @libdir_list = get_lib_dir_list( File::Spec->catfile( $top_dir, 'src', 'lib' ) );
    }
    else {
      # $top_dir refers to a library folder; e.g. MSysLib/src/lib/ToolLib
      
      # get the default header file
      @dir_list = File::Spec->splitdir( $top_dir );
      if( $#dir_list >= 3 ) {
        $repository = $dir_list[ $#dir_list - 3 ];
        $default_version_header_file = File::Spec->catfile( $repository, 'include', $repository, 'version.h' );
      }
      
      # prepare a one element list
      @libdir_list = ();
      push @libdir_list, $top_dir;
    }
    
    foreach $lib_dir (@libdir_list) {

      #check for version header
      # example:  MSysLib/src/lib/MSysLib => MSysLib/include/MSysLib/version.h

      #get the last path component
      @dir_list = File::Spec->splitdir( $lib_dir );
      
      $libname = $dir_list[ $#dir_list ];
      $version_header_file = File::Spec->catfile( $lib_dir, '..', '..', '..', 'include', $libname, 'version.h' );

      if( ! -f $version_header_file ) {
        # no library specific version information available, use the default version header
        $version_header_file = $default_version_header_file;
      }

      if( -f $version_header_file ) {
        #print "version header: $version_header_file\n";
        
        if( $bUpdateRcFile != 0 ) {
          # check for the default resource file
          $rc_file = File::Spec->catfile( $lib_dir, $libname . '.rc' );
          if( ! -f $rc_file ) {
            $rc_file = "";
          }
        }
        #print "rc_file => $rc_file\n" ;

        &get_lib_version( \$lib_version, $version_header_file, $bVerbatim, $rc_file );
        
        if( $library_version_info ne '' )
        {
          $library_version_info .= ' ';
        }
        $library_version_info .= "${libname}/$lib_version";
      }
    }
  }
  
  if( $bListVersionId ) {
    print $library_version_info;
  }
}

sub get_lib_dir_list {
  my ( $lib_dir_top ) = @_;
  my @lib_dir_list;
  my @dir_list_tmp;
  my $dirh;
  my $filesys_comp;
  my $lib_dir;
  my $f;
  
  if ( -d $lib_dir_top ) {
    #enumerate all sub directories in $lib_dir_top
    
    $dirh = new DirHandle $lib_dir_top;
    if( defined $dirh ) {
      while( defined( $filesys_comp = $dirh->read ) ) { 
        push @dir_list_tmp, $filesys_comp;
        #print "filesys_comp => $filesys_comp\n";
      }
      undef $dirh;
      
      #filter out files and hidden directories
      foreach $f (@dir_list_tmp) {
        if( $f !~ /^\./ ) {
          # the file system component does not start with a dot
          $lib_dir = File::Spec->catfile( $lib_dir_top, $f ) ;
          #print "$lib_dir\n";
          
          if( -d $lib_dir ) {
            #print "lib_dir => $lib_dir\n";
            push @lib_dir_list, $lib_dir;
          }
        }
      }
    }
  }
  return @lib_dir_list;
}

sub get_lib_version {
  my ( $r_lib_version, $header_file, $bVerbatim, $rc_file ) = @_;
  my $filename;
  my $header_file_dir;
  my $suffix;

  ( $filename, $header_file_dir, $suffix ) = fileparse( $header_file );

  #get the last path component
  my @dir_list = File::Spec->splitdir( $header_file_dir );
  my $version_include_dir;
  my $version_define_tag;
  my $version_id;
  my $windows_rc_version_id;
  my $boost_version_id;

  $version_include_dir = $dir_list[ $#dir_list - 1 ];

  $version_define_tag = uc( $version_include_dir );
  $version_define_tag .= '_VERSION';

  $version_id = extract_header_file_version( $header_file, $version_define_tag );

  if( ( $rc_file ne '' ) && -f $rc_file ) {

    #print "res file: $windows_res_filename \n" ;

    $windows_rc_version_id = get_windows_rc_version_id( $version_id );
    #print "windows version: $windows_res_version_id <= $version_id \n";
 
    &replace_windows_rc_version( $rc_file, $windows_rc_version_id );
  }

  $boost_version_id = $version_id ;
  $boost_version_id =~ s/\./_/g;
  $boost_version_id =~ s/(.*)-[0-9]+$/$1/;
  
  #Don't add an end of line to please the bjam tagging rule.
  if ( 0 != $bVerbatim ) {
    $$r_lib_version = $version_id;
  }
  else {
    $$r_lib_version = $boost_version_id;
  }
}

sub extract_header_file_version {
  my ( $header_file, $version_define_tag ) = @_;
  my $version;
  my $fh;
  my $preproc;
  my $ver;
  my $version_number = '0.0.0';
  my $line;
  
  $fh = new FileHandle $header_file, "r";
  if ( ! defined $fh ) {
    fatal_error( "cannot open $header_file for reading\n" );
  }

  while( defined( $fh ) && ( ! $fh->eof() ) ) {
    $line = $fh->getline();
    chomp;
    
    if ( $line =~ /^#define[ \t]*${version_define_tag}/ ) {
      ( $preproc, $ver, $version_number ) = split( '[ \t"]+', $line );
      undef $fh;
    }
  }

  if( defined( $fh ) ) {
    $fh->close();
  }

  return $version_number ;
}


sub get_windows_rc_version_id {
  my ( $version_id ) = @_;
  my $version_string;
  my $version_id1 = 0;
  my $version_id2 = 0;
  my $version_id3 = 0;
  my $version_id4 = 0;
  my $windows_version_string;
  
  $version_string = $version_id ;
  $version_string =~ s/_/./g;
  $version_string =~ s/-/./g;

  ( $version_id1, $version_id2, $version_id3, $version_id4 ) = split( /\./, $version_string );

  if( $version_id1 !~ /^[0-9]+$/ ) {
    #not a number, reset to zero
    $version_id1 = 0;
  }
  
  if( $version_id2 !~ /^[0-9]+$/ ) {
    #not a number, reset to zero
    $version_id2 = 0;
  }
  
  if( $version_id3 !~ /^[0-9]+$/ ) {
    #not a number, reset to zero
    $version_id3 = 0;
  }
  
  if( $version_id4 !~ /^[0-9]+$/ ) {
    #not a number, reset to zero
    $version_id4 = 0;
  }

  $windows_version_string = sprintf( '%u,%u,%u,%u', $version_id1, $version_id2, $version_id3, $version_id4 );

  return $windows_version_string ;
}


sub replace_windows_rc_version {
  my ( $rc_file, $version_string ) = @_;
  my $old_version_string;
  my $version_id1 = 0;
  my $version_id2 = 0;
  my $version_id3 = 0;
  my $version_id4 = 0;
  my $fh_in;
  my $fh_out;
  my $rc_file_tmp;
  my $bSaveTmpFile = 0;
  my $bInsideVersionInfo = 0;
  my $BeginEndBlockCnt = 0;
  my $line;
  my $version_string_unquoted;
  my $version_string_quoted;
  my $res_token;

  ( $version_id1, $version_id2, $version_id3, $version_id4 ) = split( /,/, $version_string );
  
  $version_string_quoted = sprintf( '"%u, %u, %u, %u"', $version_id1, $version_id2, $version_id3, $version_id4 );

  $fh_in = new FileHandle $rc_file, "r";
  if ( ! defined $fh_in ) {
    fatal_error( "cannot open $rc_file for reading\n" );
  }

  $rc_file_tmp = $rc_file . '.tmp' ;
  $fh_out = new FileHandle $rc_file_tmp, "w";
  if ( ! defined $fh_out ) {
    fatal_error( "cannot open $rc_file_tmp for writing\n" );
  }

  while ( <$fh_in> ) {
    chomp;
    $line = $_;

    if ( /^VS_VERSION_INFO/ ) {
      $bInsideVersionInfo = 1;
      $BeginEndBlockCnt = 0;
    }
    elsif ( $bInsideVersionInfo == 1 ) {
      
      if ( /^[ \t]*FILEVERSION\s/ ) {
        s/^[ \t]*//g;
        ( $res_token, $old_version_string ) = split( /[ \t]+/ );
        if( $old_version_string ne $version_string ) {
          #print "Updating $rc_file : FILEVERSION has been changed\n";
          $bSaveTmpFile = 1;
          $line = sprintf( ' FILEVERSION %s', $version_string );
        }
      }
      elsif ( /^[ \t]*PRODUCTVERSION\s/ ) {
        s/^[ \t]*//g;
        ( $res_token, $old_version_string ) = split( /[ \t]+/ );
        if( $old_version_string ne $version_string ) {
          #print "Updating $rc_file : PRODUCTVERSION has been changed\n";
          $bSaveTmpFile = 1;
          $line = sprintf( ' PRODUCTVERSION %s', $version_string );
        }
      }
      elsif ( /^[ \t]*VALUE[ \t]+"FileVersion",/ ) {
        ( $res_token, $old_version_string ) = split( /,/, $_, 2 );
        $old_version_string =~ s/^[ \t]+//;
        if( $old_version_string ne $version_string_quoted ) {
          #print "Updating $rc_file : FileVersion has been changed\n";
          $bSaveTmpFile = 1;
          $line = sprintf( '%s, %s', $res_token, $version_string_quoted );
        }
      }
      elsif ( /^[ \t]*VALUE[ \t]+"ProductVersion",/ ) {
        #    VALUE "ProductVersion", "0, 167, 0, 0"
        ( $res_token, $old_version_string ) = split( /,/, $_, 2 );
        $old_version_string =~ s/^[ \t]+//;
        if( $old_version_string ne $version_string_quoted ) {
          # print "Updating $rc_file : ProductVersion has been changed, '$old_version_string' '$version_string_quoted'\n";
          $bSaveTmpFile = 1;
          $line = sprintf( '%s, %s', $res_token, $version_string_quoted );
        }
      }
      elsif ( /^[ \t]*BLOCK/ ) {
        $BeginEndBlockCnt++;
      }
      elsif ( /^[ \t]*END/ ) {
        $BeginEndBlockCnt--;
        if( $BeginEndBlockCnt == 0 ) {
          #end of version resource section
          $bInsideVersionInfo = 0;
        }
      }
    }
    print $fh_out "${line}\n";
  }

  $fh_in->close();
  $fh_out->close();

  if( $bSaveTmpFile != 0 ) {
    #print "Updating $rc_file \n";
    copy( $rc_file_tmp, $rc_file ) or die "Copy( $rc_file_tmp, $rc_file ) failed: $!";
  }
  unlink( $rc_file_tmp );
}

sub show_help {
  print "\n";
  
  system( "perldoc $0" );
  if( $? != 0 ) {
    #die "error: 'perldoc $0' failed, make sure perl-doc is installed\n";
    exit 1; # LINUX: a sensible error message is displayed by default
  }
}

#
# Print usage information and exit.
#
sub usage {

  print "get_version.pl: error: illegal command line\n\n";
  
  &show_help;
  
}

#
# issue an error message and terminate the program
#
sub fatal_error {
  my ($msg) = @_;

  #if( defined( $::g_fh_error ) ) {
  #  print $::g_fh_error  "error: $msg";
  #}
  #else {
    print STDERR  "get_version: error: $msg";
  #}

  die ;
}

sub test_get_windows_rc_version_id {
   my $windows_rc_version_id;

   $windows_rc_version_id = get_windows_rc_version_id( '1.2.3-4' );
   print "test: windows version: $windows_rc_version_id <= 1.2.3-4 \n";
    
   $windows_rc_version_id = get_windows_rc_version_id( '1.2.3.4.5' );
   print "test: windows version: $windows_rc_version_id <= 1.2.3.4.5 \n";

   $windows_rc_version_id = get_windows_rc_version_id( '1' );
   print "test: windows version: $windows_rc_version_id <= 1 \n";

   $windows_rc_version_id = get_windows_rc_version_id( '1.2b' );
   print "test: windows version: $windows_rc_version_id <= 1.2b \n";

   $windows_rc_version_id = get_windows_rc_version_id( '1.2-b' );
   print "test: windows version: $windows_rc_version_id <= 1.2-b \n";
}

__END__

=pod

=head1 NAME

get_version.pl - script to extract a version number from a C/C++ header file 

=head1 SYNOPSIS

B<get_version.pl> [options] dir1...

B<get_version.pl> B<--version-header> F<file> B<--rc-file> F<file>

B<get_version.pl> B<--help>


=head1 DESCRIPTION

The script B<get_version.pl> extracts a version number from a C/C++ 
header file and prints it to stdout or updates WINDOWS resource files. 

=head1 OPTIONS

=over 2

=item B<--help> 
 
 Show documentation

=item B<--verbatim>

 Returns the version ID as found in the version header
 file. The option is off by default to return the 
 default boost library tag, which uses underscores instead of 
 dots to separate the version ID components.

=item B<--version-header> F<filename>

 Overwrites the default version header file with filename

=item B<--rc-file> F<filename>

 Updates the given resource file with the version ID found
 in the version header file.

=item B<--no-rc-update>

 Disables updating the WINDOWS resource file.

=back
 
=head1 EXAMPLES

=over 2

=item B<get_version.pl> F<MSysLib DataPacketLib>
 
 Extracts version information for all libraries found
 in MSysLib/src/lib and DataPacketLib/src/lib and
 update resource files on WINDOWS.
  
=item B<get_version.pl> --rc-file $(ProjectDir)$(ProjectName).rc
 
 Updates the resource file using the MSVC pre build event.

=back
                                         
=cut

