#!/usr/bin/perl
#
# add_pkg_prj.pl
#

use File::Basename;
use File::Copy;
use File::Spec;
use File::Path;

use DirHandle;
use FileHandle;

use Getopt::Long;
use English;

use strict 'vars';

my @todo_list;
my $jamroot_dir;
my $template_dir;
my $template_dir_makefile;
my $template_dir_debian;
my $jamroot_file;
my $output_dir = '';
my $b_show_usage = 0;
my $pkg_name;
my $pkg_dir;
my $pkg_dir_debian;
my $top_dir_rel;

my @pkg_project_files = (
  #'control.input',
  'hhi-package.install.input',
  'hhi-package.links',
  'rules'
);
my %pkg_filename_map;


my %optctl = (
  'help'           => \$b_show_usage,
  'output-dir'     => \$output_dir,
);

if( $OSNAME ne 'linux' ) {
  fatal_error( __LINE__, "Unsupported platform detected, please run this script on a LINUX system.\n" );
}

# parse the command line 
if( GetOptions( \%optctl, 'help!', 'output-dir=s' )  != 1 ) {
  fatal_error( __LINE__, "command line not understood, check the syntax\n" );
}

if( $b_show_usage ) {
  usage();
  exit( 0 );
}

if( scalar( @ARGV ) != 1 ) {
  usage();
  exit( 1 );
}

$pkg_name = $ARGV[ 0 ];

# find "BoostBuild" given the script name
$jamroot_dir = find_jamroot_dir( $0 );
# print "jamroot_dir: $jamroot_dir\n";
$jamroot_file = File::Spec->catfile( $jamroot_dir, 'Jamroot' );

if( ! -f $jamroot_file ) {
  fatal_error( __LINE__, "the file \"$jamroot_file\" does not exist\n" );
}


$template_dir = File::Spec->catdir( $jamroot_dir, 'BoostBuild', 'src', 'templates', 'deb_create_pkg_stub', 'hhi-package' );
$template_dir_makefile = File::Spec->catdir( $jamroot_dir, 'BoostBuild', 'src', 'templates', 'deb_create_pkg_stub' );
$template_dir_debian = File::Spec->catdir( $template_dir, 'debian' );


if( $output_dir eq '' ) {
  print STDERR "No package project directory specified, please use option \"--output-dir=\" to set the packge project directory.\n";
  usage();
  exit( 1 );
} 

if( ! -d $output_dir ) {
  mkpath( $output_dir, 1, 0755 );
  if( ! -d $output_dir ) {
    fatal_error( __LINE__, "the package project directory \"$output_dir\" cannot be created, check path and permissions\n" );
  }
}

$pkg_dir = File::Spec->catdir( $output_dir, $pkg_name );
$pkg_dir_debian = File::Spec->catdir( $pkg_dir, 'debian' );

if( -d $pkg_dir ) {
  fatal_error( __LINE__, "the package directory \"$pkg_dir\" already exists.\n" );
}
else {
  mkpath( $pkg_dir_debian, 1, 0755 );
  if( ! -d $pkg_dir_debian ) {
    fatal_error( __LINE__, "the package project directory \"$pkg_dir_debian\" cannot be created, check path and permissions\n" );
  }
}

# TOP will be dereferenced by the debian rules file to get to the top level directory. It has to be
# a relative path to please the debhelper scripts.
$top_dir_rel = get_top_dir_relative( $jamroot_dir, $pkg_dir );

$pkg_filename_map{ 'hhi-package.install.input' } = $pkg_name . '.install.input';
$pkg_filename_map{ 'hhi-package.links' } = $pkg_name . '.links';

copy_pkg_files( $template_dir_debian, $pkg_dir_debian, \@pkg_project_files, \%pkg_filename_map );

copy_pkg_control_file( $pkg_name, 'control.input', $template_dir_debian, $pkg_dir_debian );

compose_todo_list( $pkg_dir_debian, $pkg_name, \@todo_list );
copy_upstream_makefile( $pkg_name, 'Makefile', $template_dir_makefile, $output_dir, $top_dir_rel, \@todo_list );

if( scalar( @todo_list )  > 0 ) {
  my $line;
  
  print "\nTODO:\n\n";
  
  foreach $line ( @todo_list ) {
    print "$line\n";
  }
}

sub compose_todo_list {
  my ( $pkg_dir_debian, $pkg_name, $rl_todo_list ) = @_;
  my $msg;
  
  
  $msg = "- A debian package project has been created in \"$pkg_dir_debian\", please review the following files:\n";
  $msg .= "  control.input\n";
  $msg .= "  $pkg_name.install.input\n";
  $msg .= "  $pkg_name.links\n";
  $msg .= "  rules\n";
  
  push @$rl_todo_list, $msg;
  
  $msg = "- Add package files to $pkg_name.install.input\n";
  push @$rl_todo_list, $msg;
  
  $msg = "- Add symbolic links to $pkg_name.links and enable dh_link in rules\n";
  push @$rl_todo_list, $msg;
  
  if( ! -f '/usr/bin/dh_install' ) {
    $msg = "- Install package debhelper\n";
    push @$rl_todo_list, $msg;
  }
  
  $msg  = "- Modify Jamroot to include targets for package building if not already present:\n";
  $msg .= "  1. import notfile ;\n";
  $msg .= "  2. Add target to build the package:\n";
  $msg .= "     notfile pkg : \@build-pkg ;\n";
  $msg .= "     explicit pkg ;\n\n";
  $msg .= "     actions build-pkg\n";
  $msg .= "     {\n";
  $msg .= "       ./BoostBuild/bin/unix/make_pkg.pl --arch=i386 --bin-dir=\$(inst-bin-dir)/release-static $output_dir \n";
  $msg .= "     }\n";
  $msg .= "  3. The package architecture is required for cross compiling, otherwise optional.\n";
  $msg .= "     It is recommended to derive the package architecture using the rule hhitagging.get-target-platform.\n";
  $msg .= "  4. A Jamroot example with package building targets can be found in BoostBuild; i.e. check out BoostBuild as a standalone project.\n.";
  
  push @$rl_todo_list, $msg;
  
  $msg =  "- How to build packages with bjam:\n";
  $msg .= "  bjam -q pkg                      # to build native packages\n";
  $msg .= "  bjam -q pkg address-model=32     # to build i386 packages on the amd64 platform\n";
  $msg .= "\n";
  $msg .= "  The package files can be found under build/deb/x86 or build/deb/x86_64\n";
  
  push @$rl_todo_list, $msg;
}

sub copy_upstream_makefile {
  my ( $pkg_name, $makefile, $from_dir, $to_dir, $top_dir, $rl_todo_list ) = @_;
  my $output_makefile = File::Spec->catfile( $to_dir, $makefile );
  my $input_makefile;
  my $fh_in;
  my $fh_out;
  my $line;
  
  if( -f $output_makefile ) {
    # we don't replace any existing upstream makefile
    push @$rl_todo_list, "- check $output_makefile and add $pkg_name to the list of package directories\n";
    return;
  }
  
  $input_makefile = File::Spec->catfile( $from_dir, $makefile );
  
  $fh_in = new FileHandle $input_makefile, "r";
  if( ! defined( $fh_in ) ) {
    fatal_error( __LINE__, "cannot open input file \"$input_makefile\"\n" );
  }
  
  $fh_out = new FileHandle $output_makefile, "w";
  if( ! defined( $fh_out ) ) {
    fatal_error( __LINE__, "cannot open output file \"$output_makefile\"\n" );
  }  
  
  while( ! $fh_in->eof() ) {
    $line = $fh_in->getline;
    chomp( $line );
    if( $line =~ /^SUBDIRS\s*:=/ ) {
      $line = "SUBDIRS := $pkg_name";
    }
    elsif( $line =~ /^TOP\s*:=/ ) {
      $line = "TOP := $top_dir";
    }
    print $fh_out "$line\n";
  } 
  
  undef $fh_in;
  undef $fh_out;
}

sub copy_pkg_control_file {
  my ( $pkg_name, $control_file, $from_dir, $to_dir ) = @_;
  my $fh_in;
  my $fh_out;
  my $line;
  my $input_file = File::Spec->catfile( $from_dir, $control_file );
  my $output_file = File::Spec->catfile( $to_dir, $control_file );
  
  $fh_in = new FileHandle $input_file, "r";
  if( ! defined( $fh_in ) ) {
    fatal_error( __LINE__, "cannot open input file \"$input_file\"\n" );
  }
  
  $fh_out = new FileHandle $output_file, "w";
  if( ! defined( $fh_out ) ) {
    fatal_error( __LINE__, "cannot open output file \"$output_file\"\n" );
  }
  
  while( ! $fh_in->eof() ) {
    $line = $fh_in->getline;
    chomp( $line );
    if( $line =~ /^Package:/ ) {
      $line = "Package: $pkg_name";
    }
    print $fh_out "$line\n";
  } 
  
  undef $fh_in;
  undef $fh_out;
}



sub copy_pkg_files {
  my ( $from_dir, $to_dir, $rl_project_files, $rh_filename_map ) = @_;
  my $src_file;
  my $dest_file;
  my $file;
  
  foreach $file ( @$rl_project_files ) {
    $src_file = File::Spec->catfile( $from_dir, $file );
    if( exists( $rh_filename_map->{ $file } ) ) {
      $dest_file = File::Spec->catfile( $to_dir, $rh_filename_map->{ $file } );
    }
    else {
      $dest_file = File::Spec->catfile( $to_dir, $file );
    }
    #print "cp $src_file $dest_file\n"; 
    copy( $src_file, $dest_file );
    
    if( $file eq 'rules' ) {
      # rules must be executable but the perl copy function does not preserve the mode.
      system( "chmod 0755 $dest_file" );
      if( $? != 0 ) {
        fatal_error( __LINE__, "the command system( \"chmod 0755 $dest_file\" ) failed.\n" );
      }
    }
  }  
}

sub find_jamroot_dir {
  my ( $script ) = @_;
  my $jamroot_dir;
  my $filename;
  my $suffix;
  my $script_dir;
  my $no_file = 1;
  my $volume;
  my $directories;
  my @dir_list;

  ( $filename, $script_dir, $suffix ) = fileparse( $script );
  
  $script_dir = File::Spec->rel2abs( $script_dir );
  
  ( $volume, $directories, $filename ) = File::Spec->splitpath( $script_dir, $no_file );

  @dir_list = File::Spec->splitdir( $directories );
  
  # drop the last two path components
  pop @dir_list;
  pop @dir_list;
  pop @dir_list;
  
  # reassemble the path to get the boost build directory
  $directories = File::Spec->catdir( @dir_list );

  $jamroot_dir = File::Spec->catpath( $volume, $directories, '' );
  
  return $jamroot_dir;
}

sub get_top_dir_relative {
  my ( $top_dir_abs, $pkg_dir ) = @_;
  my $top_dir_rel;
  
  die if( ! File::Spec->file_name_is_absolute( $top_dir_abs ) ); 
 
  $top_dir_rel = File::Spec->abs2rel( $top_dir_abs, $pkg_dir );
  
  return $top_dir_rel;
}


#
# issue an error message and terminate the program
#
sub fatal_error {
  my ($line, $msg) = @_;

  #if( defined( $::g_fh_error ) ) {
  #  print $::g_fh_error  "error: $msg";
  #}
  #else {
    print STDERR  "add_pkg_prj.pl: error at line $line:  $msg";
  #}

  exit( 1 );
}

sub usage {
  my $how_to_call = <<"__END_OF_USAGE__";
Usage: add_pkg_prj.pl [options] package_name

options:=
     --help              show the usage information
     
     --output-dir        destination directory to save debian package projects
                         
                         example: 
                         
                           project_dir/installer/linux
                           
example:
  
  deb_add_pkg.pl --output=project_dir/installer/linux hhi-test1
                           
                                
__END_OF_USAGE__

  print STDERR $how_to_call;
}
