#!/usr/bin/perl
#
# deb_control_filter.pl
#
# This script supplements the debhelper script suite with advanced variable expansion in
# debian package scripts like the master control file and <package>.install or <package.links> scripts.
#
# For further information, please refer to the builtin help and the accompanying documentation.
#
use Getopt::Long;

use File::Basename;
use File::Spec;
use FileHandle;
use IO::Handle;
use DirHandle;

use English;

use strict 'vars';

my $b_help = 0;
my $b_keep_comments = 0;
my $output_file = '-';
my $input_file = '-';
my $num_args;
my $pkg_save_dir = '';
my $top_dir;
my $binary_pkg_arch = '';
my $pkg_arch;
my $platform_suffix;
my $msg;
my $builtin_test = '';

# a map of variable substitution pairs; e.g. BIN_DIR => ../bin/gcc-4.3.3/release
my %var_subst_map;

# a map of builtin tests
%::g_builtin_test_map = (
  'subst_vars'                => \&test_subst_vars,
  'get_pkg_distrib_suffix'    => \&test_get_pkg_distrib_suffix
);

my %known_pkg_arch_map = (
  'i386'  => 1,
  'amd64' => 1
);

my %optctl = (
  'help'               => \$b_help,
  'output-file'        => \$output_file,
  'keep-comments'      => \$b_keep_comments,
  'pkg-save-dir'       => \$pkg_save_dir,
  'arch'               => \$pkg_arch,
  'binary-arch'        => \$binary_pkg_arch,
  'test-case'          => \$builtin_test
);

# parse the command line 
if( GetOptions( \%optctl, 
                'help!', 
                'output-file=s', 
                'keep-comments!', 
                'pkg-save-dir=s', 
                'arch=s', 
                'binary-arch=s', 
                'test-case=s' ) != 1 ) {
  usage();
  exit 1;
}

if( $b_help ) {
  usage();
  exit 0;
}

if( $builtin_test ne '' ) {
  if( exists( $::g_builtin_test_map{ $builtin_test } ) ) {
    my $r_sub = $::g_builtin_test_map{ $builtin_test };
    $r_sub->();
    exit 0;
  }
  else {
    fatal_error( __LINE__, "\"$builtin_test\" is an unknown test case, check option \"--test-case\"\n" );
  }
}

$top_dir = get_top_dir( $0 );

if( $pkg_arch eq '' ) {
  # get the native package architecture i386 or amd64 as the default
  $pkg_arch = get_native_pkg_arch();
}

if( $pkg_arch ne 'all' ) {
  # the normal situation for a binary package is binary_pkg_arch = pkg_arch
  if( $binary_pkg_arch eq '' ) {
    $binary_pkg_arch = $pkg_arch;
  }
}

if( $binary_pkg_arch eq 'i386' ) {
  $platform_suffix = 'x86';
}
elsif( $binary_pkg_arch eq 'amd64' ) {
  $platform_suffix = 'x86_64';
}
else {
  fatal_error( __LINE__, "the package architecture \"$binary_pkg_arch\" is not supported yet.\n" );
}

if( $pkg_save_dir eq '' ) {
  $pkg_save_dir = File::Spec->catdir( $top_dir, 'build', 'deb', $platform_suffix );
}

if( ! -d $pkg_save_dir ) {
  fatal_error( __LINE__, "the package file directory \"$pkg_save_dir\" does not exist.\n" );
}


$num_args = scalar( @ARGV );

if( $num_args  < 1 ) {
  usage();
  exit 1;
}

$input_file = $ARGV[ 0 ];

if( "$input_file" ne '-' ) {
  if( ! -f "$input_file" ) {
    fatal_error( __LINE__, "the input file \"$input_file\" does not exist.\n" );
  }
}

# check for var=value expression arguments
for ( my $i = 1; $i < $num_args; $i++ ) {
  my $var;
  my $value;
  my $var_value = $ARGV[ $i ];
  
  if( $var_value !~ /=/ ) {
    fatal_error( __LINE__, "variable substitution expression \"$var_value\" is illegal, check the command line.\n" );
  }
  
  ( $var, $value ) = split /=/, $var_value;
  
  if( exists( $var_subst_map{ $var } ) ) {
    fatal_error( __LINE__, "variable \"$var\" specified twice, check the command line.\n" );
  }
  
  $var_subst_map{ $var } = $value;
} 

if( ! exists( $var_subst_map{ 'ARCH' } ) ) {
  # the variable ARCH has not been provided by the caller, add the predefined value (native architecture) 
  $var_subst_map{ 'ARCH' } = $pkg_arch;
}
elsif( $var_subst_map{ 'ARCH' } ne $pkg_arch ) {
  
  $msg = "ARCH=$var_subst_map{ 'ARCH' } conflicts with $pkg_arch, the current value of the package architecture.\n";
  $msg .= "Consider to use --arch=$var_subst_map{ 'ARCH' }\n";
  
  fatal_error( __LINE__, $msg );
}


{
  my $fh_in;
  my $fh_out;
  my $pkg_distrib_suffix;
  my @control_lines_in;
  my @control_lines_out;
  my $line;
  my $is_master_control_file = 0;
   
  if( "$input_file" eq '-' ) {
    $fh_in = IO::Handle->new_from_fd( 0, 'r' );
    
    if( ! defined( $fh_in ) ) {
      fatal_error( __LINE__, "cannot get handle of <stdin> for reading?\n" );
    }    
  }
  else {
    $fh_in = new FileHandle $input_file, "r";
    if( ! defined( $fh_in ) ) {
      fatal_error( __LINE__, "cannot open input file \"$input_file\"\n" );
    }   
  }
 
  if( $output_file eq '-' ) {
    $fh_out = IO::Handle->new_from_fd( 1, 'w' );
    
    if( ! defined( $fh_out ) ) {
      fatal_error( __LINE__, "cannot get handle of <stdout> for writing?\n" );
    } 
  }
  else {
    $fh_out = new FileHandle $output_file, "w";
    
    if( ! defined( $fh_out ) ) {
      fatal_error( __LINE__, "cannot open output file \"$output_file\"\n");
    }
  }
  
  while( defined( $fh_in ) && ( ! $fh_in->eof() ) ) {
    $line = $fh_in->getline;
    chomp( $line );
 
    if( $line =~ /^\s*#/ ) {
      if( $b_keep_comments ) {
        push @control_lines_in, $line;
      }
    }
    else {
      if( $line =~ /^\s*Package:/ ) {
        # the input looks like a debian master control file, need to filter depends fields!
        $is_master_control_file = 1;
      }
     
      # not a comment line, keep it
      push @control_lines_in, $line;
    }
  }
  
  undef $fh_in;
  
  # substitute any $VAR=value expressions
  subst_vars( \@control_lines_in, \@control_lines_out, \%var_subst_map );
  
  if( $is_master_control_file ) {
    # clear control_lines_in to receive the output of the next transformation
    @control_lines_in = ();
  
    # select depends field 
    select_best_matched_depends_field( \@control_lines_out, \@control_lines_in, $binary_pkg_arch, $pkg_save_dir, \%known_pkg_arch_map );
    
    foreach $line (@control_lines_in) {
      print $fh_out "$line\n";
    }    
  } 
  else {
    foreach $line (@control_lines_out) {
      print $fh_out "$line\n";
    }     
  }

  undef $fh_out;
}

sub get_native_pkg_arch {
  my $uname_resp;

  $uname_resp = `dpkg --print-architecture`;
  if( $? != 0 ) {
    fatal_error( __LINE__, "the command \"dpkg --print-architecture\" failed\n" );
  }
  
  #get rid of any line feeds.
  chomp $uname_resp;

  return $uname_resp;
}

sub subst_vars {
  my ( $rl_lines_in, $rl_lines_out, $rh_var_subst_map ) = @_;
  my $line;
  my $transformed_line;
  my $dollar_var;

  foreach $line (@$rl_lines_in) {
    if( $line =~ /^\s*#/ ) {
      push @$rl_lines_out, $line;
    }
    else {
      $transformed_line = $line;
      while( $transformed_line =~ /\$([^ ,:;{}()\/\t\$]+)/ ) {
        $dollar_var = $1;
        
        if( ! exists( $$rh_var_subst_map{ $dollar_var } ) ) {
          fatal_error( __LINE__, "the input contains the undefined variable \"\$$dollar_var\", please check input and command line.\n" );
        }
        else {
          $transformed_line =~ s/\$$dollar_var/$$rh_var_subst_map{ $dollar_var }/;
        }
        #print "var: \"$dollar_var\"\n";
      }
      
      # handle ${var} expression; e.g. lib${suffix}.a
      while( $transformed_line =~ /\$\{([^ ,:;}()\/\t\$]+)\}/ ) {
        $dollar_var = $1;
        
        if( ! exists( $$rh_var_subst_map{ $dollar_var } ) ) {
          fatal_error( __LINE__, "the input contains the undefined variable \"\$\{$dollar_var\}\", please check input and command line.\n" );
        }
        else {
          $transformed_line =~ s/\$\{$dollar_var\}/$$rh_var_subst_map{ $dollar_var }/;
        }
      }      
      
      push @$rl_lines_out, $transformed_line;
    }        
  }
}


sub select_best_matched_depends_field {
  my ( $rl_lines_in, $rl_lines_out, $binary_pkg_arch, $pkg_save_dir, $rh_known_pkg_arch_map ) = @_;
  my $pkg_distrib_suffix = get_pkg_distrib_suffix();
  my $line;
  my $selected_depends_field;
  my $selected_depends_line_value;
  my $depends_tag;
  my $depends_value;
  my %depends_line_map;
  my $pkg_arch = 'x';
  my $b_depends_fmt_with_arch;
  my $b_depends_field_written = 0;
  
  #print "package suffix: $pkg_distrib_suffix\n";
  
  foreach $line (@$rl_lines_in) {
    if( $line =~ /^(Depends[^:]*):(.+)$/ ) {
      $depends_tag = $1;
      $depends_value = $2;
      
      $depends_line_map{ $depends_tag } = $depends_value;
    }
    elsif( $line =~ /^Architecture:\s*(\w+)\s*$/ ) {
      $pkg_arch = $1;
    } 
  }
  
  # validate the package architecture, a mandatory field
  if( $pkg_arch eq 'x' ) {
    fatal_error( __LINE__, "the input file does not contain the package architecture field, please check.\n" );
  }
  
  if( scalar( keys( %depends_line_map ) ) == 0 ) {
    # this is valid, no depends line given and we just need to copy the control lines unmodified
    @$rl_lines_out = @$rl_lines_in;
    
    return;
  }  
  
  # We support two flavours of depends line alternatives
  #
  # Combination 1 does not include the package architecture
  #
  # Depends:
  # Depends_ubuntu-hardy:
  # Depends_ubuntu-jaunty:
  #
  # Combination 2 includes the package architecture
  #
  # Depends_i386:
  # Depends_amd64:
  # Depends_ubuntu-hardy_i386:
  # Depends_ubuntu-hardy_amd64:
  #
  
  $b_depends_fmt_with_arch = analyze_depends_lines( \%depends_line_map, $rh_known_pkg_arch_map );
  
  #print "depends with arch: $b_depends_fmt_with_arch \n";
  if( $pkg_arch eq 'all' ) {
    # platform independent packages are special because their architecture cannot be used to select the best
    # dependency field.
    #
    # example:
    #    
    #  - a script package depends on a binary package with different package names on i386 and amd64 platforms.
    #  - in this situation you can use the following depends fields:
    #    
    #    Depends_i386: packageA
    #    Depends_amd64: packageB
    # 
    $selected_depends_field = select_depends_field( \%depends_line_map, $pkg_distrib_suffix, $binary_pkg_arch, $b_depends_fmt_with_arch );
  }
  else {
    $selected_depends_field = select_depends_field( \%depends_line_map, $pkg_distrib_suffix, $pkg_arch, $b_depends_fmt_with_arch );
  }
  
  # update missing package versions in depends field
  $selected_depends_line_value = update_depends_pkg_versions( $depends_line_map{ $selected_depends_field }, $pkg_save_dir );
  
  # final step: dump the filtered control file
  foreach $line (@$rl_lines_in) {
    if( $line =~ /^(Depends[^:]*):(.+)$/ ) {
      if( ! $b_depends_field_written ) {
        # first depends line encountered, dump the selected dependency information
        push @$rl_lines_out, "Depends: $selected_depends_line_value";
        $b_depends_field_written = 1;
      }
    }
    else {
      push @$rl_lines_out, "$line";
    }
  }
}


sub select_depends_field {
  my ( $rh_depends_line_map, $pkg_distrib_suffix, $pkg_arch, $b_depends_fmt_with_arch ) = @_;
  my $depends_field;
  my @depends_field_list;
  
  if( $b_depends_fmt_with_arch ) {
    # we start with the most specific depends field
    
    $depends_field = 'Depends_' . $pkg_distrib_suffix . '_' . $pkg_arch;
    push @depends_field_list, $depends_field;
    
    $depends_field = 'Depends_' . $pkg_arch;
    push @depends_field_list, $depends_field;
  }
  else {
    # we start with the most specific depends field; e.g. Depends_ubuntu-hardy
    $depends_field = 'Depends_' . $pkg_distrib_suffix;
    push @depends_field_list, $depends_field;
    
    $depends_field = 'Depends';
    push @depends_field_list, $depends_field;    
  }
  
  while( scalar( @depends_field_list ) > 0 ) {
    $depends_field = shift @depends_field_list;
    if( exists( $$rh_depends_line_map{ $depends_field } ) ) {
      return $depends_field;
    }
  }
  
  # no matching depends field found, exit
  fatal_error( __LINE__, "no matching depends field found, please check the control file.\n" );  
}

sub analyze_depends_lines {
  my ( $rh_depends_line_map, $rh_known_pkg_arch_map ) = @_;
  my $b_depends_fmt_with_arch = 0;
  my $depends_tag;
  my $depends_last_component;
  my @depends_tag_components;
  
  foreach $depends_tag (keys %$rh_depends_line_map) {
    #print "depends_tag: $depends_tag\n";
    
    @depends_tag_components = split /_/, $depends_tag;
    if( scalar( @depends_tag_components ) > 1 ) {
      $depends_last_component = pop @depends_tag_components;
      if( exists( $$rh_known_pkg_arch_map{ $depends_last_component } ) ) {
        $b_depends_fmt_with_arch = 1;
      } 
    }    
  }
  
  if( $b_depends_fmt_with_arch ) {
    # apply consistency check to reject the combination of depends lines with and without archicture suffix
    
    foreach $depends_tag (keys %$rh_depends_line_map) {
      #print "depends_tag: $depends_tag\n";
    
      @depends_tag_components = split /_/, $depends_tag;
      if( scalar( @depends_tag_components ) > 1 ) {
        $depends_last_component = pop @depends_tag_components;
        if( exists( $$rh_known_pkg_arch_map{ $depends_last_component } ) ) {
          $b_depends_fmt_with_arch = 1;
        } 
        else {
          fatal_error( __LINE__, "the dependency field $depends_tag cannot combined with dependency fields with an architecture suffix, please check the control file\n" );
        }
      }
      else {
        fatal_error( __LINE__, "the dependency field $depends_tag cannot combined with dependency fields with an architecture suffix, please check the control file\n" );
      }    
    }    
  }
  
  
  return $b_depends_fmt_with_arch;
}

sub read_pkg_version_ids_from_dir {
  my ( $pkg_save_dir, $rh_pkg_version_id_map ) = @_;
  my $dirh;
  my $filesys_comp;
  my @dir_list;
  my @deb_file_list = ();
  my $filename;
  
  
  $dirh = new DirHandle $pkg_save_dir;
  if( ! defined $dirh ) {
    fatal_error( __LINE__, "cannot open directory \"$pkg_save_dir\" for reading." );
  }
  
  while( defined( $filesys_comp = $dirh->read ) ) { 
    #need to save the file system component temporarily and perform the file type check later on because
    #some platforms do not allow to check the file type while the directory is being searched.
    #print "filesys_comp => $filesys_comp\n";

    if( $filesys_comp =~ /.*\.deb$/i ) { 
      push @dir_list, $filesys_comp;
    }
  }
  undef $dirh;

  return @deb_file_list if( scalar( @dir_list ) == 0 );

  #found some file system components ending with .deb but this may be a directory!
  foreach $filesys_comp (@dir_list) {

    $filename = File::Spec->catfile( $pkg_save_dir, $filesys_comp );
    if( -f $filename ) {
      # found a deb file
      my $pkg_filename;
      my @pkg_filename_components;

      # strip extension
      $pkg_filename = $filesys_comp;
      $pkg_filename =~ s/\.deb$//;
      
      # $pkg_filename => pkgname_version_codename_arch
      #               => pkgname_version_arch         
      @pkg_filename_components = split /_/, $pkg_filename;
      
      if( $#pkg_filename_components < 2 ) {
        fatal_error( __LINE__, "the naming of package file \"$filename\" is not understood.\n" );
      }
      
      if( exists( $rh_pkg_version_id_map->{ $pkg_filename_components[ 0 ] } ) ) {
        fatal_error( __LINE__, "the package file \"$filename\" seems to contain a package, which is also stored in a different file in the same directory.\n" );
      }
      
      # add the package version ID to the hash.
      $rh_pkg_version_id_map->{ $pkg_filename_components[ 0 ] } = $pkg_filename_components[ 1 ]; 
    }
  }
}

sub update_depends_pkg_versions {
  my ( $depends_line, $pkg_save_dir ) = @_;
  my $depends_line_transformed = '';
  my @depends_line_token_list;
  my $depends_token;
  my %pkg_version_id_map;
  my $msg;
  
  # read package version IDs
  read_pkg_version_ids_from_dir( $pkg_save_dir, \%pkg_version_id_map );
  
  @depends_line_token_list = split /,/, $depends_line;
  
  foreach $depends_token (@depends_line_token_list) {
    my $depends_or_expr = '';
    my $depends_token2;
    
    my @depends_token_or_expr_list;
    
    $depends_token =~ s/^\s+//;
    $depends_token =~ s/\s+$//;
    
    @depends_token_or_expr_list = split /\|/, $depends_token;
    foreach $depends_token2 (@depends_token_or_expr_list) {
      my $version_expr;
      my @version_expr_list;
      my $depends_token2_transformed = '';
      my $pkg_name;
      
      $depends_token2 =~ s/^\s+//;
      $depends_token2 =~ s/\s+$//;
      
      if( $depends_token2 =~ /^([^(]+)\(([^()]+)\)/ ) {
        my $num_version_components;
        $pkg_name = $1;
        $version_expr = $2;
        
        # strip leading and trailing whitespace
        $version_expr =~ s/^\s+//;
        $version_expr =~ s/\s+$//;
        
        #print "$depends_token2 => $pkg_name($version_expr)\n";
        
        # validate version expression:
        @version_expr_list = split /\s+/, $version_expr;
        
        # a complete version expression must consist of two elements: the comparision operator and the version ID.
        $num_version_components = scalar( @version_expr_list );
        if( $num_version_components == 2 ) {
          $depends_token2_transformed = $pkg_name . '(' . $version_expr . ')';
        }
        elsif( $num_version_components == 1 ) {
          if( ! exists( $pkg_version_id_map{ $pkg_name } ) ) {
            $msg = "the directory \"$pkg_save_dir\" does not contain package $pkg_name.\n";
            $msg .= "Please check the sequence of package creation and the command line.\n";
            
            fatal_error( __LINE__,  $msg );
          }
          
          # $version_expr contains the comparision operator
          $depends_token2_transformed = $pkg_name . "($version_expr " . $pkg_version_id_map{ $pkg_name } . ')';
        } 
      }
      else {
        # no version ID supplied, keep the package name as-is.
        $depends_token2_transformed = $depends_token2;
      }
      if( $depends_or_expr ne '' ) {
        $depends_or_expr .= '|';
      }
      $depends_or_expr .= $depends_token2_transformed;
    }
    if( $depends_line_transformed ne '' ) {
      $depends_line_transformed .= ', ';
    }
    $depends_line_transformed .= $depends_or_expr;
  }
  
  return $depends_line_transformed;
}

sub get_pkg_distrib_suffix {
  my $pkg_distrib_suffix;
  my $version_file = '/etc/lsb-release';
  my $fh;
  my $distrib_id = 'x';
  my $codename = 'x';
  my $tag;
  my $value;
  my $line_cnt = 0;
  my $line;
  
  if( ! -f $version_file ) {
    fatal_error( __LINE__, "the version file \"$version_file\" does not exist.\n" );
  }

  $fh = new FileHandle $version_file, "r";
  if ( ! defined $fh ) {
    fatal_error( __LINE__, "cannot open \"$version_file\" for reading.\n" );
  }

  while ( defined( $fh ) && ( ! $fh->eof() ) ) {
    $line = $fh->getline;
    chomp( $line );
    $line_cnt++;

    if ( $line =~ /^DISTRIB_ID=/ ) {
      ( $tag, $value ) = split '=', $line;
      $distrib_id = lc( $value ) ;
    }
    elsif( $line =~ /^DISTRIB_CODENAME=/ ) {
      ( $tag, $value ) = split '=', $line;
      $codename = lc( $value ) ;
    }
  }

  undef $fh;

  if( $distrib_id eq 'x' ) {
    fatal_error( __LINE__, "DISTRIB_ID not found in version file \"$version_file\"\n" );
  }
  if( $codename eq 'x' ) {
    fatal_error( __LINE__, "DISTRIB_CODENAME not found in version file \"$version_file\"\n" );
  }


  $pkg_distrib_suffix = $distrib_id . '-' . $codename;
  
  return $pkg_distrib_suffix;
}

sub get_top_dir {
  my ( $script ) = @_;
  my $top_dir;
  my $filename;
  my $suffix;
  my $script_dir;
  my $no_file = 1;
  my $volume;
  my $directories;
  my @dir_list;
  
  ( $filename, $script_dir, $suffix ) = fileparse( $script );
  
  $script_dir = File::Spec->rel2abs( $script_dir );
  
  ( $volume, $directories, $filename ) = File::Spec->splitpath( $script_dir, $no_file );

  @dir_list = File::Spec->splitdir( $directories );
  
  #e.g. dir_list = BoostBuild/bin/linux 
  #drop the last two path components
  if( scalar( @dir_list ) < 4 ) {
    fatal_error( __LINE__, "illegal script directory encountered, please check \"$script_dir\" and its position within the project directory.\n" );
  }
  pop @dir_list;
  pop @dir_list;
  pop @dir_list;
  
  # reassemble the path to get the boost build directory
  $directories = File::Spec->catdir( @dir_list );

  $top_dir = File::Spec->catpath( $volume, $directories, '' );
  
  return $top_dir;
}

#
# issue an error message and terminate the program
#
sub fatal_error {
  my ($line, $msg) = @_;

  #if( defined( $::g_fh_error ) ) {
  #  print $::g_fh_error  "error: $msg";
  #}
  #else {
    print STDERR  "deb_control_filter.pl: error at line $line: $msg";
  #}

  exit 1;
}

sub usage {
  my $how_to_call;
  my $test_case_list;
  my $key;
  
  foreach $key ( sort keys( %::g_builtin_test_map ) ) {
    $test_case_list .= "                          $key\n";
  } 
  
  
$how_to_call = <<"__END_OF_USAGE__";
Usage: deb_control_filter.pl [options] input_file [var=value ... ]

arguments:=

    input_file          the input_file to be processed

    var=value           defines a custom variable substition 
                        If the input_file contains \$var, this variable will be 
                        replaced by value. 
  
options:=

    --help              show the usage information
    
    --arch=s            the package architecture as understood by the debian package system.
                        
                        default: i386 or amd64 
                        
                        The value of --arch is automatically associated with the variable \$ARCH and can 
                        be used in the architecture field of the control file.
                        
    --binary-arch=s     the binary architecture to be used for depends field selection if and 
                        only if the package architecture is platform independent.
                        
    --output-file=s     the output file to receive the transformed content 
    
                        default: stdout
    
    --pkg-save-dir=s    an absolute directory to save the package files. The defaults are usually 
                        sufficient for standard Jamroots. 
                        
                        default:
                          arch=i386   => <TOP>/build/deb/x86
                          arch=amd64  => <TOP>/build/deb/x86_64
                        
    --test-case=s       select a builtin test case:
    
$test_case_list          
                          
Examples:
    
    # transforms the input for the i386 architecture
    deb_control_filter.pl --arch=i386 --output-file=control control.input BIN_DIR=../../bin/gcc-4.3.3/x86/release
    
    # transforms the input for the amd64 architecture
    deb_control_filter.pl --arch=amd64 --output-file=control control.input BIN_DIR=../../bin/gcc-4.3.3/x86_64/release


__END_OF_USAGE__

  print STDERR $how_to_call;
}

#############################################################################################################
#                                         builtin tests
#############################################################################################################


sub test_subst_vars {
  my @lines_in;
  my @lines_out;
  my $line;
  my $line_out;
  my $var;
  my %var_subst_map = (
   'BIN_DIR'    => '../bin/x86', 
   'ARCH'       => 'i386',
   'LIB1'       => 'sample-gcc43-1',
   'LIB3'       => 'sample-gcc43-3',
   'LIB4'       => 'sample-gcc43-4',
   'LIB5'       => 'sample-gcc43-5'
  );
  
  print "===============================================================================================\n";
  print "testing function subst_vars\n\n";
  
  print "Listing variable map\n";
  foreach $var ( sort keys( %var_subst_map ) ) {
    print "$var=$var_subst_map{ $var }\n";
  }
  print "\n";
  
  push @lines_in, '$BIN_DIR/release/prog_sample';
  push @lines_in, '/opt/$BIN_DIR/release/prog_sample';
  push @lines_in, '/opt/${BIN_DIR}/release/prog_sample';
  push @lines_in, 'Architecture: $ARCH';
  push @lines_in, 'Depends: lib1${LIB1}.a,lib2,lib3${LIB3}-d.so,lib4${LIB4},lib5$LIB5,lib6';
  
  subst_vars( \@lines_in, \@lines_out, \%var_subst_map );
  
  foreach $line (@lines_in) {
    $line_out = shift @lines_out;
    print "$line\n$line_out\n";
  }
  print "===============================================================================================\n";
}

sub test_get_pkg_distrib_suffix {
  my $distrib_suffix;

  print "===============================================================================================\n";
  print "testing function get_pkg_distrib_suffix\n\n";
  
  $distrib_suffix = get_pkg_distrib_suffix();
  
  print "$distrib_suffix\n";
  print "===============================================================================================\n";
}