#!/usr/bin/perl
#
# get_pkg_distrib_suffix.pl
# 
# A script to return a linux distribution specific suffix that
# may be used for packaging.
#
use English;
use File::Basename;
use FileHandle;
use Getopt::Long;

use strict 'vars';

$::g_script_error = "get_pkg_distrib_suffix.pl: error:";

# e.g. ubuntu-hardy
my $pkg_distrib_suffix = '';
my $b_newline = 0; 

my $script_dir;
my $system_info;

my %optctl = (
  'newline'  => \$b_newline
);

if( $OSNAME ne 'linux' ) {
  die "$::g_script_error the platform \"$OSNAME\" is not recognized yet.";
}

if( GetOptions( \%optctl, 'newline!' ) != 1 ) {
  die "$::g_script_error command line not understood, check the syntax.";
}

$script_dir = get_script_dir( $0 );

$system_info = get_system_info( $script_dir );

$pkg_distrib_suffix = get_pkg_distrib_suffix( $system_info );

if( $b_newline != 0 ) {
  print "$pkg_distrib_suffix\n";
}
else {
  print "$pkg_distrib_suffix";
}

sub get_pkg_distrib_suffix {
  my ( $system_info, $r_pkg_distrib_suffix ) = @_;
  my $pkg_distrib_suffix;
  my @system_info_components = split /:/, $system_info;
  
  if( $system_info_components[ 3 ] ne 'unknown' ) {
    $pkg_distrib_suffix = $system_info_components[ 3 ];
    if( $system_info_components[ 4 ] ne 'unknown' ) {
      $pkg_distrib_suffix .= '-' . $system_info_components[ 4 ];
    }
  }
  
  return $pkg_distrib_suffix;
}

sub get_script_dir {
  my ( $script ) = @_;
  my $filename;
  my $suffix;
  my $script_dir;
  
  ( $filename, $script_dir, $suffix ) = fileparse( $script );
  
  $script_dir = File::Spec->rel2abs( $script_dir );
  
  return $script_dir;
}

sub get_system_info {
  my ( $script_dir ) = @_;
  my $cmd;
  my $system_info;
  my @system_info_components;
  
  $cmd = File::Spec->catfile( $script_dir, 'get_system_info_unix.pl' );

  $system_info = invoke_external_cmd( $cmd );
  chomp( $system_info );
  
  # check $system_info (number of required attributes)
  @system_info_components = split /:/, $system_info; 
  if( scalar( @system_info_components ) < 5 ) {
    die "get_system_info() has returned an unexpected string: \"$system_info\"";
  }

  return $system_info;
}

sub invoke_external_cmd {
  my ( $external_cmd ) = @_;
  my $return_value;
  my $external_cmd_quoted = quote_executable_path( $external_cmd );
  
  if( ! -f "$external_cmd" ) {
    die "error: the external command \"$external_cmd\" does not exist, check your installation";
  }
  
  $return_value = `$external_cmd_quoted`;
  if( $? != 0 ) {
    die "error: the external command \"$external_cmd\" failed unexpectedly, please contact technical support";
  }
  
  return $return_value;
}

sub quote_executable_path {
  my ( $prog ) = @_;
  
  if( $prog =~ /^"/ ) {
    # already quoted
    return $prog;
  }
  
  if( $prog =~ /[ ]+/ ) {
    $prog = "\"$prog\"";
  }
  
  return $prog;
}
