#!/usr/bin/perl
#
# get_system_info_unix.pl
#
# A script to return a single string consisting of several attributes that are separated by colons.
#
# Examples:
# 
# linux:x86:2:ubuntu:hardy      # ubuntu hardy, 2 processors
# linux:x86_64:1:ubuntu:hardy
# linux:x86_64:1:opensuse:11.1
# linux:x86_64:1:fedora:leonidas
#
# solaris:x86_64:1:unknown:5.11
#
use FileHandle;
use English;

use strict 'vars';

my %platform_support_map = (
  'linux'    => '1',
  'solaris'  => '1',
  'darwin'   => '1'   
);

my $system_info;

if( ! exists( $platform_support_map{ $OSNAME } ) ) {
  die "unsupported platform detected";
}

$system_info = get_system_info();

print "$system_info\n";


sub get_system_info {
  my $system_info;
  my $num_processors;
  
  if( $OSNAME eq 'linux' ) {
    my $linux_distro;
    my $linux_codename;
    
    $system_info = 'linux';
    
    # architecture detection
    if( is_unix_64bit() ) {
      $system_info .= ':x86_64';
    }
    else {
      $system_info .= ':x86';
    }
    
    # detection of the number of processors
    $num_processors = get_num_processors();
    $system_info .= ":$num_processors";
    
    get_linux_distro_codename( \$linux_distro, \$linux_codename );
    
    $system_info .= ":$linux_distro:$linux_codename";
      
  }
  elsif( $OSNAME eq 'solaris' ) {
    my $os_release;
    my $os_platform;
    $system_info = 'solaris';
    
    # architecture detection
    $os_platform = `uname -i`;
    chomp( $os_platform );
    
    if( $os_platform ne 'i86pc' ) {
      die "the solaris platform $os_platform is not yet supported";
    }
    
    if( is_unix_64bit() ) {
      $system_info .= ':x86_64';
    }
    else {
      $system_info .= ':x86';
    }
    
    # detection of the number of processors
    $num_processors = get_num_processors();
    $system_info .= ":$num_processors";
    
    
    # the solaris distro attribute is not yet available
    $system_info .= ':unknown';
    
    # add the release instead of a codename
    $os_release = `uname -r`;
    
    $system_info .= ":$os_release";
      
  }
  elsif( $OSNAME eq 'darwin' ) {
    $system_info = 'macos';
    
    # TODO: architecture detection on darwin
    $system_info .= 'unknown';
    
    # detection of the number of processors
    $num_processors = get_num_processors();
    $system_info .= ":$num_processors";
     
    # TODO: DISTRO  CODENAME/RELEASE on darwin
    $system_info .= ':unknown:unknown';
  }
  else {
    $system_info = 'unknown:unknown:unknown:unknown';
  }
}


sub is_unix_64bit {
  my $unix_64bit = 0;
  my $ret_val;
    
  if( $OSNAME eq 'linux' ) {
    
    $ret_val = `uname -m`;
    chomp( $ret_val );
    
    $unix_64bit = 1 if( $ret_val =~ /^x86_64/ );
  }
  elsif( $OSNAME eq 'solaris' ) {
    $ret_val = `isainfo -b`;
    chomp( $ret_val );
    
    $unix_64bit = 1 if( $ret_val == 64 );
  }
  else {
    die "unknown platform detected";
  }
  
  return $unix_64bit;  
}

sub get_linux_distro_codename {
  my ( $r_linux_distro, $r_linux_codename ) = @_;
  
  die if( $OSNAME ne 'linux' );

  $$r_linux_distro = 'unknown';
  $$r_linux_codename = 'unknown'; 
 
  if( -f '/etc/SuSE-release' ) {
    get_suse_distro_codename( '/etc/SuSE-release', $r_linux_distro, $r_linux_codename );    
  }
  elsif( -f '/usr/bin/lsb_release' ) {
    my $ret_value;
    
    $ret_value = `lsb_release -is`;
    if( $? != 0 ) {
      die "the command lsb_release -is failed";
    }
    chomp $ret_value;
    
    $$r_linux_distro = lc( $ret_value );
    
    $ret_value = `lsb_release -cs`;
    if( $? != 0 ) {
      die "the command lsb_release -cs failed";
    }    
    chomp $ret_value;
    
    $$r_linux_codename = lc( $ret_value );
  }
  else {
    die "unsupported LINUX platform encountered, lsb_release seems to be missing"; 
  }
}


sub get_suse_distro_codename {
  my ( $version_file, $r_linux_distro, $r_linux_codename ) = @_;
  
  
  $$r_linux_distro = 'suse';
  $$r_linux_codename = 'unknown';
}

sub get_num_processors {
  my $num_procs = 0;
  my $return_value;
  
  if( $OSNAME eq 'solaris' ) {
    $return_value = `/usr/sbin/psrinfo -p`;
    chomp $return_value;
    $num_procs = $return_value;
  }
  elsif( $OSNAME eq 'linux' ) {
    my $fh;
    my $line;
    
    $fh = new FileHandle '/proc/cpuinfo', "r";
    if ( ! defined $fh ) {
      fatal_error( __LINE__, "cannot open /proc/cpuinfo for reading\n" );
    }

    while( ( defined( $fh ) ) && ( ! $fh->eof() ) ) {
      $line = $fh->getline();
      chomp( $line );
      
      if ( $line =~ /^processor[ \t]*:[ \t]*[0-9]+/ ) {
        $num_procs++;
      }
    }
    undef $fh;
  }
  elsif( $OSNAME eq 'darwin' ) {
    $return_value = `/usr/sbin/sysctl hw.ncpu`;
    
    # the eclipse perl checker does not like the \1 and recommends $1 instead.
    # $return_value =~ s/^hw\.ncpu: (.*)/\1/;
    
    $return_value =~ s/^hw\.ncpu: (.*)/$1/;
    chomp $return_value;
    $num_procs = $return_value;
  }
  else {
    die "the platform $OSNAME is not supported yet";
  }
  
  return $num_procs;
}
