#!/usr/bin/perl
#
# make_pkg.pl
#
# This script is desinged to be called by bjam to build one or more debian packages. The package building steps are 
# defined by debian rules files but some essential environment variables are passed by this script to the various makefiles
# involved in building the debian binary packages. This script serves as the interface between BoostBuild and the debian package
# building environment.
#
# Further information is given in the accompanying documentation.
#
use File::Basename;
use File::Spec;
use File::Path;
use FileHandle;
use DirHandle;

use Getopt::Long;
use English;

use strict 'vars';

$::g_hor_ruler = "================================================================";

my $cwd = File::Spec->rel2abs( File::Spec->curdir() );
my $lib_dir = '';
my $bin_dir = '';
my $build_dir = '';

# bjam is supposed to invoke the script from the Jamroot directory
my $top_dir = '';
my $script_dir;
my $script_dir_unix;

my $pkg_arch = '';
my $pkg_fmt = '';
my $pkg_make_cmd = '';
my $make_cmd = 'make';

my %pkg_make_cmd_map = (
  'deb'   =>  './debian/rules',
  'rpm'   =>  './rpm/rules',
  'pkg'   =>  './pkg/rules'
);

# x86 or x86_64, derived from the package architecture
my $platform_suffix;
my $b_debug = 0;
my $b_help = 0;
my $tar_filename = '';
my $b_add_toolset_suffix = 1;
my $b_add_platform_suffix = 1;
my $b_add_linux_codename = 1;

my $toolset_suffix = '';
my $pkg_build_dir;
my $makefile;

my $system_info;
my $system_cmd_line;

# e.g. ubuntu-hardy
my $linux_codename_suffix = '<unknown>';

my %optctl = (
  'help'               => \$b_help,
  'debug'              => \$b_debug,
  'tar'                => \$tar_filename,
  'build-dir'          => \$build_dir,
  'bin-dir'            => \$bin_dir,
  'lib-dir'            => \$lib_dir,
  'add-linux-codename' => \$b_add_linux_codename,
  'toolset-suffix'     => \$b_add_toolset_suffix,
  'platform-suffix'    => \$b_add_platform_suffix,
  'arch'               => \$pkg_arch,
  'pkg-fmt'            => \$pkg_fmt
);

# parse the command line 
if( GetOptions( \%optctl, 
                'help!', 
                'debug!', 
                'tar=s', 
                'build-dir=s', 
                'bin-dir=s', 
                'lib-dir=s', 
                'add-linux-codename!', 
                'toolset-suffix!', 
                'platform-suffix!', 
                'arch=s',
                'pkg-fmt=s'  ) != 1 ) {

  fatal_error( __LINE__, "command line not understood, check the syntax\n" );
}

if( $b_help ) {
  usage();
  exit( 0 );
}

if( scalar( @ARGV ) == 0 ) {
  usage();
  exit( 1 );
}

if( $OSNAME eq 'solaris' ) {
  $make_cmd = 'gmake';
}

if( $top_dir eq '' ) {
  $top_dir = '.';
}
else {
  if( ! -d "$top_dir" ) {
    fatal_error( __LINE__, "the directory \"$top_dir\" does not exist\n" );
  }
 
  if ( File::Spec->file_name_is_absolute( $top_dir ) ) {
    #the user has provided an absolute top directory, let's make it relative to please debhelper scripts
    $top_dir = File::Spec->abs2rel( $top_dir, $cwd );
  }
}

$script_dir = File::Spec->catdir( File::Spec->rel2abs( $top_dir ), 'BoostBuild', 'bin' );
if( ! -d "$script_dir" ) {
  fatal_error( __LINE__, "the directory \"$script_dir\" does not exist\n" );
}

$script_dir_unix = File::Spec->catdir( File::Spec->rel2abs( $top_dir ), 'BoostBuild', 'bin', 'unix' );
if( ! -d "$script_dir_unix" ) {
  fatal_error( __LINE__, "the directory \"$script_dir_unix\" does not exist\n" );
}

$system_info = get_system_info( $script_dir_unix );

print "system_info: \"$system_info\"\n";
#exit( 0 );

if( $pkg_fmt eq '' ) {
  $pkg_fmt = get_native_pkg_fmt( $system_info );
}

if( ! exists( $pkg_make_cmd_map{ $pkg_fmt } ) ) {
  fatal_error( __LINE__, "package format \"$pkg_fmt\" not understood\n" );
}

if( $pkg_arch eq '' ) {
  $pkg_arch = get_native_pkg_arch( $system_info );
}

if( $pkg_arch !~ /^(i386)|(amd64)/ ) {
  fatal_error( __LINE__, "the package architecture \"$pkg_arch\" is not yet supported.\n" );
}

if( $b_add_toolset_suffix ) {
  my $gcc_version;

  $gcc_version = `gcc -dumpversion`;
  chomp $gcc_version;
  $toolset_suffix = 'gcc-' . $gcc_version;
}

if( $pkg_arch eq 'i386' ) {
  $platform_suffix = 'x86';
} 
elsif ( $pkg_arch eq 'amd64' )  {
  $platform_suffix = 'x86_64';
}
else {
  fatal_error( __LINE__, "the package architecture \"$pkg_arch\" is not yet supported.\n" );
}

#check the package building environment first
check_environment( $pkg_fmt );


if( $pkg_make_cmd eq '' ) {
  $pkg_make_cmd = $pkg_make_cmd_map{ $pkg_fmt };
}

if( $build_dir eq '' ) {
  # Note the release and debug packages are stored in the same directory to simplify 
  # automatic package version extraction. 
  #
  # example:
  # 
  # - the debug package A depends on the release package B
  # - the release packages must be built first and the package dependencies of the debug
  #   package can be updated automatically by deb_depends_filter.pl    
  
  
  #if( $b_debug != 0 ) {
    #the place to copy the debug packages to
  #  $build_dir = File::Spec->catfile( $top_dir, 'build', 'deb-dbg', $platform_suffix );
  #}
  #else {
    #the place to copy the release packages to
    $build_dir = File::Spec->catfile( $top_dir, 'build', $pkg_fmt, $platform_suffix );
  #}
}
#elsif( File::Spec->file_name_is_absolute( $build_dir ) ) {
  #the user has provided an absolute directory, let's make it relative to please debhelper scripts
#  $build_dir = File::Spec->abs2rel( $build_dir, $cwd );
#}

$build_dir = File::Spec->rel2abs( $build_dir );

if( -d $build_dir ) {
  #remove all debian packages out of the build directory to avoid archiving outdated 
  #packages.
  remove_pkg_files( $build_dir, $b_debug, $pkg_fmt );
}

if( ! -d $build_dir ) {
  # create directory the final package files
  mkpath( $build_dir, {verbose => 1} );
}


if( $bin_dir eq '' ) {
  my $bin_dir_par1;

  $bin_dir_par1 = File::Spec->catfile( $top_dir, 'bin' );
  
  if( $b_add_toolset_suffix ) {
    $bin_dir_par1 = File::Spec->catfile( $bin_dir_par1, $toolset_suffix );
  }

  if( $b_add_platform_suffix ) {
    $bin_dir_par1 = File::Spec->catfile( $bin_dir_par1, $platform_suffix );
  }
  
  if( $b_debug != 0 ) {
    $bin_dir = File::Spec->catfile( $bin_dir_par1, 'debug-static' );
  }
  else {
    $bin_dir = File::Spec->catfile( $bin_dir_par1, 'release-static' );
  }
}
elsif( File::Spec->file_name_is_absolute( $bin_dir ) ) {
  #the user has provided an absolute directory, let's make it relative to please debhelper scripts
  $bin_dir = File::Spec->abs2rel( $bin_dir, $cwd );
}

if( $lib_dir eq '' ) {
  $lib_dir = File::Spec->catfile( $top_dir, 'lib' );
  
  if( $b_add_toolset_suffix ) {
    $lib_dir = File::Spec->catfile( $lib_dir, $toolset_suffix );
  }
  
  if( $b_add_platform_suffix ) {
    $lib_dir = File::Spec->catfile( $lib_dir, $platform_suffix );
  }
}
elsif( File::Spec->file_name_is_absolute( $lib_dir ) ) {
  #the user has provided an absolute directory, let's make it relative to please debhelper scripts
  $lib_dir = File::Spec->rel2abs( $lib_dir, $cwd );
}

#pick up the linux distribution and codename
$linux_codename_suffix = get_linux_codename_suffix( $system_info );


my @pkg_list;
my $pkg_subdirs;
my $pkg_build_top_dir;

#check package building directories
foreach $pkg_build_dir (@ARGV) { 

  # we support the following (pseudo) directory names
  #
  # i)  $pkg_build_dir refers to a directory with a top level makefile
  # ii) $pkg_build_dir := <directory>:<subdir1>,<subdir2>...
  # 
  # The second alternative is intended to build only a subset of the 
  # available packages.
  #

  if( $pkg_build_dir =~ /^[^:]+:[a-zA-Z0-9_\-\.,]+$/ ) {
    
    print "Extended package build directory specification encountered: \"$pkg_build_dir\".";
    ($pkg_build_top_dir,$pkg_subdirs) = split( /:/, $pkg_build_dir );

    #@pkg_list = split( /,/, $pkg_subdirs );
    #print "Extended package build splitted: @pkg_list \n";
  
    if( ! -d $pkg_build_top_dir ) {
      fatal_error( __LINE__, "package build directory \"$pkg_build_top_dir\" not found.\n" );
    }
  
    #check for top level Makefile
    $makefile = File::Spec->catfile( $cwd, $pkg_build_top_dir, 'Makefile' );
    if( ! -f $makefile ) {
      fatal_error( __LINE__, "$makefile does not exist\n" );
    }
  }
  else {
    if( ! -d $pkg_build_dir ) {
      fatal_error( __LINE__, "package build directory \"$pkg_build_dir\" not found.\n" );
    }
  
    #check for top level Makefile
    $makefile = File::Spec->catfile( $cwd, $pkg_build_dir, 'Makefile' );
    if( ! -f $makefile ) {
      fatal_error( __LINE__, "$makefile does not exist\n" );
    }
  }
}

my $make_target;
my $make_cmd_line;
my $make_variable_defs;

print "\n$::g_hor_ruler\n";
print "Finished preparing package building environment and settings.\n";
#print "TOP:               $top_dir\n";
print "SCRIPT_DIR:        $script_dir\n";
print "SCRIPT_DIR_UNIX:   $script_dir_unix\n";
print "PKG_SAVE_DIR:      $build_dir\n";
print "BIN_DIR:           $bin_dir\n";
print "LIB_DIR:           $lib_dir\n";
print "ARCH:              $pkg_arch\n";
print "PKG_FMT:           $pkg_fmt\n";
print "PKG_MAKE_CMD:      $pkg_make_cmd\n";

if( $b_add_linux_codename != 0 ) {
  print "LINUX_CODENAME:    ${linux_codename_suffix}\n";
}

if( $tar_filename ne '' ) {
  print "TAR:               $tar_filename\n";
}
print "$::g_hor_ruler\n";
#exit 0;

#define the common make variables
$make_variable_defs = "BIN_DIR=$bin_dir LIB_DIR=$lib_dir PKG_SAVE_DIR=$build_dir SCRIPT_DIR=$script_dir SCRIPT_DIR_UNIX=$script_dir_unix ARCH=$pkg_arch";
$make_variable_defs .= " PKG_FMT=$pkg_fmt PKG_MAKE_CMD=$pkg_make_cmd";

if( $pkg_fmt eq 'deb' ) {
  $make_variable_defs .= " DH_VERBOSE=0 DH_COMPAT=6 DH_ALWAYS_EXCLUDE=\".svn\"" ;
  #$make_variable_defs .= " DH_COMPAT=6 DH_ALWAYS_EXCLUDE=\".svn\"" ;
}

if( $b_add_linux_codename != 0 ) {
  #add _ubuntu-hardy to package names
  $make_variable_defs .= " LINUX_CODENAME=_${linux_codename_suffix}";
}

#start building the packages
foreach $pkg_build_dir (@ARGV) { 
  $make_cmd_line = "make -C $pkg_build_dir $make_variable_defs";

  if( $b_debug != 0 ) {
    $make_target = 'all.dbg';
  }
  else {
    $make_target = 'all';
  }

  $system_cmd_line = $make_cmd_line . ' ' . $make_target;

  print "$::g_hor_ruler\n";
  print "Building packages in $pkg_build_dir ...\n";
  print "$system_cmd_line\n";
  print "$::g_hor_ruler\n";

  system( $system_cmd_line );
  if( $? != 0 ) {
    fatal_error( __LINE__, "building packages in \"$pkg_build_dir\" has failed.\n" );
  }
}


if( $tar_filename ne '' ) {
  my @pkg_file_list;
  my @pkg_file_list_final;
  my $filename;
  my $pkg_filename;
  my $cnt;

  if( -f $tar_filename ) {
    # remove the old tar_file first
    $cnt = unlink( $tar_filename );
    fatal_error( __LINE__, "cannot remove \"$tar_filename\"\n" ) if( $cnt != 1 ); 
  }
  
  # get the list of deb files ($build_dir may contain other file types)
  @pkg_file_list = find_pkg_files( $build_dir, $b_debug, $pkg_fmt );

  if( scalar( @pkg_file_list ) == 0 ) {
    print "\n";
    print "Warning: no package files to archive found, please check the build log\n";
    print "\n";
    exit 1;
  }

  # create a list of debian files without a leading path
  foreach $filename (@pkg_file_list) {
    $pkg_filename = basename( $filename );
    push @pkg_file_list_final, $pkg_filename;
  }

  print "\n";
  print "\nArchiving packages:\n@pkg_file_list_final => $tar_filename\n";
  print "\n";

  $system_cmd_line = "tar -zcf $tar_filename -C $build_dir @pkg_file_list_final";
  system( $system_cmd_line );
  if( $? != 0 ) {
    fatal_error( __LINE__, "archiving packages using tar has failed\n" );
  }
}

sub get_linux_codename_suffix {
  my ( $system_info, $r_pkg_distrib_suffix ) = @_;
  my $distrib_id;
  my $codename;
  my $distrib_suffix;
  
  my @system_info_components = split /:/, $system_info;

  die if( scalar( @system_info_components ) < 5 );

  $distrib_suffix = $system_info_components[ 3 ] . '-' . $system_info_components[ 4 ];
  
  return $distrib_suffix;
}


sub remove_pkg_files {
  my ( $build_dir, $b_debug, $pkg_fmt ) = @_;
  my @pkg_file_list; 
  my $filename;
  my $cnt;

  @pkg_file_list = find_pkg_files( $build_dir, $b_debug, $pkg_fmt );

  return if( scalar( @pkg_file_list ) == 0 );

  foreach $filename (@pkg_file_list) {
    print "rm $filename\n";

    $cnt = unlink $filename;
    fatal_error( __LINE__, "cannot remove \"$filename\"\n" ) if( $cnt != 1 );
  }
}

sub find_pkg_files {
  my ( $build_dir, $b_debug, $pkg_fmt ) = @_;
  my $dirh;
  my $filesys_comp;
  my @dir_list;
  my @pkg_file_list = ();
  my $filename;
  my $debug_suffix;
  my $pkg_extension;
  my %pkg_file_extension_map = ( 'deb' => 'deb', 'rpm' => 'rpm', 'pkg' => 'pkg' );
  
  if( ! exists( $pkg_file_extension_map{ $pkg_fmt } ) ) {
    fatal_error( __LINE__, "internal error: package format $pkg_fmt not understood\n" );
  }
  else {
    $pkg_extension = $pkg_file_extension_map{ $pkg_fmt };
  }
  
  $dirh = new DirHandle $build_dir;
  if( ! defined $dirh ) {
    fatal_error( __LINE__,  "cannot open directory \"$build_dir\" for reading\n" );
  }
  
  while( defined( $filesys_comp = $dirh->read ) ) { 
    # need to save the file system component temporarily and perform the file type check later on because
    # some platforms do not allow to check the file type while the directory is being searched.
    # print "filesys_comp => $filesys_comp\n";

    if( $filesys_comp =~ /.*\.$pkg_extension$/i ) { 
      #print "filesys_comp => $filesys_comp\n";
      push @dir_list, $filesys_comp;
      #print "filesys_comp => $filesys_comp\n";
    }
  }
  undef $dirh;

  return @pkg_file_list if( scalar( @dir_list ) == 0 );

  # found some file system components ending with .deb but this may be a directory!!
  foreach $filesys_comp (@dir_list) {

    $filename = File::Spec->catfile( $build_dir, $filesys_comp );
    if( -f $filename ) {
      # found a deb file
      my $pkg_filename;
      my @pkg_filename_components;
      my @pkg_name_components;
      
      # push @deb_file_list, $filename;
      
      # strip extension
      $pkg_filename = $filesys_comp;
      $pkg_filename =~ s/\.$pkg_extension$//;
      
      # analyze the filename depending on the debug flag
      @pkg_filename_components = split /_/, $pkg_filename;
      
      # split package name on '-' assuming a debug package name ends with '-dbg'
      @pkg_name_components = split /-/, $pkg_filename_components[ 0 ];
      
      $debug_suffix = pop @pkg_name_components;
      
      if( $debug_suffix eq 'dbg' ) {
        if( $b_debug ) {
          push @pkg_file_list, $filename;
        }
      }
      else {
       if( ! $b_debug ) {
          push @pkg_file_list, $filename;
        }        
      }      
    }
  }

  return @pkg_file_list;
}

#
# issue an error message and terminate the program
#
sub fatal_error {
  my ($line, $msg) = @_;

  #if( defined( $::g_fh_error ) ) {
  #  print $::g_fh_error  "error: $msg";
  #}
  #else {
    print STDERR  "make_pkg.pl: error at line $line: $msg";
  #}

  exit 1;
}


sub get_system_info {
  my ( $script_dir ) = @_;
  my $system_info;
  my $cmd = File::Spec->catfile( $script_dir, 'get_system_info_unix.pl' );
  my $cmd_quoted = quote_executable_path( $cmd );
  
  $system_info = `$cmd_quoted`;
  if( $? != 0 ) {
    fatal_error( __LINE__, "the external command \"$cmd\" failed\n" );
  }
  
  chomp( $system_info );

  return $system_info;
}

sub quote_executable_path {
  my ( $prog ) = @_;
  
  if( $prog =~ /^"/ ) {
    # already quoted
    return $prog;
  }
  
  if( $prog =~ /[ ]+/ ) {
    $prog = "\"$prog\"";
  }
  
  return $prog;
}


sub get_native_pkg_fmt {
  my ( $system_info ) = @_;
  my $pkg_fmt;
  my @system_info_components;
  
  @system_info_components = split( ':', $system_info );
   
  if( $system_info_components[ 0 ] eq 'linux' ) {
    if( scalar( @system_info_components ) < 3 ) {
      fatal_error( __LINE__, "linux system info \"$system_info\" not understood\n" );
    }
    
    if( $system_info_components[ 3 ] =~ /^ubuntu/ ) {
      $pkg_fmt = 'deb';
    }
    elsif(   ( $system_info_components[ 3 ] =~ /^fedora/ ) 
          && ( $system_info_components[ 3 ] =~ /^suse/ ) ) {
      $pkg_fmt = 'rpm';
    }
    else {
      die;
    }
  }
  elsif( $system_info_components[ 0 ] eq 'solaris' ) {
    # solaris package datastream format
    $pkg_fmt = 'pkg';    
  }  
  else {
    die;
  }
  
  return $pkg_fmt;  
}

sub get_native_pkg_arch {
  my ( $system_info ) = @_;
  my $pkg_arch;
  my @system_info_components;
  
  @system_info_components = split( ':', $system_info );

  die if( scalar( @system_info_components ) < 2 );
  
  if( $system_info_components[ 0 ] eq 'linux' ) {
    if( $system_info_components[ 1 ] eq 'x86_64' ) {
      $pkg_arch = 'amd64';
    } 
    else {      
      $pkg_arch = 'i386';
    }    
  }
  elsif( $system_info_components[ 0 ] eq 'solaris' ) {
    $pkg_arch = 'i386';
  }
  else {
    die;
  }
  
  return $pkg_arch;
}

# perform a few checks on the build environment
sub check_environment {
  my ( $pkg_fmt ) = @_;
  
  if( $OSNAME ne 'linux' ) {
    fatal_error( __LINE__, "the platform $OSNAME is not supported yet.\n" );
  }
  
  if( $pkg_fmt eq 'deb' ) {
    # we depend on the debhelper script package
    if( ! -f '/usr/bin/dh_install' ) {
      fatal_error( __LINE__, "dh_install not found, did you install the debhelper package?\n" );
    }
  
    if( ! -f '/usr/bin/dpkg' ) {
      fatal_error( __LINE__, "dpkg not found, do you run a debian style LINUX?\n" );
    }
  }
  else {
    die;
  }
}

#  'debug'              => \$b_debug,
#  'tar'                => \$tar_filename,
#  'build-dir'          => \$build_dir,
#  'bin-dir'            => \$bin_dir,
#  'lib-dir'            => \$lib_dir,
#  'add-linux-codename' => \$b_add_linux_codename,
#  'toolset-suffix'     => \$b_add_toolset_suffix,
#  'platform-suffix'    => \$b_add_platform_suffix,
#  'arch'               => \$pkg_arch,
#  'pkg-fmt'            => \$pkg_fmt

sub usage {
  my $how_to_call = <<"__END_OF_USAGE__";
Usage: make_pkg.pl [options] package_dir1 [package_dir2 ...]

options:=
     --help              show the usage information
     
     --arch=<arch>       the package architecture to be used [default: native]
     
                         example: ubuntu/64 bit => amd64 or i386
                                  ubuntu/32 bit => i386
     
     --tar=<s>           the tar file to archive the package files [default: none]
     
     --build-dir=<s>     the build directory to be used to save the package files [default: build]
     
                         N.B. the package files will be saved by default in the following directories:
                         
                         ubuntu/amd64   =>  build/deb/x86_64
                         ubuntu/i386    =>  build/deb/x86
                         
     --bin-dir=<s>       the directory containing the executables to be packaged
     
                         example:                                                 
__END_OF_USAGE__

  print STDERR $how_to_call;
}

