/**
  \ingroup UTFexampleLibTest2
  \file    UTFexampleLibTest2.cpp
  \brief   This UTFexampleLibTest2.cpp file shows some features of boost UTF
  \author  selinger@hhi.fraunhofer.de
  \date    3/3/2009

  This example is designed to demonstrate some features of the boost UTF.
  templates, defines, exceptions, expected failures, ...
  SEARCH for "FEATURE-DEMO:" in this file 

  What you get is:  a clean test

  // Find the complete documentation at:   http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/index.html
  // Find the runtime parameter reference: http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/runtime-config/reference.html
  // Find the test tool reference:         http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/testing-tools/reference.html
*/

///////////////////////////////////////////////
// Required includes and defines 
///////////////////////////////////////////////

/// 1. A must! As always I first need to include the precompiled header file
#include "UTFexampleLibTest2Pch.h"

/// 2. A must! This boost UTF macro  BOOST_TEST_MODULE  (use once and only once) generates the main entry point for this test project and names the master test suite. The name is up to you. http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/test-organization/master-test-suite.html
#define BOOST_TEST_MODULE ts_master master_test_stuite

/// 3. A must! This boost UTF header file must always be included. We use the static library variant of the UTF. It is not recommended to use the single header library variant
#include <boost/test/unit_test.hpp>

/// Include the header file(s) of the class(es) you want to test
#include "UTFexampleLib/UTFsample2.h"
#include "UTFexampleLib/UTFsampleTemplate.h"

#include <boost/mpl/list.hpp>



////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOST_AUTO_TEST_SUITE( ts1_template ) //  FEATURE-DEMO: The macro  BOOST_AUTO_TEST_SUITE  allows to group test cases into named test suites that you can choose to run separately on the the command-line. http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/test-organization/auto-test-suite.html
////////////////////////////////////////////////////////////////////////////////////////////////////////

// Definition of a list of types. Each one will result in a template test case.
typedef boost::mpl::list<unsigned char, short, unsigned int, unsigned long, float> Some_Template_Types;

// FEATURE-DEMO: This macro shows the usage of test cases run for all types of a template class. http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/test-organization/auto-test-case-template.html
// NOTE! this test case will be expanded to N test cases depending on the Some_Template_Types list size.
BOOST_AUTO_TEST_CASE_TEMPLATE( tc1_template, T, Some_Template_Types )
{
  // Instance of the class under test
  utfexamplelib::UTFsampleTemplate<T> cUTFsampleTemplate;
  // Make a test assertion
  BOOST_CHECK_EQUAL( cUTFsampleTemplate.getSizeOfT(), sizeof(T) );
}
BOOST_AUTO_TEST_SUITE_END()



////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOST_AUTO_TEST_SUITE( ts2_div_test_tools ) 
////////////////////////////////////////////////////////////////////////////////////////////////////////

// We add a new test and name it.
BOOST_AUTO_TEST_CASE( tc1_define ) 
{ 
  #define USE_FAST_CODE YES
  
  // FEATURE-DEMO: checking for a define 
  BOOST_CHECK( BOOST_IS_DEFINED( USE_FAST_CODE ) ); 
  
  // Verifying that in e.g. a release version some #define TEST_CODE_ON is not set!
  BOOST_CHECK( !BOOST_IS_DEFINED( TEST_CODE_ON ) ); 
}

// We add a new test and name it.
BOOST_AUTO_TEST_CASE( tc2_exception )
{
  // Instance of the class under test
  utfexamplelib::UTFsample2 cUTFsample2;

  // Check for a specific exception to be thrown. in this example no boost test error will occur
  BOOST_CHECK_THROW( cUTFsample2.doThrowSample2Exception(), utfexamplelib::Sample2Exception );
}

// FEATURE-DEMO: With the macro BOOST_AUTO_TEST_CASE_EXPECTED_FAILURES you can specify that you expect a test case to fail a certain amount of times
// Use with care! http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/test-organization/expected-failures.html
// NOTE!!! you can only see with --report_level=detailed    whether the expected failure occured in the expected test_case :-(
BOOST_AUTO_TEST_CASE_EXPECTED_FAILURES( tc3_expected_failure, 1 );

// We add a new test and name it.
BOOST_AUTO_TEST_CASE( tc3_expected_failure ) 
{ 
  // Instance of the class under test
  utfexamplelib::UTFsample2 cUTFsample2; 

  for( int i = 1; i < 5; i++ )
  {
    // This check will fail once and it will be reported. But as we have specified that we expect this test case to fail once, the summary 
    // of test runs will say:      "***No errors detected"   
    BOOST_CHECK_EQUAL( cUTFsample2.addNumbersWithBug( i, i ), i + i ); 
  }
}

BOOST_AUTO_TEST_SUITE_END()







// FEATURE-DEMO: this is a fixture http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/fixture/test-suite-shared.html
// a pice of data/code shared by all test cases with the members being accessible directly within all test cases. Fixtures are used for repeated identical test setup and teardown
// boost provides per test case fixtures, test suite fixtures and global fixtures
struct F 
{ 
  // This example fixture is a parameter to the test suite and then known in each test case of the suite. setup and teardown are done for each test case!
   F() : i( 0 ) { BOOST_TEST_MESSAGE( "setup fixture" ); }
  ~F()          { BOOST_TEST_MESSAGE( "teardown fixture" ); }

public:
  void incI() {i++;}
  int i;
};


// FEATURE-DEMO: test_suite macro contains _FIXTURE_ and gets F as parameter
////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOST_FIXTURE_TEST_SUITE( ts3_fixture_suite, F ) 
////////////////////////////////////////////////////////////////////////////////////////////////////////

// We add a new test and name it.
BOOST_AUTO_TEST_CASE( tc1_fixture_test ) 
{ 
 // we increment up to 3
 incI();
 incI();
 incI();
 BOOST_CHECK_EQUAL( i, 3 );
 // direct access to F members
 i += 5;
}

// We add a new test and name it.
BOOST_AUTO_TEST_CASE( tc2_fixture_test ) 
{ 
  // this test gets the same F as the previous one (after construction) start with i=0 again
  // we increment up to 3 
  incI();
  incI();
  incI();
  BOOST_CHECK_EQUAL( i, 3 );
}

BOOST_AUTO_TEST_SUITE_END()
