/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef VIDEOCODECTOOLSLIB_TOOLLIBPARSER_H
#define VIDEOCODECTOOLSLIB_TOOLLIBPARSER_H

class ToolLibStringBuffer
{
public:
  ToolLibStringBuffer()
  {
    m_cTextString  = "";
    m_cLineEndTerminator = "\r\n";
    m_uiCurrentPos = 0;
    m_uiLineNumber = 0;
    m_bInit = false;
  }

  virtual ~ToolLibStringBuffer() {}

public:
  ErrVal init( const std::string& rcTextString )
  {
    ROT( true == m_bInit );
    ROT( 0 == rcTextString.size() );

    m_cTextString  = rcTextString;
    m_uiCurrentPos = 0;
    m_uiLineNumber = 0;

#if defined( MSYS_WIN32 )
    m_cLineEndTerminator = "\r\n";
#elif defined( MSYS_LINUX )
    m_cLineEndTerminator = "\n";
#else
    ROT( true );
    m_cLineEndTerminator = "\r\n";
  #endif

    m_bInit = true;

    return Err::m_nOK;
  }

  ErrVal uninit()
  {
    ROT( false == m_bInit );

    m_cTextString  = "";
    m_uiCurrentPos = 0;
    m_uiLineNumber = 0;
    m_cLineEndTerminator = "\r\n";

    m_bInit = false;

    return Err::m_nOK;
  }

  size_t getCurrentPos() const { return m_uiCurrentPos; }

  ErrVal getLine( std::string& cLine )
  {
    ROT( false == m_bInit );

    m_uiLineNumber++;

    // initialize with all okay. can be Err or EndOfFile
    ErrVal iRet = Err::m_nOK;

    // check for end of buffer
    if( m_uiCurrentPos >= m_cTextString.size() )
    {
      cLine = "";
      return Err::m_nEndOfBuffer;
    }

    size_t uiStartPos = m_uiCurrentPos;

    // this variable is used to increment the position pointer behind the line end control characters
    int iNumberOfLineEndChars = 0;

    // find end of line allowing various kind of line end terminators
    size_t uiPos = m_cTextString.find_first_of( "\r\n", uiStartPos );

    std::string cActualLineEndTerminator = "";

    if( std::string::npos != uiPos )
    {
      // one or both of CR and LF was found
      iNumberOfLineEndChars = 1;

      if( ('\r' == m_cTextString[uiPos]) && ('\n' == m_cTextString[uiPos+1]) )
      {
        // if CR/LF was found
        iNumberOfLineEndChars = 2;
        cActualLineEndTerminator = "\r\n";
      }
      else if( ('\n' == m_cTextString[uiPos]) && ('\r' == m_cTextString[uiPos+1]) )
      {
        // if LF/CR was found
        iNumberOfLineEndChars = 2;
        cActualLineEndTerminator = "\n\r";
      }
      else if( '\n' == m_cTextString[uiPos] )
      {
        // if LF only was found. okay as it is standard for linux files
        iNumberOfLineEndChars = 1;
        cActualLineEndTerminator = "\n";
      }
      else if( '\r' == m_cTextString[uiPos] )
      {
        // if CR only was found.
        iNumberOfLineEndChars = 1;
        cActualLineEndTerminator = "\r";
      }
    }
    else if( std::string::npos == uiPos )
    {
      // handle the very end of the string which might miss a line end terminator.
      // allow this case and handle it like line that is correctly terminated by line end.
      iNumberOfLineEndChars = 0;
      cActualLineEndTerminator = m_cLineEndTerminator;
      uiPos = m_cTextString.length();
    }

    // paranoia checking that the new position is behind the old position
    ROT( uiStartPos > uiPos );

    // get the line up to the last char without line end chars
    cLine.assign( m_cTextString, uiStartPos, uiPos - uiStartPos );

    // update current position to the beginning of the next line.
    m_uiCurrentPos = uiPos + iNumberOfLineEndChars;

    if( cActualLineEndTerminator != m_cLineEndTerminator )
    {
      // signal that we encounterd a non-platform specific line end
      return Err::m_nERR;
    }

    // paranoia
    if( m_uiCurrentPos > m_cTextString.size() )
    {
      AOT( true );
      m_uiCurrentPos = m_cTextString.size();
    }

    return iRet;
  }

  size_t getLineNumber() const { return m_uiLineNumber; }

private:
  std::string m_cTextString;
  size_t m_uiCurrentPos;
  bool m_bInit;
  UInt m_uiLineNumber;
  std::string m_cLineEndTerminator;

};

#if defined( MSYS_WIN32 )
# pragma warning( disable: 4251 )
#endif

class ToolLibParser
{
public:
  ToolLibParser();
  virtual ~ToolLibParser();

public:
  ErrVal init( const std::string& cInputString, int iErrId = 0 );
  ErrVal uninit();

  void msgFatal( const std::string& rcMessage );
  void msgError( const std::string& rcMessage );
  void msgWarning( const std::string& rcMessage );
  void msgInfo( const std::string& rcMessage );
  void msgVerbose( const std::string& rcMessage );

  // gets the next token from the input to m_cToken and moves beyond it
  ErrVal getToken();

  // Similar to getToken below but takes all from inbetween the double quotes " until next " when these delimiters are used
  ErrVal getString( std::string& rcString );
  ErrVal getString( std::string& rcString, const char cOpenChar, const char cCloseChar );

  // gets the next Token into the string;
  ErrVal getToken( std::string& rcString );

  // peeks the next token into the string but does not change the parsing position
  ErrVal peekToken( std::string& rcString );

  ErrVal getToken( bool&   rVal );
  ErrVal getToken( char&   rVal );
  ErrVal getToken( short&  rVal );
  ErrVal getToken( int&  rVal );
  ErrVal getToken( Int64&  rVal );
  ErrVal getToken( UChar&  rVal );
  ErrVal getToken( UShort& rVal );
  ErrVal getToken( UInt& rVal );
  ErrVal getToken( UInt64& rVal );
  ErrVal getToken( float&  rVal );
  ErrVal getToken( double& rVal );

  // get the expected token. m_nOK if found; error message and Err::m_nErr otherwise
  ErrVal getExpected( const std::string& rcExpectedToken );
  // ErrVal getExpected( const char* pchString  );

  // get/remove the optional token if there; otherwise do nothing
  ErrVal getOptional( const std::string& rcExpectedToken );
  // ErrVal getOptional( const char* pchString  );

public:
  void setIgnoreCase( const bool bIgnoreCase ) { m_bIgnoreCase = bIgnoreCase; }
  bool getIgnoreCase() const { return m_bIgnoreCase; }

protected:

  void moveOverWhiteSpace();
  bool deleteComments();
  ErrVal getNextLine( const bool bWithErrorMessage );

  ErrVal peekToken();

  void stringToLowerCase( std::string& rcString );

  ErrVal stringToValue( const std::string& rcString, UChar&  rVal);
  ErrVal stringToValue( const std::string& rcString, UShort& rVal);
  ErrVal stringToValue( const std::string& rcString, UInt& rVal);
  ErrVal stringToValue( const std::string& rcString, UInt64& rVal );

  ErrVal stringToValue( const std::string& rcString, char&  rVal);
  ErrVal stringToValue( const std::string& rcString, short& rVal);
  ErrVal stringToValue( const std::string& rcString, int& rVal);
  ErrVal stringToValue( const std::string& rcString, Int64& rVal );
  ErrVal stringToValue( const std::string& rcString, float& rVal );
  ErrVal stringToValue( const std::string& rcString, double& rVal );


private:
  ToolLibStringBuffer m_cStringBuffer;
  std::string m_cLine;
  std::string m_cToken;
  size_t m_uiLinePos;
  bool m_bInit;
  bool m_bIgnoreCase;
  int m_iErrId;
};

#if defined( MSYS_WIN32 )
# pragma warning( default: 4251 )
#endif

#endif // VIDEOCODECTOOLSLIB_TOOLLIBPARSER_H
