/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _BLOCK_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _BLOCK_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCodecToolsLib/VideoCodecPicture.h"
#include "VideoCodecToolsLib/MSysLargeFile.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class ReuseBuffer
{
public: 
  ReuseBuffer()
    : m_pcBuffer(NULL)
    , m_uiSize(0)
    , m_bLock(false) {}

  ~ReuseBuffer()
  {
    AOT(m_bLock);
    delete [] m_pcBuffer;
  }

  unsigned char* getBuffer( const UInt uiSize )
  {
    if( uiSize > m_uiSize )
    {
      delete [] m_pcBuffer;
      m_uiSize = uiSize;
      m_pcBuffer = new unsigned char[ uiSize ];
    }
    m_bLock = true;
    return m_pcBuffer;
  }

  void freeBuffer()
  {
    m_bLock = false;
  }

private:
  unsigned char* m_pcBuffer;
  UInt m_uiSize;
  bool m_bLock;
};


typedef enum
{
  PRIMARY_RESOLUTION,
  PLANE_RESOLUTION,
} VecType;

class Ratio
{
public:
  Ratio() {}

  Ratio( int iNumerator, int iDenominator = 1 ) : m_iNumerator( iNumerator ), m_iDenominator( iDenominator ) {}

  Ratio operator*( const Ratio &r ) const { return Ratio( m_iNumerator * r.m_iNumerator, m_iDenominator * r.m_iDenominator ); }
  Ratio operator/( const Ratio &r ) const { return Ratio( m_iNumerator * r.m_iDenominator, m_iDenominator * r.m_iNumerator ); }

  Ratio operator+( const Ratio &r ) const { return Ratio (m_iNumerator * r.m_iDenominator + r.m_iNumerator * m_iDenominator, m_iDenominator * r.m_iDenominator);}
  Ratio operator-( const Ratio &r ) const { return Ratio (m_iNumerator * r.m_iDenominator - r.m_iNumerator * m_iDenominator, m_iDenominator * r.m_iDenominator);}

  bool operator< ( const Ratio &r ) const { return m_iNumerator * r.m_iDenominator <  m_iDenominator * r.m_iNumerator; }
  bool operator> ( const Ratio &r ) const { return r < *this; }
  bool operator<=( const Ratio &r ) const { return !(*this > r); }
  bool operator>=( const Ratio &r ) const { return !(*this < r); }

  bool operator==( const Ratio &r ) const { return m_iNumerator * r.m_iDenominator == m_iDenominator * r.m_iNumerator; }
  bool operator!=( const Ratio &r ) const { return !(*this == r); }

  int getNumerator  () const { return m_iNumerator; }
  int getDenominator() const { return m_iDenominator; }

private:
  int m_iNumerator;
  int m_iDenominator;
};


class ShiftScale
{

public:
  ShiftScale() {}
  explicit ShiftScale( int iDownShift ) : m_iDownShift(iDownShift)/*, m_iNumerator( 1<<gMax(0,-iDownShift) ), m_iDenominator( 1<<gMax(0,m_iDownShift)) */
  {   
  }

  int operator*( const int i ) const 
  { 
    if( 0 > m_iDownShift )
    {
      return i << -(m_iDownShift);
    }
    return i >> m_iDownShift;
  }

#if 1
  ShiftScale reciprocal() const { return ShiftScale( -m_iDownShift ); }
#else
  int operator/( const int i ) const 
  { 
    if( 0 > m_iDownShift )
    {
      return i >> -(m_iDownShift);
    }
    return i << m_iDownShift;
  }
#endif

  ShiftScale operator*( const ShiftScale &r ) const { return ShiftScale( m_iDownShift + r.m_iDownShift ); }
  ShiftScale operator/( const ShiftScale &r ) const { return ShiftScale( m_iDownShift - r.m_iDownShift ); }

  ShiftScale operator<<( const int i ) const { return ShiftScale( m_iDownShift-i ); }
  ShiftScale operator>>( const int i ) const { return ShiftScale( m_iDownShift+i ); }

  bool operator< ( const ShiftScale &r ) const { return r.m_iDownShift <  m_iDownShift; }
  bool operator> ( const ShiftScale &r ) const { return r < *this; }
  bool operator<=( const ShiftScale &r ) const { return !(*this > r); }
  bool operator>=( const ShiftScale &r ) const { return !(*this < r); }

  bool operator==( const ShiftScale &r ) const { return m_iDownShift == m_iDownShift; }
  bool operator!=( const ShiftScale &r ) const { return !(*this == r); }

  int getDownShiftMask () const { return (1<<m_iDownShift)-1; }
  int getUpShiftMask () const { return (1<<(-m_iDownShift))-1; }

  int getDownShift () const { return m_iDownShift; }
  bool isMulReversible( const int i ) { if(m_iDownShift<=0) return true; return i == ((i>>m_iDownShift)<<m_iDownShift);}
  bool isDivReversible( const int i ) { if(m_iDownShift>=0) return true; return i == ((i>>(-m_iDownShift))<<(-m_iDownShift));}

private:
  int m_iDownShift;
};


class Subsampling
{
public:
  Subsampling() {}

  Subsampling( const ShiftScale &rcX, const ShiftScale &rcY ) : m_cX( rcX ), m_cY( rcY ) {}

  Subsampling operator*( const Subsampling &r ) const { return Subsampling( m_cX * r.m_cX, m_cY * r.m_cY ); }
  Subsampling operator/( const Subsampling &r ) const { return Subsampling( m_cX / r.m_cX, m_cY / r.m_cY ); }
  Subsampling operator<<( const int i )         const { return Subsampling( m_cX << i, m_cY << i ); }
  Subsampling operator>>( const int i )         const { return Subsampling( m_cX >> i, m_cY >> i ); }

  bool operator==( const Subsampling &r ) const { return m_cX == r.m_cX && m_cY == r.m_cY; }
  bool operator!=( const Subsampling &r ) const { return !(*this == r); }


  ShiftScale getX() const { return m_cX; }
  ShiftScale getY() const { return m_cY; }

  void setX( const ShiftScale& r ) { m_cX = r; }
  void setY( const ShiftScale& r ) { m_cY = r; }

private:
  ShiftScale m_cX;
  ShiftScale m_cY;
};

template<VecType vt>
class IntVec2D
{
public:
  IntVec2D( const IntVec2D<PRIMARY_RESOLUTION> &rcPrimBlockSize) 
    : m_iX( rcPrimBlockSize.getX() ), m_iY( rcPrimBlockSize.getY() ) 
  {
  }

  IntVec2D() {}
  IntVec2D( int iX, int iY ) : m_iX( iX ), m_iY( iY ) {}

  bool     operator==( const IntVec2D &rcVec ) const { return m_iX == rcVec.m_iX && m_iY == rcVec.m_iY; }
  bool     operator!=( const IntVec2D &rcVec ) const { return !(*this == rcVec); }
  bool     operator<=( const IntVec2D &rcVec ) const { return m_iX <= rcVec.m_iX && m_iY <= rcVec.m_iY; }
  bool     operator>=( const IntVec2D &rcVec ) const { return m_iX >= rcVec.m_iX && m_iY >= rcVec.m_iY; }
  bool     operator< ( const IntVec2D &rcVec ) const { return m_iX <  rcVec.m_iX && m_iY <  rcVec.m_iY; }
  bool     operator> ( const IntVec2D &rcVec ) const { return m_iX >  rcVec.m_iX && m_iY >  rcVec.m_iY; }

  IntVec2D operator+ ( const IntVec2D &rcVec ) const { return IntVec2D( m_iX + rcVec.m_iX, m_iY + rcVec.m_iY ); }
  IntVec2D operator- ( const IntVec2D &rcVec ) const { return IntVec2D( m_iX - rcVec.m_iX, m_iY - rcVec.m_iY ); }
  IntVec2D operator* ( const IntVec2D &rcVec ) const { return IntVec2D( m_iX * rcVec.m_iX, m_iY * rcVec.m_iY ); }

  IntVec2D operator* ( int i ) const { return IntVec2D( m_iX * i, m_iY * i ); } 
  IntVec2D operator<<( int i ) const { return IntVec2D( m_iX << i, m_iY << i ); } 
  IntVec2D operator>>( int i ) const { return IntVec2D( m_iX >> i, m_iY >> i ); } 
  IntVec2D operator& ( int i ) const { return IntVec2D( m_iX &  i, m_iY &  i ); } 

  IntVec2D<PLANE_RESOLUTION> operator* ( const Subsampling &rcVec ) const
  {
    CAOF( vt == PRIMARY_RESOLUTION );
    // do not allow a rounding here

    AOF_DBG( rcVec.getX().isMulReversible( m_iX ) );
    AOF_DBG( rcVec.getY().isMulReversible( m_iY ) );

    return IntVec2D<PLANE_RESOLUTION>( rcVec.getX() * m_iX, rcVec.getY() * m_iY );
  }

  Subsampling operator/ ( const IntVec2D<PRIMARY_RESOLUTION> &rcVec ) const
  {
    CAOF( vt == PLANE_RESOLUTION );
    // do not allow a rounding here
    ShiftScale cX( gCeilLog2(rcVec.getX() / m_iX) );
    ShiftScale cY( gCeilLog2(rcVec.getY() / m_iY) );

    Subsampling cSub( cX, cY );
    AOT_DBG( rcVec.getX() % m_iX );
    AOT_DBG( rcVec.getY() % m_iY );
    return cSub;
  }


  IntVec2D<PRIMARY_RESOLUTION> operator/ ( const Subsampling &rcVec ) const
  {
    CAOF( vt == PLANE_RESOLUTION );
    // do not allow a rounding here

    AOF_DBG( rcVec.getX().isDivReversible( m_iX ) );
    AOF_DBG( rcVec.getY().isDivReversible( m_iY ) );

    return IntVec2D<PRIMARY_RESOLUTION>( rcVec.getX().reciprocal() * m_iX, rcVec.getY().reciprocal() * m_iY );
  }

  void setX( const int i) { m_iX = i; }
  void setY( const int i) { m_iY = i; }

  const int& getX() const { return m_iX; }
  const int& getY() const { return m_iY; }

  int& getX() { return m_iX; }
  int& getY() { return m_iY; }

private:
  int m_iX;
  int m_iY;
};

typedef IntVec2D<PRIMARY_RESOLUTION>  PrimBlockSize;
typedef IntVec2D<PRIMARY_RESOLUTION>  PrimPixelPos;

typedef IntVec2D<PLANE_RESOLUTION>    BlockSize;
typedef IntVec2D<PLANE_RESOLUTION>    PixelPos;

#define DEBUG_BLOCK_ACCESS 0

template<class t_El> class Block;

template<class t_El>
class Block<const t_El>
{
public:
  Block() : m_pcDataConst( NULL ), m_cSize( 0, 0 ), m_uiStride( 0 ) {}
  virtual ~Block() {}

  template<class t_El2>
  Block( const VideoCodecPlane<t_El2> &rcPlane )
    : m_pcDataConst( rcPlane.m_pData )
    , m_cSize( rcPlane.m_iWidth, rcPlane.m_iHeight )
    , m_uiStride( rcPlane.m_uiStride )
  {
    CAOF( (TypeCompare<t_El2,t_El>::m_bEqual) || (TypeCompare<t_El2,const t_El>::m_bEqual) );
  }

  Block( const t_El *pcData, const BlockSize &rcSize, UInt uiStride ) : m_pcDataConst( pcData ), m_cSize( rcSize ), m_uiStride( uiStride ) {}
  Block( const PixelPos& cPos, const BlockSize& cSize, const Block &rcRefBlock )
    : m_pcDataConst( &rcRefBlock.getEl( cPos ) )
    , m_cSize( cSize )
    , m_uiStride( rcRefBlock.getStride() )
  {
//    AOF_DBG( cPos + cSize <= rcRefBlock.getSize() );
//    AOF_DBG( cPos >= PixelPos( 0, 0 ) );
    AOF_DBG( cSize > BlockSize( 0, 0 ) );
  }

  const BlockSize& getSize()   const { return m_cSize; }
  UInt             getStride() const { return m_uiStride; }

  const t_El* getPtr() const { return m_pcDataConst; }
  const t_El& getEl( UInt ui ) const
  {
#if DEBUG_BLOCK_ACCESS
    if( xCheckRangeError( PixelPos( ui % getStride(), ui / getStride() ) ) )
    {
      AOT( 1 );
      while( true ) std::cout << "Error in getEl" << std::endl;
    }
#endif
    return m_pcDataConst[ui];
  }

  const t_El& getEl( const PixelPos &rcPos ) const
  {
#if DEBUG_BLOCK_ACCESS
    if( xCheckRangeError( rcPos ) )
    {
      AOT( 1 );
      while( true ) std::cout << "Error in getEl" << std::endl;
    }
#endif
    const Int64 iIndex = (Int64)(rcPos.getY() * (int)getStride() + rcPos.getX());
    return *(m_pcDataConst + iIndex);
  }

  const t_El& getEl( const UInt uiX, const UInt uiY ) const  { return getEl( PixelPos( uiX, uiY));  }

  Block& operator=( const Block &rcBlock )
  {
    m_pcDataConst = rcBlock.m_pcDataConst;
    m_cSize = rcBlock.m_cSize;
    m_uiStride = rcBlock.m_uiStride;
    xClearNonConst();
    return *this;
  }

protected:
  bool xCheckRangeError( const PixelPos &rcPos ) const
  {
    return ! ( PixelPos( 0, 0 ) <= rcPos && rcPos < getSize() );
  }

  void xSetBlock( const t_El *pcData, const BlockSize &rcSize, UInt uiStride )
  {
    m_pcDataConst = pcData;
    m_cSize = rcSize;
    m_uiStride = uiStride;
  }

private:
  const t_El *m_pcDataConst;
  BlockSize m_cSize;
  UInt m_uiStride;

  virtual void xClearNonConst() {}
};


template<class t_El>
class Block : public Block<const t_El>
{
public:
  Block() : Block<const t_El>(), m_pcData( NULL ) {}
  virtual ~Block() {}

  Block( const VideoCodecPlane<t_El> &rcPlane )
    : Block<const t_El>( rcPlane )
    , m_pcData( rcPlane.m_pData ) {}

  Block( t_El *pcData, const BlockSize &rcSize, UInt uiStride ) : Block<const t_El>( pcData, rcSize, uiStride ), m_pcData( pcData ) {}
  Block( const PixelPos& cPos, const BlockSize& cSize, const Block<t_El> &rcRefBlock )
    : Block<const t_El>( cPos, cSize, rcRefBlock )
    , m_pcData( &rcRefBlock.getEl( cPos ) ) {}

  t_El* getPtr() const { return m_pcData; }
  t_El& getEl( UInt ui ) const
  {
#if DEBUG_BLOCK_ACCESS
    if( Block<const t_El>::xCheckRangeError( PixelPos( ui % Block<const t_El>::getStride(), ui / Block<const t_El>::getStride() ) ) )
    {
      AOT( 1 );
      while( true ) std::cout << "Error in getEl" << std::endl;
    }
#endif
    return m_pcData[ui];
  }


  t_El& getEl( const PixelPos &rcPos ) const
  {
#if DEBUG_BLOCK_ACCESS
    if( Block<const t_El>::xCheckRangeError( rcPos ) )
    {
      AOT( 1 );
      while( true ) std::cout << "Error in getEl" << std::endl;
    }
#endif
    const Int64 iIndex = (Int64)(rcPos.getY() * (int)Block<const t_El>::getStride() + rcPos.getX());
    return *(m_pcData + iIndex);
  }

  t_El& getEl(const UInt uiX, const UInt uiY ) const { return Block<t_El>::getEl( PixelPos( uiX, uiY ) ); }

#undef DEBUG_BLOCK_ACCESS

  template<class U>
  void assignAndDivide( const Block<const U>& rcBlock, const int iDiv) const
  {
    int iShift;
    int iAdd;
    if( iDiv == 2 )
    {
      iAdd = 1;
      iShift = 1;
    }
    else if( iDiv == 4 )
    {
      iAdd = 2;
      iShift = 2;
    }
    else if( iDiv == 8 )
    {
      iAdd = 4;
      iShift = 3;
    }
    else
    {
      const U i2Div         = 2*iDiv;
      t_El* pcDesData       = m_pcData;
      const int iDesStride  = Block<const t_El>::getStride();
      const U* pcSrcData    = rcBlock.getPtr();
      const int iSrcStride  = rcBlock.getStride();
      const int iWidth      = rcBlock.getSize().getX();
      const int iHeight     = rcBlock.getSize().getY();
        
        
      for( int iY = 0; iY < iHeight; iY++ )
      {
        for( int iX = 0; iX < iWidth; iX++ )
        {
          U eU = 2*pcSrcData[iX];
          eU += pcSrcData[iX] < 0 ? -iDiv : iDiv;
          eU /= i2Div;
          pcDesData[iX] = (t_El)eU;
        }
        pcSrcData += iSrcStride;
        pcDesData += iDesStride;
      }
      return;
    }

    t_El* pcDesData       = m_pcData;
    const int iDesStride  = Block<const t_El>::getStride();
    const U* pcSrcData    = rcBlock.getPtr();
    const int iSrcStride  = rcBlock.getStride();
    const int iWidth      = rcBlock.getSize().getX();
    const int iHeight     = rcBlock.getSize().getY();
    for( int iY = 0; iY < iHeight; iY++ )
    {
      for( int iX = 0; iX < iWidth; iX++ )
      {
        U eU = pcSrcData[iX];
        eU += iAdd;
        eU >>= iShift;
        pcDesData[iX] = (t_El)eU;
      }
      pcSrcData += iSrcStride;
      pcDesData += iDesStride;
    }
  }

  template<int t_ElAdd, int t_ElShift, class t_Int, class U>
  void addShiftAndClip( const Block<const U>& rcBlock, const t_Int &t_ElMin, const t_Int &t_ElMax ) const
  {
    t_El* pcDesData       = m_pcData;
    const int iDesStride  = Block<const t_El>::getStride();
    const U* pcSrcData    = rcBlock.getPtr();
    const int iSrcStride  = rcBlock.getStride();
    const int iWidth      = rcBlock.getSize().getX();
    const int iHeight     = rcBlock.getSize().getY();

    for( int iY = 0; iY < iHeight; iY++ )
    {
      for( int iX = 0; iX < iWidth; iX++ )
      {
        t_Int eEl = (t_Int( pcDesData[iX] ) + t_Int( pcSrcData[iX] ) + t_ElAdd)>>t_ElShift;
        if( eEl < t_ElMin )
        {
          eEl = t_ElMin;
        }
        else if( eEl > t_ElMax )
        {
          eEl = t_ElMax;
        }
        pcDesData[iX] = (t_El)eEl;
      }
      pcSrcData += iSrcStride;
      pcDesData += iDesStride;
    }
  }


  template<int t_ElAdd, int t_ElShift, class t_Int, const t_Int t_ElMin, const t_Int t_ElMax, class U>
  void assignShiftAndClip( const Block<const U>& rcBlock) const
  {
    t_El* pcDesData       = m_pcData;
    const int iDesStride  = Block<const t_El>::getStride();
    const U* pcSrcData    = rcBlock.getPtr();
    const int iSrcStride  = rcBlock.getStride();
    const int iWidth      = rcBlock.getSize().getX();
    const int iHeight     = rcBlock.getSize().getY();

    for( int iY = 0; iY < iHeight; iY++ )
    {
      for( int iX = 0; iX < iWidth; iX++ )
      {
        t_Int eEl = ( t_Int( pcSrcData[iX] ) + t_ElAdd)>>t_ElShift;
        if( (UInt)eEl > t_ElMax ) 
        {
          eEl = ( eEl > t_ElMax ) ? t_ElMax : t_ElMin;
        }
/*
        if( eEl < t_ElMin )
        {
          eEl = t_ElMin;
        }
        else if( eEl > t_ElMax )
        {
          eEl = t_ElMax;
        }
*/
        pcDesData[iX] = (t_El)eEl;
      }
      pcSrcData += iSrcStride;
      pcDesData += iDesStride;
    }
  }


  void clip( const t_El &rcElMin, const t_El &rcElMax ) const
  {
    UInt uiDst = 0;
    for( int y = 0; y < Block<const t_El>::getSize().getY(); y++ )
    {
      UInt uiDstRun = uiDst;
      for( int x = 0; x < Block<const t_El>::getSize().getX(); x++ )
      {
        t_El &rcEl = getEl( uiDstRun++ );
        if( rcEl < rcElMin )
        {
          rcEl = rcElMin;
        }
        else if( rcEl > rcElMax )
        {
          rcEl = rcElMax;
        }
      }
      uiDst += Block<const t_El>::getStride();
    }
  }

  void drawBorder( bool bLeft, bool bTop, bool bRight, bool bBottom ) const
  {
    UInt uiDst = 0;
    for( int y = 0; y < Block<const t_El>::getSize().getY(); y++ )
    {
      UInt uiDstRun = uiDst;
      for( int x = 0; x < Block<const t_El>::getSize().getX(); x++ )
      {
        if ( ( bTop && y == 0 )
          || ( bLeft && x == 0 )
          || ( bRight && x == Block<const t_El>::getSize().getX() - 1 )
          || ( bBottom && y == Block<const t_El>::getSize().getY() - 1 ) )
          getEl( uiDstRun++ ) = 0;
      }
      uiDst += Block<const t_El>::getStride();
    }
  }

  template<class U>
  ErrVal assign( const Block<const U>& rcBlock ) const
  {
    AOT_DBG( rcBlock.getSize() != Block<const t_El>::getSize() );
//    ROF( rcBlock.getSize() == Block<const t_El>::getSize() );

    t_El* pDes            = getPtr();
    const int iDesStride  = Block<const t_El>::getStride();
    const U* pSrc         = rcBlock.getPtr();
    const int iSrcStride  = rcBlock.getStride();
    const int iWidth      = rcBlock.getSize().getX();
    const int iHeight     = rcBlock.getSize().getY();

    for( int y = 0; y < iHeight; y++ )
    {
      for( int x = 0; x < iWidth; x++ )
      {
        pDes[x] = (t_El)pSrc[x];
      }
      pSrc += iSrcStride;
      pDes += iDesStride;
    }

    return Err::m_nOK;
  }

  ErrVal padding( const int iMarginX, const int iMarginY ) const
  {
    const int iStride = Block<const t_El>::getStride();
    const int iWidth  = Block<const t_El>::getSize().getX();
    const int iHeight = Block<const t_El>::getSize().getY();
    t_El* pEl         = getPtr();
    for( int y = 0; y < iHeight; y++ )
    {
      // left margin
      {
        const t_El eEl = pEl[0];
        for( int n = -iMarginX; 0 > n; n++ )
        {
          pEl[n] = eEl;
        }
      }

      // right margin
      {
        const t_El eEl = pEl[iWidth-1];
        for( int n = 0; n < iMarginX; n++ )
        {
          pEl[iWidth+n] = eEl;
        }
      }
      pEl+= iStride;
    }

    const int iWidthPlusMargin = iWidth + iMarginX;
    // bottom margin
    for( int y = 0; y < iMarginY; y++ )
    {
      for( int x = -iMarginX; x < iWidthPlusMargin; x++ )
      {
        pEl[x] = pEl[x-iStride];
      }
      pEl+= iStride;
    }

    // top margin
    pEl = getPtr();
    for( int y = 0; y < iMarginY; y++ )
    {
      for( int x = -iMarginX; x < iWidthPlusMargin; x++ )
      {
        pEl[x-iStride] = pEl[x];
      }
      pEl-= iStride;
    }

    return Err::m_nOK;
  }

  template<class U>
  const Block& operator+=( const Block<const U>& rcBlock ) const
  {
    AOF_DBG( rcBlock.getSize() == Block<const t_El>::getSize() );

    t_El* pDes            = getPtr();
    const int iDesStride  = Block<const t_El>::getStride();
    const U* pSrc         = rcBlock.getPtr();
    const int iSrcStride  = rcBlock.getStride();
    const int iWidth      = rcBlock.getSize().getX();
    const int iHeight     = rcBlock.getSize().getY();

    for( int y = 0; y < iHeight; y++ )
    {
      for( int x = 0; x < iWidth; x++ )
      {
        pDes[x] += (t_El)pSrc[x];
      }
      pSrc += iSrcStride;
      pDes += iDesStride;
    }

    return *this;
  }

  template<class U>
  const Block& operator-=( const Block<const U>& rcBlock ) const
  {
    AOF_DBG( rcBlock.getSize() == Block<const t_El>::getSize() );

    t_El* pDes            = getPtr();
    const int iDesStride  = Block<const t_El>::getStride();
    const U* pSrc         = rcBlock.getPtr();
    const int iSrcStride  = rcBlock.getStride();
    const int iWidth      = rcBlock.getSize().getX();
    const int iHeight     = rcBlock.getSize().getY();

    for( int y = 0; y < iHeight; y++ )
    {
      for( int x = 0; x < iWidth; x++ )
      {
        pDes[x] -= (t_El)pSrc[x];
      }
      pSrc += iSrcStride;
      pDes += iDesStride;
    }
    return *this;
  }

  template<class U>
  const Block& operator*=( const Block<const U>& rcBlock ) const
  {
    AOF_DBG( rcBlock.getSize() == Block<const t_El>::getSize() );

    t_El* pDes            = getPtr();
    const int iDesStride  = Block<const t_El>::getStride();
    const U* pSrc         = rcBlock.getPtr();
    const int iSrcStride  = rcBlock.getStride();
    const int iWidth      = rcBlock.getSize().getX();
    const int iHeight     = rcBlock.getSize().getY();

    for( int y = 0; y < iHeight; y++ )
    {
      for( int x = 0; x < iWidth; x++ )
      {
        pDes[x] *= (t_El)pSrc[x];
      }
      pSrc += iSrcStride;
      pDes += iDesStride;
    }
    return *this;
  }

  const Block& operator*= ( const t_El &rcEl ) const;
  const Block& operator/= ( const t_El &rcEl ) const;
  const Block& operator+= ( const t_El &rcEl ) const;
  void         assign     ( const t_El &rcEl ) const;
  const Block& operator>>=( int iShift ) const;
  const Block& operator<<=( int iShift ) const;

protected:
  void xSetBlock( t_El *pcData, const BlockSize &rcSize, UInt uiStride )
  {
    m_pcData = pcData;
    Block<const t_El>::xSetBlock( pcData, rcSize, uiStride );
  }

private:
  t_El *m_pcData;

  void xClearNonConst() { m_pcData = NULL; }
};


template<class t_El>
class AllocBlock : public Block<t_El>
{
public:
  AllocBlock() : m_pcAllocatedBuffer( NULL ) {}

  AllocBlock( const BlockSize& cSize ) : Block<t_El>( new t_El[cSize.getX() * cSize.getY()], cSize, cSize.getX() )
  {
    m_pcAllocatedBuffer = &Block<t_El>::getEl( 0 );
    AOF_DBG( m_pcAllocatedBuffer );
  }

  ErrVal allocateBlock( const BlockSize& cSize )
  {
    ROT( m_pcAllocatedBuffer );
    m_pcAllocatedBuffer = new t_El[cSize.getX() * cSize.getY()];
    Block<t_El>::xSetBlock( m_pcAllocatedBuffer, cSize, cSize.getX() );
    return Err::m_nOK;
  }

  virtual ~AllocBlock()
  {
    delete[] m_pcAllocatedBuffer;
  }

  DISABLE_DEFAULT_CPYCTOR_AND_ASOP(AllocBlock);
private:
  t_El *m_pcAllocatedBuffer;
};


template<class t_El>
const Block<t_El>& Block<t_El>::operator*=( const t_El &rcEl ) const
{
  t_El* pDes            = getPtr();
  const int iDesStride  = Block<const t_El>::getStride();
  const int iWidth      = Block<const t_El>::getSize().getX();
  const int iHeight     = Block<const t_El>::getSize().getY();

  for( int y = 0; y < iHeight; y++ )
  {
    for( int x = 0; x < iWidth; x++ )
    {
      pDes[x] *= rcEl;
    }
    pDes += iDesStride;
  }

  return *this;
}

template<class t_El>
const Block<t_El>& Block<t_El>::operator/=( const t_El &rcEl ) const
{
  t_El* pDes            = getPtr();
  const int iDesStride  = Block<const t_El>::getStride();
  const int iWidth      = Block<const t_El>::getSize().getX();
  const int iHeight     = Block<const t_El>::getSize().getY();

  for( int y = 0; y < iHeight; y++ )
  {
    for( int x = 0; x < iWidth; x++ )
    {
      pDes[x] /= rcEl;
    }
    pDes += iDesStride;
  }
  return *this;
}

template<class t_El>
const Block<t_El>& Block<t_El>::operator+=( const t_El &rcEl ) const
{
  t_El* pDes            = getPtr();
  const int iDesStride  = Block<const t_El>::getStride();
  const int iWidth      = Block<const t_El>::getSize().getX();
  const int iHeight     = Block<const t_El>::getSize().getY();

  for( int y = 0; y < iHeight; y++ )
  {
    for( int x = 0; x < iWidth; x++ )
    {
      pDes[x] += rcEl;
    }
    pDes += iDesStride;
  }
  return *this;
}


template<class t_El>
void Block<t_El>::assign( const t_El &rcEl ) const
{
  t_El* pDes            = getPtr();
  const int iDesStride  = Block<const t_El>::getStride();
  const int iWidth      = Block<const t_El>::getSize().getX();
  const int iHeight     = Block<const t_El>::getSize().getY();

  for( int y = 0; y < iHeight; y++ )
  {
    for( int x = 0; x < iWidth; x++ )
    {
      pDes[x] = rcEl;
    }
    pDes += iDesStride;
  }
}

template<class t_El>
const Block<t_El>& Block<t_El>::operator>>=( int iShift ) const
{
  t_El* pDes            = getPtr();
  const int iDesStride  = Block<const t_El>::getStride();
  const int iWidth      = Block<const t_El>::getSize().getX();
  const int iHeight     = Block<const t_El>::getSize().getY();

  for( int y = 0; y < iHeight; y++ )
  {
    for( int x = 0; x < iWidth; x++ )
    {
      pDes[x] >>= iShift;
    }
    pDes += iDesStride;
  }
  return *this;
}

template<class t_El>
const Block<t_El>& Block<t_El>::operator<<=( int iShift ) const
{
  t_El* pDes            = getPtr();
  const int iDesStride  = Block<const t_El>::getStride();
  const int iWidth      = Block<const t_El>::getSize().getX();
  const int iHeight     = Block<const t_El>::getSize().getY();

  for( int y = 0; y < iHeight; y++ )
  {
    for( int x = 0; x < iWidth; x++ )
    {
      pDes[x] <<= iShift;
    }
    pDes += iDesStride;
  }
  return *this;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _BLOCK_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
