/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _INTERCOMMON_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _INTERCOMMON_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/PicBlock.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class MotionVector
{
public:
  explicit MotionVector( int iX = 0, int iY = 0, int iMvShift = 0 ) : m_cMv( iX , iY ), m_iMvShift( iMvShift ) {}

  MotionVector getMvWithPrecisionIncreasedTo( int iNewPrecision ) const
  {
    AOT_DBG( iNewPrecision < m_iMvShift );
    return MotionVector( m_cMv << ( iNewPrecision - m_iMvShift ), iNewPrecision );
  }

  MotionVector getMvWithPrecisionDecreasedTo( int iNewPrecision ) const
  {
    const int iRightShift = m_iMvShift - iNewPrecision;
    const int iAdd = 1 << iRightShift;
    AOT_DBG( iRightShift < 0 );
    return MotionVector( ( ( m_cMv << 1 ) + PrimPixelPos( iAdd, iAdd ) ) >> ( iRightShift + 1 ), iNewPrecision );
  }

  MotionVector getMvWithPrecisionModifiedTo( int iNewPrecision ) const
  {
    if( m_iMvShift <= iNewPrecision )
    {
      return getMvWithPrecisionIncreasedTo( iNewPrecision );
    }
    return getMvWithPrecisionDecreasedTo( iNewPrecision );
  }

  MotionVector plusPrimResPos ( const PrimPixelPos &rcPos ) const { return MotionVector( m_cMv + (rcPos << m_iMvShift), m_iMvShift ); }
  MotionVector minusPrimResPos( const PrimPixelPos &rcPos ) const { return MotionVector( m_cMv - (rcPos << m_iMvShift), m_iMvShift ); }

  MotionVector operator+( const MotionVector &rcRight ) const
  {
    AOF_DBG( m_iMvShift == rcRight.m_iMvShift );
    return MotionVector( m_cMv + rcRight.m_cMv, m_iMvShift );
  }

  MotionVector operator-( const MotionVector &rcRight ) const
  {
    AOF_DBG( m_iMvShift == rcRight.m_iMvShift );
    return MotionVector( m_cMv - rcRight.m_cMv, m_iMvShift );
  }

  MotionVector operator*( int i ) const
  {
    return MotionVector( m_cMv * i, m_iMvShift );
  }

  template<UInt t_uiAcc>
  ErrVal getNearestTopLeftPosAndOffset( const Subsampling &rcSub, PixelPos &rcNearestTopLeftPos, BinaryFraction<t_uiAcc> &rcOffsetX, BinaryFraction<t_uiAcc> &rcOffsetY ) const
  {
    const UInt uiAccIncrease = 1;
    PixelPos cPos = (m_cMv << uiAccIncrease) * rcSub;
    rcNearestTopLeftPos = cPos >> (m_iMvShift + uiAccIncrease);
    PixelPos cMod = cPos & ((1 << (m_iMvShift + uiAccIncrease) ) - 1);
    AOT_DBG( m_iMvShift + uiAccIncrease > int( t_uiAcc ) );
    rcOffsetX.setNumerator( Int64( cMod.getX() ) << ( t_uiAcc - (m_iMvShift + uiAccIncrease) ) );
    rcOffsetY.setNumerator( Int64( cMod.getY() ) << ( t_uiAcc - (m_iMvShift + uiAccIncrease) ) );
    return Err::m_nOK;
  } 

  const int& getX() const { return m_cMv.getX(); }
  const int& getY() const { return m_cMv.getY(); }

  int& getX() { return m_cMv.getX(); }
  int& getY() { return m_cMv.getY(); }

  const int& getPrecision() const { return m_iMvShift; }
  int& getPrecision() { return m_iMvShift; }

  bool operator==( const MotionVector &r ) const { return m_cMv == r.m_cMv && m_iMvShift == r.m_iMvShift; }
  bool operator!=( const MotionVector &r ) const { return !(*this == r); }

private:
  PrimPixelPos m_cMv;
  int m_iMvShift;

  explicit MotionVector( const PrimPixelPos& cPos, int iMvShift = 0 ) : m_cMv( cPos ), m_iMvShift( iMvShift ) {}
};


class FracPos
{
private:
  double m_dX;
  double m_dY;

public:
  FracPos() {}
  FracPos( double dX, double dY) { m_dX = dX; m_dY = dY; }
  virtual ~FracPos() {}

  double& getX() { return m_dX; }
  double& getY() { return m_dY; }

  const double& getX() const { return m_dX; }
  const double& getY() const { return m_dY; }

  FracPos operator+( const FracPos &rcRight ) const
  {
    return FracPos( m_dX + rcRight.m_dX, m_dY + rcRight.m_dY );
  }

  FracPos operator-( const FracPos &rcRight ) const
  {
    return FracPos( m_dX - rcRight.m_dX, m_dY - rcRight.m_dY );
  }

};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _INTERCOMMON_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
