/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_MULTITHREADING_H__B87E26CF_023E_4DC7_8F94_D3E38F59ABA1__INCLUDED_)
#define AFX_MULTITHREADING_H__B87E26CF_023E_4DC7_8F94_D3E38F59ABA1__INCLUDED_

#if ALLOW_BOOST_LIBRARY
#include <boost/thread.hpp>

// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class T>
class ProducerConsumerFIFO
{
  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( ProducerConsumerFIFO );
public:
  ProducerConsumerFIFO() : m_bTerminateWhenFIFOEmpty( false ) {}
  virtual ~ProducerConsumerFIFO() {}

  void terminateWhenEmpty()
  {
    {
      boost::unique_lock<boost::mutex> cLock( m_cMutex );
      m_bTerminateWhenFIFOEmpty = true;
    }
    m_cCondition.notify_all();
  }

  void pushBack( const T &r )
  {
    {
      boost::unique_lock<boost::mutex> cLock( m_cMutex );
      m_aT.push_back( r );
    }
    m_cCondition.notify_all();
  }

  T tryPopFront( bool &bSuccess, bool bWaitForOneItem = true )
  {
    boost::unique_lock<boost::mutex> cLock( m_cMutex );
    if( bWaitForOneItem )
    {
      while( m_aT.empty() && !m_bTerminateWhenFIFOEmpty )
      {
        m_cCondition.wait( cLock );
      }
    }
    if( m_aT.empty() )
    {
      bSuccess = false;
      return T();
    }
    bSuccess = true;
    T x = m_aT.front();
    m_aT.pop_front();
    return x;
  }

private:
  std::list<T> m_aT;
  boost::mutex m_cMutex;
  boost::condition_variable m_cCondition;
  bool m_bTerminateWhenFIFOEmpty;
};


template<class t_Worker>
class MultiThreadedFIFO
{
  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( MultiThreadedFIFO );

  typedef typename t_Worker::WorkerInput     InputType;
  typedef typename t_Worker::WorkerOutput    OutputType;
  typedef typename t_Worker::WorkerException ExceptionType;

public:
  MultiThreadedFIFO() {}
  virtual ~MultiThreadedFIFO() { TNOK( joinAllThreads() ); }

  ErrVal                                        addThread()                            { return xAddThread( new t_Worker() ); }
  template<class T1> ErrVal                     addThread( T1 &rT1 )                   { return xAddThread( new t_Worker( rT1 ) ); }
  template<class T1, class T2> ErrVal           addThread( T1 &rT1, T2 &rT2 )          { return xAddThread( new t_Worker( rT1, rT2 ) ); }
  template<class T1, class T2, class T3> ErrVal addThread( T1 &rT1, T2 &rT2, T3 &rT3 ) { return xAddThread( new t_Worker( rT1, rT2, rT3 ) ); }

  UInt getNumThreads() const { return UInt( m_apcThreads.size() ); }
  ErrVal joinAllThreads();

  ProducerConsumerFIFO<InputType>& getInput() { return m_cWorkFIFO; }
  ProducerConsumerFIFO<OutputType>& getOutput() { return m_cOutputFIFO; }

private:
  class WorkerThread
  {
  public:
    WorkerThread( t_Worker *p,
                  ProducerConsumerFIFO<InputType> &rcInFIFO,
                  ProducerConsumerFIFO<OutputType> &rcOutFIFO,
                  ProducerConsumerFIFO<ExceptionType> &rcExceptinFIFO )
      : m_pcWorker( p )
      , m_pcInFIFO( &rcInFIFO )
      , m_pcOutFIFO( &rcOutFIFO )
      , m_pcExceptinFIFO( &rcExceptinFIFO ) {}

    void operator() ()
    {
      while( true )
      {
        bool bSuccess = true;
        InputType cWorkItem = m_pcInFIFO->tryPopFront( bSuccess );
        if( !bSuccess)
        {
          break;
        }
        try
        {
          m_pcOutFIFO->pushBack( (*m_pcWorker)( cWorkItem ) );
        }
        catch( ExceptionType &cEx )
        {
          m_pcExceptinFIFO->pushBack( cEx );
          m_pcInFIFO->terminateWhenEmpty();
          m_pcOutFIFO->terminateWhenEmpty();
          break;
        }
        catch( ... )
        {
          m_pcExceptinFIFO->pushBack( ExceptionType() );
          m_pcInFIFO->terminateWhenEmpty();
          m_pcOutFIFO->terminateWhenEmpty();
          break;
        }
      }
    }

    t_Worker *m_pcWorker;

  private:
    ProducerConsumerFIFO<InputType> *m_pcInFIFO;
    ProducerConsumerFIFO<OutputType> *m_pcOutFIFO;
    ProducerConsumerFIFO<ExceptionType> *m_pcExceptinFIFO;
  };

  ErrVal xAddThread( t_Worker *p );

  ProducerConsumerFIFO<InputType> m_cWorkFIFO;
  ProducerConsumerFIFO<OutputType> m_cOutputFIFO;
  ProducerConsumerFIFO<ExceptionType> m_cExceptinFIFO;
  std::vector<boost::thread*> m_apcThreads;
  std::vector<WorkerThread*> m_apcWorkers;
};


template<class t_Worker>
ErrVal MultiThreadedFIFO<t_Worker>::xAddThread( t_Worker *p )
{
  m_apcWorkers.push_back( new WorkerThread( p, m_cWorkFIFO, m_cOutputFIFO, m_cExceptinFIFO ) );
  m_apcThreads.push_back( new boost::thread( boost::ref( *m_apcWorkers.back() ) ) );
  return Err::m_nOK;
}


template<class t_Worker>
ErrVal MultiThreadedFIFO<t_Worker>::joinAllThreads()
{
  getInput().terminateWhenEmpty();
  getOutput().terminateWhenEmpty();

  for( UInt ui = 0; ui < UInt( m_apcThreads.size() ); ui++ )
  {
    m_apcThreads[ui]->join();
    delete m_apcThreads[ui];
    delete m_apcWorkers[ui]->m_pcWorker;
    delete m_apcWorkers[ui];
  }
  m_apcThreads.clear();
  m_apcWorkers.clear();

  m_cExceptinFIFO.terminateWhenEmpty();

  while( true )
  {
    bool bSuccess = true;
    ExceptionType cException = m_cExceptinFIFO.tryPopFront( bSuccess );
    if( !bSuccess )
    {
      break;
    }
    throw cException; // currently one only
  }
  return Err::m_nOK;
}


class WorkerExample
{
public:
  // necessary typedefs
  typedef int    WorkerInput;
  typedef int    WorkerOutput;
  typedef ErrVal WorkerException;

  // thread function
  WorkerOutput operator() ( WorkerInput &rcWorkItem )
  {
    if( rcWorkItem == 0 )
    {
      throw Err::m_nERR;
    }
    return rcWorkItem + 7;
  }

  // excample application
  static ErrVal testWorkerExample()
  {
    MultiThreadedFIFO<WorkerExample> cExampleInstance;

    RNOK( cExampleInstance.addThread() );
    RNOK( cExampleInstance.addThread() );
    RNOK( cExampleInstance.addThread() );

    cExampleInstance.getInput().pushBack( 2 );
    cExampleInstance.getInput().pushBack( 5 );
    cExampleInstance.getInput().pushBack( 7 );

    RNOK( cExampleInstance.joinAllThreads() );

    while( true )
    {
      bool bSuccess = true;
      WorkerOutput cOutput = cExampleInstance.getOutput().tryPopFront( bSuccess );
      if( bSuccess )
      {
        std::cout << cOutput << std::endl;
      }
      else
      {
        break;
      }
    }
    return Err::m_nOK;
  }
};

// video namespace end
VIDEO_NAMESPACE_END

#endif

#endif // !defined(AFX_MULTITHREADING_H__B87E26CF_023E_4DC7_8F94_D3E38F59ABA1__INCLUDED_)
