/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _PICBLOCK_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _PICBLOCK_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "Block.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class PicBlockStats
{
public:
  PicBlockStats() {}
  PicBlockStats( const PrimBlockSize &rcPrimBlockSize, const std::vector<Subsampling> &rcSubsampling )
    : m_acSubsampling( rcSubsampling )
    , m_cPrimBlockSize( rcPrimBlockSize ) {}

  PicBlockStats( const std::vector<BlockSize> &racBlockSizes )
    : m_acSubsampling( racBlockSizes.size(), Subsampling() )
  {
    AOT( racBlockSizes.empty() );
    m_cPrimBlockSize.getX() = 0;
    m_cPrimBlockSize.getY() = 0;
    for( UInt ui = 0; ui < racBlockSizes.size(); ui++ )
    {
      m_cPrimBlockSize.getX() = gMax( m_cPrimBlockSize.getX(), racBlockSizes[ui].getX() );
      m_cPrimBlockSize.getY() = gMax( m_cPrimBlockSize.getY(), racBlockSizes[ui].getY() );
    }

    for( UInt ui = 0; ui < racBlockSizes.size(); ui++ )
    {
      m_acSubsampling[ui] = racBlockSizes[ui] / m_cPrimBlockSize;
    }
  }

  const std::vector<Subsampling>& getSubsampling()   const { return m_acSubsampling; }
  const PrimBlockSize&            getPrimBlockSize() const { return m_cPrimBlockSize; }
  UInt                            getNumPlanes()     const { return UInt( m_acSubsampling.size() ); }

  bool operator!=( const PicBlockStats &rcRight ) const { return !(*this == rcRight); }
  bool operator==( const PicBlockStats &rcRight ) const
  {
    if( m_cPrimBlockSize != rcRight.m_cPrimBlockSize )
      return false;
    if( m_acSubsampling.size() != rcRight.m_acSubsampling.size() )
      return false;
    for( UInt ui = 0; ui < m_acSubsampling.size(); ui ++ )
      if( m_acSubsampling[ui] != rcRight.m_acSubsampling[ui] )
        return false;
    return true;
  }

private:
  std::vector<Subsampling> m_acSubsampling;    //!< Contains the subsampling belonging to each plane
  PrimBlockSize            m_cPrimBlockSize;   //!< Contains the primary block size
};


template<class t_El> class PicBlock;

template<class t_El>
class PicBlock<const t_El>
{
public:
  PicBlock() {}
  virtual ~PicBlock() {}

  PicBlock( const PrimPixelPos& rcPos, const PrimPixelPos& rcSize, const PicBlock<const t_El> &rcRefPicBlock )
    : m_acBlocksConst( rcRefPicBlock.getStats().getNumPlanes() )
    , m_cStats( rcSize, rcRefPicBlock.getStats().getSubsampling() )
  {
    xCreateSubPicBlock<const t_El>( m_acBlocksConst, rcPos, rcSize, rcRefPicBlock );
  }

  PicBlock( const PicBlock<const t_El> &rcRefPicBlock, UInt uiNumPlanes )
    : m_acBlocksConst( rcRefPicBlock.m_acBlocksConst.begin(), rcRefPicBlock.m_acBlocksConst.begin() + uiNumPlanes )
    , m_cStats( xCalculateStats( rcRefPicBlock.getStats(), uiNumPlanes ) )
  { 
  }

  PicBlock( const PicBlock<const t_El> &rcRefPicBlock, UInt uiPrimPlaneId, UInt uiNumPlanes, const Subsampling& rcPrimPlaneSubsampling )
    : m_acBlocksConst( rcRefPicBlock.m_acBlocksConst.begin() + uiPrimPlaneId, rcRefPicBlock.m_acBlocksConst.begin() + uiPrimPlaneId + uiNumPlanes )
    , m_cStats( xCalculateStats( rcRefPicBlock.getStats(), uiPrimPlaneId, uiNumPlanes, rcPrimPlaneSubsampling ) )
  { 
  }

  template<class t_Block>
  PicBlock( const std::vector<t_Block> &rcBlocks, const PicBlockStats *pcStats = NULL )
    : m_acBlocksConst( rcBlocks.begin(), rcBlocks.end() )
    , m_cStats( pcStats ? *pcStats : xCalculateStats( m_acBlocksConst )  )
  {
    CAOF( (TypeCompare<t_Block,Block<t_El> >::m_bEqual) || (TypeCompare<t_Block,Block<const t_El> >::m_bEqual) ||
          (TypeCompare<t_Block,VideoCodecPlane<t_El> >::m_bEqual) || (TypeCompare<t_Block,VideoCodecPlane<const t_El> >::m_bEqual) );
  }

  PicBlock& operator=( const PicBlock &rcBlock )
  {
    m_acBlocksConst = rcBlock.m_acBlocksConst;
    m_cStats = rcBlock.m_cStats;
    xClearNonConst();
    return *this;
  }

  const Block<const t_El>& getPlane( UInt ui ) const { return m_acBlocksConst[ui]; }

  const PicBlockStats& getStats() const { return m_cStats; }

protected:
  template<class U>
  static PicBlockStats xCalculateStats( const std::vector<U>& racBlocks )
  {
    AOT( racBlocks.empty() );
    std::vector<BlockSize> acSizes( racBlocks.size() );
    for( UInt ui = 0; ui < racBlocks.size(); ui++ )
    {
      acSizes[ui] = racBlocks[ui].getSize();
    }
    return PicBlockStats( acSizes );
  }

  static PicBlockStats xCalculateStats( const PicBlockStats& rcOrgStats, UInt uiNumPlanes )
  {
    AOF( rcOrgStats.getNumPlanes() >= uiNumPlanes );
    std::vector<Subsampling> acSub( uiNumPlanes );
    for( UInt uiPlane = 0; uiPlane < uiNumPlanes; uiPlane++ )
    {
      acSub[ uiPlane ] = rcOrgStats.getSubsampling()[ uiPlane ];
    }
    return PicBlockStats( rcOrgStats.getPrimBlockSize(), acSub );
  }

  static PicBlockStats xCalculateStats( const PicBlockStats& rcOrgStats, UInt uiPrimPlaneId, UInt uiNumPlanes, const Subsampling& rcPrimPlaneSubsampling )
  {
    AOF( rcOrgStats.getNumPlanes() >= uiPrimPlaneId + uiNumPlanes );
    std::vector<Subsampling> acSub( uiNumPlanes );
    for( UInt uiPlane = 0; uiPlane < uiNumPlanes; uiPlane++ )
    {
      acSub[ uiPlane ] = rcOrgStats.getSubsampling()[ uiPlane + uiPrimPlaneId ] / rcPrimPlaneSubsampling;
    }
    BlockSize cPrimSize = rcOrgStats.getPrimBlockSize() * rcPrimPlaneSubsampling;
    return PicBlockStats( PrimBlockSize( cPrimSize.getX(), cPrimSize.getY() ), acSub );
  }

  template<class U>
  static void xCreateSubPicBlock( std::vector<Block<U> > &rcVec, const PrimPixelPos &rcPos, const PrimBlockSize &rcSize, const PicBlock<U> &rcBlock )
  {
    const PicBlockStats &rcStats = rcBlock.getStats();
    AOF_DBG( rcVec.size() == rcStats.getNumPlanes() );
//    AOF_DBG( rcPos + rcSize <= rcStats.getPrimBlockSize() );
    for( UInt ui = 0; ui < rcStats.getNumPlanes(); ui++ )
    {
      const Subsampling &rcSub = rcStats.getSubsampling()[ui];
      rcVec[ui] = Block<U>( rcPos * rcSub, rcSize * rcSub, rcBlock.getPlane( ui ) );
    }
  }

private:
  std::vector<Block<const t_El> > m_acBlocksConst;
  PicBlockStats m_cStats;

  virtual void xClearNonConst() {}
};


template<class t_El>
class PicBlock : public PicBlock<const t_El>
{
public:
  PicBlock() {}
  virtual ~PicBlock() {}

  PicBlock( const PrimPixelPos& rcPos, const PrimPixelPos& rcSize, const PicBlock<t_El> &rcRefPicBlock )
    : PicBlock<const t_El>( rcPos, rcSize, rcRefPicBlock )
    , m_acBlocks( rcRefPicBlock.getStats().getNumPlanes() )
  {
    xCreateSubPicBlock( m_acBlocks, rcPos, rcSize, rcRefPicBlock );
  }

  PicBlock( const PicBlock<t_El> &rcRefPicBlock, UInt uiNumPlanes )
    : PicBlock<const t_El>( rcRefPicBlock, uiNumPlanes )
    , m_acBlocks( rcRefPicBlock.m_acBlocks.begin(), rcRefPicBlock.m_acBlocks.begin() + uiNumPlanes )
  { 
  }

  PicBlock( const PicBlock<t_El> &rcRefPicBlock, UInt uiPrimPlaneId, UInt uiNumPlanes, const Subsampling& rcPrimPlaneSubsampling )
    : PicBlock<const t_El>( rcRefPicBlock, uiPrimPlaneId, uiNumPlanes, rcPrimPlaneSubsampling )
    , m_acBlocks( rcRefPicBlock.m_acBlocks.begin() + uiPrimPlaneId, rcRefPicBlock.m_acBlocks.begin() + uiPrimPlaneId + uiNumPlanes )
  { 
  }

  template<class t_Block>
  PicBlock( const std::vector<t_Block> &rcBlocks, const PicBlockStats *pcStats = NULL )
    : PicBlock<const t_El>( rcBlocks, pcStats )
    , m_acBlocks( rcBlocks.begin(), rcBlocks.end() )
  {
    CAOF( (TypeCompare<t_Block,Block<t_El> >::m_bEqual) || (TypeCompare<t_Block,VideoCodecPlane<t_El> >::m_bEqual) );
  }

  const Block<t_El>& getPlane( UInt ui ) const { return m_acBlocks[ui]; }


  void clip( const t_El &rcElMin, const t_El &rcElMax ) const
  {
    for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
    {
      getPlane( ui ).clip( rcElMin, rcElMax );
    }
  }

  ErrVal padding( const int iMarginX, const int iMarginY ) const
  {
    for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
    {
      RNOK( getPlane( ui ).padding( iMarginX, iMarginY ) );
    }
    return Err::m_nOK;
  }

  void drawBorder( bool bLeft, bool bTop, bool bRight, bool bBottom ) const
  {
    for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
    {
      getPlane( ui ).drawBorder( bLeft, bTop, bRight, bBottom );
    }
  }

  template<class U>
  ErrVal assign( const PicBlock<const U>& rcPicBlock ) const
  {
    AOF_DBG( PicBlock<const t_El>::getStats() == rcPicBlock.getStats() );
    for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
    {
      RNOK( getPlane( ui ).assign( rcPicBlock.getPlane( ui ) ) );
    }
    return Err::m_nOK;
  }

  template<class U>
  const PicBlock& operator+=( const PicBlock<const U>& rcPicBlock ) const
  {
    AOF_DBG( PicBlock<const t_El>::getStats() == rcPicBlock.getStats() );
    for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
    {
      getPlane( ui ) += rcPicBlock.getPlane( ui );
    }
    return *this;
  }


  template<class U>
  const PicBlock& operator-=( const PicBlock<const U>& rcPicBlock ) const
  {
    AOF_DBG( PicBlock<const t_El>::getStats() == rcPicBlock.getStats() );
    for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
    {
      getPlane( ui ) -= rcPicBlock.getPlane( ui );
    }
    return *this;
  }

  const PicBlock& operator*= ( const t_El &rcEl ) const;
  const PicBlock& operator/= ( const t_El &rcEl ) const;
  const PicBlock& operator+= ( const t_El &rcEl ) const;
  void            assign     ( const t_El &rcEl ) const;
  const PicBlock& operator>>=( int iShift ) const;
  const PicBlock& operator<<=( int iShift ) const;

  template<int t_ElAdd, int t_ElShift, class t_Int, const t_Int t_ElMin, const t_Int t_ElMax, class U>
  void assignShiftAndClip( PicBlock<const U>& rcPicBlock) const
  {
    for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
    {
      m_acBlocks[ui].assignShiftAndClip<t_ElAdd, t_ElShift, t_Int, t_ElMin, t_ElMax, U>( rcPicBlock.getPlane(ui) );
    }
  }

  template<int t_ElAdd, int t_ElShift, class t_Int, class U>
  void addShiftAndClip( const PicBlock<const U>& rcPicBlock, const t_Int &t_ElMin, const t_Int &t_ElMax) const
  {
    for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
    {
      m_acBlocks[ui].addShiftAndClip<t_ElAdd, t_ElShift, t_Int, U>( rcPicBlock.getPlane(ui), t_ElMin, t_ElMax );
    }
  }

  template<class U>
  void assignAndDivide( const PicBlock<const U>& rcPicBlock, const int iDiv ) const
  {
    for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
    {
      m_acBlocks[ui].assignAndDivide( rcPicBlock.getPlane(ui), iDiv );
    }
  }

private:
  std::vector<Block<t_El> > m_acBlocks;

  virtual void xClearNonConst() { m_acBlocks.clear(); }
};



template<class t_El>
class AllocPicBlock : public PicBlock<t_El>
{
public:
  AllocPicBlock( const PicBlockStats &rcStats )
    : PicBlock<t_El>( xAllocatePicBlockVector( rcStats ), &rcStats )
    , m_pcAllocatedBuffer( PicBlock<t_El>::getPlane( 0 ).getPtr() ) {}

  virtual ~AllocPicBlock()
  {
    delete[] m_pcAllocatedBuffer;
  }

  DISABLE_DEFAULT_CPYCTOR_AND_ASOP(AllocPicBlock);
private:
  t_El *m_pcAllocatedBuffer;

  static std::vector<Block<t_El> > xAllocatePicBlockVector( const PicBlockStats &rcStats )
  {
    std::vector<Block<t_El> > cVec( rcStats.getNumPlanes() );
    UInt uiBufferSize = 0;
    for( UInt ui = 0; ui < rcStats.getNumPlanes(); ui++ )
    {
      BlockSize cSize = rcStats.getPrimBlockSize() * rcStats.getSubsampling()[ui];
      uiBufferSize += cSize.getX() * cSize.getY();
    }
    t_El *p = new t_El[uiBufferSize];

    UInt uiOffset = 0;
    for( UInt ui = 0; ui < rcStats.getNumPlanes(); ui++ )
    {
      BlockSize cSize = rcStats.getPrimBlockSize() * rcStats.getSubsampling()[ui];
      cVec[ui] = Block<t_El>( p + uiOffset, cSize, cSize.getX() );
      uiOffset += cSize.getX() * cSize.getY();
    }
    return cVec;
  }
};

template<class t_El>
class AllocPicBlockWithMargin : public PicBlock<t_El>
{
public:
  AllocPicBlockWithMargin( const PicBlockStats& rcStats, const int iMargin )
    : PicBlock<t_El>( xAllocatePicBlockVector( rcStats, iMargin ), &rcStats )
    , m_pcAllocatedBuffer( PicBlock<t_El>::getPlane( 0 ).getPtr() ) 
  {
    BlockSize cMargin = PrimBlockSize(iMargin,iMargin) * rcStats.getSubsampling()[0];
    BlockSize cSize = rcStats.getPrimBlockSize() * rcStats.getSubsampling()[0];
    const int iBufferOffset = (2*cMargin.getX() +cSize.getX()) * cMargin.getY() + cMargin.getX();
    m_pcAllocatedBuffer -= iBufferOffset;
  }

  virtual ~AllocPicBlockWithMargin()
  {
    delete[] m_pcAllocatedBuffer;
  }

  DISABLE_DEFAULT_CPYCTOR_AND_ASOP(AllocPicBlockWithMargin);
private:
  t_El *m_pcAllocatedBuffer;

  static std::vector<Block<t_El> > xAllocatePicBlockVector( const PicBlockStats& rcStats, const UInt uiMargin )
  {
    TOT( 0 != (uiMargin & 1) );
    UInt uiBufferSize = 0;
    for( UInt ui = 0; ui < rcStats.getNumPlanes(); ui++ )
    {
      BlockSize cMargin = PrimBlockSize(uiMargin,uiMargin) * rcStats.getSubsampling()[ui];
      BlockSize cSize = rcStats.getPrimBlockSize() * rcStats.getSubsampling()[ui];
      uiBufferSize += (2*cMargin.getX() +cSize.getX()) * (2*cMargin.getY() +cSize.getY());
    }
    t_El *p = new t_El[uiBufferSize];

    std::vector<Block<t_El> > cVec( rcStats.getNumPlanes() );
    UInt uiOffset = 0;
    for( UInt ui = 0; ui < rcStats.getNumPlanes(); ui++ )
    {
      BlockSize cMargin = PrimBlockSize(uiMargin,uiMargin)* rcStats.getSubsampling()[ui];
      BlockSize cSize = rcStats.getPrimBlockSize() * rcStats.getSubsampling()[ui];
      const int iStride = 2*cMargin.getX() +cSize.getX();
      cVec[ui] = Block<t_El>( p + uiOffset + cMargin.getX() + cMargin.getY() * iStride, cSize, iStride );
      uiOffset += iStride * (2*cMargin.getY() +cSize.getY());
    }
    return cVec;
  }
};


template<class t_El>
class AllocPicBlockReuse : public PicBlock<t_El>
{
public:

  AllocPicBlockReuse( const PicBlockStats& rcStats, ReuseBuffer& rcReuseBuffer )
    : PicBlock<t_El>( xAllocatePicBlockVector( rcStats, rcReuseBuffer ), &rcStats )
    , m_pcReuseBuffer( &rcReuseBuffer ) 
  {
  }

  virtual ~AllocPicBlockReuse()
  {
    m_pcReuseBuffer->freeBuffer();
  }

  DISABLE_DEFAULT_CPYCTOR_AND_ASOP(AllocPicBlockReuse);
private:
  ReuseBuffer*m_pcReuseBuffer;

  static std::vector<Block<t_El> > xAllocatePicBlockVector( const PicBlockStats& rcStats, ReuseBuffer& rcReuseBuffer )
  {
    UInt uiBufferSize = 0;
    for( UInt ui = 0; ui < rcStats.getNumPlanes(); ui++ )
    {
      BlockSize cSize = rcStats.getPrimBlockSize() * rcStats.getSubsampling()[ui];
      uiBufferSize += cSize.getX() * cSize.getY();
    }
    t_El *p = (t_El*)rcReuseBuffer.getBuffer( sizeof(t_El)*uiBufferSize );

    std::vector<Block<t_El> > cVec( rcStats.getNumPlanes() );
    UInt uiOffset = 0;
    for( UInt ui = 0; ui < rcStats.getNumPlanes(); ui++ )
    {
      BlockSize cSize = rcStats.getPrimBlockSize() * rcStats.getSubsampling()[ui];
      cVec[ui] = Block<t_El>( p + uiOffset, cSize, cSize.getX() );
      uiOffset += cSize.getX() * cSize.getY();
    }
    return cVec;
  }

};


template<class t_El>
const PicBlock<t_El>& PicBlock<t_El>::operator*=( const t_El& rcEl ) const
{
  for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
  {
    getPlane( ui ) *= rcEl;
  }
  return *this;
}

template<class t_El>
const PicBlock<t_El>& PicBlock<t_El>::operator/=( const t_El& rcEl ) const
{
  for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
  {
    getPlane( ui ) /= rcEl;
  }
  return *this;
}

template<class t_El>
const PicBlock<t_El>& PicBlock<t_El>::operator+=( const t_El& rcEl ) const
{
  for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
  {
    getPlane( ui ) += rcEl;
  }
  return *this;
}


template<class t_El>
void PicBlock<t_El>::assign( const t_El& rcEl ) const
{
  for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
  {
    getPlane( ui ).assign( rcEl );
  }
}


template<class t_El>
const PicBlock<t_El>& PicBlock<t_El>::operator>>=( int iShift ) const
{
  for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
  {
    getPlane( ui ) >>= iShift;
  }
  return *this;
}

template<class t_El>
const PicBlock<t_El>& PicBlock<t_El>::operator<<=( int iShift ) const
{
  for( UInt ui = 0; ui < PicBlock<const t_El>::getStats().getNumPlanes(); ui++ )
  {
    getPlane( ui ) <<= iShift;
  }
  return *this;
}

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _PICBLOCK_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
