/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _PROBABILITYALLOCATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _PROBABILITYALLOCATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/InStreamIf.h"
#include "VideoCommonLib/ContextInit.h"
#include "VideoCommonLib/TraceCommon.h"
#include "VideoCommonLib/CabacContextModel.h"
#include "VideoCommonLib/CabacTables.h"

#if COLLECT_CABAC_BITS
#include "VideoCommonLib/ExtendedContextModels.h"
#endif



// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class t_ContextModel>
class ProbabilityAllocatorBase
{
protected:
  ProbabilityAllocatorBase( UInt uiNumCtx )
    : m_pacContext( new t_ContextModel[uiNumCtx] )
    , m_uiNumCtx( uiNumCtx ) {}

  virtual ~ProbabilityAllocatorBase() { delete[] m_pacContext; }
public:
  UInt getNumCtx() const { return m_uiNumCtx; }

  ErrVal setCtxStates( t_ContextModel *pacCtxStates )
  {
    ROF(( TypeCompare<t_ContextModel,CabacContextModel>::m_bEqual )); // memcpy will not work properly due to std::vector<> in t_ContextModel
    ::memcpy( m_pacContext, pacCtxStates, m_uiNumCtx * sizeof( t_ContextModel ) );
    return Err::m_nOK;
  }

  ErrVal getCtxStates( t_ContextModel *&pacCtxStates ) const
  {
    ROF(( TypeCompare<t_ContextModel,CabacContextModel>::m_bEqual )); // memcpy will not work properly due to std::vector<> in t_ContextModel
    pacCtxStates = new t_ContextModel[m_uiNumCtx];
    ::memcpy( pacCtxStates, m_pacContext, m_uiNumCtx * sizeof( t_ContextModel ) );
    return Err::m_nOK;
  }

protected:
  t_ContextModel *m_pacContext;
  const UInt m_uiNumCtx;

public:
  FORCEINLINE const t_ContextModel& getCtxModel( UInt uiCtxNumber ) const { return m_pacContext[uiCtxNumber]; }
};


template<class t_ContextModel>
class ProbabilityAllocator
  : virtual protected TraceBase
  , public ProbabilityAllocatorBase<t_ContextModel>
{
private:
  typedef ProbabilityAllocatorBase<t_ContextModel> BaseCall;

protected:
  ProbabilityAllocator( UInt uiNumCtx, TraceCommon *pcTrace )
    : ProbabilityAllocatorBase<t_ContextModel>( uiNumCtx )
#if COLLECT_CABAC_BITS
    , m_uiBinIndex( 0 )
#endif
  {
    TraceBase::setTraceCommonPtr( pcTrace );
  }

  virtual ~ProbabilityAllocator() {}

public:
  ErrVal initCtx( int iQp, const std::vector<CtxIni> &rcInit );
  void initEqualProbability();

  FORCEINLINE const t_ContextModel& getCtxDetailed( const CtxAccess &rcCtxInit, UInt uiCtxArrayIdx, UInt uiOffset ) const
  {
    return BaseCall::m_pacContext[rcCtxInit.getContextOffset( uiCtxArrayIdx, uiOffset )];
  }

#if COLLECT_CABAC_BITS
  double getSumOfBits() const
  {
    double dBits = 0.0;
    const bool bHasCounter = CheckCap<BitCounterContextModelCapIf, t_ContextModel>::m_bHasCap;
    if( bHasCounter )
    {
      for( UInt ui = 0; ui < BaseCall::getNumCtx(); ui++ )
      {
        dBits += TypeConvert<BitCounterContextModelCapIf, bHasCounter>::getIfTrue( BaseCall::m_pacContext[ui] ).getBitsCabac();
      }
    }
    return dBits;
  }

  double getBits( const CtxAccess &rcCtxInit, UInt uiCtxArrayIdx ) const
  {
    double dBits = 0.0;
    const bool bHasCounter = CheckCap<BitCounterContextModelCapIf, t_ContextModel>::m_bHasCap;
    if( bHasCounter )
    {
      for( int k = 0; k < rcCtxInit.getNumCtxElements( uiCtxArrayIdx ); k++ )
      {
        UInt uiIdx = rcCtxInit.getContextOffset( uiCtxArrayIdx );
        dBits += TypeConvert<BitCounterContextModelCapIf, bHasCounter>::getIfTrue( BaseCall::m_pacContext[uiIdx+k] ).getBitsCabac();
      }
    }
    return dBits;
  }

  UInt64 getBins( const CtxAccess &rcCtxInit, UInt uiCtxArrayIdx ) const
  {
    UInt64 uiBins = 0;
    const bool bHasCounter = CheckCap<BitCounterContextModelCapIf, t_ContextModel>::m_bHasCap;
    if( bHasCounter )
    {
      for( int k = 0; k < rcCtxInit.getNumCtxElements( uiCtxArrayIdx ); k++ )
      {
        UInt uiIdx = rcCtxInit.getContextOffset( uiCtxArrayIdx );
        uiBins += TypeConvert<BitCounterContextModelCapIf, bHasCounter>::getIfTrue( BaseCall::m_pacContext[uiIdx+k] ).getBinsCabac();
      }
    }
    return uiBins;
  }
#endif

protected:
#if COLLECT_CABAC_BITS
  UInt m_uiBinIndex;
#endif
};


template<class t_ContextModel>
ErrVal ProbabilityAllocator<t_ContextModel>::initCtx( int iQp, const std::vector<CtxIni> &racInit )
{
#if COLLECT_CABAC_BITS
  m_uiBinIndex = 0;
#endif
  ROF( UInt( racInit.size() ) == BaseCall::getNumCtx() );
  for( UInt ui = 0; ui < BaseCall::getNumCtx(); ui++ )
  {
    BaseCall::m_pacContext[ui].initCCModel( racInit[ui], iQp );
  }
  return Err::m_nOK;
}


template<class t_ContextModel>
void ProbabilityAllocator<t_ContextModel>::initEqualProbability()
{
#if COLLECT_CABAC_BITS
  m_uiBinIndex = 0;
#endif
  CtxIni cEpInit( 0, 64 );
  for( UInt ui = 0; ui < BaseCall::getNumCtx(); ui++ )
  {
    BaseCall::m_pacContext[ui].initCCModel( cEpInit, 10 ); // 10 has no influence here
  }
}


// video namespace end
VIDEO_NAMESPACE_END


#endif // #ifndef _PROBABILITYALLOCATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
