/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _QUADTREEGROUP_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _QUADTREEGROUP_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include <vector>
#include "Block.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

enum QTNodeState
{
  IS_FORBIDDEN       = 0x00,  //!< is not part of the quadtree group
  IS_INTERNAL_NODE   = 0x01,  //!< node is internal node (has child nodes)
  IS_LEAF_NODE       = 0x02,  //!< node is leaf node (does not have child nodes)
  IS_BELOW_LEAF_NODE = 0x04,  //!< node is below leaf node
};

class QuadtreeGroupStructure;
template<class T> class QuadtreeGroup;

class QTNavi
{
public:
  QTNavi() {}

  QTNavi( void *p ) : m_pcTree( NULL ) { AOT_DBG( p ); }

  QTNavi( const QuadtreeGroupStructure *pcTree, int iLayer )
  : m_pcTree( pcTree )
  , m_iLayer( iLayer )
  {
  }

  QTNavi( const QuadtreeGroupStructure *pcTree, int iX, int iY, int iLayer )
    : m_pcTree        ( pcTree )
    , m_iLayer        ( iLayer )
    , m_iX            ( iX     )
    , m_iY            ( iY     )
    , m_cPrimBlockSize( xCalculatePrimBlockSize() )
  {
  }

  typedef enum
  {
    TOP,
    BOTTOM,
    LEFT,
    RIGHT,
  } Edge;

  typedef enum
  {
    TOP_LEFT,
    TOP_RIGHT,
    BOTTOM_LEFT,
    BOTTOM_RIGHT,
  } Corner;

#define OPPOSITE_EDGE_OF(x) ( (x)==QTNavi::TOP ? QTNavi::BOTTOM : ((x)==QTNavi::LEFT ? QTNavi::RIGHT : ((x)==QTNavi::RIGHT ? QTNavi::LEFT : QTNavi::TOP) ) )

  const bool isMaxLayer() const;

  const PrimBlockSize& getPrimBlockSize() const { return m_cPrimBlockSize; }

  PrimPixelPos  getPrimPixelPos()  const;

  __inline UInt getLog2BlockSize() const;

  int  getLayer()    const { return m_iLayer; }
  int  getXCoord()   const { return m_iX; }
  int  getYCoord()   const { return m_iY; }

  const QuadtreeGroupStructure* getTreeStructurePtr() const { return m_pcTree; }

  bool liesOutside() const;

  bool isLeafNode() const;
  bool isInternalNode() const;
  bool isLeafOrInternalNode() const;

  inline QTNodeState getNodeState() const;

  QTNavi getNode     ( const PrimPixelPos &rcPos, const UInt uiLayer ) const;
  QTNavi findLeafNode( const PrimPixelPos &rcPos )                     const;

  __inline QTNavi getChild( UInt uiChildIdx ) const;
  __inline QTNavi getParent() const { return QTNavi( m_pcTree, m_iX>>1, m_iY>>1, m_iLayer - 1 ); }

  template<Edge t_eEdge, bool t_bClearVector> __inline void getEdgeNeighbours( std::vector<QTNavi> &racEdgeNeighbours, const PrimPixelPos& rcPrimPixelPos ) const;
  template<Corner t_eCorner> __inline QTNavi getCornerNeighbour( const PrimPixelPos& rcPrimPixelPos) const;

  __inline void getAllNeighbours( std::vector<QTNavi>& racAllNeighbours ) const;

  __inline QTNavi getLeft()     const { return QTNavi( m_pcTree, m_iX - 1, m_iY, m_iLayer ); }
  __inline QTNavi getRight()    const { return QTNavi( m_pcTree, m_iX + 1, m_iY, m_iLayer ); }
  __inline QTNavi getTop()      const { return QTNavi( m_pcTree, m_iX, m_iY - 1, m_iLayer ); }
  __inline QTNavi getBottom()   const { return QTNavi( m_pcTree, m_iX, m_iY + 1, m_iLayer ); }
  __inline QTNavi getTopLeft()  const { return QTNavi( m_pcTree, m_iX-1, m_iY - 1, m_iLayer ); }
  __inline QTNavi getTopRight() const { return QTNavi( m_pcTree, m_iX+1, m_iY - 1, m_iLayer ); }

  // with the following method, it is possible to determine if a block has been encoded before another
  //
  // this method should control an arbitrary coding order in future
  bool isCodedBefore( const QTNavi &rcNav ) const;
  bool isParsedBefore ( const QTNavi &rcNav ) const; // is entropy coded before

  bool operator==( const QTNavi &rcNav ) const;
  bool operator!=( const QTNavi &rcNav ) const { return !(*this == rcNav); }

private:
  PrimBlockSize xCalculatePrimBlockSize();

private:
  const QuadtreeGroupStructure *m_pcTree;
  int                           m_iLayer;
  int                           m_iX;
  int                           m_iY;
  PrimBlockSize                 m_cPrimBlockSize;
};


class QuadtreeGroupStructure
{
public:
  template<class T> friend class QuadtreeGroup;
  friend class QTNavi;

  QuadtreeGroupStructure( const QuadtreeGroupStructure &rcSource, UInt uiPower2MinBlock );
  QuadtreeGroupStructure( const PrimBlockSize& rcSize, const UInt& ruiLog2MinSize );
  QuadtreeGroupStructure( PrimBlockSize cSize, UInt uiPower2MaxBlock, UInt uiPower2MinBlock );
  virtual ~QuadtreeGroupStructure() {}

  PrimBlockSize getQuadtreeGroupSize() const { return PrimBlockSize( m_uiWidth, m_uiHeight ); }
  QTNavi getFirstQuadtreeRoot() const { return QTNavi( this, 0, 0, 0 ); }
  QTNavi getFirstQuadtreeRoot( UInt uiSlice ) const;
  QTNavi getNextQuadtreeRoot( const QTNavi &rcCurrent ) const;
  QTNavi getNextQuadtreeRoot( const QTNavi &rcCurrent, UInt uiSlice ) const;

  int getNumberOfQuadtreeRoots() const { return m_acLayers[0].getNumNodesX() * m_acLayers[0].getNumNodesY(); }
  int getNumberOfQuadtreeRootsX() const { return m_acLayers[0].getNumNodesX(); }
  int getNumberOfQuadtreeRootsY() const { return m_acLayers[0].getNumNodesY(); }

  QTNavi getNodeByPrimPixelPos( const PrimPixelPos& rcPos ) const { return getFirstQuadtreeRoot().findLeafNode( rcPos ); }

  UInt getNumEntropySlices() const;
  bool isSameEntropySlice( const QTNavi &rcQTNavi1, const QTNavi &rcQTNavi2 ) const;
  UInt getEntropySlice( const QTNavi &rcCurrent ) const;
  UInt getLog2MaxBlockSize() const { return m_acLayers[0].getDim(); }
  UInt getLog2BlockSize( UInt uiLayer ) const { return m_acLayers[uiLayer].getDim(); }

  ErrVal appendChildNodes( const QTNavi &rcNav, bool &rbChildsWereAppended );
  ErrVal removeChildNodes( const QTNavi &rcNav );
  void setEntropySliceSizeX( UInt uiSize ) { m_uiEntropySliceSizeX = uiSize; }
  void setEntropySliceSizeY( UInt uiSize ) { m_uiEntropySliceSizeY = uiSize; }

  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( QuadtreeGroupStructure );
private:

  class Layer
  {
  public:
    Layer() {}

    Layer( UInt uiNodesX, UInt uiNodesY, UInt uiDim, QTNodeState eInitNodeState )
      : m_uiNumNodesX( uiNodesX )
      , m_uiNumNodesY( uiNodesY )
      , m_uiDim( uiDim )
      , m_acNodeState( uiNodesX * uiNodesY, eInitNodeState ) {}

    UInt        getStride                 ()                 const { return getNumNodesX(); }
    QTNodeState getNodeStateNoCheck       ( int iX, int iY ) const;
    QTNodeState getNodeState              ( int iX, int iY ) const;
    void        setNodeState              ( QTNodeState eNodeState, int iX, int iY );

    UInt getNumNodesX() const { return m_uiNumNodesX; }
    UInt getNumNodesY() const { return m_uiNumNodesY; }
    UInt getDim() const { return m_uiDim; }

  private:
    UInt m_uiNumNodesX;
    UInt m_uiNumNodesY;
    UInt m_uiDim;
    std::vector<QTNodeState> m_acNodeState;
  };

  const UInt m_uiWidth;
  const UInt m_uiHeight;
  std::vector<Layer> m_acLayers;
  UInt m_uiEntropySliceSizeX;
  UInt m_uiEntropySliceSizeY;
};

class QuadtreeGroupPointerDeleter;

template<class T>
class QuadtreeGroup
{
public:
  void setAllElementsTo( const T &rcT );

  QuadtreeGroupStructure&       getTreeStructure()       { TOF( m_pcTreeStruct ); return *m_pcTreeStruct; }
  const QuadtreeGroupStructure& getTreeStructure() const { return m_rcTreeStructConst; }

  T&       getT( const QTNavi &rcNav )       { return m_apcT[rcNav.getLayer()][xGetTIdx( rcNav )]; }
  const T& getT( const QTNavi &rcNav ) const { return m_apcT[rcNav.getLayer()][xGetTIdx( rcNav )]; }

  QuadtreeGroup( QuadtreeGroupStructure *pcStructure, const QuadtreeGroupStructure &rcStructureConst );
  QuadtreeGroup( QuadtreeGroupStructure *pcStructure, const QuadtreeGroupStructure &rcStructureConst, const T &rcTInit );
  virtual ~QuadtreeGroup();

private:
  std::vector<T*> m_apcT;

  ErrVal xInit();

  int xGetNumElementsOfLayer( int iLayer ) const { return getTreeStructure().m_acLayers[iLayer].getNumNodesX() * getTreeStructure().m_acLayers[iLayer].getNumNodesY(); }
  int xGetNumLayers() const { return int( m_rcTreeStructConst.m_acLayers.size() ); }

  const QuadtreeGroupStructure &m_rcTreeStructConst;
  QuadtreeGroupStructure *m_pcTreeStruct;

  QuadtreeGroup();
  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( QuadtreeGroup );

  int xGetTIdx( const QTNavi &rcNav ) const
  {
    AOF_DBG( m_rcTreeStructConst.m_acLayers[rcNav.getLayer()].getNodeState( rcNav.getXCoord(), rcNav.getYCoord() ) & ( IS_LEAF_NODE | IS_INTERNAL_NODE ) );
    return rcNav.getYCoord() * m_rcTreeStructConst.m_acLayers[rcNav.getLayer()].getStride() + rcNav.getXCoord();
  }
};


class QuadtreeGroupPointerDeleter
{
public:
  template<class T>
  static void deleteAllPointers( QuadtreeGroup<T*> &rcThis )
  {
    for( QTNavi cNav = rcThis.getTreeStructure().getFirstQuadtreeRoot(); !cNav.liesOutside(); cNav = rcThis.getTreeStructure().getNextQuadtreeRoot( cNav ) )
    {
      if( cNav.isLeafNode() )
      {
        T*& rpT = rcThis.getT( cNav );
        delete rpT;
        rpT = NULL;
      }
      else
      {
        TNOK( deleteChildNodePointers( rcThis, cNav ) );
      }
    }
  }

  template<class T>
  static ErrVal deleteChildNodePointers( QuadtreeGroup<T*> &rcThis, const QTNavi &rcNav )
  {
    ROF( rcNav.isInternalNode() );
    for( UInt ui = 0; ui < 4; ui++ )
    {
      const QTNavi &rcChild = rcNav.getChild( ui );
      const QTNodeState eNodeState = rcChild.getNodeState();

      if( IS_INTERNAL_NODE == eNodeState )
      {
        RNOK( deleteChildNodePointers( rcThis, rcChild ) );
      }
      else if( IS_LEAF_NODE == eNodeState )
      {
        T*& rpT = rcThis.getT( rcChild );
        delete rpT;
        rpT = NULL;
      }
    }
    return Err::m_nOK;
  }
};


template<class T>
QuadtreeGroup<T>::QuadtreeGroup( QuadtreeGroupStructure *pcStructure, const QuadtreeGroupStructure &rcStructureConst )
: m_rcTreeStructConst( rcStructureConst )
, m_pcTreeStruct( pcStructure )
{
  TOF( NULL == pcStructure || &rcStructureConst == pcStructure );
  TNOK( xInit() );
}

template<class T>
QuadtreeGroup<T>::QuadtreeGroup( QuadtreeGroupStructure *pcStructure, const QuadtreeGroupStructure &rcStructureConst, const T &rcTInit )
: m_rcTreeStructConst( rcStructureConst )
, m_pcTreeStruct( pcStructure )
{
  TOF( NULL == pcStructure || &rcStructureConst == pcStructure );
  TNOK( xInit() );
  setAllElementsTo( rcTInit );
}


template<class T>
void QuadtreeGroup<T>::setAllElementsTo( const T &rcT )
{
  for( int i = 0; i < xGetNumLayers(); i++ )
  {
    const int iNumNodes = xGetNumElementsOfLayer( i );

    T *pT = m_apcT[i];
    for( int k = 0; k < iNumNodes; k++ )
    {
      pT[k] = rcT;
    }
  }
}


template<class T>
ErrVal QuadtreeGroup<T>::xInit()
{
  for( int i = 0; i < xGetNumLayers(); i++ )
  {
    const int iNumNodes = xGetNumElementsOfLayer( i );
    T *pT = new T[ iNumNodes ];
    ROF( pT );
    m_apcT.push_back( pT );
  }
  return Err::m_nOK;
}


template<class T>
QuadtreeGroup<T>::~QuadtreeGroup()
{
  while( !m_apcT.empty() )
  {
    T *pT = m_apcT.back();
    delete[] pT;
    m_apcT.pop_back();
  }
}

__inline QTNodeState QuadtreeGroupStructure::Layer::getNodeStateNoCheck       ( int iX, int iY ) const
{
  return m_acNodeState[ iY * getNumNodesX() + iX ];
}

__inline QTNodeState QuadtreeGroupStructure::Layer::getNodeState( int iX, int iY ) const
{
  if( /*iX < 0 || iY < 0 ||*/ (UInt)iX >= getNumNodesX() || (UInt)iY >= getNumNodesY() )
  {
    return IS_FORBIDDEN;
  }
  return m_acNodeState[ iY * getNumNodesX() + iX ];
}


__inline void QuadtreeGroupStructure::Layer::setNodeState( QTNodeState eNodeState, int iX, int iY )
{
  AOT_DBG( eNodeState & IS_FORBIDDEN );
  AOT_DBG( getNodeState( iX, iY ) & IS_FORBIDDEN );
  m_acNodeState[ iY * getNumNodesX() + iX ] = eNodeState;
}

inline const bool QTNavi::isMaxLayer() const { return (m_iLayer+1) == (int)m_pcTree->m_acLayers.size();}

inline PrimBlockSize QTNavi::xCalculatePrimBlockSize()
{
  const QuadtreeGroupStructure::Layer &rcLayer = m_pcTree->m_acLayers[m_iLayer];
  UInt uiWidth  = 1 << rcLayer.getDim();
  if( m_iX + 1 == (int)rcLayer.getNumNodesX() )
  {
    UInt uiBorderBlockWidth = m_pcTree->m_uiWidth & (uiWidth-1);
    if( uiBorderBlockWidth != 0 )
    {
      uiWidth = uiBorderBlockWidth;
    }
  }

  UInt uiHeight = 1 << rcLayer.getDim();
  if( m_iY + 1 == (int)rcLayer.getNumNodesY() )
  {
    UInt uiBorderBlockHeight = m_pcTree->m_uiHeight &(uiHeight-1);
    if( uiBorderBlockHeight != 0 )
    {
      uiHeight = uiBorderBlockHeight;
    }
  }
  return PrimBlockSize( uiWidth, uiHeight );
}


__inline PrimPixelPos QTNavi::getPrimPixelPos() const
{
  AOT_DBG( liesOutside() );
  const UInt uiDim = m_pcTree->m_acLayers[m_iLayer].getDim();
  AOF_DBG( (m_iX << uiDim) < (int)m_pcTree->m_uiWidth && (m_iY << uiDim) < (int)m_pcTree->m_uiHeight );
  return PrimPixelPos( m_iX << uiDim, m_iY << uiDim );
}


__inline UInt QTNavi::getLog2BlockSize() const
{
  return m_pcTree->m_acLayers[m_iLayer].getDim();
}


__inline QTNavi QTNavi::getNode( const PrimPixelPos &rcPos, const UInt uiLayer ) const
{
  AOT_DBG( uiLayer >= m_pcTree->m_acLayers.size() );
  const UInt uiDim = m_pcTree->m_acLayers[uiLayer].getDim();
  return QTNavi( m_pcTree, rcPos.getX() >> uiDim, rcPos.getY() >> uiDim, uiLayer );
}

__inline QTNavi QTNavi::getChild( UInt uiChildIdx ) const
{
  AOT_DBG( uiChildIdx > 3 );
  const int iX = (m_iX<<1) + (uiChildIdx &1);
  const int iY = (m_iY<<1) + (uiChildIdx>>1);
  return QTNavi( m_pcTree, iX, iY, m_iLayer + 1 );
}

template<QTNavi::Edge t_eEdge, bool t_bClearVector>
__inline void QTNavi::getEdgeNeighbours( std::vector<QTNavi> &racEdgeNeighbours, const PrimPixelPos& rcPrimPixelPos ) const
{
  PrimPixelPos cOffset( t_eEdge == LEFT ? -1 : (t_eEdge == RIGHT  ? getPrimBlockSize().getX() : 0 ),
                        t_eEdge == TOP  ? -1 : (t_eEdge == BOTTOM ? getPrimBlockSize().getY() : 0 ) );
  const bool bStepHor = t_eEdge == TOP || t_eEdge == BOTTOM;

  if( t_bClearVector )
  {
    racEdgeNeighbours.clear();
  }

  const QTNavi &rcFirst = findLeafNode( rcPrimPixelPos + cOffset );
  if( rcFirst.liesOutside() )
  {
    return;
  }
  racEdgeNeighbours.push_back( rcFirst );

  if( bStepHor )
  {
    cOffset.getX() += racEdgeNeighbours.back().getPrimBlockSize().getX();
    while( cOffset.getX() < getPrimBlockSize().getX() )
    {
      racEdgeNeighbours.push_back( findLeafNode( rcPrimPixelPos + cOffset ) );
      cOffset.getX() += racEdgeNeighbours.back().getPrimBlockSize().getX();
    }
  }
  else
  {
    cOffset.getY() += racEdgeNeighbours.back().getPrimBlockSize().getY();
    while( cOffset.getY() < getPrimBlockSize().getY() )
    {
      racEdgeNeighbours.push_back( findLeafNode( rcPrimPixelPos + cOffset ) );
      cOffset.getY() += racEdgeNeighbours.back().getPrimBlockSize().getY();
    }
  }
}


template<QTNavi::Corner t_eCorner>
__inline QTNavi QTNavi::getCornerNeighbour( const PrimPixelPos& rcPrimPixelPos) const
{
  const PrimPixelPos cOffset( t_eCorner == TOP_LEFT || t_eCorner == BOTTOM_LEFT ? -1 : getPrimBlockSize().getX(),
                              t_eCorner == TOP_LEFT || t_eCorner == TOP_RIGHT   ? -1 : getPrimBlockSize().getY() );
  return findLeafNode( rcPrimPixelPos + cOffset );
}

__inline void QTNavi::getAllNeighbours( std::vector<QTNavi>& racAllNeighbours ) const
{
  const PrimPixelPos cPrimPixelPos = getPrimPixelPos();
  getEdgeNeighbours<TOP,   false>( racAllNeighbours, cPrimPixelPos );
  getEdgeNeighbours<BOTTOM,false>( racAllNeighbours, cPrimPixelPos );
  getEdgeNeighbours<LEFT,  false>( racAllNeighbours, cPrimPixelPos );
  getEdgeNeighbours<RIGHT ,false>( racAllNeighbours, cPrimPixelPos );

  racAllNeighbours.push_back( getCornerNeighbour<TOP_LEFT>(cPrimPixelPos) );
  if( racAllNeighbours.back().liesOutside() )
  {
    racAllNeighbours.pop_back();
  }

  racAllNeighbours.push_back( getCornerNeighbour<TOP_RIGHT>(cPrimPixelPos) );
  if( racAllNeighbours.back().liesOutside() )
  {
    racAllNeighbours.pop_back();
  }

  racAllNeighbours.push_back( getCornerNeighbour<BOTTOM_LEFT>(cPrimPixelPos));
  if( racAllNeighbours.back().liesOutside() )
  {
    racAllNeighbours.pop_back();
  }

  racAllNeighbours.push_back( getCornerNeighbour<BOTTOM_RIGHT>(cPrimPixelPos));
  if( racAllNeighbours.back().liesOutside() )
  {
    racAllNeighbours.pop_back();
  }
}


__inline bool QTNavi::liesOutside() const
{
  return getNodeState() == IS_FORBIDDEN;
}

inline QTNodeState QTNavi::getNodeState() const
{
  if( NULL == m_pcTree )// || m_iLayer < 0 || int( m_pcTree->m_acLayers.size() ) <= m_iLayer )
  {
    return IS_FORBIDDEN;
  }
  AOT_DBG( m_iLayer < 0 || int( m_pcTree->m_acLayers.size() ) <= m_iLayer );
  return m_pcTree->m_acLayers[m_iLayer].getNodeState( m_iX, m_iY );
}

__inline bool QTNavi::isLeafNode() const
{
  return IS_LEAF_NODE == getNodeState();
}

__inline bool QTNavi::isInternalNode() const
{
  return IS_INTERNAL_NODE == getNodeState();
}

__inline bool QTNavi::isLeafOrInternalNode() const
{
  const QTNodeState eNodeState = getNodeState();
  return IS_LEAF_NODE == eNodeState || IS_INTERNAL_NODE == eNodeState;
}


// This method is to determine the coding order:
// If a Nav is smaller than another, this Nav is coded before the other one.
__inline bool QTNavi::isCodedBefore( const QTNavi &rcNav ) const
{
  AOT_DBG( liesOutside() );
  AOF_DBG( m_pcTree == rcNav.m_pcTree );

  if( *this == rcNav )
  {
    return false;
  }

  const PrimPixelPos cLeftPos  = getPrimPixelPos();
  const PrimPixelPos cRightPos = rcNav.getPrimPixelPos();

  const UInt uiDiff  = (cLeftPos.getX() ^ cRightPos.getX()) | (cLeftPos.getY() ^ cRightPos.getY());
  for( UInt ui = 0; true; ui++ )
  {
    const UInt uiMask = ((-1) << m_pcTree->m_acLayers[ui].getDim());
    if( 0 != (uiDiff & uiMask) )
    {
      const QTNavi cLeft  = getNode( cLeftPos,  ui );
      const QTNavi cRight = getNode( cRightPos, ui );
      return cLeft.m_iY < cRight.m_iY || ( cLeft.m_iY == cRight.m_iY && cLeft.m_iX < cRight.m_iX );
    }
  }
  return false;
}

__inline bool QTNavi::isParsedBefore( const QTNavi &rcNav ) const
  {
  bool bExp = isCodedBefore( rcNav );
  if( false == bExp )
  {
    return bExp;
  }

  const QTNavi cRootLeft ( m_pcTree, m_iX>>m_iLayer, m_iY>>m_iLayer, 0  );
  const QTNavi cRootRight( rcNav.m_pcTree, rcNav.m_iX>>rcNav.m_iLayer, rcNav.m_iY>>rcNav.m_iLayer, 0 );
  return m_pcTree->isSameEntropySlice( cRootLeft, cRootRight );
}

__inline bool QTNavi::operator==( const QTNavi &rcNav ) const
{
  return m_pcTree != NULL           &&
         m_pcTree == rcNav.m_pcTree &&
         m_iX     == rcNav.m_iX     &&
         m_iY     == rcNav.m_iY     &&
         m_iLayer == rcNav.m_iLayer;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _QUADTREEGROUP_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
