/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _RESTRAFOPOW2_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _RESTRAFOPOW2_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/ResidualCalculator.h"
#include "VideoCommonLib/ResTrafoCommon.h"
#include "VideoCommonLib/ContextInit.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class ResidualContextStartIndices;

class ResInfoTrafoPow2Leaf
{
public:
  ResInfoTrafoPow2Leaf( const PicBlockStats &rcStats )
    : m_cCoeffBlock( rcStats )
    , m_abCbp( new bool[ rcStats.getNumPlanes() ] )
  {
    UInt uiNumPlanes = rcStats.getNumPlanes();
    for ( UInt ui = 0; ui < uiNumPlanes; ui++ )
    {
      ShiftScale cRatioX( rcStats.getSubsampling()[ ui ].getX() );
      AOF( cRatioX == rcStats.getSubsampling()[ ui ].getY() );
    }
  }

  ~ResInfoTrafoPow2Leaf()
  {
    delete[] m_abCbp;
  }

        bool& getCbf( UInt uiPlane )       { return m_abCbp[ uiPlane ]; }
  const bool& getCbf( UInt uiPlane ) const { return m_abCbp[ uiPlane ]; }

  const PicBlock<int>&       getCoeff()       { return m_cCoeffBlock; }
  const PicBlock<const int>& getCoeff() const { return m_cCoeffBlock; }

private:
  AllocPicBlock<int>        m_cCoeffBlock;
  bool                     *m_abCbp;
};

class ResInfoTrafoPow2
  : public ResInfo, public ResTrafoCommon
{
public:
  
  const ResInfoTrafoPow2*    getResInfoTrafoPow2()        const { return this; }
  ResInfoTrafoPow2*          getResInfoTrafoPow2()              { return this; }
  
  ResInfoTrafoPow2( UInt uiNumPlanes, QuadtreeGroupStructure &rcQuadTreeGroupStructure )
    : ResTrafoCommon( uiNumPlanes )
    , m_cCoeffTree( &rcQuadTreeGroupStructure, rcQuadTreeGroupStructure, NULL )
  { /*ROF( m_cCoeffTree.QuadtreeGroupStructure.getNumberOfQuadtreeRoots() == 1 );*/ }

  virtual ~ResInfoTrafoPow2()
  {
    QuadtreeGroupPointerDeleter::deleteAllPointers( m_cCoeffTree );
    delete &(m_cCoeffTree.getTreeStructure()); // QuadtreeGroup doesn't delete the structure
  }

  ResidualMode getResidualMode() const { return ResInfo::RES_TRAFO_POW2; };

  const QuadtreeGroup<ResInfoTrafoPow2Leaf*>& getTree() const { return m_cCoeffTree; }
        QuadtreeGroup<ResInfoTrafoPow2Leaf*>& getTree()       { return m_cCoeffTree; }

  ErrVal getCbfsByPrimPixelPos( UInt uiNumPlanes, PrimPixelPos cPosInBlock, std::vector<bool> &rabCbfs ) const
  {
    rabCbfs = std::vector<bool>( uiNumPlanes, false );
    QTNavi cNav = m_cCoeffTree.getTreeStructure().getNodeByPrimPixelPos( cPosInBlock );
    ROF( cNav.isLeafNode() );
    const ResInfoTrafoPow2Leaf* p = m_cCoeffTree.getT( cNav );
    for( UInt ui = 0; ui < uiNumPlanes; ui++ )
    {
      rabCbfs[ui] = p->getCbf( ui );
    }
    return Err::m_nOK;
  }

private:
  QuadtreeGroup<ResInfoTrafoPow2Leaf*> m_cCoeffTree;
};

// class that holds the configuration of one reconstructResidual() calculation for fast access.
// ResTrafoPow2 can't hold that state itself because reconstructResidual() is const.
class ResTrafoPow2Configuration
{
public:
  ResTrafoPow2Configuration( const ResInfoTrafoPow2 *pcResInfo,
                             const PicBlock<DPel> &rcResBlock,
                             const bool bAddDecRes )
    : m_raiPlaneQps( pcResInfo->getQps() )
    , m_rcCoeffTree( pcResInfo->getTree() )
    , m_rcResBlock( rcResBlock )
    , m_bAddDecRes( bAddDecRes ) {}

  const std::vector<int> &m_raiPlaneQps;
  const QuadtreeGroup<ResInfoTrafoPow2Leaf*> &m_rcCoeffTree;
  const PicBlock<DPel> &m_rcResBlock;
  const bool m_bAddDecRes;
};

class ResTrafoPow2
  : public virtual ResidualCalculator
{
public:
  explicit ResTrafoPow2( UInt uiLog2MaxBlockSize )
  {
    AOF( uiLog2MaxBlockSize <= sm_uiLog2MaxTSize );
    AOT( xInitCoeffArray() );
  }
  virtual ~ResTrafoPow2() {}
  ErrVal reconstructResidual( const ResInfo &rcResInfo,
                              const PicBlock<DPel> &rcResBlock,
                              const bool bAddDecRes );

  static ErrVal appendTrafoPow2ContextsAndSetIndices( ResidualContextStartIndices &rcResCtxIndices,
                                                      UInt uiLog2MinTrafoSize,
                                                      UInt uiLog2MaxTrafoSize,
                                                      CtxInitArray &rcCtxInit );

  static ErrVal setScanPattern ( std::vector<UInt>& rauiScanPattern, const UInt uiLog2BlockSize, const bool bDownLeft );
  static UInt   getSigCtxInc   ( const Block<const int>& rcBlock, const UInt uiPosX, const UInt uiPosY, const UInt uiLog2BlkSize, const bool bDownLeft );
  static UInt   getLastCtxInc  ( const Block<const int>& rcBlock, const UInt uiPosX, const UInt uiPosY, const UInt uiLog2BlkSize );

protected:
  template< UInt uiBlockSize, UInt uiLog2TSize, bool bHor >
  void xFast1dIDCT( int* paiData, const UInt uiInc, UInt uiFinalShift );
  template<UInt uiBlockSize, UInt uiLog2TSize > 
  void xFast2dIDCT( const Block<int>& rcIn, const Block<DPel>& rcOut, bool bAddDecRes );

  ErrVal  xFast2dIDCT ( const Block<int>& rcIn, const Block<DPel>& rcOut, bool bAddDecRes );
  ErrVal  xFastIScale ( int iQp, const Block<const int> &rcIn, const Block<int> &rcOut );

  static const UInt           sm_uiLog2MaxTSize;
  static const UInt           sm_uiCoeffAccessMask;
  static const UInt           sm_uiDCTPrecision;
  static const UInt           sm_uiIDCTPrecision;
  static const UInt           sm_uiQBits;
  static const int            sm_aiQStep        [6];
  static const int            sm_aiQFactor      [6];
  static const int            sm_aiDCTCoeffs    [128];
  static const int            sm_aiIDCTCoeffs   [128];
  int                         m_aiDCTCoeffArray [512];
  int                         m_aiIDCTCoeffArray[512];
  int                         m_aiAuxArray      [128];
  int                         m_aiAux2Array      [128];
  int                         m_aucTempIntBuffer[128*128];

private:
  ErrVal xInitCoeffArray();
  ErrVal xReconstructNode( const QTNavi &rcNav, const ResTrafoPow2Configuration &rcConfig );

  static ErrVal xAppendSigCtx   ( CtxInitArray &rcCtxInit, UInt uiLog2MinTrafoSize, UInt uiLog2MaxTrafoSize );
  static ErrVal xAppendLastCtx  ( CtxInitArray &rcCtxInit, UInt uiLog2MinTrafoSize, UInt uiLog2MaxTrafoSize );
};

VIDEO_NAMESPACE_END

#endif // #ifndef _RESTRAFOPOW2_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
