/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _SPLINES_H__
#define _SPLINES_H__

#include "InterpolationIf.h"
#include "InterCommon.h"
#include <fstream>

// video namespace begin
VIDEO_NAMESPACE_BEGIN

typedef short CoeffT;

template<class t_TSrc, class t_TDst>
class CubicSplineFixed
  : public InterpolationIf<t_TSrc, t_TDst>
{
public:
  CubicSplineFixed() : m_iSplPadding( 2 ), m_iQ25SplGain( 1207959552 ), m_i16Q2KernelGain( short( 6*(1<<2) ) ), m_i16Q15Z1(short(8780*-1)), m_i16Q15Z1_c(9459) {}
  CubicSplineFixed( const int iQ25SplGain, const short i16Q2KernelGain, const short i16Q15Z1, const short i16Q15Z1_c ) : m_iSplPadding( 2 ), m_iQ25SplGain( iQ25SplGain ), m_i16Q2KernelGain( i16Q2KernelGain ), m_i16Q15Z1( i16Q15Z1 ), m_i16Q15Z1_c( i16Q15Z1_c ) {}
  virtual ~CubicSplineFixed()
  { 
    std::vector<Block<short>*>::iterator it;
    for (it=m_pi16Q6SplCoeffFP.begin(); it!=m_pi16Q6SplCoeffFP.end(); ++it)
    {
      delete (*it);
    }
    m_pi16Q6SplCoeffFP.clear();
  }

  // Allocate memory for spline_coeffs + 2*padding
  ErrVal allocCoeffs(const PrimBlockSize& rcPbSize, const std::vector<Subsampling> &racOrigSub )
  {
    for( UInt ui = 0; ui < racOrigSub.size(); ui++ )
    {
      const BlockSize cSize = rcPbSize * racOrigSub[ui];
      m_pi16Q6SplCoeffFP.push_back( new AllocBlock<short>( BlockSize(cSize.getX()+2*m_iSplPadding, cSize.getY()+2*m_iSplPadding) ) );
      ROF( m_pi16Q6SplCoeffFP[ui] );
    }
    return Err::m_nOK;
  }

  // function to calculate spline coeffs
  void calcCoeffs(const Block<const t_TSrc> &rcSrc, UInt uiPl);

  // function to interpolate using spline coeffs
  void interpolate00(UInt uiPl, const Block<t_TDst> &rcDst, int iFac) const;
  void interpolate(UInt uiPl, const Block<t_TDst> &rcDst, int iFac);
  void interpolateBlock(Block<t_TDst>& rcDstBlock, 
                        const UInt uiPlane, 
                        const PixelPos &rcPosInUps,
                        BinaryFraction<15> &rcOffX,
                        BinaryFraction<15> &rcOffY) const 
  {  
    xInterpolateBlock( rcDstBlock, uiPlane, rcPosInUps, rcOffX, rcOffY);
  }


private:
public:
  enum PaddingMode 
  {
    INVALID_PADDING_MODE = -1,
    NO_PADDING = 0 ,
    PAD_RIGHT_BORDER = 1,
    PAD_RIGHT_AND_TOP_BORDER = 5,
    PAD_RIGHT_AND_BOTTOM_BORDER = 9,
    PAD_LEFT_BORDER = 2,
    PAD_LEFT_AND_TOP_BORDER = 6,
    PAD_LEFT_AND_BOTTOM_BORDER = 10,
    PAD_TOP_BORDER = 4,
    PAD_BOTTOM_BORDER = 8,
  };

  void xCopyFromSrcBuffer( Block<short>*& rpcDestBlock, const Block<short> &rcSrc, const PixelPos& rcPixelPos, PaddingMode& rePaddingMode, BlockSize& rcPaddingSize, BlockSize& rcValidSize, int iHorAdd, int iVerAdd ) const ;
  void xFilterHorizontal( Block<short> &rcDst, const Block<const short> &rcSrc, int iFilterPos) const;
  void xFilterVertical( Block<short> &rcDst, const Block<const short> &rcSrc, int iFilterPos) const;
  void xApplyPadding( Block<short> &rcDst, PaddingMode ePaddingMode, const BlockSize& rcPaddingSize ) const;
  void xInterpolateBlock(Block<t_TDst>& rcDstBlock,
                          const UInt uiPlane, 
                          const PixelPos &rcPosInUps,
                          BinaryFraction<15> &rcOffX,
                          BinaryFraction<15> &rcOffY) const ;

  void xCopyTo(Block<t_TDst>& rcDst, 
                const Block<t_TDst>& rcSrc, 
                const int iXPos,
                const int iYPos,
                const int iXAdd,
                const int iYAdd) const;

  void xFilterInternal( Block<short>& rcDest, const short* picSrc, const int iSrcStride, const int iHorFrac, const int iVerFrac, const PaddingMode ePaddingMode, const BlockSize& cInsideSize, const BlockSize& cValidSize, UInt uiVerySpecialFilterPosHandling) const; 

#if ! USE_SEPARABLE_SPLINE_UPSAMPLING
  template<int iFac>
  void xInterpolate(UInt uiPl, const Block<t_TDst> &rcDst);
#endif
  short xInitRowCausalCoeff(const Block<const t_TSrc> &rcSrc, int iY, int N);
  short xInitColCausalCoeff(const Block<const short> &rcSrc, int iX, int N);

  __inline virtual int   xGetSplineVal       ( int iX, int iY )         const { return int(sm_aai16Q15SplVal8PelLUT[iY][iX]); }
  __inline         short xInt16Multiplication( const short& i16A, const short& i16B ) const { return short( ( ( int(i16A)*int(i16B) )+(1<<14) )>>15); }
  virtual const short* xGetSplineBuf()      const { return sm_aai16Q15SplVal8PelLUT[0]; }
  virtual const int xGetSplineBufStride() const { return 32; }
  __inline virtual const CoeffT* xGetFilterCoeffs( const int iFilter ) const { return sm_aai16Q15SplVal4PelLUT[iFilter]; }

  // table look-up for quarter-pel kernel values
  __inline virtual short xGetKernel( int iDx, int iXX ) const { return sm_aai16Q15SplVal4PelLUT[iDx][iXX]; }

  const int                   m_iSplPadding;
  const int                   m_iQ25SplGain;
  const int                 m_i16Q2KernelGain;
  std::vector<Block<short>*>  m_pi16Q6SplCoeffFP;
  static const short          sm_aai16Q15SplVal8PelLUT[32][32];
  static const CoeffT          sm_aai16Q15SplVal4PelLUT[4][4];
  const short                 m_i16Q15Z1;
  const short                 m_i16Q15Z1_c;
};

// Q15 LUT 4x4 spline value for 16 different (dx,dxx) 
template<class t_TSrc, class t_TDst>
const CoeffT CubicSplineFixed<t_TSrc, t_TDst>::sm_aai16Q15SplVal4PelLUT[4][4]=
{
  { 5461,   21845,  5461,   0   , },
  { 2304,   20053,  10325,  85  , },
  { 683,    15701,  15701,  683 , },
  { 85,     10325,  20053,  2304, }, 
};

// Q15 LUT 4x4 spline value for every 32 1/8 Pel (dx,dy) pairs
template<class t_TSrc, class t_TDst>
const short CubicSplineFixed<t_TSrc, t_TDst>::sm_aai16Q15SplVal8PelLUT[32][32]=
{
  {      910,    3641,     910,       0,     610,    3561,    1289,       2,     384,    3342,    1721,      14,     222,    3017,    2174,      48,     114,    2617,    2617,     114,      48,    2174,    3017,     222,      14,    1721,    3342,     384,       2,    1289,    3561,     610, },
  {     3641,   14564,    3641,       0,    2439,   14244,    5156,       7,    1536,   13369,    6884,      57,     889,   12068,    8697,     192,     455,   10468,   10468,     455,     192,    8697,   12068,     889,      57,    6884,   13369,    1536,       7,    5156,   14244,    2439, },
  {      910,    3641,     910,       0,     610,    3561,    1289,       2,     384,    3342,    1721,      14,     222,    3017,    2174,      48,     114,    2617,    2617,     114,      48,    2174,    3017,     222,      14,    1721,    3342,     384,       2,    1289,    3561,     610, },
  {        0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0, },
  {      610,    2439,     610,       0,     409,    2386,     863,       1,     257,    2239,    1153,      10,     149,    2021,    1457,      32,      76,    1753,    1753,      76,      32,    1457,    2021,     149,      10,    1153,    2239,     257,       1,     863,    2386,     409, },
  {     3561,   14244,    3561,       0,    2386,   13931,    5042,       7,    1502,   13075,    6732,      56,     869,   11802,    8506,     188,     445,   10238,   10238,     445,     188,    8506,   11802,     869,      56,    6732,   13075,    1502,       7,    5042,   13931,    2386, },
  {     1289,    5156,    1289,       0,     863,    5042,    1825,       3,     544,    4733,    2437,      20,     315,    4272,    3079,      68,     161,    3706,    3706,     161,      68,    3079,    4272,     315,      20,    2437,    4733,     544,       3,    1825,    5042,     863, },
  {        2,       7,       2,       0,       1,       7,       3,       0,       1,       7,       3,       0,       0,       6,       4,       0,       0,       5,       5,       0,       0,       4,       6,       0,       0,       3,       7,       1,       0,       3,       7,       1, },
  {      384,    1536,     384,       0,     257,    1502,     544,       1,     162,    1410,     726,       6,      94,    1273,     917,      20,      48,    1104,    1104,      48,      20,     917,    1273,      94,       6,     726,    1410,     162,       1,     544,    1502,     257, },
  {     3342,   13369,    3342,       0,    2239,   13075,    4733,       7,    1410,   12272,    6319,      52,     816,   11078,    7983,     176,     418,    9609,    9609,     418,     176,    7983,   11078,     816,      52,    6319,   12272,    1410,       7,    4733,   13075,    2239, },
  {     1721,    6884,    1721,       0,    1153,    6732,    2437,       3,     726,    6319,    3254,      27,     420,    5704,    4111,      91,     215,    4948,    4948,     215,      91,    4111,    5704,     420,      27,    3254,    6319,     726,       3,    2437,    6732,    1153, },
  {       14,      57,      14,       0,      10,      56,      20,       0,       6,      52,      27,       0,       3,      47,      34,       1,       2,      41,      41,       2,       1,      34,      47,       3,       0,      27,      52,       6,       0,      20,      56,      10, },
  {      222,     889,     222,       0,     149,     869,     315,       0,      94,     816,     420,       3,      54,     737,     531,      12,      28,     639,     639,      28,      12,     531,     737,      54,       3,     420,     816,      94,       0,     315,     869,     149, },
  {     3017,   12068,    3017,       0,    2021,   11802,    4272,       6,    1273,   11078,    5704,      47,     737,    9999,    7206,     159,     377,    8674,    8674,     377,     159,    7206,    9999,     737,      47,    5704,   11078,    1273,       6,    4272,   11802,    2021, },
  {     2174,    8697,    2174,       0,    1457,    8506,    3079,       4,     917,    7983,    4111,      34,     531,    7206,    5194,     115,     272,    6251,    6251,     272,     115,    5194,    7206,     531,      34,    4111,    7983,     917,       4,    3079,    8506,    1457, },
  {       48,     192,      48,       0,      32,     188,      68,       0,      20,     176,      91,       1,      12,     159,     115,       3,       6,     138,     138,       6,       3,     115,     159,      12,       1,      91,     176,      20,       0,      68,     188,      32, },
  {      114,     455,     114,       0,      76,     445,     161,       0,      48,     418,     215,       2,      28,     377,     272,       6,      14,     327,     327,      14,       6,     272,     377,      28,       2,     215,     418,      48,       0,     161,     445,      76, },
  {     2617,   10468,    2617,       0,    1753,   10238,    3706,       5,    1104,    9609,    4948,      41,     639,    8674,    6251,     138,     327,    7524,    7524,     327,     138,    6251,    8674,     639,      41,    4948,    9609,    1104,       5,    3706,   10238,    1753, },
  {     2617,   10468,    2617,       0,    1753,   10238,    3706,       5,    1104,    9609,    4948,      41,     639,    8674,    6251,     138,     327,    7524,    7524,     327,     138,    6251,    8674,     639,      41,    4948,    9609,    1104,       5,    3706,   10238,    1753, },
  {      114,     455,     114,       0,      76,     445,     161,       0,      48,     418,     215,       2,      28,     377,     272,       6,      14,     327,     327,      14,       6,     272,     377,      28,       2,     215,     418,      48,       0,     161,     445,      76, },
  {       48,     192,      48,       0,      32,     188,      68,       0,      20,     176,      91,       1,      12,     159,     115,       3,       6,     138,     138,       6,       3,     115,     159,      12,       1,      91,     176,      20,       0,      68,     188,      32, },
  {     2174,    8697,    2174,       0,    1457,    8506,    3079,       4,     917,    7983,    4111,      34,     531,    7206,    5194,     115,     272,    6251,    6251,     272,     115,    5194,    7206,     531,      34,    4111,    7983,     917,       4,    3079,    8506,    1457, },
  {     3017,   12068,    3017,       0,    2021,   11802,    4272,       6,    1273,   11078,    5704,      47,     737,    9999,    7206,     159,     377,    8674,    8674,     377,     159,    7206,    9999,     737,      47,    5704,   11078,    1273,       6,    4272,   11802,    2021, },
  {      222,     889,     222,       0,     149,     869,     315,       0,      94,     816,     420,       3,      54,     737,     531,      12,      28,     639,     639,      28,      12,     531,     737,      54,       3,     420,     816,      94,       0,     315,     869,     149, },
  {       14,      57,      14,       0,      10,      56,      20,       0,       6,      52,      27,       0,       3,      47,      34,       1,       2,      41,      41,       2,       1,      34,      47,       3,       0,      27,      52,       6,       0,      20,      56,      10, },
  {     1721,    6884,    1721,       0,    1153,    6732,    2437,       3,     726,    6319,    3254,      27,     420,    5704,    4111,      91,     215,    4948,    4948,     215,      91,    4111,    5704,     420,      27,    3254,    6319,     726,       3,    2437,    6732,    1153, },
  {     3342,   13369,    3342,       0,    2239,   13075,    4733,       7,    1410,   12272,    6319,      52,     816,   11078,    7983,     176,     418,    9609,    9609,     418,     176,    7983,   11078,     816,      52,    6319,   12272,    1410,       7,    4733,   13075,    2239, },
  {      384,    1536,     384,       0,     257,    1502,     544,       1,     162,    1410,     726,       6,      94,    1273,     917,      20,      48,    1104,    1104,      48,      20,     917,    1273,      94,       6,     726,    1410,     162,       1,     544,    1502,     257, },
  {        2,       7,       2,       0,       1,       7,       3,       0,       1,       7,       3,       0,       0,       6,       4,       0,       0,       5,       5,       0,       0,       4,       6,       0,       0,       3,       7,       1,       0,       3,       7,       1, },
  {     1289,    5156,    1289,       0,     863,    5042,    1825,       3,     544,    4733,    2437,      20,     315,    4272,    3079,      68,     161,    3706,    3706,     161,      68,    3079,    4272,     315,      20,    2437,    4733,     544,       3,    1825,    5042,     863, },
  {     3561,   14244,    3561,       0,    2386,   13931,    5042,       7,    1502,   13075,    6732,      56,     869,   11802,    8506,     188,     445,   10238,   10238,     445,     188,    8506,   11802,     869,      56,    6732,   13075,    1502,       7,    5042,   13931,    2386, },
  {      610,    2439,     610,       0,     409,    2386,     863,       1,     257,    2239,    1153,      10,     149,    2021,    1457,      32,      76,    1753,    1753,      76,      32,    1457,    2021,     149,      10,    1153,    2239,     257,       1,     863,    2386,     409, },
};


#if USE_SEPARABLE_SPLINE_UPSAMPLING

#if USE_NON_INTERLEAVED
template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::interpolate(UInt uiPl, const Block<t_TDst> &rcDst, int iFac)
{

  TOF( iFac == 4 );

  int aiCoeff[4][4];
  // get the coeffs
  {
    for ( int iY = 0; iY < 4; iY++ )
    {
      for ( int iX = 0; iX < 4; iX++ )
      {
        aiCoeff[iY][iX] = int( xGetKernel ( iY, iX) );
      }
    }
  }
  const Block<short> &ri16Q6SplCoeffs = *m_pi16Q6SplCoeffFP[uiPl];
  const int iSrcYMax                        = ri16Q6SplCoeffs.getSize().getY() - 2*m_iSplPadding;
  const int iDstYMax                        = rcDst.getSize().getY();
  const int iDstXMax                        = rcDst.getSize().getX();


  // temp variables
  AllocBlock<short> cQ6HorUpsPlane( BlockSize( iDstXMax, iSrcYMax + 2*m_iSplPadding ) );
  Block<short> cQ6TmpPlane( PixelPos( 0, m_iSplPadding ), BlockSize( iDstXMax, iSrcYMax ), cQ6HorUpsPlane );

  short* piTmp = cQ6TmpPlane.getPtr();
  const int iTmpStride = cQ6TmpPlane.getStride();
  const int iSrcStride = ri16Q6SplCoeffs.getStride();
  const short* piSrcLn = &ri16Q6SplCoeffs.getEl( m_iSplPadding-1, m_iSplPadding );
  // interpolation in horizontal direction
  for ( int iY = 0; iY < iSrcYMax; iY++ )
  {
    for ( int iX = 0; iX < iDstXMax; iX++ )
    {
      const int iSrcX   = iX >> 2;
      const int iDx     = iX - ( iSrcX << 2 );
      const short* piSrc = piSrcLn + iSrcX;

      // 1d convolution
      int i32Q21Sum = 0;
      for ( int iXX = 0; iXX < 4; iXX++ )
      {
        i32Q21Sum += aiCoeff[iDx][iXX] * int( piSrc[iXX] );
      }
      const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
      piTmp[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }

  // extend using mirror symmetry
  {

    // bottom
    short* piTmp = &cQ6HorUpsPlane.getEl( 0, m_iSplPadding+iSrcYMax-1 );
    const int iTmpStride = cQ6HorUpsPlane.getStride();
    int iOffset = 0;
    for( int iY = 0; iY < 2; iY++ ) 
    {
      iOffset += iTmpStride;
      ::memcpy( piTmp + iOffset, piTmp - iOffset, iTmpStride*sizeof(short) );
    }

    // top
    piTmp = &cQ6HorUpsPlane.getEl( 0, m_iSplPadding );
    iOffset = 0;
    for( int iY = 0; iY < 1; iY++ ) // 2
    {
      iOffset += iTmpStride;
      ::memcpy( piTmp - iOffset, piTmp + iOffset, iTmpStride*sizeof(short) );
    }
  }


  {
    short* piDes = rcDst.getPtr();
    const int iDesStride = rcDst.getStride();
    const short* piSrcLn = &cQ6HorUpsPlane.getEl( 0, m_iSplPadding -1 );
    const int iSrcStride = cQ6TmpPlane.getStride();
    for ( int iY = 0; iY < iDstYMax; iY+=4 )
    {
      for ( int iDy = 0; iDy < 4; iDy++ )
      {
        const int* piCoeff = aiCoeff[iDy];
        for ( int iX = 0; iX < iDstXMax; iX++ )
        {
          const short* piSrc = piSrcLn + iX;
        // 1d convolution
          int i32Q21Sum = 0;
          for ( int iYY = 0; iYY < 4; iYY++ )
          {
            i32Q21Sum += piCoeff[iYY] * int( *piSrc );
            piSrc += iSrcStride;
          }

          const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
          piDes[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
        }
        piDes += iDesStride;
      }
      piSrcLn += iSrcStride;
    }
  }


  {
    short* piDes = rcDst.getPtr();
    const int iDesStride = rcDst.getStride();
    const short* piSrcLn = &cQ6HorUpsPlane.getEl( 0, m_iSplPadding -1 );
    const int iSrcStride = cQ6TmpPlane.getStride();
    for ( int iY = 0; iY < iDstYMax; iY+=4 )
    {
      for ( int iDy = 0; iDy < 4; iDy++ )
      {
        const int* piCoeff = aiCoeff[iDy];
        for ( int iX = 0; iX < iDstXMax; iX++ )
        {
          const short* piSrc = piSrcLn + iX;
        // 1d convolution
          int i32Q21Sum = 0;
          for ( int iYY = 0; iYY < 4; iYY++ )
          {
            i32Q21Sum += piCoeff[iYY] * int( *piSrc );
            piSrc += iSrcStride;
          }

          const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
          piDes[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
        }
        piDes += iDesStride;
      }
      piSrcLn += iSrcStride;
    }
  }

}

#else 

// this interleaved filter improves cache data locality

template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::interpolate(UInt uiPl, const Block<t_TDst> &rcDst, int iFac)
{

  TOF( iFac == 4 );

  int aiCoeff[4][4];
  // get the coeffs
  {
    for ( int iY = 0; iY < 4; iY++ )
    {
      for ( int iX = 0; iX < 4; iX++ )
      {
        aiCoeff[iY][iX] = int( xGetKernel ( iY, iX) );
      }
    }
  }

  const Block<short> &ri16Q6SplCoeffs = *m_pi16Q6SplCoeffFP[uiPl];
  const int iSrcYMax                        = ri16Q6SplCoeffs.getSize().getY() - 2*m_iSplPadding;
  const int iDstXMax                        = rcDst.getSize().getX();

  // temp variables
  AllocBlock<short> cQ6HorUpsPlane( BlockSize( iDstXMax, iSrcYMax + 2*m_iSplPadding ) );
  Block<short> cQ6TmpPlane( PixelPos( 0, m_iSplPadding ), BlockSize( iDstXMax, iSrcYMax ), cQ6HorUpsPlane );

  short* piTmp = cQ6TmpPlane.getPtr();
  const int iTmpStride = cQ6TmpPlane.getStride();
  const int iSrcStride = ri16Q6SplCoeffs.getStride();
  const short* piSrcLn = &ri16Q6SplCoeffs.getEl( m_iSplPadding-1, m_iSplPadding );

  // interpolation 2 lines in horizontal direction
  for ( int iY = 0; iY < 2; iY++ )
  {
    for ( int iX = 0; iX < iDstXMax; iX++ )
    {
      const int iSrcX   = iX >> 2;
      const int iDx     = iX - ( iSrcX << 2 );
      const short* piSrc = piSrcLn + iSrcX;

      // 1d convolution
      int i32Q21Sum = 0;
      for ( int iXX = 0; iXX < 4; iXX++ )
      {
        i32Q21Sum += aiCoeff[iDx][iXX] * int( piSrc[iXX] );
      }
      const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
      piTmp[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }

  {
    // mirror top
    short* piTmp = &cQ6HorUpsPlane.getEl( 0, m_iSplPadding );
    int iOffset = 0;
    for( int iY = 0; iY < 1; iY++ ) // 2
    {
      iOffset += iTmpStride;
      ::memcpy( piTmp - iOffset, piTmp + iOffset, iTmpStride*sizeof(short) );
    }
  }

  short* piDes = rcDst.getPtr();
  const int iDesStride = rcDst.getStride();

  for ( int iY = 2; iY < iSrcYMax; iY++ )
  {
    // horizontal 
    {
      for ( int iX = 0; iX < iDstXMax; iX++ )
      {
        const int iSrcX   = iX >> 2;
        const int iDx     = iX - ( iSrcX << 2 );
        const short* piSrc = piSrcLn + iSrcX;

        // 1d convolution
        int i32Q21Sum = 0;
        for ( int iXX = 0; iXX < 4; iXX++ )
        {
          i32Q21Sum += aiCoeff[iDx][iXX] * int( piSrc[iXX] );
        }
        const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
        piTmp[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
      }
      piTmp += iTmpStride;
      piSrcLn += iSrcStride;
    }

    // vertical
    {
      const short* piSrcLnV = piTmp-4*iTmpStride;
      for ( int iDy = 0; iDy < 4; iDy++ )
      {
        const int* piCoeff = aiCoeff[iDy];
        for ( int iX = 0; iX < iDstXMax; iX++ )
        {
          const short* piSrc = piSrcLnV + iX;
          // 1d convolution
          int i32Q21Sum = 0;
          for ( int iYY = 0; iYY < 4; iYY++ )
          {
            i32Q21Sum += piCoeff[iYY] * int( *piSrc );
            piSrc += iTmpStride;
          }
          const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
          piDes[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
        }
        piDes += iDesStride;
      }
    }
  }

  // extend using mirror symmetry
  {

    // bottom mirror
    short* piTmp = &cQ6HorUpsPlane.getEl( 0, m_iSplPadding+iSrcYMax-1 );
    const int iTmpStride = cQ6HorUpsPlane.getStride();
    int iOffset = 0;
    for( int iY = 0; iY < 2; iY++ ) 
    {
      iOffset += iTmpStride;
      ::memcpy( piTmp + iOffset, piTmp - iOffset, iTmpStride*sizeof(short) );
    }
  }


  for ( int iY = 0; iY < 2; iY++ )
  {
    piTmp += iTmpStride;
    // finish vertical
    {
      const short* piSrcLnV = piTmp-4*iTmpStride;
      for ( int iDy = 0; iDy < 4; iDy++ )
      {
        const int* piCoeff = aiCoeff[iDy];
        for ( int iX = 0; iX < iDstXMax; iX++ )
        {
          const short* piSrc = piSrcLnV + iX;
          // 1d convolution
          int i32Q21Sum = 0;
          for ( int iYY = 0; iYY < 4; iYY++ )
          {
            i32Q21Sum += piCoeff[iYY] * int( *piSrc );
            piSrc += iTmpStride;
          }
          const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
          piDes[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
        }
        piDes += iDesStride;
      }
    }
  }
}

template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::interpolate00(UInt uiPl, const Block<t_TDst> &rcDst, int iFac)  const
{

  TOF( iFac == 1 );

  const CoeffT* piCoeff = xGetFilterCoeffs(0);

  const Block<short> &ri16Q6SplCoeffs = *m_pi16Q6SplCoeffFP[uiPl];
  const int iSrcYMax                        = ri16Q6SplCoeffs.getSize().getY() - 2*m_iSplPadding;
  const int iDstXMax                        = rcDst.getSize().getX();

  // temp variables
  AllocBlock<short> cQ6HorUpsPlane( BlockSize( iDstXMax, iSrcYMax + 2*m_iSplPadding ) );
  Block<short> cQ6TmpPlane( PixelPos( 0, m_iSplPadding ), BlockSize( iDstXMax, iSrcYMax ), cQ6HorUpsPlane );

  short* piTmp = cQ6TmpPlane.getPtr();
  const int iTmpStride = cQ6TmpPlane.getStride();
  const int iSrcStride = ri16Q6SplCoeffs.getStride();
  const short* piSrcLn = &ri16Q6SplCoeffs.getEl( m_iSplPadding-1, m_iSplPadding );

  // interpolation 2 lines in horizontal direction
  for ( int iY = 0; iY < 2; iY++ )
  {
    for ( int iX = 0; iX < iDstXMax; iX++ )
    {
      const short* piSrc = piSrcLn + iX;

      // 1d convolution
      int i32Q21Sum = 0;
      for ( int iXX = 0; iXX < 4; iXX++ )
      {
        i32Q21Sum += piCoeff[iXX] * int( piSrc[iXX] );
      }
      const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
      piTmp[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }

  {
    // mirror top
    short* piTmp = &cQ6HorUpsPlane.getEl( 0, m_iSplPadding );
    int iOffset = 0;
    for( int iY = 0; iY < 1; iY++ ) // 2
    {
      iOffset += iTmpStride;
      ::memcpy( piTmp - iOffset, piTmp + iOffset, iTmpStride*sizeof(short) );
    }
  }

  short* piDes = rcDst.getPtr();
  const int iDesStride = rcDst.getStride();

  for ( int iY = 2; iY < iSrcYMax; iY++ )
  {
    // horizontal 
    {
      for ( int iX = 0; iX < iDstXMax; iX++ )
      {
        const short* piSrc = piSrcLn + iX;

        // 1d convolution
        int i32Q21Sum = 0;
        for ( int iXX = 0; iXX < 4; iXX++ )
        {
          i32Q21Sum += piCoeff[iXX] * int( piSrc[iXX] );
        }
        const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
        piTmp[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
      }
      piTmp += iTmpStride;
      piSrcLn += iSrcStride;
    }

    // vertical
    {
      const short* piSrcLnV = piTmp-4*iTmpStride;
      for ( int iX = 0; iX < iDstXMax; iX++ )
      {
        const short* piSrc = piSrcLnV + iX;
        // 1d convolution
        int i32Q21Sum = 0;
        for ( int iYY = 0; iYY < 4; iYY++ )
        {
          i32Q21Sum += piCoeff[iYY] * int( *piSrc );
          piSrc += iTmpStride;
        }
        const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
        piDes[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
      }
      piDes += iDesStride;
    }
  }

  // extend using mirror symmetry
  {

    // bottom mirror
    short* piTmp = &cQ6HorUpsPlane.getEl( 0, m_iSplPadding+iSrcYMax-1 );
    const int iTmpStride = cQ6HorUpsPlane.getStride();
    int iOffset = 0;
    for( int iY = 0; iY < 2; iY++ ) 
    {
      iOffset += iTmpStride;
      ::memcpy( piTmp + iOffset, piTmp - iOffset, iTmpStride*sizeof(short) );
    }
  }


  for ( int iY = 0; iY < 2; iY++ )
  {
    piTmp += iTmpStride;
    // finish vertical
    {
      const short* piSrcLnV = piTmp-4*iTmpStride;
      for ( int iX = 0; iX < iDstXMax; iX++ )
      {
        const short* piSrc = piSrcLnV + iX;
        // 1d convolution
        int i32Q21Sum = 0;
        for ( int iYY = 0; iYY < 4; iYY++ )
        {
          i32Q21Sum += piCoeff[iYY] * int( *piSrc );
          piSrc += iTmpStride;
        }
        const short i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
        piDes[iX] = short( ( int(m_i16Q2KernelGain) * int(i16Q6Sum) + (1<<1) ) >> 2 );
      }
      piDes += iDesStride;
    }
  }
}



#endif

#else

template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::interpolate(UInt uiPl, const Block<t_TDst> &rcDst, int iFac)
{
  TOF( (iFac == 2) || (iFac == 4) || (iFac == 8) );
  switch( iFac )
  {
  case 2:
    xInterpolate<2>( uiPl, rcDst);
    break;
  case 4:
    xInterpolate<4>( uiPl, rcDst);
    break;
  case 5:
    xInterpolate<8>( uiPl, rcDst);
    break;
  default:
    TOT(1);
  }
}

template<class t_TSrc, class t_TDst>
template<int iFac>
void CubicSplineFixed<t_TSrc, t_TDst>::xInterpolate(UInt uiPl, const Block<t_TDst> &rcDst)
{
//  TOF( (iFac == 2) || (iFac == 4) || (iFac == 8) );

  const int iFacShift = gCeilLog2( iFac );
  const int iMask = (1<<iFacShift)-1;

  const Block<short> &cSplCoeffs = *m_pi16Q6SplCoeffFP[uiPl];
  const int iDstYMax=rcDst.getSize().getY();
  const int iDstXMax=rcDst.getSize().getX();

  const short* pSpl = xGetSplineBuf();
  const int iSplBufStride = xGetSplineBufStride();
  const int iSrcBufStride = cSplCoeffs.getStride();

  const short* pSrc = &cSplCoeffs.getEl(m_iSplPadding-1,m_iSplPadding-1);
  t_TDst* pDes = &rcDst.getEl(0,0);
  const int iDesStride = rcDst.getStride();
  const int iSplGain = (m_iQ25SplGain>>15);
  const int iFacMul = (32>>iFacShift);

  // Perform Interpolation
  for( int iY=0; iY<iDstYMax; iY++)
  {
    const int iSrcY=iY>>iFacShift;
    const int iDy= (iY&iMask)*iFacMul;

    const short* pSrcLn = pSrc + iSrcY * iSrcBufStride;
    const short* pSplLn = pSpl + iDy * iSplBufStride;

    for( int iX=0; iX<iDstXMax; iX++)
    {
      const int iSrcX=iX>>iFacShift;
      const int iDx= (iX&iMask)*iFacMul;

      const short* pSrcTmp = pSrcLn + iSrcX;
      const short* pSplTmp = pSplLn + iDx;
      // 2d convolution
      int iSum = 0;
      for(int iYY=0; iYY<4; iYY++)
      {
        for(int iXX=0; iXX<4; iXX++)
        {
          iSum += int( pSrcTmp[iXX]) * int(pSplTmp[iXX]);
        }
        pSplTmp += iSplBufStride;
        pSrcTmp += iSrcBufStride;
      }

      pDes[iX]=t_TDst(short( ( ((iSum+(1<<14))>>15) * iSplGain + (1<<9) )>>10));
    }
    pDes += iDesStride;
  }
}
#endif

// internal function for spline initialization
template<class t_TSrc, class t_TDst>
short CubicSplineFixed<t_TSrc, t_TDst>::xInitRowCausalCoeff(const Block<const t_TSrc> &rcSrc, int iY, int N)
{
  short i16Q15Z_n=m_i16Q15Z1;
  // only sum up 16 value
  const int iXMax = (N > 16) ? 16 : N;
  const t_TSrc* pcSrc = &rcSrc.getEl(0,iY);
  short i16Q6Sum = short(pcSrc[0])<<6;

  for (int iX = 1; iX < iXMax; iX++)
  {
    i16Q6Sum += xInt16Multiplication( i16Q15Z_n, short(pcSrc[iX])<<6 );
    i16Q15Z_n = xInt16Multiplication( i16Q15Z_n, m_i16Q15Z1 );
  }
  return i16Q6Sum;
}

// internal function for spline initialization
template<class t_TSrc, class t_TDst>
short CubicSplineFixed<t_TSrc, t_TDst>::xInitColCausalCoeff(const Block<const short> &ri16Q6Src, int iX, int N)
{
  short i16Q15Z_n = m_i16Q15Z1;
  // only sum up 16 value
  const int iYMax = (N > 16) ? 16 : N;
  const int iSrcStride = ri16Q6Src.getStride();
  const short* pcSrc = &ri16Q6Src.getEl(iX,0);
  short i16Q6Sum = *pcSrc;
  for (int iY = 1; iY < iYMax; iY++)
  {
    pcSrc += iSrcStride;
    i16Q6Sum += xInt16Multiplication( i16Q15Z_n, *pcSrc );
    i16Q15Z_n = xInt16Multiplication( i16Q15Z_n, m_i16Q15Z1 );
  }
  return i16Q6Sum;
}


template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::calcCoeffs(const Block<const t_TSrc> &rcSrc, UInt uiPl )
{
  int iY, iX;
  const Block<short> &ri16Q6SplCoeffs = *m_pi16Q6SplCoeffFP[uiPl];
  const int iXMax = ri16Q6SplCoeffs.getSize().getX()-2*m_iSplPadding; 
  const int iYMax = ri16Q6SplCoeffs.getSize().getY()-2*m_iSplPadding;
  Block<short> i16Q6Spl(PixelPos(m_iSplPadding,m_iSplPadding), BlockSize(iXMax, iYMax), ri16Q6SplCoeffs);
//  short i16Q15Z2 = m_i16Q15Z1_c;

  const t_TSrc* pSrc    = rcSrc.getPtr();
  const int iSrcStride  = rcSrc.getStride();
  short* pSpl           = i16Q6Spl.getPtr();
  const int iSplStride  = i16Q6Spl.getStride();
  register short r;

  // filter along rows
  for (iY=0; iY<iYMax; iY++)
  {
    // init causal coeff
    pSpl[0] = r = xInitRowCausalCoeff(rcSrc, iY, iXMax);

    // causal filtering
    {
      const int iFactor = m_i16Q15Z1;
      for (iX=1; iX<iXMax; iX++)
      {
        pSpl[iX] = r = xInt16Multiplication( r, iFactor ) + ( short(pSrc[iX])<<6 );
      }
    }

    // init anti-causal coeff
    pSpl[iXMax-1] = r = xInt16Multiplication( xInt16Multiplication( pSpl[iXMax-2], m_i16Q15Z1 )+pSpl[iXMax-1], m_i16Q15Z1_c );

    // anti-causal filtering
    {
      const int iFactor = m_i16Q15Z1;
      for (iX=iXMax-2; iX>=0; iX--)
      {
        pSpl[iX] = r = xInt16Multiplication( r-pSpl[iX], iFactor );
      }
    }
    pSpl += iSplStride;
    pSrc += iSrcStride;
  }



  // filter along columns
  pSpl           = i16Q6Spl.getPtr();

  for (iX=0; iX<iXMax; iX++)
  {
    // init causal coeff
    pSpl[iX]=xInitColCausalCoeff(i16Q6Spl, iX, iYMax);
  }

  // causal filtering
  for (iY=1; iY<iYMax; iY++)
  {
    pSpl += iSplStride;
    {
      const int iFactor = m_i16Q15Z1;
      for (iX=0; iX<iXMax; iX++)
      {
        pSpl[iX] += xInt16Multiplication( pSpl[iX-iSplStride], iFactor );
      }
    }
  }


  for (iX=0; iX<iXMax; iX++)
  {
    // init anti-causal coeff
    pSpl[iX] = xInt16Multiplication( xInt16Multiplication( pSpl[iX-iSplStride], m_i16Q15Z1 )+pSpl[iX], m_i16Q15Z1_c );
  }

  // anti-causal filtering
  for (iY=iYMax-2; iY>=0; iY--)
  {
    pSpl -= iSplStride;
    {
      const int iFactor = m_i16Q15Z1;
      for (iX=0; iX<iXMax; iX++)
      {
        pSpl[iX] = xInt16Multiplication( pSpl[iX+iSplStride]-pSpl[iX], iFactor );
      }
    }
  }

  // extend using mirror symmetry
  {
    short* pSpl           = i16Q6Spl.getPtr();
    const int iStride     = i16Q6Spl.getStride();
    for(iY=0; iY<iYMax; iY++)
    {
      //left
      for( iX=1; iX<=m_iSplPadding; iX++ )
      {
        pSpl[-iX] = pSpl[ iX ];
      }

      // right
      pSpl += iXMax-1;
      for( iX=1; iX<=m_iSplPadding; iX++ )
      {
        pSpl[iX] = pSpl[ -iX ];
      }
      pSpl += iStride -(iXMax-1);
    }

//#if ! USE_SEPARABLE_SPLINE_UPSAMPLING
    //bottom
    pSpl -= m_iSplPadding + iStride;
    int iOffset = iStride;
    for(iY=0; iY<m_iSplPadding; iY++)
    {
      for(iX=0; iX<iStride; iX++)
      {
        pSpl[iX + iOffset] = pSpl[ iX - iOffset];
      }
      iOffset += iStride;
    }

    //top
    pSpl = i16Q6Spl.getPtr()-m_iSplPadding;
    iOffset = iStride;
    for(iY=0; iY<m_iSplPadding; iY++)
    {
      for(iX=0; iX<iStride; iX++)
      {
        pSpl[iX - iOffset] = pSpl[ iX + iOffset];
      }
      iOffset += iStride;
    }
//#endif
  }
}


template<class t_TSrc, class t_TDst>
class CubicSpline
  : public InterpolationIf<t_TSrc, t_TDst>
{
public:
  CubicSpline() : m_iSplPadding( 2 ), m_dSplGain( 36.0 ), m_dZ1(-2+sqrt(double(3.0))) {}
  CubicSpline( const double dSplGain, const double dZ1 ) : m_iSplPadding( 2 ), m_dSplGain( dSplGain ), m_dZ1( dZ1 ) {}
  virtual ~CubicSpline()
  {
    std::vector<Block<double>*>::iterator it;
    for (it=m_pcSplCoeff.begin(); it!=m_pcSplCoeff.end(); ++it)
    {
      delete (*it);
    }
    m_pcSplCoeff.clear();
  }

  // Allocate memory for spline_coeffs + 2*padding
  ErrVal allocCoeffs(const PrimBlockSize& rcPbSize, const std::vector<Subsampling> &racOrigSub)
  {
    for( UInt ui = 0; ui < racOrigSub.size(); ui++ )
    {
      BlockSize cSize = rcPbSize * racOrigSub[ui];
      m_pcSplCoeff.push_back( new AllocBlock<double>( BlockSize(cSize.getX()+2*m_iSplPadding, cSize.getY()+2*m_iSplPadding) ) );
      ROF( m_pcSplCoeff[ui] );
    }
    return Err::m_nOK;
  }

  // function to calculate spline coeffs
  void calcCoeffs(const Block<const t_TSrc> &rcSrc, UInt uiPl);
  // function to interpolate using spline coeffs
  void interpolate(UInt uiPl, const Block<t_TDst> &rcDst, int iFac);

  double splineInterp(UInt uiPl, FracPos cPos, const BlockSize &rcPicSize) const;

  int clpMirSym(int iLoc, int iPicMax) const;

  void computeGradients( double dXLoc,
                         double dYLoc,
                         std::vector<double> &racGrVecX1,
                         std::vector<double> &racGrVecY1,
                         std::vector<double> &racGrVecX2,
                         std::vector<double> &racGrVecY2 ) const;
  void computeGradCoeffs( const FracPos &rcPos, const Block<double> &rcGradCoeffsX, const Block<double> &rcGradCoeffsY) const;

  void gradCalc( UInt uiPl, const FracPos &rcPos, const BlockSize &rcPicSize, double &rcGradX, double &rcGradY ) const;

  const BlockSize getOrigPlaneSize(UInt uiPlane) const { return  m_pcSplCoeff[uiPlane]->getSize() - BlockSize(2*m_iSplPadding, 2*m_iSplPadding); }

  const Block<double>* getPlane(UInt uiPl) const { return m_pcSplCoeff[uiPl]; }

private:

  double xInitRowCausalCoeff(const Block<const t_TSrc> &rcSrc, int iY, int N);
  double xInitColCausalCoeff(const Block<const double> &rcSrc, int iX, int N);

  virtual void   xSplineVal       ( const Block<double> &rcSplVal, double dDx, double dDy) const;
  double xConv            ( FracPos cPos, const Block<double> &rcSplCoeffs, const BlockSize &rcPicSize, const Block<double> &rcVal) const;

  const int                   m_iSplPadding;
  const double                m_dSplGain;
  std::vector<Block<double>*> m_pcSplCoeff;
  const double                m_dZ1;
};

// interpolation of pixel using spline coeffs
template<class t_TSrc, class t_TDst>
double CubicSpline<t_TSrc, t_TDst>::splineInterp(UInt uiPl, FracPos cPos, const BlockSize &rcPicSize) const
{
  double dSum;
  double dDx, dDy;
  AllocBlock<double> cSplVal( BlockSize(4, 4) );
  const Block<double> &cSplCoeffs = *m_pcSplCoeff[uiPl];

  dDx = cPos.getX() - floor( cPos.getX() );
  dDy = cPos.getY() - floor( cPos.getY() );

  xSplineVal(cSplVal, dDx, dDy);
  dSum = xConv(cPos, cSplCoeffs, rcPicSize, cSplVal);

  return dSum;
}

template<class t_TSrc, class t_TDst>
void CubicSpline<t_TSrc, t_TDst>::computeGradients( double dXLoc,
                                    double dYLoc,
                                    std::vector<double> &racGrVecX1,
                                    std::vector<double> &racGrVecY1,
                                    std::vector<double> &racGrVecX2,
                                    std::vector<double> &racGrVecY2 ) const
{
  double dx=dXLoc-(int)floor(dXLoc);
  double dy=dYLoc-(int)floor(dYLoc);

  racGrVecX1[0] = -(pow((1-dx),2))/2.0; racGrVecX1[1] = (3.0/2.0*dx*dx-2.0*dx); racGrVecX1[2] = (-3.0/2.0*dx*dx + dx + 1.0/2.0); racGrVecX1[3] = (dx*dx)/2.0;
  racGrVecY1[0] = 1.0/6.0*pow((1-dy),3); racGrVecY1[1] = (2.0/3.0-dy*dy+1.0/2.0*dy*dy*dy); racGrVecY1[2] = (1.0/6.0+1.0/2.0*dy*dy+1.0/2.0*dy-1.0/2.0*dy*dy*dy); racGrVecY1[3] = (1.0/6.0*dy*dy*dy);
  racGrVecX2[0] = 1.0/6.0*pow((1-dx),3); racGrVecX2[1] = (2.0/3.0-dx*dx+1.0/2.0*dx*dx*dx); racGrVecX2[2] = (1.0/6.0+1.0/2.0*dx*dx+1.0/2.0*dx-1.0/2.0*dx*dx*dx); racGrVecX2[3] = (1.0/6.0*dx*dx*dx);
  racGrVecY2[0] = -(pow((1-dy),2))/2.0; racGrVecY2[1] = (3.0/2.0*dy*dy-2.0*dy); racGrVecY2[2] = (-3.0/2.0*dy*dy + dy + 1.0/2.0); racGrVecY2[3] = (dy*dy)/2.0;
}

template<class t_TSrc, class t_TDst>
void CubicSpline<t_TSrc, t_TDst>::computeGradCoeffs( const FracPos &rcPos, const Block<double> &rcGradCoeffsX, const Block<double> &rcGradCoeffsY) const
{
  double dx = rcPos.getX() - int( floor( rcPos.getX() ) );
  double dy = rcPos.getY() - int( floor( rcPos.getY() ) );
  std::vector<double> acGrVecX1(4), acGrVecY1(4), acGrVecX2(4), acGrVecY2(4);

  acGrVecX1[0] = -(pow((1-dx),2))/2.0;  acGrVecX1[1] = (3.0/2.0*dx*dx-2.0*dx);           acGrVecX1[2] = (-3.0/2.0*dx*dx + dx + 1.0/2.0);                     acGrVecX1[3] = (dx*dx)/2.0;
  acGrVecY1[0] = 1.0/6.0*pow((1-dy),3); acGrVecY1[1] = (2.0/3.0-dy*dy+1.0/2.0*dy*dy*dy); acGrVecY1[2] = (1.0/6.0+1.0/2.0*dy*dy+1.0/2.0*dy-1.0/2.0*dy*dy*dy); acGrVecY1[3] = (1.0/6.0*dy*dy*dy);
  acGrVecX2[0] = 1.0/6.0*pow((1-dx),3); acGrVecX2[1] = (2.0/3.0-dx*dx+1.0/2.0*dx*dx*dx); acGrVecX2[2] = (1.0/6.0+1.0/2.0*dx*dx+1.0/2.0*dx-1.0/2.0*dx*dx*dx); acGrVecX2[3] = (1.0/6.0*dx*dx*dx);
  acGrVecY2[0] = -(pow((1-dy),2))/2.0;  acGrVecY2[1] = (3.0/2.0*dy*dy-2.0*dy);           acGrVecY2[2] = (-3.0/2.0*dy*dy + dy + 1.0/2.0);                     acGrVecY2[3] = (dy*dy)/2.0;

  PixelPos cPos;
  for (cPos.getY() = 0; cPos.getY() < 4; cPos.getY()++)
  {
    for (cPos.getX() = 0; cPos.getX() < 4; cPos.getX()++)
    {
      rcGradCoeffsX.getEl(cPos) = acGrVecX1[cPos.getX()] * acGrVecY1[cPos.getY()];
      rcGradCoeffsY.getEl(cPos) = acGrVecX2[cPos.getX()] * acGrVecY2[cPos.getY()];
    }
  }

}

template<class t_TSrc, class t_TDst>
void CubicSpline<t_TSrc, t_TDst>::gradCalc( UInt uiPl, const FracPos &rcPos, const BlockSize &rcPicSize, double &rcGradX, double &rcGradY ) const
{
  const Block<double> &cSplCoeffs = *m_pcSplCoeff[uiPl];
  AllocBlock<double> cGradCoeffsX( BlockSize(4, 4) ), cGradCoeffsY( BlockSize(4, 4) );

 computeGradCoeffs( rcPos, cGradCoeffsX, cGradCoeffsY );
 rcGradX = xConv(rcPos, cSplCoeffs, rcPicSize, cGradCoeffsX);
 rcGradY = xConv(rcPos, cSplCoeffs, rcPicSize, cGradCoeffsY);
}

template<class t_TSrc, class t_TDst>
int CubicSpline<t_TSrc, t_TDst>::clpMirSym(int iLoc, int iPicMax) const
{
  if(iLoc >= iPicMax)
  {
    iLoc = (iPicMax - 1) - (iLoc - (iPicMax - 1));
  }
  return iLoc;
}

template<class t_TSrc, class t_TDst>
double CubicSpline<t_TSrc, t_TDst>::xConv(FracPos cPos, const Block<double> &rcSplCoeffs, const BlockSize &rcPicSize, const Block<double> &rcVal) const
{
  int iXX, iYY;
  double dSum = 0;
  int iXClip, iYClip;
  double dXSplLoc, dYSplLoc;

  for(iYY = -1; iYY <3; iYY++)
  {
    for(iXX = -1; iXX < 3; iXX++)
    {
      dXSplLoc = cPos.getX() + iXX;
      dYSplLoc = cPos.getY() + iYY;

      iXClip = abs( int( floor(dXSplLoc) ) );
      iYClip = abs( int( floor(dYSplLoc) ) );

      iXClip = clpMirSym(iXClip, rcPicSize.getX());
      iYClip = clpMirSym(iYClip, rcPicSize.getY());

      dSum = dSum + rcSplCoeffs.getEl(iXClip+m_iSplPadding, iYClip+m_iSplPadding) * rcVal.getEl(iXX+1, iYY+1);
    }
  }

  dSum *= m_dSplGain;

  return dSum;
}

// return 4x4 spline values for a given position (dx,dy)
template<class t_TSrc, class t_TDst>
void CubicSpline<t_TSrc, t_TDst>::xSplineVal(const Block<double> &rcSplVal, double dDx, double dDy) const
{
  int iX, iY;

  double dValX,dValY;
  double dSplX,dSplY;

  for(iY=-1; iY<=2; iY++)
  {
    // get Y component of spline
    dValY=fabs(iY-dDy);
    if(dValY >= 2)
      dSplY = 0;
    else if (dValY>=1 && dValY < 2)
      dSplY = pow((2-dValY),3)/6;
    else
      dSplY = 2.0/3 - pow(dValY,2) + pow(dValY,3)/2;

    for(iX=-1; iX<=2; iX++)
    {
      // get X component of spline
      dValX=fabs(iX-dDx);
      if(dValX >= 2)
        dSplX = 0;
      else if (dValX>=1 && dValX < 2)
        dSplX = pow((2-dValX),3)/6;
      else
        dSplX = 2.0/3 - pow(dValX,2) + pow(dValX,3)/2;

      rcSplVal.getEl(iX+1, iY+1) = dSplX*dSplY;
    }
  }
}

// interpolation of block using spline coeffs
template<class t_TSrc, class t_TDst>
void CubicSpline<t_TSrc, t_TDst>::interpolate(UInt uiPl, const Block<t_TDst> &rcDst, int iFac)
{
  int iX, iY;
  const Block<double> &cSplCoeffs = *m_pcSplCoeff[uiPl];
  int iDstYMax=rcDst.getSize().getY(), iDstXMax=rcDst.getSize().getX();

  // Perform Interpolation
  for(iY=0; iY<iDstYMax; iY++)
  {
    int iSrcY=int(iY/iFac);
    double dDy=double(iY-iSrcY*iFac)/iFac;

    for(iX=0; iX<iDstXMax; iX++)
    {
      int iXX, iYY;
      double dSum=0;
      int iSrcX=int(iX/iFac);
      AllocBlock<double> dSplVal( BlockSize(4, 4) );
      double dDx=double(iX-iSrcX*iFac)/iFac;

      // choose spline values
      xSplineVal(dSplVal, dDx, dDy);

      // 2d convolution
      for(iYY=0; iYY<4; iYY++)
      {
        for(iXX=0; iXX<4; iXX++)
        {
          dSum = dSum + m_dSplGain*cSplCoeffs.getEl(m_iSplPadding+iSrcX-1+iXX,m_iSplPadding+iSrcY-1+iYY) * dSplVal.getEl(iXX, iYY);
        }
      }
      rcDst.getEl(iX,iY)=t_TDst(pow(2.0,EXTRA_PRECISION_BITS)*dSum+0.5);

    }
  }
}

// internal function for spline initialization
template<class t_TSrc, class t_TDst>
double CubicSpline<t_TSrc, t_TDst>::xInitRowCausalCoeff(const Block<const t_TSrc> &rcSrc, int iY, int N)
{
  int iX;
  double Sum, c_0;
  double zn, iz, z2n;

  zn=m_dZ1;
  iz = 1.0 / m_dZ1;
  z2n = pow(m_dZ1, (N - 1));
  Sum = rcSrc.getEl(0,iY) + z2n * rcSrc.getEl(N-1,iY);
  z2n = z2n * z2n * iz;

  for (iX = 1; iX < N-1; iX++)
  {
    Sum = Sum + (zn + z2n) * rcSrc.getEl(iX,iY);
    zn = zn * m_dZ1;
    z2n = z2n * iz;
  }

  c_0 = Sum/(1.0 - zn * zn);

  return c_0;
}

// internal function for spline initialization
template<class t_TSrc, class t_TDst>
double CubicSpline<t_TSrc, t_TDst>::xInitColCausalCoeff(const Block<const double> &rcSrc, int iX, int N)
{
  int iY;
  double Sum, c_0;
  double zn, iz, z2n;

  zn=m_dZ1;
  iz = 1.0 / m_dZ1;
  z2n = pow(m_dZ1, (N - 1));

  Sum = rcSrc.getEl(iX,0) + z2n * rcSrc.getEl(iX,N-1);
  z2n = z2n * z2n * iz;

  for (iY = 1; iY < N-1; iY++)
  {
    Sum = Sum + (zn + z2n) * rcSrc.getEl(iX,iY);
    zn = zn * m_dZ1;
    z2n = z2n * iz;
  }

  c_0 = Sum/(1.0 - zn * zn);

  return c_0;
}

// Convert image samples to spline coefficients
template<class t_TSrc, class t_TDst>
void CubicSpline<t_TSrc, t_TDst>::calcCoeffs(const Block<const t_TSrc> &rcSrc, UInt uiPl)
{
  int iY, iX;
  double dZ1_c = (m_dZ1 / (m_dZ1 * m_dZ1 - 1.0));
  const Block<double> &cSplCoeffs = *m_pcSplCoeff[uiPl];
  int iXMax=cSplCoeffs.getSize().getX()-2*m_iSplPadding, iYMax=cSplCoeffs.getSize().getY()-2*m_iSplPadding;
  Block<double> cSpl(PixelPos(m_iSplPadding,m_iSplPadding), BlockSize(iXMax, iYMax), cSplCoeffs);

  // filter along rows
  for (iY=0; iY<iYMax; iY++)
  {
    // init causal coeff
    cSpl.getEl(0,iY)=xInitRowCausalCoeff(rcSrc, iY, iXMax);
    // causal filtering
    for (iX=1; iX<iXMax; iX++)
    {
      cSpl.getEl(iX,iY)= rcSrc.getEl(iX,iY) + m_dZ1 * cSpl.getEl(iX-1,iY);
    }

    // init anti-causal coeff
    cSpl.getEl(iXMax-1,iY) = dZ1_c * (m_dZ1 * cSpl.getEl(iXMax-2,iY) + cSpl.getEl(iXMax-1,iY));

    // anti-causal filtering
    for (iX=iXMax-2; iX>=0; iX--)
    {
      cSpl.getEl(iX,iY) = m_dZ1 * (cSpl.getEl(iX+1,iY) - cSpl.getEl(iX,iY));
    }
  }

  // filter along columns
  for (iX=0; iX<iXMax; iX++)
  {
    // init causal coeff
    cSpl.getEl(iX,0)=xInitColCausalCoeff(cSpl, iX, iYMax);

    // causal filtering
    for (iY=1; iY<iYMax; iY++)
    {
      cSpl.getEl(iX,iY) = cSpl.getEl(iX,iY) + m_dZ1 * cSpl.getEl(iX,iY-1);
    }

    // init anti-causal coeff
    cSpl.getEl(iX,iYMax-1) = dZ1_c * (m_dZ1 * cSpl.getEl(iX,iYMax-2) + cSpl.getEl(iX,iYMax-1));

    // anti-causal filtering
    for (iY=iYMax-2; iY>=0; iY--)
    {
      cSpl.getEl(iX,iY) = m_dZ1 * (cSpl.getEl(iX,iY+1) - cSpl.getEl(iX,iY));
    }
  }

  // extend using mirror symmetry
  int iSrcX, iDstX, iSrcY, iDstY;
  // left
  for(iSrcX=1, iDstX=m_iSplPadding-1; iSrcX<=m_iSplPadding; iSrcX++, iDstX--)
  {
    for(iSrcY=0, iDstY=m_iSplPadding; iSrcY<iYMax; iSrcY++, iDstY++)
    {
      cSplCoeffs.getEl(iDstX,iDstY)=cSpl.getEl(iSrcX,iSrcY);
    }
  }

  // right
  for(iSrcX=iXMax-2, iDstX=m_iSplPadding+iXMax; iSrcX>iXMax-2-m_iSplPadding; iSrcX--, iDstX++)
  {
    for(iSrcY=0, iDstY=m_iSplPadding; iSrcY<iYMax; iSrcY++, iDstY++)
    {
      cSplCoeffs.getEl(iDstX,iDstY)=cSpl.getEl(iSrcX,iSrcY);
    }
  }

  // top
  for(iSrcY=m_iSplPadding+1, iDstY=m_iSplPadding-1; iDstY>=0; iSrcY++, iDstY--)
  {
    for(iX=0; iX<iXMax+2*m_iSplPadding; iX++)
    {
      cSplCoeffs.getEl(iX,iDstY)=cSplCoeffs.getEl(iX,iSrcY);
    }
  }

  // bottom
  for(iSrcY=m_iSplPadding+iYMax-2, iDstY=m_iSplPadding+iYMax; iDstY<iYMax+2*m_iSplPadding; iSrcY--, iDstY++)
  {
    for(iX=0; iX<iXMax+2*m_iSplPadding; iX++)
    {
      cSplCoeffs.getEl(iX,iDstY)=cSplCoeffs.getEl(iX,iSrcY);
    }
  }
}

template<class t_TSrc, class t_TDst>
class QuadraticSpline
  : public CubicSpline<t_TSrc, t_TDst>
{
public:
  QuadraticSpline() : CubicSpline<t_TSrc, t_TDst>( 64.0, -3+2*sqrt(2.0) ) {}
  virtual ~QuadraticSpline() {}
private:
  void xSplineVal ( const Block<double> &rcSplVal, double dDx, double dDy) const;

  static const double sm_aadSplValQPelLUT[16][16];
};

template<class t_TSrc, class t_TDst>
const double QuadraticSpline<t_TSrc, t_TDst>::sm_aadSplValQPelLUT[16][16]=
{
  {     1.562500000000000e-02,    9.375000000000000e-02,    1.562500000000000e-02,    0.000000000000000e+00,    3.906250000000000e-03,    8.593750000000000e-02,    3.515625000000000e-02,    0.000000000000000e+00,    0.000000000000000e+00,    6.250000000000000e-02,    6.250000000000000e-02,    0.000000000000000e+00,    0.000000000000000e+00,    3.515625000000000e-02,    8.593750000000000e-02,    3.906250000000000e-03, },
  {     9.375000000000000e-02,    5.625000000000000e-01,    9.375000000000000e-02,    0.000000000000000e+00,    2.343750000000000e-02,    5.156250000000000e-01,    2.109375000000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    3.750000000000000e-01,    3.750000000000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    2.109375000000000e-01,    5.156250000000000e-01,    2.343750000000000e-02, },
  {     1.562500000000000e-02,    9.375000000000000e-02,    1.562500000000000e-02,    0.000000000000000e+00,    3.906250000000000e-03,    8.593750000000000e-02,    3.515625000000000e-02,    0.000000000000000e+00,    0.000000000000000e+00,    6.250000000000000e-02,    6.250000000000000e-02,    0.000000000000000e+00,    0.000000000000000e+00,    3.515625000000000e-02,    8.593750000000000e-02,    3.906250000000000e-03, },
  {     0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00, },
  {     3.906250000000000e-03,    2.343750000000000e-02,    3.906250000000000e-03,    0.000000000000000e+00,    9.765625000000000e-04,    2.148437500000000e-02,    8.789062500000000e-03,    0.000000000000000e+00,    0.000000000000000e+00,    1.562500000000000e-02,    1.562500000000000e-02,    0.000000000000000e+00,    0.000000000000000e+00,    8.789062500000000e-03,    2.148437500000000e-02,    9.765625000000000e-04, },
  {     8.593750000000000e-02,    5.156250000000000e-01,    8.593750000000000e-02,    0.000000000000000e+00,    2.148437500000000e-02,    4.726562500000000e-01,    1.933593750000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    3.437500000000000e-01,    3.437500000000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    1.933593750000000e-01,    4.726562500000000e-01,    2.148437500000000e-02, },
  {     3.515625000000000e-02,    2.109375000000000e-01,    3.515625000000000e-02,    0.000000000000000e+00,    8.789062500000000e-03,    1.933593750000000e-01,    7.910156250000000e-02,    0.000000000000000e+00,    0.000000000000000e+00,    1.406250000000000e-01,    1.406250000000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    7.910156250000000e-02,    1.933593750000000e-01,    8.789062500000000e-03, },
  {     0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00, },
  {     0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00, },
  {     6.250000000000000e-02,    3.750000000000000e-01,    6.250000000000000e-02,    0.000000000000000e+00,    1.562500000000000e-02,    3.437500000000000e-01,    1.406250000000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    2.500000000000000e-01,    2.500000000000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    1.406250000000000e-01,    3.437500000000000e-01,    1.562500000000000e-02, },
  {     6.250000000000000e-02,    3.750000000000000e-01,    6.250000000000000e-02,    0.000000000000000e+00,    1.562500000000000e-02,    3.437500000000000e-01,    1.406250000000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    2.500000000000000e-01,    2.500000000000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    1.406250000000000e-01,    3.437500000000000e-01,    1.562500000000000e-02, },
  {     0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00, },
  {     0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00, },
  {     3.515625000000000e-02,    2.109375000000000e-01,    3.515625000000000e-02,    0.000000000000000e+00,    8.789062500000000e-03,    1.933593750000000e-01,    7.910156250000000e-02,    0.000000000000000e+00,    0.000000000000000e+00,    1.406250000000000e-01,    1.406250000000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    7.910156250000000e-02,    1.933593750000000e-01,    8.789062500000000e-03, },
  {     8.593750000000000e-02,    5.156250000000000e-01,    8.593750000000000e-02,    0.000000000000000e+00,    2.148437500000000e-02,    4.726562500000000e-01,    1.933593750000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    3.437500000000000e-01,    3.437500000000000e-01,    0.000000000000000e+00,    0.000000000000000e+00,    1.933593750000000e-01,    4.726562500000000e-01,    2.148437500000000e-02, },
  {     3.906250000000000e-03,    2.343750000000000e-02,    3.906250000000000e-03,    0.000000000000000e+00,    9.765625000000000e-04,    2.148437500000000e-02,    8.789062500000000e-03,    0.000000000000000e+00,    0.000000000000000e+00,    1.562500000000000e-02,    1.562500000000000e-02,    0.000000000000000e+00,    0.000000000000000e+00,    8.789062500000000e-03,    2.148437500000000e-02,    9.765625000000000e-04, },
};

// return 4x4 spline values for a given position (dx,dy)
template<class t_TSrc, class t_TDst>
void QuadraticSpline<t_TSrc, t_TDst>::xSplineVal(const Block<double> &rcSplVal, double dDx, double dDy) const
{
  int iX, iY;

  int iDx = int(dDx*4);
  int iDy = int(dDy*4);

  for(iY=-1; iY<=2; iY++)
  {
    for(iX=-1; iX<=2; iX++)
    {
      rcSplVal.getEl(iX+1, iY+1) = sm_aadSplValQPelLUT[iDx*4+iX+1][iDy*4+iY+1];
    }
  }
}


template<class t_TSrc, class t_TDst>
class QuinticSplineFixed
  : public InterpolationIf<t_TSrc, t_TDst>
{
public:
  QuinticSplineFixed()
  : m_iSplPadding(3)
  , m_iSplGain(120*120)
  , m_iQSplGain(0)
  , m_i16Q15Z1        ( short(-0.430575347099978*pow(2.0,15)+0.5) )
  , m_i16Q15Z1_c      ( short( 0.528569571222538*pow(2.0,15)+0.5) )
  , m_i16Q15Z2        ( short(-0.043096288203263*pow(2.0,15)+0.5) )
  , m_i16Q15Z2_c      ( short( 0.043176479449236*pow(2.0,15)+0.5) )
  , m_iQRowCausal2    (  7 )
  , m_iQRowAntiCausal2(  9 )
  , m_iQColCausal2    ( 11 )
  , m_iQColAntiCausal2( 14 )
  , m_iSumShift       (  9 )
  , m_iQVal           ( 16 ) {}
  QuinticSplineFixed( const int iSplGain, const int iQSplGain, const short i16Q15Z1, const short i16Q15Z1_c, const short i16Q15Z2, const short i16Q15Z2_c, const int iQRowCausal2, const int iQRowAntiCausal2, const int iQColCausal2, const int iQColAntiCausal2, const int iSumShift, const int iQVal )
  : m_iSplPadding     (3)
  , m_iSplGain        ( iSplGain )
  , m_iQSplGain       ( iQSplGain )
  , m_i16Q15Z1        ( i16Q15Z1 )
  , m_i16Q15Z1_c      ( i16Q15Z1_c )
  , m_i16Q15Z2        ( i16Q15Z2 )
  , m_i16Q15Z2_c      ( i16Q15Z2_c )
  , m_iQRowCausal2    ( iQRowCausal2 )
  , m_iQRowAntiCausal2( iQRowAntiCausal2 )
  , m_iQColCausal2    ( iQColCausal2 )
  , m_iQColAntiCausal2( iQColAntiCausal2 )
  , m_iSumShift       ( iSumShift )
  , m_iQVal           ( iQVal ) {}

  virtual ~QuinticSplineFixed()
  {
    std::vector<Block<short>*>::iterator it;
    for (it=m_pi16Q6SplCoeff.begin(); it!=m_pi16Q6SplCoeff.end(); ++it)
    {
      delete (*it);
    }
    m_pi16Q6SplCoeff.clear();
  }

  // Allocate memory for spline_coeffs + 2*padding
  ErrVal allocCoeffs(const PrimBlockSize& cPbSize, const std::vector<Subsampling> &racOrigSub)
  {
    for( UInt ui = 0; ui < racOrigSub.size(); ui++ )
    {
      BlockSize cSize = cPbSize * racOrigSub[ui];
      m_pi16Q6SplCoeff.push_back( new AllocBlock<short>( BlockSize(cSize.getX()+2*m_iSplPadding, cSize.getY()+2*m_iSplPadding) ) );
      ROF( m_pi16Q6SplCoeff[ui] );
    }
    return Err::m_nOK;
  }

  // function to calculate spline coeffs
  void calcCoeffs(const Block<const t_TSrc> &rcSrc, UInt uiPl);

  // function to interpolate using spline coeffs
  void interpolate(UInt uiPl, const Block<t_TDst> &rcDst, int iFac);

private:

  template<int iFac>
  void xInterpolate(UInt uiPl, const Block<t_TDst> &rcDst);

  template<class t_T>
  short  xInitRowCausalCoeff(const Block<const t_T> &rcSrc, const int iSrcShift, const int iY, const int N, const short i16Q15Z);
  short  xInitColCausalCoeff(const Block<const short> &ri16Q6Src, const int iX, const int N, const short i16Q15Z);
  template<class t_T>
  void xSplineCoeffs( const Block<const t_T> &rcSrc, const int iSrcShift, Block<short> &raai16Q6Spl, const short i16Q15Z, const short i16Q15Z_c, const int iQRowCausal, const int iQRowAntiCausal, const int iQColCausal, const int iQColAntiCausal  );
  __inline virtual int   xGetSplineVal       ( int iX, int iY )         const { return int(sm_aai16Q16SplVal8PelLUT[iX][iY]); }
  __inline         short xInt16Multiplication( const short i16A, const short i16B ) const { return short( ( ( int(i16A)*int(i16B) )+(1<<14) )>>15); }
  virtual const short* xGetSplineBuf()      const { return sm_aai16Q16SplVal8PelLUT[0]; }
  virtual const int xGetSplineBufStride() const { return 48; }

  const int                   m_iSplPadding;
  const int                   m_iSplGain;
  const int                   m_iQSplGain;
  std::vector<Block<short>*>  m_pi16Q6SplCoeff;
  const short                 m_i16Q15Z1;
  const short                 m_i16Q15Z1_c;
  const short                 m_i16Q15Z2;
  const short                 m_i16Q15Z2_c;
  const int                   m_iQRowCausal2;
  const int                   m_iQRowAntiCausal2;
  const int                   m_iQColCausal2;
  const int                   m_iQColAntiCausal2;
  const int                   m_iSumShift;
  const int                   m_iQVal;
  static const short         sm_aai16Q16SplVal8PelLUT[48][48];
};

template<class t_TSrc, class t_TDst>
const short QuinticSplineFixed<t_TSrc, t_TDst>::sm_aai16Q16SplVal8PelLUT[48][48]=
{
    {        5,     118,     300,     118,       5,       0,       2,      91,     296,     148,       8,       0,       1,      68,     284,     179,      14,       0,       0,      49,     264,     210,      22,       0,       0,      34,     239,     239,      34,       0,       0,      22,     210,     264,      49,       0,       0,      14,     179,     284,      68,       1,       0,       8,     148,     296,      91,       2, },
    {      118,    3077,    7810,    3077,     118,       0,      61,    2378,    7700,    3848,     213,       0,      28,    1774,    7379,    4659,     360,       0,      11,    1273,    6873,    5465,     576,       1,       4,     876,    6220,    6220,     876,       4,       1,     576,    5465,    6873,    1273,      11,       0,     360,    4659,    7379,    1774,      28,       0,     213,    3848,    7700,    2378,      61, },
    {      300,    7810,   19825,    7810,     300,       0,     154,    6037,   19545,    9768,     541,       0,      71,    4502,   18731,   11825,     915,       0,      29,    3232,   17446,   13873,    1463,       2,       9,    2225,   15788,   15788,    2225,       9,       2,    1463,   13873,   17446,    3232,      29,       0,     915,   11825,   18731,    4502,      71,       0,     541,    9768,   19545,    6037,     154, },
    {      118,    3077,    7810,    3077,     118,       0,      61,    2378,    7700,    3848,     213,       0,      28,    1774,    7379,    4659,     360,       0,      11,    1273,    6873,    5465,     576,       1,       4,     876,    6220,    6220,     876,       4,       1,     576,    5465,    6873,    1273,      11,       0,     360,    4659,    7379,    1774,      28,       0,     213,    3848,    7700,    2378,      61, },
    {        5,     118,     300,     118,       5,       0,       2,      91,     296,     148,       8,       0,       1,      68,     284,     179,      14,       0,       0,      49,     264,     210,      22,       0,       0,      34,     239,     239,      34,       0,       0,      22,     210,     264,      49,       0,       0,      14,     179,     284,      68,       1,       0,       8,     148,     296,      91,       2, },
  {        0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0, },
    {        2,      61,     154,      61,       2,       0,       1,      47,     152,      76,       4,       0,       1,      35,     146,      92,       7,       0,       0,      25,     136,     108,      11,       0,       0,      17,     123,     123,      17,       0,       0,      11,     108,     136,      25,       0,       0,       7,      92,     146,      35,       1,       0,       4,      76,     152,      47,       1, },
    {       91,    2378,    6037,    2378,      91,       0,      47,    1838,    5951,    2974,     165,       0,      22,    1371,    5703,    3601,     279,       0,       9,     984,    5312,    4224,     445,       1,       3,     677,    4808,    4808,     677,       3,       1,     445,    4224,    5312,     984,       9,       0,     279,    3601,    5703,    1371,      22,       0,     165,    2974,    5951,    1838,      47, },
    {      296,    7700,   19545,    7700,     296,       0,     152,    5951,   19270,    9630,     534,       0,      70,    4439,   18466,   11659,     902,       0,      28,    3186,   17200,   13678,    1442,       2,       9,    2193,   15566,   15566,    2193,       9,       2,    1442,   13678,   17200,    3186,      28,       0,     902,   11659,   18466,    4439,      70,       0,     534,    9630,   19270,    5951,     152, },
    {      148,    3848,    9768,    3848,     148,       0,      76,    2974,    9630,    4813,     267,       0,      35,    2218,    9229,    5827,     451,       0,      14,    1592,    8596,    6835,     721,       1,       5,    1096,    7779,    7779,    1096,       5,       1,     721,    6835,    8596,    1592,      14,       0,     451,    5827,    9229,    2218,      35,       0,     267,    4813,    9630,    2974,      76, },
    {        8,     213,     541,     213,       8,       0,       4,     165,     534,     267,      15,       0,       2,     123,     511,     323,      25,       0,       1,      88,     476,     379,      40,       0,       0,      61,     431,     431,      61,       0,       0,      40,     379,     476,      88,       1,       0,      25,     323,     511,     123,       2,       0,      15,     267,     534,     165,       4, },
  {        0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0, },
    {        1,      28,      71,      28,       1,       0,       1,      22,      70,      35,       2,       0,       0,      16,      67,      43,       3,       0,       0,      12,      63,      50,       5,       0,       0,       8,      57,      57,       8,       0,       0,       5,      50,      63,      12,       0,       0,       3,      43,      67,      16,       0,       0,       2,      35,      70,      22,       1, },
    {       68,    1774,    4502,    1774,      68,       0,      35,    1371,    4439,    2218,     123,       0,      16,    1023,    4254,    2686,     208,       0,       7,     734,    3962,    3151,     332,       1,       2,     505,    3586,    3586,     505,       2,       1,     332,    3151,    3962,     734,       7,       0,     208,    2686,    4254,    1023,      16,       0,     123,    2218,    4439,    1371,      35, },
    {      284,    7379,   18731,    7379,     284,       0,     146,    5703,   18466,    9229,     511,       0,      67,    4254,   17697,   11173,     864,       0,      27,    3053,   16483,   13108,    1382,       2,       9,    2102,   14917,   14917,    2102,       9,       2,    1382,   13108,   16483,    3053,      27,       0,     864,   11173,   17697,    4254,      67,       0,     511,    9229,   18466,    5703,     146, },
    {      179,    4659,   11825,    4659,     179,       0,      92,    3601,   11659,    5827,     323,       0,      43,    2686,   11173,    7054,     546,       0,      17,    1928,   10407,    8275,     873,       1,       6,    1327,    9418,    9418,    1327,       6,       1,     873,    8275,   10407,    1928,      17,       0,     546,    7054,   11173,    2686,      43,       0,     323,    5827,   11659,    3601,      92, },
    {       14,     360,     915,     360,      14,       0,       7,     279,     902,     451,      25,       0,       3,     208,     864,     546,      42,       0,       1,     149,     805,     640,      68,       0,       0,     103,     729,     729,     103,       0,       0,      68,     640,     805,     149,       1,       0,      42,     546,     864,     208,       3,       0,      25,     451,     902,     279,       7, },
  {        0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0, },
    {        0,      11,      29,      11,       0,       0,       0,       9,      28,      14,       1,       0,       0,       7,      27,      17,       1,       0,       0,       5,      25,      20,       2,       0,       0,       3,      23,      23,       3,       0,       0,       2,      20,      25,       5,       0,       0,       1,      17,      27,       7,       0,       0,       1,      14,      28,       9,       0, },
    {       49,    1273,    3232,    1273,      49,       0,      25,     984,    3186,    1592,      88,       0,      12,     734,    3053,    1928,     149,       0,       5,     527,    2844,    2261,     238,       0,       2,     363,    2574,    2574,     363,       2,       0,     238,    2261,    2844,     527,       5,       0,     149,    1928,    3053,     734,      12,       0,      88,    1592,    3186,     984,      25, },
    {      264,    6873,   17446,    6873,     264,       0,     136,    5312,   17200,    8596,     476,       0,      63,    3962,   16483,   10407,     805,       0,      25,    2844,   15353,   12209,    1287,       2,       8,    1958,   13894,   13894,    1958,       8,       2,    1287,   12209,   15353,    2844,      25,       0,     805,   10407,   16483,    3962,      63,       0,     476,    8596,   17200,    5312,     136, },
    {      210,    5465,   13873,    5465,     210,       0,     108,    4224,   13678,    6835,     379,       0,      50,    3151,   13108,    8275,     640,       0,      20,    2261,   12209,    9708,    1024,       2,       7,    1557,   11049,   11049,    1557,       7,       2,    1024,    9708,   12209,    2261,      20,       0,     640,    8275,   13108,    3151,      50,       0,     379,    6835,   13678,    4224,     108, },
    {       22,     576,    1463,     576,      22,       0,      11,     445,    1442,     721,      40,       0,       5,     332,    1382,     873,      68,       0,       2,     238,    1287,    1024,     108,       0,       1,     164,    1165,    1165,     164,       1,       0,     108,    1024,    1287,     238,       2,       0,      68,     873,    1382,     332,       5,       0,      40,     721,    1442,     445,      11, },
    {        0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0, },
    {        0,       4,       9,       4,       0,       0,       0,       3,       9,       5,       0,       0,       0,       2,       9,       6,       0,       0,       0,       2,       8,       7,       1,       0,       0,       1,       7,       7,       1,       0,       0,       1,       7,       8,       2,       0,       0,       0,       6,       9,       2,       0,       0,       0,       5,       9,       3,       0, },
    {       34,     876,    2225,     876,      34,       0,      17,     677,    2193,    1096,      61,       0,       8,     505,    2102,    1327,     103,       0,       3,     363,    1958,    1557,     164,       0,       1,     250,    1772,    1772,     250,       1,       0,     164,    1557,    1958,     363,       3,       0,     103,    1327,    2102,     505,       8,       0,      61,    1096,    2193,     677,      17, },
    {      239,    6220,   15788,    6220,     239,       0,     123,    4808,   15566,    7779,     431,       0,      57,    3586,   14917,    9418,     729,       0,      23,    2574,   13894,   11049,    1165,       2,       7,    1772,   12574,   12574,    1772,       7,       2,    1165,   11049,   13894,    2574,      23,       0,     729,    9418,   14917,    3586,      57,       0,     431,    7779,   15566,    4808,     123, },
    {      239,    6220,   15788,    6220,     239,       0,     123,    4808,   15566,    7779,     431,       0,      57,    3586,   14917,    9418,     729,       0,      23,    2574,   13894,   11049,    1165,       2,       7,    1772,   12574,   12574,    1772,       7,       2,    1165,   11049,   13894,    2574,      23,       0,     729,    9418,   14917,    3586,      57,       0,     431,    7779,   15566,    4808,     123, },
    {       34,     876,    2225,     876,      34,       0,      17,     677,    2193,    1096,      61,       0,       8,     505,    2102,    1327,     103,       0,       3,     363,    1958,    1557,     164,       0,       1,     250,    1772,    1772,     250,       1,       0,     164,    1557,    1958,     363,       3,       0,     103,    1327,    2102,     505,       8,       0,      61,    1096,    2193,     677,      17, },
    {        0,       4,       9,       4,       0,       0,       0,       3,       9,       5,       0,       0,       0,       2,       9,       6,       0,       0,       0,       2,       8,       7,       1,       0,       0,       1,       7,       7,       1,       0,       0,       1,       7,       8,       2,       0,       0,       0,       6,       9,       2,       0,       0,       0,       5,       9,       3,       0, },
    {        0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0, },
    {       22,     576,    1463,     576,      22,       0,      11,     445,    1442,     721,      40,       0,       5,     332,    1382,     873,      68,       0,       2,     238,    1287,    1024,     108,       0,       1,     164,    1165,    1165,     164,       1,       0,     108,    1024,    1287,     238,       2,       0,      68,     873,    1382,     332,       5,       0,      40,     721,    1442,     445,      11, },
    {      210,    5465,   13873,    5465,     210,       0,     108,    4224,   13678,    6835,     379,       0,      50,    3151,   13108,    8275,     640,       0,      20,    2261,   12209,    9708,    1024,       2,       7,    1557,   11049,   11049,    1557,       7,       2,    1024,    9708,   12209,    2261,      20,       0,     640,    8275,   13108,    3151,      50,       0,     379,    6835,   13678,    4224,     108, },
    {      264,    6873,   17446,    6873,     264,       0,     136,    5312,   17200,    8596,     476,       0,      63,    3962,   16483,   10407,     805,       0,      25,    2844,   15353,   12209,    1287,       2,       8,    1958,   13894,   13894,    1958,       8,       2,    1287,   12209,   15353,    2844,      25,       0,     805,   10407,   16483,    3962,      63,       0,     476,    8596,   17200,    5312,     136, },
    {       49,    1273,    3232,    1273,      49,       0,      25,     984,    3186,    1592,      88,       0,      12,     734,    3053,    1928,     149,       0,       5,     527,    2844,    2261,     238,       0,       2,     363,    2574,    2574,     363,       2,       0,     238,    2261,    2844,     527,       5,       0,     149,    1928,    3053,     734,      12,       0,      88,    1592,    3186,     984,      25, },
    {        0,      11,      29,      11,       0,       0,       0,       9,      28,      14,       1,       0,       0,       7,      27,      17,       1,       0,       0,       5,      25,      20,       2,       0,       0,       3,      23,      23,       3,       0,       0,       2,      20,      25,       5,       0,       0,       1,      17,      27,       7,       0,       0,       1,      14,      28,       9,       0, },
  {        0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0, },
    {       14,     360,     915,     360,      14,       0,       7,     279,     902,     451,      25,       0,       3,     208,     864,     546,      42,       0,       1,     149,     805,     640,      68,       0,       0,     103,     729,     729,     103,       0,       0,      68,     640,     805,     149,       1,       0,      42,     546,     864,     208,       3,       0,      25,     451,     902,     279,       7, },
    {      179,    4659,   11825,    4659,     179,       0,      92,    3601,   11659,    5827,     323,       0,      43,    2686,   11173,    7054,     546,       0,      17,    1928,   10407,    8275,     873,       1,       6,    1327,    9418,    9418,    1327,       6,       1,     873,    8275,   10407,    1928,      17,       0,     546,    7054,   11173,    2686,      43,       0,     323,    5827,   11659,    3601,      92, },
    {      284,    7379,   18731,    7379,     284,       0,     146,    5703,   18466,    9229,     511,       0,      67,    4254,   17697,   11173,     864,       0,      27,    3053,   16483,   13108,    1382,       2,       9,    2102,   14917,   14917,    2102,       9,       2,    1382,   13108,   16483,    3053,      27,       0,     864,   11173,   17697,    4254,      67,       0,     511,    9229,   18466,    5703,     146, },
    {       68,    1774,    4502,    1774,      68,       0,      35,    1371,    4439,    2218,     123,       0,      16,    1023,    4254,    2686,     208,       0,       7,     734,    3962,    3151,     332,       1,       2,     505,    3586,    3586,     505,       2,       1,     332,    3151,    3962,     734,       7,       0,     208,    2686,    4254,    1023,      16,       0,     123,    2218,    4439,    1371,      35, },
    {        1,      28,      71,      28,       1,       0,       1,      22,      70,      35,       2,       0,       0,      16,      67,      43,       3,       0,       0,      12,      63,      50,       5,       0,       0,       8,      57,      57,       8,       0,       0,       5,      50,      63,      12,       0,       0,       3,      43,      67,      16,       0,       0,       2,      35,      70,      22,       1, },
  {        0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0, },
    {        8,     213,     541,     213,       8,       0,       4,     165,     534,     267,      15,       0,       2,     123,     511,     323,      25,       0,       1,      88,     476,     379,      40,       0,       0,      61,     431,     431,      61,       0,       0,      40,     379,     476,      88,       1,       0,      25,     323,     511,     123,       2,       0,      15,     267,     534,     165,       4, },
    {      148,    3848,    9768,    3848,     148,       0,      76,    2974,    9630,    4813,     267,       0,      35,    2218,    9229,    5827,     451,       0,      14,    1592,    8596,    6835,     721,       1,       5,    1096,    7779,    7779,    1096,       5,       1,     721,    6835,    8596,    1592,      14,       0,     451,    5827,    9229,    2218,      35,       0,     267,    4813,    9630,    2974,      76, },
    {      296,    7700,   19545,    7700,     296,       0,     152,    5951,   19270,    9630,     534,       0,      70,    4439,   18466,   11659,     902,       0,      28,    3186,   17200,   13678,    1442,       2,       9,    2193,   15566,   15566,    2193,       9,       2,    1442,   13678,   17200,    3186,      28,       0,     902,   11659,   18466,    4439,      70,       0,     534,    9630,   19270,    5951,     152, },
    {       91,    2378,    6037,    2378,      91,       0,      47,    1838,    5951,    2974,     165,       0,      22,    1371,    5703,    3601,     279,       0,       9,     984,    5312,    4224,     445,       1,       3,     677,    4808,    4808,     677,       3,       1,     445,    4224,    5312,     984,       9,       0,     279,    3601,    5703,    1371,      22,       0,     165,    2974,    5951,    1838,      47, },
    {        2,      61,     154,      61,       2,       0,       1,      47,     152,      76,       4,       0,       1,      35,     146,      92,       7,       0,       0,      25,     136,     108,      11,       0,       0,      17,     123,     123,      17,       0,       0,      11,     108,     136,      25,       0,       0,       7,      92,     146,      35,       1,       0,       4,      76,     152,      47,       1, },
};

template<class t_TSrc, class t_TDst>
void QuinticSplineFixed<t_TSrc, t_TDst>::interpolate(UInt uiPl, const Block<t_TDst> &rcDst, int iFac)
{
  TOF( (iFac == 2) || (iFac == 4) || (iFac == 8) );
  switch( iFac )
  {
  case 2:
    xInterpolate<2>( uiPl, rcDst);
    break;
  case 4:
    xInterpolate<4>( uiPl, rcDst);
    break;
  case 5:
    xInterpolate<8>( uiPl, rcDst);
    break;
  default:
    TOT(1);
  }
}

// interpolation of block using spline coeffs
template<class t_TSrc, class t_TDst>
template<int iFac>
void QuinticSplineFixed<t_TSrc, t_TDst>::xInterpolate(UInt uiPl, const Block<t_TDst> &rcDst)
{
//  TOF( (iFac == 2) || (iFac == 4) || (iFac == 8) );

  const int iFacShift = gCeilLog2( iFac );
  const int iMask = (1<<iFacShift)-1;

  const Block<short> &cSplCoeffs = *m_pi16Q6SplCoeff[uiPl];
  const int iDstYMax=rcDst.getSize().getY();
  const int iDstXMax=rcDst.getSize().getX();

  const short* pSpl = xGetSplineBuf();
  const int iSplBufStride = xGetSplineBufStride();
  const int iSrcBufStride = cSplCoeffs.getStride();

  const short* pSrc = &cSplCoeffs.getEl(m_iSplPadding-2,m_iSplPadding-2);
  t_TDst* pDes = &rcDst.getEl(0,0);
  const int iDesStride = rcDst.getStride();


  const int iFinalShift = (m_iQColAntiCausal2+m_iQVal+m_iQSplGain-m_iSumShift)-EXTRA_PRECISION_BITS;
  const int iFinalAdd   = (1<<(iFinalShift-1));
  const int iSumShift   = m_iSumShift;
  const int iSumAdd     = (1<<(m_iSumShift-1));
  const int iSplGain    = m_iSplGain;
  const int iMulFac     = (48>>iFacShift);

  // Perform Interpolation
  for( int iY=0; iY<iDstYMax; iY++)
  {
    const int iSrcY=iY>>iFacShift;
    const int iDy= (iY&iMask)*iMulFac;

    const short* pSrcLn = pSrc + iSrcY * iSrcBufStride;
    const short* pSplLn = pSpl + iDy * iSplBufStride;

    for( int iX=0; iX<iDstXMax; iX++)
    {
      const int iSrcX=iX>>iFacShift;
      const int iDx= (iX&iMask)*iMulFac;

      const short* pSrcTmp = pSrcLn + iSrcX;
      const short* pSplTmp = pSplLn + iDx;
      // 2d convolution
      int iSum = 0;
      for(int iYY=0; iYY<6; iYY++)
      {
        for(int iXX=0; iXX<6; iXX++)
        {
          iSum += int( pSrcTmp[iXX]) * int(pSplTmp[iXX]);
        }
        pSplTmp += iSplBufStride;
        pSrcTmp += iSrcBufStride;
      }

      pDes[iX]=t_TDst((((iSum+iSumAdd)>>iSumShift) * iSplGain + iFinalAdd )>>iFinalShift);
    }
    pDes += iDesStride;
  }
}


// internal function for spline initialization
template<class t_TSrc, class t_TDst>
template<class t_T>
short QuinticSplineFixed<t_TSrc, t_TDst>::xInitRowCausalCoeff(const Block<const t_T> &rcSrc, const int iSrcShift, const int iY, const int N, const short i16Q15Z)
{
  // only sum up 16 value
  const int iXMax  = (N > 16) ? 16 : N;
  const t_T* pcSrc = &rcSrc.getEl(0,iY);
  short i16Q6Sum   = short(pcSrc[0])<<iSrcShift;

  short i16Q15Z_n=i16Q15Z;
  for( int iX = 1; iX < iXMax; iX++)
  {
    i16Q6Sum += xInt16Multiplication( i16Q15Z_n, short(pcSrc[iX])<<iSrcShift );
    i16Q15Z_n = xInt16Multiplication( i16Q15Z_n, i16Q15Z );
  }
  return i16Q6Sum;
}

// internal function for spline initialization
template<class t_TSrc, class t_TDst>
short QuinticSplineFixed<t_TSrc, t_TDst>::xInitColCausalCoeff(const Block<const short> &ri16Q6Src, const int iX, const int N, const short i16Q15Z)
{
  // only sum up 16 value
  const int iYMax     = (N > 16) ? 16 : N;
  const short* pcSrc  = &ri16Q6Src.getEl(iX,0);
  const int iStride   = ri16Q6Src.getStride();
  short i16Q6Sum      = *pcSrc;

  short i16Q15Z_n = i16Q15Z;
  for( int iY = 1; iY < iYMax; iY++)
  {
    pcSrc += iStride;
    i16Q6Sum += xInt16Multiplication( *pcSrc, i16Q15Z_n );
    i16Q15Z_n = xInt16Multiplication( i16Q15Z_n, i16Q15Z );
  }
  return i16Q6Sum;
}

template<class t_TSrc, class t_TDst>
template<class t_T>
void QuinticSplineFixed<t_TSrc, t_TDst>::xSplineCoeffs( const Block<const t_T> &rcSrc, const int iSrcShift, Block<short> &raai16Q6Spl, short i16Q15Z, const short i16Q15Z_c, const int iQRowCausal, const int iQRowAntiCausal, const int iQColCausal, const int iQColAntiCausal )
{
  int iY, iX;
  const int iXMax=rcSrc.getSize().getX();
  const int iYMax=rcSrc.getSize().getY();
  short* pSpl = &raai16Q6Spl.getEl(0,0);
  const int iStride = raai16Q6Spl.getStride();

  //limit the scope of variables
  {
    //Q-change;
    const int iQDiffRowCausal = iQRowCausal - 6;
    TOT( iQDiffRowCausal < 0 );
    const int iQDiffRowAntiCausal = iQRowAntiCausal - iQRowCausal;
    TOT( iQDiffRowAntiCausal < 0 );

    const t_T* pSrc         = &rcSrc.getEl(0,0);
    const int  iSrcStride = rcSrc.getStride();
    const int iSrcShiftSh   = iSrcShift + iQDiffRowCausal;
    const int i16Q15Z_c_sh  = i16Q15Z_c<<iQDiffRowAntiCausal;
    const int i16Q15Z_sh    = i16Q15Z<<iQDiffRowAntiCausal;

    // filter along rows
    for( iY=0; iY<iYMax; iY++)
    {
      // init causal coeff
      pSpl[0] = xInitRowCausalCoeff<t_T>(rcSrc, iSrcShift, iY, iXMax, i16Q15Z ) << iQDiffRowCausal;

      // causal filtering
      for(iX=1; iX<iXMax; iX++)
      {
        pSpl[iX] = xInt16Multiplication( pSpl[iX-1], i16Q15Z ) + ((short)pSrc[iX] << iSrcShiftSh);
      }

      // init anti-causal coeff
      pSpl[iXMax-1] = xInt16Multiplication( xInt16Multiplication( pSpl[iXMax-2], i16Q15Z ) + pSpl[iXMax-1], i16Q15Z_c_sh );

      // anti-causal filtering
      for (iX=iXMax-2; iX>=0; iX--)
      {
        pSpl[iX] = xInt16Multiplication( (pSpl[iX+1]>>iQDiffRowAntiCausal) - pSpl[iX], i16Q15Z_sh );
      }
      pSpl += iStride;
      pSrc += iSrcStride;
    }
  }

  {
    //Q-change;
    const int iQDiffColCausal = iQColCausal - iQRowAntiCausal;
    TOT( iQDiffColCausal < 0 );
    const int iQDiffColAntiCausal = iQColAntiCausal - iQColCausal;
    TOT( iQDiffColAntiCausal < 0 );

    pSpl = &raai16Q6Spl.getEl(0,0);

    // filter along columns
    for (iX=0; iX<iXMax; iX++)
    {
      // init causal coeff
      pSpl[iX] =( xInitColCausalCoeff( raai16Q6Spl, iX, iYMax, i16Q15Z ) << iQDiffColCausal );
    }

    // causal filtering
    for (iY=1; iY<iYMax; iY++)
    {
      pSpl += iStride;
      for (iX=0; iX<iXMax; iX++)
      {
        pSpl[iX] = ( pSpl[iX] << iQDiffColCausal ) + xInt16Multiplication( pSpl[iX-iStride], i16Q15Z );
      }
    }
    const int i16Q15Z_c_sh = i16Q15Z_c<<iQDiffColAntiCausal;
    // init anti-causal coeff
    for (iX=0; iX<iXMax; iX++)
    {
      pSpl[iX] = xInt16Multiplication( xInt16Multiplication( pSpl[iX-iStride], i16Q15Z ) + pSpl[iX], i16Q15Z_c_sh );
    }

    const int i16Q15Z_sh = i16Q15Z<<iQDiffColAntiCausal;
    // anti-causal filtering
    for (iY=iYMax-2; iY>=0; iY--)
    {
      pSpl -= iStride;
      for (iX=0; iX<iXMax; iX++)
      {
        pSpl[iX] = xInt16Multiplication( (pSpl[iX+iStride]>>iQDiffColAntiCausal)-pSpl[iX], i16Q15Z_sh );
      }
    }
  }
}



// Convert image samples to spline coefficients
template<class t_TSrc, class t_TDst>
void QuinticSplineFixed<t_TSrc, t_TDst>::calcCoeffs(const Block<const t_TSrc> &rcSrc, UInt uiPl)
{
  const int iXMax=rcSrc.getSize().getX();
  const int iYMax=rcSrc.getSize().getY();

  // filter for pole z1
  AllocBlock<short> cTemp( rcSrc.getSize() );
  xSplineCoeffs<t_TSrc>( rcSrc, 6, cTemp, m_i16Q15Z1, m_i16Q15Z1_c, 6, 6, 6, 6 );

  // filter for pole z2
  const Block<short> &cSplCoeffs = *m_pi16Q6SplCoeff[uiPl];
  Block<short> cSpl(PixelPos(m_iSplPadding,m_iSplPadding), BlockSize(iXMax, iYMax), cSplCoeffs);;
  xSplineCoeffs<short>( cTemp, 0, cSpl, m_i16Q15Z2, m_i16Q15Z2_c, m_iQRowCausal2, m_iQRowAntiCausal2, m_iQColCausal2, m_iQColAntiCausal2 );

  // extend using mirror symmetry
  {
    int iX, iY;
    short* pSpl           = &cSpl.getEl(0,0);
    const int iStride     = cSpl.getStride();
    for(iY=0; iY<iYMax; iY++)
    {
      //left
      for( iX=1; iX<=m_iSplPadding; iX++ )
      {
        pSpl[-iX] = pSpl[ iX ];
      }

      // right
      pSpl += iXMax-1;
      for( iX=1; iX<=m_iSplPadding; iX++ )
      {
        pSpl[iX] = pSpl[ -iX ];
      }
      pSpl += iStride -(iXMax-1);
    }

    //bottom
    pSpl -= m_iSplPadding + iStride;
    int iOffset = iStride;
    for(iY=0; iY<m_iSplPadding; iY++)
    {
      for(iX=0; iX<iStride; iX++)
      {
        pSpl[iX + iOffset] = pSpl[ iX - iOffset];
      }
      iOffset += iStride;
    }

    //top
    pSpl = &cSpl.getEl(0,0)-m_iSplPadding;
    iOffset = iStride;
    for(iY=0; iY<m_iSplPadding; iY++)
    {
      for(iX=0; iX<iStride; iX++)
      {
        pSpl[iX - iOffset] = pSpl[ iX + iOffset];
      }
      iOffset += iStride;
    }
  }
}

template<class t_TSrc, class t_TDst>
class QuinticSpline
  : public InterpolationIf<t_TSrc, t_TDst>
{
public:
  QuinticSpline() : m_iSplPadding(3), m_dSplGain(120.0*120.0), m_dZ1(-0.430575347099978), m_dZ2(-0.043096288203263) {}
  QuinticSpline( const double dSplGain, const double dZ1, const double dZ2 ) : m_iSplPadding(3), m_dSplGain(dSplGain), m_dZ1(dZ1), m_dZ2(dZ2) {}

  virtual ~QuinticSpline()
  {
    std::vector<Block<double>*>::iterator it;
    for (it=m_pcSplCoeff.begin(); it!=m_pcSplCoeff.end(); ++it)
    {
      delete (*it);
    }
    m_pcSplCoeff.clear();
  }

  // Allocate memory for spline_coeffs + 2*padding
  ErrVal allocCoeffs(const PrimBlockSize& rcPbSize, const std::vector<Subsampling> &racOrigSub)
  {
    for( UInt ui = 0; ui < racOrigSub.size(); ui++ )
    {
      BlockSize cSize = rcPbSize * racOrigSub[ui];
      m_pcSplCoeff.push_back( new AllocBlock<double>( BlockSize(cSize.getX()+2*m_iSplPadding, cSize.getY()+2*m_iSplPadding) ) );
      ROF( m_pcSplCoeff[ui] );
    }
    return Err::m_nOK;
  }

  // function to calculate spline coeffs
  void calcCoeffs(const Block<const t_TSrc> &rcSrc, UInt uiPl);

  // function to interpolate using spline coeffs
  void interpolate(UInt uiPl, const Block<t_TDst> &rcDst, int iFac);

private:

  template<class t_T>
  double xInitRowCausalCoeff(const Block<const t_T> &rcSrc, const int iY, const int N, const double z1 );
  double xInitColCausalCoeff(const Block<const double> &rcSrc, const int iX, const int N, const double z1 );
  template<class t_T>
  void xSplineCoeffs( const Block<const t_T> &rcSrc, Block<double> &rcDst, const double dZ );
  virtual double xSplineVal ( int iX, int iY) const;

  const int                   m_iSplPadding;
  const double                m_dSplGain;
  std::vector<Block<double>*> m_pcSplCoeff;
  const double                m_dZ1;
  const double                m_dZ2;
  static const double         sm_aadSplValQPelLUT[24][24];
};

template<class t_TSrc, class t_TDst>
const double QuinticSpline<t_TSrc, t_TDst>::sm_aadSplValQPelLUT[24][24]=
{
    {     6.944444444444444e-05,    1.805555555555556e-03,    4.583333333333333e-03,    1.805555555555556e-03,    6.944444444444444e-05,    0.000000000000000e+00,    1.647949218750000e-05,    1.040920681423611e-03,    4.330376519097222e-03,    2.733968098958333e-03,    2.115207248263889e-04,    6.781684027777778e-08,    2.170138888888889e-06,    5.143229166666667e-04,    3.650173611111111e-03,    3.650173611111111e-03,    5.143229166666667e-04,    2.170138888888889e-06,    6.781684027777778e-08,    2.115207248263889e-04,    2.733968098958333e-03,    4.330376519097222e-03,    1.040920681423611e-03,    1.647949218750000e-05, },
    {     1.805555555555556e-03,    4.694444444444445e-02,    1.191666666666667e-01,    4.694444444444445e-02,    1.805555555555556e-03,    0.000000000000000e+00,    4.284667968750000e-04,    2.706393771701389e-02,    1.125897894965278e-01,    7.108317057291667e-02,    5.499538845486112e-03,    1.763237847222222e-06,    5.642361111111112e-05,    1.337239583333334e-02,    9.490451388888890e-02,    9.490451388888890e-02,    1.337239583333334e-02,    5.642361111111112e-05,    1.763237847222222e-06,    5.499538845486112e-03,    7.108317057291667e-02,    1.125897894965278e-01,    2.706393771701389e-02,    4.284667968750000e-04, },
    {     4.583333333333333e-03,    1.191666666666667e-01,    3.025000000000000e-01,    1.191666666666667e-01,    4.583333333333333e-03,    0.000000000000000e+00,    1.087646484375000e-03,    6.870076497395834e-02,    2.858048502604167e-01,    1.804418945312500e-01,    1.396036783854167e-02,    4.475911458333333e-06,    1.432291666666667e-04,    3.394531250000000e-02,    2.409114583333334e-01,    2.409114583333334e-01,    3.394531250000000e-02,    1.432291666666667e-04,    4.475911458333333e-06,    1.396036783854167e-02,    1.804418945312500e-01,    2.858048502604167e-01,    6.870076497395834e-02,    1.087646484375000e-03, },
    {     1.805555555555556e-03,    4.694444444444445e-02,    1.191666666666667e-01,    4.694444444444445e-02,    1.805555555555556e-03,    0.000000000000000e+00,    4.284667968750000e-04,    2.706393771701389e-02,    1.125897894965278e-01,    7.108317057291667e-02,    5.499538845486112e-03,    1.763237847222222e-06,    5.642361111111112e-05,    1.337239583333334e-02,    9.490451388888890e-02,    9.490451388888890e-02,    1.337239583333334e-02,    5.642361111111112e-05,    1.763237847222222e-06,    5.499538845486112e-03,    7.108317057291667e-02,    1.125897894965278e-01,    2.706393771701389e-02,    4.284667968750000e-04, },
    {     6.944444444444444e-05,    1.805555555555556e-03,    4.583333333333333e-03,    1.805555555555556e-03,    6.944444444444444e-05,    0.000000000000000e+00,    1.647949218750000e-05,    1.040920681423611e-03,    4.330376519097222e-03,    2.733968098958333e-03,    2.115207248263889e-04,    6.781684027777778e-08,    2.170138888888889e-06,    5.143229166666667e-04,    3.650173611111111e-03,    3.650173611111111e-03,    5.143229166666667e-04,    2.170138888888889e-06,    6.781684027777778e-08,    2.115207248263889e-04,    2.733968098958333e-03,    4.330376519097222e-03,    1.040920681423611e-03,    1.647949218750000e-05, },
    {     0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00, },
    {     1.647949218750000e-05,    4.284667968750000e-04,    1.087646484375000e-03,    4.284667968750000e-04,    1.647949218750000e-05,    0.000000000000000e+00,    3.910660743713379e-06,    2.470153570175171e-04,    1.027618646621704e-03,    6.487834453582763e-04,    5.019485950469971e-05,    1.609325408935547e-08,    5.149841308593749e-07,    1.220512390136719e-04,    8.662033081054687e-04,    8.662033081054687e-04,    1.220512390136719e-04,    5.149841308593749e-07,    1.609325408935547e-08,    5.019485950469971e-05,    6.487834453582763e-04,    1.027618646621704e-03,    2.470153570175171e-04,    3.910660743713379e-06, },
    {     1.040920681423611e-03,    2.706393771701389e-02,    6.870076497395834e-02,    2.706393771701389e-02,    1.040920681423611e-03,    0.000000000000000e+00,    2.470153570175171e-04,    1.560262845622169e-02,    6.490913006994459e-02,    4.098015268643697e-02,    3.170538677109612e-03,    1.016524102952745e-06,    3.252877129448784e-05,    7.709318796793620e-03,    5.471339331732856e-02,    5.471339331732856e-02,    7.709318796793620e-03,    3.252877129448784e-05,    1.016524102952745e-06,    3.170538677109612e-03,    4.098015268643697e-02,    6.490913006994459e-02,    1.560262845622169e-02,    2.470153570175171e-04, },
    {     4.330376519097222e-03,    1.125897894965278e-01,    2.858048502604167e-01,    1.125897894965278e-01,    4.330376519097222e-03,    0.000000000000000e+00,    1.027618646621704e-03,    6.490913006994459e-02,    2.700311154789395e-01,    1.704832021395365e-01,    1.318988707330492e-02,    4.228883319430881e-06,    1.353242662217882e-04,    3.207185109456381e-02,    2.276154157850478e-01,    2.276154157850478e-01,    3.207185109456381e-02,    1.353242662217882e-04,    4.228883319430881e-06,    1.318988707330492e-02,    1.704832021395365e-01,    2.700311154789395e-01,    6.490913006994459e-02,    1.027618646621704e-03, },
    {     2.733968098958333e-03,    7.108317057291667e-02,    1.804418945312500e-01,    7.108317057291667e-02,    2.733968098958333e-03,    0.000000000000000e+00,    6.487834453582763e-04,    4.098015268643697e-02,    1.704832021395365e-01,    1.076339745521545e-01,    8.327389160792033e-03,    2.669890721638997e-06,    8.543650309244791e-05,    2.024845123291016e-02,    1.437041982014974e-01,    1.437041982014974e-01,    2.024845123291016e-02,    8.543650309244791e-05,    2.669890721638997e-06,    8.327389160792033e-03,    1.076339745521545e-01,    1.704832021395365e-01,    4.098015268643697e-02,    6.487834453582763e-04, },
    {     2.115207248263889e-04,    5.499538845486112e-03,    1.396036783854167e-02,    5.499538845486112e-03,    2.115207248263889e-04,    0.000000000000000e+00,    5.019485950469971e-05,    3.170538677109612e-03,    1.318988707330492e-02,    8.327389160792033e-03,    6.442706452475654e-04,    2.065632078382704e-07,    6.610022650824653e-06,    1.566575368245443e-03,    1.111805809868707e-02,    1.111805809868707e-02,    1.566575368245443e-03,    6.610022650824653e-06,    2.065632078382704e-07,    6.442706452475654e-04,    8.327389160792033e-03,    1.318988707330492e-02,    3.170538677109612e-03,    5.019485950469971e-05, },
    {     6.781684027777778e-08,    1.763237847222222e-06,    4.475911458333333e-06,    1.763237847222222e-06,    6.781684027777778e-08,    0.000000000000000e+00,    1.609325408935547e-08,    1.016524102952745e-06,    4.228883319430881e-06,    2.669890721638997e-06,    2.065632078382704e-07,    6.622738308376736e-11,    2.119276258680556e-09,    5.022684733072917e-07,    3.564622667100695e-06,    3.564622667100695e-06,    5.022684733072917e-07,    2.119276258680556e-09,    6.622738308376736e-11,    2.065632078382704e-07,    2.669890721638997e-06,    4.228883319430881e-06,    1.016524102952745e-06,    1.609325408935547e-08, },
    {     2.170138888888889e-06,    5.642361111111112e-05,    1.432291666666667e-04,    5.642361111111112e-05,    2.170138888888889e-06,    0.000000000000000e+00,    5.149841308593749e-07,    3.252877129448784e-05,    1.353242662217882e-04,    8.543650309244791e-05,    6.610022650824653e-06,    2.119276258680556e-09,    6.781684027777778e-08,    1.607259114583334e-05,    1.140679253472222e-04,    1.140679253472222e-04,    1.607259114583334e-05,    6.781684027777778e-08,    2.119276258680556e-09,    6.610022650824653e-06,    8.543650309244791e-05,    1.353242662217882e-04,    3.252877129448784e-05,    5.149841308593749e-07, },
    {     5.143229166666667e-04,    1.337239583333334e-02,    3.394531250000000e-02,    1.337239583333334e-02,    5.143229166666667e-04,    0.000000000000000e+00,    1.220512390136719e-04,    7.709318796793620e-03,    3.207185109456381e-02,    2.024845123291016e-02,    1.566575368245443e-03,    5.022684733072917e-07,    1.607259114583334e-05,    3.809204101562500e-03,    2.703409830729167e-02,    2.703409830729167e-02,    3.809204101562500e-03,    1.607259114583334e-05,    5.022684733072917e-07,    1.566575368245443e-03,    2.024845123291016e-02,    3.207185109456381e-02,    7.709318796793620e-03,    1.220512390136719e-04, },
    {     3.650173611111111e-03,    9.490451388888890e-02,    2.409114583333334e-01,    9.490451388888890e-02,    3.650173611111111e-03,    0.000000000000000e+00,    8.662033081054687e-04,    5.471339331732856e-02,    2.276154157850478e-01,    1.437041982014974e-01,    1.111805809868707e-02,    3.564622667100695e-06,    1.140679253472222e-04,    2.703409830729167e-02,    1.918622504340278e-01,    1.918622504340278e-01,    2.703409830729167e-02,    1.140679253472222e-04,    3.564622667100695e-06,    1.111805809868707e-02,    1.437041982014974e-01,    2.276154157850478e-01,    5.471339331732856e-02,    8.662033081054687e-04, },
    {     3.650173611111111e-03,    9.490451388888890e-02,    2.409114583333334e-01,    9.490451388888890e-02,    3.650173611111111e-03,    0.000000000000000e+00,    8.662033081054687e-04,    5.471339331732856e-02,    2.276154157850478e-01,    1.437041982014974e-01,    1.111805809868707e-02,    3.564622667100695e-06,    1.140679253472222e-04,    2.703409830729167e-02,    1.918622504340278e-01,    1.918622504340278e-01,    2.703409830729167e-02,    1.140679253472222e-04,    3.564622667100695e-06,    1.111805809868707e-02,    1.437041982014974e-01,    2.276154157850478e-01,    5.471339331732856e-02,    8.662033081054687e-04, },
    {     5.143229166666667e-04,    1.337239583333334e-02,    3.394531250000000e-02,    1.337239583333334e-02,    5.143229166666667e-04,    0.000000000000000e+00,    1.220512390136719e-04,    7.709318796793620e-03,    3.207185109456381e-02,    2.024845123291016e-02,    1.566575368245443e-03,    5.022684733072917e-07,    1.607259114583334e-05,    3.809204101562500e-03,    2.703409830729167e-02,    2.703409830729167e-02,    3.809204101562500e-03,    1.607259114583334e-05,    5.022684733072917e-07,    1.566575368245443e-03,    2.024845123291016e-02,    3.207185109456381e-02,    7.709318796793620e-03,    1.220512390136719e-04, },
    {     2.170138888888889e-06,    5.642361111111112e-05,    1.432291666666667e-04,    5.642361111111112e-05,    2.170138888888889e-06,    0.000000000000000e+00,    5.149841308593749e-07,    3.252877129448784e-05,    1.353242662217882e-04,    8.543650309244791e-05,    6.610022650824653e-06,    2.119276258680556e-09,    6.781684027777778e-08,    1.607259114583334e-05,    1.140679253472222e-04,    1.140679253472222e-04,    1.607259114583334e-05,    6.781684027777778e-08,    2.119276258680556e-09,    6.610022650824653e-06,    8.543650309244791e-05,    1.353242662217882e-04,    3.252877129448784e-05,    5.149841308593749e-07, },
    {     6.781684027777778e-08,    1.763237847222222e-06,    4.475911458333333e-06,    1.763237847222222e-06,    6.781684027777778e-08,    0.000000000000000e+00,    1.609325408935547e-08,    1.016524102952745e-06,    4.228883319430881e-06,    2.669890721638997e-06,    2.065632078382704e-07,    6.622738308376736e-11,    2.119276258680556e-09,    5.022684733072917e-07,    3.564622667100695e-06,    3.564622667100695e-06,    5.022684733072917e-07,    2.119276258680556e-09,    6.622738308376736e-11,    2.065632078382704e-07,    2.669890721638997e-06,    4.228883319430881e-06,    1.016524102952745e-06,    1.609325408935547e-08, },
    {     2.115207248263889e-04,    5.499538845486112e-03,    1.396036783854167e-02,    5.499538845486112e-03,    2.115207248263889e-04,    0.000000000000000e+00,    5.019485950469971e-05,    3.170538677109612e-03,    1.318988707330492e-02,    8.327389160792033e-03,    6.442706452475654e-04,    2.065632078382704e-07,    6.610022650824653e-06,    1.566575368245443e-03,    1.111805809868707e-02,    1.111805809868707e-02,    1.566575368245443e-03,    6.610022650824653e-06,    2.065632078382704e-07,    6.442706452475654e-04,    8.327389160792033e-03,    1.318988707330492e-02,    3.170538677109612e-03,    5.019485950469971e-05, },
    {     2.733968098958333e-03,    7.108317057291667e-02,    1.804418945312500e-01,    7.108317057291667e-02,    2.733968098958333e-03,    0.000000000000000e+00,    6.487834453582763e-04,    4.098015268643697e-02,    1.704832021395365e-01,    1.076339745521545e-01,    8.327389160792033e-03,    2.669890721638997e-06,    8.543650309244791e-05,    2.024845123291016e-02,    1.437041982014974e-01,    1.437041982014974e-01,    2.024845123291016e-02,    8.543650309244791e-05,    2.669890721638997e-06,    8.327389160792033e-03,    1.076339745521545e-01,    1.704832021395365e-01,    4.098015268643697e-02,    6.487834453582763e-04, },
    {     4.330376519097222e-03,    1.125897894965278e-01,    2.858048502604167e-01,    1.125897894965278e-01,    4.330376519097222e-03,    0.000000000000000e+00,    1.027618646621704e-03,    6.490913006994459e-02,    2.700311154789395e-01,    1.704832021395365e-01,    1.318988707330492e-02,    4.228883319430881e-06,    1.353242662217882e-04,    3.207185109456381e-02,    2.276154157850478e-01,    2.276154157850478e-01,    3.207185109456381e-02,    1.353242662217882e-04,    4.228883319430881e-06,    1.318988707330492e-02,    1.704832021395365e-01,    2.700311154789395e-01,    6.490913006994459e-02,    1.027618646621704e-03, },
    {     1.040920681423611e-03,    2.706393771701389e-02,    6.870076497395834e-02,    2.706393771701389e-02,    1.040920681423611e-03,    0.000000000000000e+00,    2.470153570175171e-04,    1.560262845622169e-02,    6.490913006994459e-02,    4.098015268643697e-02,    3.170538677109612e-03,    1.016524102952745e-06,    3.252877129448784e-05,    7.709318796793620e-03,    5.471339331732856e-02,    5.471339331732856e-02,    7.709318796793620e-03,    3.252877129448784e-05,    1.016524102952745e-06,    3.170538677109612e-03,    4.098015268643697e-02,    6.490913006994459e-02,    1.560262845622169e-02,    2.470153570175171e-04, },
    {     1.647949218750000e-05,    4.284667968750000e-04,    1.087646484375000e-03,    4.284667968750000e-04,    1.647949218750000e-05,    0.000000000000000e+00,    3.910660743713379e-06,    2.470153570175171e-04,    1.027618646621704e-03,    6.487834453582763e-04,    5.019485950469971e-05,    1.609325408935547e-08,    5.149841308593749e-07,    1.220512390136719e-04,    8.662033081054687e-04,    8.662033081054687e-04,    1.220512390136719e-04,    5.149841308593749e-07,    1.609325408935547e-08,    5.019485950469971e-05,    6.487834453582763e-04,    1.027618646621704e-03,    2.470153570175171e-04,    3.910660743713379e-06, },
};

// return 6x6 spline values for a given position (dx,dy)
template<class t_TSrc, class t_TDst>
double QuinticSpline<t_TSrc, t_TDst>::xSplineVal ( int iX, int iY) const
{
  return sm_aadSplValQPelLUT[iX][iY];
}

// interpolation of block using spline coeffs
template<class t_TSrc, class t_TDst>
void QuinticSpline<t_TSrc, t_TDst>::interpolate(UInt uiPl, const Block<t_TDst> &rcDst, int iFac)
{
  AOF( iFac == 4 );

  int iX, iY;
  const Block<double> &cSplCoeffs = *m_pcSplCoeff[uiPl];
  int iDstYMax=rcDst.getSize().getY(), iDstXMax=rcDst.getSize().getX();

  // Perform Interpolation
  for(iY=0; iY<iDstYMax; iY++)
  {
    int iSrcY=int(iY/iFac);
    int iDy= iY % iFac;

    for(iX=0; iX<iDstXMax; iX++)
    {
      int iXX, iYY;
      double dSum=0;
      int iSrcX=int(iX/iFac);
      int iDx= iX % iFac;

      // 2d convolution
      for(iYY=0; iYY<6; iYY++)
      {
        for(iXX=0; iXX<6; iXX++)
        {
          dSum = dSum + m_dSplGain*cSplCoeffs.getEl(m_iSplPadding+iSrcX-2+iXX,m_iSplPadding+iSrcY-2+iYY) * xSplineVal(iDx*6+iXX,iDy*6+iYY);
        }
      }
      rcDst.getEl(iX,iY)=t_TDst(pow(2.0,EXTRA_PRECISION_BITS)*dSum+0.5);

    }
  }
}

// internal function for spline initialization
template<class t_TSrc, class t_TDst>
template<class t_T>
double QuinticSpline<t_TSrc, t_TDst>::xInitRowCausalCoeff(const Block<const t_T> &rcSrc, const int iY, const int N, const double z1)
{
  int iX;
  double Sum, c_0;
  double zn, iz, z2n;

  zn=z1;
  iz = 1.0 / z1;
  z2n = pow(z1, (N - 1));
  Sum = rcSrc.getEl(0,iY) + z2n * rcSrc.getEl(N-1,iY);
  z2n = z2n * z2n * iz;

  for (iX = 1; iX < N-1; iX++)
  {
    Sum = Sum + (zn + z2n) * rcSrc.getEl(iX,iY);
    zn = zn * z1;
    z2n = z2n * iz;
  }

  c_0 = Sum/(1.0 - zn * zn);

  return c_0;
}

// internal function for spline initialization
template<class t_TSrc, class t_TDst>
double QuinticSpline<t_TSrc, t_TDst>::xInitColCausalCoeff(const Block<const double> &rcSrc, const int iX, const int N, const double z1)
{
  int iY;
  double Sum, c_0;
  double zn, iz, z2n;

  zn=z1;
  iz = 1.0 / z1;
  z2n = pow(z1, (N - 1));

  Sum = rcSrc.getEl(iX,0) + z2n * rcSrc.getEl(iX,N-1);
  z2n = z2n * z2n * iz;

  for (iY = 1; iY < N-1; iY++)
  {
    Sum = Sum + (zn + z2n) * rcSrc.getEl(iX,iY);
    zn = zn * z1;
    z2n = z2n * iz;
  }

  c_0 = Sum/(1.0 - zn * zn);

  return c_0;
}

template<class t_TSrc, class t_TDst>
template<class t_T>
void QuinticSpline<t_TSrc, t_TDst>::xSplineCoeffs( const Block<const t_T> &rcSrc, Block<double> &rcDst, const double dZ )
{
  int iY, iX;
  double dZ_c = dZ / ( dZ*dZ - 1.0 );
  int iXMax=rcSrc.getSize().getX();
  int iYMax=rcSrc.getSize().getY();

  // filter along rows
  for (iY=0; iY<iYMax; iY++)
  {
    // init causal coeff
    rcDst.getEl(0,iY)=xInitRowCausalCoeff<t_T>(rcSrc, iY, iXMax, dZ);
    // causal filtering
    for (iX=1; iX<iXMax; iX++)
    {
      rcDst.getEl(iX,iY)= rcSrc.getEl(iX,iY) + dZ * rcDst.getEl(iX-1,iY);
    }

    // init anti-causal coeff
    rcDst.getEl(iXMax-1,iY) = dZ_c * (dZ * rcDst.getEl(iXMax-2,iY) + rcDst.getEl(iXMax-1,iY));

    // anti-causal filtering
    for (iX=iXMax-2; iX>=0; iX--)
    {
      rcDst.getEl(iX,iY) = dZ * (rcDst.getEl(iX+1,iY) - rcDst.getEl(iX,iY));
    }
  }

  // filter along columns
  for (iX=0; iX<iXMax; iX++)
  {
    // init causal coeff
    rcDst.getEl(iX,0)=xInitColCausalCoeff(rcDst, iX, iYMax, dZ);

    // causal filtering
    for (iY=1; iY<iYMax; iY++)
    {
      rcDst.getEl(iX,iY) = rcDst.getEl(iX,iY) + dZ * rcDst.getEl(iX,iY-1);
    }

    // init anti-causal coeff
    rcDst.getEl(iX,iYMax-1) = dZ_c * (dZ * rcDst.getEl(iX,iYMax-2) + rcDst.getEl(iX,iYMax-1));

    // anti-causal filtering
    for (iY=iYMax-2; iY>=0; iY--)
    {
      rcDst.getEl(iX,iY) = dZ * (rcDst.getEl(iX,iY+1) - rcDst.getEl(iX,iY));
    }
  }
}

// Convert image samples to spline coefficients
template<class t_TSrc, class t_TDst>
void QuinticSpline<t_TSrc, t_TDst>::calcCoeffs(const Block<const t_TSrc> &rcSrc, UInt uiPl)
{
  // filter for pole z1
  AllocBlock<double> cTemp( rcSrc.getSize() );
  xSplineCoeffs<t_TSrc>( rcSrc, cTemp, m_dZ1 );

  // filter for pole z2
  const Block<double> &cSplCoeffs = *m_pcSplCoeff[uiPl];
  int iXMax=rcSrc.getSize().getX();
  int iYMax=rcSrc.getSize().getY();
  Block<double> cSpl(PixelPos(m_iSplPadding,m_iSplPadding), BlockSize(iXMax, iYMax), cSplCoeffs);
  xSplineCoeffs<double>( cTemp, cSpl, m_dZ2 );

  // extend using mirror symmetry
  int iSrcX, iDstX, iSrcY, iDstY, iX;
  // left
  for(iSrcX=1, iDstX=m_iSplPadding-1; iSrcX<=m_iSplPadding; iSrcX++, iDstX--)
  {
    for(iSrcY=0, iDstY=m_iSplPadding; iSrcY<iYMax; iSrcY++, iDstY++)
    {
      cSplCoeffs.getEl(iDstX,iDstY)=cSpl.getEl(iSrcX,iSrcY);
    }
  }
  // right
  for(iSrcX=iXMax-2, iDstX=m_iSplPadding+iXMax; iSrcX>iXMax-2-m_iSplPadding; iSrcX--, iDstX++)
  {
    for(iSrcY=0, iDstY=m_iSplPadding; iSrcY<iYMax; iSrcY++, iDstY++)
    {
      cSplCoeffs.getEl(iDstX,iDstY)=cSpl.getEl(iSrcX,iSrcY);
    }
  }
  // top
  for(iSrcY=m_iSplPadding+1, iDstY=m_iSplPadding-1; iDstY>=0; iSrcY++, iDstY--)
  {
    for(iX=0; iX<iXMax+2*m_iSplPadding; iX++)
    {
      cSplCoeffs.getEl(iX,iDstY)=cSplCoeffs.getEl(iX,iSrcY);
    }
  }
  // bottom
  for(iSrcY=m_iSplPadding+iYMax-2, iDstY=m_iSplPadding+iYMax; iDstY<iYMax+2*m_iSplPadding; iSrcY--, iDstY++)
  {
    for(iX=0; iX<iXMax+2*m_iSplPadding; iX++)
    {
      cSplCoeffs.getEl(iX,iDstY)=cSplCoeffs.getEl(iX,iSrcY);
    }
  }
}


template<VecType vt>
void dump( const IntVec2D<vt>& rcVec, std::ostream& rcOs)
{
  rcOs << " (" << rcVec.getX() << " : " << rcVec.getY() << ") ";
}
/*
template<class T>
void dump( const Block<T> rcBlock, std::ostream& rcOs)
{
  const BlockSize& rcBlockSize = rcBlock.getSize();
  rcOs << std::endl;
  for( int y= 0; y < rcBlockSize.getY(); y++)
  {
    for( int x= 0; x < rcBlockSize.getX(); x++)
    {
      rcOs << " 0x" << std::setw(4) << std::setfill('0') << std::hex << (short)rcBlock.getEl(x,y);
    }
    rcOs << std::dec << std::endl;
  }
  rcOs << std::endl;
}
*/

template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::xCopyTo(Block<t_TDst>& rcDst, 
                                               const Block<t_TDst>& rcSrc, 
                                               const int iXPos,
                                               const int iYPos,
                                               const int iXAdd,
                                               const int iYAdd) const 
{
  const BlockSize& rcSrcSize = rcSrc.getSize();


  for( int y = 0, yo = iYPos; y < rcSrcSize.getY(); y++, yo += iYAdd )
  {
    for( int x = 0, xo = iXPos; x < rcSrcSize.getY(); x++, xo += iXAdd )
    {
      rcDst.getEl( xo, yo ) = rcSrc.getEl( x,y);
    }
  }
}


//#define TEST( exp ) 

template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::xInterpolateBlock(Block<t_TDst>& rcDstBlock, 
                                                         const UInt uiPlane, 
                                                         const PixelPos &rcPosInUps,
                                                         BinaryFraction<15> &rcOffX,
                                                         BinaryFraction<15> &rcOffY) const

{ 
  const UInt uiInterPolFacLog2 = 2;
  const int iVerPos = int((rcOffY.getNumerator())>>(15-uiInterPolFacLog2));
  const int iHorPos = int((rcOffX.getNumerator())>>(15-uiInterPolFacLog2));

  rcOffX.setNumerator( (rcOffX.getNumerator() - (iHorPos<<(15-uiInterPolFacLog2)))<<uiInterPolFacLog2 );
  rcOffY.setNumerator( (rcOffY.getNumerator() - (iVerPos<<(15-uiInterPolFacLog2)))<<uiInterPolFacLog2 );

  // this add size we need for further chroma subsampling
  const int iAddSizeX = 0 != rcOffX.getNumerator() ? 1 : 0;
  const int iAddSizeY = 0 != rcOffY.getNumerator() ? 1 : 0;

  AOT_DBG( iVerPos < 0 || iVerPos > 3);
  AOT_DBG( iHorPos < 0 || iHorPos > 3);

  const BlockSize& rcDestSize = rcDstBlock.getSize();

  BlockSize cInsideSize = rcDestSize;
  BlockSize cValidSize;
  PaddingMode ePaddingMode;

  Block<short>* pcTempHorBlock = NULL;

  const Block<short>& rcSrcBlock = *m_pi16Q6SplCoeffFP[uiPlane];
  xCopyFromSrcBuffer( pcTempHorBlock, rcSrcBlock, rcPosInUps, ePaddingMode, cInsideSize, cValidSize, (iHorPos == 3)?1:0, (iVerPos == 3)?1:0 );

  Block<short>& cTempHorBlock = *pcTempHorBlock;
  if( NO_PADDING == ePaddingMode )
  {
    if( 0 == (iAddSizeX + iAddSizeY) )
    {
      AllocBlock<short> cTempVerBlock( BlockSize(rcDestSize.getX(), rcDestSize.getY()+3 ) );
      xFilterHorizontal( cTempVerBlock, cTempHorBlock, iHorPos );
      xFilterVertical( rcDstBlock, cTempVerBlock, iVerPos);
      delete pcTempHorBlock;
      return;
    }

    const int iNewHorPos = (iHorPos +1)&3;
    const int iNewVerPos = (iVerPos +1)&3;

    int iShift = 1;
    AllocBlock<int> cSum( rcDestSize );
    const int iAddYLine = 0 == iNewVerPos ? 1:0;
    const BlockSize cVerBlockSize(rcDestSize.getX(), rcDestSize.getY()+3+iAddYLine);
    AllocBlock<short> cTempVerBlock( cVerBlockSize );
    const int iVerOffset = iAddYLine * cTempVerBlock.getStride();

    Block<short> cTempVerBlock2( &cTempVerBlock.getEl(iVerOffset), BlockSize( rcDestSize.getX(), rcDestSize.getY()+3), cTempVerBlock.getStride());
    {
      xFilterHorizontal( cTempVerBlock, cTempHorBlock, iHorPos );

      xFilterVertical( rcDstBlock, cTempVerBlock, iVerPos);
      cSum.assign( rcDstBlock );

      if( 0 != iAddSizeY)
      {
        xFilterVertical( rcDstBlock, cTempVerBlock2, iNewVerPos);
        cSum += rcDstBlock;
      }
    }

    if( 0 != iAddSizeX )
    {
      const int iAddXCol = 0 == iNewHorPos ? 1:0;
      Block<short> cTempHorBlock1( cTempHorBlock.getPtr()+iAddXCol, BlockSize( rcDestSize.getX()+3, rcDestSize.getY()+3+iAddYLine), cTempHorBlock.getStride());

      xFilterHorizontal( cTempVerBlock, cTempHorBlock1, iNewHorPos );

      xFilterVertical( rcDstBlock, cTempVerBlock, iVerPos);
      cSum += rcDstBlock;

      if( 0 != iAddSizeY)
      {
        xFilterVertical( rcDstBlock, cTempVerBlock2, iNewVerPos);
        cSum += rcDstBlock;
        iShift = 2;
      }
    }
    cSum += iShift;
    cSum >>= iShift;
    (rcDstBlock).assign(cSum);
    delete pcTempHorBlock;
    return;
  }

//  else
  {
    int iPaddingSizeX = cValidSize.getX();
    int iPaddingSizeY = cValidSize.getY();

    const short* piSrcPtr = cTempHorBlock.getPtr();
    const int iSrcStride = cTempHorBlock.getStride();

    // here we need some extra storage
    xFilterInternal( rcDstBlock, piSrcPtr, iSrcStride, iHorPos, iVerPos, ePaddingMode, cInsideSize, cValidSize, 0 );

    if( 0 != (iAddSizeX + iAddSizeY) )
    {
      AllocBlock<int> cSum( rcDestSize );
      cSum.assign( rcDstBlock );
      const int iNewHorPos = (iHorPos +1)&3;
      const int iNewVerPos = (iVerPos +1)&3;

      int iInsideSizeX = cInsideSize.getX();
      int iInsideSizeY = cInsideSize.getY();
      int iValidSizeX = cValidSize.getX();
      int iValidSizeY = cValidSize.getY();    
      int iHorOffset = 0;
      int iVerOffset = 0;

      if( iAddSizeX )
      {
        if( 0 == iNewHorPos )
        {
          iHorOffset = 1;
          if( ePaddingMode & PAD_LEFT_BORDER )
          {
            iInsideSizeX++;
            // here we have to force a different 1st row
          }
          else if( ePaddingMode & PAD_RIGHT_BORDER )
          {
            if( 1 < iValidSizeX )
            {
              iInsideSizeX--;
              iValidSizeX--;
              iPaddingSizeX--;
            }
            else if( 1 == iValidSizeX )
            {
              iHorOffset = 0;
              iValidSizeX--;
              iPaddingSizeX--;
            }
            else
            {
              iHorOffset = 0;
            }
            // here we have to force a different 1st row
          }
        }
        BlockSize cNewInsideSize( iInsideSizeX, cInsideSize.getY());
        BlockSize cNewValidSize( iValidSizeX, cValidSize.getY());
        xFilterInternal( rcDstBlock, piSrcPtr + iHorOffset, iSrcStride, iNewHorPos, iVerPos, ePaddingMode, cNewInsideSize, cNewValidSize, 1 );
        cSum += rcDstBlock;
      }

      if( iAddSizeY )
      {
        if(0 == iNewVerPos)
        {
          iVerOffset = iSrcStride;
          if( ePaddingMode & PAD_TOP_BORDER )
          {
            iInsideSizeY++;
            // here we have to force a different 1st line
          }
          else if( ePaddingMode & PAD_BOTTOM_BORDER )
          {
            if(1 < iValidSizeY)
            {
              iInsideSizeY--;
              iValidSizeY--;
              iPaddingSizeY--;
            }
            else if(1 ==iValidSizeY)
            {
              iPaddingSizeY--;
              iValidSizeY--;
              iVerOffset = 0;
            }
            else
            {
              iVerOffset = 0;
            }
            // here we have to force a different 1st line
          }
        }
        BlockSize cNewInsideSize( cInsideSize.getX(), iInsideSizeY);
        BlockSize cNewValidSize( cValidSize.getX(), iValidSizeY);
        xFilterInternal( rcDstBlock, piSrcPtr + iVerOffset, iSrcStride, iHorPos, iNewVerPos, ePaddingMode, cNewInsideSize, cNewValidSize, 2 );
        cSum += rcDstBlock;
      }

      if( iAddSizeX && iAddSizeY )
      {
        BlockSize cNewInsideSize( iInsideSizeX, iInsideSizeY);
        BlockSize cNewValidSize( iValidSizeX, iValidSizeY);
        xFilterInternal( rcDstBlock, piSrcPtr + iHorOffset + iVerOffset, iSrcStride, iNewHorPos, iNewVerPos, ePaddingMode, cNewInsideSize, cNewValidSize, 3 );
        cSum += rcDstBlock;
        cSum += 2;
        cSum >>= 2;
      }
      else
      {
        cSum += 1;
        cSum >>= 1;
      }
      (rcDstBlock).assign(cSum);
    }

    BlockSize cNewInsideSize( std::min( rcDestSize.getX(), iPaddingSizeX+1), std::min( rcDestSize.getY(), iPaddingSizeY+1));
    xApplyPadding( rcDstBlock, ePaddingMode, cNewInsideSize );
  }
  delete pcTempHorBlock;
}


template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::xFilterInternal( Block<short>& rcDest, const short* piSrc, const int iSrcStride, const int iHorFrac, const int iVerFrac, const PaddingMode ePaddingMode, const BlockSize& cInsideSize, const BlockSize& cValidSize, UInt uiVerySpecialFilterPosHandling) const 
{
  const BlockSize rcDestSize = rcDest.getSize();
  const BlockSize cVerBlockSize(rcDestSize.getX(), rcDestSize.getY()+3);

  AllocBlock<short> cTempVerBlock( cVerBlockSize );
  AllocBlock<short> cTempVerBlockTopBot   ( cVerBlockSize );
  AllocBlock<short> cTempVerBlockLeftRight( cVerBlockSize );

  AllocBlock<short> cDestTopBot( rcDestSize );
  AllocBlock<short> cDestCorner( rcDestSize );
  AllocBlock<short> cDestLeftRight( rcDestSize );


  const Block<const short> cTempHorBlock( piSrc, BlockSize( rcDestSize.getX()+3, rcDestSize.getY()+3), iSrcStride);
  xFilterHorizontal( cTempVerBlock, cTempHorBlock, iHorFrac );

  int iXInsertPos = 0;
  int iXSrcOffset = 0;
  int iLeftRightFilter = -1; 
  if( 0 != (ePaddingMode & PAD_LEFT_BORDER) )
  {
    iXSrcOffset = 1+ cValidSize.getX() - cInsideSize.getX();
    iXInsertPos = rcDestSize.getX()-cValidSize.getX()-1;
    iLeftRightFilter = uiVerySpecialFilterPosHandling & 1 ? 1 : 0;
    xFilterHorizontal( cTempVerBlockLeftRight, cTempHorBlock, iLeftRightFilter);
  }
  else if( 0 != (ePaddingMode & PAD_RIGHT_BORDER) )
  {
    iXInsertPos = std::min( rcDestSize.getX()-1, cValidSize.getX());
    if( cValidSize.getX() != rcDestSize.getX() )
    {
      iXSrcOffset = cInsideSize.getX()-cValidSize.getX()-1;
    }

    iLeftRightFilter = 3; // don't care about the filter position just go for it
    xFilterHorizontal( cTempVerBlockLeftRight, cTempHorBlock, iLeftRightFilter);
  }


  // vertical filtering
  xFilterVertical( rcDest, cTempVerBlock, iVerFrac);

  int iYInsertPos = 0;
  int iYSrcOffset = 0;
  int iTopBottomFilter = -1; 
  if( 0 != (ePaddingMode & PAD_TOP_BORDER) )
  {
    iYSrcOffset = 1+ cValidSize.getY() - cInsideSize.getY();
    iYInsertPos = rcDestSize.getY()-cValidSize.getY()-1;
    iTopBottomFilter = uiVerySpecialFilterPosHandling & 2 ? 1 : 0;
    xFilterVertical( cDestTopBot, cTempVerBlock, iTopBottomFilter);
  }
  else if( 0 != (ePaddingMode & PAD_BOTTOM_BORDER) )
  {
    if( cValidSize.getY() != rcDestSize.getY() )
    {
      iYSrcOffset = cInsideSize.getY()-cValidSize.getY()-1;
    }
    iYInsertPos = std::min( rcDestSize.getY()-1, cValidSize.getY());
    iTopBottomFilter = 3; // again don't care about the filter position just go for it
    xFilterVertical( cDestTopBot, cTempVerBlock, iTopBottomFilter);
  }

  if( -1 != iLeftRightFilter )
  {
    xFilterVertical( cDestLeftRight, cTempVerBlockLeftRight, iVerFrac);

    // copy the left right col to the right pos
    for( int y = 0; y < rcDestSize.getY(); y++)
    {
      rcDest.getEl(iXInsertPos,y) = cDestLeftRight.getEl(std::max(0,iXInsertPos+iXSrcOffset),y);
    }

    if( -1 != iTopBottomFilter )
    {
      //we have to filter an corner block
      xFilterVertical( cDestCorner, cTempVerBlockLeftRight, iTopBottomFilter);

      // copy the top bottom line to the right pos
      ::memcpy( &rcDest.getEl(0,iYInsertPos), &cDestTopBot.getEl(0,std::max(0,iYInsertPos+iYSrcOffset)), rcDestSize.getX()*sizeof(short));

      // copy the corner Pixel to the right pos
      rcDest.getEl(iXInsertPos,iYInsertPos) = cDestCorner.getEl(std::max(0,iXInsertPos+iXSrcOffset),std::max(0,iYInsertPos+iYSrcOffset));
    }

  }
  else if( -1 != iTopBottomFilter )
  {
    // copy the top bottom line to the right pos
    ::memcpy( &rcDest.getEl(0,iYInsertPos), &cDestTopBot.getEl(0,std::max(0,iYInsertPos+iYSrcOffset)), rcDestSize.getX()*sizeof(short));
  }
}




template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::xCopyFromSrcBuffer( Block<short>*& rpcDestBlock, const Block<short> &rcSrc, const PixelPos& rcPixelPosIn, PaddingMode& rePaddingMode, BlockSize& rcInsideSize, BlockSize& rcValidSize, int iHorAdd, int iVerAdd ) const 
{
  BlockSize rcDestSize = rcInsideSize;
  const BlockSize cSrcSize = rcSrc.getSize()-BlockSize(4,4); 

  int iXSize = 0;
  int iYSize = 0;

  int iXInsideSize = 0;
  int iYInsideSize = 0;

  int iDstXOffset = 0;
  int iDstYOffset = 0;

  int iAddXRead = 0;
  int iAddYRead = 0;
  PixelPos cPixelPos = rcPixelPosIn;

  rePaddingMode = INVALID_PADDING_MODE;
  if( 0 > cPixelPos.getY() )
  {
    AOT_DBG( (cPixelPos.getY() + rcDestSize.getY() + iVerAdd) > cSrcSize.getY() ); // Padding on both borders not supported
    iAddYRead = 1;
    iYInsideSize = std::max( 0, rcDestSize.getY() + cPixelPos.getY());
    iYSize = std::max( 1, rcDestSize.getY() + cPixelPos.getY());
    iDstYOffset = rcDestSize.getY() - iYSize;
    cPixelPos.setY( std::max(-iDstYOffset, 0) );  // at least we need one pixel of the src

    if( 0 > cPixelPos.getX() )
    {
      iAddXRead = 1;
      iXInsideSize = std::max( 0, rcDestSize.getX() + cPixelPos.getX());
      iXSize = std::max( 1, rcDestSize.getX() + cPixelPos.getX());
      iDstXOffset = rcDestSize.getX() - iXSize;
      cPixelPos.setX( std::max(-iDstXOffset, 0) );  // at least we need one pixel of the src
      rePaddingMode = PAD_LEFT_AND_TOP_BORDER;
      AOT_DBG( (cPixelPos.getX() + rcDestSize.getX() + iHorAdd) > cSrcSize.getX() ); // Padding on both borders not supported
    }
    else if( ( cPixelPos.getX() + rcDestSize.getX() + iHorAdd) > cSrcSize.getX() )
    {
      iXInsideSize = std::max( 0, cSrcSize.getX() - cPixelPos.getX());
      iXSize = std::max( 1, cSrcSize.getX() - cPixelPos.getX());
      cPixelPos.setX( cSrcSize.getX() - iXSize );
      rePaddingMode = PAD_RIGHT_AND_TOP_BORDER;
    }
    else
    {
      rePaddingMode = PAD_TOP_BORDER;
      iXSize = rcDestSize.getX();
      iXInsideSize = rcDestSize.getX();
    }
  }
  else if( (cPixelPos.getY() + rcDestSize.getY() + iVerAdd) > cSrcSize.getY() )
  {
    iYInsideSize = std::max( 0, cSrcSize.getY() - cPixelPos.getY());
    iYSize = std::max( 1, cSrcSize.getY() - cPixelPos.getY());
    cPixelPos.setY( cSrcSize.getY() - iYSize );

    if( 0 > cPixelPos.getX() )
    {
      iAddXRead = 1;
      iXInsideSize = std::max( 0, rcDestSize.getX() + cPixelPos.getX());
      iXSize = std::max( 1, rcDestSize.getX() + cPixelPos.getX());
      iDstXOffset = rcDestSize.getX() - iXSize;
      cPixelPos.setX( std::max(-iDstXOffset, 0) );  // at least we need one pixel of the src
      rePaddingMode = PAD_LEFT_AND_BOTTOM_BORDER;
      AOT_DBG( (cPixelPos.getX() + rcDestSize.getX() + iHorAdd) > cSrcSize.getX() ); // Padding on both borders not supported
    }
    else if( (cPixelPos.getX() + rcDestSize.getX()+ iHorAdd) > cSrcSize.getX() )
    {
      iXInsideSize = std::max( 0, cSrcSize.getX() - cPixelPos.getX());
      iXSize = std::max( 1, cSrcSize.getX() - cPixelPos.getX());
      cPixelPos.setX( cSrcSize.getX() - iXSize );
      rePaddingMode = PAD_RIGHT_AND_BOTTOM_BORDER;
    }
    else
    {
      rePaddingMode = PAD_BOTTOM_BORDER;
      iXSize = rcDestSize.getX();
      iXInsideSize = rcDestSize.getX();
    }
  }
  else
  {
    if( 0 > cPixelPos.getX() )
    {
      iAddXRead = 1;
      iXInsideSize = std::max( 0, rcDestSize.getX() + cPixelPos.getX());
      iYInsideSize = rcDestSize.getY();
      iYSize = rcDestSize.getY();
      iXSize = std::max( 1, rcDestSize.getX() + cPixelPos.getX());
      iDstXOffset = rcDestSize.getX() - iXSize;
      cPixelPos.setX( std::max(-iDstXOffset, 0) );  // at least we need one pixel of the src
      rePaddingMode = PAD_LEFT_BORDER;
      AOT_DBG( (cPixelPos.getX() + rcDestSize.getX() + iHorAdd) > cSrcSize.getX() ); // Padding on both borders not supported
    }
    else if( (cPixelPos.getX() + rcDestSize.getX()+ iHorAdd) > cSrcSize.getX() )
    {
      iXInsideSize = std::max( 0, cSrcSize.getX() - cPixelPos.getX());
      iYInsideSize = rcDestSize.getY();
      iYSize = rcDestSize.getY();
      iXSize = std::max( 1, cSrcSize.getX() - cPixelPos.getX());
      cPixelPos.setX( cSrcSize.getX() - iXSize );
      rePaddingMode = PAD_RIGHT_BORDER;
    }
    else
    {
      rePaddingMode = NO_PADDING;
    }
  }

  if( NO_PADDING == rePaddingMode )
  {
    rpcDestBlock = new Block<short>( PixelPos(cPixelPos.getX()-1+2, cPixelPos.getY()-1+2), BlockSize( rcDestSize.getX() + 4, rcDestSize.getY() + 4), rcSrc );
    return;
  }

  AOT_DBG( rcDestSize.getY() < iYSize || 0 >= iYSize);
  AOT_DBG( rcDestSize.getX() < iXSize || 0 >= iXSize);

  AOT_DBG( rcDestSize.getY() < iDstYOffset + iYSize);
  AOT_DBG( rcDestSize.getX() < iDstXOffset + iXSize);

  rcValidSize.setX( iXInsideSize );
  rcValidSize.setY( iYInsideSize );

  rcInsideSize.setX( iXSize );
  rcInsideSize.setY( iYSize );

  rpcDestBlock = new AllocBlock<short>( BlockSize( rcDestSize.getX() + 4, rcDestSize.getY() + 4) );

  short* piDest = &rpcDestBlock->getEl( iDstXOffset, iDstYOffset );
  const int iDstStride = rpcDestBlock->getStride();

  const short* piSrc = &rcSrc.getEl( cPixelPos.getX()-1+2, cPixelPos.getY()-1+2 );
  const int iSize = (iXSize+3+iAddXRead) * sizeof(short);
  const int iSrcStride = rcSrc.getStride();

  for( int y = 0; y < iYSize+3+iAddYRead; y++ )
  {
    ::memcpy( piDest, piSrc, iSize );
    piDest += iDstStride;
    piSrc += iSrcStride;
  }
}


template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::xFilterHorizontal( Block<short> &rcDst, const Block<const short> &rcSrc, int iFilterPos) const 
{
  const BlockSize& rcDestSize = rcDst.getSize();

  short* piTmp = &rcDst.getEl( 0 );
  const int iTmpStride = rcDst.getStride();
  const int iSrcStride = rcSrc.getStride();
  const short* piSrc = &rcSrc.getEl( 0 );
  const int iDstYMax = rcDestSize.getY();
  const int iDstXMax = rcDestSize.getX();
  const CoeffT* piCoeff = xGetFilterCoeffs(iFilterPos);

  // interpolation in horizontal direction
  for( int iY = 0; iY < iDstYMax; iY++ )
  {
    for ( int iX = 0; iX < iDstXMax; iX++ )
    {
      const int i32Q21Sum = piCoeff[0] * int( piSrc[0+iX] )+ piCoeff[1] * int( piSrc[1+iX] )
                            + piCoeff[2] * int( piSrc[2+iX] )+ piCoeff[3] * int( piSrc[3+iX] );
      const int i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
      piTmp[iX] = short( ( m_i16Q2KernelGain * i16Q6Sum + (1<<1) ) >> 2 );
    }
    piTmp += iTmpStride;
    piSrc += iSrcStride;
  }
}

template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::xFilterVertical( Block<short> &rcDst, const Block<const short> &rcSrc, int iFilterPos) const 
{
  const BlockSize& rcDestSize = rcDst.getSize();

  short* piDes = &rcDst.getEl( 0 );
  const int iDesStride = rcDst.getStride();
  const int iSrcStride = rcSrc.getStride();
  const short* piSrcLn = &rcSrc.getEl( 0 ) + iSrcStride ; // (0,1)
  const int iDstYMax = rcDestSize.getY();
  const int iDstXMax = rcDestSize.getX();
  const CoeffT* piCoeff = xGetFilterCoeffs(iFilterPos);

  for ( int iY = 0; iY < iDstYMax; iY++ )
  {
    for ( int iX = 0; iX < iDstXMax; iX++ )
    {
      // 1d convolution
      const short* piSrc = piSrcLn + iX;
      const int i32Q21Sum = piCoeff[0] * int( piSrc[-iSrcStride] ) + piCoeff[1] * int( piSrc[0] )
                            + piCoeff[2] * int( piSrc[iSrcStride] ) + piCoeff[3] * int( piSrc[2*iSrcStride] );

      const int i16Q6Sum = ( ( i32Q21Sum + (1<<14) ) >> 15 );
      piDes[iX] = short( ( m_i16Q2KernelGain * i16Q6Sum + (1<<1) ) >> 2 );
    }
    piDes += iDesStride;
    piSrcLn += iSrcStride;
  }
}

template<class t_TSrc, class t_TDst>
void CubicSplineFixed<t_TSrc, t_TDst>::xApplyPadding( Block<short> &rcDst, PaddingMode ePaddingMode, const BlockSize& rcInsideSize ) const 
{
  const BlockSize& rcDestSize = rcDst.getSize();
  const int iStride = rcDst.getStride();

  switch ( ePaddingMode )
  {
  case PAD_RIGHT_BORDER:
    {
      short* pDes = &rcDst.getEl( 0 );
      for( int y = 0; y < rcDestSize.getY(); y++ )
      {
        for( int x = rcInsideSize.getX(); x < rcDestSize.getX(); x++ )
        {
          pDes[x] = pDes[x-1];
        }
        pDes += iStride;
      }
    }
    break;
  case PAD_RIGHT_AND_TOP_BORDER:
    {
      const int iYPadSize = rcDestSize.getY() - rcInsideSize.getY();
      short* pDes = &rcDst.getEl( 0, iYPadSize);

      for( int y = 0; y < rcInsideSize.getY(); y++ )
      {
        for( int x = rcInsideSize.getX(); x < rcDestSize.getX(); x++ )
        {
          pDes[x] = pDes[x-1];
        }
        pDes += iStride;
      }

      pDes = &rcDst.getEl( 0, iYPadSize);
      const int iSize = rcDestSize.getX() * sizeof(short);
      for( int y = 0; y < iYPadSize; y++ )
      {
        pDes -= iStride;
        ::memcpy( pDes, pDes +iStride, iSize );
      }
    }
    break;
  case PAD_RIGHT_AND_BOTTOM_BORDER:
    {
      short* pDes = &rcDst.getEl( 0 );
      for( int y = 0; y < rcInsideSize.getY(); y++ )
      {
        for( int x = rcInsideSize.getX(); x < rcDestSize.getX(); x++ )
        {
          pDes[x] = pDes[x-1];
        }
        pDes += iStride;
      }

      pDes = &rcDst.getEl( 0, rcInsideSize.getY());
      const int iYPadSize = rcDestSize.getY() - rcInsideSize.getY();
      const int iSize = rcDestSize.getX() * sizeof(short);
      for( int y = 0; y < iYPadSize; y++ )
      {
        ::memcpy( pDes, pDes -iStride, iSize );
          pDes += iStride;
      }
    }
    break;
  case PAD_LEFT_BORDER:
    {
      short* pDes = &rcDst.getEl( 0 );
      const int iXSize = rcDestSize.getX() - rcInsideSize.getX();
      for( int y = 0; y < rcDestSize.getY(); y++ )
      {
        for( int x = iXSize; x > 0; x-- )
        {
          pDes[x-1] = pDes[x];
        }
        pDes += iStride;
      }
    }
    break;
  case PAD_LEFT_AND_TOP_BORDER:
    {
      const int iYPadSize = rcDestSize.getY() - rcInsideSize.getY();
      short* pDes = &rcDst.getEl( 0, iYPadSize);

      const int iXSize = rcDestSize.getX() - rcInsideSize.getX();
      for( int y = 0; y < rcInsideSize.getY(); y++ )
      {
        for( int x = iXSize; x > 0; x-- )
        {
          pDes[x-1] = pDes[x];
        }
        pDes += iStride;
      }

      pDes = &rcDst.getEl( 0, iYPadSize);
      const int iSize = rcDestSize.getX() * sizeof(short);
      for( int y = 0; y < iYPadSize; y++ )
      {
        pDes -= iStride;
        ::memcpy( pDes, pDes +iStride, iSize );
      }
    }
    break;
  case PAD_LEFT_AND_BOTTOM_BORDER:
    {
      short* pDes = &rcDst.getEl( 0 );
      const int iXSize = rcDestSize.getX() - rcInsideSize.getX();
      for( int y = 0; y < rcInsideSize.getY(); y++ )
      {
        for( int x = iXSize; x > 0; x-- )
        {
          pDes[x-1] = pDes[x];
        }
        pDes += iStride;
      }

      const int iSize = rcDestSize.getX() * sizeof(short);
      const int iYPadSize = rcDestSize.getY() - rcInsideSize.getY();
      pDes = &rcDst.getEl( 0, rcInsideSize.getY());
      for( int y = 0; y < iYPadSize; y++ )
      {
        ::memcpy( pDes, pDes -iStride, iSize );
          pDes += iStride;
      }
    }
    break;
  case PAD_TOP_BORDER:
    {
      const int iSize = rcDestSize.getX() * sizeof(short);
      const int iYPadSize = rcDestSize.getY() - rcInsideSize.getY();
      short* pDes = &rcDst.getEl( 0, iYPadSize);
      for( int y = 0; y < iYPadSize; y++ )
      {
        pDes -= iStride;
        ::memcpy( pDes, pDes +iStride, iSize );
      }
    }
    break;
  case PAD_BOTTOM_BORDER:
    {
      const int iSize = rcDestSize.getX() * sizeof(short);
      const int iYPadSize = rcDestSize.getY() - rcInsideSize.getY();
      short* pDes = &rcDst.getEl( 0, rcInsideSize.getY());
      for( int y = 0; y < iYPadSize; y++ )
      {
        ::memcpy( pDes, pDes -iStride, iSize ); 
          pDes += iStride;
      }
    }
    break;
  case NO_PADDING:
  default:
    {
      TOT(1)
    }
    break;
  };

}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _SPLINES_H__
