/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef WIENERFILTERPARAMETER_H_
#define WIENERFILTERPARAMETER_H_


#include "VideoCommonLib/VideoCommonLibDecl.h"
#include "QuadtreeGroup.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class WienerBlockPar
{
public:
  WienerBlockPar() : m_FilterBlock(false) {}
  bool m_FilterBlock;
};

class FilterCoefficients
  {
  public:
    FilterCoefficients()
    {
      m_uiNumQuantBits = 14;
    }

    void setFilterTypeSep(UInt uiTyp,UInt uiLength);
    const int getFilterType() const {return m_uiFilterType;}
    const UInt getFilterLength() const {return m_uiFilterLength;}
    const UInt  getNumberOfCoeffs() const {return m_uiNumberOfCoeffs; }
    double quantFilterCoeffs(UInt uiNumQuantBits);
    double getDcOffset() const { return m_adNotQuantCoeff.back(); }
    int getQuantDcOffset() const { return m_aiQuantCoeff.back(); }

    int getMarginOverlap() const { return m_uiFilterLength >> 1; }

    const std::vector<int>& getQuantCoeffs() const {return m_aiQuantCoeff;}

    std::vector<UInt> getMappingRule() const {return m_auiMapping;}
    void setQuantCoeffs(std::vector<int> rQuantCoeff) { m_aiQuantCoeff = rQuantCoeff; }
    void pushBackQuantDcOffset( int iDcOffset) { m_aiQuantCoeff.push_back( iDcOffset ); }
    UInt getNumberOfQuantizationBits() const {return m_uiNumQuantBits ;}



    bool getIsHorizontalFilter()  const {return m_bIsHorizontalFilter ;}
    bool getIsVerticalFilter()    const {return m_bIsVerticalFilter ;} ;
    void setHorizontal()
    {
      m_bIsHorizontalFilter  = true ;
      m_bIsVerticalFilter    = false ;
    }
    void setVertical()
    {
      m_bIsHorizontalFilter  = false ;
      m_bIsVerticalFilter    = true ;
    }


    void setCoefficients(std::vector<double> adCoeffs)
    {
      m_adNotQuantCoeff = adCoeffs;
      quantFilterCoeffs(m_uiNumQuantBits);
    }

    int getNumberOfQuantBits() const {return m_uiNumQuantBits;}
    int getQuantCoefficient(int iFilterPos) const {return m_aiQuantCoeff[m_auiMapping[iFilterPos]]  ;}

    int getQuantFilterCoeff(UInt uiPosition) const
    {
      if (uiPosition < m_uiFilterLength )
        return m_aiQuantCoeff[m_auiMapping[uiPosition]] ;
      else
        RNOK( -1 ) ;
    }
    // debug
    const void printQuantFilterCoeffs()const
    {
      for (UInt i = 0 ; i < m_aiQuantCoeff.size(); i++)
        std::cout << m_aiQuantCoeff[i] << "\t" ;
      std::cout << std::endl ;
    }
    const void printFilterCoeffs() const
    {
      for (UInt i = 0 ; i < m_adNotQuantCoeff.size(); i++)
        std::cout <<  m_adNotQuantCoeff[i] << "\t" ;
      std::cout << std::endl ;
    }

    void pushNewQuantCoeffPredicted( UInt uiCoeffIdx , int iCoeff )
    {
      if ( uiCoeffIdx == m_uiCenterPos )
      {
        m_aiQuantCoeff.push_back(  iCoeff + ( 1 << m_uiNumQuantBits) );
      }
      else
      {
        m_aiQuantCoeff.push_back( iCoeff );
      }
    }

    int getPredQuantCoeffTransmit( UInt uiCoeffIdx ) const
    {
      if ( uiCoeffIdx == m_uiCenterPos )
      {
        return m_aiQuantCoeff[uiCoeffIdx]  - ( 1 << m_uiNumQuantBits );
      }
      else
      {
        return m_aiQuantCoeff[uiCoeffIdx];
      }
    }

   int getPredQuantDcTransmit() const
    {
     return m_aiQuantCoeff.back();
    }

//    double getCoefficient(int iFilterPos) const {return m_aiQuantCoeff[iFilterPos]/(double) (1 << m_iNumQuantBits);}

    void setCorrelationMatrix( const std::vector< std::vector<double> >& aiCorrMat ) { m_adCorrelationMatrix = aiCorrMat ; }
    const std::vector< std::vector<double> >& getCorrelationMatrix( ) { return m_adCorrelationMatrix ; }

  private:

    void calcFilterHisto();

    std::vector<double> m_adNotQuantCoeff;
    std::vector<int>    m_aiQuantCoeff;

    bool m_bIsHorizontalFilter ;
    bool m_bIsVerticalFilter ;
    UInt m_uiFilterLength;
    UInt m_uiNumQuantBits;
    UInt m_uiNumberOfCoeffs;
    UInt m_uiFilterType;
    UInt m_uiCenterPos ;

    std::vector<UInt> m_auiFilterHisto;
    std::vector<UInt> m_auiMapping ;

    std::vector< std::vector<double> > m_adCorrelationMatrix ;
  };


class WienerFilterParameters
{
public:
  WienerFilterParameters()
  {
    m_pcQuadtreeGroupStructure = NULL;
    m_pcQTBlockPars = NULL;
    m_bUseFilter = false;
    m_bUseBlocks = false;
    m_abFilterPlane[0] = false;
    m_abFilterPlane[1] = false;
    m_abFilterPlane[2] = false;
    m_auiDefaultSet[0] = 0;
    m_auiDefaultSet[1] = 0;
    m_auiDefaultSet[2] = 0;
    m_bQuadTreeUsedExtern = false;
  }

  WienerFilterParameters(BlockSize cSize,UInt uiPower2MaxBlock, UInt uiPower2MinBlock)
  {
    initQuadtreeGroupStructure(PrimBlockSize(cSize.getX(),cSize.getY()), uiPower2MaxBlock,uiPower2MinBlock);
    m_bUseFilter = true;
    m_bUseBlocks = false;
  }


  ~WienerFilterParameters()
  {
    if ( !m_bQuadTreeUsedExtern )
    {
      if (m_pcQTBlockPars != NULL)
      {
        QuadtreeGroupPointerDeleter::deleteAllPointers( *m_pcQTBlockPars );
        delete m_pcQTBlockPars;
      }
      if (m_pcQuadtreeGroupStructure != NULL)
        delete m_pcQuadtreeGroupStructure;
    }
  }

  void setUseFilterFlag(bool bUseFilter) {m_bUseFilter = bUseFilter;}
  void setUseBlocksFlag(bool bUseBlocks) {m_bUseBlocks = bUseBlocks;}
  const bool getUseFilterFlag() const {return m_bUseFilter;}
  bool getUseBlocksFlag() const {return m_bUseBlocks;}

  bool getFilterPlaneFlag( UInt uiPlaneNum ) const {return m_abFilterPlane[uiPlaneNum];}
  void setFilterPlaneFlag( UInt uiPlaneNum, bool bFlag ) {m_abFilterPlane[ uiPlaneNum] = bFlag;}
  void setFilterSetForPlane( UInt uiPlaneNum, int iFilterSet) {m_auiDefaultSet[uiPlaneNum] = iFilterSet;}
  const UInt getSetNumOfPlane ( UInt uiPlaneNum) const { return m_auiDefaultSet[ uiPlaneNum ]; }
  const UInt getNewSetNumberSep () const {return UInt( m_acHorizontalFilterSets.size() ); }
  const UInt getNewSetNumber () const {return UInt( m_acFilterSets.size() ); }

  UInt getNumberOfPlanesWhichUsesSet( UInt uiSetNum ) const
  {
    UInt uiCount = 0;
    for( UInt uiPlane=0; uiPlane < 3; uiPlane++ )
    {
      if( m_abFilterPlane[ uiPlane ] && m_auiDefaultSet[ uiPlane ] == uiSetNum )
      {
        uiCount++;
      }
    }
    return uiCount;
  }

  FilterCoefficients& getFilterCoeff( UInt iSetNum )
  {
    while ( m_acFilterSets.size() <= iSetNum )
      m_acFilterSets.push_back( FilterCoefficients() );
    return m_acFilterSets[ iSetNum ];
  }

  const FilterCoefficients& getFilterCoeff( UInt iSetNum ) const
  {
    AOT ( m_acFilterSets.size() < iSetNum )
    return m_acFilterSets[ iSetNum ];
  }


  FilterCoefficients& getFilterCoeff() {return m_acFilterSets[0];}
  const FilterCoefficients& getFilterCoeff() const {return m_acFilterSets[0];}

  const FilterCoefficients& getCoeffOfPlane( UInt uiPlane) const {return m_acFilterSets[ m_auiDefaultSet[ uiPlane ] ];}
  FilterCoefficients& getCoeffOfPlane( UInt uiPlane) {return m_acFilterSets[ m_auiDefaultSet[ uiPlane ] ];}
  std::vector<FilterCoefficients>& getFilterSets() { return m_acFilterSets; }

  UInt getNumberOfUsedSets() const
  {
    UInt uiCount=0;
    for (UInt uiSet=0; uiSet < m_acFilterSets.size(); uiSet++ )
    {
      if ( getNumberOfPlanesWhichUsesSet( uiSet ) > 0 )
      {
        uiCount++;
      }
    }
    return uiCount;
  }

  UInt getNumberOfUsedSetsSep() const
  {
    UInt uiCount=0;
    for (UInt uiSet=0; uiSet < m_acHorizontalFilterSets.size(); uiSet++ )
    {
      if ( getNumberOfPlanesWhichUsesSet( uiSet ) > 0 )
      {
        uiCount++;
      }
    }
    return uiCount;
  }


  void removeLastSetSep() { m_acHorizontalFilterSets.pop_back();
                            m_acVerticalFilterSets.pop_back(); }
  void removeLastSet() { m_acFilterSets.pop_back();}



  FilterCoefficients& getHorizontalFilterCoeff( UInt iSetNum )
  {
    while ( m_acHorizontalFilterSets.size() <= iSetNum )
      m_acHorizontalFilterSets.push_back( FilterCoefficients() );
    return m_acHorizontalFilterSets[ iSetNum ];
  }

  const FilterCoefficients& getHorizontalFilterCoeff( UInt iSetNum ) const
  {
    AOT ( m_acHorizontalFilterSets.size() < iSetNum )
    return m_acHorizontalFilterSets[ iSetNum ];
  }

  FilterCoefficients& getHorizontalFilterCoeff() {return m_acHorizontalFilterSets[0];}
  const FilterCoefficients& getHorizontalFilterCoeff() const {return m_acHorizontalFilterSets[0];}


  FilterCoefficients& getVerticalFilterCoeff( UInt iSetNum )
  {
    while ( m_acVerticalFilterSets.size() <= iSetNum )
      m_acVerticalFilterSets.push_back( FilterCoefficients() );
    return m_acVerticalFilterSets[ iSetNum ];
      }

  const FilterCoefficients& getVerticalFilterCoeff( UInt iSetNum ) const
  {
    AOT ( m_acVerticalFilterSets.size() < iSetNum )
    return m_acVerticalFilterSets[ iSetNum ];
  }

  FilterCoefficients& getVerticalFilterCoeff() {return m_acHorizontalFilterSets[0];}
  const FilterCoefficients& getVerticalFilterCoeff() const {return m_acVerticalFilterSets[0];}


  const FilterCoefficients& getHorizontalCoeffOfPlane( UInt uiPlane) const {return m_acHorizontalFilterSets[ m_auiDefaultSet[ uiPlane ] ];}
  FilterCoefficients& getHorizontalCoeffOfPlane( UInt uiPlane) {return m_acHorizontalFilterSets[ m_auiDefaultSet[ uiPlane ] ];}

  const FilterCoefficients& getVerticalCoeffOfPlane( UInt uiPlane) const {return m_acVerticalFilterSets[ m_auiDefaultSet[ uiPlane ] ];}
  FilterCoefficients& getVerticalCoeffOfPlane( UInt uiPlane) {return m_acVerticalFilterSets[ m_auiDefaultSet[ uiPlane ] ];}

  std::vector<FilterCoefficients>& getHorizontalFilterSets() { return m_acHorizontalFilterSets; }
  std::vector<FilterCoefficients>& getVerticalFilterSets() { return m_acVerticalFilterSets; }






  //QUADTREE
  void initQuadtreeGroupStructure(PrimBlockSize cSize,UInt uiPower2MaxBlock, UInt uiPower2MinBlock);
  void initQuadtreeGroupStructure(PrimBlockSize cSize) { initQuadtreeGroupStructure(cSize, m_uiPower2MaxBlock , m_uiPower2MinBlock ); }
  void setQuadTreeUsedExternFlag ( bool bUsedExt ) { m_bQuadTreeUsedExtern = bUsedExt; }

  QuadtreeGroup<WienerBlockPar*>& getQTBlockPars(){return *m_pcQTBlockPars;}
  const QuadtreeGroup<WienerBlockPar*>& getQTBlockPars() const {return *m_pcQTBlockPars;}
  QuadtreeGroup<WienerBlockPar*>* getQTBlockParsPointer() const {return m_pcQTBlockPars;}
  void setQTBlockPars( QuadtreeGroup<WienerBlockPar*>* pcQTGroup) {m_pcQTBlockPars = pcQTGroup;}

  QuadtreeGroupStructure&  getQuadtreeGroupStructure() { return *m_pcQuadtreeGroupStructure; }
  const QuadtreeGroupStructure&  getQuadtreeGroupStructure() const { return *m_pcQuadtreeGroupStructure; }
  QuadtreeGroupStructure* getQuadtreeGroupStructurePointer() {return m_pcQuadtreeGroupStructure; }
  void setQuadtreeGroupStructure(QuadtreeGroupStructure* pcQTStructure ) { m_pcQuadtreeGroupStructure = pcQTStructure;}

  UInt getPower2MaxBlock() const {return m_uiPower2MaxBlock;}
  UInt getPower2MinBlock() const {return m_uiPower2MinBlock;}

  void setPower2MaxBlock( UInt uiMaxBlock) { m_uiPower2MaxBlock = uiMaxBlock;}
  void setPower2MinBlock( UInt uiMinBlock) { m_uiPower2MinBlock = uiMinBlock;}


  ErrVal deleteTree()
  {
    if (m_pcQTBlockPars != NULL)
    {
      QuadtreeGroupPointerDeleter::deleteAllPointers( *m_pcQTBlockPars );
      delete m_pcQTBlockPars;
    }
    if (m_pcQuadtreeGroupStructure != NULL)
      delete m_pcQuadtreeGroupStructure;
    return Err::m_nOK;
  }


private:

  std::vector<FilterCoefficients> m_acHorizontalFilterSets;
  std::vector<FilterCoefficients> m_acVerticalFilterSets;
  std::vector<FilterCoefficients> m_acFilterSets;

  QuadtreeGroupStructure* m_pcQuadtreeGroupStructure;
  QuadtreeGroup<WienerBlockPar*>* m_pcQTBlockPars;
  UInt m_uiPower2MaxBlock;
  UInt m_uiPower2MinBlock;

  bool m_abFilterPlane[3];
  UInt m_auiDefaultSet[3];
  bool m_bUseFilter;
  bool m_bUseBlocks;
  bool m_bQuadTreeUsedExtern;
};










// video namespace end
VIDEO_NAMESPACE_END


#endif /* WIENERFILTER_H_ */
