/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoDecoderPch.h"
#include "VideoDecoderApp.h"
#include "VideoCodecToolsLib/YuvPictureWriter.h"
#include "VideoCodecToolsLib/PacketParser.h"
#include "VideoDecoderLib/VideoDecoderInstance.h"
#include "VideoCodecToolsLib/MSysTimeStamp.h"

#include<iomanip>

VideoDecoderApp::VideoDecoderApp()
: m_pcVideoDecoderInstance( NULL )
{
}

VideoDecoderApp::~VideoDecoderApp()
{
}

ErrVal VideoDecoderApp::runAlternative( int argc, char **argv, bool &rbExit )
{
  RNOK( VideoDecoderInstance::runAlternative( argc, argv, rbExit ) );
  return Err::m_nOK;
}


ErrVal VideoDecoderApp::create( VideoDecoderApp*& rpcVideoDecoderApp )
{
  rpcVideoDecoderApp = new VideoDecoderApp;
  ROT( NULL == rpcVideoDecoderApp );

  return Err::m_nOK;
}


ErrVal VideoDecoderApp::destroy()
{
  if( NULL != m_pcVideoDecoderInstance )
  {
    RNOK( m_pcVideoDecoderInstance->destroy() );
    m_pcVideoDecoderInstance = NULL;
  }
  if( m_cDecoded.is_open() )
  {
    m_cDecoded.close();
  }
  if ( !m_apcOverlay.empty() )
  {
   for ( UInt uiPlaPaId = 0; uiPlaPaId < m_apcOverlay.size(); uiPlaPaId++ )
    {
      if ( m_apcOverlay[uiPlaPaId]->is_open() )
      {
        m_apcOverlay[uiPlaPaId]->close();
      }
      delete m_apcOverlay[uiPlaPaId];
    }
  }
  delete this;
  return Err::m_nOK;
}

void VideoDecoderApp::xDisplayUsage( char *pcArgv0 )
{
  std::cout << "Usage: " << pcArgv0 << " stream dec [overlayPic [...]]" << std::endl << std::endl;
  std::cout << "stream        bitstream to decode" << std::endl;
  std::cout << "dec           decoded stream" << std::endl;
  std::cout << "overlayPic    decoded stream with informative overlay (\"none\" disables the overlay picture)" << std::endl;
  std::cout << "...           arbitrary number of additional streams to decode" << std::endl << std::endl;
}


ErrVal VideoDecoderApp::init( int argc, char**argv )
{
  if( ( argc < 3 ) )
  {
    xDisplayUsage( argv[0] );
    return Err::m_nERR;
  }
  if( Err::m_nOK != m_cDecoded.open( argv[2], LargeFile::OM_WRITEONLY ) )
  {
    std::cout << "Could not open decoded file!" << std::endl;
    xDisplayUsage( argv[0] );
    return Err::m_nERR;
  }
  m_cOverlayFilename = "none";
  if( argc >= 4 && std::string( argv[3] ) != std::string( "none" ) )
  {
    m_cOverlayFilename = std::string( argv[3] );
  }

  LargeFile cTest;
  for( UInt ui = 1; ui < UInt( argc ); ui += ui == 1 ? 3 : 1 )
  {
    if( Err::m_nOK == cTest.open( argv[ui], LargeFile::OM_READONLY ) )
    {
      m_acBitstreams.push_back( argv[ui] );
      cTest.close();
    }
    else
    {
      std::cout << "Could not open Bitstream file: " << argv[ui] << std::endl;
      xDisplayUsage( argv[0] );
      return Err::m_nERR;
    }
  }
  RNOK( VideoDecoderInstance::create( m_pcVideoDecoderInstance ) );
  return Err::m_nOK;
} 


class AllocVideoCodecPicture
  : public VideoCodecPicture<VideoCodecPel>
{
public:
  ErrVal allocateBuffer()
  {
    for( UInt ui = 0; ui < m_acPlanes.size(); ui++ )
    {
      ROT( m_acPlanes[ui].m_pData );
      ROF( m_acPlanes[ui].m_pData = new VideoCodecPel[m_acPlanes[ui].m_iWidth * m_acPlanes[ui].m_iHeight] );
      m_acPlanes[ui].m_uiStride = m_acPlanes[ui].m_iWidth;
    }
    return Err::m_nOK;
  }

  virtual ~AllocVideoCodecPicture()
  {
    for( UInt ui = 0; ui < m_acPlanes.size(); ui++ )
    {
      delete[] m_acPlanes[ui].m_pData;
    }
  }
};

ErrVal VideoDecoderApp::go()
{
  YuvPictureWriter<VideoCodecPel> cDecodedWriter( m_cDecoded );
  std::vector<VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*>*> apcOverlayWriter;

  TimeStamp cT1;
  TimeStamp cT2; 

  cT1.stamp();

  UInt uiDecCount = 0;
  for( UInt ui = 0; ui < m_acBitstreams.size(); ui++ )
  {
    std::cout << std::endl << "Beginning decoding of file " << std::setw( 4 ) << ( ui + 1 ) << " of " << std::setw( 4 ) << m_acBitstreams.size() << ": " << m_acBitstreams[ui] << std::endl;
    LargeFile cFile;
    ErrVal nRet = cFile.open( m_acBitstreams[ui], LargeFile::OM_READONLY );
    if( Err::m_nOK != nRet )
    {
      std::cout << "Error while opening: " << m_acBitstreams[ui] << std::endl;
      return nRet;
    }
    PacketParserStartCode<LargeFile> cParser( cFile );

    UInt uiFileDecCount = 0;

    RNOK( m_pcVideoDecoderInstance->getVideoDecoderIf()->prepareForNewStream() );
    while( Err::m_nOK == cParser.readNextPacket() )
    {
      UInt uiPlanePackages;
      AllocVideoCodecPicture *pcDecodedPicture = new AllocVideoCodecPicture;
      ErrVal nRet = m_pcVideoDecoderInstance->getVideoDecoderIf()->getPictureStats( cParser.getVector(), *pcDecodedPicture, uiPlanePackages );

      if( Err::m_nSizeNotAvailable != nRet )
      {
        RNOK( nRet );

        std::vector<AllocVideoCodecPicture*> apcAllocOverlayPicture;
        std::vector<VideoCodecPicture<VideoCodecPel>*> apcOverlayPicture;
        if( m_cOverlayFilename != "none" )
        {
          for ( UInt uiPlaPaId = 0; uiPlaPaId < uiPlanePackages; uiPlaPaId++ )
          {
            if ( m_apcOverlay.size() < uiPlanePackages )
            {
              m_apcOverlay.push_back( new LargeFile );
              std::ostringstream cOverlayFilename;
              cOverlayFilename << m_cOverlayFilename << "." << uiPlaPaId << ".yuv";
              if( Err::m_nOK != m_apcOverlay[uiPlaPaId]->open( cOverlayFilename.str(), LargeFile::OM_WRITEONLY ) )
              {
                std::cout << "Could not open decoded file with informative overlay!" << std::endl;
                return Err::m_nERR;
              }
              m_pcVideoDecoderInstance->getVideoDecoderIf()->createOverlaidDecToDispBuffer();
              apcOverlayWriter.push_back( new YuvPictureWriter<VideoCodecPel>( *(m_apcOverlay[uiPlaPaId]) ) );
            }
            apcAllocOverlayPicture.push_back( new AllocVideoCodecPicture );
            apcAllocOverlayPicture[uiPlaPaId]->m_acPlanes = std::vector< VideoCodecPlane<VideoCodecPel> >( 3, pcDecodedPicture->m_acPlanes[0] );
            RNOK( apcAllocOverlayPicture[uiPlaPaId]->allocateBuffer() );
            apcOverlayPicture.push_back( apcAllocOverlayPicture[uiPlaPaId] );
          }
        }

        RNOK( pcDecodedPicture->allocateBuffer() );
        RNOK( m_pcVideoDecoderInstance->getVideoDecoderIf()->decodePicture( cParser.getVector(),
                                                                            pcDecodedPicture,
                                                                            cDecodedWriter,
                                                                            apcOverlayPicture,
                                                                            apcOverlayWriter ) );
        uiDecCount++;
        uiFileDecCount++;
        pcDecodedPicture = NULL;
      }
      else
      {
        delete pcDecodedPicture;
      }
    }
    cFile.close();
  }

  cT2.stamp();

  std::cout << std::endl;
  std::cout << "Number of decoded frames: " << std::setw( 5 ) << uiDecCount << std::endl;
  std::cout << "Average frame rate:       " << std::fixed << std::setprecision( 3 ) << std::setw( 9 ) << (uiDecCount * 1000.0 ) / double( (cT2-cT1).getMSec() ) << " fps" << std::endl;
  return Err::m_nOK;
}
