/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoEncoderPch.h"

#include "VideoCodecToolsLib/YuvPictureWriter.h"
#include "VideoCodecToolsLib/YuvPictureReader.h"
#include "VideoCodecToolsLib/BitstreamSink.h"
#include "VideoCodecToolsLib/MSysTimeStamp.h"
#include "VideoEncoderApp.h"

#include <cmath>
#include <iomanip>

#define ERROR_MSG(x) (std::cerr << x)



VideoEncoderApp::VideoEncoderApp() 
  : m_pcVideoEncoderInstance( NULL )
  , m_pcVideoEncoderIf      ( NULL )
  , m_pcEncoderParameter    ( NULL )
{
}


VideoEncoderApp::~VideoEncoderApp()
{
}


ErrVal VideoEncoderApp::runAlternative( int argc, char **argv, bool &rbExit )
{
  RNOK( VideoEncoderInstance::runAlternative( argc, argv, rbExit ) );
  return Err::m_nOK;
}


ErrVal VideoEncoderApp::create( VideoEncoderApp*& rpcVideoEncoderApp )
{
  rpcVideoEncoderApp = new VideoEncoderApp;

  return Err::m_nOK;
}


ErrVal VideoEncoderApp::xPrintSummary( UInt64 ui64RunTime, size_t uiTotalBytes, UInt uiFrame )
{
  std::vector<double> adPSNR;
  m_pcVideoEncoderIf->getPSNR( adPSNR );

  std::cout << std::endl;
  std::cout << "Average values:    Bytes/Pic      PSNR" << std::endl;
  std::cout << "======================================" << std::endl;
  std::cout << std::setw( 27 ) << uiTotalBytes/uiFrame;
  for( UInt ui = 0; ui < adPSNR.size(); ui++ )
  {
    std::cout << std::fixed << std::setprecision( 4 ) << std::setw( 11 ) << (adPSNR[ui] > 1000.0 ? 0.0 : adPSNR[ui]);
  }
  std::cout << std::endl << std::endl;

  std::cout << "Total bytes:                " << uiTotalBytes << std::endl;
  std::cout << "Wall clock time:            " << std::setprecision( 2 ) << std::fixed << double( ui64RunTime )                                           / 1000.0 << std::endl;
  std::cout << "CPU time:                   " << std::setprecision( 2 ) << std::fixed << double( m_pcVideoEncoderIf->getFrameEncodingDurationSumMSec() ) / 1000.0 << std::endl;
  std::cout << "Motion estimation CPU Time: " << std::setprecision( 2 ) << std::fixed << double( m_pcVideoEncoderIf->getFrameMEDurationSumMSec() )       / 1000.0 << std::endl;
  std::cout << "CPU time / wall clock Time: "  << std::setprecision( 4 ) << std::fixed << double( m_pcVideoEncoderIf->getFrameEncodingDurationSumMSec() ) / double( ui64RunTime ) << std::endl;
  std::cout << "Number of encoded frames:   " << uiFrame << std::endl;
//  std::cout << "fps: " << std::setprecision( 6 ) << std::fixed << ((double)uiFrame * 1000) / (UInt)ui64RunTime << std::endl << std::endl;

  return Err::m_nOK;
}


ErrVal VideoEncoderApp::destroy()
{
  //set the interface NULL
  m_pcVideoEncoderIf = NULL;

  // check each object before calling destroy
  if( NULL != m_pcVideoEncoderInstance )
  {
    RNOK( m_pcVideoEncoderInstance->destroy() );
    m_pcVideoEncoderInstance = NULL;
  }

  if( m_cBitstreamFile.is_open() )
  {
    RNOK( m_cBitstreamFile.close() );
  }

  RNOK( m_pcEncoderParameter->destroy());

  delete this;
  return Err::m_nOK;
}


ErrVal VideoEncoderApp::init( int argc, char**argv )
{
  // create Encoder parameter 
  RNOK( VideoEncoderParameter::create(m_pcEncoderParameter ) );
  RNOKRS( m_pcEncoderParameter->initCommandLineArgs( argc, argv ), -3);

  const VideoEncoderParameter::GeneralParams& rcGeneralParams = m_pcEncoderParameter->getGeneralParams();

  if( rcGeneralParams.getInputName().empty() || rcGeneralParams.getBitstreamName().empty() )
  {
    ERROR_MSG( "no input or bitstream file " );
    return -3;
  }

  if( Err::m_nOK != m_cBitstreamFile.open( rcGeneralParams.getBitstreamName(), LargeFile::OM_WRITEONLY ) )
  {
    std::cerr << "Failed to create output bitstream " << rcGeneralParams.getBitstreamName() << std::endl;
    return Err::m_nERR;
  }

  
  RNOK( VideoEncoderInstance::create( m_pcVideoEncoderInstance ) );

  m_pcVideoEncoderIf = m_pcVideoEncoderInstance->getVideoEncoderIf();

  RNOKS( m_pcVideoEncoderIf->setEncoderParameter( m_pcEncoderParameter ) );

  return Err::m_nOK;
} 


ErrVal VideoEncoderApp::go()
{
  TimeStamp cT1;
  TimeStamp cT2; 

  const VideoEncoderParameter::GeneralParams& rcGeneralParams = m_pcEncoderParameter->getGeneralParams();
  const UInt uiWidth = m_pcEncoderParameter->getGeneralParams().getInputPicSizeX();
  const UInt uiHeight = m_pcEncoderParameter->getGeneralParams().getInputPicSizeY();
  const UInt uiChromaFormatIdc = m_pcEncoderParameter->getGeneralParams().getChromaFormatIdc();

  LargeFile cYuvOutputFile, cYuvInputFile, cYuvPredictionFile, cYuvResidualFile;
  if( Err::m_nOK != cYuvOutputFile.open( m_pcEncoderParameter->getGeneralParams().getOutputName(), LargeFile::OM_WRITEONLY ) )
  {
    std::cerr << "failed to open yuv output file" << std::endl;
    return Err::m_nERR;
  }
  YuvPictureWriter<VideoCodecPel> cYuvWriter( cYuvOutputFile );
  if( Err::m_nOK != cYuvInputFile.open( m_pcEncoderParameter->getGeneralParams().getInputName(), LargeFile::OM_READONLY ) )
  {
    std::cerr << "failed to open yuv input file" << std::endl;
    return Err::m_nERR;
  }
  YuvPictureWriter<VideoCodecPel>* pcYuvRecPreWriter = NULL;
  if ( m_pcEncoderParameter->getGeneralParams().getPredictionName() != "none" )
  {
    if( Err::m_nOK != cYuvPredictionFile.open( m_pcEncoderParameter->getGeneralParams().getPredictionName(), LargeFile::OM_WRITEONLY ) )
    {
      std::cerr << "failed to open yuv prediction file" << std::endl;
      return Err::m_nERR;
    }
    pcYuvRecPreWriter = new YuvPictureWriter<VideoCodecPel>( cYuvPredictionFile );
  }
  YuvPictureWriter<VideoCodecPel>* pcYuvRecResWriter = NULL;
  if ( m_pcEncoderParameter->getGeneralParams().getResidualName() != "none" )
  {
    if( Err::m_nOK != cYuvResidualFile.open( m_pcEncoderParameter->getGeneralParams().getResidualName(), LargeFile::OM_WRITEONLY ) )
    {
      std::cerr << "failed to open yuv residual file" << std::endl;
      return Err::m_nERR;
    }
    pcYuvRecResWriter = new YuvPictureWriter<VideoCodecPel>( cYuvResidualFile );
  }

  YuvPictureReader cYuvReader( cYuvInputFile );

  // get the start time
  cT1.stamp();

  const UInt uiMaxFrame = rcGeneralParams.getTotalInputFrames();
  std::cout << "FrameNum  Qp  Mode      Bytes   Time(s) MeTime(s)      PSNR" << std::endl;
  std::cout << "===========================================================" << std::endl;

  BitstreamSinkStartCode cBitstreamSink( m_cBitstreamFile );

  UInt uiFrame;
  for( uiFrame = 0; uiFrame < uiMaxFrame; uiFrame++)
  {
    AllocVideoCodecPicture<VideoCodecPel> *pcNonConstOriginalPicture = xAllocateVideoCodecPicture<VideoCodecPel>( uiWidth, uiHeight, uiChromaFormatIdc );
    VideoCodecPicture<VideoCodecPel>      *pcReconstructedPicture    = xAllocateVideoCodecPicture<VideoCodecPel>( uiWidth, uiHeight, uiChromaFormatIdc );
    VideoCodecPicture<VideoCodecPel>      *pcReconstructedPrePicture = pcYuvRecPreWriter ? xAllocateVideoCodecPicture<VideoCodecPel>( uiWidth, uiHeight, uiChromaFormatIdc ) : NULL;
    VideoCodecPicture<VideoCodecPel>      *pcReconstructedResPicture = pcYuvRecResWriter ? xAllocateVideoCodecPicture<VideoCodecPel>( uiWidth, uiHeight, uiChromaFormatIdc ) : NULL;

    ErrVal nRet;
    if( Err::m_nOK != (nRet = cYuvReader.read( *pcNonConstOriginalPicture )) )
    {
      std::cerr << "Could not read from input file." << std::endl;
      return nRet;
    }

    AllocVideoCodecPicture<const VideoCodecPel> *pcOriginalPicture = pcNonConstOriginalPicture->getConstCopy();
    pcOriginalPicture->pushBackPointerToDelete( pcNonConstOriginalPicture );
    pcNonConstOriginalPicture = NULL;
    
    PictureSink<const VideoCodecPel> cReleaseOriginalPictureSink;

    RNOK( m_pcVideoEncoderIf->encodePicture( pcOriginalPicture,
                                             pcReconstructedPicture,
                                             pcReconstructedPrePicture,
                                             pcReconstructedResPicture,
                                             cReleaseOriginalPictureSink,
                                             cYuvWriter,
                                             pcYuvRecPreWriter,
                                             pcYuvRecResWriter,
                                             cBitstreamSink ) );

    if( uiFrame == uiMaxFrame - 1 )
    {
      // flush the encoder
      RNOK( m_pcVideoEncoderIf->encodePicture( NULL,
                                               NULL,
                                               NULL,
                                               NULL,
                                               cReleaseOriginalPictureSink,
                                               cYuvWriter,
                                               pcYuvRecPreWriter,
                                               pcYuvRecResWriter,
                                               cBitstreamSink ) );
    }
  }

  // get the stop time
  cT2.stamp();

  RNOK( xPrintSummary( ( cT2 - cT1 ).getMSec(), cBitstreamSink.getTotalNumBytes(), uiFrame ) );
  delete pcYuvRecPreWriter;
  delete pcYuvRecResWriter;
  return Err::m_nOK;
}
