/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef __VIDEOENCODERAPP_H_D65BE9B4_A8DA_11D3_AFE7_005004464B79
#define __VIDEOENCODERAPP_H_D65BE9B4_A8DA_11D3_AFE7_005004464B79

#include "VideoCodecToolsLib/MSysLargeFile.h"
#include "VideoEncoderLib/VideoEncoderInstance.h"
#include "VideoEncoderLib/VideoEncoderParameter.h"
#include "VideoEncoderLib/VideoEncoderIf.h"


class AutoDeleteOnDestruction
{
public:
  AutoDeleteOnDestruction() {}

  template<class T> void pushBackPointerToDelete( const T *p ) { m_apPointersToDelete.push_back( new DeletePtrClass<T>( p, false ) ); }
  template<class T> void pushBackArrayPointerToDelete( const T *p ) { m_apPointersToDelete.push_back( new DeletePtrClass<T>( p, true ) ); }

  virtual ~AutoDeleteOnDestruction()
  {
    while( !m_apPointersToDelete.empty() )
    {
      delete m_apPointersToDelete.back();
      m_apPointersToDelete.pop_back();
    }
  }

private:
  AutoDeleteOnDestruction( const AutoDeleteOnDestruction &r );
  const AutoDeleteOnDestruction& operator=( const AutoDeleteOnDestruction &r );

  class BaseClass { public: virtual ~BaseClass() {} };

  template<class T>
  class DeletePtrClass : public BaseClass
  {
  public:
    DeletePtrClass( const T *p, bool bArray ) : m_pT( p ), m_bArray( bArray ) {}
    virtual ~DeletePtrClass()
    {
      if( m_bArray )
      {
        delete[] m_pT;
      }
      else
      {
        delete m_pT;
      }
    }

  private:
    const T * const m_pT;
    const bool m_bArray;
  };

  std::vector<const BaseClass*> m_apPointersToDelete;
};


template<class t_Pel>
class AllocVideoCodecPicture
  : public VideoCodecPicture<t_Pel>
  , public AutoDeleteOnDestruction
{
public:
  AllocVideoCodecPicture<const t_Pel>* getConstCopy() const
  {
    AllocVideoCodecPicture<const t_Pel> *pcCopy = new AllocVideoCodecPicture<const t_Pel>;
    pcCopy->m_acPlanes.resize( VideoCodecPicture<t_Pel>::m_acPlanes.size() );
    for( UInt ui = 0; ui < VideoCodecPicture<t_Pel>::m_acPlanes.size(); ui++ )
    {
      pcCopy->m_acPlanes[ui].m_pData = VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_pData;
      pcCopy->m_acPlanes[ui].m_iWidth = VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_iWidth;
      pcCopy->m_acPlanes[ui].m_iHeight = VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_iHeight;
      pcCopy->m_acPlanes[ui].m_uiStride = VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_uiStride;
    }
    return pcCopy;
  }
  
  ErrVal allocateBuffer()
  {
    for( UInt ui = 0; ui < VideoCodecPicture<t_Pel>::m_acPlanes.size(); ui++ )
    {
      ROT( VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_pData );
      ROF( VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_pData = new t_Pel[VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_iWidth * VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_iHeight] );
      pushBackArrayPointerToDelete( VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_pData );
      VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_uiStride = VideoCodecPicture<t_Pel>::m_acPlanes[ui].m_iWidth;
    }
    return Err::m_nOK;
  }
};


//!\ingroup VideoEncoderApp
//!Sample application that demonstrates the usage of the H264EncoderLib. 
class VideoEncoderApp
{
protected:
	VideoEncoderApp();
	virtual ~VideoEncoderApp();

public:
  static ErrVal runAlternative( int argc, char **argv, bool &rbExit );

  //! creates a new instance of VideoEncoderApp
  static ErrVal create( VideoEncoderApp*& rpcVideoEncoderApp );

  //! initializes the class 
  ErrVal init( int argc, char**argv );

  //! runs the encoder
  ErrVal go();

  //! destroys the VideoEncoderApp object
  ErrVal destroy();

private:

  template <class t_Pel>
  class PictureSink : public VideoCodecSinkIf<const VideoCodecPicture<t_Pel>*>
  {
  public:
    ErrVal receive( const VideoCodecPicture<t_Pel>* pcPicture )
    {
      delete pcPicture;
      return Err::m_nOK;
    }
  };

  //! prints a summary of encoded sequence
  ErrVal xPrintSummary( UInt64 ui64RunTime, size_t uiTotalBytes, UInt uiFrame );

  template<class t_Pel>
  static AllocVideoCodecPicture<t_Pel>* xAllocateVideoCodecPicture( UInt uiWidth, UInt uiHeight, UInt uiChromaFormatIdc )
  {
    AllocVideoCodecPicture<t_Pel>* pcPicture = new AllocVideoCodecPicture<t_Pel>;
    pcPicture->m_acPlanes.push_back( VideoCodecPlane<t_Pel>( NULL, uiWidth, uiHeight, 0 ) );
    if( uiChromaFormatIdc == 1 )
    {
      pcPicture->m_acPlanes.push_back( VideoCodecPlane<t_Pel>( NULL, uiWidth>>1, uiHeight>>1, 0 ) );
      pcPicture->m_acPlanes.push_back( VideoCodecPlane<t_Pel>( NULL, uiWidth>>1, uiHeight>>1, 0 ) );
    }
    else if( uiChromaFormatIdc > 1 )
    {
      AOT( 1 );
    }
    pcPicture->allocateBuffer();
    return pcPicture;
  }


protected:
  VideoEncoderInstance*    m_pcVideoEncoderInstance;      //!< pointer to encoder creator class
  VideoEncoderIf*          m_pcVideoEncoderIf;            //!< pointer to encoder interface
  VideoEncoderParameter*   m_pcEncoderParameter;          //!< pointer to encoder parameters
  LargeFile m_cBitstreamFile;
};

#endif //__VIDEOENCODERAPP_H_D65BE9B4_A8DA_11D3_AFE7_005004464B79
