/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCodecToolsLibPch.h"
#include "VideoCodecToolsLib/MSysLargeFile.h"

#include <errno.h>

#if defined( MSYS_WIN32 )

# include <io.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <share.h>

#else

# include <unistd.h>

#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MSysLargeFile::MSysLargeFile() :
  m_iFileHandle( -1 )
{
}

MSysLargeFile::~MSysLargeFile()
{
  if( -1 != m_iFileHandle )
  {
#if defined( MSYS_WIN32 )
    ::_close( m_iFileHandle );
#elif defined( MSYS_LINUX )
    ::close( m_iFileHandle );
#endif
  }
}

ErrVal MSysLargeFile::open( const std::string& rcFilename, enum OpenMode eOpenMode, int iPermMode )
{
  ROT( rcFilename.empty() );
  ROF( -1 == m_iFileHandle );

  int iOpenMode;

#if defined( MSYS_WIN32 )
  int iShareMode = _SH_DENYWR;

  //The UNIX style iPermMode parameter is only meaningful to provide source code compatibility but cannot 
  //be used to configure the file protection. Furthermore, the secure VC8.x CRT asserts even in read only mode if 
  //an unsupported permission mode is specified.
  if( eOpenMode == OM_READONLY )
  {
    iOpenMode = _O_RDONLY;
    iPermMode = _S_IREAD;
  }
  else if( eOpenMode == OM_WRITEONLY )
  {
    iOpenMode = _O_CREAT | _O_TRUNC | _O_WRONLY;
    iPermMode = _S_IWRITE;
  }
  else if( eOpenMode == OM_APPEND )
  {
    //append mode does not imply write access and the create flag
    //simplifies the program's logic (in most cases).
    iOpenMode = _O_APPEND | _O_CREAT | _O_WRONLY;

    iPermMode = _S_IWRITE;
  }
  else if( eOpenMode == OM_READWRITE )
  {
    iOpenMode = _O_CREAT | _O_RDWR;

    iPermMode = _S_IREAD | _S_IWRITE;
  }
  else
  {
    AOT( 1 );
    return Err::m_nERR;
  }

  iOpenMode |= _O_SEQUENTIAL | _O_BINARY;

  // m_iFileHandle = ::_open( rcFilename.c_str(), iOpenMode, iPermMode );
#if( _MSC_VER >= 1400 )
  if( 0 !=  ::_sopen_s( &m_iFileHandle, rcFilename.c_str(), iOpenMode, iShareMode, iPermMode ) )  { m_iFileHandle = -1; }
#else
  m_iFileHandle = ::_sopen( rcFilename.c_str(), iOpenMode, iShareMode, iPermMode );
#endif
  
#elif defined( MSYS_LINUX )

  if( eOpenMode == OM_READONLY )
  {
    iOpenMode = O_RDONLY;
  }
  else if( eOpenMode == OM_WRITEONLY )
  {
    iOpenMode = O_CREAT | O_TRUNC | O_WRONLY;
  }
  else if( eOpenMode == OM_APPEND )
  {
    iOpenMode = O_APPEND | O_CREAT | O_WRONLY;
  }
  else if( eOpenMode == OM_READWRITE )
  {
    iOpenMode = O_CREAT | O_RDWR;
  }
  else
  {
    AOT( 1 );
    return Err::m_nERR;
  }

  m_iFileHandle = open64( rcFilename.c_str(), iOpenMode, iPermMode );

#endif

  // check if file is really open
  ROTS( -1 == m_iFileHandle );

  // and return
  return Err::m_nOK;
}

ErrVal MSysLargeFile::close()
{
  int iRetv;

  ROTS( -1 == m_iFileHandle );
  
#if defined( MSYS_WIN32 )
  iRetv = ::_close( m_iFileHandle );
#elif defined( MSYS_LINUX )
  iRetv = ::close( m_iFileHandle );
#endif

  m_iFileHandle = -1;

  return ( iRetv == 0 ) ? Err::m_nOK : Err::m_nERR;
}

ErrVal MSysLargeFile::seek( Int64 iOffset, int iOrigin )
{ 
  Int64 iNewOffset;
  ROT( -1 == m_iFileHandle );

#if defined( MSYS_WIN32 )
  iNewOffset = _lseeki64( m_iFileHandle, iOffset, iOrigin );
#elif defined( MSYS_LINUX )
  iNewOffset = lseek64( m_iFileHandle, iOffset, iOrigin );
#endif

  return (( iNewOffset == -1 ) ? (Err::m_nERR) : (Err::m_nOK));
}

Int64 MSysLargeFile::tell()
{
  ROTR( -1 == m_iFileHandle, -1 );
  Int64 iOffset;
#if defined( MSYS_WIN32 )
  iOffset = _telli64( m_iFileHandle );
#elif defined( MSYS_LINUX )
  iOffset = lseek64( m_iFileHandle, 0, SEEK_CUR );
#endif
  ROT( iOffset == -1 );

  // and return
  return iOffset;
}


// the following read method tries to read uiCount bytes
// if end of file is reached ruiBytesRead is set to the number of bytes read
// to get EOF return value the method has to be called again
ErrVal MSysLargeFile::read( void *pvBuffer, UInt uiCount, UInt& ruiBytesRead )
{
  int iRetv;

  // paranoia checks
  ROT( -1 == m_iFileHandle );
  ROT( 0 == uiCount );
  
  // init
  ruiBytesRead = 0;

  while( 0 != uiCount )
  {
    // try to read from file
#if defined( MSYS_WIN32 )
    iRetv = ::_read( m_iFileHandle, (void*)((char*)pvBuffer + ruiBytesRead), uiCount );
#elif defined( MSYS_LINUX )
    iRetv = ::read( m_iFileHandle, (void*)((char*)pvBuffer + ruiBytesRead), uiCount );
#endif

    // return errno if error occurred
    ROTRS( -1 == iRetv, errno );

    // check if we reached end of file
    if( 0 == iRetv )
    {
      // return EOF if we read nothing so far
      ROTRS( 0 == ruiBytesRead, Err::m_nEndOfFile );

      // if we already read something -> return 
      return Err::m_nOK;
    }

    // unknown return value -> fire assertion
    ROT( -1 > iRetv );

    // calculate the needed values
    uiCount -= iRetv;
    ruiBytesRead += iRetv;
  }

  // and return
  return Err::m_nOK;
}

ErrVal MSysLargeFile::read( void *pvBuffer, UInt uiCount )
{
  UInt uiBytesRead;
  return read( pvBuffer, uiCount, uiBytesRead );
}


ErrVal MSysLargeFile::write( const void *pvBuffer, UInt uiCount )
{
  int iRetv;

  ROT( -1 == m_iFileHandle );
  ROT( 0 == uiCount );

#if defined( MSYS_WIN32 )
  iRetv = ::_write( m_iFileHandle, pvBuffer, uiCount );
#elif defined( MSYS_LINUX )
  iRetv = ::write( m_iFileHandle, pvBuffer, uiCount );
#endif
  ROF( iRetv == (int)uiCount );

  return Err::m_nOK;
}
