/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCodecToolsLibPch.h"
#include "VideoCodecToolsLib/ToolLibInitOptionParser.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ToolLibInitOptionParser::ToolLibInitOptionParser()
: m_bInit (false)
{
}

ToolLibInitOptionParser::~ToolLibInitOptionParser()
{
}


ErrVal ToolLibInitOptionParser::init( const std::string& cInputString, int iErrId )
{
  ROT(m_bInit);

  RNOK(ToolLibParser::init( cInputString, iErrId ));

  RNOK(xReadOptions());

  m_bInit = true;

  return Err::m_nOK;

}

ErrVal ToolLibInitOptionParser::uninit( bool printWarning )
{
  ROF(m_bInit);

  if (printWarning)
  {
    xPrintWarningUnusedOptions();
  }

  m_cOptionTokenMap.clear();

  RNOK(ToolLibParser::uninit());

  m_bInit = false;

  return Err::m_nOK;

}

ErrVal ToolLibInitOptionParser::xReadOptions()
{
  std::string cOption;
  std::string cValue;

  while( Err::m_nEndOfFile != getNextLine(false))
  {
    if( Err::m_nOK != getToken( cOption ) )
    {
      msgWarning( "unexpected line in config file. ignoring line" );
    }
    else if( Err::m_nOK != getExpected( "=" ) )
    {
      msgWarning( "ignoring line" );
    }
    else if (Err::m_nOK != getString(cValue))
    {
      msgWarning( "unexpected value after '=' in config file. ignoring line" );
    }
    else
    {
      if(getIgnoreCase())
      {
        stringToLowerCase(cOption);
      }

      if (m_cOptionTokenMap.end()!=m_cOptionTokenMap.find(cOption))
      {
        msgWarning( "double definition of " + cOption + ". Using latest value." );
      }
      m_cOptionTokenMap[cOption] = cValue;
    }
  }
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::getValue (const std::string& cOption, std::string& cValue, bool bWarnIfNotAvail)
{
  std::string cVarOption;

  cVarOption = cOption;

  if (getIgnoreCase())
  {
     stringToLowerCase(cVarOption);
  }

  OptionTokenMap::iterator cToken = m_cOptionTokenMap.find(cVarOption);

  if (m_cOptionTokenMap.end()==cToken)
  {
    if (bWarnIfNotAvail)
    {
      msgError("Option " + cOption + " required");
    }
    return Err::m_nERR;
  }

  cValue = m_cOptionTokenMap[cVarOption];

  m_cOptionTokenMap.erase(cToken);

  return Err::m_nOK;
}

ErrVal ToolLibInitOptionParser::getValue( const std::string& cOption, bool&   rVal, bool bWarnIfNotAvail)
{
  std::string cValue;
  UInt uiValue;

  RNOKS(getValue(cOption, cValue, bWarnIfNotAvail));

  RNOK(stringToValue(cValue, uiValue));

  rVal=(uiValue>0)?true:false;
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::getValue( const std::string& cOption, char&   rVal, bool bWarnIfNotAvail )
{
  std::string cValue;
  RNOKS(getValue(cOption, cValue, bWarnIfNotAvail));

  RNOK(stringToValue(cValue, rVal));
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::getValue( const std::string& cOption, short&  rVal, bool bWarnIfNotAvail )
{
  std::string cValue;
  RNOKS(getValue(cOption, cValue, bWarnIfNotAvail));

  RNOK(stringToValue(cValue, rVal));
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::getValue( const std::string& cOption, int&  rVal, bool bWarnIfNotAvail )
{
  std::string cValue;
  RNOKS(getValue(cOption, cValue, bWarnIfNotAvail));

  RNOK(stringToValue(cValue, rVal));
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::getValue( const std::string& cOption, Int64&  rVal, bool bWarnIfNotAvail )
{
  std::string cValue;
  RNOKS(getValue(cOption, cValue, bWarnIfNotAvail));

  RNOK(stringToValue(cValue, rVal));
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::getValue( const std::string& cOption, UChar&  rVal, bool bWarnIfNotAvail )
{
  std::string cValue;
  RNOKS(getValue(cOption, cValue, bWarnIfNotAvail));

  RNOK(stringToValue(cValue, rVal));
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::getValue( const std::string& cOption, UShort& rVal, bool bWarnIfNotAvail )
{
  std::string cValue;
  RNOKS(getValue(cOption, cValue, bWarnIfNotAvail));

  RNOK(stringToValue(cValue, rVal));
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::getValue( const std::string& cOption, UInt& rVal, bool bWarnIfNotAvail )
{
  std::string cValue;
  RNOKS(getValue(cOption, cValue, bWarnIfNotAvail));

  RNOK(stringToValue(cValue, rVal));
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::getValue( const std::string& cOption, UInt64& rVal, bool bWarnIfNotAvail )
{
  std::string cValue;
  RNOKS(getValue(cOption, cValue, bWarnIfNotAvail));

  RNOK(stringToValue(cValue, rVal));
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::getValue( const std::string& cOption, float& rVal, bool bWarnIfNotAvail )
{
  std::string cValue;
  RNOKS(getValue(cOption, cValue, bWarnIfNotAvail));

  RNOK(stringToValue(cValue, rVal));
  return Err::m_nOK;
}


ErrVal ToolLibInitOptionParser::xPrintWarningUnusedOptions ()
{
  OptionTokenMap::iterator cToken;

  if (!m_cOptionTokenMap.empty())
  {
    std::cerr<<"Warning: the following Options were ignored:" << std::endl;

      for (cToken = m_cOptionTokenMap.begin();cToken != m_cOptionTokenMap.end();cToken++)
      {
        std::cerr<<"  "<< cToken->first << std::endl;
      }
  }

  return Err::m_nOK;
}

