/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCodecToolsLibPch.h"
#include "VideoCodecToolsLib/ToolLibParser.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#if defined( MSYS_WIN32 )
# define atoi64 _atoi64
#elif defined( MSYS_LINUX )
# define atoi64 atoll
#endif

ToolLibParser::ToolLibParser()
: m_cLine        ("")
, m_cToken       ("")
, m_uiLinePos    (0)
, m_bInit        (false)
, m_bIgnoreCase  (false)
, m_iErrId       (0)
{
}

ToolLibParser::~ToolLibParser()
{
  AOT( true == m_bInit)
}

ErrVal ToolLibParser::init( const std::string& cInputString, int iErrId )
{
  AOT( true == m_bInit );
  ROT( 0 == cInputString.size() );

  RNOK( m_cStringBuffer.init( cInputString ) );

  m_cLine        = "";
  m_cToken       = "";
  m_uiLinePos    = 0;
  m_iErrId       = iErrId;

  m_bInit = true;

  return Err::m_nOK;
}

ErrVal ToolLibParser::uninit()
{
  AOT( false == m_bInit );

  ANOK( m_cStringBuffer.uninit() );

  m_cLine        = "";
  m_cToken       = "";
  m_uiLinePos    = 0;
  m_bIgnoreCase  = false;

  m_bInit = false;

  return Err::m_nOK;
}

void ToolLibParser::msgFatal( const std::string& rcMessage )
{
  std::cerr << "Fatal error in line: " << m_cStringBuffer.getLineNumber() << " " << rcMessage << std::endl;
}

void ToolLibParser::msgError( const std::string& rcMessage )
{
  std::cerr << "Error in line: " << m_cStringBuffer.getLineNumber() << " " << rcMessage << std::endl;
}

void ToolLibParser::msgWarning( const std::string& rcMessage )
{
  std::cerr << "Warning in line: " << m_cStringBuffer.getLineNumber() << " " << rcMessage << std::endl;
}

void ToolLibParser::msgVerbose( const std::string& rcMessage )
{
  std::cerr << "Info in line: " << m_cStringBuffer.getLineNumber() << " " << rcMessage << std::endl;
}

void ToolLibParser::stringToLowerCase( std::string& rcString )
{
  UInt i = 0;

  while( rcString.length() > i )
  {
    rcString[i] = tolower( rcString[i] );
    i++;
  }
}

ErrVal ToolLibParser::stringToValue( const std::string& rcString, char& rVal )
{
  int nVal;

  RNOKS( stringToValue( rcString, nVal ) );

  rVal = (char) nVal;

  return Err::m_nOK;
}

ErrVal ToolLibParser::stringToValue( const std::string& rcString, short& rVal )
{
  int nVal;

  RNOKS( stringToValue( rcString, nVal ) );

  rVal = (short) nVal;

  return Err::m_nOK;
}

ErrVal ToolLibParser::stringToValue( const std::string& rcString, int& rVal )
{
  rVal = strtol( rcString.c_str(), NULL, 0 );

  // returns 0 in case of non-convertible string check whether the string was really 0
  // if not 0 the string contained some characters that prevented conversion
  if( 0 == rVal)
  {
    if( ("0" != rcString ) && ("0x0" != rcString) && ("0X0" != rcString) )
    {
      return Err::m_nERR;
    }
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::stringToValue( const std::string& rcString, UChar& rVal )
{
  UInt nVal;

  RNOKS( stringToValue( rcString, nVal ) );

  if( nVal > 0xFF )
  {
    return Err::m_nERR;
  }

  rVal = (UChar) nVal;

  return Err::m_nOK;
}

ErrVal ToolLibParser::stringToValue( const std::string& rcString, UShort& rVal )
{
  UInt nVal;

  RNOKS( stringToValue( rcString, nVal ) );

  if( nVal > 0xFFFF )
  {
    return Err::m_nERR;
  }

  rVal = (UShort) nVal;

  return Err::m_nOK;
}

ErrVal ToolLibParser::stringToValue( const std::string& rcString, UInt& rVal )
{
  rVal = strtoul( rcString.c_str(), NULL, 0 );

  if( '-' == rcString.c_str()[0] )
  {
    return Err::m_nERR;
  }

  // returns 0 in case of non-convertible string check whether the string was really 0
  // if not 0 the string contained some characters that prevented conversion
  if( 0 == rVal)
  {
    if( ("0" != rcString ) && ("0x0" != rcString) && ("0X0" != rcString) )
    {
      return Err::m_nERR;
    }
  }

  return Err::m_nOK;
}

// 64 bit conversion is handled separate. Unfortunately it does not support hex numbers
ErrVal ToolLibParser::stringToValue( const std::string& rcString, UInt64& rVal )
{
  rVal = atoi64( rcString.c_str() );

  if( '-' == rcString.c_str()[0] )
  {
    return Err::m_nERR;
  }

  // returns 0 in case of non-convertible string check whether the string was really 0
  // if not 0 the string contained some characters that prevented conversion
  if( 0 == rVal)
  {
    if( "0" != rcString )
    {
      return Err::m_nERR;
    }
  }
  return Err::m_nOK;
}

ErrVal ToolLibParser::stringToValue( const std::string& rcString, Int64& rVal )
{
  rVal = atoi64( rcString.c_str() );

  // returns 0 in case of non-convertible string check whether the string was really 0
  // if not 0 the string contained some characters that prevented conversion
  if( 0 == rVal)
  {
    if( "0" != rcString )
    {
      return Err::m_nERR;
    }
  }
  return Err::m_nOK;
}


ErrVal ToolLibParser::stringToValue( const std::string& rcString, float& rVal )
{
  std::string cToken = rcString;

  int  i = 0;
  bool b_NotANumber = false;

  while (cToken[i] != '\0')
  {
    if (isalpha(cToken[i]))
    {
      if (tolower(cToken[i]) != 'e' && tolower(cToken[i]) != 'd')
      {
        b_NotANumber = true;
      }
    }
    i++;
  }

  // returns 0 in case of non-convertible string check if it was really 0
  rVal = (float) atof( cToken.c_str() );

  if( 0 == rVal ) // perform some  plausibility check if zero: .0 or 0.0
  {
    if (cToken[0] == '-' || cToken[0] == '+')
    {
      i=1;
      while( cToken[i] != '\0' )
      {
        if ( !((i==1 && cToken[1] == '.') || (i==2 && cToken[2] == '.')) )
        {
          if( cToken[i] != '0' )
          {
            b_NotANumber = true;
          }
        }
        i++;
      }
      if( (i==1) || (i==3 && cToken[2] == '.') || (i==2 && cToken[1] == '.') )
      {
        b_NotANumber = true;
      }
    }
    else
    {
      i=0;
      while( cToken[i] != '\0' )
      {
        if( !( (i==0 && cToken[0] == '.') || (i==1 && cToken[1] == '.')) )
        {
          if( cToken[i] != '0' )
          {
            b_NotANumber = true;
          }
        }
        i++;
      }

      if( (i==0) || (i==2 && cToken[1] == '.') || (i==1 && cToken[0] == '.') )
      {
        b_NotANumber = true;
      }
    }
  }

  if( true == b_NotANumber)
  {
    return Err::m_nERR;
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::stringToValue( const std::string& rcString, double& rVal )
{
  std::string cToken = rcString;

  int i;
  bool b_NotANumber = false;

  i=0;
  while (m_cToken[i] != '\0')
  {
    if (isalpha(m_cToken[i]))
    {
      if (tolower(m_cToken[i]) != 'e' && tolower(m_cToken[i]) != 'd')
      {
        b_NotANumber = true;
      }
    }
    i++;
  }

  // returns 0 in case of non-convertible string check if it was really 0
  rVal = atof( m_cToken.c_str() );

  if( 0 == rVal ) // perform some  plausibility check if zero: .0 or 0.0
  {
    if (m_cToken[0] == '-' || m_cToken[0] == '+')
    {
      i=1;
      while( m_cToken[i] != '\0' )
      {
        if ( !((i==1 && m_cToken[1] == '.') || (i==2 && m_cToken[2] == '.')) )
        {
          if( m_cToken[i] != '0' )
          {
            b_NotANumber = true;
          }
        }
        i++;
      }
      if( (i==1) || (i==3 && m_cToken[2] == '.') || (i==2 && m_cToken[1] == '.') )
      {
        b_NotANumber = true;
      }
    }
    else
    {
      i=0;
      while( m_cToken[i] != '\0' )
      {
        if( !( (i==0 && m_cToken[0] == '.') || (i==1 && m_cToken[1] == '.')) )
        {
          if( m_cToken[i] != '0' )
          {
            b_NotANumber = true;
          }
        }
        i++;
      }

      if( (i==0) || (i==2 && m_cToken[1] == '.') || (i==1 && m_cToken[0] == '.') )
      {
        b_NotANumber = true;
      }
    }
  }

  if( true == b_NotANumber)
  {
    return Err::m_nERR;
  }

  return Err::m_nOK;
}

/*-------------------------------------------------------------------------------
Returns true when a blank line or a pure comment line is encountered.
Returns false when a normal line is encountered (may have comment at the end of line)
Comments at the end of a line are cut off by setting a '\0' there.
-------------------------------------------------------------------------------*/
bool ToolLibParser::deleteComments()
{
  // blank line
  if( true == m_cLine.empty() )
  {
    return true;
  }

  // check for line with blanks or tabs only
  size_t uiPos = 0;

  while( (m_cLine[uiPos] == ' ') || (m_cLine[uiPos] == '\t') )
  {
    uiPos++;
  }

  // only blanks and tabs
  if( m_cLine.length() == uiPos )
  {
    return true;
  }

  // pure comment line
  if( '#' == m_cLine[uiPos] )
  {
    return true;
  }

  // check for comment at end of line.
//  uiPos = m_cLine.find_first_of( "#" );
  bool bString = false;
  for( ; uiPos<m_cLine.length(); uiPos++ )
  {
    if ( '"' == m_cLine[uiPos] )
    {
      bString = !bString;
    }
    if ( '#' == m_cLine[uiPos] && (!bString) )
    {
      break;
    }

  }

  if( std::string::npos != uiPos )
  {
    // normal comment somewhere in the line used to terminate the string
    m_cLine.assign( (char*) m_cLine.c_str(), m_uiLinePos, uiPos - m_uiLinePos );
  }

  return false;
}


/*-------------------------------------------------------------------------------
Returns either okay or end of file
-------------------------------------------------------------------------------*/
ErrVal ToolLibParser::getNextLine( const bool bWithErrorMessage )
{
  ErrVal iRet;

  iRet = m_cStringBuffer.getLine( m_cLine );

  if( Err::m_nEndOfBuffer == iRet )
  {
    if( bWithErrorMessage )
    {
      msgWarning( "end of file encountered" );
    }

    m_uiLinePos = 0;

    // using end of file in the parser for file end
    return Err::m_nEndOfFile;
  }


  if( Err::m_nERR == iRet )
  {
    //msgWarning( "line end character differs from platform line-end type" );
    // ToolLibParser carries on working even with strange line end.
    iRet = Err::m_nOK;
  }


  // actually we will not return here just a paranoia
  RNOKR( iRet, iRet );

  // we have a new line. set the position in the line to 0
  m_uiLinePos = 0;

  // check if line is pure comment or whitespace or empty line
  if( true == deleteComments() )
  {
    iRet = getNextLine( bWithErrorMessage );
  }

  return iRet;
}


void ToolLibParser::moveOverWhiteSpace()
{
  while( (m_cLine[m_uiLinePos] == ' ')  ||
         (m_cLine[m_uiLinePos] == ',') ||
         (m_cLine[m_uiLinePos] == '\t') )
  {
    m_uiLinePos++; // skip over blanks and tabs in new line
  }
}

/*-------------------------------------------------------------------------------
Extracts a Token but leaves the position in the file as if the Token has not
been read. Any call of GetToken or TestToken after a Call of TestToken delivers
the same Token again.
TRUE when okay
FALSE for EOF
-------------------------------------------------------------------------------*/
ErrVal ToolLibParser::peekToken()
{
  ErrVal nRet = Err::m_nERR;

  if( m_cLine.length() == m_uiLinePos )
  {
    nRet = getNextLine( 0 );  // parameter 0 supresses EOF error message

    if( Err::m_nERR == nRet )
    {
      msgError( "unknown error while reading file" );
      return nRet;
    }

    if( Err::m_nEndOfFile == nRet )
    {
      return nRet;
    }
  }

  // skip over leading white space until first token in line
  moveOverWhiteSpace();

  size_t uiPos = 0;

  // find whitespace that separate the tokens
  uiPos = m_cLine.find_first_of( " \t\n{[()]},", m_uiLinePos );

  if( std::string::npos == uiPos )
  {
    uiPos = m_cLine.length();
  }

  // check for any kind of bracket which maybe without blanks as separators
  if( '{' == m_cLine[m_uiLinePos] || '[' == m_cLine[m_uiLinePos] || '(' == m_cLine[m_uiLinePos]  ||
      '}' == m_cLine[m_uiLinePos] || ']' == m_cLine[m_uiLinePos] || ')' == m_cLine[m_uiLinePos] )
  {
    // found a bracket only (not as a delimiter) put iPos behind the character for extraction
    if( uiPos == m_uiLinePos )
    {
      uiPos = m_uiLinePos + 1;
    }
  }

  ROTS( m_uiLinePos >= uiPos );

  m_cToken.assign( (char*) m_cLine.c_str(), m_uiLinePos, uiPos - m_uiLinePos );

  if( true == m_bIgnoreCase )
  {
    stringToLowerCase( m_cToken );
  }

  return Err::m_nOK;
}

/*-------------------------------------------------------------------------------
Gets next token to m_cToken and move behind it
-------------------------------------------------------------------------------*/
ErrVal ToolLibParser::getToken()
{
  ErrVal nRet;

  nRet = peekToken();

  if( Err::m_nEndOfFile == nRet )
  {
    msgError( "unexpected end of file encountered" );
    return nRet;
  }

  if( Err::m_nERR == nRet )
  {
    msgError( "getToken parsing error" );
    return nRet;
  }

  m_uiLinePos += m_cToken.length();

  // got a token move over whitespace to beginning of next token or end of line
  moveOverWhiteSpace();

  return Err::m_nOK;
}

/*-------------------------------------------------------------------------------
Peeks the next token ant puts it to m_cToken. Leaves position as is
-------------------------------------------------------------------------------*/
ErrVal ToolLibParser::peekToken( std::string& rcString )
{
  ErrVal nRet;

  rcString = "";

  RNOKRS( nRet = peekToken(), nRet );

  rcString = m_cToken;

  return Err::m_nOK;
}

/*-------------------------------------------------------------------------------
gets the next token and puts it into the string
-------------------------------------------------------------------------------*/
ErrVal ToolLibParser::getToken( std::string& rcString )
{
  ErrVal nRet;

  rcString = "";

  RNOKRS( nRet = getToken(), nRet );

  rcString = m_cToken;

  return Err::m_nOK;
}

ErrVal ToolLibParser::getString( std::string& rcString )
{
  ErrVal nRet = Err::m_nERR;

  rcString = "";

  if( m_cLine.length() == m_uiLinePos )
  {
    nRet = getNextLine( 0 ); // parameter 0 supresses EOF error message

    if( Err::m_nERR == nRet )
    {
      msgError( "unknown error while reading file" );
      return nRet;
    }

    if( Err::m_nEndOfFile == nRet )
    {
      return nRet;
    }
  }

  // skip over leading white space
  moveOverWhiteSpace();

  // special handling if the string is enclosed in ""
  if( '"' == m_cLine[m_uiLinePos] )
  {
    size_t uiPos = 0;
    // search for the next "
    uiPos = m_cLine.find_first_of( "\"", m_uiLinePos + 1 );

    if( std::string::npos == uiPos )
    {
      msgError( "missing string termination symbol \"" );

      uiPos = m_cLine.length();

      return Err::m_nERR;
    }

    ROT( m_uiLinePos >= uiPos );

    // assign string without quotes at start and end
    m_cToken.assign( (char*) m_cLine.c_str(), m_uiLinePos + 1, uiPos - m_uiLinePos - 1 );

    rcString = m_cToken;
    m_uiLinePos += m_cToken.length() + 2; // +2 for the quotes
  }
  else
  {
    RNOKS( getToken( rcString ) );
  }

  // got a string move over whitespace to beginning of next token or end of line
  moveOverWhiteSpace();

  if( true == m_bIgnoreCase )
  {
    stringToLowerCase( rcString );
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::getString( std::string& rcString, const char cOpenChar, const char cCloseChar )
{
  ErrVal nRet = Err::m_nERR;

  rcString = "";

  if( m_cLine.length() == m_uiLinePos )
  {
    nRet = getNextLine( 0 ); // parameter 0 supresses EOF error message

    if( Err::m_nERR == nRet )
    {
      msgError( "unknown error while reading file" );
      return nRet;
    }

    if( Err::m_nEndOfFile == nRet )
    {
      return nRet;
    }
  }

  // skip over leading white space
  moveOverWhiteSpace();

  size_t uiPosStart = m_uiLinePos;
  size_t uiCountStarts = 0;
  size_t uiCountEnd    = 0;
  size_t uiPos;
  for( uiPos = uiPosStart; uiPos < m_cLine.length(); uiPos++ )
  {
    if(m_cLine[uiPos] == cOpenChar)
    {
      uiCountStarts++;
    }
    else if(m_cLine[uiPos] == cCloseChar)
    {
      uiCountEnd++;
      if( uiCountEnd == uiCountStarts)
      {
        break;
      }
    }
  }

  if( std::string::npos == uiPos )
  {
    msgError( "missing string termination symbol \"" );

    uiPos = m_cLine.length();

    return Err::m_nERR;
  }

  ROT( m_uiLinePos >= uiPos );

  // assign string without quotes at start and end
  m_cToken.assign( (char*) m_cLine.c_str(), m_uiLinePos + 1, uiPos - m_uiLinePos - 1 );

  rcString = m_cToken;
  m_uiLinePos += m_cToken.length() + 2; // +2 for the quotes


  if( true == m_bIgnoreCase )
  {
    stringToLowerCase( rcString );
  }

  return Err::m_nOK;
}

/*-------------------------------------------------------------------------------
gets the next token and checks whether it is the expected one. error otherwise
-------------------------------------------------------------------------------*/
ErrVal ToolLibParser::getExpected( const std::string& rcExpectedToken )
{
  ErrVal iRet;

  std::string cOrigExpectedToken =   rcExpectedToken;
  std::string cCopiedExpectedToken = rcExpectedToken;

  if( true == m_bIgnoreCase )
  {
    stringToLowerCase( cCopiedExpectedToken );
  }

  std::string cToken;

  iRet = getToken( cToken );

  if( Err::m_nEndOfFile == iRet )
  {
    msgError( "end of file encountered. Expected: " + cOrigExpectedToken );
    return Err::m_nERR;
  }
  else if( cToken != cCopiedExpectedToken )
  {
    msgError( "unexpected token: " + cToken + " Expected: " + cOrigExpectedToken );
    return Err::m_nERR;
  }
  else if( Err::m_nOK != iRet )
  {
    msgError( "error occurred. Expected: " + cOrigExpectedToken );
    return Err::m_nERR;
  }

  return Err::m_nOK;
}


/*-------------------------------------------------------------------------------
gets the next token if it is the optional one. do nothing otherwise
-------------------------------------------------------------------------------*/

ErrVal ToolLibParser::getOptional( const std::string& rcExpectedToken )
{
  std::string cToken;
  std::string cCopiedExpectedToken = rcExpectedToken;

  RNOKS( peekToken( cToken ) );

  if( true == m_bIgnoreCase )
  {
    stringToLowerCase( cToken );
    stringToLowerCase( cCopiedExpectedToken );
  }

  if( cToken == cCopiedExpectedToken )
  {
    RNOKS( getToken() );
  }

  return Err::m_nOK;
}


ErrVal ToolLibParser::getToken( bool& rVal )
{
  ErrVal nRet;

  RNOKRS( nRet = getToken(), nRet );

  stringToLowerCase( m_cToken );

  if( ("0" == m_cToken) || ("false" == m_cToken) )
  {
    rVal = false;
  }
  else if( ("1" == m_cToken) || ("true" == m_cToken) )
  {
    rVal = true;
  }
  else
  {
    msgError( "expected one of TRUE|FALSE|true|false|1|0" );
    return Err::m_nERR;
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::getToken( char& rVal )
{
  ErrVal nRet;

  RNOKRS( nRet = getToken(), nRet );

  if( Err::m_nERR == stringToValue( m_cToken, rVal ) )
  {
    msgError( "char expected. Found: " + m_cToken );
    return Err::m_nERR;
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::getToken( short& rVal )
{
  RNOKS( getToken() );

  if( Err::m_nERR == stringToValue( m_cToken, rVal ) )
  {
    msgError( "short expected. Found: " +  m_cToken );
    return Err::m_nERR;
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::getToken( int& rVal )
{
  RNOKS( getToken() );

  if( Err::m_nERR == stringToValue( m_cToken, rVal ) )
  {
    msgError( "int expected. Found: " +  m_cToken );
    return Err::m_nERR;
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::getToken( Int64& rVal )
{
  ErrVal nRet;

  RNOKRS( nRet = getToken(), nRet );

  if( Err::m_nERR == stringToValue( m_cToken, rVal ) )
  {
    msgError( "Int64 expected in decimal format. Found: " +  m_cToken );
    return Err::m_nERR;
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::getToken( UChar& rVal )
{
  ErrVal nRet;

  RNOKRS( nRet = getToken(), nRet );

  if( Err::m_nERR == stringToValue( m_cToken, rVal ) )
  {
    msgError( "UChar expected. Found: " +  m_cToken );
    return Err::m_nERR;
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::getToken( UShort& rVal )
{
  ErrVal nRet;

  RNOKRS( nRet = getToken(), nRet );

  if( Err::m_nERR == stringToValue( m_cToken, rVal ) )
  {
    msgError( "UShort expected. Found: " +  m_cToken );
    return Err::m_nERR;
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::getToken( UInt& rVal )
{
  ErrVal nRet;

  RNOKRS( nRet = getToken(), nRet );

  if( Err::m_nERR == stringToValue( m_cToken, rVal ) )
  {
    msgError( "UInt expected. Found: " +  m_cToken );
    return Err::m_nERR;
  }

  return Err::m_nOK;
}

ErrVal ToolLibParser::getToken( UInt64& rVal )
{
  ErrVal nRet;

  RNOKRS( nRet = getToken(), nRet );

  if( Err::m_nERR == stringToValue( m_cToken, rVal ) )
  {
    msgError( "UInt64 expected in decimal format. Found: " +  m_cToken );
    return Err::m_nERR;
  }

  return Err::m_nOK;
}



ErrVal ToolLibParser::getToken( float& rVal )
{
  ErrVal nRet;

  RNOKRS( nRet = getToken(), nRet );

  if( Err::m_nERR == stringToValue( m_cToken, rVal ) )
  {
    msgError( "incorrect number format: not a 'float' value: " +  m_cToken );
    return Err::m_nERR;
  }

  return nRet;
}

ErrVal ToolLibParser::getToken( double& rVal )
{
  ErrVal nRet;
  RNOKRS( nRet = getToken(), nRet );

  if( Err::m_nERR == stringToValue( m_cToken, rVal ) )
  {
    msgError( "incorrect number format: not a 'double' value: " +  m_cToken );
    return Err::m_nERR;
  }

  return nRet;
}

