/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/ResTrafoPow2.h"
#include "VideoCommonLib/ContextDataDeriver.h"
#include "VideoCommonLib/SliceData.h"
#include "VideoCommonLib/IntraDirectional.h"
#include "VideoCommonLib/InterTranslational.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


ErrVal ContextDataDeriver::deriveCbfByResInfoPtr( const PlanePackageParameters &rcPlaPaPars,
                                                  const ResInfo* pcResInfo,
                                                  const PrimPixelPos& cPosInResInfo,
                                                  std::vector<bool> &rabCbf )
{
  rabCbf = std::vector<bool>( rcPlaPaPars.getNumPlanes(), false );
  ROTRS( pcResInfo->getResidualMode() == ResInfo::RES_TRAFO_SKIP, Err::m_nOK );

  if( pcResInfo->getResidualMode() == ResInfo::RES_TRAFO_POW2 )
  {
    const ResInfoTrafoPow2 *p = pcResInfo->getResInfoTrafoPow2();
    const QTNavi &rcTrafoPow2LeafNav = p->getTree().getTreeStructure().getNodeByPrimPixelPos( cPosInResInfo );

    for( UInt uiPl = 0; uiPl < rcPlaPaPars.getNumPlanes(); uiPl++ )
    {
      rabCbf[uiPl] = p->getTree().getT( rcTrafoPow2LeafNav )->getCbf( uiPl );
    }
  }
  else
  {
    ROT( 1 );
  }
  return Err::m_nOK;
}


ErrVal ContextDataDeriver::deriveCbfByPrimPixelPos( const PlanePackageParameters &rcPlaPaPars,
                                                    const SliceData &rcSliceData,
                                                    const PrimPixelPos& cPos,
                                                    std::vector<bool> &rabCbf )
{
  const QTNavi &rcNav = rcSliceData.getQuadtreeGroupStructure().getNodeByPrimPixelPos( cPos );

  if( rcNav.isLeafNode() )
  {
    const ResInfo *pcResInfo = rcSliceData.getResInfoTree().getT( rcNav );
    RNOK( deriveCbfByResInfoPtr( rcPlaPaPars, pcResInfo, cPos - rcNav.getPrimPixelPos(), rabCbf ) );
  }
  else
  {
    rabCbf = std::vector<bool>( rcPlaPaPars.getNumPlanes(), false );
  }
  return Err::m_nOK;
}


ErrVal ContextDataDeriver::derivePow2CbfCtx( const QTNavi &rcSubLeafPos,
                                             const QTNavi &rcCtxNav,
                                             const AccessUnitParameters &rcAUPars,
                                             const PlanePackageParameters &rcPlaPaPars,
                                             const SliceData &rcSliceData,
                                             const QuadtreeGroup<ResInfoTrafoPow2Leaf*> &rcTreeCurr,
                                             std::vector<UInt> &rauiPow2CbfCtx )
{
  const UInt uiNumPlanes = rcPlaPaPars.getNumPlanes();

  if( rcSubLeafPos.getLayer() == 0 )
  {
    const PrimPixelPos cPosOfPredBlock = rcCtxNav.getPrimPixelPos();
    const PrimPixelPos cPosInPredBlock = rcSubLeafPos.getPrimPixelPos();

    const bool bLeftInside = cPosInPredBlock.getX() > 0 && cPosInPredBlock.getY() >= 0;
    const bool bTopInside = cPosInPredBlock.getY() > 0 && cPosInPredBlock.getX() >= 0;

    std::vector<bool> abLeftCbf( uiNumPlanes, false );
    std::vector<bool> abTopCbf( uiNumPlanes, false );

    if( bLeftInside )
    {
      const QTNavi &rcNav = rcTreeCurr.getTreeStructure().getNodeByPrimPixelPos( cPosInPredBlock + PrimPixelPos( -1, 0 ) );
      ROF( rcNav.isLeafNode() );
      const ResInfoTrafoPow2Leaf* p = rcTreeCurr.getT( rcNav );
      for( UInt ui = 0; ui < uiNumPlanes; ui++ )
      {
        abLeftCbf[ui] = p->getCbf( ui );
      }
    }
    else
    {
      const PrimPixelPos cLeftPos = cPosOfPredBlock + cPosInPredBlock + PrimPixelPos( -2, 0 ); // (-2) in order to avoid assertion in multiplication with Subsampling
      const QTNavi cLeft = rcSliceData.getQuadtreeGroupStructure().getNodeByPrimPixelPos( cLeftPos );
      if ( !cLeft.liesOutside() && cLeft.isParsedBefore( rcCtxNav ) )
      {
        RNOK( ContextDataDeriver::deriveCbfByPrimPixelPos( rcPlaPaPars, rcSliceData, cLeftPos, abLeftCbf ) );
      }
    }

    if( bTopInside )
    {
      const QTNavi &rcNav = rcTreeCurr.getTreeStructure().getNodeByPrimPixelPos( cPosInPredBlock + PrimPixelPos( 0, -1 ) );
      ROF( rcNav.isLeafNode() );
      const ResInfoTrafoPow2Leaf* p = rcTreeCurr.getT( rcNav );
      for( UInt ui = 0; ui < uiNumPlanes; ui++ )
      {
        abTopCbf[ui] = p->getCbf( ui );
      }
    }
    else
    {
      const PrimPixelPos cTopPos = cPosOfPredBlock + cPosInPredBlock + PrimPixelPos( 0, -2 ); // (-2) in order to avoid assertion in multiplication with Subsampling
      const QTNavi cTop  = rcSliceData.getQuadtreeGroupStructure().getNodeByPrimPixelPos( cTopPos  );
      if ( !cTop.liesOutside() && cTop.isParsedBefore( rcCtxNav ) )
      {
        RNOK( ContextDataDeriver::deriveCbfByPrimPixelPos( rcPlaPaPars, rcSliceData, cTopPos, abTopCbf ) );
      }
    }

    for( UInt ui = 0; ui < uiNumPlanes; ui++ )
    {
      rauiPow2CbfCtx[ui] = 1 + (abLeftCbf[ui] ? 1 : 0) + (abTopCbf[ui] ? 2 : 0);
    }
  }
  else
  {
    for( UInt ui = 0; ui < uiNumPlanes; ui++ )
    {
      rauiPow2CbfCtx[ui] = 0;
    }
  }
  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END
