/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/IntraDC.h"
#include "VideoCommonLib/IntraDirectional.h"
#include "VideoCommonLib/InterTranslational.h"
#include "VideoCommonLib/InterPlanePackage.h"
#include "VideoCommonLib/SliceData.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN



ErrVal
PredInfoInterPlanePackage::xCreatePredTree( const QuadtreeGroup<PredInfo*>& rcPredTreeSrc )
{
  QTNavi  cNavSrc = rcPredTreeSrc.getTreeStructure().getFirstQuadtreeRoot();
  QTNavi  cNavDes = m_cQTGS                         .getFirstQuadtreeRoot();
  RNOK( QuadtreeGroupPointerDeleter::deleteChildNodePointers( m_cPredTree, cNavDes ) );
  RNOK( m_cPredTree.getTreeStructure().removeChildNodes( cNavDes ) );
  RNOK( xCopyQTNode( cNavDes, rcPredTreeSrc, cNavSrc ) );
  ROF ( rcPredTreeSrc.getTreeStructure().getNextQuadtreeRoot( cNavSrc ).liesOutside() ); // just a check
  ROF ( m_cQTGS                         .getNextQuadtreeRoot( cNavDes ).liesOutside() ); // just a check
  return Err::m_nOK;
}

ErrVal
PredInfoInterPlanePackage::xCopyQTNode( const QTNavi &rcNavDes, const QuadtreeGroup<PredInfo*>& rcPredTreeSrc, const QTNavi &rcNavSrc )
{
  if( rcNavSrc.isLeafNode() )
  {
    PredInfo*       pcPredInfoDes = m_cPredTree  .getT( rcNavDes );
    const PredInfo* pcPredInfoSrc = rcPredTreeSrc.getT( rcNavSrc );
    ROT( pcPredInfoDes );
    if ( pcPredInfoSrc )
    {
      switch( pcPredInfoSrc->getPredictionMode() )
      {
      case PredInfo::INTRA_DC:                    pcPredInfoDes = new PredSyntaxIntraDC                ( *pcPredInfoSrc->getPredInfoIntraDC                () );  break;
      case PredInfo::INTRA_DIRECTIONAL:           pcPredInfoDes = new PredSyntaxIntraDirectional       ( *pcPredInfoSrc->getPredInfoIntraDirectional       () );  break;
      case PredInfo::INTER_TRANSLATIONAL:           pcPredInfoDes = new PredSyntaxInterTranslational       ( *pcPredInfoSrc->getPredInfoInterTranslational       () );  break;
      default:                                    return Err::m_nERR;
      }
    }
  }
  else
  {
    bool bOk = false;
    RNOK( m_cQTGS.appendChildNodes( rcNavDes, bOk ) );
    ROF ( bOk );
    for ( UInt uiIdx = 0; uiIdx < 4; uiIdx++ )
    {
      const QTNavi &rcChildDes = rcNavDes.getChild( uiIdx );
      const QTNavi &rcChildSrc = rcNavSrc.getChild( uiIdx );
      if( rcChildDes.isLeafOrInternalNode() )
      {
        ROF ( rcChildSrc.isLeafOrInternalNode() );
        RNOK( xCopyQTNode( rcChildDes, rcPredTreeSrc, rcChildSrc ) );
      }
    }
  }
  return Err::m_nOK;
}



ErrVal
PredInfoInterPlanePackage::xSetPredInfo( const QTNavi &rcNav, const PlanePackageParameters& rcPlaPaPars, const PlanePackage& rcRefPlaPa )
{
  // tree must be empty
  const QTNavi &rcNavRoot = m_cQTGS.getFirstQuadtreeRoot();
  ROFR ( rcNavRoot   .isLeafNode() , Err::m_nERR);
  ROT ( m_cPredTree.getT      ( rcNavRoot ) );
  RNOK( xCreatePredTree( rcNav, rcPlaPaPars, rcRefPlaPa ) );
  return Err::m_nOK;
}

ErrVal
PredInfoInterPlanePackage::xCreatePredTree( const QTNavi &rcNav, const PlanePackageParameters& rcPlaPaPars, const PlanePackage& rcRefPlaPa )
{
  const QTNavi      &rcNavDes        = m_cQTGS.getFirstQuadtreeRoot();
  const Subsampling &rcSub           = rcRefPlaPa.getPars().getPrimPlaneSubsampling() / rcPlaPaPars.getPrimPlaneSubsampling();
  UInt               uiLog2ResFactor = -rcSub.getX().getDownShift();
//th
//  ROF( ( 1 << uiLog2ResFactor ) * rcSub.getX().getDenominator() == rcSub.getX().getNumerator() );
//  ROF( ( 1 << uiLog2ResFactor ) * rcSub.getY().getDenominator() == rcSub.getY().getNumerator() );
  RNOK( xDeriveQTNode( rcNav, rcNavDes, uiLog2ResFactor, rcRefPlaPa ) );
  return Err::m_nOK;
}

bool
PredInfoInterPlanePackage::xIsAllIntra( const QTNavi&       rcNavPrd,
                                        const QTNavi&       rcNavDes,
                                        const UInt          uiLog2ResFactor,
                                        const PlanePackage& rcRefPlaPa )
{
  TOF( rcNavPrd.isLeafNode() );
  TOF( rcNavDes.isLeafNode() );
  const PrimPixelPos  &rcPrimPosCurr = rcNavPrd.getPrimPixelPos() + rcNavDes.getPrimPixelPos();
  const PrimPixelPos  cPrimPosRef( rcPrimPosCurr.getX() << uiLog2ResFactor, rcPrimPosCurr.getY() << uiLog2ResFactor );
  const QTNavi        &rcNavRef      = rcRefPlaPa.getCurrentSlice().getPredInfoTree().getTreeStructure().getNodeByPrimPixelPos( cPrimPosRef );
  const bool          bIsLeafNode    = ( rcNavRef.getLog2BlockSize() >= rcNavDes.getLog2BlockSize() + uiLog2ResFactor );
  bool                bAllIntra      = true;
  if( bIsLeafNode )
  {
    const PredInfo*   pcPredInfoSrc = rcRefPlaPa.getCurrentSlice().getPredInfoTree().getT( rcNavRef );
    TOF   ( pcPredInfoSrc );
    switch( pcPredInfoSrc->getPredictionMode() )
    {
    case PredInfo::INTER_TRANSLATIONAL:
      bAllIntra = false;
      break;
    default:
      break;
    }
  }
  else
  {
    bool bOk = false;
    TOF( Err::m_nOK == m_cQTGS.appendChildNodes( rcNavDes, bOk ) );
    TOF( bOk );
    for( UInt uiIdx = 0; uiIdx < 4; uiIdx++ )
    {
      const QTNavi &rcNavChild = rcNavDes.getChild( uiIdx );
      if( rcNavChild.isLeafOrInternalNode() )
      {
        bool bIntra = xIsAllIntra( rcNavPrd, rcNavChild, uiLog2ResFactor, rcRefPlaPa );
        if( !bIntra )
        {
          bAllIntra = false;
          break;
        }
      }
    }
    TOF( Err::m_nOK == m_cQTGS.removeChildNodes( rcNavDes ) );
    TOF( rcNavDes.isLeafNode() );
  }
  return bAllIntra;
}

ErrVal
PredInfoInterPlanePackage::xDeriveQTNode( const QTNavi&       rcNavPrd,
                                          const QTNavi&       rcNavDes,
                                          const UInt          uiLog2ResFactor,
                                          const PlanePackage& rcRefPlaPa )
{
  ROF( rcNavPrd.isLeafNode() );
  ROF( rcNavDes.isLeafNode() );
  const PrimPixelPos  &rcPrimPosCurr = rcNavPrd.getPrimPixelPos() + rcNavDes.getPrimPixelPos();
  const PrimPixelPos  cPrimPosRef( rcPrimPosCurr.getX() << uiLog2ResFactor, rcPrimPosCurr.getY() << uiLog2ResFactor );
  const QTNavi        &rcNavRef      = rcRefPlaPa.getCurrentSlice().getPredInfoTree().getTreeStructure().getNodeByPrimPixelPos( cPrimPosRef );
  const bool          bIsAllIntra    = xIsAllIntra( rcNavPrd, rcNavDes, uiLog2ResFactor, rcRefPlaPa );
  const bool          bIsLeafNode    = ( rcNavRef.getLog2BlockSize() >= rcNavDes.getLog2BlockSize() + uiLog2ResFactor );

  if( bIsAllIntra )
  { // mark intra leaf nodes by null pointer (pointer must be equal to null at this place)
    ROT( m_cPredTree.getT( rcNavDes ) );
    return Err::m_nOK;
  }
  
  if( bIsLeafNode )
  { // create new pred info with scaled motion vectors
    const PredInfo*                 pcPredInfoSrc   = rcRefPlaPa.getCurrentSlice().getPredInfoTree().getT( rcNavRef );
    ROF( pcPredInfoSrc );
    const PredInfoInterTranslational* pcPredInfoTrans = pcPredInfoSrc->getPredInfoInterTranslational();
    ROF( pcPredInfoTrans );
    PredInfoInterTranslational*       pcScaledPrdInfo = new PredSyntaxInterTranslational( *pcPredInfoTrans );
    for( UInt ui = 0; ui < pcScaledPrdInfo->getNumMv(); ui++ )
    {
      pcScaledPrdInfo->getMv( ui ).getPrecision() += int( uiLog2ResFactor );
    }
    m_cPredTree.getT( rcNavDes ) = pcScaledPrdInfo;
    return Err::m_nOK;
  }

  { // append child nodes and derive pred info for child nodes
    bool bOk = false;
    RNOK( m_cQTGS.appendChildNodes( rcNavDes, bOk ) );
    ROF ( bOk );
    for( UInt uiIdx = 0; uiIdx < 4; uiIdx++ )
    {
      const QTNavi &rcNavChild = rcNavDes.getChild( uiIdx );
      if( rcNavChild.isLeafOrInternalNode() )
      {
        RNOK( xDeriveQTNode( rcNavPrd, rcNavChild, uiLog2ResFactor, rcRefPlaPa ) );
      }
    }
    RNOK( xReplaceIntra( rcNavDes ) );
  }
  return Err::m_nOK;
}


ErrVal PredInfoInterPlanePackage::xReplaceIntra( const QTNavi &rcNavDes )
{
  ROT( rcNavDes.isLeafNode() );
  for( UInt uiIdx = 0; uiIdx < 4; uiIdx++ )
  {
    const QTNavi &rcNavChild = rcNavDes.getChild( uiIdx );
    if( rcNavChild.isLeafNode() && m_cPredTree.getT( rcNavChild ) == 0 )
    {
      int iSize = ( 1 << rcNavChild.getLog2BlockSize() );
      int iHorOffset = ( uiIdx & 1 ? -1 : iSize );
      const PrimPixelPos  cHorPos = rcNavChild.getPrimPixelPos() + PrimPixelPos( iHorOffset, 0 );
      const QTNavi  cNavHor = m_cQTGS.getNodeByPrimPixelPos( cHorPos );
      if( cNavHor.isLeafNode() && m_cPredTree.getT( cNavHor ) )
      {
        const PredInfoInterTranslational* pcPredInfoSrc = m_cPredTree.getT( cNavHor )->getPredInfoInterTranslational();
        ROF( pcPredInfoSrc );
        m_cPredTree.getT( rcNavChild ) = new PredSyntaxInterTranslational( *pcPredInfoSrc );
      }
      else
      {
        int iVerOffset = ( uiIdx > 1 ? -1 : iSize );
        const PrimPixelPos cVerPos = rcNavChild.getPrimPixelPos() + PrimPixelPos( 0, iVerOffset );
        const QTNavi cNavVer = m_cQTGS.getNodeByPrimPixelPos( cVerPos );
        if( cNavVer.isLeafNode() && m_cPredTree.getT( cNavVer ) )
        {
          const PredInfoInterTranslational* pcPredInfoSrc = m_cPredTree.getT( cNavVer )->getPredInfoInterTranslational();
          ROF( pcPredInfoSrc );
          m_cPredTree.getT( rcNavChild ) = new PredSyntaxInterTranslational( *pcPredInfoSrc );
        }
        else
        {
          const PrimPixelPos cDiaPos = rcNavChild.getPrimPixelPos() + PrimPixelPos( iHorOffset, iVerOffset );
          const QTNavi cNavDia = m_cQTGS.getNodeByPrimPixelPos( cDiaPos );
          if( cNavDia.isLeafNode() && m_cPredTree.getT( cNavDia ) )
          {
            const PredInfoInterTranslational* pcPredInfoSrc = m_cPredTree.getT( cNavDia )->getPredInfoInterTranslational();
            ROF( pcPredInfoSrc );
            m_cPredTree.getT( rcNavChild ) = new PredSyntaxInterTranslational( *pcPredInfoSrc );
          }
          else
          {
            ROT( 1 ); // should never happen (I think)
          }
        }
      }
    }
  }
  return Err::m_nOK;
}


bool PredInfoInterPlanePackage::xIsIntra() const
{
  const QTNavi &rcNavRoot = m_cQTGS.getFirstQuadtreeRoot();
  ROFRS( rcNavRoot.isLeafNode(), false );
  return ( m_cPredTree.getT( rcNavRoot ) == 0 );
}


ErrVal PredSyntaxInterPlanePackage::xSetComplementData( const QTNavi&                 rcPos,
                                                        const PlanePackageParameters& rcPlaPaPars,
                                                        const AccessUnitParameters&   rcAUPars,
                                                        const SliceData&              rcSliceData,
                                                        const PlanePackage*           pcRefPlanePackage )
{
  ROF( pcRefPlanePackage );
  if ( xIsPredSyntaxOnly() )
  { // syntax to data conversion
    RNOK( xSetPredInfo( rcPos, rcPlaPaPars, *pcRefPlanePackage ) );
  }
  else
  { // data to syntax conversion
  }
  return Err::m_nOK;
}


ErrVal InterPlanePackage::calculatePrediction( const PredInfo &rcPredInfo,
                                               const QTNavi &rcNav,
                                               const PrimPixelPos& cTopLeftPosOfPrediction,
                                               const PicBlock<const Pel> *pcEncPic,
                                               const CodingState &rcCodingState,
                                               const SliceData &rcCurrSlice,
                                               const PicBlock<DPel> &rcPredPicBlock,
                                               bool &rbPredictionValid )
{
  const PredInfoInterPlanePackage*  pcPredInfoIPP = rcPredInfo.getPredInfoInterPlanePackage();
  ROF( pcPredInfoIPP );
  const QuadtreeGroup<PredInfo*>&   rcPredTree    = pcPredInfoIPP->getPredTree();
  QTNavi                            cNavPred      = rcPredTree.getTreeStructure().getFirstQuadtreeRoot();
  RNOK( xCalcPredForNode( rcPredTree, cNavPred, cTopLeftPosOfPrediction, rcCodingState, rcCurrSlice, rcPredPicBlock, rbPredictionValid ) );
  return Err::m_nOK;
}

ErrVal
InterPlanePackage::xCalcPredForNode( const QuadtreeGroup<PredInfo*>&  rcPredTree,
                                     const QTNavi&                    rcNavPredTree,
                                     const PrimPixelPos&              cTopLeftPos,
                                     const CodingState&               rcCodingState,
                                     const SliceData&                 rcCurrSlice,
                                     const PicBlock<DPel>&            rcPredBlock,
                                     bool&                            rbPredictionValid )
{
  if( rcNavPredTree.isLeafNode() )
  {
    QTNavi                      cDummyNav( NULL );
    const PicBlock<const Pel>*  pcEncBlock  = 0;
    const PredInfo*             pcPredInfo  = rcPredTree.getT( rcNavPredTree );
    ROF ( pcPredInfo );
    ROF ( pcPredInfo->getPredictionMode() == PredInfo::INTER_TRANSLATIONAL );
    RNOK( m_rcInterTranslational.calculatePrediction( *pcPredInfo, cDummyNav, cTopLeftPos, pcEncBlock, rcCodingState, rcCurrSlice, rcPredBlock, rbPredictionValid ) );
    return Err::m_nOK;
  }

  rbPredictionValid       = true;
  PrimPixelPos  cTreePos  = rcNavPredTree.getPrimPixelPos();
  for( UInt uiIdx = 0; uiIdx < 4; uiIdx++ )
  {
    const QTNavi &rcNavChild = rcNavPredTree.getChild( uiIdx );
    if( rcNavChild.isLeafOrInternalNode() )
    {
      bool                  bPredValid    = true;
      PrimPixelPos          cChildPos     = rcNavChild.getPrimPixelPos () - cTreePos;
      PrimBlockSize         cChildSize    = rcNavChild.getPrimBlockSize();
      PrimPixelPos          cChildTopLeft = cTopLeftPos + cChildPos;
      const PicBlock<DPel>  cChildBlock   ( cChildPos, cChildSize, rcPredBlock );
      RNOK( xCalcPredForNode( rcPredTree, rcNavChild, cChildTopLeft, rcCodingState, rcCurrSlice, cChildBlock, bPredValid ) );
      if( !bPredValid )
      {
        rbPredictionValid = false;
      }
    }
  }
  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END
