/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/InterTranslational.h"
#include "VideoCommonLib/SliceData.h"
#include "VideoCommonLib/ReferencePictureAccessor.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


ErrVal
PredInfoInterTranslational::setDirectModeData( const QTNavi&                 rcPos,
                                             const PlanePackageParameters& rcPlaPaPars,
                                             const AccessUnitParameters&   rcAUPars,
                                             const SliceData&              rcSliceData )
{
  std::vector<MotionVector> cMvVector(4);
  setNumMv( rcAUPars.getMaxMvsPerBlk() );
  for( UInt uiIdx = 0; uiIdx < getNumMv(); uiIdx++ )
  {
    PicOrderCnt&  rcPoc = getRefPoc( uiIdx ) = rcAUPars.getRefPoc( uiIdx );
    MotionVector& rcMv  = getMv    ( uiIdx ) = MotionVector( 0, 0, rcPlaPaPars.getMvAccuracy() ); 
    MotionVector  cMvPred( 0, 0, rcMv.getPrecision() ); // specify precision for motion vector prediction !!!!
    RNOK( InterTranslational::getMvPred( rcPoc, rcPos, rcSliceData.getPredInfoTree(), cMvPred, cMvVector ) );
    rcMv = rcMv + cMvPred;
  }
  return Err::m_nOK;
}


bool
PredInfoInterTranslational::operator==( const PredInfoInterTranslational &rcComp ) const
{
  if( rcComp.m_acMvAndRefPoc.size() != m_acMvAndRefPoc.size() )
  {
    return false;
  }
  std::vector<MvAndRefPoc> acCompMvAndRefPoc( rcComp.m_acMvAndRefPoc );
  for( UInt ui = 0; ui < acCompMvAndRefPoc.size(); ui++ )
  {
    std::vector<MvAndRefPoc>::iterator cIter = std::find( acCompMvAndRefPoc.begin(), acCompMvAndRefPoc.end(), m_acMvAndRefPoc[ui] );
    if( cIter == acCompMvAndRefPoc.end() )
    {
      return false;
    }
    acCompMvAndRefPoc.erase( cIter );
  }
  return true;
}

bool PredInfoInterTranslational::compareMvAndRefPocVec( const std::vector<MvAndRefPoc>& racVecA, const std::vector<MvAndRefPoc>& racVecB )
{
  bool bIsSame = racVecA.size() == racVecB.size();
  std::vector<bool> auiIndexMatched( racVecB.size(), false );
  if ( bIsSame )
  {
    for ( UInt uiVectorAIdx = 0; bIsSame && ( uiVectorAIdx < racVecA.size() ); uiVectorAIdx++ )
    {
      bool bMvAndRefPocAPresentInVectorB = false;
      for ( UInt uiVectorBIdx = 0; !bMvAndRefPocAPresentInVectorB && ( uiVectorBIdx < racVecB.size() ); uiVectorBIdx++ )
      {
        if ( !auiIndexMatched[ uiVectorBIdx ] )
        {
          bMvAndRefPocAPresentInVectorB = ( racVecA[ uiVectorAIdx ] == racVecB[ uiVectorBIdx ] );
          auiIndexMatched[ uiVectorBIdx ] = bMvAndRefPocAPresentInVectorB;
        }
      }
      bIsSame = bMvAndRefPocAPresentInVectorB;
    }
  }

  return bIsSame;
}

ErrVal PredInfoInterTranslational::xFindLeftAndTopMergeTargets( const QTNavi& rcPos, const QuadtreeGroup<PredInfo*> &rcTree, bool& rbLeftAvailable, bool& rbTopAvailable, const PredInfoInterTranslational*& rpcPredInfoITTop, const PredInfoInterTranslational*& rpcPredInfoITLeft )
{
  const PrimPixelPos  cNavPos        = rcPos.getPrimPixelPos();
  const PrimPixelPos  cTopPos        = cNavPos + PrimPixelPos(  0,                              -1 );
  const PrimPixelPos  cLeftPos       = cNavPos + PrimPixelPos( -1,                               0 );
  const QTNavi        cTopNavi       = rcPos.findLeafNode( cTopPos      );
  const QTNavi        cLeftNavi      = rcPos.findLeafNode( cLeftPos     );
  
  rbTopAvailable  = !cTopNavi.liesOutside() && cTopNavi.isParsedBefore( rcPos );
  rbLeftAvailable = !cLeftNavi.liesOutside() && cLeftNavi.isParsedBefore( rcPos );

  if ( rbTopAvailable )
  {
    rbTopAvailable = ( rcTree.getT( cTopNavi )->getPredictionMode() == PredInfo::INTER_TRANSLATIONAL );
    if ( rbTopAvailable )
    {
      rpcPredInfoITTop = rcTree.getT( cTopNavi )->getPredInfoInterTranslational();
      // information of neighboring prediction object must be present in data mode
      ROFR( rpcPredInfoITTop, Err::m_nERR );
    }
  }
  if ( rbLeftAvailable )
  {
    rbLeftAvailable = ( rcTree.getT( cLeftNavi )->getPredictionMode() == PredInfo::INTER_TRANSLATIONAL );
    if ( rbLeftAvailable )
    { 
      rpcPredInfoITLeft = rcTree.getT( cLeftNavi )->getPredInfoInterTranslational();
      // information of neighboring prediction object must be present in data mode
      ROFR( rpcPredInfoITLeft, Err::m_nERR );
    }
  }

  return Err::m_nOK;
}

ErrVal
PredSyntaxInterTranslational::xSetComplementData( const QTNavi&                 rcPos,
                                                const PlanePackageParameters& rcPlaPaPars,
                                                const AccessUnitParameters&   rcAUPars,
                                                const SliceData&              rcSliceData,
                                                const PlanePackage*           pcRefPlanePackage )
{
  if( xIsPredSyntaxOnly() )
  { // syntax to data conversion
    if( rcAUPars.getEntropySliceFlag() && getDirectModeFlag() )
    {
      RNOK( PredInfoInterTranslational::setDirectModeData( rcPos, rcPlaPaPars, rcAUPars, rcSliceData ) );
    }
    else
    {
      PredInfoInterTranslational::setNumMv( PredSyntaxInterTranslational::getNumEntries() );

      if ( !rcAUPars.getEntropySliceFlag() && getMergeFlag() )
      {
        // Copy motion vectors from neighboring region
        bool bLeftAvailable;
        bool bTopAvailable;
        const PredInfoInterTranslational* pcPredInfoITLeft = NULL;
        const PredInfoInterTranslational* pcPredInfoITTop = NULL;
        RNOK( xFindLeftAndTopMergeTargets( rcPos, rcSliceData.getPredInfoTree(), bLeftAvailable, bTopAvailable, pcPredInfoITTop, pcPredInfoITLeft ) );

        ROFR( bLeftAvailable || bTopAvailable, Err::m_nERR ); // Block can only be merged if at least one neighboring region is present

        bool bMergeLeftNotTop = false; // Merge to top by default
        bool bSameRegion = false;
        if ( bLeftAvailable && bTopAvailable )
        {
          // check for equality
          bSameRegion = PredInfoInterTranslational::compareMvAndRefPocVec( pcPredInfoITLeft->getMvAndRefPocVec(), pcPredInfoITTop->getMvAndRefPocVec() );
          if ( !bSameRegion )
          {
            // two possible merge targets, merge left flag is present
            bMergeLeftNotTop = PredSyntaxInterTranslational::getMergeLeftFlag();
          }
        }
        else if ( bTopAvailable )
        {
          bMergeLeftNotTop = false;
        }
        else if ( bLeftAvailable )
        {
          bMergeLeftNotTop = true;
        }

        if ( bMergeLeftNotTop )
        {
          // copy vectors from left region
          PredInfoInterTranslational::getMvAndRefPocVec() = pcPredInfoITLeft->getMvAndRefPocVec();
        }
        else
        {
          // copy vectors from top region
          PredInfoInterTranslational::getMvAndRefPocVec() = pcPredInfoITTop->getMvAndRefPocVec();
        }

      }
      else
      {
        // calculate motion vectors from predictor and difference
        for( UInt uiIdx = 0; uiIdx < PredInfoInterTranslational::getNumMv(); uiIdx++ )
        {
          PicOrderCnt&  rcPoc = PredInfoInterTranslational::getRefPoc( uiIdx ) = rcAUPars.getRefPoc( getRefIdx( uiIdx ) );
          MotionVector& rcMv  = PredInfoInterTranslational::getMv    ( uiIdx ) =                     getMvDiff( uiIdx ); 
          if( rcPlaPaPars.getMvCompetition() )
          {
            std::vector<MotionVector>& cMvPredLst = m_acMvVector;
            RNOK( InterTranslational::getMvPredLst( rcPoc, rcPos, rcSliceData.getPredInfoTree(), cMvPredLst, rcMv.getPrecision() ) );
            ROF ( getMvPredIdx( uiIdx ) < cMvPredLst.size() );
            rcMv = rcMv + cMvPredLst[ getMvPredIdx( uiIdx ) ];
          }
          else
          {
            int iMvCompPredY = 0;
            int iMvCompPredX = 0;
            RNOK( InterTranslational::getMvPredYInd( rcPoc, rcPos, rcSliceData.getPredInfoTree(), rcMv.getPrecision(),              iMvCompPredY ) );
            rcMv.getY() += iMvCompPredY;
            RNOK( InterTranslational::getMvPredXDep( rcPoc, rcPos, rcSliceData.getPredInfoTree(), rcMv.getPrecision(), rcMv.getY(), iMvCompPredX ) );
            rcMv.getX() += iMvCompPredX;
          }
        }
      }

    }
  }
  else
  { // data to syntax conversion

    //--- check for direct mode ---
    bool bIsDirectMode = ( rcAUPars.isDirectModeFlagPresent( rcPlaPaPars.getPlaPaId() ) && PredInfoInterTranslational::getNumMv() == rcAUPars.getMaxMvsPerBlk() );
    if( bIsDirectMode )
    {
      std::vector<MotionVector> cMvVector;
      UInt uiPredSyntaxIdx = 0;
      PredSyntaxInterTranslational::setNumEntries( PredInfoInterTranslational::getNumMv() );
      for( UInt uiRefIdx = 0; uiRefIdx < rcAUPars.getRefListSize() && bIsDirectMode; uiRefIdx++ )
      {
        const PicOrderCnt cRefPoc = rcAUPars.getRefPoc( uiRefIdx );
        for( UInt uiPredInfoIdx = 0; uiPredInfoIdx < PredInfoInterTranslational::getNumMv() && bIsDirectMode; uiPredInfoIdx++ )
        {
          if( PredInfoInterTranslational::getRefPoc( uiPredInfoIdx ) == cRefPoc )
          {
            getRefIdx( uiPredSyntaxIdx ) = uiRefIdx;
            MotionVector& rcMv = getMvDiff( uiPredSyntaxIdx ) = PredInfoInterTranslational::getMv( uiPredInfoIdx );
            MotionVector cMvPred( 0, 0, PredInfoInterTranslational::getMv( uiPredInfoIdx ).getPrecision() ); // specify precision for motion vector prediction !!!!
            RNOK( InterTranslational::getMvPred( cRefPoc, rcPos, rcSliceData.getPredInfoTree(), cMvPred, cMvVector ) );
            rcMv = rcMv - cMvPred;
            if( uiRefIdx != uiPredSyntaxIdx || rcMv.getX() != 0 || rcMv.getY() != 0 )
            {
              bIsDirectMode = false;
            }
            uiPredSyntaxIdx++;
          }
        }
      }
    }
    setDirectModeFlag( bIsDirectMode );

    setMVAddress( rcPos.getPrimPixelPos() );

    if( !bIsDirectMode )
    {
      bool bMvDiffPresent = false;
      if( !rcAUPars.getEntropySliceFlag() )
      {
        // merge mode
        bool bLeftAvailable;
        bool bTopAvailable;
        const PredInfoInterTranslational* pcPredInfoITLeft = NULL;
        const PredInfoInterTranslational* pcPredInfoITTop = NULL;
        RNOK( xFindLeftAndTopMergeTargets( rcPos, rcSliceData.getPredInfoTree(), bLeftAvailable, bTopAvailable, pcPredInfoITTop, pcPredInfoITLeft ) );
        bool bTopParameterIsSame = false;
        bool bLeftParameterIsSame = false;
        if( bTopAvailable )
        {
          bTopParameterIsSame = PredInfoInterTranslational::compareMvAndRefPocVec( PredInfoInterTranslational::getMvAndRefPocVec(), pcPredInfoITTop->getMvAndRefPocVec() );
        }
        if( bLeftAvailable )
        {
          bLeftParameterIsSame = PredInfoInterTranslational::compareMvAndRefPocVec( PredInfoInterTranslational::getMvAndRefPocVec(), pcPredInfoITLeft->getMvAndRefPocVec() );
        }
        setMergeFlag( ( bTopParameterIsSame || bLeftParameterIsSame ) );
        if( getMergeFlag() )
        {
          // coded as being part of an already existing motion region with known vector
          bool bMergeLeftNotTop = !bTopParameterIsSame; // merge to top region if both regions have the same parameters.
          setMergeLeftFlag( bMergeLeftNotTop );
          setMVAddress( bMergeLeftNotTop ? pcPredInfoITLeft->getPredSyntax()->getPredSyntaxInterTranslational()->getMVAddress() : pcPredInfoITTop->getPredSyntax()->getPredSyntaxInterTranslational()->getMVAddress() );
        }
        else
        {
          bMvDiffPresent = true;
        }
      }
      else
      {
        bMvDiffPresent = true;
      }

      if( bMvDiffPresent )
      {
        //--- normal inter mode ---
        // coded as difference to predictor
        UInt  uiPredSyntaxIdx       = 0;
        PredSyntaxInterTranslational::setNumEntries( PredInfoInterTranslational::getNumMv() );
        for( UInt uiRefIdx = 0; uiRefIdx < rcAUPars.getRefListSize(); uiRefIdx++ )
        {
          const PicOrderCnt cRefPoc = rcAUPars.getRefPoc( uiRefIdx );
          for( UInt uiPredInfoIdx = 0; uiPredInfoIdx < PredInfoInterTranslational::getNumMv(); uiPredInfoIdx++ )
          {
            if( PredInfoInterTranslational::getRefPoc( uiPredInfoIdx ) == cRefPoc )
            {
              getRefIdx( uiPredSyntaxIdx ) = uiRefIdx;
              MotionVector& rcMv = getMvDiff( uiPredSyntaxIdx ) = PredInfoInterTranslational::getMv( uiPredInfoIdx );
              if( rcPlaPaPars.getMvCompetition() )
              {
                std::vector<MotionVector>& cMvPredLst = m_acMvVector;
                RNOK( InterTranslational::getMvPredLst( cRefPoc, rcPos, rcSliceData.getPredInfoTree(), cMvPredLst, rcMv.getPrecision() ) );
                UInt  uiBestIdx = TypeLimits<UInt>::m_iMax;
                UInt  uiMinCost = TypeLimits<UInt>::m_iMax;
                for( UInt uiIdx = 0; uiIdx < cMvPredLst.size(); uiIdx++ )
                {
                  MotionVector cMvd   = rcMv - cMvPredLst[ uiIdx ];
                  UInt       uiCost = xGetMvCompCost( cMvd.getX() ) + xGetMvCompCost( cMvd.getY() ) + ( cMvPredLst.size() > 2 && uiIdx > 0 ? 1 : 0 );
                  //UInt       uiCost = gAbs( cMvd.getX() ) + gAbs( cMvd.getY() );
                  if( uiCost < uiMinCost )
                  {
                    uiBestIdx = uiIdx;
                    uiMinCost = uiCost;
                  }
                }
                ROT( uiMinCost == TypeLimits<UInt>::m_iMax );
                rcMv = rcMv - cMvPredLst[ uiBestIdx ];
                getMvPredIdx( uiPredSyntaxIdx ) = uiBestIdx;
              }
              else
              {
                int iMvCompPredY = 0;
                int iMvCompPredX = 0;
                RNOK( InterTranslational::getMvPredYInd( cRefPoc, rcPos, rcSliceData.getPredInfoTree(), rcMv.getPrecision(),              iMvCompPredY ) );
                RNOK( InterTranslational::getMvPredXDep( cRefPoc, rcPos, rcSliceData.getPredInfoTree(), rcMv.getPrecision(), rcMv.getY(), iMvCompPredX ) );
                rcMv.getY() -= iMvCompPredY;
                rcMv.getX() -= iMvCompPredX;
              }
              uiPredSyntaxIdx++;
            }
          }
        }
      }
    }
  }

  return Err::m_nOK;
}

UInt
PredSyntaxInterTranslational::xGetMvCompCost( int iMvComp ) // count CABAC bins
{
  const UInt  uiMaxVal  = 9;
  const UInt  uiInitCnt = 3;

  UInt  uiAbs   = abs( iMvComp );
  UInt  uiVal   = 0;
  bool  bMore   = false;
  // unary part
  do  { bMore   = ( uiAbs > uiVal ); uiVal++; } while( bMore && uiVal < uiMaxVal );
  UInt  uiBits  = uiVal + ( uiAbs > 0 ? 1 : 0 );
  // exp-Golomb part
  UInt uiCnt = uiInitCnt;
  if( bMore )
  {
    uiVal = uiAbs - uiMaxVal;
    while( uiVal >= UInt( 1 << uiCnt ) )
    {
      uiVal -= UInt( 1 << uiCnt );
      uiCnt ++;
      uiBits++;
    }
    uiBits += 1 + uiCnt;
  }
  return uiBits;
}


ErrVal InterTranslational::calculatePrediction( const PredInfo &rcPredInfo,
                                              const QTNavi &rcNav,
                                              const PrimPixelPos&  cTopLeftPosOfPrediction,
                                              const PicBlock<const Pel> *pcEncPic,
                                              const CodingState &rcCodingState,
                                              const SliceData &rcCurrSlice,
                                              const PicBlock<DPel> &rcPredPicBlock,
                                              bool &rbPredictionValid )
{
  rbPredictionValid = true;
  const PredInfoInterTranslational*   const pcPredInfo      = rcPredInfo.getPredInfoInterTranslational();
  const ReferencePlanePackage<Pel>* const pcFirstRefPlaPa = rcCodingState.getRefPlanePackage( pcPredInfo->getRefPoc( 0 ) );
  ROF( pcFirstRefPlaPa );

  if( 1 == pcPredInfo->getNumMv() )
  {
    ReferencePlanePackageAccessor<SetPredictionToPicBlock<DPel>,Pel> cRefPlaPaAccessor( *pcFirstRefPlaPa, SetPredictionToPicBlock<DPel>( rcPredPicBlock ) );
    RNOK( cRefPlaPaAccessor.getPrediction( rcPredPicBlock.getStats(), pcPredInfo->getMv( 0 ).plusPrimResPos( cTopLeftPosOfPrediction ) ) );
  }
  else
  {
    //(HK) consider replacing the AllocPicBlock<int> below
//    const AllocPicBlock<int> cCumulatedPrediction( rcPredPicBlock.getStats() );//th
    const AllocPicBlockReuse<int> cCumulatedPrediction( rcPredPicBlock.getStats(), m_cReuseBuffer );
    ReferencePlanePackageAccessor<SetPredictionToPicBlock<int>,Pel> cRefPlaPaAccessor( *pcFirstRefPlaPa, SetPredictionToPicBlock<int>( cCumulatedPrediction ) );
    RNOK( cRefPlaPaAccessor.getPrediction( rcPredPicBlock.getStats(), pcPredInfo->getMv( 0 ).plusPrimResPos( cTopLeftPosOfPrediction ) ) );

    // sum all motion-compensated prediction signals
    for( UInt ui = 1; ui < pcPredInfo->getNumMv(); ui++ )
    {
      const ReferencePlanePackage<Pel>* const pcRefPlaPa = rcCodingState.getRefPlanePackage( pcPredInfo->getRefPoc( ui ) );
      ROF( pcRefPlaPa );
      ReferencePlanePackageAccessor<AddPredictionToPicBlock<int>,Pel> cRefPlaPaAccessor( *pcRefPlaPa, AddPredictionToPicBlock<int>( cCumulatedPrediction ) );
      RNOK( cRefPlaPaAccessor.getPrediction( rcPredPicBlock.getStats(), pcPredInfo->getMv( ui ).plusPrimResPos( cTopLeftPosOfPrediction ) ) );
    }
    rcPredPicBlock.assignAndDivide( cCumulatedPrediction, pcPredInfo->getNumMv() );
  }
  return Err::m_nOK;
}


ErrVal InterTranslational::xInsertReferenceMvs( PicOrderCnt cRefPoc,
                                              const QTNavi &rcNav,
                                              const QTNavi &rcNeighbour,
                                              const QuadtreeGroup<PredInfo*> &rcTree,
                                              int iMvAccuracy,
                                              std::vector<MotionVector> &rcReferenceMvs,
                                              bool bFirstOnly )
{
  const bool bAvailable = !rcNeighbour.liesOutside() && rcNeighbour.isCodedBefore( rcNav ) && rcTree.getT( rcNeighbour )->getPredictionMode() == PredInfo::INTER_TRANSLATIONAL;
  if( bAvailable )
  {
    const PredInfoInterTranslational &rcPredInfo = *rcTree.getT( rcNeighbour )->getPredInfoInterTranslational();
    const UInt uiNumMv = rcPredInfo.getNumMv();
    for( UInt ui = 0; ui < uiNumMv; ui++ )
    {
      if( rcPredInfo.getRefPoc( ui ) == cRefPoc )
      {
        rcReferenceMvs.push_back( rcPredInfo.getMv( ui ).getMvWithPrecisionModifiedTo( iMvAccuracy ) );
        ROTRS( bFirstOnly, Err::m_nOK );
      }
    }
  }
  return Err::m_nOK;
}


int
InterTranslational::xGetMvPredXCompInd( const std::vector<MotionVector>& rcMvPredCandLst )
{
  ROT( rcMvPredCandLst.empty() );
  if(  rcMvPredCandLst.size() == 1 )
  {
    return rcMvPredCandLst[0].getX();
  }
  if ( rcMvPredCandLst.size() == 2 )
  {
    return ( rcMvPredCandLst[0].getX() + rcMvPredCandLst[1].getX() + 1 ) >> 1;
  }
  const UInt uiNumEntries = UInt( rcMvPredCandLst.size() );
  std::vector<int>          aiX ( rcMvPredCandLst.size() );
  for( UInt ui = 0; ui < uiNumEntries; ui++ )
  {
    aiX[ui] = rcMvPredCandLst[ui].getX();
  }
  sort ( aiX.begin(), aiX.end() );
  return aiX[ uiNumEntries >> 1 ];
}

int
InterTranslational::xGetMvPredYCompInd( const std::vector<MotionVector>& rcMvPredCandLst )
{
  ROT( rcMvPredCandLst.empty() );
  if(  rcMvPredCandLst.size() == 1 )
  {
    return rcMvPredCandLst[0].getY();
  }
  if ( rcMvPredCandLst.size() == 2 )
  {
    return ( rcMvPredCandLst[0].getY() + rcMvPredCandLst[1].getY() + 1 ) >> 1;
  }
  const UInt uiNumEntries = UInt( rcMvPredCandLst.size() );
  std::vector<int>          aiY ( rcMvPredCandLst.size() );
  for( UInt ui = 0; ui < uiNumEntries; ui++ )
  {
    aiY[ui] = rcMvPredCandLst[ui].getY();
  }
  sort ( aiY.begin(), aiY.end() );
  return aiY[ uiNumEntries >> 1 ];
}

int
InterTranslational::xGetMvPredXCompDep( const std::vector<int>& rcYThresLst, 
                                      const std::vector<int>& rcXPredLst,
                                      const int               iYComponent )
{
  TOT( rcYThresLst.empty() );
  TOF( rcYThresLst.size () == rcXPredLst.size() );
  const UInt uiSM1 = UInt( rcYThresLst.size() ) - 1;
  for(  UInt uiIdx = 0; uiIdx < uiSM1; uiIdx++ )
  {
    if( iYComponent <= rcYThresLst[ uiIdx ] )
    {
      return rcXPredLst[ uiIdx ];
    }
  }
  return rcXPredLst[ uiSM1 ];
}

int
InterTranslational::xGetMvPredYCompDep( const std::vector<int>& rcXThresLst, 
                                      const std::vector<int>& rcYPredLst,
                                      const int               iXComponent )
{
  TOT( rcXThresLst.empty() );
  TOF( rcXThresLst.size () == rcYPredLst.size() );
  const UInt uiSM1 = UInt( rcXThresLst.size() ) - 1;
  for(  UInt uiIdx = 0; uiIdx < uiSM1; uiIdx++ )
  {
    if( iXComponent <= rcXThresLst[ uiIdx ] )
    {
      return rcYPredLst[ uiIdx ];
    }
  }
  return rcYPredLst[ uiSM1 ];
}

ErrVal
InterTranslational::xSetYThresXPredLists( const std::vector<MotionVector>& rcMvPredCandLst,
                                        std::vector<int>&                rcYThresLst,
                                        std::vector<int>&                rcXPredLst )
{
  ROT( rcMvPredCandLst.empty() );
  ROF( rcYThresLst    .empty() );
  ROF( rcXPredLst     .empty() );
  
  //===== get sorted list of y-components (without multiple occurrences) =====
  std::vector<int>  aiSortedYValues;
  for( UInt uiCandIdx   = 0; uiCandIdx < UInt( rcMvPredCandLst.size() ); uiCandIdx++ )
  {
    const int iYComp  = rcMvPredCandLst[ uiCandIdx ].getY();
    bool      bInsert = true;
    for( UInt uiIdx   = 0; uiIdx < UInt( aiSortedYValues.size() ); uiIdx++ )
    {
      if( aiSortedYValues[ uiIdx ] == iYComp )
      {
        bInsert = false;
        break;
      }
    }
    if( bInsert )
    {
      aiSortedYValues.push_back( iYComp );
    }
  }
  sort( aiSortedYValues.begin(), aiSortedYValues.end() );
  ROT ( aiSortedYValues.empty() ); // paranoia

  //===== set x-prediction for sorted values =====
  for( UInt uiIdx = 0; uiIdx < UInt( aiSortedYValues.size() ); uiIdx++ )
  {
    std::vector<int>  aiX;
    const int         iYComp  = aiSortedYValues[ uiIdx ];
    //--- create list of x-components for given y-component ---
    for( UInt uiCandIdx = 0; uiCandIdx < UInt( rcMvPredCandLst.size() ); uiCandIdx++ )
    {
      const MotionVector& rcCandMv = rcMvPredCandLst[ uiCandIdx ];
      if( iYComp == rcCandMv.getY() )
      {
        aiX.push_back( rcCandMv.getX() );
      }
    }
    ROT( aiX.empty() ); // paranoia
    //--- derive predictor for x ---
    int iMvPredX = 0;
    if ( aiX.size() == 1 )
    {
      iMvPredX  = aiX[0];
    }
    else if( aiX.size() == 2 )
    {
      iMvPredX  = ( aiX[0] + aiX[1] + 1 ) >> 1;
    }
    else
    {
      sort( aiX.begin(), aiX.end() );
      UInt uiId = UInt( aiX.size() ) >> 1;
      iMvPredX  = aiX[ uiId ];
    }
    //--- insert x-component predictor ---
    rcXPredLst.push_back( iMvPredX );
  }

  //===== determine decisions thresholds for y-component -----
  UInt uiSizeMin1 = UInt( aiSortedYValues.size() ) - 1;
  for( UInt uiIdx = 0; uiIdx < uiSizeMin1; uiIdx++ )
  {
    int iYThreshold = ( aiSortedYValues[ uiIdx ] + aiSortedYValues[ uiIdx + 1 ] ) >> 1;
    rcYThresLst.push_back( iYThreshold );
  }
  rcYThresLst.push_back( TypeLimits<int>::getMax() );
  ROF( rcYThresLst.size () == rcXPredLst.size() ); // paranoia
  ROT( rcYThresLst.empty() );
  return Err::m_nOK;
}

ErrVal
InterTranslational::xSetXThresYPredLists( const std::vector<MotionVector>& rcMvPredCandLst,
                                        std::vector<int>&                rcXThresLst,
                                        std::vector<int>&                rcYPredLst )
{
  ROT( rcMvPredCandLst.empty() );
  ROF( rcXThresLst    .empty() );
  ROF( rcYPredLst     .empty() );
  
  //===== get sorted list of x-components (without multiple occurrences) =====
  std::vector<int>  aiSortedXValues;
  for( UInt uiCandIdx   = 0; uiCandIdx < UInt( rcMvPredCandLst.size() ); uiCandIdx++ )
  {
    const int iXComp  = rcMvPredCandLst[ uiCandIdx ].getX();
    bool      bInsert = true;
    for( UInt uiIdx   = 0; uiIdx < UInt( aiSortedXValues.size() ); uiIdx++ )
    {
      if( aiSortedXValues[ uiIdx ] == iXComp )
      {
        bInsert = false;
        break;
      }
    }
    if( bInsert )
    {
      aiSortedXValues.push_back( iXComp );
    }
  }
  sort( aiSortedXValues.begin(), aiSortedXValues.end() );
  ROT ( aiSortedXValues.empty() ); // paranoia

  //===== set y-prediction for sorted values =====
  for( UInt uiIdx = 0; uiIdx < UInt( aiSortedXValues.size() ); uiIdx++ )
  {
    std::vector<int>  aiY;
    const int         iXComp  = aiSortedXValues[ uiIdx ];
    //--- create list of x-components for given y-component ---
    for( UInt uiCandIdx = 0; uiCandIdx < UInt( rcMvPredCandLst.size() ); uiCandIdx++ )
    {
      const MotionVector& rcCandMv = rcMvPredCandLst[ uiCandIdx ];
      if( iXComp == rcCandMv.getX() )
      {
        aiY.push_back( rcCandMv.getY() );
      }
    }
    ROT( aiY.empty() ); // paranoia
    //--- derive predictor for y ---
    int iMvPredY = 0;
    if ( aiY.size() == 1 )
    {
      iMvPredY  = aiY[0];
    }
    else if( aiY.size() == 2 )
    {
      iMvPredY  = ( aiY[0] + aiY[1] + 1 ) >> 1;
    }
    else
    {
      sort( aiY.begin(), aiY.end() );
      UInt uiId = UInt( aiY.size() ) >> 1;
      iMvPredY  = aiY[ uiId ];
    }
    //--- insert x-component predictor ---
    rcYPredLst.push_back( iMvPredY );
  }

  //===== determine decisions thresholds for x-component -----
  UInt uiSizeMin1 = UInt( aiSortedXValues.size() ) - 1;
  for( UInt uiIdx = 0; uiIdx < uiSizeMin1; uiIdx++ )
  {
    int iXThreshold = ( aiSortedXValues[ uiIdx ] + aiSortedXValues[ uiIdx + 1 ] ) >> 1;
    rcXThresLst.push_back( iXThreshold );
  }
  rcXThresLst.push_back( TypeLimits<int>::getMax() );
  ROF( rcXThresLst.size () == rcYPredLst.size() ); // paranoia
  ROT( rcXThresLst.empty() );
  return Err::m_nOK;
}

ErrVal
InterTranslational::xGetMvPredCandLstStd( const PicOrderCnt&              rcRefPoc,
                                        const QTNavi&                   rcNav,
                                        const QuadtreeGroup<PredInfo*>& rcTree,
                                        const int                       iMvPrecision,
                                        std::vector<MotionVector>&      rcMvPredCandLst )
{
  PrimPixelPos cLeftPos     = rcNav.getPrimPixelPos() + PrimPixelPos( -1,                                0 );
  PrimPixelPos cTopPos      = rcNav.getPrimPixelPos() + PrimPixelPos(  0,                               -1 );
  PrimPixelPos cTopRightPos = rcNav.getPrimPixelPos() + PrimPixelPos(  rcNav.getPrimBlockSize().getX(), -1 );
  PrimPixelPos cTopLeftPos  = rcNav.getPrimPixelPos() + PrimPixelPos( -1,                               -1 );
  QTNavi       cLeft        = rcNav.findLeafNode( cLeftPos     );
  QTNavi       cTop         = rcNav.findLeafNode( cTopPos      );
  QTNavi       cTopRight    = rcNav.findLeafNode( cTopRightPos );
  QTNavi       cTopLeft     = rcNav.findLeafNode( cTopLeftPos  );

  RNOK  ( xInsertNeighbourMvs( rcRefPoc, rcNav, cLeft,      rcTree, iMvPrecision, false, rcMvPredCandLst ) );
  RNOK  ( xInsertNeighbourMvs( rcRefPoc, rcNav, cTop,       rcTree, iMvPrecision, false, rcMvPredCandLst ) );
  const size_t uiNumMvsBefore = rcMvPredCandLst.size();
  RNOK  ( xInsertNeighbourMvs( rcRefPoc, rcNav, cTopRight,  rcTree, iMvPrecision, false, rcMvPredCandLst ) );
  if( rcMvPredCandLst.size() == uiNumMvsBefore )
  {
    RNOK( xInsertNeighbourMvs( rcRefPoc, rcNav, cTopLeft,   rcTree, iMvPrecision, false, rcMvPredCandLst ) );
  }
  if( rcMvPredCandLst.empty() )
  {
    rcMvPredCandLst.push_back( MotionVector( 0, 0, iMvPrecision ) );
  }
  return Err::m_nOK;
}

ErrVal
InterTranslational::xGetMvPredCandLstExt( const PicOrderCnt&              rcRefPoc,
                                        const QTNavi&                   rcNav,
                                        const QuadtreeGroup<PredInfo*>& rcTree,
                                        const int                       iMvPrecision,
                                        std::vector<MotionVector>&      rcMvPredCandLst )
{
  const PrimPixelPos&   rcPrimPixPos  = rcNav.getPrimPixelPos ();
  const PrimBlockSize&  rcPrimBlkSize = rcNav.getPrimBlockSize();
  //----- left neighbors -----
  for( int iY = 0; iY <  rcPrimBlkSize.getY(); iY += 4 )
  {
    const PrimPixelPos  cNeighbourPos = rcPrimPixPos + PrimPixelPos( -1, iY );
    const QTNavi        cNeighbourNav = rcNav.findLeafNode( cNeighbourPos );
    RNOK( xInsertNeighbourMvs( rcRefPoc, rcNav, cNeighbourNav, rcTree, iMvPrecision, true, rcMvPredCandLst ) );
  }
  //----- top-left -----
  {
    const PrimPixelPos  cNeighbourPos = rcPrimPixPos + PrimPixelPos( -1, -1 );
    const QTNavi        cNeighbourNav = rcNav.findLeafNode( cNeighbourPos );
    RNOK( xInsertNeighbourMvs( rcRefPoc, rcNav, cNeighbourNav, rcTree, iMvPrecision, true, rcMvPredCandLst ) );
  }
  //----- top and top-right -----
  for( int iX = 0; iX <= rcPrimBlkSize.getX(); iX += 4 )
  {
    const PrimPixelPos  cNeighbourPos = rcPrimPixPos + PrimPixelPos( iX, -1 );
    const QTNavi        cNeighbourNav = rcNav.findLeafNode( cNeighbourPos );
    RNOK( xInsertNeighbourMvs( rcRefPoc, rcNav, cNeighbourNav, rcTree, iMvPrecision, true, rcMvPredCandLst ) );
  }
  //----- check for empty list -----
  if( rcMvPredCandLst.empty() )
  {
    rcMvPredCandLst.push_back( MotionVector( 0, 0, iMvPrecision ) );
  }
  return Err::m_nOK;
}

ErrVal 
InterTranslational::xInsertNeighbourMvs( const PicOrderCnt&               rcRefPoc,
                                       const QTNavi&                    rcNav,
                                       const QTNavi&                    rcNeighbour,
                                       const QuadtreeGroup<PredInfo*>&  rcTree,
                                       const int                        iMvAccuracy,
                                       const bool                       bAvoidMultipleInsertion,
                                       std::vector<MotionVector>&       racNeighbourMvs )
{
  const bool bAvailable = ( !rcNeighbour.liesOutside   ()         &&
                             rcNeighbour.isCodedBefore ( rcNav )  &&
                             rcTree.getT( rcNeighbour )->getPredictionMode() == PredInfo::INTER_TRANSLATIONAL );
  if( bAvailable )
  {
    const PredInfoInterTranslational& rcPredInfo  = *rcTree.getT( rcNeighbour )->getPredInfoInterTranslational();
    const UInt                      uiNumMv     = rcPredInfo.getNumMv();
    for( UInt ui = 0; ui < uiNumMv; ui++ )
    {
      if( rcPredInfo.getRefPoc( ui ) == rcRefPoc )
      {
        MotionVector  cMv     = rcPredInfo.getMv( ui ).getMvWithPrecisionModifiedTo( iMvAccuracy );
        bool          bInsert = true;
        if( bAvoidMultipleInsertion )
        {
          for( UInt uiIdx = 0; uiIdx < UInt( racNeighbourMvs.size() ); uiIdx++ )
          {
            if( racNeighbourMvs[ uiIdx ] == cMv )
            {
              bInsert = false;
              break;
            }
          }
        }
        if( bInsert )
        {
          racNeighbourMvs.push_back( cMv );
        }
      }
    }
  }
  return Err::m_nOK;
}


ErrVal InterTranslational::xIsRefIdxAvailable( UInt uiRefIdx,
                                             const QTNavi &rcNav,
                                             const QTNavi &rcNeighbour,
                                             const QuadtreeGroup<PredInfo*> &rcTree,
                                             bool& rbAvailable )
{
  rbAvailable = false;
  ROTRS( rcNeighbour.liesOutside   (),        Err::m_nOK );
  ROFRS( rcNeighbour.isCodedBefore ( rcNav ), Err::m_nOK );
  const PredSyntax* pcPredSyntax = rcTree.getT( rcNeighbour )->getPredSyntax();
  ROF  ( pcPredSyntax );
  if   ( pcPredSyntax->getCodingMode() == PredInfo::INTER_PLANE_PACKAGE )
  {
    rbAvailable = true; // it may be available, we don't know it during parsing
    return Err::m_nOK;
  }
  ROFRS( pcPredSyntax->getCodingMode() == PredInfo::INTER_TRANSLATIONAL, Err::m_nOK );
  const PredSyntaxInterTranslational* pcPredSyntaxIT = pcPredSyntax->getPredSyntaxInterTranslational();
  ROF( pcPredSyntaxIT ); 
  
  if ( pcPredSyntaxIT->getMergeFlag() )
  {
    // The current block only stores the merge information only.
    // Get syntax object of block that stores MV syntax information for current block.
    PrimPixelPos cRegionOrigin( pcPredSyntaxIT->getMVAddress() );
    pcPredSyntaxIT = rcTree.getT( rcNav.findLeafNode( cRegionOrigin ) )->getPredSyntax()->getPredSyntaxInterTranslational();
  }
  
  // only non-merged blocks store MV syntax information
  ROTRS( pcPredSyntaxIT->getMergeFlag(), Err::m_nERR );

  for( UInt uiIdx = 0; uiIdx < pcPredSyntaxIT->getNumEntries(); uiIdx++ )
  {
    if( pcPredSyntaxIT->getRefIdx( uiIdx ) == uiRefIdx )
    {
      rbAvailable = true;
      return Err::m_nOK;
    }
  }
  return Err::m_nOK;
}

ErrVal InterTranslational::getMvPred( PicOrderCnt cRefPoc,
                                    const QTNavi &rcNav,
                                    const QuadtreeGroup<PredInfo*> &rcTree,
                                    MotionVector &rcMvPred,
                                    std::vector<MotionVector>& rcReferenceMvs) // NOTE: motion vector precision is specified by the input rcMvPred !!!!
{
  const PrimPixelPos &rcNavPos      = rcNav.getPrimPixelPos();
  const PrimPixelPos &rcLeftPos     = rcNavPos + PrimPixelPos( -1,                                0 );
  const PrimPixelPos &rcTopPos      = rcNavPos + PrimPixelPos(  0,                               -1 );
  const PrimPixelPos &rcTopRightPos = rcNavPos + PrimPixelPos(  rcNav.getPrimBlockSize().getX(), -1 );
  const PrimPixelPos &rcTopLeftPos  = rcNavPos + PrimPixelPos( -1                              , -1 );

  const QTNavi &rcLeft     = rcNav.findLeafNode( rcLeftPos     );
  const QTNavi &rcTop      = rcNav.findLeafNode( rcTopPos      );
  const QTNavi &rcTopRight = rcNav.findLeafNode( rcTopRightPos );
  const QTNavi &rcTopLeft  = rcNav.findLeafNode( rcTopLeftPos  );

  rcReferenceMvs.clear();
  //(HK) consider getting a temp vector from outside or have it as member
//  std::vector<MotionVector> acReferenceMvs;

  RNOK  ( xInsertReferenceMvs( cRefPoc, rcNav, rcLeft,      rcTree, rcMvPred.getPrecision(), rcReferenceMvs ) );
  RNOK  ( xInsertReferenceMvs( cRefPoc, rcNav, rcTop,       rcTree, rcMvPred.getPrecision(), rcReferenceMvs ) );

  const size_t uiNumRefMvBefore = rcReferenceMvs.size();
  RNOK  ( xInsertReferenceMvs( cRefPoc, rcNav, rcTopRight,  rcTree, rcMvPred.getPrecision(), rcReferenceMvs ) );
  if( rcReferenceMvs.size() == uiNumRefMvBefore )
  {
    RNOK( xInsertReferenceMvs( cRefPoc, rcNav, rcTopLeft,   rcTree, rcMvPred.getPrecision(), rcReferenceMvs ) );
  }

  // derive motion vector predictor
  if( rcReferenceMvs.size() == 0 )
  {
    rcMvPred = MotionVector( 0, 0, rcMvPred.getPrecision() );
  }
  else if( rcReferenceMvs.size() == 1 )
  {
    rcMvPred = rcReferenceMvs[0];
  }
  else
  {
    if( rcReferenceMvs.size() == 2 )
    {
      const int iX = ( rcReferenceMvs[0].getX() + rcReferenceMvs[1].getX() + 1 ) >> 1;
      const int iY = ( rcReferenceMvs[0].getY() + rcReferenceMvs[1].getY() + 1 ) >> 1;
      rcMvPred = MotionVector( iX, iY, rcMvPred.getPrecision() );
    }
    else
    {
      std::vector<int> aiX( rcReferenceMvs.size() ), aiY( rcReferenceMvs.size() );
      for( UInt ui = 0; ui < rcReferenceMvs.size(); ui++ )
      {
        aiX[ui] = rcReferenceMvs[ui].getX();
        aiY[ui] = rcReferenceMvs[ui].getY();
      }
      sort( aiX.begin(), aiX.end() );
      sort( aiY.begin(), aiY.end() );

      const UInt uiMedianElement = UInt( rcReferenceMvs.size()>>1 );
      rcMvPred = MotionVector( aiX[uiMedianElement], aiY[uiMedianElement], rcMvPred.getPrecision() );
    }
  }

  return Err::m_nOK;
}


ErrVal
InterTranslational::getMvPredXInd( const PicOrderCnt&              rcRefPoc, 
                                 const QTNavi&                   rcNav,
                                 const QuadtreeGroup<PredInfo*>& rcTree,
                                 const int                       iMvPrecision,
                                 int&                            riMvPredX )
{
  std::vector<MotionVector> acMvPredCandList;
  RNOK( xGetMvPredCandLstStd( rcRefPoc, rcNav, rcTree, iMvPrecision, acMvPredCandList ) );
  riMvPredX = xGetMvPredXCompInd( acMvPredCandList );
  return Err::m_nOK;
}

ErrVal
InterTranslational::getMvPredYInd( const PicOrderCnt&              rcRefPoc, 
                                 const QTNavi&                   rcNav,
                                 const QuadtreeGroup<PredInfo*>& rcTree,
                                 const int                       iMvPrecision,
                                 int&                            riMvPredY )
{
  std::vector<MotionVector> acMvPredCandList;
  RNOK( xGetMvPredCandLstStd( rcRefPoc, rcNav, rcTree, iMvPrecision, acMvPredCandList ) );
  riMvPredY = xGetMvPredYCompInd( acMvPredCandList );
  return Err::m_nOK;
}

ErrVal
InterTranslational::getMvPredXDep( const PicOrderCnt&              rcRefPoc, 
                                 const QTNavi&                   rcNav,
                                 const QuadtreeGroup<PredInfo*>& rcTree,
                                 const int                       iMvPrecision,
                                 const int                       iMvY,
                                 int&                            riMvPredX )
{
  std::vector<MotionVector> acMvPredCandList;
  std::vector<int>          acYThresList;
  std::vector<int>          acXPredList;
  RNOK( xGetMvPredCandLstExt( rcRefPoc, rcNav, rcTree, iMvPrecision, acMvPredCandList ) );
  RNOK( xSetYThresXPredLists( acMvPredCandList, acYThresList, acXPredList ) );
  riMvPredX = xGetMvPredXCompDep( acYThresList, acXPredList, iMvY );
  return Err::m_nOK;
}

ErrVal
InterTranslational::getMvPredYDep( const PicOrderCnt&              rcRefPoc, 
                                 const QTNavi&                   rcNav,
                                 const QuadtreeGroup<PredInfo*>& rcTree,
                                 const int                       iMvPrecision,
                                 const int                       iMvX,
                                 int&                            riMvPredY )
{
  std::vector<MotionVector> acMvPredCandList;
  std::vector<int>          acXThresList;
  std::vector<int>          acYPredList;
  RNOK( xGetMvPredCandLstExt( rcRefPoc, rcNav, rcTree, iMvPrecision, acMvPredCandList ) );
  RNOK( xSetXThresYPredLists( acMvPredCandList, acXThresList, acYPredList ) );
  riMvPredY = xGetMvPredYCompDep( acXThresList, acYPredList, iMvX );
  return Err::m_nOK;
}

ErrVal
InterTranslational::getYThresXPredLists( const PicOrderCnt&              rcRefPoc,
                                       const QTNavi&                   rcNav,
                                       const QuadtreeGroup<PredInfo*>& rcTree,
                                       const int                       iMvAccuracy,
                                       int&                            riYPred,
                                       std::vector<int>&               racYThresList,
                                       std::vector<int>&               racXPredList )
{
  racYThresList.clear();
  racXPredList .clear();
  std::vector<MotionVector> acMvPredCandYList;
  std::vector<MotionVector> acMvPredCandXList;
  RNOK(     xGetMvPredCandLstStd( rcRefPoc, rcNav, rcTree, iMvAccuracy, acMvPredCandYList ) );
  RNOK(     xGetMvPredCandLstExt( rcRefPoc, rcNav, rcTree, iMvAccuracy, acMvPredCandXList ) );
  RNOK(     xSetYThresXPredLists( acMvPredCandXList, racYThresList, racXPredList ) );
  riYPred = xGetMvPredYCompInd  ( acMvPredCandYList );
  return Err::m_nOK;
}

ErrVal
InterTranslational::getXThresYPredLists( const PicOrderCnt&              rcRefPoc,
                                       const QTNavi&                   rcNav,
                                       const QuadtreeGroup<PredInfo*>& rcTree,
                                       const int                       iMvAccuracy,
                                       int&                            riXPred,
                                       std::vector<int>&               racXThresList,
                                       std::vector<int>&               racYPredList )
{
  racXThresList.clear();
  racYPredList .clear();
  std::vector<MotionVector> acMvPredCandXList;
  std::vector<MotionVector> acMvPredCandYList;
  RNOK(     xGetMvPredCandLstStd( rcRefPoc, rcNav, rcTree, iMvAccuracy, acMvPredCandXList ) );
  RNOK(     xGetMvPredCandLstExt( rcRefPoc, rcNav, rcTree, iMvAccuracy, acMvPredCandYList ) );
  RNOK(     xSetXThresYPredLists( acMvPredCandYList, racXThresList, racYPredList ) );
  riXPred = xGetMvPredXCompInd  ( acMvPredCandXList );
  return Err::m_nOK;
}



ErrVal InterTranslational::getMvPredLst( PicOrderCnt cRefPoc,
                                       const QTNavi &rcNav,
                                       const QuadtreeGroup<PredInfo*> &rcTree,
                                       std::vector<MotionVector>& rcMvPredLst,
                                       int iMvPrecision )
{
  rcMvPredLst.clear();

  const PrimPixelPos  &rcNavPos       = rcNav.getPrimPixelPos();
  const PrimPixelPos  &rcLeftPos      = rcNavPos + PrimPixelPos( -1,                                0 );
  const PrimPixelPos  &rcTopPos       = rcNavPos + PrimPixelPos(  0,                               -1 );
  const PrimPixelPos  &rcTopRightPos  = rcNavPos + PrimPixelPos(  rcNav.getPrimBlockSize().getX(), -1 );
  const PrimPixelPos  &rcTopLeftPos   = rcNavPos + PrimPixelPos( -1                              , -1 );
  const QTNavi        &rcLeft         = rcNav.findLeafNode( rcLeftPos     );
  const QTNavi        &rcTop          = rcNav.findLeafNode( rcTopPos      );
  const QTNavi        &rcTopRight     = rcNav.findLeafNode( rcTopRightPos );
  const QTNavi        &rcTopLeft      = rcNav.findLeafNode( rcTopLeftPos  );
 
  RNOK  ( xInsertReferenceMvs( cRefPoc, rcNav, rcTop,       rcTree, iMvPrecision, rcMvPredLst, true ) );
  RNOK  ( xInsertReferenceMvs( cRefPoc, rcNav, rcLeft,      rcTree, iMvPrecision, rcMvPredLst, true ) );
  RNOK  ( xInsertReferenceMvs( cRefPoc, rcNav, rcTopRight,  rcTree, iMvPrecision, rcMvPredLst, true ) );
  if( rcMvPredLst.empty() )
  {
    RNOK( xInsertReferenceMvs( cRefPoc, rcNav, rcTopLeft,   rcTree, iMvPrecision, rcMvPredLst, true ) );
  }
  if( rcMvPredLst.empty() )
  {
    rcMvPredLst.push_back( MotionVector( 0, 0, iMvPrecision ) );
  }
  ROF( rcMvPredLst.size() <= 3 );
  return Err::m_nOK;
}


ErrVal InterTranslational::getNumMvPred( UInt uiRefIdx,
                                       const QTNavi &rcNav,
                                       const QuadtreeGroup<PredInfo*> &rcTree,
                                       UInt& ruiNumMvPred,
                                       UInt* auiCtx )
{
  const PrimPixelPos  &rcNavPos       = rcNav.getPrimPixelPos();
  const PrimPixelPos  &rcLeftPos      = rcNavPos + PrimPixelPos( -1,                                0 );
  const PrimPixelPos  &rcTopPos       = rcNavPos + PrimPixelPos(  0,                               -1 );
  const PrimPixelPos  &rcTopRightPos  = rcNavPos + PrimPixelPos(  rcNav.getPrimBlockSize().getX(), -1 );
  const QTNavi        &rcLeft         = rcNav.findLeafNode( rcLeftPos     );
  const QTNavi        &rcTop          = rcNav.findLeafNode( rcTopPos      );
  const QTNavi        &rcTopRight     = rcNav.findLeafNode( rcTopRightPos );
  bool          bLeft         = !rcLeft.liesOutside() && rcLeft.isCodedBefore( rcNav );
  bool          bTop          = !rcTop.liesOutside() && rcTop.isCodedBefore( rcNav );
  bool          bTopRight     = !rcTopRight.liesOutside() && rcTopRight.isCodedBefore( rcNav );

  if ( bLeft )
  {
    if ( rcLeft.isParsedBefore( rcNav ) )
    {
      RNOK( xIsRefIdxAvailable( uiRefIdx, rcNav, rcLeft, rcTree, bLeft ) );
    }
  }

  if ( bTop )
  {
    if ( rcTop.isParsedBefore( rcNav ) )
    {
      RNOK( xIsRefIdxAvailable( uiRefIdx, rcNav, rcTop, rcTree, bTop ) );
    }
  }

  if ( bTopRight )
  {
    if ( rcTopRight.isParsedBefore( rcNav ) )
    {
      RNOK( xIsRefIdxAvailable( uiRefIdx, rcNav, rcTopRight, rcTree, bTopRight ) );
    }
  }

  ruiNumMvPred = ( bLeft ? 1 : 0 ) + ( bTop ? 1 : 0 ) + ( bTopRight ? 1 : 0 );
  ruiNumMvPred = gMax( ruiNumMvPred, UInt(1) );
  auiCtx[0]    = ( bTop ? 0 : 1 );
  auiCtx[1]    = 1;
  return Err::m_nOK;
}

ErrVal InterTranslational::getMvD( const PredSyntaxInterTranslational& rcPredSyntax,
                                 UInt uiRefIdx,
                                 MotionVector& rcMvD ) // NOTE: motion vector precision is specified by the input rcMvD !!!!
{
  UInt uiNumFound = 0;
  rcMvD = MotionVector( 0, 0, rcMvD.getPrecision() );
  for( UInt ui = 0; ui < rcPredSyntax.getNumEntries(); ui++ )
  {
    if( rcPredSyntax.getRefIdx( ui ) == uiRefIdx )
    {
      rcMvD = rcMvD + rcPredSyntax.getMvDiff( ui ).getMvWithPrecisionModifiedTo( rcMvD.getPrecision() );
      uiNumFound++;
    }
  }
  if( uiNumFound > 1 )
  {
    const int iX =  gSign( rcMvD.getX() ) * ( abs( rcMvD.getX() << 1 ) + uiNumFound ) / ( uiNumFound << 1 );
    const int iY =  gSign( rcMvD.getY() ) * ( abs( rcMvD.getY() << 1 ) + uiNumFound ) / ( uiNumFound << 1 );
    rcMvD = MotionVector( iX, iY, rcMvD.getPrecision() );
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
