/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/IntraDC.h"
#include "VideoCommonLib/PicBlock.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


ErrVal
PredSyntaxIntraDC::xSetComplementData( const QTNavi&                 rcPos,
                                       const PlanePackageParameters& rcPlaPaPars,
                                       const AccessUnitParameters&   rcAUPars,
                                       const SliceData&              rcSliceData,
                                       const PlanePackage*           pcRefPlanePackage )
{
  return Err::m_nOK;
}


ErrVal IntraDC::calculatePrediction( const PredInfo &rcPredInfo,
                                     const QTNavi &rcNav,
                                     const PrimPixelPos&  cTopLeftPosOfPrediction,
                                     const PicBlock<const Pel> *pcEncPic,
                                     const CodingState &rcCodingState,
                                     const SliceData &rcCurrSlice,
                                     const PicBlock<DPel> &rcPredPicBlock,
                                     bool &rbPredictionValid )
{
  // DC prediction is always valid
  rbPredictionValid = true;

  const PrimBlockSize cPrimBlockSize = rcNav.getPrimBlockSize();
  const PrimPixelPos cPrimPixelPos = rcNav.getPrimPixelPos();

  ROF( rcPredPicBlock.getStats().getPrimBlockSize() == cPrimBlockSize );
  ROF( cPrimPixelPos == cTopLeftPosOfPrediction );

  for( UInt ui = 0; ui < pcEncPic->getStats().getNumPlanes(); ui++ )
  {
    const Block<const Pel> &rcEncPic    = pcEncPic     ->getPlane( ui );
    const Block<DPel>      &rcPredBlock = rcPredPicBlock.getPlane( ui );
    int iDCVal = 0;
    int iNumValues = 0;

    PixelPos  cPos  = cPrimPixelPos  * pcEncPic->getStats().getSubsampling()[ui];
    BlockSize cSize = cPrimBlockSize * pcEncPic->getStats().getSubsampling()[ui];

    // Sum over left neighboring pixel
    if( xHasLeft( rcNav ) )
    {
      AllocBlock<DPel> cLeftBlock;
      RNOK( xGetLeft( rcNav, false, cPos, cSize, rcEncPic, cLeftBlock ) );
      xCalcSumOverElements( cLeftBlock, iDCVal, iNumValues );
    }

    // Sum over top neighboring pixel
    if( xHasTop( rcNav ) )
    {
      AllocBlock<DPel> cTopBlock;
      RNOK( xGetTop( rcNav, false, cPos, cSize, rcEncPic, cTopBlock ) );
      xCalcSumOverElements( cTopBlock, iDCVal, iNumValues );
    }

    if( 0 == iNumValues )
    {
      iDCVal = 128 << EXTRA_PRECISION_BITS; // DC-Pred-value
    }
    else
    {
      // calculate average
      iDCVal <<= 1;
      iDCVal  += iNumValues;
      iDCVal  /= iNumValues << 1;
    }

    // assign prediction
    rcPredBlock.assign( iDCVal );
  }

  return Err::m_nOK;
}

template<class t_Pel>
void IntraDC::xCalcSumOverElements( const Block<const t_Pel> &rcBlock, int &riSum, int &riNumElements ) const
{
  const BlockSize& rcSize = rcBlock.getSize();
  riNumElements += rcSize.getX() * rcSize.getY();
  const t_Pel* pPel = &rcBlock.getEl(0);
  const int iStride = rcBlock.getStride();
  int iSum = 0;
  for( int iY = 0; iY < rcSize.getY(); iY++ )
  {
    for( int iX = 0; iX < rcSize.getX(); iX++ )
    {
      iSum += pPel[iX];
    }
    pPel += iStride;
  }
  riSum += iSum;
}


// video namespace end
VIDEO_NAMESPACE_END
