/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/IntraDirectional.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


ErrVal
PredSyntaxIntraDirectional::xSetComplementData( const QTNavi&                 rcPos,
                                                const PlanePackageParameters& rcPlaPaPars,
                                                const AccessUnitParameters&   rcAUPars,
                                                const SliceData&              rcSliceData,
                                                const PlanePackage*           pcRefPlanePackage )
{
  return Err::m_nOK;
}


ErrVal IntraDirectional::calculatePrediction( const PredInfo &rcPredInfo,
                                              const QTNavi &rcNav,
                                              const PrimPixelPos&  cTopLeftPosOfPrediction,
                                              const PicBlock<const Pel> *pcEncPic,
                                              const CodingState &rcCodingState,
                                              const SliceData &rcCurrSlice,
                                              const PicBlock<DPel> &rcPredPicBlock,
                                              bool &rbPredictionValid )
{
  rbPredictionValid = false;
  const PredInfoIntraDirectional *pcPredInfo = rcPredInfo.getPredInfoIntraDirectional();

  ROF( rcPredPicBlock.getStats().getPrimBlockSize() == rcNav.getPrimBlockSize() );
  ROF( rcNav.getPrimPixelPos() == cTopLeftPosOfPrediction );

  const bool bSmoothRefPixel = pcPredInfo->getSmoothedRefPixelFlag();

  switch( pcPredInfo->getIntraDirection() )
  {
  case PredInfoIntraDirectional::INTRA_WEST_SOUTHWEST:
    RNOK( xPredWestSouthwest( rcNav, pcEncPic, rcPredPicBlock, rbPredictionValid, bSmoothRefPixel ) );
    break;
  case PredInfoIntraDirectional::INTRA_WEST:
    RNOK( xPredWest( rcNav, pcEncPic, rcPredPicBlock, rbPredictionValid, bSmoothRefPixel ) );
    break;
  case PredInfoIntraDirectional::INTRA_WEST_NORTHWEST:
    RNOK( xPredWestNorthwest( rcNav, pcEncPic, rcPredPicBlock, rbPredictionValid, bSmoothRefPixel ) );
    break;
  case PredInfoIntraDirectional::INTRA_NORTHWEST:
    RNOK( xPredNorthwest( rcNav, pcEncPic, rcPredPicBlock, rbPredictionValid, bSmoothRefPixel ) );
    break;
  case PredInfoIntraDirectional::INTRA_NORTH_NORTHWEST:
    RNOK( xPredNorthNorthwest( rcNav, pcEncPic, rcPredPicBlock, rbPredictionValid, bSmoothRefPixel ) );
    break;
  case PredInfoIntraDirectional::INTRA_NORTH:
    RNOK( xPredNorth( rcNav, pcEncPic, rcPredPicBlock, rbPredictionValid, bSmoothRefPixel ) );
    break;
  case PredInfoIntraDirectional::INTRA_NORTH_NORTHEAST:
    RNOK( xPredNorthNortheast( rcNav, pcEncPic, rcPredPicBlock, rbPredictionValid, bSmoothRefPixel ) );
    break;
  case PredInfoIntraDirectional::INTRA_NORTHEAST:
    RNOK( xPredNortheast( rcNav, pcEncPic, rcPredPicBlock, rbPredictionValid, bSmoothRefPixel ) );
    break;
  default:
    ROT( 1 );
  }
  return Err::m_nOK;
}


ErrVal IntraDirectional::xPredWestSouthwest( const QTNavi                &rcNav,
                                             const PicBlock<const Pel>   *pcEncPic,
                                             const PicBlock<DPel>        &rcPredPicBlock,
                                             bool                        &rbPredictionValid,
                                             bool                         bSmoothRefPixel ) const
{
  rbPredictionValid = false;
  ROFRS( xHasLeft( rcNav ), Err::m_nOK );

  const PrimBlockSize cPrimBlockSize = rcNav.getPrimBlockSize();
  const PrimPixelPos cPrimPixelPos = rcNav.getPrimPixelPos();
  for( UInt ui = 0; ui < rcPredPicBlock.getStats().getNumPlanes(); ui++ )
  {
    const PixelPos  cCurrPos  = cPrimPixelPos  * rcPredPicBlock.getStats().getSubsampling()[ui];
    const BlockSize cCurrSize = cPrimBlockSize * rcPredPicBlock.getStats().getSubsampling()[ui];
    
    AllocBlock<DPel> cLeft;
    RNOK( xGetLeft( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ), cLeft ) );
    AllocBlock<DPel> cFilteredLine( BlockSize( 1, cCurrSize.getY() + cCurrSize.getX() / 2 ) );

    const int iMaxYInLeft = cLeft.getSize().getY() - 1;
    // create bilinearly filtered pixel
    for( int i = 0; i < (int)cFilteredLine.getSize().getY(); i++ )
    {
      cFilteredLine.getEl( PixelPos( 0, i ) ) = ( (int)cLeft.getEl( PixelPos( 0, gMin( i    , iMaxYInLeft ) ) ) +
                                                  (int)cLeft.getEl( PixelPos( 0, gMin( i + 1, iMaxYInLeft ) ) ) + 1 ) >> 1;
    }

    // assign bilinearly filtered pixel to even columns
    const Block<DPel> &rcPredBlock = rcPredPicBlock.getPlane( ui );
    for( int iX = 0, k = 0; iX < rcPredBlock.getSize().getX(); iX+=2, k++ )
    {
      Block<DPel>       cDst( PixelPos( iX, 0 ), BlockSize( 1, rcPredBlock.getSize().getY() ), rcPredBlock   );
      Block<const DPel> cSrc( PixelPos( 0,  k ), BlockSize( 1, cLeft.getSize().getY()       ), cFilteredLine );
      RNOK( cDst.assign( cSrc ) );
    }

    // create 1-2-1 filtered pixel
    for( int i = 0; i < (int)cFilteredLine.getSize().getY(); i++ )
    {
      cFilteredLine.getEl( PixelPos( 0, i ) ) = (    (int)cLeft.getEl( PixelPos( 0, gMin( i    , iMaxYInLeft ) ) ) +
                                                 2 * (int)cLeft.getEl( PixelPos( 0, gMin( i + 1, iMaxYInLeft ) ) ) +
                                                     (int)cLeft.getEl( PixelPos( 0, gMin( i + 2, iMaxYInLeft ) ) ) + 2 ) >> 2;
    }

    // assign 1-2-1 filtered pixel to odd columns
    for( int iX = 1, k = 0; iX < rcPredBlock.getSize().getX(); iX+=2, k++ )
    {
      Block<DPel>       cDst( PixelPos( iX, 0 ), BlockSize( 1, rcPredBlock.getSize().getY() ), rcPredBlock   );
      Block<const DPel> cSrc( PixelPos( 0,  k ), BlockSize( 1, cLeft.getSize().getY()       ), cFilteredLine );
      RNOK( cDst.assign( cSrc ) );
    }
  }
  rbPredictionValid = true;
  return Err::m_nOK;
}


ErrVal IntraDirectional::xPredWest( const QTNavi                &rcNav,
                                    const PicBlock<const Pel>   *pcEncPic,
                                    const PicBlock<DPel>        &rcPredPicBlock,
                                    bool                        &rbPredictionValid,
                                    bool                         bSmoothRefPixel ) const
{
  rbPredictionValid = false;
  ROFRS( xHasLeft( rcNav ), Err::m_nOK );

  const PrimBlockSize cPrimBlockSize = rcNav.getPrimBlockSize();
  const PrimPixelPos cPrimPixelPos = rcNav.getPrimPixelPos();
  for( UInt ui = 0; ui < rcPredPicBlock.getStats().getNumPlanes(); ui++ )
  {
    AllocBlock<DPel> cLeft;
    RNOK( xGetLeft( rcNav, bSmoothRefPixel, 
                    cPrimPixelPos  * rcPredPicBlock.getStats().getSubsampling()[ui],
                    cPrimBlockSize * rcPredPicBlock.getStats().getSubsampling()[ui],
                    pcEncPic->getPlane( ui ), cLeft ) );
    const Block<DPel> &rcPredBlock = rcPredPicBlock.getPlane( ui );
    for( int iX = 0; iX < rcPredBlock.getSize().getX(); iX++ )
    {
      Block<DPel> cDst( PixelPos( iX, 0 ), BlockSize( 1, rcPredBlock.getSize().getY() ), rcPredBlock );
      RNOK( cDst.assign( cLeft ) );
    }
  }
  rbPredictionValid = true;
  return Err::m_nOK;
}


ErrVal IntraDirectional::xPredWestNorthwest( const QTNavi                &rcNav,
                                             const PicBlock<const Pel>   *pcEncPic,
                                             const PicBlock<DPel>        &rcPredPicBlock,
                                             bool                        &rbPredictionValid,
                                             bool                         bSmoothRefPixel ) const
{
  rbPredictionValid = false;
  ROFRS( xHasLeft( rcNav ) && xHasTop( rcNav ) && xHasTopLeft( rcNav ), Err::m_nOK );

  const PrimBlockSize cPrimBlockSize = rcNav.getPrimBlockSize();
  const PrimPixelPos cPrimPixelPos = rcNav.getPrimPixelPos();
  for( UInt ui = 0; ui < rcPredPicBlock.getStats().getNumPlanes(); ui++ )
  {
    const PixelPos  cCurrPos  = cPrimPixelPos  * rcPredPicBlock.getStats().getSubsampling()[ui];
    const BlockSize cCurrSize = cPrimBlockSize * rcPredPicBlock.getStats().getSubsampling()[ui];

    AllocBlock<DPel> cLeft;
    RNOK( xGetLeft( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ), cLeft ) );
    const DPel cTopLeft = xGetTopLeft( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ) );
    AllocBlock<DPel> cFilteredLineV( BlockSize( 1, cCurrSize.getY() ) );

    // create bilinearly filtered pixel
    cFilteredLineV.getEl( 0 ) = ( (int)cTopLeft + (int)cLeft.getEl( 0 ) + 1 ) >> 1;
    for( int i = 1; i < (int)cFilteredLineV.getSize().getY(); i++ )
    {
      cFilteredLineV.getEl( PixelPos( 0, i ) ) = ( (int)cLeft.getEl( PixelPos( 0, i - 1 ) ) +
                                                   (int)cLeft.getEl( PixelPos( 0, i     ) ) + 1 ) >> 1;
    }

    // assign bilinearly filtered pixel to even columns
    const Block<DPel> &rcPredBlock = rcPredPicBlock.getPlane( ui );
    for( int iX = 0, k = 0; iX < rcPredBlock.getSize().getX() && k != cLeft.getSize().getY(); iX+=2, k++ )
    {
      Block<DPel>       cDst( PixelPos( iX, k ), BlockSize( 1, rcPredBlock.getSize().getY() - k ), rcPredBlock    );
      Block<const DPel> cSrc( PixelPos( 0,  0 ), BlockSize( 1, cLeft.getSize().getY()       - k ), cFilteredLineV );
      RNOK( cDst.assign( cSrc ) );
    }

    // create 1-2-1 filtered pixel
    cFilteredLineV.getEl( 0 ) = (    (int)cTopLeft +
                                 2 * (int)cLeft.getEl( PixelPos( 0, 0 ) ) +
                                     (int)cLeft.getEl( PixelPos( 0, gMin( 1, cLeft.getSize().getY() - 1 ) ) ) + 2 ) >> 2;
    for( int i = 1; i < (int)cLeft.getSize().getY() - 1; i++ )
    {
      cFilteredLineV.getEl( PixelPos( 0, i ) ) = (     (int)cLeft.getEl( PixelPos( 0, i-1 ) ) +
                                                   2 * (int)cLeft.getEl( PixelPos( 0, i   ) ) +
                                                       (int)cLeft.getEl( PixelPos( 0, i+1 ) ) + 2 ) >> 2;
    }

    // assign 1-2-1 filtered pixel to odd columns
    for( int iX = 1, k = 1; iX < rcPredBlock.getSize().getX() && k != cLeft.getSize().getY(); iX+=2, k++ )
    {
      Block<DPel>       cDst( PixelPos( iX, k ), BlockSize( 1, rcPredBlock.getSize().getY() - k ), rcPredBlock    );
      Block<const DPel> cSrc( PixelPos( 0,  0 ), BlockSize( 1, cLeft.getSize().getY()       - k ), cFilteredLineV );
      RNOK( cDst.assign( cSrc ) );
    }

    if( cCurrSize.getX() > 1 )
    {
      AllocBlock<DPel> cTop;
      RNOK( xGetTop( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ), cTop ) );
      AllocBlock<DPel> cFilteredLineH( BlockSize( cTop.getSize().getX() - 1, 1 ) );

      // create 1-2-1 filtered pixel
      cFilteredLineH.getEl( 0 ) = ( (int)cLeft.getEl( PixelPos( 0, 0 ) ) + 2 * (int)cTopLeft + (int)cTop.getEl( PixelPos( 0, 0 ) ) + 2 ) >> 2;
      if( cFilteredLineH.getSize().getX() > 1 )
      {
        cFilteredLineH.getEl( 1 ) = ( (int)cTopLeft + 2 * (int)cTop.getEl( PixelPos( 0, 0 ) )  + (int)cTop.getEl( PixelPos( gMin( 1, cTop.getSize().getX() - 1), 0 ) ) + 2 ) >> 2;
      }
      for( int i = 2; i < (int)cFilteredLineH.getSize().getX(); i++ )
      {
        cFilteredLineH.getEl( PixelPos( i, 0 ) ) = (    (int)cTop.getEl( PixelPos( i-2, 0 ) ) +
                                                    2 * (int)cTop.getEl( PixelPos( i-1, 0 ) ) +
                                                        (int)cTop.getEl( PixelPos( i,   0 ) ) + 2 ) >> 2;
      }

      // assign 1-2-1 filtered horizontal pixel to remaining rows
      for( int iX = 1, k = 0; iX < rcPredBlock.getSize().getX() && k != rcPredBlock.getSize().getY(); iX+=2, k++ )
      {
        Block<DPel>       cDst( PixelPos( iX, k ), BlockSize( rcPredBlock.getSize().getX() - iX, 1 ), rcPredBlock    );
        Block<const DPel> cSrc( PixelPos( 0,  0 ), BlockSize( rcPredBlock.getSize().getX() - iX, 1 ), cFilteredLineH );
        RNOK( cDst.assign( cSrc ) );
      }
    }
  }
  rbPredictionValid = true;
  return Err::m_nOK;
}


ErrVal IntraDirectional::xPredNorthwest( const QTNavi                &rcNav,
                                         const PicBlock<const Pel>   *pcEncPic,
                                         const PicBlock<DPel>        &rcPredPicBlock,
                                         bool                        &rbPredictionValid,
                                         bool                         bSmoothRefPixel ) const
{
  rbPredictionValid = false;
  ROFRS( xHasLeft( rcNav ) && xHasTop( rcNav ) && xHasTopLeft( rcNav ), Err::m_nOK );

  const PrimBlockSize cPrimBlockSize = rcNav.getPrimBlockSize();
  const PrimPixelPos cPrimPixelPos = rcNav.getPrimPixelPos();
  for( UInt ui = 0; ui < rcPredPicBlock.getStats().getNumPlanes(); ui++ )
  {
    const PixelPos  cCurrPos  = cPrimPixelPos  * rcPredPicBlock.getStats().getSubsampling()[ui];
    const BlockSize cCurrSize = cPrimBlockSize * rcPredPicBlock.getStats().getSubsampling()[ui];

    AllocBlock<DPel> cLeft;
    RNOK( xGetLeft( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ), cLeft ) );
    AllocBlock<DPel> cTop;
    RNOK( xGetTop( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ), cTop ) );
    const DPel cTopLeft = xGetTopLeft( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ) );

    // collect predline
    AllocBlock<DPel> cPredLine( BlockSize( cCurrSize.getX() + cCurrSize.getY() + 1, 1 ) );
    for( int i = 0, k = cCurrSize.getY() - 1; k >= 0; i++, k-- )
    {
      cPredLine.getEl( i ) = cLeft.getEl( PixelPos( 0, k ) );
    }
    cPredLine.getEl( cCurrSize.getY() ) = cTopLeft;

    {
      const Block<DPel> cDst( PixelPos( cCurrSize.getY() + 1, 0 ), BlockSize( cCurrSize.getX(), 1 ), cPredLine );
      RNOK( cDst.assign( cTop ) );
    }

    // filter predline
    AllocBlock<DPel> cPredLineFiltered( BlockSize( cCurrSize.getX() + cCurrSize.getY() - 1, 1 ) );
    for( int i = 0; i < (int)cPredLineFiltered.getSize().getX(); i++ )
    {
      cPredLineFiltered.getEl( i ) = ( (int)cPredLine.getEl( i ) + 2 * (int)cPredLine.getEl( i+1 ) + (int)cPredLine.getEl( i+2 ) + 2 ) >> 2;
    }

    // apply northwest
    const Block<DPel> &rcPredBlock = rcPredPicBlock.getPlane( ui );
    for( int iX = cCurrSize.getY() - 1, iY = 0; iX >= 0; iX--, iY++ )
    {
      Block<DPel>       cDst( PixelPos( 0, iY ), BlockSize( cCurrSize.getX(), 1 ), rcPredBlock       );
      Block<const DPel> cSrc( PixelPos( iX, 0 ), BlockSize( cCurrSize.getX(), 1 ), cPredLineFiltered );
      RNOK( cDst.assign( cSrc ) );
    }
  }
  rbPredictionValid = true;
  return Err::m_nOK;
}


ErrVal IntraDirectional::xPredNorthNorthwest( const QTNavi                &rcNav,
                                              const PicBlock<const Pel>   *pcEncPic,
                                              const PicBlock<DPel>        &rcPredPicBlock,
                                              bool                        &rbPredictionValid,
                                              bool                         bSmoothRefPixel ) const
{
  rbPredictionValid = false;
  ROFRS( xHasTop( rcNav ) && xHasLeft( rcNav ) && xHasTopLeft( rcNav ), Err::m_nOK );

  const PrimBlockSize cPrimBlockSize = rcNav.getPrimBlockSize();
  const PrimPixelPos cPrimPixelPos = rcNav.getPrimPixelPos();
  for( UInt ui = 0; ui < rcPredPicBlock.getStats().getNumPlanes(); ui++ )
  {
    const PixelPos  cCurrPos  = cPrimPixelPos  * rcPredPicBlock.getStats().getSubsampling()[ui];
    const BlockSize cCurrSize = cPrimBlockSize * rcPredPicBlock.getStats().getSubsampling()[ui];

    AllocBlock<DPel> cTop;
    RNOK( xGetTop( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ), cTop ) );
    const DPel cTopLeft = xGetTopLeft( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ) );
    const AllocBlock<DPel> cFilteredLineH( BlockSize( cCurrSize.getX(), 1 ) );

    // create bilinearly filtered pixel
    cFilteredLineH.getEl( 0 ) = ( (int)cTopLeft + (int)cTop.getEl( 0 ) + 1 ) >> 1;
    for( int i = 1; i < (int)cFilteredLineH.getSize().getX(); i++ )
    {
      cFilteredLineH.getEl( PixelPos( i, 0 ) ) = ( (int)cTop.getEl( PixelPos( i - 1, 0 ) ) +
                                                   (int)cTop.getEl( PixelPos( i,     0 ) ) + 1 ) >> 1;
    }

    // assign bilinearly filtered pixel to even rows
    const Block<DPel> &rcPredBlock = rcPredPicBlock.getPlane( ui );
    for( int iY = 0, k = 0; iY < rcPredBlock.getSize().getY() && k != cTop.getSize().getX(); iY+=2, k++ )
    {
      Block<DPel>       cDst( PixelPos( k, iY ), BlockSize( rcPredBlock.getSize().getX() - k, 1 ), rcPredBlock    );
      Block<const DPel> cSrc( PixelPos( 0,  0 ), BlockSize( cTop.getSize().getX()        - k, 1 ), cFilteredLineH );
      RNOK( cDst.assign( cSrc ) );
    }

    // create 1-2-1 filtered pixel
    cFilteredLineH.getEl( 0 ) = ( (int)cTopLeft + 2 * (int)cTop.getEl( PixelPos( 0, 0 ) ) + (int)cTop.getEl( PixelPos( gMin( 1, cTop.getSize().getX() - 1), 0 ) ) + 2 ) >> 2;
    for( int i = 1; i < (int)cTop.getSize().getX() - 1; i++ )
    {
      cFilteredLineH.getEl( PixelPos( i, 0 ) ) = (    (int)cTop.getEl( PixelPos( i-1, 0 ) ) +
                                                  2 * (int)cTop.getEl( PixelPos( i,   0 ) ) +
                                                      (int)cTop.getEl( PixelPos( i+1, 0 ) ) + 2 ) >> 2;
    }

    // assign 1-2-1 filtered pixel to odd rows
    for( int iY = 1, k = 1; iY < rcPredBlock.getSize().getY() && k != cTop.getSize().getX(); iY+=2, k++ )
    {
      const Block<DPel>       cDst( PixelPos( k, iY ), BlockSize( rcPredBlock.getSize().getX() - k, 1 ), rcPredBlock    );
      const Block<const DPel> cSrc( PixelPos( 0,  0 ), BlockSize( cTop.getSize().getX()        - k, 1 ), cFilteredLineH );
      RNOK( cDst.assign( cSrc ) );
    }

    if( cCurrSize.getY() > 1 )
    {
      AllocBlock<DPel> cLeft;
      RNOK( xGetLeft( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ), cLeft ) );
      const AllocBlock<DPel> cFilteredLineV( BlockSize( 1, cLeft.getSize().getY() - 1 ) );

      // create 1-2-1 filtered pixel
      cFilteredLineV.getEl( PixelPos( 0, 0 ) ) = ( (int)cTop.getEl( PixelPos( 0, 0 ) ) + 2 * (int)cTopLeft  + (int)cLeft.getEl( PixelPos( 0, 0 ) ) + 2 ) >> 2;
      if( cFilteredLineV.getSize().getY() > 1 )
      {
        cFilteredLineV.getEl( PixelPos( 0, 1 ) ) = (    (int)cTopLeft +
                                                    2 * (int)cLeft.getEl( PixelPos( 0, 0 ) ) +
                                                        (int)cLeft.getEl( PixelPos( 0, gMin( 1, cLeft.getSize().getY() - 1) ) ) + 2 ) >> 2;
      }

      for( int i = 2; i < (int)cFilteredLineV.getSize().getY(); i++ )
      {
        cFilteredLineV.getEl( PixelPos( 0, i ) ) = (    (int)cLeft.getEl( PixelPos( 0, i-2 ) ) +
                                                    2 * (int)cLeft.getEl( PixelPos( 0, i-1 ) ) +
                                                        (int)cLeft.getEl( PixelPos( 0, i   ) ) + 2 ) >> 2;
      }

      // assign 1-2-1 filtered vertical pixel to remaining columns
      for( int iY = 1, k = 0; iY < rcPredBlock.getSize().getY() && k != rcPredBlock.getSize().getX(); iY+=2, k++ )
      {
        const Block<DPel>       cDst( PixelPos( k, iY ), BlockSize( 1, rcPredBlock.getSize().getY() - iY ), rcPredBlock    );
        const Block<const DPel> cSrc( PixelPos( 0, 0  ), BlockSize( 1, rcPredBlock.getSize().getY() - iY ), cFilteredLineV );
        RNOK( cDst.assign( cSrc ) );
      }
    }
  }
  rbPredictionValid = true;
  return Err::m_nOK;
}


ErrVal IntraDirectional::xPredNorth( const QTNavi                &rcNav,
                                     const PicBlock<const Pel>   *pcEncPic,
                                     const PicBlock<DPel>        &rcPredPicBlock,
                                     bool                        &rbPredictionValid,
                                     bool                         bSmoothRefPixel ) const
{
  rbPredictionValid = false;
  ROFRS( xHasTop( rcNav ), Err::m_nOK );

  const PrimBlockSize cPrimBlockSize = rcNav.getPrimBlockSize();
  const PrimPixelPos cPrimPixelPos = rcNav.getPrimPixelPos();
  for( UInt ui = 0; ui < rcPredPicBlock.getStats().getNumPlanes(); ui++ )
  {
    AllocBlock<DPel> cTop;
    RNOK( xGetTop( rcNav, bSmoothRefPixel, 
                   cPrimPixelPos  * rcPredPicBlock.getStats().getSubsampling()[ui],
                   cPrimBlockSize * rcPredPicBlock.getStats().getSubsampling()[ui],
                   pcEncPic->getPlane( ui ), cTop ) );
    const Block<DPel> &rcPredBlock = rcPredPicBlock.getPlane( ui );
    for( int iY = 0; iY < rcPredBlock.getSize().getY(); iY++ )
    {
      Block<DPel> cDst( PixelPos( 0, iY ), BlockSize( rcPredBlock.getSize().getX(), 1 ), rcPredBlock );
      RNOK( cDst.assign( cTop ) );
    }
  }
  rbPredictionValid = true;
  return Err::m_nOK;
}


ErrVal IntraDirectional::xPredNorthNortheast( const QTNavi                &rcNav,
                                              const PicBlock<const Pel>   *pcEncPic,
                                              const PicBlock<DPel>        &rcPredPicBlock,
                                              bool                        &rbPredictionValid,
                                              bool                         bSmoothRefPixel ) const
{
  rbPredictionValid = false;
  ROFRS( xHasTop( rcNav ), Err::m_nOK );

  const PrimBlockSize cPrimBlockSize = rcNav.getPrimBlockSize();
  const PrimPixelPos cPrimPixelPos = rcNav.getPrimPixelPos();
  for( UInt ui = 0; ui < rcPredPicBlock.getStats().getNumPlanes(); ui++ )
  {
    const Subsampling &rcSub = rcPredPicBlock.getStats().getSubsampling()[ui];
    const PixelPos  cCurrPos       = cPrimPixelPos  * rcSub;
    const BlockSize cCurrSize      = cPrimBlockSize * rcSub;

    AllocBlock<DPel> cTop;
    RNOK( xGetTop( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ), cTop ) );

    AllocBlock<DPel> cPredLine( BlockSize( cCurrSize.getX() + cCurrSize.getY() / 2 + 1, 1 ) );
    {
      Block<DPel> cDst( PixelPos( 0, 0 ), cTop.getSize(), cPredLine );
      RNOK( cDst.assign( cTop ) );
    }

    int iPaddingStart = cCurrSize.getX();
    if( xHasTopRight( rcNav ) )
    {
      BlockSize cTopRightSize = rcNav.getTopRight().getPrimBlockSize() * rcSub;
      cTopRightSize.getX() = gMin( cTopRightSize.getX(), cCurrSize.getY() / 2 + 1 );
      AllocBlock<DPel> cTopRight;
      RNOK( xGetTopRight( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, cTopRightSize, pcEncPic->getPlane( ui ), cTopRight ) );
      Block<DPel> cDst( PixelPos( cCurrSize.getX(), 0 ), BlockSize( cTopRightSize.getX(), 1 ), cPredLine );
      RNOK( cDst.assign( cTopRight ) );
      iPaddingStart += cTopRightSize.getX();
    }

    // padding
    if( (int)cPredLine.getSize().getX() > iPaddingStart )
    {
      Block<DPel> cDst( PixelPos( iPaddingStart, 0 ), BlockSize( cPredLine.getSize().getX() - iPaddingStart, 1 ), cPredLine );
      cDst.assign( cPredLine.getEl( iPaddingStart - 1 ) );
    }

    // bilinearly filter pixel
    AllocBlock<DPel> cFilteredPredLine( BlockSize( cPredLine.getSize().getX() - 1, 1 ) );
    for( int i = 0; i < (int)cFilteredPredLine.getSize().getX(); i++ )
    {
      cFilteredPredLine.getEl( i ) = ( (int)cPredLine.getEl( i ) + (int)cPredLine.getEl( i + 1 ) + 1 ) >> 1;
    }

    const Block<DPel> &rcPredBlock = rcPredPicBlock.getPlane( ui );

    // assign bilinearly filtered line to even rows
    for( int i = 0, k = 0; i < cCurrSize.getY(); i+=2, k++ )
    {
      Block<DPel>       cDst( PixelPos( 0, i ), BlockSize( rcPredBlock.getSize().getX(), 1 ), rcPredBlock       );
      Block<const DPel> cSrc( PixelPos( k, 0 ), BlockSize( rcPredBlock.getSize().getX(), 1 ), cFilteredPredLine );
      RNOK( cDst.assign( cSrc ) );
    }

    // apply 1-2-1 filter pixel
    for( int i = 0; i < (int)cFilteredPredLine.getSize().getX() - 1; i++ )
    {
      cFilteredPredLine.getEl( i ) = ( (int)cPredLine.getEl( i ) + 2 * (int)cPredLine.getEl( i + 1 ) + (int)cPredLine.getEl( i + 2 ) + 2 ) >> 2;
    }

    // assign 1-2-1 filtered line to odd rows
    for( int i = 1, k = 0; i < cCurrSize.getY(); i+=2, k++ )
    {
      Block<DPel>       cDst( PixelPos( 0, i ), BlockSize( rcPredBlock.getSize().getX(), 1 ), rcPredBlock       );
      Block<const DPel> cSrc( PixelPos( k, 0 ), BlockSize( rcPredBlock.getSize().getX(), 1 ), cFilteredPredLine );
      RNOK( cDst.assign( cSrc ) );
    }
  }
  rbPredictionValid = true;
  return Err::m_nOK;
}


ErrVal IntraDirectional::xPredNortheast( const QTNavi                &rcNav,
                                         const PicBlock<const Pel>   *pcEncPic,
                                         const PicBlock<DPel>        &rcPredPicBlock,
                                         bool                        &rbPredictionValid,
                                         bool                         bSmoothRefPixel ) const
{
  rbPredictionValid = false;
  ROFRS( xHasTop( rcNav ), Err::m_nOK );

  const PrimBlockSize cPrimBlockSize = rcNav.getPrimBlockSize();
  const PrimPixelPos cPrimPixelPos = rcNav.getPrimPixelPos();
  for( UInt ui = 0; ui < rcPredPicBlock.getStats().getNumPlanes(); ui++ )
  {
    const Subsampling &rcSub = rcPredPicBlock.getStats().getSubsampling()[ui];
    const PixelPos cCurrPos = cPrimPixelPos * rcSub;
    const BlockSize cCurrSize = cPrimBlockSize * rcSub;

    AllocBlock<DPel> cTop;
    RNOK( xGetTop( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, pcEncPic->getPlane( ui ), cTop ) );

    // collect predline
    AllocBlock<DPel> cPredLine( BlockSize( cCurrSize.getX() + cCurrSize.getY() + 1, 1 ) );
    {
      Block<DPel> cDst( PixelPos( 0, 0 ), BlockSize( cCurrSize.getX(), 1 ), cPredLine );
      RNOK( cDst.assign( cTop ) );
    }

    int iUninitPixelXStart;
    if( xHasTopRight( rcNav ) )
    {
      BlockSize cTopRightSize = rcNav.getTopRight().getPrimBlockSize() * rcSub;
      cTopRightSize.getX() = gMin( cTopRightSize.getX(), cCurrSize.getY() + 1 );

      AllocBlock<DPel> cTopRight;
      RNOK( xGetTopRight( rcNav, bSmoothRefPixel, cCurrPos, cCurrSize, cTopRightSize, pcEncPic->getPlane( ui ), cTopRight ) );
      Block<DPel> cDst( PixelPos( cCurrSize.getX(), 0 ), BlockSize( cTopRightSize.getX(), 1 ), cPredLine );
      RNOK( cDst.assign( cTopRight ) );
      iUninitPixelXStart = cCurrSize.getX() + cTopRightSize.getX();
    }
    else
    {
      iUninitPixelXStart = cCurrSize.getX();
    }
    
    ROT( iUninitPixelXStart == 0 );
    for( int i = iUninitPixelXStart; i < (int)cPredLine.getSize().getX(); i++ )
    {
      // repeat missing pixel
      cPredLine.getEl( i ) = cPredLine.getEl( i - 1 );
    }

    // filter predline
    AllocBlock<DPel> cPredLineFiltered( BlockSize( cCurrSize.getX() + cCurrSize.getY() - 1, 1 ) );
    for( int i = 0; i < (int)cPredLineFiltered.getSize().getX(); i++ )
    {
      cPredLineFiltered.getEl( i ) = ( (int)cPredLine.getEl( i ) + 2 * (int)cPredLine.getEl( i+1 ) + (int)cPredLine.getEl( i+2 ) + 2 ) >> 2;
    }

    // apply northeast
    const Block<DPel> &rcPredBlock = rcPredPicBlock.getPlane( ui );
    for( int i = 0; i < (int)rcPredBlock.getSize().getY(); i++ )
    {
      Block<DPel>       cDst( PixelPos( 0, i ), BlockSize( cCurrSize.getX(), 1 ), rcPredBlock       );
      Block<const DPel> cSrc( PixelPos( i, 0 ), BlockSize( cCurrSize.getX(), 1 ), cPredLineFiltered );
      RNOK( cDst.assign( cSrc ) );
    }
  }
  rbPredictionValid = true;
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
