/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/ModePictureGenerator.h"
#include "VideoCommonLib/PicBlock.h"
#include "VideoCommonLib/IntraDirectional.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

ErrVal ModePictureGenerator::generatePicture( const Block<const Pel> &rcBackgroundPicture, const PicBlock<Pel> &rcModePicture, const Subsampling &rcPrimPlaneSubsampling )
{
  AOF( rcModePicture.getStats().getNumPlanes() == 3 );

  m_pcColorMap = new AllocBlock<Pel>( rcBackgroundPicture.getSize() );
  m_pcColorMap->assign( 0 ); // set background color ( 255 = transparent )

  SliceDataPlotter cSliceDataPlotter( *this, m_rcMPGParams, m_rcSliceData );
  cSliceDataPlotter.processTree();
  
  int iSizeX = rcBackgroundPicture.getSize().getX();
  int iSizeY = rcBackgroundPicture.getSize().getY();
  int iDenomX = 1<<rcPrimPlaneSubsampling.getX().getDownShift();
  int iDenomY = 1<<rcPrimPlaneSubsampling.getY().getDownShift();
  rcModePicture.getPlane( 0 ).assign( rcBackgroundPicture );
  rcModePicture.getPlane( 1 ).assign( 128 );
  rcModePicture.getPlane( 2 ).assign( 128 );

  for ( UInt uiPlane = 0; uiPlane < 3; uiPlane++ )
  {
    Block<Pel> cModePicturePlane = rcModePicture.getPlane( uiPlane );
    for ( int y = 0; y < iSizeY; y += iDenomY )
    {
      for ( int x = 0; x < iSizeX; x += iDenomX )
      {
        Pel cPel = m_pcColorMap->getEl( PixelPos( x / iDenomX, y / iDenomY ) );
        Pel cAlpha = m_rcMPGParams.getAlpha();
        for ( int uiXOffset = 0; uiXOffset < iDenomX; uiXOffset++ )
        {
          for ( int uiYOffset = 0; uiYOffset < iDenomY; uiYOffset++ )
          {
            cModePicturePlane.getEl( PixelPos( x + uiXOffset, y + uiYOffset ) ) = ( cAlpha*(m_aacPaletteMap[ cPel ])[uiPlane] + ( 255 - cAlpha ) * cModePicturePlane.getEl( PixelPos( x + uiXOffset, y + uiYOffset ) ) + 128 ) / 255;
          }
        }
      }
    }
  }

  delete m_pcColorMap;
  return Err::m_nOK;
}

ErrVal ModePictureGenerator::SliceDataPlotter::xProcessLeafNode( const QTNavi &rcPos )
{
  const PredInfo *pcPredInfo = m_rcSliceData.getPredInfoTree().getT( rcPos );
  const ResInfo *pcResInfo = m_rcSliceData.getResInfoTree().getT( rcPos );

  if ( m_rcMPGParams.getResMode() )
  {
    ROF( pcResInfo->getResidualMode() == ResInfo::RES_TRAFO_POW2);
    ModePictureGenerator::TrafoPow2Plotter cTrafoPow2Plotter( m_rcPicGen, m_rcMPGParams, *pcResInfo->getResInfoTrafoPow2(), rcPos.getPrimPixelPos() );
    cTrafoPow2Plotter.processTree();
  }
  else
  {
    UInt uiColor = 0;
    if ( pcPredInfo->isIntraPredicted() )
    {
      uiColor = 128;
      if ( pcPredInfo->getPredictionMode() == PredInfo::INTRA_DC )
        uiColor += 2;
      else
        uiColor += pcPredInfo->getPredSyntax()->getPredSyntaxIntraDirectional()->getIntraDirection();
    }
    else
    {
      uiColor = 128 + pcPredInfo->getPredictionMode();
    }

    m_rcPicGen.xPlotBlockArea( PixelPos( rcPos.getPrimPixelPos().getX(), rcPos.getPrimPixelPos().getY() ),
                               BlockSize( rcPos.getPrimBlockSize().getX(), rcPos.getPrimBlockSize().getY() ),
                               uiColor );

    m_rcPicGen.xPlotBlockBorder( PixelPos( rcPos.getPrimPixelPos().getX(), rcPos.getPrimPixelPos().getY() ),
                                 BlockSize( rcPos.getPrimBlockSize().getX(), rcPos.getPrimBlockSize().getY() ),
                                 m_rcMPGParams.getBlockBorders() );
  }

  return Err::m_nOK;
}

__inline void ModePictureGenerator::xPlotBlockArea( const PixelPos& cPixelPos, const BlockSize& cBlockSize, UInt uiColor )
{
  Block<Pel> cTargetBlock( cPixelPos, cBlockSize, *m_pcColorMap );
  cTargetBlock.assign( uiColor );
}

__inline void ModePictureGenerator::xPlotBlockBorder( const PixelPos& cPixelPos, const BlockSize& cBlockSize, UInt uiColor )
{ 
  Block<Pel> cBorderH( cPixelPos + PixelPos( 0, cBlockSize.getY()-1 ), BlockSize( cBlockSize.getX(), 1 ), *m_pcColorMap );
  Block<Pel> cBorderV( cPixelPos + PixelPos( cBlockSize.getX()-1, 0 ), BlockSize( 1, cBlockSize.getY() ), *m_pcColorMap );
  cBorderH.assign( uiColor );
  cBorderV.assign( uiColor );
}

__inline void ModePictureGenerator::xDrawLine( const PrimPixelPos& cEndA, const PrimPixelPos& cEndB, UInt uiColor )
{
  int iXA = cEndA.getX();
  int iXB = cEndB.getX();
  int iYA = cEndA.getY();
  int iYB = cEndB.getY();

  int iXMax = m_pcColorMap->getSize().getX() - 1;
  int iYMax = m_pcColorMap->getSize().getY() - 1;

  if ( iXA > iXMax || iXA < 0 ) return;
  if ( iYA > iYMax || iYA < 0 ) return;
  if ( iXB > iXMax || iXB < 0 ) return;
  if ( iYB > iYMax || iYB < 0 ) return;

  int iDiffX = iXA - iXB;
  int iDiffY = iYA - iYB;
  UInt uiDiffX = abs( iDiffX );
  UInt uiDiffY = abs( iDiffY );

  UInt uiStride;
  UInt uiDelta;
  int iFastPos;
  int iSlowPos;
  int iSlowDirection;

  bool bIterateX = ( uiDiffX > uiDiffY );
  if ( bIterateX )
  {
    uiStride = uiDiffX;
    uiDelta = uiDiffY;
    if ( iDiffX > 0 )
    {
      iFastPos = iXB;
      iSlowPos = iYB;
      iSlowDirection = iDiffY > 0 ? 1 : -1;
    }
    else
    {
      iFastPos = iXA;
      iSlowPos = iYA;
      iSlowDirection = iDiffY > 0 ? -1 : 1;
    }
  }
  else
  {
    uiStride = uiDiffY;
    uiDelta = uiDiffX;
    if ( iDiffY > 0 )
    {
      iFastPos = iYB;
      iSlowPos = iXB;
      iSlowDirection = iDiffX > 0 ? 1 : -1;
    }
    else
    {
      iFastPos = iYA;
      iSlowPos = iXA;
      iSlowDirection = iDiffX > 0 ? -1 : 1;
    }
  }

  UInt uiAcc = 0;

  UInt ui = 0;
  for (;;)
  {
    PixelPos cPixelPos;
    if ( bIterateX )
    {
      cPixelPos = PixelPos( iFastPos, iSlowPos );
    }
    else
    {
      cPixelPos = PixelPos( iSlowPos, iFastPos );
    }

    m_pcColorMap->getEl( cPixelPos ) = uiColor;

    if ( ui == uiStride ) break;
    ui++;

    iFastPos++;
    uiAcc += uiDelta;
    if ( uiAcc >= uiStride )
    {
      uiAcc -= uiStride;
      iSlowPos += iSlowDirection;
    }
  }
}

ErrVal ModePictureGenerator::TrafoPow2Plotter::xProcessLeafNode( const QTNavi &rcPos )
{
  PrimPixelPos cPrimPixelPos = m_cOrigin + rcPos.getPrimPixelPos();
  PixelPos cPixelPos( cPrimPixelPos.getX(), cPrimPixelPos.getY() );
  BlockSize cBlockSize( rcPos.getPrimBlockSize().getX(), rcPos.getPrimBlockSize().getY() );
  Block<Pel> cTargetBlock( cPixelPos, cBlockSize, *m_rcPicGen.m_pcColorMap );
  const PicBlock<const int> &cCoeffBlock ( m_rcResInfo.getTree().getT( rcPos )->getCoeff() );
  int iSizeX = cTargetBlock.getSize().getX();
  int iSizeY = cTargetBlock.getSize().getY();

  for ( int y = 0; y < iSizeY; y++ )
  {
    for ( int x = 0; x < iSizeX; x++ )
    {
      int iLevel = cCoeffBlock.getPlane( 0 ).getEl( PixelPos( x, y ) );
      if (iLevel < 0)
      {
        cTargetBlock.getEl( PixelPos( x, y ) ) = Pel( 140 + PredInfo::INTER_TRANSLATIONAL ); // cyan
      }
      else
      {
        cTargetBlock.getEl( PixelPos( x, y ) ) = Pel( 140 + PredInfo::INTRA_DIRECTIONAL ); // yellow
      }
    }
  }

  m_rcPicGen.xPlotBlockBorder( cPixelPos, cBlockSize, m_rcMPGParams.getBlockBorders() );
  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END
