/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"

#include "VideoCommonLib/QuadtreeGroup.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

QTNavi QTNavi::findLeafNode( const PrimPixelPos &rcPos ) const
{
  if( /*0 > rcPos.getX() || 0 > rcPos.getY() || */((UInt)rcPos.getX() >= m_pcTree->m_uiWidth ) || ((UInt)rcPos.getY() >= m_pcTree->m_uiHeight) )
  {
    return QTNavi( NULL );
  }

  const std::vector<QuadtreeGroupStructure::Layer>& racLayers = m_pcTree->m_acLayers;
  UInt uiDim = racLayers[m_iLayer].getDim();
  int iX = rcPos.getX() >> uiDim;
  int iY = rcPos.getY() >> uiDim;
  const QTNodeState eNodeState = racLayers[m_iLayer].getNodeStateNoCheck( iX, iY );
  if( IS_LEAF_NODE == eNodeState )
  {
    return QTNavi( m_pcTree, iX, iY, m_iLayer );
  }
  else if( eNodeState == IS_INTERNAL_NODE )
  {
    // go down
    int iLayer = m_iLayer;
    do
    {
      ++iLayer;
      uiDim = racLayers[iLayer].getDim();
      iX = rcPos.getX() >> uiDim;
      iY = rcPos.getY() >> uiDim;
    } while( IS_LEAF_NODE != racLayers[iLayer].getNodeStateNoCheck( iX, iY ) );
    return QTNavi( m_pcTree, iX, iY, iLayer );
  }
  else if( eNodeState == IS_BELOW_LEAF_NODE )
  {
    // go up
    int iLayer = m_iLayer;
    do
    {
      --iLayer;
      uiDim = racLayers[iLayer].getDim();
      iX = rcPos.getX() >> uiDim;
      iY = rcPos.getY() >> uiDim;
    } while( IS_LEAF_NODE != racLayers[iLayer].getNodeStateNoCheck( iX, iY ) );
    return QTNavi( m_pcTree, iX, iY, iLayer );
  }
  return QTNavi( NULL );
}


QuadtreeGroupStructure::QuadtreeGroupStructure( PrimBlockSize cSize, UInt uiPower2MaxBlock, UInt uiPower2MinBlock )
: m_uiWidth( cSize.getX() )
, m_uiHeight( cSize.getY() )
, m_uiEntropySliceSizeX( 0 )
, m_uiEntropySliceSizeY( 0 )
{
  QTNodeState eInitNodeState = IS_LEAF_NODE;
  for( int iDim = uiPower2MaxBlock; iDim >= (int)uiPower2MinBlock; iDim-- )
  {
    UInt uiNodesX = (m_uiWidth  >> iDim) + (m_uiWidth  % (1<<iDim) ? 1:0);
    UInt uiNodesY = (m_uiHeight >> iDim) + (m_uiHeight % (1<<iDim) ? 1:0);
    m_acLayers.push_back( Layer( uiNodesX, uiNodesY, iDim, eInitNodeState ) );
    eInitNodeState = IS_BELOW_LEAF_NODE;
  }
}

QuadtreeGroupStructure::QuadtreeGroupStructure( const PrimBlockSize& rcSize, const UInt& ruiLog2MinSize )
: m_uiWidth ( rcSize.getX() )
, m_uiHeight( rcSize.getY() )
, m_uiEntropySliceSizeX( 0 )
, m_uiEntropySliceSizeY( 0 )
{
  UInt        uiLog2MaxSize  = gCeilLog2( gMax( rcSize.getX(), rcSize.getY() )  );
  QTNodeState eInitNodeState = IS_LEAF_NODE;
  for( int iDim = (int)uiLog2MaxSize; iDim >= (int)ruiLog2MinSize; iDim-- )
  {
    UInt uiNodesX = ( m_uiWidth  >> iDim ) + ( m_uiWidth  % ( 1 << iDim ) ? 1 : 0 );
    UInt uiNodesY = ( m_uiHeight >> iDim ) + ( m_uiHeight % ( 1 << iDim ) ? 1 : 0 );
    m_acLayers.push_back( Layer( uiNodesX, uiNodesY, iDim, eInitNodeState ) );
    eInitNodeState = IS_BELOW_LEAF_NODE;
  }
}

QuadtreeGroupStructure::QuadtreeGroupStructure( const QuadtreeGroupStructure &rcSource, UInt uiPower2MinBlock )
: m_uiWidth( rcSource.m_uiWidth )
, m_uiHeight( rcSource.m_uiHeight )
, m_acLayers( rcSource.m_acLayers )
, m_uiEntropySliceSizeX( rcSource.m_uiEntropySliceSizeX )
, m_uiEntropySliceSizeY( rcSource.m_uiEntropySliceSizeY )
{
  QTNodeState eInitNodeState = IS_BELOW_LEAF_NODE;
  for( int iDim = int( m_acLayers.back().getDim() ) - 1; iDim >= (int)uiPower2MinBlock; iDim-- )
  {
    UInt uiNodesX = (m_uiWidth  >> iDim) + (m_uiWidth  % (1<<iDim) ? 1:0);
    UInt uiNodesY = (m_uiHeight >> iDim) + (m_uiHeight % (1<<iDim) ? 1:0);
    m_acLayers.push_back( Layer( uiNodesX, uiNodesY, iDim, eInitNodeState ) );
  }
}

/*
QTNavi QuadtreeGroupStructure::getFirstQuadtreeRoot() const
{
  return getNextQuadtreeRoot( QTNavi( this, -1, 0, 0 ) );
}
*/
QTNavi QuadtreeGroupStructure::getFirstQuadtreeRoot( UInt uiSlice ) const
{
  UInt uiSizeX = m_uiEntropySliceSizeX == 0 ? m_acLayers[0].getNumNodesX() : m_uiEntropySliceSizeX;
  UInt uiSliceDimX = UInt( ceil( m_acLayers[0].getNumNodesX() / (double) uiSizeX ) );
  UInt uiSliceX = uiSlice % uiSliceDimX;
  UInt uiSliceY = uiSlice / uiSliceDimX;
  return QTNavi( this, uiSliceX * m_uiEntropySliceSizeX, uiSliceY * m_uiEntropySliceSizeY, 0 );
}

QTNavi QuadtreeGroupStructure::getNextQuadtreeRoot( const QTNavi &rcCurrent ) const
{
  AOF_DBG( rcCurrent.getTreeStructurePtr() == this );  // ensure that the navigator points to the root layer
  AOF_DBG( rcCurrent.getLayer() == 0 );

  UInt uiXStart = rcCurrent.getXCoord() + 1;
  for( UInt uiY = rcCurrent.getYCoord(); uiY < m_acLayers[0].getNumNodesY(); uiY++ )
  {
    for( UInt uiX = uiXStart; uiX < m_acLayers[0].getNumNodesX(); uiX++ )
    {
      if( 0 == (m_acLayers[0].getNodeState( uiX, uiY ) & IS_FORBIDDEN) )
      {
        return QTNavi( this, uiX, uiY, 0 );
      }
    }
    uiXStart = 0;
  }
  return QTNavi( NULL );
}

QTNavi QuadtreeGroupStructure::getNextQuadtreeRoot( const QTNavi &rcCurrent, UInt uiSlice ) const
{
  AOF_DBG( rcCurrent.getTreeStructurePtr() == this );  // ensure that the navigator points to the root layer
  AOF_DBG( rcCurrent.getLayer() == 0 );

  UInt uiXStart = rcCurrent.getXCoord() + 1;
  UInt uiSizeX = m_uiEntropySliceSizeX == 0 ? m_acLayers[0].getNumNodesX() : m_uiEntropySliceSizeX;
  UInt uiSizeY = m_uiEntropySliceSizeY == 0 ? m_acLayers[0].getNumNodesY() : m_uiEntropySliceSizeY;
  UInt uiSliceDimX = UInt( ceil( m_acLayers[0].getNumNodesX() / (double) uiSizeX ) );
  UInt uiSliceDimY = UInt( ceil( m_acLayers[0].getNumNodesY() / (double) uiSizeY ) );
  UInt uiSliceX = uiSlice % uiSliceDimX;
  UInt uiSliceY = ( uiSlice - uiSliceX ) / uiSliceDimY;
  for( UInt uiY = rcCurrent.getYCoord(); uiY < m_uiEntropySliceSizeY * ( uiSliceY + 1); uiY++ )
  {
    for( UInt uiX = uiXStart; uiX < m_uiEntropySliceSizeX * ( uiSliceX + 1 ); uiX++ )
    {
      if( 0 == (m_acLayers[0].getNodeState( uiX, uiY ) & IS_FORBIDDEN) )
      {
        QTNavi cNav( this, uiX, uiY, 0 );
        if ( getEntropySlice(cNav) == uiSlice )
          return cNav;
      }
    }
    uiXStart = m_uiEntropySliceSizeX * uiSliceX;
  }
  return QTNavi( NULL );
}

UInt QuadtreeGroupStructure::getNumEntropySlices() const
{
  UInt uiSizeX = m_uiEntropySliceSizeX == 0 ? m_acLayers[0].getNumNodesX() : m_uiEntropySliceSizeX;
  UInt uiSizeY = m_uiEntropySliceSizeY == 0 ? m_acLayers[0].getNumNodesY() : m_uiEntropySliceSizeY;
  UInt uiSliceDimX = UInt( ceil( m_acLayers[0].getNumNodesX() / (double) uiSizeX ) );
  UInt uiSliceDimY = UInt( ceil( m_acLayers[0].getNumNodesY() / (double) uiSizeY ) );
  return uiSliceDimX * uiSliceDimY;
}

bool QuadtreeGroupStructure::isSameEntropySlice( const QTNavi &rcQTNavi1, const QTNavi &rcQTNavi2 ) const
{
  if( 0 != m_uiEntropySliceSizeX )
  {
    if( rcQTNavi1.getXCoord() / m_uiEntropySliceSizeX != rcQTNavi2.getXCoord() / m_uiEntropySliceSizeX)
    {
      return false;
    }
  }

  if( 0 != m_uiEntropySliceSizeY )
  {
    if( rcQTNavi1.getYCoord() / m_uiEntropySliceSizeY != rcQTNavi2.getYCoord() / m_uiEntropySliceSizeY)
    {
      return false;
    }
  }

  return true;
}

UInt QuadtreeGroupStructure::getEntropySlice( const QTNavi &rcCurrent ) const
{
  UInt uiPosX = rcCurrent.getXCoord();
  UInt uiPosY = rcCurrent.getYCoord();
  UInt uiSizeX = m_uiEntropySliceSizeX == 0 ? m_acLayers[0].getNumNodesX() : m_uiEntropySliceSizeX;
  UInt uiSizeY = m_uiEntropySliceSizeY == 0 ? m_acLayers[0].getNumNodesY() : m_uiEntropySliceSizeY;
  UInt uiSliceX = uiPosX / uiSizeX;
  UInt uiSliceY = uiPosY / uiSizeY;
  UInt uiSliceDimX = UInt( ceil( m_acLayers[0].getNumNodesX() / (double) uiSizeX ) );
  return ( uiSliceY * uiSliceDimX ) + uiSliceX;
}

ErrVal QuadtreeGroupStructure::appendChildNodes( const QTNavi &rcNav, bool &rbChildsWereAppended )
{
  ROF( rcNav.getTreeStructurePtr() == this );
  ROF( rcNav.isLeafNode() );
  if( rcNav.getLayer() + 1 >= (int)rcNav.getTreeStructurePtr()->m_acLayers.size() )
  {
    rbChildsWereAppended = false;
  }
  else
  {
    m_acLayers[rcNav.getLayer()].setNodeState( IS_INTERNAL_NODE, rcNav.getXCoord(), rcNav.getYCoord() );
    for( UInt ui = 0; ui < 4; ui++ )
    {
      const QTNavi &rcChild = rcNav.getChild( ui );
      if( !rcChild.liesOutside() )
      {
        m_acLayers[rcChild.getLayer()].setNodeState( IS_LEAF_NODE, rcChild.getXCoord(), rcChild.getYCoord() );
      }
    }
    rbChildsWereAppended = true;
  }
  return Err::m_nOK;
}


ErrVal QuadtreeGroupStructure::removeChildNodes( const QTNavi &rcNav )
{
  ROF( rcNav.getTreeStructurePtr() == this );
  ROF( rcNav.isInternalNode() );
  m_acLayers[rcNav.getLayer()].setNodeState( IS_LEAF_NODE, rcNav.getXCoord(), rcNav.getYCoord() );
  for( UInt ui = 0; ui < 4; ui++ )
  {
    const QTNavi &rcChild = rcNav.getChild( ui );
    if( rcChild.isInternalNode() )
    {
      RNOK( removeChildNodes( rcChild ) );
    }
    if( rcChild.isLeafNode() )
    {
      m_acLayers[rcChild.getLayer()].setNodeState( IS_BELOW_LEAF_NODE, rcChild.getXCoord(), rcChild.getYCoord() );
    }
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
