/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/ResTrafoPow2.h"
#include "VideoCommonLib/SyntaxElementCoder.h"
#include "VideoCommonLib/ContextTables.h"
#include <math.h>

// video namespace begin
VIDEO_NAMESPACE_BEGIN


//==== NOTE: don't change any of the static constants (unless you know what you are doing) =====
const UInt  ResTrafoPow2::sm_uiLog2MaxTSize     = 7;
const UInt  ResTrafoPow2::sm_uiCoeffAccessMask  = 0x01FF; // 4*2^m_suiLog2MaxTSize-1
const UInt  ResTrafoPow2::sm_uiDCTPrecision     = 8;      // 2..min(8,20-m_suiLog2MaxTSize)
const UInt  ResTrafoPow2::sm_uiIDCTPrecision    = 12;     // 3..19-m_suiLog2MaxTSize
const UInt  ResTrafoPow2::sm_uiQBits            = 15;     // 0..22-m_suiLog2MaxTSize
const int   ResTrafoPow2::sm_aiQStep      [6]   =         // as in H.264/AVC 
{
  16,   18,   20,   23,   25,   29 
};
const int   ResTrafoPow2::sm_aiQFactor    [6]   =         // ( 1 << m_suiQBits ) / m_saiQStep[i],
{
  2048, 1820, 1638, 1424, 1310, 1129
};
const int   ResTrafoPow2::sm_aiDCTCoeffs  [128] =         // round( 2^(sm_uiDCTPrecision+4)*sqrt(1/5)*cos(2*pi*k/(2^(m_suiLog2MaxTSize+2)))*(k==0?sqrt(1/2):1) )
{
 1295, 1832, 1831, 1831, 1830, 1828, 1827, 1825, 1823, 1821, 1818, 1815, 1812, 1809, 1805, 1801,
 1797, 1792, 1787, 1782, 1777, 1771, 1765, 1759, 1753, 1746, 1739, 1732, 1725, 1717, 1709, 1701,
 1692, 1684, 1675, 1665, 1656, 1646, 1636, 1626, 1615, 1605, 1594, 1583, 1571, 1560, 1548, 1535,
 1523, 1510, 1498, 1485, 1471, 1458, 1444, 1430, 1416, 1402, 1387, 1372, 1357, 1342, 1327, 1311,
 1295, 1279, 1263, 1247, 1230, 1213, 1196, 1179, 1162, 1145, 1127, 1109, 1091, 1073, 1055, 1036,
 1018,  999,  980,  961,  942,  922,  903,  883,  863,  844,  824,  803,  783,  763,  742,  722,
  701,  680,  659,  638,  617,  596,  575,  553,  532,  510,  489,  467,  445,  423,  401,  379,
  357,  335,  313,  291,  269,  247,  224,  202,  180,  157,  135,  112,   90,   67,   45,   22
};
const int   ResTrafoPow2::sm_aiIDCTCoeffs [128] =         // round( 2^(sm_uiIDCTPrecision-3)*sqrt(5)*cos(2*pi*k/(2^(m_suiLog2MaxTSize+2)))*(k==0?sqrt(1/2):1) )
{
  810, 1145, 1145, 1144, 1143, 1143, 1142, 1141, 1139, 1138, 1136, 1134, 1132, 1130, 1128, 1126,
 1123, 1120, 1117, 1114, 1111, 1107, 1103, 1100, 1096, 1091, 1087, 1083, 1078, 1073, 1068, 1063,
 1058, 1052, 1047, 1041, 1035, 1029, 1023, 1016, 1010, 1003,  996,  989,  982,  975,  967,  960,
  952,  944,  936,  928,  920,  911,  903,  894,  885,  876,  867,  858,  848,  839,  829,  819,
  810,  800,  789,  779,  769,  758,  748,  737,  726,  715,  704,  693,  682,  671,  659,  648,
  636,  624,  613,  601,  589,  576,  564,  552,  540,  527,  515,  502,  489,  477,  464,  451,
  438,  425,  412,  399,  386,  372,  359,  346,  332,  319,  305,  292,  278,  265,  251,  237,
  223,  210,  196,  182,  168,  154,  140,  126,  112,   98,   84,   70,   56,   42,   28,   14
};


ErrVal ResTrafoPow2::reconstructResidual( const ResInfo &rcResInfo,
                                          const PicBlock<DPel> &rcResBlock,
                                          const bool bAddDecRes )
{
  const  ResInfoTrafoPow2 *pcResInfo = rcResInfo.getResInfoTrafoPow2();
  ROF( pcResInfo );

  for( QTNavi cNav = pcResInfo->getTree().getTreeStructure().getFirstQuadtreeRoot(); !cNav.liesOutside(); cNav = pcResInfo->getTree().getTreeStructure().getNextQuadtreeRoot( cNav ) )
  {
    ResTrafoPow2Configuration cConfig( pcResInfo, rcResBlock, bAddDecRes );
    RNOK( xReconstructNode( cNav, cConfig ) );
  }

  return Err::m_nOK;
}


ErrVal ResTrafoPow2::appendTrafoPow2ContextsAndSetIndices( ResidualContextStartIndices &rcResCtxIndices,
                                                           UInt uiLog2MinTrafoSize,
                                                           UInt uiLog2MaxTrafoSize,
                                                           CtxInitArray &rcCtxInit )
{
  rcResCtxIndices.m_uiStartTrafoPow2Sig = rcCtxInit.getCtxAccess().getNumCtxInitArrays();
  RNOK( xAppendSigCtx ( rcCtxInit, uiLog2MinTrafoSize, uiLog2MaxTrafoSize ) );
  rcResCtxIndices.m_uiStartTrafoPow2Last = rcCtxInit.getCtxAccess().getNumCtxInitArrays();
  RNOK( xAppendLastCtx( rcCtxInit, uiLog2MinTrafoSize, uiLog2MaxTrafoSize ) );
  return Err::m_nOK;  
}


ErrVal ResTrafoPow2::xAppendSigCtx( CtxInitArray &rcCtxInit, UInt uiLog2MinTrafoSize, UInt uiLog2MaxTrafoSize )
{
  for( UInt k = 0; k <= uiLog2MaxTrafoSize; k++ )
  {
    if( k >= uiLog2MinTrafoSize )
    {
      if( k <= 3 )
      {
        rcCtxInit = rcCtxInit << ContextTables::m_cCtxAccessSig4x4;
      }
      else
      {
        rcCtxInit = rcCtxInit << ContextTables::m_cCtxAccessSig;
      }
    }
    else
    {
      rcCtxInit   = rcCtxInit << CtxInitArray( "CTX_START_POW2_SIG_DUMMY", std::vector<CtxIni>() );
    }
  }
  return Err::m_nOK;
}

ErrVal ResTrafoPow2::xAppendLastCtx( CtxInitArray &rcCtxInit, UInt uiLog2MinTrafoSize, UInt uiLog2MaxTrafoSize )
{
  for( UInt k = 0; k <= uiLog2MaxTrafoSize; k++ )
  {
    if( k >= uiLog2MinTrafoSize )
    {
      if( k <= 3 )
      {
        rcCtxInit = rcCtxInit << ContextTables::m_cCtxAccessLast4x4;
      }
      else
      {
        rcCtxInit = rcCtxInit << ContextTables::m_cCtxAccessLast;
      }
    }
    else
    {
      rcCtxInit   = rcCtxInit << CtxInitArray( "CTX_START_POW2_LAST_DUMMY", std::vector<CtxIni>() );
    }
  }
  return Err::m_nOK;
}

ErrVal ResTrafoPow2::xReconstructNode( const QTNavi &rcNav, const ResTrafoPow2Configuration &rcConfig )
{
  const QTNodeState eNodeState = rcNav.getNodeState();
  if( IS_INTERNAL_NODE == eNodeState )
  {
    for (UInt uiChildIndex = 0; uiChildIndex < 4; uiChildIndex++ )
    {
      RNOK( xReconstructNode( rcNav.getChild( uiChildIndex ), rcConfig ) );
    }
  }
  else if( IS_LEAF_NODE == eNodeState )
  {
    const ResInfoTrafoPow2Leaf* pcLeaf =  rcConfig.m_rcCoeffTree.getT( rcNav );
    const PicBlock<DPel> cDestinationPicBlock( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), rcConfig.m_rcResBlock );
//    AOF( cDestinationPicBlock.getStats() == pcLeaf->getCoeff().getStats() );
    UInt uiNumPlanes = pcLeaf->getCoeff().getStats().getNumPlanes();
    for( UInt uiPlane = 0; uiPlane < uiNumPlanes; uiPlane++ )
    {
      if ( !( pcLeaf->getCbf( uiPlane ) ) )
      {
        if ( !rcConfig.m_bAddDecRes )
        {
          cDestinationPicBlock.getPlane( uiPlane ).assign( 0 );
        }
      }
      else
      {
        const BlockSize rcBlockSize = pcLeaf->getCoeff().getPlane( uiPlane ).getSize();
        Block<int> cCoeffs( m_aucTempIntBuffer, rcBlockSize, rcBlockSize.getX() );

//          AllocBlock<int> cCoeffs( pcLeaf->getCoeff().getPlane( uiPlane ).getSize() );
        RNOK( xFastIScale( rcConfig.m_raiPlaneQps[uiPlane], pcLeaf->getCoeff().getPlane( uiPlane ), cCoeffs ) );
        RNOK( xFast2dIDCT( cCoeffs, cDestinationPicBlock.getPlane( uiPlane ), rcConfig.m_bAddDecRes ) );
      }
    }
  }
  return Err::m_nOK;
}


ErrVal ResTrafoPow2::xInitCoeffArray()
{
  UInt  uiMax = 1 << sm_uiLog2MaxTSize;
  UInt  ui1   = 0;
  UInt  ui2   = uiMax << 1;
  UInt  ui3   = uiMax << 1;
  UInt  ui4   = uiMax << 2;
  {
    m_aiDCTCoeffArray  [ui1] =  sm_aiDCTCoeffs  [ui1];
    m_aiDCTCoeffArray  [ui3] = -sm_aiDCTCoeffs  [ui1];
    m_aiIDCTCoeffArray [ui1] =  sm_aiIDCTCoeffs [ui1];
    m_aiIDCTCoeffArray [ui3] = -sm_aiIDCTCoeffs [ui1];
  }
  for( ui1++, ui2--, ui3++, ui4--; ui1 < uiMax; ui1++, ui2--, ui3++, ui4-- )
  {
    m_aiDCTCoeffArray  [ui1] = m_aiDCTCoeffArray  [ui4] =  sm_aiDCTCoeffs [ui1];
    m_aiDCTCoeffArray  [ui2] = m_aiDCTCoeffArray  [ui3] = -sm_aiDCTCoeffs [ui1];
    m_aiIDCTCoeffArray [ui1] = m_aiIDCTCoeffArray [ui4] =  sm_aiIDCTCoeffs[ui1];
    m_aiIDCTCoeffArray [ui2] = m_aiIDCTCoeffArray [ui3] = -sm_aiIDCTCoeffs[ui1];
  }
  {
    m_aiDCTCoeffArray  [ui1] = m_aiDCTCoeffArray  [ui3] =  0;
    m_aiIDCTCoeffArray [ui1] = m_aiIDCTCoeffArray [ui3] =  0;
  }
  return Err::m_nOK;
}


template<UInt uiBlockSize,  UInt uiLog2TSize, bool bHor >
__inline void ResTrafoPow2::xFast1dIDCT( int* paiData, const UInt uiInc1, const UInt uiFinalShift ) 
{
  if( false == bHor )
  {
  //==== copy and init with DC coefficient =====
    for( UInt uiIdx = 0, uiDataIdx = 0; uiIdx < uiBlockSize; uiIdx++, uiDataIdx += uiInc1 )
  {
    m_aiAuxArray[ uiIdx ] = paiData[ uiDataIdx ];
  }
  }

  {
    //==== init with DC coefficient =====
    const int iDCContribution = paiData[ 0 ] * m_aiIDCTCoeffArray[ 0 ];
    for( UInt uiIdx = 0; uiIdx < uiBlockSize; uiIdx++ )
    {
      m_aiAux2Array[ uiIdx ] = iDCContribution;
    }
  }

  const int* piSrc = bHor ? paiData : m_aiAuxArray;

  if( 0 < uiLog2TSize )
  {
    const UInt uiIndex = 1 << (uiLog2TSize -1);
    {
      int iACContribution = 810 * piSrc[ uiIndex ];
      for( UInt uiFwdDataIdx = 0; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 2 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 1 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
  }
    

  if( 1 < uiLog2TSize )
  {
    const UInt uiIndex = 1 << gMax(0,((int)uiLog2TSize -2));
    {
      int iACContribution = 1058 * piSrc[ uiIndex ] + 438 * piSrc[ 3*uiIndex ];
      for( UInt uiFwdDataIdx = 0; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 4 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 3 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }

    {
      int iACContribution = 438 * piSrc[ uiIndex ] - 1058 * piSrc[ 3*uiIndex ];
      for( UInt uiFwdDataIdx = 1; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 4 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 1 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
  }


  if( 2 < uiLog2TSize )
  {
    const UInt uiIndex = 1 << gMax(0,((int)uiLog2TSize -3));
    {
      int iACContribution = 1123 * piSrc[ uiIndex ] + 952 * piSrc[ 3 * uiIndex ] + 636 * piSrc[ 5 * uiIndex ] + 223 * piSrc[ 7 * uiIndex ];
      for( UInt uiFwdDataIdx = 0; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 8 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 7 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }

    {
      int iACContribution = 952 * piSrc[ 1 * uiIndex ] - 223 * piSrc[ 3 * uiIndex ] - 1123 * piSrc[ 5 * uiIndex ] - 636 * piSrc[ 7 * uiIndex ];
      for( UInt uiFwdDataIdx = 1; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 8 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 5 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }

    {
      int iACContribution = 636 * piSrc[ 1 * uiIndex ] - 1123 * piSrc[ 3 * uiIndex ] + 223 * piSrc[ 5 * uiIndex ] + 952 * piSrc[ 7 * uiIndex ];
      for( UInt uiFwdDataIdx = 2; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 8 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 3 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }

    {
      int iACContribution = 223 * piSrc[ 1 * uiIndex ] - 636 * piSrc[ 3 * uiIndex ] + 952 * piSrc[ 5 * uiIndex ] - 1123 * piSrc[ 7 * uiIndex ];
      for( UInt uiFwdDataIdx = 3; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 8 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 1 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
  }

  if( 3 < uiLog2TSize )
  {

    const UInt uiIndex = 1 << gMax(0,((int)uiLog2TSize -4));
    {
      int iACContribution = 1139 * piSrc[ uiIndex ] + 1096 * piSrc[ 3 * uiIndex ] + 1010 * piSrc[ 5 * uiIndex ] + 885 * piSrc[ 7 * uiIndex ] 
                          + 726 * piSrc[ 9 * uiIndex ] + 540 * piSrc[ 11 * uiIndex ] + 332 * piSrc[ 13 * uiIndex ] + 112 * piSrc[ 15 * uiIndex ];
      for( UInt uiFwdDataIdx = 0; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 16 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 15 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }

    {
      int iACContribution = 1096 * piSrc[ uiIndex ] + 726 * piSrc[ 3 * uiIndex ] + 112 * piSrc[ 5 * uiIndex ] - 540 * piSrc[ 7 * uiIndex ] 
                           -1010 * piSrc[ 9 * uiIndex ] -1139 * piSrc[ 11 * uiIndex ] -885 * piSrc[ 13 * uiIndex ] -332 * piSrc[ 15 * uiIndex ];
      for( UInt uiFwdDataIdx = 1; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 16 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 13 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }

    {
      int iACContribution = 1010 * piSrc[ uiIndex ] + 112 * piSrc[ 3 * uiIndex ] - 885 * piSrc[ 5 * uiIndex ] -1096 * piSrc[ 7 * uiIndex ] 
                           -332 * piSrc[ 9 * uiIndex ] + 726 * piSrc[ 11 * uiIndex ] +1139 * piSrc[ 13 * uiIndex ] + 540 * piSrc[ 15 * uiIndex ];
      for( UInt uiFwdDataIdx = 2; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 16 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 11 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }

    {
      int iACContribution = 885 * piSrc[ uiIndex ] - 540 * piSrc[ 3 * uiIndex ] - 1096 * piSrc[ 5 * uiIndex ] + 112 * piSrc[ 7 * uiIndex ] 
                          + 1139 * piSrc[ 9 * uiIndex ] + 332 * piSrc[ 11 * uiIndex ] -1010 * piSrc[ 13 * uiIndex ] - 726 * piSrc[ 15 * uiIndex ];
      for( UInt uiFwdDataIdx = 3; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 16 )
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 9 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
     {
       int iACContribution = 726 * piSrc[ uiIndex ] - 1010 * piSrc[ 3 * uiIndex ] -332 * piSrc[ 5 * uiIndex ] + 1139 * piSrc[ 7 * uiIndex ] 
                           -112 * piSrc[ 9 * uiIndex ] -1096 * piSrc[ 11 * uiIndex ] + 540 * piSrc[ 13 * uiIndex ] + 885 * piSrc[ 15 * uiIndex ];
       for( UInt uiFwdDataIdx = 4; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 16 )
       {
         m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
         m_aiAux2Array[ uiFwdDataIdx + 7 ] -= iACContribution;
         iACContribution = -iACContribution;
       }
     }

     {
       int iACContribution = 540 * piSrc[ uiIndex ] -1139 * piSrc[ 3 * uiIndex ] + 726 * piSrc[ 5 * uiIndex ] + 332 * piSrc[ 7 * uiIndex ] 
                           -1096 * piSrc[ 9 * uiIndex ] +885 * piSrc[ 11 * uiIndex ] +112 * piSrc[ 13 * uiIndex ] -1010 * piSrc[ 15 * uiIndex ];
       for( UInt uiFwdDataIdx = 5; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 16 )
       {
         m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
         m_aiAux2Array[ uiFwdDataIdx + 5 ] -= iACContribution;
         iACContribution = -iACContribution;
       }
     }

     {
       int iACContribution = 332 * piSrc[ uiIndex ] -885 * piSrc[ 3 * uiIndex ] +1139 * piSrc[ 5 * uiIndex ] -1010 * piSrc[ 7 * uiIndex ] 
                            +540 * piSrc[ 9 * uiIndex ] + 112 * piSrc[ 11 * uiIndex ] -726 * piSrc[ 13 * uiIndex ] + 1096 * piSrc[ 15 * uiIndex ];
       for( UInt uiFwdDataIdx = 6; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 16 )
       {
         m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
         m_aiAux2Array[ uiFwdDataIdx + 3 ] -= iACContribution;
         iACContribution = -iACContribution;
       }
     }

     {
       int iACContribution = 112 * piSrc[ uiIndex ] - 332 * piSrc[ 3 * uiIndex ] + 540 * piSrc[ 5 * uiIndex ] -726 * piSrc[ 7 * uiIndex ] 
                          + 885 * piSrc[ 9 * uiIndex ] -1010 * piSrc[ 11 * uiIndex ] + 1096 * piSrc[ 13 * uiIndex ] - 1139 * piSrc[ 15 * uiIndex ];
       for( UInt uiFwdDataIdx = 7; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 16 )
       {
         m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
         m_aiAux2Array[ uiFwdDataIdx + 1 ] -= iACContribution;
         iACContribution = -iACContribution;
       }
     }
  }


/*  
  if( 4 < uiLog2TSize )
  { 
    const UInt  uiCoeffAccessShift  = sm_uiLog2MaxTSize - uiLog2TSize;
    //==== AC coefficients =====
//    for( UInt uiLog2X = uiLog2TSize-4; uiLog2X > 0; uiLog2X-- )
    {
      UInt uiLog2X = uiLog2TSize-4;
      const UInt  uiDiffLog2        = uiLog2TSize - uiLog2X;
      const UInt  uiMaxMIdx         = 1 << uiDiffLog2;
      const UInt  uiDataStepSize    = 2 << uiDiffLog2;

      const UInt  uiStep            = 1 << uiLog2X;
      const UInt  uiStart           = uiStep >> 1;
      UInt        uiCoeffStart      = uiStart << uiCoeffAccessShift;
      const UInt  uiCoeffMAdd       = uiCoeffStart << 1;

      UInt  uiFwdDataIdxStart = 0;
      UInt  uiBwdOffset = uiDataStepSize - 1;
      for( UInt uiMIdx = 0; uiMIdx < uiMaxMIdx; uiMIdx++)
      {
        std::cout << "{" << std::endl;
        std::cout << "int iACContribution = ";
        int   iACContribution = 0;
        {
          const UInt uiCoeffAdd      = uiCoeffStart << 1;
          for( UInt uiIdx = uiStart, uiCoeffIdx = uiCoeffStart; uiIdx < uiBlockSize; uiIdx += uiStep )
          {
            std::cout << " +" << m_aiIDCTCoeffArray[ uiCoeffIdx & sm_uiCoeffAccessMask ] << " * piSrc[ " << uiIdx/2 << " * uiIndex ]" << std::endl; 
            iACContribution += piSrc[ uiIdx ] * m_aiIDCTCoeffArray[ uiCoeffIdx & sm_uiCoeffAccessMask ];
            uiCoeffIdx += uiCoeffAdd;
          }
        }

                std::cout << ";" << std::endl << std::endl;
                std::cout << "   for( UInt uiFwdDataIdx = " << uiFwdDataIdxStart << "; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += " << uiDataStepSize << ") " << std::endl;

                std::cout << "   { " << std::endl;
                std::cout << "     m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;" << std::endl;
                std::cout << "     m_aiAux2Array[ uiFwdDataIdx + " << uiBwdOffset << " ] -= iACContribution;" << std::endl;
                std::cout << "     iACContribution = -iACContribution;" << std::endl;
                std::cout << "   }" << std::endl;
                std::cout << "}" << std::endl;


        {
          for( UInt uiFwdDataIdx = uiFwdDataIdxStart; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += uiDataStepSize )
          {
            m_aiAux2Array[ uiFwdDataIdx ]               += iACContribution;
            m_aiAux2Array[ uiFwdDataIdx + uiBwdOffset ] -= iACContribution;
            iACContribution = -iACContribution;
          }
        }
        uiCoeffStart += uiCoeffMAdd, 
          uiFwdDataIdxStart += 1; 
        uiBwdOffset -= 2;
      }
    }
  }
  */

  if( 4 < uiLog2TSize )
  {

    const UInt uiIndex = 1 << gMax(0,((int)uiLog2TSize -5));

    {
      int iACContribution =  +1143 * piSrc[ 1 * uiIndex ]
      +1132 * piSrc[ 3 * uiIndex ]
      +1111 * piSrc[ 5 * uiIndex ]
      +1078 * piSrc[ 7 * uiIndex ]
      +1035 * piSrc[ 9 * uiIndex ]
      +982 * piSrc[ 11 * uiIndex ]
      +920 * piSrc[ 13 * uiIndex ]
      +848 * piSrc[ 15 * uiIndex ]
      +769 * piSrc[ 17 * uiIndex ]
      +682 * piSrc[ 19 * uiIndex ]
      +589 * piSrc[ 21 * uiIndex ]
      +489 * piSrc[ 23 * uiIndex ]
      +386 * piSrc[ 25 * uiIndex ]
      +278 * piSrc[ 27 * uiIndex ]
      +168 * piSrc[ 29 * uiIndex ]
      +56 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 0; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 31 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +1132 * piSrc[ 1 * uiIndex ]
      +1035 * piSrc[ 3 * uiIndex ]
      +848 * piSrc[ 5 * uiIndex ]
      +589 * piSrc[ 7 * uiIndex ]
      +278 * piSrc[ 9 * uiIndex ]
      -56 * piSrc[ 11 * uiIndex ]
      -386 * piSrc[ 13 * uiIndex ]
      -682 * piSrc[ 15 * uiIndex ]
      -920 * piSrc[ 17 * uiIndex ]
      -1078 * piSrc[ 19 * uiIndex ]
      -1143 * piSrc[ 21 * uiIndex ]
      -1111 * piSrc[ 23 * uiIndex ]
      -982 * piSrc[ 25 * uiIndex ]
      -769 * piSrc[ 27 * uiIndex ]
      -489 * piSrc[ 29 * uiIndex ]
      -168 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 1; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 29 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +1111 * piSrc[ 1 * uiIndex ]
      +848 * piSrc[ 3 * uiIndex ]
      +386 * piSrc[ 5 * uiIndex ]
      -168 * piSrc[ 7 * uiIndex ]
      -682 * piSrc[ 9 * uiIndex ]
      -1035 * piSrc[ 11 * uiIndex ]
      -1143 * piSrc[ 13 * uiIndex ]
      -982 * piSrc[ 15 * uiIndex ]
      -589 * piSrc[ 17 * uiIndex ]
      -56 * piSrc[ 19 * uiIndex ]
      +489 * piSrc[ 21 * uiIndex ]
      +920 * piSrc[ 23 * uiIndex ]
      +1132 * piSrc[ 25 * uiIndex ]
      +1078 * piSrc[ 27 * uiIndex ]
      +769 * piSrc[ 29 * uiIndex ]
      +278 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 2; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 27 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +1078 * piSrc[ 1 * uiIndex ]
      +589 * piSrc[ 3 * uiIndex ]
      -168 * piSrc[ 5 * uiIndex ]
      -848 * piSrc[ 7 * uiIndex ]
      -1143 * piSrc[ 9 * uiIndex ]
      -920 * piSrc[ 11 * uiIndex ]
      -278 * piSrc[ 13 * uiIndex ]
      +489 * piSrc[ 15 * uiIndex ]
      +1035 * piSrc[ 17 * uiIndex ]
      +1111 * piSrc[ 19 * uiIndex ]
      +682 * piSrc[ 21 * uiIndex ]
      -56 * piSrc[ 23 * uiIndex ]
      -769 * piSrc[ 25 * uiIndex ]
      -1132 * piSrc[ 27 * uiIndex ]
      -982 * piSrc[ 29 * uiIndex ]
      -386 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 3; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 25 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +1035 * piSrc[ 1 * uiIndex ]
      +278 * piSrc[ 3 * uiIndex ]
      -682 * piSrc[ 5 * uiIndex ]
      -1143 * piSrc[ 7 * uiIndex ]
      -769 * piSrc[ 9 * uiIndex ]
      +168 * piSrc[ 11 * uiIndex ]
      +982 * piSrc[ 13 * uiIndex ]
      +1078 * piSrc[ 15 * uiIndex ]
      +386 * piSrc[ 17 * uiIndex ]
      -589 * piSrc[ 19 * uiIndex ]
      -1132 * piSrc[ 21 * uiIndex ]
      -848 * piSrc[ 23 * uiIndex ]
      +56 * piSrc[ 25 * uiIndex ]
      +920 * piSrc[ 27 * uiIndex ]
      +1111 * piSrc[ 29 * uiIndex ]
      +489 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 4; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 23 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +982 * piSrc[ 1 * uiIndex ]
      -56 * piSrc[ 3 * uiIndex ]
      -1035 * piSrc[ 5 * uiIndex ]
      -920 * piSrc[ 7 * uiIndex ]
      +168 * piSrc[ 9 * uiIndex ]
      +1078 * piSrc[ 11 * uiIndex ]
      +848 * piSrc[ 13 * uiIndex ]
      -278 * piSrc[ 15 * uiIndex ]
      -1111 * piSrc[ 17 * uiIndex ]
      -769 * piSrc[ 19 * uiIndex ]
      +386 * piSrc[ 21 * uiIndex ]
      +1132 * piSrc[ 23 * uiIndex ]
      +682 * piSrc[ 25 * uiIndex ]
      -489 * piSrc[ 27 * uiIndex ]
      -1143 * piSrc[ 29 * uiIndex ]
      -589 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 5; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 21 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +920 * piSrc[ 1 * uiIndex ]
      -386 * piSrc[ 3 * uiIndex ]
      -1143 * piSrc[ 5 * uiIndex ]
      -278 * piSrc[ 7 * uiIndex ]
      +982 * piSrc[ 9 * uiIndex ]
      +848 * piSrc[ 11 * uiIndex ]
      -489 * piSrc[ 13 * uiIndex ]
      -1132 * piSrc[ 15 * uiIndex ]
      -168 * piSrc[ 17 * uiIndex ]
      +1035 * piSrc[ 19 * uiIndex ]
      +769 * piSrc[ 21 * uiIndex ]
      -589 * piSrc[ 23 * uiIndex ]
      -1111 * piSrc[ 25 * uiIndex ]
      -56 * piSrc[ 27 * uiIndex ]
      +1078 * piSrc[ 29 * uiIndex ]
      +682 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 6; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 19 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +848 * piSrc[ 1 * uiIndex ]
      -682 * piSrc[ 3 * uiIndex ]
      -982 * piSrc[ 5 * uiIndex ]
      +489 * piSrc[ 7 * uiIndex ]
      +1078 * piSrc[ 9 * uiIndex ]
      -278 * piSrc[ 11 * uiIndex ]
      -1132 * piSrc[ 13 * uiIndex ]
      +56 * piSrc[ 15 * uiIndex ]
      +1143 * piSrc[ 17 * uiIndex ]
      +168 * piSrc[ 19 * uiIndex ]
      -1111 * piSrc[ 21 * uiIndex ]
      -386 * piSrc[ 23 * uiIndex ]
      +1035 * piSrc[ 25 * uiIndex ]
      +589 * piSrc[ 27 * uiIndex ]
      -920 * piSrc[ 29 * uiIndex ]
      -769 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 7; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 17 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +769 * piSrc[ 1 * uiIndex ]
      -920 * piSrc[ 3 * uiIndex ]
      -589 * piSrc[ 5 * uiIndex ]
      +1035 * piSrc[ 7 * uiIndex ]
      +386 * piSrc[ 9 * uiIndex ]
      -1111 * piSrc[ 11 * uiIndex ]
      -168 * piSrc[ 13 * uiIndex ]
      +1143 * piSrc[ 15 * uiIndex ]
      -56 * piSrc[ 17 * uiIndex ]
      -1132 * piSrc[ 19 * uiIndex ]
      +278 * piSrc[ 21 * uiIndex ]
      +1078 * piSrc[ 23 * uiIndex ]
      -489 * piSrc[ 25 * uiIndex ]
      -982 * piSrc[ 27 * uiIndex ]
      +682 * piSrc[ 29 * uiIndex ]
      +848 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 8; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 15 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +682 * piSrc[ 1 * uiIndex ]
      -1078 * piSrc[ 3 * uiIndex ]
      -56 * piSrc[ 5 * uiIndex ]
      +1111 * piSrc[ 7 * uiIndex ]
      -589 * piSrc[ 9 * uiIndex ]
      -769 * piSrc[ 11 * uiIndex ]
      +1035 * piSrc[ 13 * uiIndex ]
      +168 * piSrc[ 15 * uiIndex ]
      -1132 * piSrc[ 17 * uiIndex ]
      +489 * piSrc[ 19 * uiIndex ]
      +848 * piSrc[ 21 * uiIndex ]
      -982 * piSrc[ 23 * uiIndex ]
      -278 * piSrc[ 25 * uiIndex ]
      +1143 * piSrc[ 27 * uiIndex ]
      -386 * piSrc[ 29 * uiIndex ]
      -920 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 9; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 13 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +589 * piSrc[ 1 * uiIndex ]
      -1143 * piSrc[ 3 * uiIndex ]
      +489 * piSrc[ 5 * uiIndex ]
      +682 * piSrc[ 7 * uiIndex ]
      -1132 * piSrc[ 9 * uiIndex ]
      +386 * piSrc[ 11 * uiIndex ]
      +769 * piSrc[ 13 * uiIndex ]
      -1111 * piSrc[ 15 * uiIndex ]
      +278 * piSrc[ 17 * uiIndex ]
      +848 * piSrc[ 19 * uiIndex ]
      -1078 * piSrc[ 21 * uiIndex ]
      +168 * piSrc[ 23 * uiIndex ]
      +920 * piSrc[ 25 * uiIndex ]
      -1035 * piSrc[ 27 * uiIndex ]
      +56 * piSrc[ 29 * uiIndex ]
      +982 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 10; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 11 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +489 * piSrc[ 1 * uiIndex ]
      -1111 * piSrc[ 3 * uiIndex ]
      +920 * piSrc[ 5 * uiIndex ]
      -56 * piSrc[ 7 * uiIndex ]
      -848 * piSrc[ 9 * uiIndex ]
      +1132 * piSrc[ 11 * uiIndex ]
      -589 * piSrc[ 13 * uiIndex ]
      -386 * piSrc[ 15 * uiIndex ]
      +1078 * piSrc[ 17 * uiIndex ]
      -982 * piSrc[ 19 * uiIndex ]
      +168 * piSrc[ 21 * uiIndex ]
      +769 * piSrc[ 23 * uiIndex ]
      -1143 * piSrc[ 25 * uiIndex ]
      +682 * piSrc[ 27 * uiIndex ]
      +278 * piSrc[ 29 * uiIndex ]
      -1035 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 11; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 9 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +386 * piSrc[ 1 * uiIndex ]
      -982 * piSrc[ 3 * uiIndex ]
      +1132 * piSrc[ 5 * uiIndex ]
      -769 * piSrc[ 7 * uiIndex ]
      +56 * piSrc[ 9 * uiIndex ]
      +682 * piSrc[ 11 * uiIndex ]
      -1111 * piSrc[ 13 * uiIndex ]
      +1035 * piSrc[ 15 * uiIndex ]
      -489 * piSrc[ 17 * uiIndex ]
      -278 * piSrc[ 19 * uiIndex ]
      +920 * piSrc[ 21 * uiIndex ]
      -1143 * piSrc[ 23 * uiIndex ]
      +848 * piSrc[ 25 * uiIndex ]
      -168 * piSrc[ 27 * uiIndex ]
      -589 * piSrc[ 29 * uiIndex ]
      +1078 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 12; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 7 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +278 * piSrc[ 1 * uiIndex ]
      -769 * piSrc[ 3 * uiIndex ]
      +1078 * piSrc[ 5 * uiIndex ]
      -1132 * piSrc[ 7 * uiIndex ]
      +920 * piSrc[ 9 * uiIndex ]
      -489 * piSrc[ 11 * uiIndex ]
      -56 * piSrc[ 13 * uiIndex ]
      +589 * piSrc[ 15 * uiIndex ]
      -982 * piSrc[ 17 * uiIndex ]
      +1143 * piSrc[ 19 * uiIndex ]
      -1035 * piSrc[ 21 * uiIndex ]
      +682 * piSrc[ 23 * uiIndex ]
      -168 * piSrc[ 25 * uiIndex ]
      -386 * piSrc[ 27 * uiIndex ]
      +848 * piSrc[ 29 * uiIndex ]
      -1111 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 13; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 5 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +168 * piSrc[ 1 * uiIndex ]
      -489 * piSrc[ 3 * uiIndex ]
      +769 * piSrc[ 5 * uiIndex ]
      -982 * piSrc[ 7 * uiIndex ]
      +1111 * piSrc[ 9 * uiIndex ]
      -1143 * piSrc[ 11 * uiIndex ]
      +1078 * piSrc[ 13 * uiIndex ]
      -920 * piSrc[ 15 * uiIndex ]
      +682 * piSrc[ 17 * uiIndex ]
      -386 * piSrc[ 19 * uiIndex ]
      +56 * piSrc[ 21 * uiIndex ]
      +278 * piSrc[ 23 * uiIndex ]
      -589 * piSrc[ 25 * uiIndex ]
      +848 * piSrc[ 27 * uiIndex ]
      -1035 * piSrc[ 29 * uiIndex ]
      +1132 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 14; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 3 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
    {
      int iACContribution =  +56 * piSrc[ 1 * uiIndex ]
      -168 * piSrc[ 3 * uiIndex ]
      +278 * piSrc[ 5 * uiIndex ]
      -386 * piSrc[ 7 * uiIndex ]
      +489 * piSrc[ 9 * uiIndex ]
      -589 * piSrc[ 11 * uiIndex ]
      +682 * piSrc[ 13 * uiIndex ]
      -769 * piSrc[ 15 * uiIndex ]
      +848 * piSrc[ 17 * uiIndex ]
      -920 * piSrc[ 19 * uiIndex ]
      +982 * piSrc[ 21 * uiIndex ]
      -1035 * piSrc[ 23 * uiIndex ]
      +1078 * piSrc[ 25 * uiIndex ]
      -1111 * piSrc[ 27 * uiIndex ]
      +1132 * piSrc[ 29 * uiIndex ]
      -1143 * piSrc[ 31 * uiIndex ]
      ;

      for( UInt uiFwdDataIdx = 15; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 32)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 1 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }
  }



    if( 5 < uiLog2TSize )
    {

      const UInt uiIndex = 1 << gMax(0,((int)uiLog2TSize -6));

      int iACContribution = 1145 * piSrc[ 1 * uiIndex ] + 1142 * piSrc[ 3 * uiIndex ] + 1136 * piSrc[ 5 * uiIndex ] + 1128 * piSrc[ 7 * uiIndex ] + 1117 * piSrc[ 9 * uiIndex ] + 1103 * piSrc[ 11 * uiIndex ] + 1087 * piSrc[ 13 * uiIndex ] + 1068 * piSrc[ 15 * uiIndex ] + 1047 * piSrc[ 17 * uiIndex ] + 1023 * piSrc[ 19 * uiIndex ] + 996 * piSrc[ 21 * uiIndex ] + 967 * piSrc[ 23 * uiIndex ] + 936 * piSrc[ 25 * uiIndex ] + 903 * piSrc[ 27 * uiIndex ] + 867 * piSrc[ 29 * uiIndex ] + 829 * piSrc[ 31 * uiIndex ] + 789 * piSrc[ 33 * uiIndex ] + 748 * piSrc[ 35 * uiIndex ] + 704 * piSrc[ 37 * uiIndex ] + 659 * piSrc[ 39 * uiIndex ] + 613 * piSrc[ 41 * uiIndex ] + 564 * piSrc[ 43 * uiIndex ] + 515 * piSrc[ 45 * uiIndex ] + 464 * piSrc[ 47 * uiIndex ] + 412 * piSrc[ 49 * uiIndex ] + 359 * piSrc[ 51 * uiIndex ] + 305 * piSrc[ 53 * uiIndex ] + 251 * piSrc[ 55 * uiIndex ] + 196 * piSrc[ 57 * uiIndex ] + 140 * piSrc[ 59 * uiIndex ] + 84 * piSrc[ 61 * uiIndex ] + 28 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 0; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 63 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 1142 * piSrc[ 1 * uiIndex ] + 1117 * piSrc[ 3 * uiIndex ] + 1068 * piSrc[ 5 * uiIndex ] + 996 * piSrc[ 7 * uiIndex ] + 903 * piSrc[ 9 * uiIndex ] + 789 * piSrc[ 11 * uiIndex ] + 659 * piSrc[ 13 * uiIndex ] + 515 * piSrc[ 15 * uiIndex ] + 359 * piSrc[ 17 * uiIndex ] + 196 * piSrc[ 19 * uiIndex ] + 28 * piSrc[ 21 * uiIndex ] -140 * piSrc[ 23 * uiIndex ] -305 * piSrc[ 25 * uiIndex ]  -464 * piSrc[ 27 * uiIndex ] -613 * piSrc[ 29 * uiIndex ] -748 * piSrc[ 31 * uiIndex ] -867 * piSrc[ 33 * uiIndex ] -967 * piSrc[ 35 * uiIndex ] -1047 * piSrc[ 37 * uiIndex ] -1103 * piSrc[ 39 * uiIndex ] -1136 * piSrc[ 41 * uiIndex ] -1145 * piSrc[ 43 * uiIndex ] -1128 * piSrc[ 45 * uiIndex ] -1087 * piSrc[ 47 * uiIndex ] -1023 * piSrc[ 49 * uiIndex ] -936 * piSrc[ 51 * uiIndex ] -829 * piSrc[ 53 * uiIndex ] -704 * piSrc[ 55 * uiIndex ] -564 * piSrc[ 57 * uiIndex ] -412 * piSrc[ 59 * uiIndex ] -251 * piSrc[ 61 * uiIndex ] -84 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 1; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 61 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 1136 * piSrc[ 1 * uiIndex ] + 1068 * piSrc[ 3 * uiIndex ] + 936 * piSrc[ 5 * uiIndex ] + 748 * piSrc[ 7 * uiIndex ] + 515 * piSrc[ 9 * uiIndex ] + 251 * piSrc[ 11 * uiIndex ] - 28 * piSrc[ 13 * uiIndex ] -305 * piSrc[ 15 * uiIndex ] -564 * piSrc[ 17 * uiIndex ] -789 * piSrc[ 19 * uiIndex ] -967 * piSrc[ 21 * uiIndex ] -1087 * piSrc[ 23 * uiIndex ] -1142 * piSrc[ 25 * uiIndex ] -1128 * piSrc[ 27 * uiIndex ] -1047 * piSrc[ 29 * uiIndex ] -903 * piSrc[ 31 * uiIndex ] -704 * piSrc[ 33 * uiIndex ] -464 * piSrc[ 35 * uiIndex ] -196 * piSrc[ 37 * uiIndex ] + 84 * piSrc[ 39 * uiIndex ] + 359 * piSrc[ 41 * uiIndex ] + 613 * piSrc[ 43 * uiIndex ] + 829 * piSrc[ 45 * uiIndex ] + 996 * piSrc[ 47 * uiIndex ] + 1103 * piSrc[ 49 * uiIndex ] + 1145 * piSrc[ 51 * uiIndex ] + 1117 * piSrc[ 53 * uiIndex ] + 1023 * piSrc[ 55 * uiIndex ] + 867 * piSrc[ 57 * uiIndex ] + 659 * piSrc[ 59 * uiIndex ] + 412 * piSrc[ 61 * uiIndex ] + 140 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 2; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 59 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 1128 * piSrc[ 1 * uiIndex ] + 996 * piSrc[ 3 * uiIndex ] + 748 * piSrc[ 5 * uiIndex ] + 412 * piSrc[ 7 * uiIndex ] + 28 * piSrc[ 9 * uiIndex ] -359 * piSrc[ 11 * uiIndex ] -704 * piSrc[ 13 * uiIndex ] -967 * piSrc[ 15 * uiIndex ] -1117 * piSrc[ 17 * uiIndex ] -1136 * piSrc[ 19 * uiIndex ] -1023 * piSrc[ 21 * uiIndex ] -789 * piSrc[ 23 * uiIndex ] -464 * piSrc[ 25 * uiIndex ] -84 * piSrc[ 27 * uiIndex ] + 305 * piSrc[ 29 * uiIndex ] + 659 * piSrc[ 31 * uiIndex ] + 936 * piSrc[ 33 * uiIndex ] + 1103 * piSrc[ 35 * uiIndex ] + 1142 * piSrc[ 37 * uiIndex ] + 1047 * piSrc[ 39 * uiIndex ] + 829 * piSrc[ 41 * uiIndex ] + 515 * piSrc[ 43 * uiIndex ] + 140 * piSrc[ 45 * uiIndex ] -251 * piSrc[ 47 * uiIndex ] -613 * piSrc[ 49 * uiIndex ] -903 * piSrc[ 51 * uiIndex ] -1087 * piSrc[ 53 * uiIndex ] -1145 * piSrc[ 55 * uiIndex ] -1068 * piSrc[ 57 * uiIndex ] -867 * piSrc[ 59 * uiIndex ] -564 * piSrc[ 61 * uiIndex ] -196 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 3; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 57 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 1117 * piSrc[ 1 * uiIndex ] + 903 * piSrc[ 3 * uiIndex ] + 515 * piSrc[ 5 * uiIndex ] + 28 * piSrc[ 7 * uiIndex ] -464 * piSrc[ 9 * uiIndex ] -867 * piSrc[ 11 * uiIndex ] - 1103 * piSrc[ 13 * uiIndex ] -1128 * piSrc[ 15 * uiIndex ] -936 * piSrc[ 17 * uiIndex ] -564 * piSrc[ 19 * uiIndex ] -84 * piSrc[ 21 * uiIndex ] + 412 * piSrc[ 23 * uiIndex ] + 829 * piSrc[ 25 * uiIndex ] + 1087 * piSrc[ 27 * uiIndex ] + 1136 * piSrc[ 29 * uiIndex ] + 967 * piSrc[ 31 * uiIndex ] + 613 * piSrc[ 33 * uiIndex ] + 140 * piSrc[ 35 * uiIndex ] -359 * piSrc[ 37 * uiIndex ] -789 * piSrc[ 39 * uiIndex ] -1068 * piSrc[ 41 * uiIndex ] -1142 * piSrc[ 43 * uiIndex ] -996 * piSrc[ 45 * uiIndex ] -659 * piSrc[ 47 * uiIndex ] -196 * piSrc[ 49 * uiIndex ] + 305 * piSrc[ 51 * uiIndex ] + 748 * piSrc[ 53 * uiIndex ] + 1047 * piSrc[ 55 * uiIndex ] + 1145 * piSrc[ 57 * uiIndex ] + 1023 * piSrc[ 59 * uiIndex ] + 704 * piSrc[ 61 * uiIndex ] + 251 * piSrc[ 63 * uiIndex ];


      for( UInt uiFwdDataIdx = 4; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 55 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 1103 * piSrc[ 1 * uiIndex ] + 789 * piSrc[ 3 * uiIndex ] + 251 * piSrc[ 5 * uiIndex ] -359 * piSrc[ 7 * uiIndex ] -867 * piSrc[ 9 * uiIndex ] -1128 * piSrc[ 11 * uiIndex ] -1068 * piSrc[ 13 * uiIndex ] -704 * piSrc[ 15 * uiIndex ] -140 * piSrc[ 17 * uiIndex ] + 464 * piSrc[ 19 * uiIndex ] + 936 * piSrc[ 21 * uiIndex ] + 1142 * piSrc[ 23 * uiIndex ] + 1023 * piSrc[ 25 * uiIndex ] + 613 * piSrc[ 27 * uiIndex ] + 28 * piSrc[ 29 * uiIndex ] -564 * piSrc[ 31 * uiIndex ] -996 * piSrc[ 33 * uiIndex ] -1145 * piSrc[ 35 * uiIndex ] -967 * piSrc[ 37 * uiIndex ] -515 * piSrc[ 39 * uiIndex ] + 84 * piSrc[ 41 * uiIndex ] + 659 * piSrc[ 43 * uiIndex ] + 1047 * piSrc[ 45 * uiIndex ] + 1136 * piSrc[ 47 * uiIndex ] + 903 * piSrc[ 49 * uiIndex ] + 412 * piSrc[ 51 * uiIndex ] -196 * piSrc[ 53 * uiIndex ] -748 * piSrc[ 55 * uiIndex ] -1087 * piSrc[ 57 * uiIndex ] -1117 * piSrc[ 59 * uiIndex ] -829 * piSrc[ 61 * uiIndex ] -305 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 5; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 53 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 1087 * piSrc[ 1 * uiIndex ] + 659 * piSrc[ 3 * uiIndex ] -28 * piSrc[ 5 * uiIndex ] -704 * piSrc[ 7 * uiIndex ] -1103 * piSrc[ 9 * uiIndex ] -1068 * piSrc[ 11 * uiIndex ] -613 * piSrc[ 13 * uiIndex ] + 84 * piSrc[ 15 * uiIndex ] + 748 * piSrc[ 17 * uiIndex ] + 1117 * piSrc[ 19 * uiIndex ] + 1047 * piSrc[ 21 * uiIndex ] + 564 * piSrc[ 23 * uiIndex ] -140 * piSrc[ 25 * uiIndex ] -789 * piSrc[ 27 * uiIndex ] -1128 * piSrc[ 29 * uiIndex ] -1023 * piSrc[ 31 * uiIndex ] -515 * piSrc[ 33 * uiIndex ] + 196 * piSrc[ 35 * uiIndex ] + 829 * piSrc[ 37 * uiIndex ] + 1136 * piSrc[ 39 * uiIndex ] + 996 * piSrc[ 41 * uiIndex ] + 464 * piSrc[ 43 * uiIndex ] -251 * piSrc[ 45 * uiIndex ] -867 * piSrc[ 47 * uiIndex ] -1142 * piSrc[ 49 * uiIndex ] -967 * piSrc[ 51 * uiIndex ] -412 * piSrc[ 53 * uiIndex ] + 305 * piSrc[ 55 * uiIndex ] + 903 * piSrc[ 57 * uiIndex ] + 1145 * piSrc[ 59 * uiIndex ] + 936 * piSrc[ 61 * uiIndex ] + 359 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 6; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 51 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 1068 * piSrc[ 1 * uiIndex ] + 515 * piSrc[ 3 * uiIndex ] -305 * piSrc[ 5 * uiIndex ] -967 * piSrc[ 7 * uiIndex ] -1128 * piSrc[ 9 * uiIndex ] -704 * piSrc[ 11 * uiIndex ] + 84 * piSrc[ 13 * uiIndex ] + 829 * piSrc[ 15 * uiIndex ] + 1145 * piSrc[ 17 * uiIndex ] + 867 * piSrc[ 19 * uiIndex ] + 140 * piSrc[ 21 * uiIndex ] -659 * piSrc[ 23 * uiIndex ] -1117 * piSrc[ 25 * uiIndex ] -996 * piSrc[ 27 * uiIndex ] -359 * piSrc[ 29 * uiIndex ] + 464 * piSrc[ 31 * uiIndex ] + 1047 * piSrc[ 33 * uiIndex ] + 1087 * piSrc[ 35 * uiIndex ] + 564 * piSrc[ 37 * uiIndex ] + -251 * piSrc[ 39 * uiIndex ] -936 * piSrc[ 41 * uiIndex ] -1136 * piSrc[ 43 * uiIndex ] -748 * piSrc[ 45 * uiIndex ] + 28 * piSrc[ 47 * uiIndex ] + 789 * piSrc[ 49 * uiIndex ] + 1142 * piSrc[ 51 * uiIndex ] + 903 * piSrc[ 53 * uiIndex ] + 196 * piSrc[ 55 * uiIndex ] -613 * piSrc[ 57 * uiIndex ] -1103 * piSrc[ 59 * uiIndex ] -1023 * piSrc[ 61 * uiIndex ] -412 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 7; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 49 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 1047 * piSrc[ 1 * uiIndex ] + 359 * piSrc[ 3 * uiIndex ] -564 * piSrc[ 5 * uiIndex ] -1117 * piSrc[ 7 * uiIndex ] -936 * piSrc[ 9 * uiIndex ] -140 * piSrc[ 11 * uiIndex ] + 748 * piSrc[ 13 * uiIndex ] + 1145 * piSrc[ 15 * uiIndex ] + 789 * piSrc[ 17 * uiIndex ] -84 * piSrc[ 19 * uiIndex ] -903 * piSrc[ 21 * uiIndex ] -1128 * piSrc[ 23 * uiIndex ] -613 * piSrc[25 * uiIndex ] + 305 * piSrc[ 27 * uiIndex ] + 1023 * piSrc[ 29 * uiIndex ] + 1068 * piSrc[ 31 * uiIndex ] + 412 * piSrc[ 33 * uiIndex ] -515 * piSrc[ 35 * uiIndex ] -1103 * piSrc[ 37 * uiIndex ] -967 * piSrc[ 39 * uiIndex ] -196 * piSrc[ 41 * uiIndex ] + 704 * piSrc[ 43 * uiIndex ] + 1142 * piSrc[ 45 * uiIndex ] + 829 * piSrc[ 47 * uiIndex ] -28 * piSrc[ 49 * uiIndex ] -867 * piSrc[51 * uiIndex ] -1136 * piSrc[ 53 * uiIndex ] -659 * piSrc[ 55 * uiIndex ] + 251 * piSrc[ 57 * uiIndex ] + 996 * piSrc[ 59 * uiIndex ] + 1087 * piSrc[ 61 * uiIndex ] + 464 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 8; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 47 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 1023 * piSrc[ 1 * uiIndex ] + 196 * piSrc[ 3 * uiIndex ] -789 * piSrc[ 5 * uiIndex ] -1136 * piSrc[ 7 * uiIndex ] -564 * piSrc[ 9 * uiIndex ] + 464 * piSrc[ 11 * uiIndex ] + 1117 * piSrc[ 13 * uiIndex ] + 867 * piSrc[ 15 * uiIndex ] -84 * piSrc[ 17 * uiIndex ] -967 * piSrc[ 19 * uiIndex ] -1068 * piSrc[ 21 * uiIndex ] -305 * piSrc[ 23 * uiIndex ] + 704 * piSrc[25 * uiIndex ] + 1145 * piSrc[ 27 * uiIndex ] + 659 * piSrc[ 29 * uiIndex ] -359 * piSrc[ 31 * uiIndex ] -1087 * piSrc[ 33 * uiIndex ] -936 * piSrc[ 35 * uiIndex ] -28 * piSrc[ 37 * uiIndex ] + 903 * piSrc[ 39 * uiIndex ] + 1103 * piSrc[ 41 * uiIndex ] + 412 * piSrc[ 43 * uiIndex ] -613 * piSrc[ 45 * uiIndex ] -1142 * piSrc[ 47 * uiIndex ] -748 * piSrc[ 49 * uiIndex ] + 251 * piSrc[ 51 * uiIndex ] + 1047 * piSrc[ 53 * uiIndex ] + 996 * piSrc[ 55 * uiIndex ] + 140 * piSrc[ 57 * uiIndex ] -829 * piSrc[ 59 * uiIndex ] -1128 * piSrc[ 61 * uiIndex ] -515 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 9; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 45 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 996 * piSrc[ 1 * uiIndex ] + 28 * piSrc[ 3 * uiIndex ] -967 * piSrc[ 5 * uiIndex ] -1023 * piSrc[ 7 * uiIndex ] -84 * piSrc[ 9 * uiIndex ] + 936 * piSrc[ 11 * uiIndex ] + 1047 * piSrc[ 13 * uiIndex ] + 140 * piSrc[ 15 * uiIndex ] -903 * piSrc[ 17 * uiIndex ] -1068 * piSrc[ 19 * uiIndex ] -196 * piSrc[ 21 * uiIndex ] + 867 * piSrc[ 23 * uiIndex ] + 1087 * piSrc[ 25 * uiIndex ] + 251 * piSrc[ 27 * uiIndex ] -829 * piSrc[ 29 * uiIndex ] -1103 * piSrc[ 31 * uiIndex ] -305 * piSrc[ 33 * uiIndex ] + 789 * piSrc[ 35 * uiIndex ] + 1117 * piSrc[ 37 * uiIndex ] + 359 * piSrc[ 39 * uiIndex ] -748 * piSrc[ 41 * uiIndex ] -1128 * piSrc[ 43 * uiIndex ] -412 * piSrc[ 45 * uiIndex ] + 704 * piSrc[ 47 * uiIndex ] + 1136 * piSrc[ 49 * uiIndex ] + 464 * piSrc[ 51 * uiIndex ] -659 * piSrc[ 53 * uiIndex ] -1142 * piSrc[ 55 * uiIndex ] -515 * piSrc[ 57 * uiIndex ] + 613 * piSrc[ 59 * uiIndex ] + 1145 * piSrc[ 61 * uiIndex ] + 564 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 10; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 43 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 967 * piSrc[ 1 * uiIndex ] -140 * piSrc[ 3 * uiIndex ] -1087 * piSrc[ 5 * uiIndex ] -789 * piSrc[ 7 * uiIndex ] + 412 * piSrc[ 9 * uiIndex ] + 1142 * piSrc[ 11 * uiIndex ] + 564 * piSrc[ 13 * uiIndex ] -659 * piSrc[ 15 * uiIndex ] -1128 * piSrc[ 17 * uiIndex ] -305 * piSrc[ 19 * uiIndex ] + 867 * piSrc[ 21 * uiIndex ] + 1047 * piSrc[ 23 * uiIndex ] + 28 * piSrc[ 25 * uiIndex ] -1023 * piSrc[ 27 * uiIndex ] -903 * piSrc[ 29 * uiIndex ] + 251 * piSrc[ 31 * uiIndex ] + 1117 * piSrc[ 33 * uiIndex ] + 704 * piSrc[ 35 * uiIndex ] -515 * piSrc[ 37 * uiIndex ] -1145 * piSrc[ 39 * uiIndex ] -464 * piSrc[ 41 * uiIndex ] + 748 * piSrc[ 43 * uiIndex ] + 1103 * piSrc[ 45 * uiIndex ] + 196 * piSrc[ 47 * uiIndex ] -936 * piSrc[ 49 * uiIndex ] -996 * piSrc[ 51 * uiIndex ] + 84 * piSrc[ 53 * uiIndex ] + 1068 * piSrc[ 55 * uiIndex ] + 829 * piSrc[ 57 * uiIndex ] -359 * piSrc[ 59 * uiIndex ] -1136 * piSrc[ 61 * uiIndex ] -613 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 11; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 41 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 936 * piSrc[ 1 * uiIndex ] -305 * piSrc[ 3 * uiIndex ] -1142 * piSrc[ 5 * uiIndex ] -464 * piSrc[ 7 * uiIndex ] + 829 * piSrc[ 9 * uiIndex ] + 1023 * piSrc[ 11 * uiIndex ] -140 * piSrc[ 13 * uiIndex ] -1117 * piSrc[ 15 * uiIndex ] -613 * piSrc[ 17 * uiIndex ] + 704 * piSrc[ 19 * uiIndex ] + 1087 * piSrc[ 21 * uiIndex ] + 28 * piSrc[ 23 * uiIndex ] -1068 * piSrc[ 25 * uiIndex ] -748 * piSrc[ 27 * uiIndex ] + 564 * piSrc[ 29 * uiIndex ] + 1128 * piSrc[ 31 * uiIndex ] + 196 * piSrc[ 33 * uiIndex ] -996 * piSrc[ 35 * uiIndex ] -867 * piSrc[ 37 * uiIndex ] + 412 * piSrc[ 39 * uiIndex ] + 1145 * piSrc[ 41 * uiIndex ] + 359 * piSrc[ 43 * uiIndex ] -903 * piSrc[ 45 * uiIndex ] -967 * piSrc[ 47 * uiIndex ] + 251 * piSrc[ 49 * uiIndex ] + 1136 * piSrc[ 51 * uiIndex ] + 515 * piSrc[ 53 * uiIndex ] -789 * piSrc[ 55 * uiIndex ] -1047 * piSrc[ 57 * uiIndex ] + 84 * piSrc[ 59 * uiIndex ] + 1103 * piSrc[ 61 * uiIndex ] + 659 * piSrc[ 63 * uiIndex ];


      for( UInt uiFwdDataIdx = 12; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 39 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 903 * piSrc[ 1 * uiIndex ] -464 * piSrc[ 3 * uiIndex ] -1128 * piSrc[ 5 * uiIndex ] -84 * piSrc[ 7 * uiIndex ] + 1087 * piSrc[ 9 * uiIndex ] + 613 * piSrc[ 11 * uiIndex ] -789 * piSrc[ 13 * uiIndex ] -996 * piSrc[ 15 * uiIndex ] + 305 * piSrc[ 17 * uiIndex ] + 1145 * piSrc[ 19 * uiIndex ] + 251 * piSrc[ 21 * uiIndex ] -1023 * piSrc[ 23 * uiIndex ] -748 * piSrc[25 * uiIndex ] + 659 * piSrc[ 27 * uiIndex ] + 1068 * piSrc[ 29 * uiIndex ] -140 * piSrc[ 31 * uiIndex ] -1136 * piSrc[ 33 * uiIndex ] -412 * piSrc[ 35 * uiIndex ] + 936 * piSrc[ 37 * uiIndex ] + 867 * piSrc[ 39 * uiIndex ] -515 * piSrc[ 41 * uiIndex ] -1117 * piSrc[ 43 * uiIndex ] -28 * piSrc[ 45 * uiIndex ] + 1103 * piSrc[ 47 * uiIndex ] + 564 * piSrc[ 49 * uiIndex ] -829 * piSrc[51 * uiIndex ] -967 * piSrc[ 53 * uiIndex ] + 359 * piSrc[ 55 * uiIndex ] + 1142 * piSrc[ 57 * uiIndex ] + 196 * piSrc[ 59 * uiIndex ] -1047 * piSrc[ 61 * uiIndex ] -704 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 13; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 37 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 867 * piSrc[ 1 * uiIndex ] -613 * piSrc[ 3 * uiIndex ] -1047 * piSrc[ 5 * uiIndex ] + 305 * piSrc[ 7 * uiIndex ] + 1136 * piSrc[ 9 * uiIndex ] + 28 * piSrc[ 11 * uiIndex ] -1128 * piSrc[ 13 * uiIndex ] -359 * piSrc[ 15 * uiIndex ] + 1023 * piSrc[ 17 * uiIndex ] + 659 * piSrc[ 19 * uiIndex ] -829 * piSrc[ 21 * uiIndex ] -903 * piSrc[ 23 * uiIndex ] + 564 * piSrc[ 25 * uiIndex ] + 1068 * piSrc[ 27 * uiIndex ] -251 * piSrc[ 29 * uiIndex ] -1142 * piSrc[ 31 * uiIndex ] -84 * piSrc[ 33 * uiIndex ] + 1117 * piSrc[ 35 * uiIndex ] + 412 * piSrc[ 37 * uiIndex ] -996 * piSrc[ 39 * uiIndex ] -704 * piSrc[ 41 * uiIndex ] + 789 * piSrc[ 43 * uiIndex ] + 936 * piSrc[ 45 * uiIndex ] -515 * piSrc[ 47 * uiIndex ] -1087 * piSrc[ 49 * uiIndex ] + 196 * piSrc[ 51 * uiIndex ] + 1145 * piSrc[ 53 * uiIndex ] + 140 * piSrc[ 55 * uiIndex ] -1103 * piSrc[ 57 * uiIndex ] -464 * piSrc[ 59 * uiIndex ] + 967 * piSrc[ 61 * uiIndex ] + 748 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 14; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 35 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 829 * piSrc[ 1 * uiIndex ] -748 * piSrc[ 3 * uiIndex ] -903 * piSrc[ 5 * uiIndex ] + 659 * piSrc[ 7 * uiIndex ] + 967 * piSrc[ 9 * uiIndex ] -564 * piSrc[ 11 * uiIndex ] -1023 * piSrc[ 13 * uiIndex ] + 464 * piSrc[ 15 * uiIndex ] + 1068 * piSrc[ 17 * uiIndex ] -359 * piSrc[ 19 * uiIndex ] -1103 * piSrc[ 21 * uiIndex ] + 251 * piSrc[ 23 * uiIndex ] + 1128 * piSrc[25 * uiIndex ] -140 * piSrc[ 27 * uiIndex ] -1142 * piSrc[ 29 * uiIndex ] + 28 * piSrc[ 31 * uiIndex ] + 1145 * piSrc[ 33 * uiIndex ] + 84 * piSrc[ 35 * uiIndex ] -1136 * piSrc[ 37 * uiIndex ] -196 * piSrc[ 39 * uiIndex ] + 1117 * piSrc[ 41 * uiIndex ] + 305 * piSrc[ 43 * uiIndex ] -1087 * piSrc[ 45 * uiIndex ] -412 * piSrc[ 47 * uiIndex ] + 1047 * piSrc[ 49 * uiIndex ] + 515 * piSrc[51 * uiIndex ] -996 * piSrc[ 53 * uiIndex ] -613 * piSrc[ 55 * uiIndex ] + 936 * piSrc[ 57 * uiIndex ] + 704 * piSrc[ 59 * uiIndex ] -867 * piSrc[ 61 * uiIndex ] -789 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 15; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 33 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 789 * piSrc[ 1 * uiIndex ] -867 * piSrc[ 3 * uiIndex ] -704 * piSrc[ 5 * uiIndex ] + 936 * piSrc[ 7 * uiIndex ] + 613 * piSrc[ 9 * uiIndex ] -996 * piSrc[ 11 * uiIndex ] -515 * piSrc[ 13 * uiIndex ] + 1047 * piSrc[ 15 * uiIndex ] + 412 * piSrc[ 17 * uiIndex ] -1087 * piSrc[ 19 * uiIndex ] -305 * piSrc[ 21 * uiIndex ] + 1117 * piSrc[ 23 * uiIndex ] + 196 * piSrc[ 25 * uiIndex ] -1136 * piSrc[ 27 * uiIndex ] -84 * piSrc[ 29 * uiIndex ] + 1145 * piSrc[ 31 * uiIndex ] -28 * piSrc[ 33 * uiIndex ] -1142 * piSrc[ 35 * uiIndex ] + 140 * piSrc[ 37 * uiIndex ] + 1128 * piSrc[ 39 * uiIndex ] -251 * piSrc[ 41 * uiIndex ] -1103 * piSrc[ 43 * uiIndex ] + 359 * piSrc[ 45 * uiIndex ] + 1068 * piSrc[ 47 * uiIndex ] -464 * piSrc[ 49 * uiIndex ] -1023 * piSrc[ 51 * uiIndex ] + 564 * piSrc[ 53 * uiIndex ] + 967 * piSrc[ 55 * uiIndex ] -659 * piSrc[ 57 * uiIndex ] -903 * piSrc[ 59 * uiIndex ] + 748 * piSrc[ 61 * uiIndex ] + 829 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 16; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 31 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 748 * piSrc[ 1 * uiIndex ] -967 * piSrc[ 3 * uiIndex ] -464 * piSrc[ 5 * uiIndex ] + 1103 * piSrc[ 7 * uiIndex ] + 140 * piSrc[ 9 * uiIndex ] -1145 * piSrc[ 11 * uiIndex ] + 196 * piSrc[ 13 * uiIndex ] + 1087 * piSrc[ 15 * uiIndex ] -515 * piSrc[ 17 * uiIndex ] -936 * piSrc[ 19 * uiIndex ] + 789 * piSrc[ 21 * uiIndex ] + 704 * piSrc[ 23 * uiIndex ] -996 * piSrc[ 25 * uiIndex ] -412 * piSrc[ 27 * uiIndex ] + 1117 * piSrc[ 29 * uiIndex ] + 84 * piSrc[ 31 * uiIndex ] -1142 * piSrc[ 33 * uiIndex ] + 251 * piSrc[ 35 * uiIndex ] + 1068 * piSrc[ 37 * uiIndex ] -564 * piSrc[ 39 * uiIndex ] -903 * piSrc[ 41 * uiIndex ] + 829 * piSrc[ 43 * uiIndex ] + 659 * piSrc[ 45 * uiIndex ] -1023 * piSrc[ 47 * uiIndex ] -359 * piSrc[ 49 * uiIndex ] + 1128 * piSrc[ 51 * uiIndex ] + 28 * piSrc[ 53 * uiIndex ] -1136 * piSrc[ 55 * uiIndex ] + 305 * piSrc[ 57 * uiIndex ] + 1047 * piSrc[ 59 * uiIndex ] -613 * piSrc[ 61 * uiIndex ] -867 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 17; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 29 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 704 * piSrc[ 1 * uiIndex ] -1047 * piSrc[ 3 * uiIndex ] -196 * piSrc[ 5 * uiIndex ] + 1142 * piSrc[ 7 * uiIndex ] -359 * piSrc[ 9 * uiIndex ] -967 * piSrc[ 11 * uiIndex ] + 829 * piSrc[ 13 * uiIndex ] + 564 * piSrc[ 15 * uiIndex ] -1103 * piSrc[ 17 * uiIndex ] -28 * piSrc[ 19 * uiIndex ] + 1117 * piSrc[ 21 * uiIndex ] -515 * piSrc[ 23 * uiIndex ] -867 * piSrc[25 * uiIndex ] + 936 * piSrc[ 27 * uiIndex ] + 412 * piSrc[ 29 * uiIndex ] -1136 * piSrc[ 31 * uiIndex ] + 140 * piSrc[ 33 * uiIndex ] + 1068 * piSrc[ 35 * uiIndex ] -659 * piSrc[ 37 * uiIndex ] -748 * piSrc[ 39 * uiIndex ] + 1023 * piSrc[ 41 * uiIndex ] + 251 * piSrc[ 43 * uiIndex ] -1145 * piSrc[ 45 * uiIndex ] + 305 * piSrc[ 47 * uiIndex ] + 996 * piSrc[ 49 * uiIndex ] -789 * piSrc[51 * uiIndex ] -613 * piSrc[ 53 * uiIndex ] + 1087 * piSrc[ 55 * uiIndex ] + 84 * piSrc[ 57 * uiIndex ] -1128 * piSrc[ 59 * uiIndex ] + 464 * piSrc[ 61 * uiIndex ] + 903 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 18; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 27 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 659 * piSrc[ 1 * uiIndex ] -1103 * piSrc[ 3 * uiIndex ] + 84 * piSrc[ 5 * uiIndex ] + 1047 * piSrc[ 7 * uiIndex ] -789 * piSrc[ 9 * uiIndex ] -515 * piSrc[ 11 * uiIndex ] + 1136 * piSrc[ 13 * uiIndex ] -251 * piSrc[ 15 * uiIndex ] -967 * piSrc[ 17 * uiIndex ] + 903 * piSrc[ 19 * uiIndex ] + 359 * piSrc[ 21 * uiIndex ] -1145 * piSrc[ 23 * uiIndex ] + 412 * piSrc[ 25 * uiIndex ] + 867 * piSrc[ 27 * uiIndex ] -996 * piSrc[ 29 * uiIndex ] -196 * piSrc[ 31 * uiIndex ] + 1128 * piSrc[ 33 * uiIndex ] -564 * piSrc[ 35 * uiIndex ] -748 * piSrc[ 37 * uiIndex ] +1068 * piSrc[ 39 * uiIndex ] + 28 * piSrc[ 41 * uiIndex ] -1087 * piSrc[ 43 * uiIndex ] + 704 * piSrc[ 45 * uiIndex ] + 613 * piSrc[ 47 * uiIndex ] -1117 * piSrc[ 49 * uiIndex ] + 140 * piSrc[ 51* uiIndex ] + 1023 * piSrc[ 53 * uiIndex ] -829 * piSrc[ 55 * uiIndex ] -464 * piSrc[ 57 * uiIndex ] + 1142 * piSrc[ 59 * uiIndex ] -305 * piSrc[ 61 * uiIndex ] -936 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 19; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 25 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 613 * piSrc[ 1 * uiIndex ] -1136 * piSrc[ 3 * uiIndex ] + 359 * piSrc[ 5 * uiIndex ] + 829 * piSrc[ 7 * uiIndex ] -1068 * piSrc[ 9 * uiIndex ] + 84 * piSrc[ 11 * uiIndex ] + 996 * piSrc[ 13 * uiIndex ] -936 * piSrc[ 15 * uiIndex ] -196 * piSrc[ 17 * uiIndex ] + 1103 * piSrc[ 19 * uiIndex ] -748 * piSrc[ 21 * uiIndex ] -464 * piSrc[ 23 * uiIndex ] + 1145 * piSrc[ 25 * uiIndex ] -515 * piSrc[ 27 * uiIndex ] -704 * piSrc[ 29 * uiIndex ] + 1117 * piSrc[ 31 * uiIndex ] -251 * piSrc[ 33 * uiIndex ] -903 * piSrc[ 35 * uiIndex ] + 1023 * piSrc[ 37 * uiIndex ] + 28 * piSrc[ 39 * uiIndex ] -1047 * piSrc[ 41 * uiIndex ] + 867 * piSrc[ 43 * uiIndex ] + 305 * piSrc[ 45 * uiIndex ] -1128 * piSrc[ 47 * uiIndex ] + 659 * piSrc[ 49 * uiIndex ] + 564 * piSrc[ 51 * uiIndex ] -1142 * piSrc[ 53 * uiIndex ] + 412 * piSrc[ 55 * uiIndex ] + 789 * piSrc[ 57 * uiIndex ] -1087 * piSrc[ 59 * uiIndex ] + 140 * piSrc[ 61 * uiIndex ] + 967 * piSrc[ 63 * uiIndex ];


      for( UInt uiFwdDataIdx = 20; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 23 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 564 * piSrc[ 1 * uiIndex ] -1145 * piSrc[ 3 * uiIndex ] + 613 * piSrc[ 5 * uiIndex ] + 515 * piSrc[ 7 * uiIndex ] -1142 * piSrc[ 9 * uiIndex ] + 659 * piSrc[ 11 * uiIndex ] + 464 * piSrc[ 13 * uiIndex ] -1136 * piSrc[ 15 * uiIndex ] + 704 * piSrc[ 17 * uiIndex ] + 412 * piSrc[ 19 * uiIndex ] -1128 * piSrc[ 21 * uiIndex ] + 748 * piSrc[ 23 * uiIndex ] + 359 * piSrc[ 25 * uiIndex ] -1117 * piSrc[ 27 * uiIndex ] + 789 * piSrc[ 29 * uiIndex ] + 305 * piSrc[ 31 * uiIndex ] -1103 * piSrc[ 33 * uiIndex ] + 829 * piSrc[ 35 * uiIndex ] + 251 * piSrc[ 37 * uiIndex ] - 1087 * piSrc[ 39 * uiIndex ] + 867 * piSrc[ 41 * uiIndex ] + 196 * piSrc[ 43 * uiIndex ] -1068 * piSrc[ 45 * uiIndex ] + 903 * piSrc[ 47 * uiIndex ] + 140 * piSrc[ 49 * uiIndex ] -1047 * piSrc[ 51 * uiIndex ] + 936 * piSrc[ 53 * uiIndex ] + 84 * piSrc[ 55 * uiIndex ] -1023 * piSrc[ 57 * uiIndex ] + 967 * piSrc[ 59 * uiIndex ] + 28 * piSrc[ 61 * uiIndex ] -996 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 21; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 21 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 515 * piSrc[ 1 * uiIndex ] -1128 * piSrc[ 3 * uiIndex ] + 829 * piSrc[ 5 * uiIndex ] + 140 * piSrc[ 7 * uiIndex ] -996 * piSrc[ 9 * uiIndex ] + 1047 * piSrc[ 11 * uiIndex ] -251 * piSrc[ 13 * uiIndex ] -748 * piSrc[ 15 * uiIndex ] + 1142 * piSrc[ 17 * uiIndex ] -613 * piSrc[ 19 * uiIndex ] -412 * piSrc[ 21 * uiIndex ] + 1103 * piSrc[ 23 * uiIndex ] -903 * piSrc[25 * uiIndex ] -28 * piSrc[ 27 * uiIndex ] + 936 * piSrc[ 29 * uiIndex ] -1087 * piSrc[ 31 * uiIndex ] + 359 * piSrc[ 33 * uiIndex ] + 659 * piSrc[ 35 * uiIndex ] -1145 * piSrc[ 37 * uiIndex ]+ 704 * piSrc[ 39 * uiIndex ] + 305 * piSrc[ 41 * uiIndex ] -1068 * piSrc[ 43 * uiIndex ] + 967 * piSrc[ 45 * uiIndex ] -84 * piSrc[ 47 * uiIndex ] -867 * piSrc[ 49 * uiIndex ] + 1117 * piSrc[ 51 * uiIndex ] -464 * piSrc[ 53 * uiIndex ] -564 * piSrc[ 55 * uiIndex ] + 1136 * piSrc[ 57 * uiIndex ] -789 * piSrc[ 59 * uiIndex ] -196 * piSrc[ 61 * uiIndex ] + 1023 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 22; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 19 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 464 * piSrc[ 1 * uiIndex ] -1087 * piSrc[ 3 * uiIndex ] + 996 * piSrc[ 5 * uiIndex ] -251 * piSrc[ 7 * uiIndex ] -659 * piSrc[ 9 * uiIndex ] + 1136 * piSrc[ 11 * uiIndex ] -867 * piSrc[ 13 * uiIndex ] + 28 * piSrc[ 15 * uiIndex ] + 829 * piSrc[ 17 * uiIndex ] -1142 * piSrc[ 19 * uiIndex ] + 704 * piSrc[ 21 * uiIndex ] + 196 * piSrc[ 23 * uiIndex ] -967 * piSrc[ 25 * uiIndex ] + 1103 * piSrc[ 27 * uiIndex ] -515 * piSrc[ 29 * uiIndex ] -412 * piSrc[ 31 * uiIndex ] + 1068 * piSrc[ 33 * uiIndex ] -1023 * piSrc[ 35 * uiIndex ] + 305 * piSrc[ 37 * uiIndex ] + 613 * piSrc[ 39 * uiIndex ] -1128 * piSrc[ 41 * uiIndex ] + 903 * piSrc[ 43 * uiIndex ] -84 * piSrc[ 45 * uiIndex ] -789 * piSrc[ 47 * uiIndex ] + 1145 * piSrc[ 49 * uiIndex ] -748 * piSrc[ 51 * uiIndex ] -140 * piSrc[ 53 * uiIndex ] + 936 * piSrc[ 55 * uiIndex ] -1117 * piSrc[ 57 * uiIndex ] + 564 * piSrc[ 59 * uiIndex ] + 359 * piSrc[ 61 * uiIndex ] -1047 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 23; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 17 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 412 * piSrc[ 1 * uiIndex ] -1023 * piSrc[ 3 * uiIndex ] + 1103 * piSrc[ 5 * uiIndex ] -613 * piSrc[ 7 * uiIndex ] -196 * piSrc[ 9 * uiIndex ] + 903 * piSrc[ 11 * uiIndex ] -1142 * piSrc[ 13 * uiIndex ] + 789 * piSrc[ 15 * uiIndex ] -28 * piSrc[ 17 * uiIndex ] -748 * piSrc[ 19 * uiIndex ] + 1136 * piSrc[ 21 * uiIndex ] -936 * piSrc[ 23 * uiIndex ] + 251 * piSrc[25 * uiIndex ] + 564 * piSrc[ 27 * uiIndex ] -1087 * piSrc[ 29 * uiIndex ] + 1047 * piSrc[ 31 * uiIndex ] -464 * piSrc[ 33 * uiIndex ] -359 * piSrc[ 35 * uiIndex ] + 996 * piSrc[ 37 * uiIndex ] -1117 * piSrc[ 39 * uiIndex ] + 659 * piSrc[ 41 * uiIndex ] + 140 * piSrc[ 43 * uiIndex ] -867 * piSrc[ 45 * uiIndex ] + 1145 * piSrc[ 47 * uiIndex ] -829 * piSrc[ 49 * uiIndex ] + 84 * piSrc[ 51 * uiIndex ] + 704 * piSrc[ 53 * uiIndex ] -1128 * piSrc[ 55 * uiIndex ] + 967 * piSrc[ 57 * uiIndex ] -305 * piSrc[ 59 * uiIndex ] -515 * piSrc[ 61 * uiIndex ] + 1068 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 24; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 15 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 359 * piSrc[ 1 * uiIndex ] -936 * piSrc[ 3 * uiIndex ] + 1145 * piSrc[ 5 * uiIndex ] -903 * piSrc[ 7 * uiIndex ] + 305 * piSrc[ 9 * uiIndex ] + 412 * piSrc[ 11 * uiIndex ] -967 * piSrc[ 13 * uiIndex ] + 1142 * piSrc[ 15 * uiIndex ] -867 * piSrc[ 17 * uiIndex ] + 251 * piSrc[ 19 * uiIndex ] + 464 * piSrc[ 21 * uiIndex ] -996 * piSrc[ 23 * uiIndex ] + 1136 * piSrc[ 25 * uiIndex ] -829 * piSrc[ 27 * uiIndex ] + 196 * piSrc[ 29 * uiIndex ] + 515 * piSrc[ 31 * uiIndex ] -1023 * piSrc[ 33 * uiIndex ] + 1128 * piSrc[ 35 * uiIndex ] -789 * piSrc[ 37 * uiIndex ] + 140 * piSrc[ 39 * uiIndex ] + 564 * piSrc[ 41 * uiIndex ] -1047 * piSrc[ 43 * uiIndex ] + 1117 * piSrc[ 45 * uiIndex ] -748 * piSrc[ 47 * uiIndex ] + 84 * piSrc[ 49 * uiIndex ] + 613 * piSrc[ 51 *uiIndex ] -1068 * piSrc[ 53 * uiIndex ] + 1103 * piSrc[ 55 * uiIndex ] -704 * piSrc[ 57 * uiIndex ] + 28 * piSrc[ 59 * uiIndex ] + 659 * piSrc[ 61 * uiIndex ] -1087 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 25; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 13 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 305 * piSrc[ 1 * uiIndex ] -829 * piSrc[ 3 * uiIndex ] + 1117 * piSrc[ 5 * uiIndex ] -1087 * piSrc[ 7 * uiIndex ] + 748 * piSrc[ 9 * uiIndex ] -196 * piSrc[ 11 * uiIndex ] -412 * piSrc[ 13 * uiIndex ] + 903 * piSrc[ 15 * uiIndex ] -1136 * piSrc[ 17 * uiIndex ] + 1047 * piSrc[ 19 * uiIndex ] -659 * piSrc[ 21 * uiIndex ] + 84 * piSrc[ 23 * uiIndex ] + 515 * piSrc[ 25 * uiIndex ] -967 * piSrc[ 27 * uiIndex ] + 1145 * piSrc[ 29 * uiIndex ] -996 * piSrc[ 31 * uiIndex ] + 564 * piSrc[ 33 * uiIndex ] + 28 * piSrc[ 35 * uiIndex ] -613 * piSrc[ 37 * uiIndex ] + 1023 * piSrc[ 39 * uiIndex ] -1142 * piSrc[ 41 * uiIndex ] + 936 * piSrc[ 43 * uiIndex ] -464 * piSrc[ 45 * uiIndex ] -140 * piSrc[ 47 * uiIndex ] + 704 * piSrc[ 49 * uiIndex ] -1068 * piSrc[ 51 * uiIndex ] + 1128 * piSrc[ 53 * uiIndex ] -867 * piSrc[ 55 * uiIndex ] + 359 * piSrc[ 57 * uiIndex ] + 251 * piSrc[ 59 * uiIndex ] -789 * piSrc[ 61 * uiIndex ] + 1103 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 26; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 11 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 251 * piSrc[ 1 * uiIndex ] -704 * piSrc[ 3 * uiIndex ] + 1023 * piSrc[ 5 * uiIndex ] -1145 * piSrc[ 7 * uiIndex ] + 1047 * piSrc[ 9 * uiIndex ] -748 * piSrc[ 11 * uiIndex ] + 305 * piSrc[ 13 * uiIndex ] + 196 * piSrc[ 15 * uiIndex ] -659 * piSrc[ 17 * uiIndex ] + 996 * piSrc[ 19 * uiIndex ] -1142 * piSrc[ 21 * uiIndex ] + 1068 * piSrc[ 23 * uiIndex ] -789 * piSrc[25 * uiIndex ] + 359 * piSrc[ 27 * uiIndex ] + 140 * piSrc[ 29 * uiIndex ] -613 * piSrc[ 31 * uiIndex ] + 967 * piSrc[ 33 * uiIndex ] -1136 * piSrc[ 35 * uiIndex ] + 1087 * piSrc[ 37 * uiIndex ] -829 * piSrc[ 39 * uiIndex ] + 412 * piSrc[ 41 * uiIndex ] + 84 * piSrc[ 43 * uiIndex ] -564 * piSrc[ 45 * uiIndex ] + 936 * piSrc[ 47 * uiIndex ] -1128 * piSrc[ 49 * uiIndex ] + 1103 * piSrc[ 51 * uiIndex ] -867 * piSrc[ 53 * uiIndex ] + 464 * piSrc[ 55 * uiIndex ] + 28 * piSrc[ 57 * uiIndex ] -515 * piSrc[ 59 * uiIndex ] + 903 * piSrc[ 61 * uiIndex ] -1117 * piSrc[ 63 * uiIndex ];


      for( UInt uiFwdDataIdx = 27; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 9 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      
      iACContribution = 196 * piSrc[ 1 * uiIndex ] -564 * piSrc[ 3 * uiIndex ] + 867 * piSrc[ 5 * uiIndex ] -1068 * piSrc[ 7 * uiIndex ] + 1145 * piSrc[ 9 * uiIndex ] -1087 * piSrc[ 11 * uiIndex ] + 903 * piSrc[ 13 * uiIndex ] -613 * piSrc[ 15 * uiIndex ] + 251 * piSrc[ 17 * uiIndex ] + 140 * piSrc[ 19 * uiIndex ] -515 * piSrc[ 21 * uiIndex ] + 829 * piSrc[ 23 * uiIndex ] -1047 * piSrc[25 * uiIndex ] + 1142 * piSrc[ 27 * uiIndex ] -1103 * piSrc[ 29 * uiIndex ] + 936 * piSrc[ 31 * uiIndex ] -659 * piSrc[ 33 * uiIndex ] + 305 * piSrc[ 35 * uiIndex ] + 84 * piSrc[ 37 * uiIndex ] -464 * piSrc[ 39 * uiIndex ] + 789 * piSrc[ 41 * uiIndex ] -1023 * piSrc[ 43 * uiIndex ] + 1136 * piSrc[ 45 * uiIndex ] -1117 * piSrc[ 47 * uiIndex ] + 967 * piSrc[ 49 * uiIndex ] -704 * piSrc[ 51 * uiIndex ] + 359 * piSrc[ 53 * uiIndex ] + 28 * piSrc[ 55 * uiIndex ] -412 * piSrc[ 57 * uiIndex ] + 748 * piSrc[ 59 * uiIndex ] -996 * piSrc[ 61 * uiIndex ] + 1128 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 28; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 7 ] -= iACContribution;
        iACContribution = -iACContribution;
      }

      iACContribution = 140 * piSrc[ 1 * uiIndex ] -412 * piSrc[ 3 * uiIndex ] + 659 * piSrc[ 5 * uiIndex ] -867 * piSrc[ 7 * uiIndex ] + 1023 * piSrc[ 9 * uiIndex ] -1117 * piSrc[ 11 * uiIndex ] + 1145 * piSrc[ 13 * uiIndex ] -1103 * piSrc[ 15 * uiIndex ] + 996 * piSrc[ 17 * uiIndex ] -829 * piSrc[ 19 * uiIndex ] + 613 * piSrc[ 21 * uiIndex ] -359 * piSrc[ 23 * uiIndex ] + 84 * piSrc[ 25 * uiIndex ] + 196 * piSrc[ 27 * uiIndex ] -464 * piSrc[ 29 * uiIndex ] + 704 * piSrc[ 31 * uiIndex ] -903 * piSrc[ 33 * uiIndex ] + 1047 * piSrc[ 35 * uiIndex ] -1128 * piSrc[ 37 * uiIndex ] + 1142 * piSrc[ 39 * uiIndex ] -1087 * piSrc[ 41 * uiIndex ] + 967 * piSrc[ 43 * uiIndex ] -789 * piSrc[ 45 * uiIndex ] + 564 * piSrc[ 47 * uiIndex ] -305 * piSrc[ 49 * uiIndex ] + 28 * piSrc[ 51 * uiIndex ] + 251 * piSrc[ 53 * uiIndex ] -515 * piSrc[ 55 * uiIndex ] + 748 * piSrc[ 57 * uiIndex ] -936 * piSrc[ 59 * uiIndex ] + 1068 * piSrc[ 61 * uiIndex ] -1136 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 29; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 5 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 84 * piSrc[ 1 * uiIndex ] -251 * piSrc[ 3 * uiIndex ] + 412 * piSrc[ 5 * uiIndex ] -564 * piSrc[ 7 * uiIndex ] + 704 * piSrc[ 9 * uiIndex ] -829 * piSrc[ 11 * uiIndex ] + 936 * piSrc[ 13 * uiIndex ] -1023 * piSrc[ 15 * uiIndex ] + 1087 * piSrc[ 17 * uiIndex ] -1128 * piSrc[ 19 * uiIndex ] + 1145 * piSrc[ 21 * uiIndex ] -1136 * piSrc[ 23 * uiIndex ] + 1103 * piSrc[ 25 * uiIndex ] -1047 * piSrc[ 27 * uiIndex ] + 967 * piSrc[ 29 * uiIndex ] -867 * piSrc[ 31 * uiIndex ] + 748 * piSrc[ 33 * uiIndex ] -613 * piSrc[ 35 * uiIndex ] + 464 * piSrc[ 37 * uiIndex ] -305 * piSrc[ 39 * uiIndex ] + 140 * piSrc[ 41 * uiIndex ] + 28 * piSrc[ 43 * uiIndex ] -196 * piSrc[ 45 * uiIndex ] + 359 * piSrc[ 47 * uiIndex ] -515 * piSrc[ 49 * uiIndex ] + 659 * piSrc[ 51 * uiIndex ] -789 * piSrc[ 53 * uiIndex ] + 903 * piSrc[ 55 * uiIndex ] -996 * piSrc[ 57 * uiIndex ] + 1068 * piSrc[ 59 * uiIndex ] -1117 * piSrc[ 61 * uiIndex ] + 1142 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 30; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 3 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
      iACContribution = 28 * piSrc[ 1 * uiIndex ] -84 * piSrc[ 3 * uiIndex ] + 140 * piSrc[ 5 * uiIndex ] -196 * piSrc[ 7 * uiIndex ] + 251 * piSrc[ 9 * uiIndex ] -305 * piSrc[ 11 * uiIndex ] + 359 * piSrc[ 13 * uiIndex ] -412 * piSrc[ 15 * uiIndex ] + 464 * piSrc[ 17 * uiIndex ] -515 * piSrc[ 19 * uiIndex ] + 564 * piSrc[ 21 * uiIndex ] -613 * piSrc[ 23 * uiIndex ] + 659 * piSrc[ 25 * uiIndex ] -704 * piSrc[ 27 * uiIndex ] + 748 * piSrc[ 29 * uiIndex ] -789 * piSrc[ 31 * uiIndex ] + 829 * piSrc[ 33 * uiIndex ] -867 * piSrc[ 35 * uiIndex ] + 903 * piSrc[ 37 * uiIndex ] -936 * piSrc[ 39 * uiIndex ] + 967 * piSrc[ 41 * uiIndex ] -996 * piSrc[ 43 * uiIndex ] + 1023 * piSrc[ 45 * uiIndex ] -1047 * piSrc[ 47 * uiIndex ] + 1068 * piSrc[ 49 * uiIndex ] -1087 * piSrc[ 51 * uiIndex ] + 1103 * piSrc[ 53 * uiIndex ] -1117 * piSrc[ 55 * uiIndex ] + 1128 * piSrc[ 57 * uiIndex ] -1136 * piSrc[ 59 * uiIndex ] + 1142 * piSrc[ 61 * uiIndex ] -1145 * piSrc[ 63 * uiIndex ];

      for( UInt uiFwdDataIdx = 31; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += 64)
      {
        m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;
        m_aiAux2Array[ uiFwdDataIdx + 1 ] -= iACContribution;
        iACContribution = -iACContribution;
      }
    }

   if( 6 < uiLog2TSize )
   { 
    const UInt  uiCoeffAccessShift  = sm_uiLog2MaxTSize - uiLog2TSize;
    //==== AC coefficients =====
    for( UInt uiLog2X = uiLog2TSize-4; uiLog2X > 0; uiLog2X-- )
    {
      const UInt  uiDiffLog2        = uiLog2TSize - uiLog2X;
      const UInt  uiMaxMIdx         = 1 << uiDiffLog2;
      const UInt  uiDataStepSize    = 2 << uiDiffLog2;

      const UInt  uiStep            = 1 << uiLog2X;
      const UInt  uiStart           = uiStep >> 1;
      UInt        uiCoeffStart      = uiStart << uiCoeffAccessShift;
      const UInt  uiCoeffMAdd       = uiCoeffStart << 1;

      UInt  uiFwdDataIdxStart = 0;
      UInt  uiBwdOffset = uiDataStepSize - 1;
      for( UInt uiMIdx = 0; uiMIdx < uiMaxMIdx; uiMIdx++)
      {
//        std::cout << "int iACContribution = ";
        int   iACContribution = 0;
        {
          const UInt uiCoeffAdd      = uiCoeffStart << 1;
          for( UInt uiIdx = uiStart, uiCoeffIdx = uiCoeffStart; uiIdx < uiBlockSize; uiIdx += uiStep )
          {
//            std::cout << m_aiIDCTCoeffArray[ uiCoeffIdx & sm_uiCoeffAccessMask ] << " * piSrc[ " << uiIdx <<" * uiIndex ] + "; 
            iACContribution += piSrc[ uiIdx ] * m_aiIDCTCoeffArray[ uiCoeffIdx & sm_uiCoeffAccessMask ];
            uiCoeffIdx += uiCoeffAdd;
          }
        }

//        std::cout << ";" << std::endl << std::endl;
//        std::cout << " for( UInt uiFwdDataIdx = " << uiFwdDataIdxStart << "; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += " << uiDataStepSize << ") " << std::endl;

//        std::cout << " { " << std::endl;
//        std::cout << "   m_aiAux2Array[ uiFwdDataIdx ]     += iACContribution;" << std::endl;
//        std::cout << "   m_aiAux2Array[ uiFwdDataIdx + " << uiBwdOffset << " ] -= iACContribution;" << std::endl;
//        std::cout << "   iACContribution = -iACContribution;" << std::endl;
//        std::cout << " }" << std::endl;


        {
          for( UInt uiFwdDataIdx = uiFwdDataIdxStart; uiFwdDataIdx < uiBlockSize; uiFwdDataIdx += uiDataStepSize )
          {
            m_aiAux2Array[ uiFwdDataIdx ]               += iACContribution;
            m_aiAux2Array[ uiFwdDataIdx + uiBwdOffset ] -= iACContribution;
            iACContribution = -iACContribution;
          }
        }
        uiCoeffStart += uiCoeffMAdd, 
        uiFwdDataIdxStart += 1; 
        uiBwdOffset -= 2;
      }
    }
  }


  if( uiFinalShift )
  {
    //===== final shift =====
    const UInt uiAdd = 1 << (uiFinalShift - 1 );
    for( UInt uiIdx = 0, uiDataIdx = 0; uiIdx < uiBlockSize; uiIdx++, uiDataIdx += uiInc1 )
    {
      int iTemp = m_aiAux2Array[uiIdx];
      iTemp  += uiAdd;
      iTemp >>= uiFinalShift;
      paiData[ uiDataIdx ] = iTemp;
    }
  }
  else
  {
    for( UInt uiIdx = 0, uiDataIdx = 0; uiIdx < uiBlockSize; uiIdx++, uiDataIdx += uiInc1 )
    {
      paiData[ uiDataIdx ] = m_aiAux2Array[uiIdx];
    }
  }
}


template<UInt uiBlockSize, UInt uiLog2TSize > 
__inline void ResTrafoPow2::xFast2dIDCT( const Block<int>& rcIn, const Block<DPel>& rcOut, bool bAddDecRes ) 
{
  const UInt  uiStride    = rcIn.getStride();
  int*  paiTCoeff         = rcIn.getPtr();
  const int   iFullShift  = ( (int)sm_uiIDCTPrecision << 1 ) + (int)uiLog2TSize - EXTRA_PRECISION_BITS;
  const int   iShift1     = gMax( 0, iFullShift - 22 + EXTRA_PRECISION_BITS );
  const int   iShift2     = iFullShift - iShift1;
  AOF( rcIn .getSize().getX() == int( uiBlockSize ) );
  AOF( rcIn .getSize().getY() == int( uiBlockSize ) );
  //===== horizontal transform =====
  {
    int* paiData = paiTCoeff;
    for( UInt uiRow = 0; uiRow < uiBlockSize; uiRow++, paiData += uiStride )
    {
      xFast1dIDCT<uiBlockSize, uiLog2TSize, true>( paiData, 1, (UInt)iShift1 );
    }
  }
  //===== vertical transform =====
  {
    int* paiData = paiTCoeff;
    for( UInt uiCol = 0; uiCol < uiBlockSize; uiCol++, paiData++ )
    {
      xFast1dIDCT<uiBlockSize, uiLog2TSize, false>( paiData, uiStride, (UInt)iShift2 );
    }
  }
  //===== update data =====
  {
    const int*  paiData     = paiTCoeff;
    const UInt  uiOutStride = rcOut.getStride();
    DPel* paiOutData        = rcOut.getPtr();
    if( bAddDecRes )
    {
      for( UInt uiRow = 0; uiRow < UInt( rcOut.getSize().getY() ); uiRow++, paiOutData += uiOutStride, paiData += uiStride )
      {
        for( UInt uiCol = 0; uiCol < UInt( rcOut.getSize().getX() ); uiCol++ )
        {
          paiOutData[ uiCol ] += paiData[ uiCol ];
        }
      }
    }
    else
    {
      for( UInt uiRow = 0; uiRow < UInt( rcOut.getSize().getY() ); uiRow++, paiOutData += uiOutStride, paiData += uiStride )
      {
        for( UInt uiCol = 0; uiCol < UInt( rcOut.getSize().getX() ); uiCol++ )
        {
          paiOutData[ uiCol ]  = paiData[ uiCol ];
        }
      }
    }
  }
}

ErrVal
ResTrafoPow2::xFast2dIDCT( const Block<int>& rcIn, const Block<DPel>& rcOut, bool bAddDecRes )
{
  const int iBlkSize = rcIn.getSize().getX();
  if( 2 == iBlkSize )
  {
    xFast2dIDCT<2,1>( rcIn, rcOut, bAddDecRes );
  }
  else if( 4 == iBlkSize )
  {
    xFast2dIDCT<4,2>( rcIn, rcOut, bAddDecRes );
  }
  else if( 8 == iBlkSize )
  {
    xFast2dIDCT<8,3>( rcIn, rcOut, bAddDecRes );
  }
  else if( 16 == iBlkSize )
  {
    xFast2dIDCT<16,4>( rcIn, rcOut, bAddDecRes );
  }
  else if( 32 == iBlkSize )
  {
    xFast2dIDCT<32,5>( rcIn, rcOut, bAddDecRes );
  }
  else if( 64 == iBlkSize )
  {
    xFast2dIDCT<64,6>( rcIn, rcOut, bAddDecRes );
  }
/*
  else if( 128 == iBlkSize )
  {
    xFast2dIDCT<128,7>( rcIn, rcOut, bAddDecRes );
  }
*/
  else
  {
    ROT(1);
  }
  return Err::m_nOK;
}

ErrVal
ResTrafoPow2::xFastIScale( int iQp, const Block<const int> &rcIn, const Block<int> &rcOut )  
{
  ROT( iQp < 0 );
  const UInt  uiNumCols   = rcIn .getSize().getX();
  const UInt  uiNumRows   = rcIn .getSize().getY();
  const UInt  uiInStride  = rcIn .getStride();
  const UInt  uiOutStride = rcOut.getStride();
  const int*  paiInData   = rcIn .getPtr();
  int*        paiOutData  = rcOut.getPtr();
  const int   iQStep      = sm_aiQStep[ iQp % 6 ] << ( iQp / 6 );

  for( UInt uiRow = 0; uiRow < uiNumRows; uiRow++, paiInData += uiInStride, paiOutData += uiOutStride )
  {
    for( UInt uiCol = 0; uiCol < uiNumCols; uiCol++ )
    {
      paiOutData[ uiCol ] = paiInData[ uiCol ] * iQStep;
    }
  }
  return Err::m_nOK;
}



ErrVal
ResTrafoPow2::setScanPattern( std::vector<UInt>& rauiScanPattern,
                             const UInt         uiLog2BlockSize,
                             const bool         bDownLeft )
{
  const int   iBlockSize    = 1 << uiLog2BlockSize;
  const UInt  uiNumScanPos  = UInt( iBlockSize * iBlockSize );
  UInt        uiNextScanPos = 0;
  rauiScanPattern.resize( uiNumScanPos );

  for( UInt uiScanLine = 0; uiNextScanPos < uiNumScanPos; uiScanLine++ )
  {
    int    iPrimDim  = int( uiScanLine );
    int    iScndDim  = 0;
    while( iPrimDim >= iBlockSize )
    {
      iScndDim++;
      iPrimDim--;
    }
    while( iPrimDim >= 0 && iScndDim < iBlockSize )
    {
      rauiScanPattern[ uiNextScanPos++ ] = ( bDownLeft ? iScndDim * iBlockSize + iPrimDim : iPrimDim * iBlockSize + iScndDim );
      iScndDim++;
      iPrimDim--;
    }
  }
  return Err::m_nOK;
}

UInt 
ResTrafoPow2::getSigCtxInc( const Block<const int>&  rcBlock,
                            const UInt               uiPosX,
                            const UInt               uiPosY,
                            const UInt               uiLog2BlkSize,
                            const bool               bDownLeft )
{
  UInt  uiCtxInc  = 0;
  UInt  uiSizeM1  = ( 1 << uiLog2BlkSize ) - 1;
  if( uiLog2BlkSize <= 3 )
  {
    UInt  uiShift = gMax<int>( 0, uiLog2BlkSize - 2 );
    uiCtxInc      = ( ( uiPosY >> uiShift ) << 2 ) + ( uiPosX >> uiShift );
  }
  else if( uiPosX <= 1 && uiPosY <= 1 )
  {
    uiCtxInc            = ( uiPosY << 1 ) + uiPosX;
  }
  else if( uiPosY == 0 )
  {
    const int*  pData   = &rcBlock.getEl    ( uiPosX, uiPosY );
    int         iStride =  rcBlock.getStride();
    UInt        uiCnt   = ( pData[         -1 ] ? 1 : 0 );
    uiCnt              += ( pData[         -2 ] ? 1 : 0 );
    uiCnt              += ( pData[ iStride -2 ] ? 1 : 0 );
    if( ! bDownLeft )
    {
      uiCnt            += ( pData[ iStride -1 ] ? 1 : 0 );
    }
    uiCtxInc            = 4 + ( ( uiCnt + 1 ) >> 1 );
  }
  else if( uiPosX == 0 )
  {
    const int*  pData   = &rcBlock.getEl    ( uiPosX, uiPosY );
    int         iStride =  rcBlock.getStride();
    int         iStride2=  iStride << 1;
    UInt        uiCnt   = ( pData[  -iStride  ] ? 1 : 0 );
    uiCnt              += ( pData[  -iStride2 ] ? 1 : 0 );
    uiCnt              += ( pData[ 1-iStride2 ] ? 1 : 0 );
    if( bDownLeft )
    {
      uiCnt            += ( pData[ 1-iStride  ] ? 1 : 0 );
    }
    uiCtxInc            = 7 + ( ( uiCnt + 1 ) >> 1 );
  }
  else
  {
    const int*  pData   = &rcBlock.getEl    ( uiPosX, uiPosY );
    int         iStride =  rcBlock.getStride();
    int         iStride2=  iStride << 1;
    UInt        uiCnt   = ( pData[    -iStride  ] ? 1 : 0 );
    uiCnt              += ( pData[ -1           ] ? 1 : 0 );
    uiCnt              += ( pData[ -1 -iStride  ] ? 1 : 0 );
    if( uiPosX > 1 )
    {
      uiCnt            += ( pData[ -2           ] ? 1 : 0 );
      uiCnt            += ( pData[ -2 -iStride  ] ? 1 : 0 );
      if( uiPosY < uiSizeM1 )
      {
        uiCnt          += ( pData[ -2 +iStride  ] ? 1 : 0 );
      }
    }
    if( uiPosY > 1 )
    {
      uiCnt            += ( pData[    -iStride2 ] ? 1 : 0 );
      uiCnt            += ( pData[ -1 -iStride2 ] ? 1 : 0 );
      if( uiPosX < uiSizeM1 )
      {
        uiCnt          += ( pData[  1 -iStride2 ] ? 1 : 0 );
      }
    }
    if( bDownLeft )
    {
      if( uiPosX < uiSizeM1 )
      {
        uiCnt          += ( pData[  1 -iStride  ] ? 1 : 0 );
      }
    }
    else
    {
      if( uiPosY < uiSizeM1 )
      {
        uiCnt          += ( pData[ -1 +iStride  ] ? 1 : 0 );
      }
    }
    uiCtxInc      = 10 + gMin<UInt>( 4, ( uiCnt + 1 ) >> 1 );
  }
  return uiCtxInc;
}

UInt 
ResTrafoPow2::getLastCtxInc( const Block<const int>&  rcBlock,
                             const UInt               uiPosX,
                             const UInt               uiPosY,
                             const UInt               uiLog2BlkSize )
{
  UInt uiCtxInc  = 0;
  if( uiLog2BlkSize <= 2 )
  {
    UInt  uiShift = gMax<int>( 0, uiLog2BlkSize - 2 );
    uiCtxInc      = ( ( uiPosY >> uiShift ) << 2 ) + ( uiPosX >> uiShift );
  }
  else
  {
    uiCtxInc      = ( uiPosX + uiPosY ) >> ( uiLog2BlkSize - 3 );
  }
  return uiCtxInc;
}


VIDEO_NAMESPACE_END
