/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/VLCodingTables.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


VLCTreeNode::VLCTreeNode( VLCTreeNode* pcRootNode, VLCTreeNode* pcParentNode, UInt& ruiNextNodeIdx )
: m_uiNodeIdx           ( ruiNextNodeIdx++ )
, m_uiCodeIdx           ( TypeLimits<UInt>::getMax() )
, m_uiCodeLength        ( TypeLimits<UInt>::getMax() )
, m_pcRoot              ( pcRootNode )
, m_pcParent            ( pcParentNode )
, m_pcChild0            ( 0 )
, m_pcChild1            ( 0 )
, m_pcMinLengthLeafNode ( 0 )
{
}

VLCTreeNode::~VLCTreeNode()
{
  delete  m_pcChild0;
  delete  m_pcChild1;
}

ErrVal
VLCTreeNode::xCheckInitTreeRoot( const UInt uiNumCodewords, UInt& ruiNumNodes )
{
  UInt  uiMaxNodeIdx    = 0;
  UInt  uiMaxCodeIdx    = 0;
  ruiNumNodes           = 0;
  RNOK( xCheckInitTreeNode( this, 0, uiMaxNodeIdx, uiMaxCodeIdx, ruiNumNodes ) );
  ROF ( uiNumCodewords == uiMaxCodeIdx + 1 );
  ROF ( ruiNumNodes    == uiMaxNodeIdx + 1 );
  return Err::m_nOK;
}

ErrVal
VLCTreeNode::xCheckInitTreeNode( const VLCTreeNode* pcRootNode, const VLCTreeNode* pcParentNode, UInt& ruiMaxNodeIdx, UInt& ruiMaxCodeIdx, UInt& ruiNumNodes )
{
  ROF( m_pcRoot   == pcRootNode   );
  ROF( m_pcParent == pcParentNode );
  if ( m_pcChild0 == 0 || m_pcChild1 == 0 )
  { // leaf node 
    ROF ( m_uiCodeIdx     != TypeLimits<UInt>::getMax() );
    ROF ( m_uiCodeLength  != TypeLimits<UInt>::getMax() );
    ruiMaxCodeIdx = gMax<UInt>( ruiMaxCodeIdx, m_uiCodeIdx );
  }
  else
  { // internal node
    ROF ( m_uiCodeIdx     == TypeLimits<UInt>::getMax() );
    ROF ( m_uiCodeLength  == TypeLimits<UInt>::getMax() );
    ROF ( m_pcChild0 );
    ROF ( m_pcChild1 );
    RNOK( m_pcChild0->xCheckInitTreeNode( pcRootNode, this, ruiMaxNodeIdx, ruiMaxCodeIdx, ruiNumNodes ) );
    RNOK( m_pcChild1->xCheckInitTreeNode( pcRootNode, this, ruiMaxNodeIdx, ruiMaxCodeIdx, ruiNumNodes ) );
  }
  ruiMaxNodeIdx = gMax<UInt>( ruiMaxNodeIdx, m_uiNodeIdx );
  ruiNumNodes++;
  return Err::m_nOK;
}


ErrVal
VLCTreeNode::xAddMinLengthNodes()
{
  ROF( this == m_pcRoot );
  for( VLCTreeNode* pcNextNode = this; pcNextNode; pcNextNode = pcNextNode->xGetNextNode( this ) )
  {
    RNOK( pcNextNode->xAddMinLengthNode() );
  }
  return Err::m_nOK;
}


ErrVal
VLCTreeNode::xAddMinLengthNode()
{
  ROT( m_pcMinLengthLeafNode );
  if ( m_pcChild0 == 0 )
  { // leaf node
    m_pcMinLengthLeafNode = m_pcRoot;
    return Err::m_nOK;
  }
  UInt uiMinCodeLength = TypeLimits<UInt>::getMax();
  for( VLCTreeNode* pcNextLeafNode = xGetNextLeafNode( this ); pcNextLeafNode; pcNextLeafNode = pcNextLeafNode->xGetNextLeafNode( this ) )
  {
    if( pcNextLeafNode->m_uiCodeLength < uiMinCodeLength )
    {
      uiMinCodeLength       = pcNextLeafNode->m_uiCodeLength;
      m_pcMinLengthLeafNode = pcNextLeafNode;
    }
  }
  ROF( m_pcMinLengthLeafNode );
  return Err::m_nOK;
}


const VLCTreeNode*
VLCTreeNode::xGetNextLeafNode( const VLCTreeNode* pcSubTreeRoot ) const
{
  for( const VLCTreeNode* pcNextNode = xGetNextNode( pcSubTreeRoot ); pcNextNode; pcNextNode = pcNextNode->xGetNextNode( pcSubTreeRoot ) )
  {
    ROTRS( pcNextNode->m_pcChild0 == 0, pcNextNode );
  }
  return 0;
}
VLCTreeNode*
VLCTreeNode::xGetNextLeafNode( const VLCTreeNode* pcSubTreeRoot )
{
  for( VLCTreeNode* pcNextNode = xGetNextNode( pcSubTreeRoot ); pcNextNode; pcNextNode = pcNextNode->xGetNextNode( pcSubTreeRoot ) )
  {
    ROTRS( pcNextNode->m_pcChild0 == 0, pcNextNode );
  }
  return 0;
}

const VLCTreeNode*
VLCTreeNode::xGetNextNode( const VLCTreeNode* pcSubTreeRoot ) const
{
  ROTRS( m_pcChild0,            m_pcChild0  );
  ROTRS( this == pcSubTreeRoot, 0 );
  return m_pcParent->xGetNextChild1Node( pcSubTreeRoot, this );
}
VLCTreeNode*
VLCTreeNode::xGetNextNode( const VLCTreeNode* pcSubTreeRoot )
{
  ROTRS( m_pcChild0,            m_pcChild0  );
  ROTRS( this == pcSubTreeRoot, 0 );
  return m_pcParent->xGetNextChild1Node( pcSubTreeRoot, this );
}

const VLCTreeNode*
VLCTreeNode::xGetNextChild1Node( const VLCTreeNode* pcSubTreeRoot, const VLCTreeNode* pcPrevNode ) const
{
  ROTRS( m_pcChild1 != pcPrevNode,    m_pcChild1 );
  ROTRS( this       == pcSubTreeRoot, 0 );
  return m_pcParent->xGetNextChild1Node( pcSubTreeRoot, this );
}
VLCTreeNode*
VLCTreeNode::xGetNextChild1Node( const VLCTreeNode* pcSubTreeRoot, const VLCTreeNode* pcPrevNode )
{
  ROTRS( m_pcChild1 != pcPrevNode,    m_pcChild1 );
  ROTRS( this       == pcSubTreeRoot, 0 );
  return m_pcParent->xGetNextChild1Node( pcSubTreeRoot, this );
}


// video namespace end
VIDEO_NAMESPACE_END

