/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoDecoderLibPch.h"
#include "VideoCodecToolsLib/MSysLargeFile.h"
#include "VideoCommonLib/CabacMultiCoder.h"

#if COLLECT_CABAC_BITS
#include "VideoCommonLib/ExtendedContextModels.h"
#endif

#include "SyntaxElementDecoder.h"

#include "AccessUnitDecoder.h"

#include "VideoCommonLib/ContextTables.h"
#include "VideoCommonLib/ExtendedCabacTables.h"
#include "VideoCommonLib/BitWriter.h"

#include "SliceParser.h"
#include "SliceConverter.h"
#include "SliceDecoder.h"

#include "ProbabilityAllocatorDec.h"
#include "VideoCommonLib/TraceCabac.h"
#include "VideoCommonLib/ModePictureGenerator.h"
#include "VideoCommonLib/CodingState.h"
#include "VideoCommonLib/DeblockingFilter.h"
#include "WienerFilterDecoder.h"
#include "SliceStream.h"

#include "VideoCodecToolsLib/PacketParser.h"

#include "VideoCommonLib/H264SequenceParameterSet.h"
#include "VideoCommonLib/H264PictureParameterSet.h"
#include "VideoCommonLib/PocAndFrameNum.h"
#include "H264NalUnitParser.h"

#if RANDOM_SLICE_PARSING
#include "time.h"
#endif

#include "WienerParser.h"



// video namespace begin
VIDEO_NAMESPACE_BEGIN


ErrVal AccessUnitDecoder::parseAccessUnit( const std::vector<UChar> &raucBitstream,
                                           const CodedPictureStoreBase &rcCodedPictureStore,
                                           AccessUnit *pcAU,
                                           TraceCabacVecIf *pcCabacTraceVecIf,
                                           PicOrderCnt &rcPoc,
                                           TraceCommon *pcTrace )
{
  BitReader<UChar, UInt> cHeaderReader( true, raucBitstream.begin(), raucBitstream.end() );
  UInt uiPkgSize = UInt( cHeaderReader.getNumRemainingBits() / 8 );

  RNOK( pcAU->createPlanePackages() );

  rcPoc = pcAU->getAUPars().getPicOrderCnt();
  UInt uiHeaderSize = UInt( uiPkgSize - UInt( cHeaderReader.getNumRemainingBits() / 8 ) );

  if( pcAU->getAUPars().getWienerFilterParameters().getUseFilterFlag() )
  {
    BitReader<UChar, UInt> cWFBitReader( true, raucBitstream.begin() + uiHeaderSize, raucBitstream.end() );
    ProbBinStreamDecIf*     pcProbBinStreamDecIf  = 0;
    CabaCoderStartFinishIf* pcCoderStartFinishIf  = 0;
    if( pcAU->getAUPars().getVLCFlag() )
    {
      VLCDecoder<VLCodeType>* pDec  = new VLCDecoder<VLCodeType>( cWFBitReader, pcAU->getAUPars().getVLCMaxDelayInBits() );
      pcProbBinStreamDecIf          = pDec;
      pcCoderStartFinishIf          = pDec;
    }
    else
    {
      CabaDecoder<>* pDec = new CabaDecoder<>( cWFBitReader );
      pcProbBinStreamDecIf  = pDec;
      pcCoderStartFinishIf  = pDec;
    }
    RNOK( pcCoderStartFinishIf->startCabac() );
    if( pcTrace )
    {
      pcTrace->resetCount();
    }
    RNOK( xParseWienerParameters( *pcProbBinStreamDecIf, pcAU->getAUPars(), pcCabacTraceVecIf, TRACE_HEADERS ? pcTrace : NULL ) );
    RNOK( pcCoderStartFinishIf->finishCabac() );
    delete pcCoderStartFinishIf;

    const UInt uiNumBits = cWFBitReader.getBitsUntilByteAligned();
    if( 0 != uiNumBits )
    {
      UInt uiDummy;
      cWFBitReader.read( uiDummy, uiNumBits );
    }

    uiHeaderSize = UInt( uiPkgSize - UInt( cWFBitReader.getNumRemainingBits() / 8 ) );
  }

  if( pcTrace )
  {
    pcTrace->resetCount();
  }

  if ( pcAU->getAUPars().getEntropySliceFlag() )
  {
    UInt uiSlices = 0;
    SliceStream cSliceStream( &raucBitstream[uiHeaderSize] );
    PacketParser<SliceStream> cPacketParser( cSliceStream );
    std::vector< std::vector<UChar> > acBitstreams;

    for ( UInt uiPlaPaId = 0; uiPlaPaId < pcAU->getAUPars().getNumPlanePacks(); uiPlaPaId++ )
    {
      uiSlices += pcAU->getPlanePackage( uiPlaPaId )->getCurrentSlice().getQuadtreeGroupStructure().getNumEntropySlices();
    }

    for ( UInt ui = 0; ui < uiSlices; ui++ )
    {
      RNOK( cPacketParser.readNextPacket() );
      acBitstreams.push_back( cPacketParser.getVector() );
    }

    RNOK( xParsePlanePackages( acBitstreams, rcCodedPictureStore, *pcAU, pcCabacTraceVecIf, pcTrace ) );
  }
  else
  {
    BitReader<UChar, UInt> cBitReader( true, raucBitstream.begin() + uiHeaderSize, raucBitstream.end() );

    CabaCoderStartFinishIf *pcCabaDecoder;
    ProbBinStreamDecIf *pcProbBinStreamDec;

    std::vector<std::list<std::vector<UChar> > > aaaucBins;
    std::list<std::vector<UChar>::const_iterator> aacIterEpStart, aacIterEpEnd;
    MultiVectorBitReader<UChar, UInt> cEqProbReader;

    const ExtendedCabacTables eTables = pcAU->getAUPars().getExtendedCabacTables();

    if( pcAU->getAUPars().getVLCFlag() )
    {
      VLCDecoder<VLCodeType> *pVLCDecoder = new VLCDecoder<VLCodeType>( cBitReader, pcAU->getAUPars().getVLCMaxDelayInBits() );
      RNOK( pVLCDecoder->startCabac() );
      pcCabaDecoder       = pVLCDecoder;
      pcProbBinStreamDec  = pVLCDecoder;
    }
    else if( eTables == AGGREGATE_CABAC_TABLES12 )
    {
      RNOK( xReadMultiplex<AggregateCabacTables12>( cBitReader, raucBitstream.end(), aaaucBins, aacIterEpStart, aacIterEpEnd, pcAU->getAUPars().getPicOrderCnt() ) );
      MultiVectorBitReader<UChar, UInt>::InitPars cEPReaderPars( aacIterEpStart, aacIterEpEnd );
      CabacMultiCoder<CabacPassThrough, BitReaderIf<UInt>, AggregateCabacTables12, false> *pcCabaMultiDecoder;
      pcCabaMultiDecoder = new CabacMultiCoder<CabacPassThrough, BitReaderIf<UInt>, AggregateCabacTables12, false>( pcAU->getAUPars().getPicOrderCnt() );
      RNOK( cEqProbReader.init( cEPReaderPars ) );
      pcCabaMultiDecoder->setEqProbBitStream( &cEqProbReader );
      for( UInt ui = 1; ui < AggregateCabacTables12::eNumAggregateStates; ++ui )
      {
        MultiVectorBitReader<UChar, UInt>::InitPars cPars( aaaucBins[ui].begin(), aaaucBins[ui].end() );
        RNOK( pcCabaMultiDecoder->addCabacCoder( cPars ) );
      }
      pcCabaDecoder = pcCabaMultiDecoder;
      pcProbBinStreamDec = pcCabaMultiDecoder;
    }
    else
    {
      CabaDecoder<> *p = new CabaDecoder<>( cBitReader );
      p->startCabac();
      pcCabaDecoder = p;
      pcProbBinStreamDec = p;
    }

    for( UInt uiRuns = 0; uiRuns < pcAU->getAUPars().getNumPlanePacks(); uiRuns++ )
    {
      RNOK( xParsePlanePackage( uiRuns,
                                rcCodedPictureStore,
                                *pcProbBinStreamDec,
                                *pcAU,
                                pcCabacTraceVecIf,
                                pcTrace ) );
    }
    RNOK( pcCabaDecoder->finishCabac() );
    delete pcCabaDecoder;
  }

  return Err::m_nOK;
}


ErrVal AccessUnitDecoder::decodeAccessUnit( const PicBlock<Pel> &rcDec,
                                            const std::vector<PicBlock<Pel> > &racOverlaidDec,
                                            const CodedPictureStore &rcCodedPictureStore,
                                            const AccessUnit *pcAU )
{
  for( UInt uiRuns = 0; uiRuns < pcAU->getAUPars().getNumPlanePacks(); uiRuns++ )
  {
    PicBlock<Pel> cDecPlanePackage;

    RNOK( pcAU->extractPlanePackagePicBlock( uiRuns, rcDec, cDecPlanePackage ) );
    RNOK( decodePlanePackage( uiRuns,
                              cDecPlanePackage,
                              CodingState( rcCodedPictureStore.getPlanePackageVec<ReferencePlanePackage<Pel> >( uiRuns, pcAU->getAUPars() ), rcCodedPictureStore.getRefCtxStore( uiRuns, pcAU->getAUPars() ) ),
                              *pcAU ) );

    if( pcAU->getAUPars().getDeblockingIdc() == 0 )
    {
      std::vector<bool> abIsLuma( cDecPlanePackage.getStats().getNumPlanes(), false );
      if( uiRuns == 0 )
      {
        abIsLuma[0] = true;
      }
      UInt uiLog2MinBlkSize = ( abIsLuma[0] ? 2 : 1 ); // allow 2x2 for chroma plane packages (req. for inter plane prediction)
      RNOK( DeblockingFilter::deblockPlanePackage<Pel>( *pcAU->getPlanePackage( uiRuns ),
                                                        cDecPlanePackage,
                                                        abIsLuma,
                                                        gMin<int>( uiLog2MinBlkSize, pcAU->getAUPars().getPlaPaPars( uiRuns ).getLog2MinPow2TrafoBlkSize() ),
                                                        2 * pcAU->getAUPars().getDeblockAlphaOffs(),
                                                        2 * pcAU->getAUPars().getDeblockBetaOffs() ) );
    }

    if( !racOverlaidDec.empty() )
    {
      ModePictureGeneratorParams cMPGParams( 128, false, true, 0 );
      ModePictureGenerator cModePicGen( cMPGParams, pcAU->getPlanePackage( uiRuns )->getSlice( 0 ) );
      cModePicGen.generatePicture( rcDec.getPlane( 0 ), racOverlaidDec[uiRuns], pcAU->getPlanePackage( uiRuns )->getPars().getPrimPlaneSubsampling() );
    }
  }

  WienerFilterDecoder::applyFilterToPicBlockDec( pcAU->getAUPars(), rcDec);

  return Err::m_nOK;
}


ErrVal AccessUnitDecoder::xParsePlanePackage( UInt uiPlaPaId,
                                              const CodedPictureStoreBase &rcCodedPictureStore,
                                              ProbBinStreamDecIf &rcProbBinInput,
                                              AccessUnit &rcAU,
                                              TraceCabacVecIf *pcCabacTraceVecIf,
                                              TraceCommon *pcTrace )
{
  const AccessUnitParameters& rcAUPars    =  rcAU.getAUPars          ();
  PlanePackage&               rcPlaPaCurr = *rcAU.getPlanePackage    ( uiPlaPaId );
  const PlanePackage*         pcRefPlaPa  =  rcAU.getRefPlanePackage ( uiPlaPaId );

#if COLLECT_CABAC_BITS
  typedef BitCounterContextModel<CabacContextModel> ContextModelCurr;
#else
  typedef CabacContextModel ContextModelCurr;
#endif

  ProbabilityAllocatorDec<ContextModelCurr> cProbAllocDec( rcProbBinInput, UInt( rcAU.getCtxStore( uiPlaPaId ).getCtxInit().getCtxInit().size() ), TRACE_PROB_BINS ? pcTrace : NULL );
  RNOK( cProbAllocDec.initCtx( rcPlaPaCurr.getCurrentSlice().getSliceQps()[0], rcAU.getCtxStore( uiPlaPaId ).getCtxInit().getCtxInit() ) );
  BinStreamProcessor<CtxBin&, false, EPCtxBin&>::CtorPars cDefaultBinProcPars;
  SyntaxElementDecoder<> cSEDecoder( rcAU.getCtxStore( uiPlaPaId ).getCtxInit().getCtxAccess(), rcAU.getCtxStore( uiPlaPaId ).getCtxIndices(), rcPlaPaCurr.getPars(), pcTrace, cDefaultBinProcPars );
  cSEDecoder.InStreamProcessor<CtxBin&>::setInStreamIf( cProbAllocDec.getInStreamIf() );
  cSEDecoder.InStreamProcessor<EPCtxBin&>::setInStreamIf( cProbAllocDec.getInStreamIf() );

  SliceParserAndConverter cSliceParserAndConverter( 0, rcPlaPaCurr.getCurrentSlice(), rcPlaPaCurr.getPars().getSubsampling(), &cSEDecoder, rcPlaPaCurr.getPars(), rcAUPars, pcRefPlaPa, pcTrace );
  RNOK( cSliceParserAndConverter.processTree() );

#if TRACK_CABAC_BITS
  if( pcCabacTraceVecIf )
  {
    TraceCabacCategorizer cCabCat;
    cCabCat.setAUPars( rcAUPars );

    RNOK( pcCabacTraceVecIf->getTraceCabac( uiPlaPaId == 0 ? 0 : 5 + uiPlaPaId ).storeCabacBits( cCabCat, "unchanged", rcAU.getCtxStore( uiPlaPaId ).getCtxInit().getCtxAccess(), cProbAllocDec ) );
  }
#endif

  return Err::m_nOK;
}


ErrVal AccessUnitDecoder::xParsePlanePackages( const std::vector< std::vector<UChar> > &rcBitstream,
                                               const CodedPictureStoreBase &rcCodedPictureStore,
                                               AccessUnit &rcAU,
                                               TraceCabacVecIf *pcCabacTraceVecIf,
                                               TraceCommon *pcTrace )
{
  UInt uiPgkIdx = 0;

  for ( UInt uiPlaPaId = 0; uiPlaPaId < rcAU.getAUPars().getNumPlanePacks(); uiPlaPaId++ )
  {
    const AccessUnitParameters& rcAUPars    =  rcAU.getAUPars          ();
    PlanePackage&               rcPlaPaCurr = *rcAU.getPlanePackage    ( uiPlaPaId );
    const PlanePackage*         pcRefPlaPa  =  rcAU.getRefPlanePackage ( uiPlaPaId );

#if COLLECT_CABAC_BITS
    typedef BitCounterContextModel<CabacContextModel> ContextModelCurr;
#else
    typedef CabacContextModel ContextModelCurr;
#endif

    ContextStore<CabacContextModel> *pcRefCtxStore = rcAU.getAUPars().getCtxStatesReuseFlag() ? rcCodedPictureStore.getRefCtxStore( uiPlaPaId, rcAU.getAUPars() ) : NULL;

#if RANDOM_SLICE_PARSING
    std::vector<UInt> acSlices;
    for ( UInt uiSlice = 0; uiSlice < rcPlaPaCurr.getCurrentSlice().getQuadtreeGroupStructure().getNumEntropySlices(); uiSlice++ )
      acSlices.push_back( uiSlice );

    srand( (UInt) time( NULL ) );
    std::random_shuffle( acSlices.begin(), acSlices.end() );

    for ( UInt ui = 0; ui < rcPlaPaCurr.getCurrentSlice().getQuadtreeGroupStructure().getNumEntropySlices(); ui++ )
    {
      UInt uiSlice = acSlices[ui];
#else
    for ( UInt uiSlice = 0; uiSlice < rcPlaPaCurr.getCurrentSlice().getQuadtreeGroupStructure().getNumEntropySlices(); uiSlice++ )
    {
#endif
      CabacContextModel *pacCtxStates;
      BitReader<UChar, UInt> cBitReader( true, rcBitstream[uiSlice + uiPgkIdx].begin(), rcBitstream[uiSlice + uiPgkIdx].end() );
      ProbBinStreamDecIf*     pcProbBinStreamDecIf  = 0;
      CabaCoderStartFinishIf* pcCoderStartFinishIf  = 0;
      if( rcAU.getAUPars().getVLCFlag() )
      {
        VLCDecoder<VLCodeType>* pDec  = new VLCDecoder<VLCodeType>( cBitReader, rcAU.getAUPars().getVLCMaxDelayInBits() );
        pcProbBinStreamDecIf          = pDec;
        pcCoderStartFinishIf          = pDec;
      }
      else
      {
        CabaDecoder<>* pDec = new CabaDecoder<>( cBitReader );
        pcProbBinStreamDecIf  = pDec;
        pcCoderStartFinishIf  = pDec;
      }
      RNOK( pcCoderStartFinishIf->startCabac() );

      ProbabilityAllocatorDec<ContextModelCurr> cProbAllocDec( *pcProbBinStreamDecIf, UInt( rcAU.getCtxStore( uiPlaPaId ).getCtxInit().getCtxInit().size() ), TRACE_PROB_BINS ? pcTrace : NULL );
      cProbAllocDec.initCtx( rcPlaPaCurr.getCurrentSlice().getSliceQps()[0], rcAU.getCtxStore( uiPlaPaId ).getCtxInit().getCtxInit() );
      BinStreamProcessor<CtxBin&, false, EPCtxBin&>::CtorPars cDefaultBinProcPars;
      SyntaxElementDecoder<> cSEDecoder( rcAU.getCtxStore( uiPlaPaId ).getCtxInit().getCtxAccess(), rcAU.getCtxStore( uiPlaPaId ).getCtxIndices(), rcPlaPaCurr.getPars(), pcTrace, cDefaultBinProcPars );
      cSEDecoder.InStreamProcessor<CtxBin&>::setInStreamIf( cProbAllocDec.getInStreamIf() );
      cSEDecoder.InStreamProcessor<EPCtxBin&>::setInStreamIf( cProbAllocDec.getInStreamIf() );

      if ( pcRefCtxStore )
      {
        RNOK( pcRefCtxStore->getCtxStates( uiSlice, pacCtxStates ) );
#if COLLECT_CABAC_BITS
        TOT( 1 ); // currently not supported
#else
        RNOK( cProbAllocDec.setCtxStates( pacCtxStates ) );
#endif
      }

      SliceParser cSliceParser( uiSlice, rcPlaPaCurr.getCurrentSlice(), rcPlaPaCurr.getPars().getSubsampling(), &cSEDecoder, rcPlaPaCurr.getPars(), rcAUPars, pcTrace );
      RNOK( cSliceParser.processTree() );
      RNOK( pcCoderStartFinishIf->finishCabac() );

#if COLLECT_CABAC_BITS
      TOT( 1 ); // currently not supported
#else
      RNOK( cProbAllocDec.getCtxStates( pacCtxStates ) );
#endif
      RNOK( rcAU.getCtxStore( uiPlaPaId ).setCtxStates( uiSlice, pacCtxStates ) );
    }

    SliceConverter cSliceConverter( rcPlaPaCurr.getPars(), rcAUPars, pcRefPlaPa, rcPlaPaCurr.getCurrentSlice() );
    RNOK( cSliceConverter.processTree() );

    uiPgkIdx += rcPlaPaCurr.getCurrentSlice().getQuadtreeGroupStructure().getNumEntropySlices();

#if TRACK_CABAC_BITS
    if( pcCabacTraceVecIf )
    {
      TraceCabacCategorizer cCabCat;
      cCabCat.setAUPars( rcAUPars );
      TOT( 1 );
      //RNOK( pcCabacTraceVecIf->getTraceCabac( uiPlaPaId == 0 ? 0 : 5 + uiPlaPaId ).storeCabacBits( cCabCat, "unchanged", cCtxInit.getCtxAccess(), cProbAllocDec ) );
    }
#endif
  }

  return Err::m_nOK;
}


ErrVal AccessUnitDecoder::xParseWienerParameters( ProbBinStreamDecIf& rcProbBinStreamDecIf,
                                                  AccessUnitParameters &rcAUPars,
                                                  TraceCabacVecIf *pcCabacTraceVecIf,
                                                  TraceCommon *pcTrace )
{
  ContextArrayStartIndices cCtxIndices ;
  cCtxIndices.m_uiStartRefIdx = 0;

  CtxInitArray cCtxInit ;
  cCtxInit = ContextTables::m_cCtxWiener;
//  cCtxIndices.m_uiStartRefIdx = cCtxInit.getCtxAccess().getNumCtxInitArrays();


#if COLLECT_CABAC_BITS
  typedef BitCounterContextModel<CabacContextModel> ContextModelCurr;
#else
  typedef CabacContextModel ContextModelCurr;
#endif

  ProbabilityAllocatorDec<> cProbAllocDec( rcProbBinStreamDecIf, UInt( cCtxInit.getCtxInit().size() ), TRACE_PROB_BINS ? pcTrace : NULL );

  cProbAllocDec.initCtx( rcAUPars.getPlaPaPars(0).getPlaneQp(0), cCtxInit.getCtxInit() );

  BinStreamProcessor<CtxBin&, false, EPCtxBin&>::CtorPars cDefaultBinProcPars;
  SyntaxElementDecoder<> cSEDecoder( cCtxInit.getCtxAccess(), cCtxIndices, rcAUPars.getPlaPaPars(0), TRACE_HEADERS ? pcTrace : NULL, cDefaultBinProcPars );
  cSEDecoder.InStreamProcessor<CtxBin&>::setInStreamIf( cProbAllocDec.getInStreamIf() );
  cSEDecoder.InStreamProcessor<EPCtxBin&>::setInStreamIf( cProbAllocDec.getInStreamIf() );

  //TODO read max and min size from cubic stream

  RNOK( cSEDecoder.processWienerParameters(rcAUPars.getWienerFilterParameters(), rcAUPars.getNumPlanes() ) );

  if( rcAUPars.getWienerFilterParameters().getUseBlocksFlag() )
  {
    rcAUPars.getWienerFilterParameters().initQuadtreeGroupStructure(rcAUPars.getPicSize());
    WienerParser cWfParser( &cSEDecoder, rcAUPars.getWienerFilterParameters(),pcTrace);
    RNOK ( cWfParser.processTree() );
  }

  return Err::m_nOK;
}


ErrVal AccessUnitDecoder::decodePlanePackage( UInt uiPlaPaId,
                                              const PicBlock<Pel> &rcDec,
                                              const CodingState &rcCodingState,
                                              const AccessUnit &rcAU )
{
  const PlanePackage& rcPlaPaCurr = *rcAU.getPlanePackage( uiPlaPaId );
  SliceDecoder cSliceDecoder( rcDec, rcCodingState, rcPlaPaCurr.getPars(), rcPlaPaCurr.getCurrentSlice() );
  RNOK( cSliceDecoder.processTree() );
  return Err::m_nOK;
}


ErrVal AccessUnitDecoder::decodeNalUnit( AccessUnitParameters& rcAUPars, std::vector<UChar> *pcRawData, const std::vector<UChar> &rcNalUnit, bool& rbSizeInfoAvailable, TraceCommon* pcTrace, SequenceParameterSet*& rpcSPS, PictureParameterSet*& rpcPPS)
{
  rbSizeInfoAvailable = false;
  std::vector<UChar> cTempData;
  cTempData.reserve( rcNalUnit.size()+ 256);
  RNOK( NalUnitParser::convert( cTempData, rcNalUnit));
  BitReader<UChar,UInt> cBitReader( true, cTempData.begin(), cTempData.end() );
  UvlcReader cUvlcReader( cBitReader );

#if TRACE_HEADERS
  cUvlcReader.setTraceCommonPtr( pcTrace );
#endif

  UInt uiFirstByte;
  cBitReader.read( uiFirstByte, 8 );
  const NalUnitType eNalUnitType = NalUnitType( uiFirstByte & 0x1F );
  const NalRefIdc eNalRefIdc     = NalRefIdc( uiFirstByte >> 5 );

  switch( eNalUnitType )
  {
  case NAL_UNIT_SPS:
    {
      if( NULL == rpcSPS )
      {
        RNOK( SequenceParameterSet::create(rpcSPS));
      }
      RNOK( rpcSPS->read( &cUvlcReader) );
      return Err::m_nOK;
    }
  case NAL_UNIT_PPS:
    {
      if( NULL == rpcPPS )
      {
        RNOK( PictureParameterSet::create(rpcPPS));
      }
      RNOK( rpcPPS->read( &cUvlcReader) );
      return Err::m_nOK;
    }
  case NAL_UNIT_UNSPECIFIED_30:
  case NAL_UNIT_UNSPECIFIED_31:
    {
      ROT( NULL == rpcSPS);
      ROT( NULL == rpcPPS);
      // the following method creates a new AUPars instance because ...

      SliceHeader cSliceHeader( *rpcSPS, *rpcPPS);
      cSliceHeader.setNalUnitType( eNalUnitType );
      cSliceHeader.setNalRefIdc( eNalRefIdc );
      RNOK( cSliceHeader.readSH( &cUvlcReader, rcAUPars ) );
      cUvlcReader.readByteAlignOnes();


      if( pcRawData )
      {
        int iOffset = (int)cBitReader.getNumBitsRead() /8;
        pcRawData->resize(0);
        pcRawData->insert( pcRawData->end(), cTempData.begin()+iOffset, cTempData.end());
      }
      rbSizeInfoAvailable = true;
      return Err::m_nOK;
    }
  default:
    {
      ROT(1);
    }
  }

  return Err::m_nERR;
}

// video namespace end
VIDEO_NAMESPACE_END
