/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _BITREADER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _BITREADER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "BitReaderIf.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_UnsignedStoreType, class t_UnsignedReadType>
class BitReader
  : public BitReaderIf<t_UnsignedReadType>
{
public:
  struct InitPars
  {
    bool m_bConvertRBSPToSOBD;
    typename std::vector<t_UnsignedStoreType>::const_iterator m_cBegin;
    typename std::vector<t_UnsignedStoreType>::const_iterator m_cEnd;
  };

  ErrVal init( InitPars &r );

  BitReader() {}

  BitReader( bool bConvertRBSPToSOBD,
             typename std::vector<t_UnsignedStoreType>::const_iterator cBegin,
             typename std::vector<t_UnsignedStoreType>::const_iterator cEnd )
  {
    InitPars cPars = { bConvertRBSPToSOBD, cBegin, cEnd };
    TNOK( init( cPars ) );
  }


  int getBitsUntilByteAligned() const { return 7 & m_uiNextBitPosInIter; }
  UInt64 getNumBitsRead() { return m_uiBitsRead; }

  ErrVal read( t_UnsignedReadType &ruiBits, UInt uiNumberOfBits = 1 ) { return readWithSpecifiedReadTyp<t_UnsignedReadType>( ruiBits, uiNumberOfBits ); }

  template<class t_ReadType>
  ErrVal readWithSpecifiedReadTyp( t_ReadType &ruiBits, UInt uiNumberOfBits = 1 );

  UInt64 getNumRemainingBits() { return m_iBitsLeft; }

  UInt64 getNumRemainingBitsToEnd() { return gMax<Int64>( 0, ( Int64( m_cEnd - m_cIter ) - 1 ) * 8 * Int64( sizeof( t_UnsignedStoreType ) ) + Int64( m_uiNextBitPosInIter ) ); }

private:
  ErrVal xConvertRBSPToSODB( size_t& riBitsInPacket ) const;

protected:
  typename std::vector<t_UnsignedStoreType>::const_iterator m_cIter;
  typename std::vector<t_UnsignedStoreType>::const_iterator m_cEnd;

private:
  virtual void xNextIter() { ++m_cIter; }
  UInt m_uiNextBitPosInIter;
  UInt m_uiBitsRead;
  size_t m_iBitsLeft;
};


template<class t_UnsignedStoreType, class t_UnsignedReadType>
ErrVal BitReader<t_UnsignedStoreType, t_UnsignedReadType>::init( InitPars &r )
{
  m_cIter = r.m_cBegin;
  m_cEnd = r.m_cEnd;
  m_uiNextBitPosInIter = 8 * sizeof( t_UnsignedStoreType );
  m_uiBitsRead = 0;
  m_iBitsLeft = (r.m_cEnd - r.m_cBegin) * 8 * sizeof( t_UnsignedStoreType );

  if( r.m_bConvertRBSPToSOBD )
  {
    RNOK( xConvertRBSPToSODB( m_iBitsLeft ) );
  }
  return Err::m_nOK;
}


template<class t_UnsignedStoreType, class t_UnsignedReadType>
template<class t_ReadType>
ErrVal BitReader<t_UnsignedStoreType, t_UnsignedReadType>::readWithSpecifiedReadTyp( t_ReadType &ruiBits, UInt uiNumberOfBits )
{
  m_uiBitsRead += uiNumberOfBits;
  ROT( uiNumberOfBits > 8 * sizeof( t_ReadType ) );
  ROT( uiNumberOfBits == 0 );

  m_iBitsLeft -= uiNumberOfBits;
  ROT( 0 > m_iBitsLeft);

  ruiBits = 0;

  if( uiNumberOfBits < m_uiNextBitPosInIter )
  {
    m_uiNextBitPosInIter -= uiNumberOfBits;
    ROTS( m_cIter == m_cEnd );
    ruiBits = (*m_cIter) >> m_uiNextBitPosInIter;
  }
  else
  {
    UInt uiShift = uiNumberOfBits - m_uiNextBitPosInIter;
    ruiBits = t_ReadType(*m_cIter) << uiShift;

    while( uiShift >= 8 * sizeof( t_UnsignedStoreType ) )
    {
      uiShift -= 8 * sizeof( t_UnsignedStoreType );
      xNextIter();
      ROTS( m_cIter == m_cEnd );
      ruiBits |= t_ReadType(*m_cIter) << uiShift;
    }

    xNextIter();

    m_uiNextBitPosInIter = 8 * sizeof( t_UnsignedStoreType ) - uiShift;
    if( uiShift != 0 )
    {
      ROTS( m_cIter == m_cEnd );
      ruiBits |= (*m_cIter) >> m_uiNextBitPosInIter;
    }
  }
  const t_ReadType uiMask = ((( t_ReadType(1) << (uiNumberOfBits-1) ) - 1 ) << 1 ) | 1;
  ruiBits &= uiMask;
  return Err::m_nOK;
}


template<class t_UnsignedStoreType, class t_UnsignedReadType>
ErrVal BitReader<t_UnsignedStoreType, t_UnsignedReadType>::xConvertRBSPToSODB( size_t& riBitsInPacket ) const 
{
  size_t uiPacketLength = (m_cEnd - m_cIter)-1;
//  UInt uiPacketLength = rcRawData.size()-1;

  CAOF(( TypeCompare<t_UnsignedStoreType, UChar>::m_bEqual ));

  const UChar *puc = &(m_cIter[0]);

  //remove zero bytes at the end of the stream
  while (puc[uiPacketLength] == 0x00)
  {
    uiPacketLength-=1;
    // ROT(uiPacketLength<0);
  }

  // find the first non-zero bit
  UChar ucLastByte=puc[uiPacketLength];
  int i = 0;

  for (i=0; (ucLastByte&1)==0; i++)
  {
    ucLastByte>>=1;
    AOT_DBG (i>7);
  }

  riBitsInPacket = (uiPacketLength << 3) + 8-i;

  return Err::m_nOK;
}


template<class t_UnsignedStoreType, class t_UnsignedReadType>
class MultiVectorBitReader
  : public BitReader<t_UnsignedStoreType, t_UnsignedReadType>
{
public:
  struct InitPars
  {
    InitPars( const std::list<typename std::vector<t_UnsignedStoreType>::const_iterator> &rcBegin,
              const std::list<typename std::vector<t_UnsignedStoreType>::const_iterator> &rcEnd ) : m_cBegin( rcBegin ), m_cEnd( rcEnd ) {}

    InitPars( typename std::list<std::vector<t_UnsignedStoreType> >::const_iterator cIter,
              const typename std::list<std::vector<t_UnsignedStoreType> >::const_iterator &rcEnd )
    {
      for( ; cIter != rcEnd; ++cIter )
      {
        m_cBegin.push_back( (*cIter).begin() );
        m_cEnd.push_back( (*cIter).end() );
      }
    }

    std::list<typename std::vector<t_UnsignedStoreType>::const_iterator> m_cBegin;
    std::list<typename std::vector<t_UnsignedStoreType>::const_iterator> m_cEnd;
  };

  ErrVal init( InitPars &r )
  {
    m_cBeginList = r.m_cBegin;
    m_cEndList   = r.m_cEnd;

    ROF( m_cBeginList.size() == m_cEndList.size() );

    m_cBeginListIter = m_cBeginList.begin();
    m_cEndListIter   = m_cEndList.begin();

    if( !m_cBeginList.empty() )
    {
      typename BitReader<t_UnsignedStoreType, t_UnsignedReadType>::InitPars cPars = { false, *m_cBeginListIter, *m_cEndListIter };
      RNOK(( BitReader<t_UnsignedStoreType, t_UnsignedReadType>::init( cPars ) ));
    }
    return Err::m_nOK;
  }

  void xNextIter()
  {
    if( ++BitReader<t_UnsignedStoreType, t_UnsignedReadType>::m_cIter == BitReader<t_UnsignedStoreType, t_UnsignedReadType>::m_cEnd )
    {
      ++m_cBeginListIter;
      ++m_cEndListIter;
      if( m_cBeginListIter != m_cBeginList.end() )
      {
        BitReader<t_UnsignedStoreType, t_UnsignedReadType>::m_cIter = *m_cBeginListIter;
        BitReader<t_UnsignedStoreType, t_UnsignedReadType>::m_cEnd  = *m_cEndListIter;
      }
    }
  }

private:
  std::list<typename std::vector<t_UnsignedStoreType>::const_iterator> m_cBeginList;
  std::list<typename std::vector<t_UnsignedStoreType>::const_iterator> m_cEndList;
  typename std::list<typename std::vector<t_UnsignedStoreType>::const_iterator>::const_iterator m_cBeginListIter;
  typename std::list<typename std::vector<t_UnsignedStoreType>::const_iterator>::const_iterator m_cEndListIter;
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _BITREADER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
