/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoDecoderLibPch.h"
#include "UvlcReader.h"
#include "BitReaderIf.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN


__inline 
ErrVal UvlcReader::xGetCode( UInt& ruiCode, UInt uiLength )
{
  ErrVal retVal = m_pcBitReaderIf->read( ruiCode, uiLength );
  return retVal;
}

__inline 
ErrVal UvlcReader::xGetFlag( UInt& ruiCode )
{
  ErrVal retVal = m_pcBitReaderIf->read( ruiCode );
  return retVal;
}


ErrVal UvlcReader::xGetUvlcCode( UInt& ruiVal)
{
  UInt uiVal = 0;
  UInt uiCode = 0;
  UInt uiLength;

  RNOK( m_pcBitReaderIf->read( uiCode ) );

  if( 0 == uiCode )
  {
    uiLength = 0;

    while( 0 == uiCode )
    {
      RNOK( m_pcBitReaderIf->read( uiCode ) );
      uiLength++;
    }


    RNOK( m_pcBitReaderIf->read( uiVal, uiLength ) );

    uiVal += (1 << uiLength)-1;
  }

  ruiVal = uiVal;


  return Err::m_nOK;
}


ErrVal UvlcReader::xGetUvlc64Code( UInt64& ruiVal)
{
  UInt64 uiVal = 0;
  UInt uiCode = 0;
  UInt uiLength;

  RNOK( m_pcBitReaderIf->read( uiCode ) );

  if( 0 == uiCode )
  {
    uiLength = 0;

    while( 0 == uiCode )
    {
      RNOK( m_pcBitReaderIf->read( uiCode ) );
      uiLength++;
    }

    for( UInt uiBit = 0; uiBit < uiLength; uiBit++ )
    {
      RNOK( m_pcBitReaderIf->read( uiCode ) );
      uiVal = ( uiVal << 1 ) | uiCode;
    }

    uiVal += (1 << uiLength)-1;
  }

  ruiVal = uiVal;


  return Err::m_nOK;
}




ErrVal UvlcReader::xGetSvlcCode( int& riVal)
{
  UInt uiBits = 0;

  RNOK( m_pcBitReaderIf->read( uiBits ) );

  if( 0 == uiBits )
  {
    UInt uiLength = 0;

    while( 0 == uiBits )
    {
      RNOK( m_pcBitReaderIf->read( uiBits ) );
      uiLength++;
    }

    RNOK( m_pcBitReaderIf->read( uiBits, uiLength ) );
    uiBits += (1 << uiLength);
    riVal = uiBits>>1;
    riVal = ( uiBits & 1) ? -riVal : riVal;
  }
  else
  {
    riVal = 0;
  }
  return Err::m_nOK;
}



ErrVal UvlcReader::getFlag( bool& rbFlag, const char* pcTraceString )
{
  UInt uiCode;
  RNOK( xGetFlag( uiCode ) );

  rbFlag = (1 == uiCode);
  xTraceString( pcTraceString, rbFlag );
  return Err::m_nOK;
}



ErrVal UvlcReader::getSCode( int& riCode, UInt uiLength, const char* pcTraceString )
{
  UInt uiCode;
  RNOK( m_pcBitReaderIf->read( uiCode, uiLength ) );

  UInt uiShift = 32 - uiLength;
  riCode = ((int)(uiCode << uiShift)) >> uiShift;
  xTraceString( pcTraceString, riCode );
  return Err::m_nOK;
}


ErrVal UvlcReader::getCode( UInt& ruiCode, UInt uiLength, const char* pcTraceString )
{
  RNOK( m_pcBitReaderIf->read( ruiCode, uiLength ) );

  xTraceString( pcTraceString, ruiCode );
  return Err::m_nOK;
}


ErrVal UvlcReader::getUvlc( UInt& ruiCode, const char* pcTraceString )
{
  RNOK( xGetUvlcCode( ruiCode ) );
  xTraceString( pcTraceString, ruiCode );
  return Err::m_nOK;
}


ErrVal UvlcReader::getUvlc64( UInt64& ruiCode, const char* pcTraceString )
{
  RNOK( xGetUvlc64Code( ruiCode ) );
  xTraceString( pcTraceString, ruiCode );
  return Err::m_nOK;
}


ErrVal UvlcReader::getSvlc( int& riCode, const char* pcTraceString )
{
  RNOK( xGetSvlcCode( riCode ) );
  xTraceString( pcTraceString, riCode );
  return Err::m_nOK;
}

bool UvlcReader::moreRBSPData()
{
  return 1 < m_pcBitReaderIf->getNumRemainingBits();
}

ErrVal UvlcReader::readByteAlignOnes()
{
  UInt uiCode;
  UInt uiLength = m_pcBitReaderIf->getBitsUntilByteAligned();

  ROTRS( 0 == uiLength, Err::m_nOK );

  RNOK( m_pcBitReaderIf->read( uiCode, uiLength ) );
  xTraceString( "read_byte_align_ones", uiCode );

  ROTR( (UInt(1<<uiLength)-1) != uiCode, Err::m_nInvalidParameter );;

  return Err::m_nOK;
}

ErrVal UvlcReader::readByteAlign()
{
  UInt uiCode;
  UInt uiLength = m_pcBitReaderIf->getBitsUntilByteAligned();

  ROTRS( 0 == uiLength, Err::m_nOK );

  RNOK( m_pcBitReaderIf->read( uiCode, uiLength ) );
  xTraceString( "read_byte_align", uiCode );

  ROTR( (UInt(1<<uiLength)>>1) != uiCode, Err::m_nInvalidParameter );;

  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
