/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoEncoderLibPch.h"
#include "AccessUnitEncoder.h"
#include "SliceEncoder.h"
#include "SliceEstimator.h"
#include "VideoCommonLib/CodingState.h"
#include "VideoCommonLib/SliceData.h"
#include "UvlcWriter.h"
#include "VideoCommonLib/BitWriter.h"
#include "VideoCommonLib/AccessUnit.h"
#include "VideoCommonLib/CodedPictureStore.h"
#include "ReferencePictureEnc.h"
#include "TraceEnc.h"
#include "VideoCommonLib/DeblockingFilter.h"
#include "WienerFilterEncoder.h"

#include "../VideoDecoderLib/AccessUnitDecoder.h"
#include "../VideoDecoderLib/WienerFilterDecoder.h"

#include "H264NalUnitEncoder.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

ErrVal AccessUnitEncoder::parsePrecodedAccessUnit( PacketParser<LargeFile> &rcPrecodedBitstreamParser,
                                                   const CodedPictureStoreBase &rcCodedPictureStore,
                                                   bool bParse,
#if MULTICABAC_TRANSCODING_HACK
                                                   UInt& ruiParsedPacketSize,
#endif
                                                   const AccessUnitParameters &rcAUPars,
                                                   AccessUnit *pcAU,
                                                   TraceEnc *pcTraceEnc,
                                                   SequenceParameterSet*& rpcSequenceParameterSet,
                                                   PictureParameterSet*& rpcPictureParameterSet)
{
  ErrVal nRet;
  bool bSizeInfoAvailable = false;
  std::vector<UChar> cRawData;
  while( false == bSizeInfoAvailable )
  {
    nRet = rcPrecodedBitstreamParser.readNextPacket();
#if MULTICABAC_TRANSCODING_HACK
    ruiParsedPacketSize = UInt( rcPrecodedBitstreamParser.getVector().size() );
#endif
    ROFRS( bParse, Err::m_nOK );
    
    if( Err::m_nOK != nRet )
    {
      std::cerr << "Could not parse packet" << std::endl;
      return Err::m_nERR;
    }

    cRawData.clear();
    RNOK( AccessUnitDecoder::decodeNalUnit( pcAU->getAUPars(), &cRawData, rcPrecodedBitstreamParser.getVector(), bSizeInfoAvailable, pcTraceEnc, rpcSequenceParameterSet, rpcPictureParameterSet)); 
  }


  PicOrderCnt cPoc = 0;
  RNOK( AccessUnitDecoder::parseAccessUnit( cRawData,
                                            rcCodedPictureStore,
                                            pcAU,
                                            NULL,
                                            cPoc,
                                            pcTraceEnc ) );
  return Err::m_nOK;
}


ErrVal AccessUnitEncoder::xWriteMultiplex( std::vector<MultiVectorBitWriter<UChar,UInt>*> &racPackets,
                                           std::list<std::vector<UChar> > &rcBitstreamData,
                                           BitWriterIf<UInt> &rcBitWriter,
                                           UInt &ruiBytesInMultiplex )
{
  ruiBytesInMultiplex = 0;
  UvlcWriter cPacketizerWriter( rcBitWriter );

  UInt uiLastSize = 0;
  std::vector<UInt> auiTotalSizes( racPackets.size(), 0 );
  for( UInt ui = 0; ui < UInt( racPackets.size() ); ++ui )
  {
    std::list<std::vector<UChar> > &racFragments = racPackets[ui]->getVectors();
    for( std::list<std::vector<UChar> >::iterator cIter = racFragments.begin(); cIter != racFragments.end(); ++cIter )
    {
      auiTotalSizes[ui] += UInt( (*cIter).size() );
    }
    RNOK( cPacketizerWriter.writeSvlc( int( auiTotalSizes[ui] ) - int( uiLastSize ), "diff_fragment_size_sum_in_bytes" ) );
    uiLastSize = auiTotalSizes[ui];
    ruiBytesInMultiplex += auiTotalSizes[ui];
  }

  bool bFragmented = false;
  std::vector<UInt> auiFragmentSizesCheck( racPackets.size(), 0 );
  for( UInt ui = 0; ui < UInt( racPackets.size() ); ++ui )
  {
    if( auiTotalSizes[ui] )
    {
      auiFragmentSizesCheck[ui] = UInt( racPackets[ui]->getVectors().front().size() );
      UInt uiCode = 0;
      if( auiFragmentSizesCheck[ui] != auiTotalSizes[ui] )
      {
        uiCode = auiFragmentSizesCheck[ui];
      }
      RNOK( cPacketizerWriter.writeUvlc( uiCode, "fragment_size_in_bytes" ) );
      if( uiCode )
      {
        bFragmented = true;
      }
    }
  }

  if( bFragmented )
  {
    RNOK( cPacketizerWriter.writeFlag( true, "fragments_are_independently_decodable" ) );
  }

  bool bAllDone = false;
  while( !bAllDone )
  {
    bAllDone = true;
    for( UInt ui = 0; ui < UInt( racPackets.size() ); ++ui )
    {
      if( auiTotalSizes[ui] )
      {
        bAllDone = false;
        std::list<std::vector<UChar> > &racFragments = racPackets[ui]->getVectors();
        rcBitstreamData.push_back( std::vector<UChar>() );
        rcBitstreamData.back().swap( racFragments.front() );
        racFragments.pop_front();

        if( auiTotalSizes[ui] >= auiFragmentSizesCheck[ui] )
        {
          ROF( UInt( rcBitstreamData.back().size() ) == auiFragmentSizesCheck[ui] );
          auiTotalSizes[ui] -= auiFragmentSizesCheck[ui];
        }
        else
        {
          ROF( UInt( rcBitstreamData.back().size() ) == auiTotalSizes[ui] );
          auiTotalSizes[ui] = 0;
        }
      }
    }
  }

  return Err::m_nOK;
}


ErrVal AccessUnitEncoder::encodeAccessUnit( const PicBlock<const Pel> &rcOrig,
                                            const PicBlock<Pel> &rcEnc,
                                            const PicBlock<Pel> *pcEncPre,
                                            const PicBlock<Pel> *pcEncRes,
                                            const AccessUnitParameters &rcAUPars,
                                            const AccessUnitEncoderParameters &rcAUEstPar,
                                            const std::vector<const CodingState*> &rapcCodingState,
                                            PicOrderCnt iLastKeyPicPoc0,
                                            PicOrderCnt iLastKeyPicPoc1,
                                            AccessUnit *pcAU,
                                            bool bAUWasParsed,
#if MULTICABAC_TRANSCODING_HACK
                                            const UInt uiParsedPacketSize,
#endif
                                            std::list<std::vector<UChar> > &rcBitstreamData,
                                            std::vector<double> &radPSNR,
                                            std::ostream &rcOut,
                                            UInt64 &ruiEncodingDurationMSec,
                                            UInt64 &ruiMEDurationMSec,
                                            TraceEnc *pcTraceEnc )
{
  rcOut << std::setw(8) << rcAUPars.getPicOrderCnt();

  TimeStamp cStart;
  cStart.stamp();

  ROF( rcOrig.getStats() == rcEnc.getStats() );

  if( !bAUWasParsed )
  {
    pcAU->getAUPars() = rcAUPars;
    RNOK( pcAU->createPlanePackages() );
  }
  else
  {
    pcAU->getAUPars().setVLCFlag              ( rcAUPars.getVLCFlag             () );
    pcAU->getAUPars().setVLCMaxDelayParameter ( rcAUPars.getVLCMaxDelayParameter() );
#if MULTICABAC_TRANSCODING_HACK
    pcAU->getAUPars().setExtendedCabacTables  ( rcAUPars.getExtendedCabacTables () );
#endif
  }

  ExtendedTablesCabaEncoder *pcExtTabCabaEncoder = new ExtendedTablesCabaEncoder;
  RNOK( pcExtTabCabaEncoder->createEncoder( pcAU->getAUPars() ) );

  ROF( rcBitstreamData.empty() );

  if( !pcAU->getAUPars().getVLCFlag() && pcAU->getAUPars().getExtendedCabacTables() != DEFAULT_CABAC_TABLES && pcAU->getAUPars().getNumPlanePacks() != 1 )
  {
    std::cout << std::endl << "currently, only one plane package is possible" << std::endl;
    TOT( 1 );
  }

  UInt64 uiTotalMEDuration = 0;
  for( UInt uiRuns = 0; uiRuns < pcAU->getAUPars().getNumPlanePacks(); uiRuns++ )
  {
    SliceEstimator::Parameters cSliceEstPars;
    PicBlock<const Pel> cOrigPlanePackage;
    PicBlock<Pel> cEncPlanePackage;
    PicBlock<Pel> *pcEncPrePlanePackage = pcEncPre ? new PicBlock<Pel> : NULL;
    PicBlock<Pel> *pcEncResPlanePackage = pcEncRes ? new PicBlock<Pel> : NULL;
    std::stringstream cConsoleOutput;

    cSliceEstPars.setLambda( rcAUPars,
                             iLastKeyPicPoc0,
                             iLastKeyPicPoc1,
                             rcAUEstPar.getNextPoc(),
                             rcAUEstPar.getLambdaCalculationMode(),
                             rcAUEstPar.getQuantizationMode( uiRuns ),
                             rcAUEstPar.getPrimQp() );
    RNOK( pcAU->extractPlanePackagePicBlock( uiRuns, rcOrig, cOrigPlanePackage ) );
    RNOK( pcAU->extractPlanePackagePicBlock( uiRuns, rcEnc, cEncPlanePackage ) );

    if ( pcEncPrePlanePackage )
    {
      RNOK( pcAU->extractPlanePackagePicBlock( uiRuns, *pcEncPre, *pcEncPrePlanePackage ) );
    }

    if ( pcEncResPlanePackage )
    {
      RNOK( pcAU->extractPlanePackagePicBlock( uiRuns, *pcEncRes, *pcEncResPlanePackage ) );
    }

    UInt64 uiPlaPaMEDuration = 0;

    const bool bTwoPassEncode = rcAUEstPar.getTwoPassEncodingFlag();

    BitRate cEstimatedBitrate( 0 );

    if( bAUWasParsed )
    {
      RNOK( AccessUnitDecoder::decodePlanePackage( uiRuns, cEncPlanePackage, *rapcCodingState[uiRuns], *pcAU ) );
#if MULTICABAC_TRANSCODING_HACK
      cEstimatedBitrate = BitRate( 8 * uiParsedPacketSize );
#endif
      cConsoleOutput << std::setw( 4 ) << pcAU->getPlanePackage( uiRuns )->getCurrentSlice().getSliceQps()[0] << ( rcAUPars.getRefPicFlag() ? " PARSED" : " parsed" );
    }
    else
    {
      if( bTwoPassEncode )
      {
//        RNOK( xEstimatePlanePackage<false>( uiRuns, pcAU->getAUPars().getExtendedCabacTables(), cOrigPlanePackage, cEncPlanePackage, pcEncPrePlanePackage, pcEncResPlanePackage, *pcAU, cSliceEstPars, rcAUEstPar,
//                                            *rapcCodingState[uiRuns], NULL, uiPlaPaMEDuration, cConsoleOutput, NULL ) );
        RNOK( xEstimatePlanePackage<false>( uiRuns, DEFAULT_CABAC_TABLES, cOrigPlanePackage, cEncPlanePackage, pcEncPrePlanePackage, pcEncResPlanePackage, *pcAU, cSliceEstPars, rcAUEstPar,
                                            *rapcCodingState[uiRuns], NULL, uiPlaPaMEDuration, cConsoleOutput, NULL, cEstimatedBitrate ) );
      }
      else
      {
#if USE_CABAC12_TABLE_IN_ENCODER
        TOT( 1 );
#endif
//        RNOK( xEstimatePlanePackage<true>( uiRuns, pcAU->getAUPars().getExtendedCabacTables(), cOrigPlanePackage, cEncPlanePackage, pcEncPrePlanePackage, pcEncResPlanePackage, *pcAU, cSliceEstPars, rcAUEstPar,
//                                           *rapcCodingState[uiRuns], pcExtTabCabaEncoder.getProbBinStreamEncIf(), uiPlaPaMEDuration, cConsoleOutput, pcTraceEnc ) );
        RNOK( xEstimatePlanePackage<true>( uiRuns, DEFAULT_CABAC_TABLES, cOrigPlanePackage, cEncPlanePackage, pcEncPrePlanePackage, pcEncResPlanePackage, *pcAU, cSliceEstPars, rcAUEstPar,
                                           *rapcCodingState[uiRuns], &pcExtTabCabaEncoder->getProbBinStreamEncIf(), uiPlaPaMEDuration, cConsoleOutput, pcTraceEnc, cEstimatedBitrate ) );
      }
    }
    if( bTwoPassEncode )
    {
      RNOK( pcAU->getCtxStore( uiRuns ).clearCtxStates() );

      const ContextStore<CabacContextModel>* pcRefCtxStore = NULL;
      if( rcAUPars.getCtxStatesReuseFlag() && rapcCodingState[uiRuns]->getRefCtxStore() != NULL )
      {
        pcRefCtxStore = rapcCodingState[uiRuns]->getRefCtxStore();
      }

      const BitRate cThreshold( 8 * 12000 );
#if MULTICABAC_TRANSCODING_HACK
      if( !pcAU->getAUPars().getVLCFlag() && pcAU->getAUPars().getExtendedCabacTables() != DEFAULT_CABAC_TABLES && cEstimatedBitrate < cThreshold )
#else
      if( !pcAU->getAUPars().getVLCFlag() && !bAUWasParsed && pcAU->getAUPars().getExtendedCabacTables() != DEFAULT_CABAC_TABLES && cEstimatedBitrate < cThreshold )
#endif
      {
        // switch to DEFAULT_CABAC_TABLES
        pcAU->getAUPars().setExtendedCabacTables( DEFAULT_CABAC_TABLES );
        delete pcExtTabCabaEncoder;
        pcExtTabCabaEncoder = new ExtendedTablesCabaEncoder;
        RNOK( pcExtTabCabaEncoder->createEncoder( pcAU->getAUPars() ) );
      }

      RNOK( encodePlanePackage( uiRuns,
                                *pcAU,
                                pcAU->getCtxStore( uiRuns ),
                                pcRefCtxStore,
                                pcExtTabCabaEncoder->getProbBinStreamEncIf(),
                                pcExtTabCabaEncoder->getCabaCoderStartFinishIf(),
                                pcExtTabCabaEncoder->getDefaultCabacBitWriter(),
                                rcBitstreamData,
                                pcTraceEnc ) );
    }
    uiTotalMEDuration += uiPlaPaMEDuration;
    if( uiRuns == 0 )
    {
      rcOut << cConsoleOutput.str();
    }

    if( pcAU->getAUPars().getDeblockingIdc() == 0 )
    {
      std::vector<bool> abIsLuma( cEncPlanePackage.getStats().getNumPlanes(), false );
      if( uiRuns == 0 )
      {
        abIsLuma[0] = true;
      }
      UInt uiLog2MinBlkSize = ( abIsLuma[0] ? 2 : 1 ); // allow 2x2 for chroma plane packages (req. for inter plane prediction)
      RNOK( DeblockingFilter::deblockPlanePackage<Pel>( *pcAU->getPlanePackage( uiRuns ),
                                                         cEncPlanePackage,
                                                         abIsLuma,
                                                         gMin<int>( uiLog2MinBlkSize, pcAU->getAUPars().getPlaPaPars( uiRuns ).getLog2MinPow2TrafoBlkSize() ),
                                                         2 * pcAU->getAUPars().getDeblockAlphaOffs(),
                                                         2 * pcAU->getAUPars().getDeblockBetaOffs() ) );
    }
    delete pcEncPrePlanePackage;
    delete pcEncResPlanePackage;
  }

  size_t uiEncFrameSize = 0;
  if( !pcAU->getAUPars().getEntropySliceFlag() )
  {
    TOF( rcBitstreamData.empty() );
    RNOK( pcExtTabCabaEncoder->getCabaCoderStartFinishIf().finishCabac() );
    if( pcAU->getAUPars().getVLCFlag() || pcAU->getAUPars().getExtendedCabacTables() == DEFAULT_CABAC_TABLES )
    {
      uiEncFrameSize += pcExtTabCabaEncoder->getDefaultCabacBitstream().size();
      rcBitstreamData.push_back( std::vector<UChar>() );
      rcBitstreamData.back().swap( pcExtTabCabaEncoder->getDefaultCabacBitstream() );
    }
  }
  else
  {
    std::list< std::vector< UChar > >::iterator it = rcBitstreamData.begin();
    while ( it != rcBitstreamData.end() )
    {
      uiEncFrameSize += it->size();
      uiEncFrameSize += 4;
      {
        std::vector<UChar> cSliceSize;
        const UInt uiSize = UInt( it->size() );
        CAOF( sizeof( UInt ) == 4 );
        ROF( size_t( uiSize ) == it->size() );
        for( UInt ui = 0; ui < 4; ui++ )
        {
          const UChar ucByte = uiSize >> ui*8;
          cSliceSize.push_back( ucByte );
        }
        rcBitstreamData.insert( it, cSliceSize );
      }
      it++;
    }
  }

  SliceEstimator::Parameters cSliceEstPars;
  cSliceEstPars.setLambda( rcAUPars,
               iLastKeyPicPoc0,
               iLastKeyPicPoc1,
               rcAUEstPar.getNextPoc(),
               rcAUEstPar.getLambdaCalculationMode(),
               rcAUEstPar.getQuantizationMode( 0 ),
               rcAUEstPar.getPrimQp() );

  AccessUnitParameters      cAUParsCopy = rcAUPars; cAUParsCopy.setExtendedCabacTables( DEFAULT_CABAC_TABLES );
  ExtendedTablesCabaEncoder cExtTabCabaEncoder;
  cExtTabCabaEncoder.createEncoder( cAUParsCopy );

  int iWfQp =  pcAU->getPlanePackage( 0 )->getCurrentSlice().getSliceQps()[0];
  
  Block<double> *pcInLoopFiltMatrix   = NULL;
  bool bAdapInterpSwitch              = ((rcAUPars.getMaxMvsPerBlk() < 2) && (rcAUPars.getAdapInterpSwitch() != 0));

  cExtTabCabaEncoder.getCabaCoderStartFinishIf().startCabac();
  if( pcTraceEnc )
  {
    pcTraceEnc->resetCount();
  }
  if( bAUWasParsed )
  {
    WienerFilterDecoder::applyFilterToPicBlockDec( pcAU->getAUPars(), rcEnc);
    WienerFilterEncoder<Pel>::encodeInLoopFilter(pcAU->getAUPars(), &cExtTabCabaEncoder.getProbBinStreamEncIf(), iWfQp, TRACE_HEADERS ? pcTraceEnc : NULL ) ;
  }
  else
  {
    WienerFilterEncoder<Pel>::estimateAndEncodeInLoopFilter( pcAU->getAUPars(), rcAUEstPar, pcAU->getCtxStore(0), rcEnc, rcOrig, &cExtTabCabaEncoder.getProbBinStreamEncIf(), iWfQp , cSliceEstPars.getLambda(), TRACE_HEADERS ? pcTraceEnc : NULL, pcInLoopFiltMatrix, bAdapInterpSwitch );
  }
  cExtTabCabaEncoder.getCabaCoderStartFinishIf().finishCabac();

  if ( bAdapInterpSwitch && pcInLoopFiltMatrix != NULL )
  {
    UInt uiUpsMethodAdap = rcAUPars.getInterpolationMethod();
    RNOK( xEstimateInterpolationMethod( pcInLoopFiltMatrix, uiUpsMethodAdap ) );
    pcAU->getAUPars().setInterpolationMethod( uiUpsMethodAdap );
  }
  delete pcInLoopFiltMatrix;

  std::vector<UChar> aucHeaderBitstream;
  BitWriter<UChar,UInt> cBitWriterHeader( aucHeaderBitstream );
  UvlcWriter cUvlcWriterHeader(cBitWriterHeader);

#if TRACE_HEADERS
  if( pcTraceEnc )
  {
    pcTraceEnc->resetCount();
  }
  cUvlcWriterHeader.setTraceCommonPtr( pcTraceEnc );
#endif

  NalUnitEncoder cNalUnitEncoder;
  RNOK( cNalUnitEncoder.init( &cBitWriterHeader, &cUvlcWriterHeader ) );
  RNOK( cNalUnitEncoder.initNalUnit());

  SliceHeader cSliceHeader( *(rcAUPars.getSPS()), *(rcAUPars.getPPS()));
  RNOK( cSliceHeader.setAccessUnitParameters( pcAU->getAUPars() ) );
  RNOK( cNalUnitEncoder.writeSliceHeader( cSliceHeader ) );
  RNOK( cBitWriterHeader.writeAlignOne() );

  if( !pcAU->getAUPars().getEntropySliceFlag() && !pcAU->getAUPars().getVLCFlag() && pcAU->getAUPars().getExtendedCabacTables() != DEFAULT_CABAC_TABLES )
  {
    TOF( rcBitstreamData.empty() );
    rcBitstreamData.push_back( std::vector<UChar>() );
    BitWriter<UChar,UInt> cBitWriterMultiplexHeader( rcBitstreamData.front() );
    UInt uiBitstreamSize = 0;
    RNOK( xWriteMultiplex( pcExtTabCabaEncoder->getMultiBitWriters(), rcBitstreamData, cBitWriterMultiplexHeader, uiBitstreamSize ) );
    uiEncFrameSize += uiBitstreamSize;
    uiEncFrameSize += rcBitstreamData.front().size();
  }

  if( pcAU->getAUPars().getWienerFilterParameters().getUseFilterFlag() )
  {
    uiEncFrameSize += cExtTabCabaEncoder.getDefaultCabacBitstream().size();
    rcBitstreamData.push_front( std::vector<UChar>() );
    rcBitstreamData.front().swap( cExtTabCabaEncoder.getDefaultCabacBitstream() );
  }

  uiEncFrameSize += aucHeaderBitstream.size();
  rcBitstreamData.push_front( std::vector<UChar>() );
  rcBitstreamData.front().swap( cBitWriterHeader.getVectorRef() );
  cBitWriterHeader.setVector( cBitWriterHeader.getVectorRef() );

  if( rcAUPars.getRefPicFlag() )
  {
    UInt uiShift;
    UInt uiMargin;
    ReferencePicture<Pel>::UpsMethod cUpsMethod;

    pcAU->getAUPars().getInterpolationParameters( uiShift, uiMargin, cUpsMethod, false );

    //th potential cubic problem
    UInt uiExtraMargin = 2*((rcAUEstPar.getSearchRange()+3)/2);
    bool  bStoreDecPic  = ( rcAUEstPar.getSearchModeFullPel() == 2 );
    bool  bStoreNorms   = ( rcAUEstPar.getSearchModeFullPel() != 2 );
    ReferencePictureEnc<Pel>* pcRefPicture = new ReferencePictureEnc<Pel>( rcAUPars.getPicOrderCnt(), rcAUPars.getAUStats(), uiShift, uiMargin, uiExtraMargin, bStoreDecPic, bStoreNorms, cUpsMethod );
    for( UInt uiRuns = 0; uiRuns < rcAUPars.getNumPlanePacks(); uiRuns++ )
    {
      PicBlock<Pel> cEncPlanePackage;
      RNOK( pcAU->extractPlanePackagePicBlock( uiRuns, rcEnc, cEncPlanePackage ) );
      RNOK( pcRefPicture->assignPlanePackage( pcAU->getPlanePackage( uiRuns )->getPars().getPrimPlaneId(), cEncPlanePackage, cUpsMethod, rcAUEstPar.getSearchRange() ) );
    }
    RNOK( pcAU->setRefPicture( pcRefPicture ) );
  }

  RNOK( xCalcPSNR( rcOrig, rcEnc, radPSNR ) );

  TimeStamp cEnd;
  cEnd.stamp();
  UInt64 uiDuration = (cEnd - cStart).getMSec();

  ruiEncodingDurationMSec = uiDuration;
  ruiMEDurationMSec       = uiTotalMEDuration;

  RNOK( cNalUnitEncoder.closeNalUnit( rcBitstreamData ) );
  ROF( rcBitstreamData.size() == 1 );
  uiEncFrameSize = rcBitstreamData.front().size();

  rcOut << " " << std::setw( 9 ) << uiEncFrameSize;
  rcOut << " " << std::fixed << std::setw( 9 ) << std::setprecision( 2 ) << double( uiDuration        ) / 1000.0;
  rcOut << " " << std::fixed << std::setw( 9 ) << std::setprecision( 2 ) << double( uiTotalMEDuration ) / 1000.0;

  for( UInt uiPlane = 0; uiPlane < rcOrig.getStats().getNumPlanes(); uiPlane++ )
  {
    rcOut << "  " << std::fixed << std::setprecision( 4 ) << std::setw( 8 ) << radPSNR[uiPlane];
  }
  rcOut << std::endl;

  delete pcExtTabCabaEncoder;

  return Err::m_nOK;
}


ErrVal AccessUnitEncoder::encodePlanePackage( UInt uiPlaPaId,
                                              const AccessUnit &rcAU,
                                              ContextStore<CabacContextModel> &rcContextStore,
                                              const ContextStore<CabacContextModel> *pcRefCtxStore,
                                              ProbBinStreamEncIf &rcProbBinOut,
                                              CabaCoderStartFinishIf &rcCabaEncoder,
                                              BitWriter<UChar,UInt> &rcBitWriter,
                                              std::list< std::vector<UChar> > &rcBitstreamData,
                                              TraceEnc *pcTrace )
{
  const AccessUnitParameters& rcAUPars    =  rcAU.getAUPars      ();
  const PlanePackage&         rcPlaPaCurr = *rcAU.getPlanePackage( uiPlaPaId );

  SyntaxElementProcessingChainEnc<false> cFinalSEProcChainEnc( NULL, &rcProbBinOut, rcContextStore.getCtxIndices(), rcContextStore.getCtxInit(), rcPlaPaCurr.getPars(), rcPlaPaCurr.getCurrentSlice().getSliceQps()[0], pcTrace );

  if( rcAUPars.getEntropySliceFlag() )
  {
    CabacContextModel *pacCtxStates;

    if( !pcRefCtxStore && !rcContextStore.isCtxInitStates() )
    {
      RNOK( cFinalSEProcChainEnc.getProbAlloc().getCtxStates( pacCtxStates ) );
      RNOK( rcContextStore.setCtxInitStates( pacCtxStates ) );
    }

    for( UInt uiSlice = 0; uiSlice < rcPlaPaCurr.getCurrentSlice().getQuadtreeGroupStructure().getNumEntropySlices(); uiSlice++ )
    {
      TOF( rcBitWriter.getVectorRef().empty() );
      RNOK( rcCabaEncoder.startCabac() );

      if( pcRefCtxStore )
      {
        RNOK( pcRefCtxStore->getCtxStates( uiSlice, pacCtxStates ) );
      }
      else
      {
        RNOK( rcContextStore.getCtxStates( uiSlice, pacCtxStates ) );
      }

      RNOK( cFinalSEProcChainEnc.getProbAlloc().setCtxStates( pacCtxStates ) );
      SliceEncoder cSliceEncoderFinal( rcAUPars, rcPlaPaCurr.getPars(), rcPlaPaCurr.getCurrentSlice(), &cFinalSEProcChainEnc.getSyntaxElementEncoderIf(), uiSlice, pcTrace );
      RNOK( cSliceEncoderFinal.processTree() );
      RNOK( rcCabaEncoder.finishCabac() );
      rcBitstreamData.push_back( std::vector<UChar>() );
      rcBitstreamData.back().swap( rcBitWriter.getVectorRef() );
      rcBitWriter.setVector( rcBitWriter.getVectorRef() );
      RNOK( cFinalSEProcChainEnc.getProbAlloc().getCtxStates( pacCtxStates ) );
      RNOK( rcContextStore.setCtxStates( uiSlice, pacCtxStates ) );
    }
  }
  else
  {
    SliceEncoder cSliceEncoderFinal( rcAUPars, rcPlaPaCurr.getPars(), rcPlaPaCurr.getCurrentSlice(), &cFinalSEProcChainEnc.getSyntaxElementEncoderIf(), 0, pcTrace );
    RNOK( cSliceEncoderFinal.processTree() );
  }
  return Err::m_nOK;
}


ErrVal AccessUnitEncoder::xPrintQpAndFrameType( std::ostream &rcOut,
                                                const PlanePackage &rcPlaPaCurr,
                                                const AccessUnitParameters &rcAUPars,
                                                bool bTestIntraModes,
                                                bool bTestInterModes )
{
  rcOut << std::setw( 4 ) << rcPlaPaCurr.getCurrentSlice().getSliceQps()[0] << "  ";
  rcOut << ( bTestIntraModes ? ( rcAUPars.getRefPicFlag() ? "I" : "i" ) : " " );
  if( bTestInterModes )
  {
    char cFrameType = rcAUPars.getMaxMvsPerBlk() > 2 ? 'N' : (rcAUPars.getMaxMvsPerBlk() > 1 ? 'B' : 'P');
    if( !rcAUPars.getRefPicFlag() )
    {
      cFrameType = tolower( cFrameType );
    }
    rcOut << cFrameType;
  }
  else
  {
    rcOut << " ";
  }
  rcOut << "   ";
  return Err::m_nOK;
}


ErrVal AccessUnitEncoder::xCalcPSNR( const PicBlock<const Pel> &rcOrig,
                                     const PicBlock<const Pel> &rcEnc,
                                     std::vector<double> &radPSNR )
{
  ROF( rcOrig.getStats().getNumPlanes() == rcEnc.getStats().getNumPlanes() );

  radPSNR.resize( rcOrig.getStats().getNumPlanes() );
  for( UInt uiPlane = 0; uiPlane < rcOrig.getStats().getNumPlanes(); uiPlane++ )
  {
    const Block<const Pel>& rcOrigBlock = rcOrig.getPlane( uiPlane );
    const Block<const Pel>& rcEncBlock  = rcEnc.getPlane( uiPlane );
    UInt64 uiSSD = 0;
    for( int iY = 0; iY < rcOrigBlock.getSize().getY(); iY++ )
    {
      for( int iX = 0; iX < rcOrigBlock.getSize().getX(); iX++ )
      {
        const int iDiff = int( rcOrigBlock.getEl( PixelPos( iX, iY ) ) ) - int( rcEncBlock.getEl( PixelPos( iX, iY ) ) );
        uiSSD += iDiff * iDiff;
      }
    }
    const double dMSE = double( uiSSD ) / ( rcOrigBlock.getSize().getX() * rcOrigBlock.getSize().getY() );
    const double dPSNR = 10 * log10( 255. * 255. / dMSE );
    radPSNR[uiPlane] = uiSSD != 0 ? dPSNR : 0.0;
  }
  return Err::m_nOK;
}

ErrVal AccessUnitEncoder::xEstimateInterpolationMethod( Block<double>* &rpcInLoopFiltMatrix,
                                                        UInt           &ruiUpsMethodEst )
{
  int iX, iY;
  double dGain = 0;
  const double dOmega = 0.75, dPi = 3.14159265;

 // compute attenuation due to inloop filter
  for ( iX = 0; iX < rpcInLoopFiltMatrix->getSize().getX(); iX++ )
  {
    for ( iY = 0; iY < rpcInLoopFiltMatrix->getSize().getY(); iY++ )
    {
      dGain += rpcInLoopFiltMatrix->getEl( iX, iY ) * cos( dPi * iX * dOmega  ) * cos( dPi * iY * dOmega  );
    }
  }

  // If attenuation is high, switch to H.264 6-tap fiilter
  if ( 2.0 * dGain < 0.3 )
  {
    ruiUpsMethodEst = 0;
  }

  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END
