/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_CABAENCODER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
#define AFX_CABAENCODER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_

#include "VideoCommonLib/InStreamIf.h"
#include "VideoCommonLib/BitWriterIf.h"
#include "ProbabilityAllocatorEnc.h"
#include "VideoCommonLib/ExtendedCabacTables.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

//!\ingroup VideoEncoderLib
//!\CabaEncoder
#if USE_CABAC12_TABLE_IN_ENCODER
template<class t_Tables = AggregateCabacTables12>
#else
template<class t_Tables = DefaultCabacTables>
#endif
class CabaEncoder
  : public ProbBinStreamEncIf
  , public CabaCoderStartFinishIf
{
public:
  typedef typename t_Tables::Tables TablesType;

  CabaEncoder( BitWriterIf<UInt> &rcBitWriterIf );
  virtual ~CabaEncoder() {}

  ErrVal startCabac();

  ErrVal pushBack( const ProbBin &rcT );
  ErrVal pushBack( const bool &rcT );

  ErrVal finishCabac();

  UInt getNumBinsWritten() const { return m_uiBinsWritten; }
  void setUseShortFinishFlag( bool b ) { m_bUseShortFinish = b; }

  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( CabaEncoder );

private:
  ErrVal xWriteTerminatingBit( UInt uiBit );
  ErrVal xWriteBit( UInt uiBit);
  ErrVal xWriteBitAndBitsToFollow( UInt uiBit);

  BitWriterIf<UInt>* m_pcBitWriterIf;
  UInt m_uiRange;
  UInt m_uiLow;
  UInt m_uiByte;
  UInt m_uiBitsLeft;
  UInt m_uiBitsToFollow;
  UInt m_uiBinsWritten;
  bool m_bUseShortFinish;
};


template<class t_Tables>
FORCEINLINE ErrVal CabaEncoder<t_Tables>::xWriteBit( UInt uiBit)
{
  m_uiByte += m_uiByte + uiBit;
  if( ! --m_uiBitsLeft )
  {
    const UInt uiByte = m_uiByte;
    m_uiBitsLeft = 8;
    m_uiByte     = 0;
    return m_pcBitWriterIf->write( uiByte, 8);
  }
  return Err::m_nOK;
}


template<class t_Tables>
FORCEINLINE ErrVal CabaEncoder<t_Tables>::xWriteBitAndBitsToFollow( UInt uiBit)
{
  RNOK( xWriteBit( uiBit ) );
  // invert bit
  uiBit = 1-uiBit;

  while( m_uiBitsToFollow > 0)
  {
    m_uiBitsToFollow--;
    RNOK( xWriteBit( uiBit ) );
  }
  return Err::m_nOK;
}


template<class t_Tables>
CabaEncoder<t_Tables>::CabaEncoder( BitWriterIf<UInt> &rcBitWriterIf )
: m_pcBitWriterIf  ( &rcBitWriterIf )
, m_uiRange        ( 0 )
, m_uiLow          ( 0 )
, m_uiByte         ( 0 )
, m_uiBitsLeft     ( 0 )
, m_uiBitsToFollow ( 0 )
, m_uiBinsWritten  ( 0 )
, m_bUseShortFinish( false )
{
}


template<class t_Tables>
ErrVal CabaEncoder<t_Tables>::startCabac()
{
  m_uiRange = TablesType::sm_uiHalf - 2;
  m_uiLow = 0;
  m_uiBitsToFollow = 0;
  m_uiByte = 0;
  m_uiBitsLeft = 9;
  m_uiBinsWritten = 0;

  return Err::m_nOK;
}


template<class t_Tables>
ErrVal CabaEncoder<t_Tables>::finishCabac()
{
  if( m_bUseShortFinish )
  {
    UInt uiHigh = m_uiLow + m_uiRange - 1;
    if( uiHigh >= TablesType::sm_uiHalf )
    {
      RNOK( xWriteBitAndBitsToFollow( 1 ) );
      uiHigh -= TablesType::sm_uiHalf;
    }
    else
    {
      RNOK( xWriteBitAndBitsToFollow( 0 ) );
    }
    if( uiHigh >= TablesType::sm_uiQuarter )
    {
      RNOK( xWriteBit( 1 ) );
    }
  }
  else
  {
    RNOK( xWriteTerminatingBit( 1 ) );
    RNOK( xWriteBitAndBitsToFollow( (m_uiLow >> (TablesType::sm_uiBits - 1)) & 1 ) );
    RNOK( xWriteBit(                (m_uiLow >> (TablesType::sm_uiBits - 2)) & 1 ) );
    RNOK( xWriteBit(                1 ) );
  }

  if( 8 - m_uiBitsLeft  != 0 )
  {
    RNOK( m_pcBitWriterIf->write( m_uiByte, 8 - m_uiBitsLeft ) );
  }

  return Err::m_nOK;
}


template<class t_Tables>
ErrVal CabaEncoder<t_Tables>::pushBack( const ProbBin &rcT )
{
  UInt uiIdx = t_Tables::sm_auiStateToIdx[rcT.getStateIdx()];
  if( !TypeCompare<t_Tables,DefaultCabacTables>::m_bEqual && uiIdx == 0 )
  {
    return pushBack( rcT.isLps() );
  }

  UInt uiLPS = TablesType::sm_auiLPSTable[t_Tables::sm_auiIdxToAggregateState[uiIdx]][(m_uiRange>>(6+TablesType::sm_uiDifBits)) & 3];
//  UInt uiLPS = TablesType::sm_auiLPSTable[rcT.getStateIdx()][(m_uiRange>>(6+TablesType::sm_uiDifBits)) & 3];
  ++m_uiBinsWritten;

  m_uiRange -= uiLPS;
  if( rcT.isLps() )
  {
    m_uiLow += m_uiRange;
    m_uiRange = uiLPS;
  }

  while( m_uiRange < TablesType::sm_uiQuarter )
  {
    if( m_uiLow >= TablesType::sm_uiHalf )
    {
      RNOK( xWriteBitAndBitsToFollow( 1 ) );
      m_uiLow -= TablesType::sm_uiHalf;
    }
    else if( m_uiLow < TablesType::sm_uiQuarter )
    {
      RNOK( xWriteBitAndBitsToFollow( 0 ) );
    }
    else
    {
      m_uiBitsToFollow++;
      m_uiLow -= TablesType::sm_uiQuarter;
    }
    m_uiLow   <<= 1;
    m_uiRange <<= 1;
  }

  return Err::m_nOK;
}


template<class t_Tables>
ErrVal CabaEncoder<t_Tables>::pushBack( const bool &rcT )
{
  UInt uiLow = m_uiLow<<1;

  ++m_uiBinsWritten;

  if( rcT )
  {
    uiLow += m_uiRange;
  }

  if (uiLow >= TablesType::sm_uiOne )
  {
    RNOK( xWriteBitAndBitsToFollow( 1 ) );
    uiLow -= TablesType::sm_uiOne;
  }
  else if (uiLow < TablesType::sm_uiHalf )
  {
    RNOK( xWriteBitAndBitsToFollow( 0 ) );
  }
  else
  {
    m_uiBitsToFollow++;
    uiLow -= TablesType::sm_uiHalf;
  }

  m_uiLow = uiLow;
  return Err::m_nOK;
}


template<class t_Tables>
ErrVal CabaEncoder<t_Tables>::xWriteTerminatingBit( UInt uiBit )
{
  UInt uiRange = m_uiRange - 2;
  UInt uiLow   = m_uiLow;

  if( uiBit )
  {
    uiLow += uiRange;
    uiRange = 2;
  }

  while( uiRange < TablesType::sm_uiQuarter )
  {
    if( uiLow >= TablesType::sm_uiHalf )
    {
      RNOK( xWriteBitAndBitsToFollow( 1 ) );
      uiLow -= TablesType::sm_uiHalf;
    }
    else if( uiLow < TablesType::sm_uiQuarter )
    {
      RNOK( xWriteBitAndBitsToFollow( 0 ) );
    }
    else
    {
      m_uiBitsToFollow++;
      uiLow -= TablesType::sm_uiQuarter;
    }
    uiLow   <<= 1;
    uiRange <<= 1;
  }

  m_uiRange = uiRange;
  m_uiLow   = uiLow;

  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_CABAENCODER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
