/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _FASTFULLSEARCH_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _FASTFULLSEARCH_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "MotionSearchIf.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

#define MAX_4X4_BLOCKS 256

template<UInt t_uiArraySize>
class MvCostArray
{
public:
  MvCostArray() : m_pucZero( &m_aucCost[t_uiArraySize-1] )
  {
    m_pucZero[0] = 1;
    for(UInt uiNumBits = 3; true; uiNumBits += 2 )
    {
      UInt uiStop = 1 << (uiNumBits >> 1);
      UInt uiStart = uiStop >> 1;

      uiStop = gMin( t_uiArraySize, uiStop );

      for( UInt ui = uiStart; ui < uiStop; ui++ )
      {
        m_pucZero[ui] = uiNumBits;
        m_pucZero[-int(ui)] = uiNumBits;
      }

      if( uiStop == t_uiArraySize )
      {
        break;
      }
    }
  }
  
  __inline UChar getCostChecked( UInt ui ) const { return m_pucZero[gMin( t_uiArraySize - 1, ui )]; }
  __inline UChar getCostInt( int i ) const { return m_pucZero[i]; }
  __inline UChar getCostIntChecked( int i ) const
  {
#if 0
    // fast under Linux
    const UInt uiAbs = abs( i );
    if( uiAbs >= t_uiArraySize )
    {
      return m_pucZero[t_uiArraySize - 1];
    }
    return m_pucZero[uiAbs];
#else
    // fast under Windows
    if( UInt( abs(i) ) >= t_uiArraySize )
    {
      return m_pucZero[t_uiArraySize - 1];
    }
    return m_pucZero[i];
#endif
  }

private:
  UChar m_aucCost[2 * t_uiArraySize - 1];
  UChar *m_pucZero;
};


class MvMeasureInit
{
public:
  int                       m_iMvScale;
  bool                      m_bMvCompetition;
  std::vector<PrimPixelPos> m_acMvCompOrigins;
  int                       m_iYOrigin;
  std::vector<int>          m_aiYThreshold;
  std::vector<int>          m_aiXOrigin;
};

template<UInt t_uiMaxMv = 256>
class MvMeasure
{
public:
  MvMeasure() : m_bMvCompetition( true ), m_iMvScale( 0 ) { m_acMvCompOrigins.push_back( PrimPixelPos( 0, 0 ) ); }
  virtual ~MvMeasure() {}

  ErrVal init( const MvMeasureInit& rcMvMeasureInit )
  {
    m_iMvScale        = rcMvMeasureInit.m_iMvScale;
    m_bMvCompetition  = rcMvMeasureInit.m_bMvCompetition;
    if( m_bMvCompetition )
    {
      ROT( rcMvMeasureInit.m_acMvCompOrigins.empty() );
      m_acMvCompOrigins = rcMvMeasureInit.m_acMvCompOrigins;
    }
    else
    {
      ROT( rcMvMeasureInit.m_aiYThreshold.empty() );
      ROF( rcMvMeasureInit.m_aiYThreshold.size () == rcMvMeasureInit.m_aiXOrigin.size() );
      m_iYOrigin        = rcMvMeasureInit.m_iYOrigin;
      m_iLstSizeMinus1  = int( rcMvMeasureInit.m_aiYThreshold.size() ) - 1;
      m_aiYThreshold    = rcMvMeasureInit.m_aiYThreshold;
      m_aiXOrigin       = rcMvMeasureInit.m_aiXOrigin;
    }
    return Err::m_nOK;
  }

  __inline BitRate getRate( int iMvCompX, int iMvCompY )  const
  {
    int iMvDCost = TypeLimits<int>::m_iMax;
    if( m_bMvCompetition )
    {
      for( UInt uiIdx = 0; uiIdx < m_acMvCompOrigins.size(); uiIdx++ )
      {
        const int iMvDX = ( iMvCompX << m_iMvScale) + m_acMvCompOrigins[uiIdx].getX();
        const int iMvDY = ( iMvCompY << m_iMvScale) + m_acMvCompOrigins[uiIdx].getY();
        const int iCost = int( m_cCost.getCostIntChecked( iMvDX ) ) + int( m_cCost.getCostIntChecked( iMvDY ) );
        iMvDCost        = gMin<int>( iMvDCost, iCost );
      }
    }
    else
    {
      const int iMvX      = ( iMvCompX << m_iMvScale );
      const int iMvY      = ( iMvCompY << m_iMvScale );
      int       iXOrigin  = m_aiXOrigin [ m_iLstSizeMinus1 ];
      for(  int iLstIdx   = 0;  iLstIdx < m_iLstSizeMinus1; iLstIdx++ )
      {
        if( iMvY  <= m_aiYThreshold[ iLstIdx ] )
        {
          iXOrigin = m_aiXOrigin   [ iLstIdx ];
          break;
        }
      }
      const int iMvDX     = iMvX + iXOrigin;
      const int iMvDY     = iMvY + m_iYOrigin;
      iMvDCost            = int( m_cCost.getCostIntChecked( iMvDX ) ) + int( m_cCost.getCostIntChecked( iMvDY ) );
    }
    return iMvDCost;
  }

private:
  const MvCostArray<256>    m_cCost;
  bool                      m_bMvCompetition;
  int                       m_iMvScale;
  std::vector<PrimPixelPos> m_acMvCompOrigins;
  int                       m_iYOrigin;
  int                       m_iLstSizeMinus1;
  std::vector<int>          m_aiYThreshold;
  std::vector<int>          m_aiXOrigin;
};


class SearchScaleParameters // for fast (multiplication-free) conversion with down rounding
{
public:
  SearchScaleParameters( const PrimBlockSize& rcPrimBlkSize, const Subsampling& rcRefSubsampling, const Subsampling& rcOrgSubsampling )
    : m_cBlockSize    ( rcPrimBlkSize * rcRefSubsampling )
    , m_c4x4BlkSize   ( m_cBlockSize >> 2 )
    , m_uiPosShiftX   ( xRatioToShift( rcRefSubsampling.getX() ) )
    , m_uiPosShiftY   ( xRatioToShift( rcRefSubsampling.getY() ) )
    , m_iAddRefShiftX ( xRatioToShift( rcOrgSubsampling.getX() ) - m_uiPosShiftX )
    , m_iAddRefShiftY ( xRatioToShift( rcOrgSubsampling.getY() ) - m_uiPosShiftY )
  {
    AOT( m_iAddRefShiftX < 0 );
    AOT( m_iAddRefShiftY < 0 );
  }

  BlockSize getBlockSize    ()                                const { return m_cBlockSize; }
  BlockSize get4x4BlkSize   ()                                const { return m_c4x4BlkSize; }
  PixelPos  getPixelPos     ( const PrimPixelPos& rcPrimPos ) const { return PixelPos( rcPrimPos.getX() >> m_uiPosShiftX, rcPrimPos.getY() >> m_uiPosShiftY ); }
  UInt      getAddRefShiftX ()                                const { return (UInt)m_iAddRefShiftX; }
  UInt      getAddRefShiftY ()                                const { return (UInt)m_iAddRefShiftY; }

private:
  static UInt xRatioToShift( const ShiftScale& rcRatio )
  {
    return rcRatio.getDownShift();
  }
private:
  BlockSize m_cBlockSize;
  BlockSize m_c4x4BlkSize;
  UInt      m_uiPosShiftX;
  UInt      m_uiPosShiftY;
  int       m_iAddRefShiftX;
  int       m_iAddRefShiftY;
};


template<class t_OrigPel, class t_RefPel, DistortionMeasure t_eDistMeasure> class DistortionComputation;

template<class t_OrigPel, class t_RefPel> class DistortionComputation<t_OrigPel, t_RefPel, DIST_SAD>
{
public:
  static FORCEINLINE bool trueCostIsLower( const PrimPixelPos cCand,
                                           const PicBlock<const t_OrigPel> &rcOrig,
                                           const PicBlock<const t_RefPel> &rcRef,
                                           const std::vector<SearchScaleParameters>& racSearchScalePars,
                                           UInt uiLog2RefStep,
                                           const LambdaTimesRate cBestCost,
                                           LambdaTimesRate &rcCurrCost,
                                           const Int64 *piPseudoSAD4x4RowValues,
                                           UInt &ruiNumEliminated )
  {
    UInt  uiSAD = 0;
    for( UInt uiPlane = 0; uiPlane < racSearchScalePars.size(); uiPlane++, piPseudoSAD4x4RowValues += MAX_4X4_BLOCKS )
    {
      const SearchScaleParameters&  rcSSP     = racSearchScalePars[ uiPlane ];
      const Block<const t_OrigPel>  cOrgBlk   = rcOrig.getPlane( uiPlane );
      const Block<const t_RefPel>   cRefBlk( rcSSP.getPixelPos( cCand ), rcSSP.getBlockSize(), rcRef.getPlane( uiPlane ) );
      const t_OrigPel*              pcBlk1    = &cOrgBlk.getEl( 0 );
      const t_RefPel*               pcBlk2    = &cRefBlk.getEl( 0 );
      const int                     iStride1  = cOrgBlk.getStride();
      const int                     iStride2  = cRefBlk.getStride() << uiLog2RefStep;
      const int                     iHeight   = cOrgBlk.getSize().getY();
      const int                     iWidth    = cOrgBlk.getSize().getX();
      ruiNumEliminated                       += iWidth * iHeight >> 4;

      for( int y = 0; y < iHeight; y++, pcBlk1 += iStride1, pcBlk2 += iStride2 )
      {
        for( int x = 0; x < iWidth; x++ )
        {
          uiSAD += abs( int( pcBlk1[ x ] ) - int( pcBlk2[ x << uiLog2RefStep ] ) );
        }
        if( ( y & 3 ) == 3 )
        {
          ruiNumEliminated -= iWidth >> 2;
          rcCurrCost       -= piPseudoSAD4x4RowValues[ y >> 2 ];
          rcCurrCost       += uiSAD;
          if( cBestCost <= rcCurrCost )
          {
            return false;
          }
          uiSAD = 0;
        }
      }
    }
    return true;
  }

  static FORCEINLINE UInt computeDist( const Block<const t_OrigPel> &rcOrig, const Block<const t_RefPel> &rcRef )
  {
    UInt uiSAD = 0;
    const t_OrigPel*  pcBlk1    = &rcOrig.getEl( 0 );
    const t_RefPel*   pcBlk2    = &rcRef .getEl( 0 );
    const int         iStride1  =  rcOrig.getStride();
    const int         iStride2  =  rcRef .getStride();
    const t_OrigPel*  pcBlk1End = pcBlk1 + rcOrig.getSize().getY() * iStride1;
    const int         iWidth    = rcOrig.getSize().getX();
    for( ; pcBlk1 < pcBlk1End; pcBlk1 += iStride1, pcBlk2 += iStride2 )
    {
      for( int x = 0; x < iWidth; x++ )
      {
        uiSAD += abs( int( pcBlk1[ x ] ) - int( pcBlk2[ x ] ) );
      }
    }
    return uiSAD;
  }

  static FORCEINLINE UInt computeDist( const Block<const t_OrigPel> &rcOrig, const Block<const t_RefPel> &rcRef, UInt uiLog2RefStepX, UInt uiLog2RefStepY )
  {
    UInt uiSAD = 0;
    const t_OrigPel*  pcBlk1    = &rcOrig.getEl( 0 );
    const t_RefPel*   pcBlk2    = &rcRef .getEl( 0 );
    const int         iStride1  =  rcOrig.getStride();
    const int         iStride2  =  rcRef .getStride() << uiLog2RefStepY;
    const t_OrigPel*  pcBlk1End = pcBlk1 + rcOrig.getSize().getY() * iStride1;
    const int         iWidth    = rcOrig.getSize().getX();
    for( ; pcBlk1 < pcBlk1End; pcBlk1 += iStride1, pcBlk2 += iStride2 )
    {
      for( int x = 0; x < iWidth; x++ )
      {
        uiSAD += abs( int( pcBlk1[ x ] ) - int( pcBlk2[ x << uiLog2RefStepX ] ) );
      }
    }
    return uiSAD;
  }
};

template<class t_OrigPel, class t_RefPel> class DistortionComputation<t_OrigPel, t_RefPel, DIST_SSD>
{
public:
  static __inline UInt computeDist( const Block<const t_OrigPel> &rcOrig, const Block<const t_RefPel> &rcRef, UInt uiLog2RefStepX, UInt uiLog2RefStepY )
  {
    UInt uiDist = 0;
    const t_OrigPel*  pcBlk1    = &rcOrig.getEl( 0 );
    const t_RefPel*   pcBlk2    = &rcRef .getEl( 0 );
    const int         iStride1  =  rcOrig.getStride();
    const int         iStride2  =  rcRef .getStride() << uiLog2RefStepY;
    const t_OrigPel*  pcBlk1End = pcBlk1 + rcOrig.getSize().getY() * iStride1;
    const int         iWidth    = rcOrig.getSize().getX();
    for( ; pcBlk1 < pcBlk1End; pcBlk1 += iStride1, pcBlk2 += iStride2 )
    {
      for( int x = 0; x < iWidth; x++ )
      {
        const int iDiff =  int( pcBlk1[ x ] ) - int( pcBlk2[ x << uiLog2RefStepX ] );
        uiDist         += UInt( iDiff * iDiff );
      }
    }
    return uiDist;
  }

  static __inline UInt computeDist( const Block<const t_OrigPel> &rcOrig, const Block<const t_RefPel> &rcRef )
  {
    UInt uiDist = 0;
    const t_OrigPel*  pcBlk1    = &rcOrig.getEl( 0 );
    const t_RefPel*   pcBlk2    = &rcRef .getEl( 0 );
    const int         iStride1  =  rcOrig.getStride();
    const int         iStride2  =  rcRef .getStride();
    const t_OrigPel*  pcBlk1End = pcBlk1 + rcOrig.getSize().getY() * iStride1;
    const int         iWidth    = rcOrig.getSize().getX();
    for( ; pcBlk1 < pcBlk1End; pcBlk1 += iStride1, pcBlk2 += iStride2 )
    {
      for( int x = 0; x < iWidth; x++ )
      {
        const int iDiff =  int( pcBlk1[ x ] ) - int( pcBlk2[ x ] );
        uiDist         += UInt( iDiff * iDiff );
      }
    }
    return uiDist;
  }
};


template<DistortionMeasure t_eDistMeasure, class t_MvMeasure, class t_OrigPel, class t_RefPel>
class FullSearch
  : public t_MvMeasure
{
public:
  __inline ErrVal searchMvFullPel( UInt uiSearchMode,
                                   const PicBlock<const t_OrigPel> &rcOrig,
                                   const PicBlock<const t_RefPel> &rcRef,
                                   PrimPixelPos &rcBestPos,
                                   const Lambda cLambda,
                                   LambdaTimesRate &rcBestCostVal,
                                   UInt &ruiNumEliminated, UInt &ruiTotal,
                                   const MvMeasureInit& rcMvMeasureInit,
                                   const PicBlock<const UInt> &rcNorms );
  __inline bool searchMV( const PicBlock<const t_OrigPel> &rcOrig,
                          const PicBlock<const t_RefPel> &rcRef,
                          PrimPixelPos &rcBestPos,
                          const Lambda cLambda,
                          LambdaTimesRate &rcBestCostVal,
                          const MvMeasureInit& rcMvMeasureInit,
                          const UInt uiLog2RefStep );

protected:
  __inline ErrVal fullPelBlockSearch  ( const PicBlock<const t_OrigPel> &rcOrig,
                                        const PicBlock<const t_RefPel> &rcRef,
                                        PrimPixelPos &rcBestPos,
                                        const Lambda cLambda,
                                        LambdaTimesRate &rcBestCostVal,
                                        UInt &ruiNumEliminated,
                                        UInt &ruiTotal,
                                        const PicBlock<const UInt> &rcNorms,
                                        const std::vector<SearchScaleParameters>& racSearchScalePars,
                                        const PicBlock<const UInt>& rcOrig4x4L1Norm );
  __inline ErrVal fullPelSpiralSearch ( const PicBlock<const t_OrigPel> &rcOrig,
                                        const PicBlock<const t_RefPel> &rcRef,
                                        PrimPixelPos &rcBestPos,
                                        const Lambda cLambda,
                                        LambdaTimesRate &rcBestCostVal,
                                        UInt &ruiNumEliminated,
                                        UInt &ruiTotal,
                                        const PicBlock<const UInt> &rcNorms,
                                        const std::vector<SearchScaleParameters>& racSearchScalePars,
                                        const PicBlock<const UInt>& rcOrig4x4L1Norm );

private:
  static FORCEINLINE bool xNeedsTrueRDCost( const PrimPixelPos cPos,
                                            const std::vector<SearchScaleParameters>& racSearchScalePars,
                                            const PicBlock<const UInt> &rcRefNorms,
                                            const PicBlock<const UInt> &rcOrgNorms,
                                            const LambdaTimesRate cBestCost,
                                            LambdaTimesRate &rcCurrLambdaTimesRate,
                                            Int64* piPseudoSAD4x4RowValues )
  {
    for( UInt uiPlane = 0; uiPlane < racSearchScalePars.size(); uiPlane++, piPseudoSAD4x4RowValues += MAX_4X4_BLOCKS )
    {
      const SearchScaleParameters&  rcSSP         = racSearchScalePars  [ uiPlane ];
      const Block<const UInt>&      rcOrgNormBlk  = rcOrgNorms.getPlane ( uiPlane );
      const Block<const UInt>&      rcRefNormBlk  = rcRefNorms.getPlane ( uiPlane );
      const PixelPos                cRefBasePos   = rcSSP.getPixelPos   ( cPos );
      for( int y = 0; y < rcSSP.get4x4BlkSize().getY(); y++ )
      {
        piPseudoSAD4x4RowValues[ y ] = 0;
        for( int x = 0; x < rcSSP.get4x4BlkSize().getX(); x++ )
        {
          const PixelPos  c4x4Blk( x, y );
          const PixelPos  cRefNormPos   = cRefBasePos + ( c4x4Blk << 2 );
          const UInt      uiPseudoSAD   = abs( int( rcOrgNormBlk.getEl( c4x4Blk ) ) - int( rcRefNormBlk.getEl( cRefNormPos ) ) );
          rcCurrLambdaTimesRate        += uiPseudoSAD;
          piPseudoSAD4x4RowValues[ y ] += uiPseudoSAD;
          ROTRS( cBestCost < rcCurrLambdaTimesRate, false );
        }
      }
    }
    return true;
  }

public:
  static FORCEINLINE bool xCheckCandidate( const PrimPixelPos cCand,
                                           const PicBlock<const t_OrigPel> &rcOrig,
                                           const PicBlock<const t_RefPel> &rcRef,
                                           const std::vector<SearchScaleParameters>& racSearchScalePars,
                                           const PicBlock<const UInt> &rcRefNorms,
                                           const PicBlock<const UInt> &rcOrgNorms,
                                           LambdaTimesRate cCurrLambdaTimesRate,
                                           Int64* piPseudoSAD4x4RowValues,
                                           LambdaTimesRate &rcBestCost,
                                           PrimPixelPos &rcBestPos )
  {
    if( xNeedsTrueRDCost( cCand, racSearchScalePars, rcRefNorms, rcOrgNorms, rcBestCost, cCurrLambdaTimesRate, piPseudoSAD4x4RowValues ) )
    {
      UInt uiDummy = 0;
      if( DistortionComputation<t_OrigPel,t_RefPel,t_eDistMeasure>::trueCostIsLower( cCand, rcOrig, rcRef, racSearchScalePars, 0, rcBestCost, cCurrLambdaTimesRate, piPseudoSAD4x4RowValues, uiDummy ) )
      {
        rcBestCost  = cCurrLambdaTimesRate;
        rcBestPos   = cCand;
      }
      return false;
    }
    return true;
  }
};


template<DistortionMeasure t_eDistMeasure, class t_MvMeasure, class t_OrigPel, class t_RefPel>
FORCEINLINE ErrVal 
FullSearch<t_eDistMeasure,t_MvMeasure,t_OrigPel,t_RefPel>::searchMvFullPel( UInt uiSearchMode,
                                                                            const PicBlock<const t_OrigPel> &rcOrig,
                                                                            const PicBlock<const t_RefPel> &rcRef,
                                                                            PrimPixelPos &rcBestPos,
                                                                            const Lambda cLambda,
                                                                            LambdaTimesRate &rcBestCostVal,
                                                                            UInt &ruiNumEliminated,
                                                                            UInt &ruiTotal,
                                                                            const MvMeasureInit& rcMvMeasureInit,
                                                                            const PicBlock<const UInt> &rcNorms )
{
  t_MvMeasure::init( rcMvMeasureInit );
  CAOF( t_eDistMeasure == DIST_SAD );
  for( UInt uiPlane = 0; uiPlane < rcRef.getStats().getNumPlanes(); uiPlane++ )
  {
    ROF( rcOrig.getPlane( uiPlane ).getSize().getX() % 4 == 0 && rcOrig.getPlane( uiPlane ).getSize().getY() % 4 == 0 );
    ROF( rcOrig.getStats().getSubsampling()[ uiPlane ] == rcRef.getStats().getSubsampling()[ uiPlane ] );
  }

  const PrimBlockSize cSizeInRef    = rcOrig.getStats().getPrimBlockSize();
  const PrimBlockSize cOrig4x4Size  = rcOrig.getStats().getPrimBlockSize() >> 2;

  std::vector<SearchScaleParameters> acSearchScalePars;
  for( UInt uiPlane = 0; uiPlane < rcRef.getStats().getNumPlanes(); uiPlane++ )
  {
    acSearchScalePars.push_back( SearchScaleParameters( cSizeInRef, rcRef.getStats().getSubsampling()[ uiPlane ], rcOrig.getStats().getSubsampling()[ uiPlane ] ) );
  }

  AllocPicBlock<UInt> cOrig4x4L1Norm( PicBlockStats( cOrig4x4Size, rcRef.getStats().getSubsampling() ) );
  for( UInt uiPlane = 0; uiPlane < acSearchScalePars.size(); uiPlane++ )
  {
    const Block<const t_OrigPel>& rcOrigBlk = rcOrig        .getPlane( uiPlane );
    const Block<UInt>&            rcNormBlk = cOrig4x4L1Norm.getPlane( uiPlane );
    for( int y = 0; y < acSearchScalePars[ uiPlane ].get4x4BlkSize().getY(); y++ )
    {
      for( int x = 0; x < acSearchScalePars[ uiPlane ].get4x4BlkSize().getX(); x++ )
      {
        UInt uiBlockNorm = 0;
        for( int yy = 0; yy < 4; yy++ )
        {
          for( int xx = 0; xx < 4; xx++ )
          {
            uiBlockNorm += abs( rcOrigBlk.getEl( PixelPos( ( x << 2 ) + xx, ( y << 2 ) + yy ) ) );
          }
        }
        rcNormBlk.getEl( PixelPos( x, y ) ) = uiBlockNorm;
      }
    }
  }

  if( uiSearchMode == 0 )
  {
    RNOK( fullPelBlockSearch  ( rcOrig, rcRef, rcBestPos, cLambda, rcBestCostVal, ruiNumEliminated, ruiTotal, rcNorms, acSearchScalePars, cOrig4x4L1Norm ) );
  }
  else if( uiSearchMode == 1 )
  {
    RNOK( fullPelSpiralSearch ( rcOrig, rcRef, rcBestPos, cLambda, rcBestCostVal, ruiNumEliminated, ruiTotal, rcNorms, acSearchScalePars, cOrig4x4L1Norm ) );
  }
  else
  {
    ROT(1);
  }
  return Err::m_nOK;
}


template<DistortionMeasure t_eDistMeasure, class t_MvMeasure, class t_OrigPel, class t_RefPel>
__inline ErrVal 
FullSearch<t_eDistMeasure,t_MvMeasure,t_OrigPel,t_RefPel>::fullPelBlockSearch( const PicBlock<const t_OrigPel> &rcOrig,
                                                                               const PicBlock<const t_RefPel> &rcRef,
                                                                               PrimPixelPos &rcBestPos,
                                                                               const Lambda cLambda,
                                                                               LambdaTimesRate &rcBestCostVal,
                                                                               UInt &ruiNumEliminated,
                                                                               UInt &ruiTotal,
                                                                               const PicBlock<const UInt> &rcNorms,
                                                                               const std::vector<SearchScaleParameters>& racSearchScalePars,
                                                                               const PicBlock<const UInt>& rcOrig4x4L1Norm )
{
  Int64 aiPseudoSAD4x4RowValues[ 4 * MAX_4X4_BLOCKS ]; // Only works for block heights up to 1024 and up to 4 planes
  ROF( racSearchScalePars.size()                            <= 4 );
  ROF( rcOrig4x4L1Norm.getStats().getPrimBlockSize().getY() <= MAX_4X4_BLOCKS );

  const int iYMax = rcRef.getStats().getPrimBlockSize().getY() - rcOrig.getStats().getPrimBlockSize().getY() + 1;
  const int iXMax = rcRef.getStats().getPrimBlockSize().getX() - rcOrig.getStats().getPrimBlockSize().getX() + 1;
  ruiTotal       += iXMax * iYMax;

  //===== init cost with center positition =====
  rcBestCostVal = cLambda * t_MvMeasure::getRate( rcBestPos.getX(), rcBestPos.getY() );
  for( UInt uiPlane = 0; uiPlane < racSearchScalePars.size(); uiPlane++ )
  {
    const SearchScaleParameters&  rcSSP    = racSearchScalePars[ uiPlane ];  
    Block<const int>              cRefStart( rcSSP.getPixelPos( rcBestPos ), rcSSP.getBlockSize(), rcRef.getPlane( uiPlane ) );
    rcBestCostVal += DistortionComputation<t_OrigPel,t_RefPel,t_eDistMeasure>::computeDist( rcOrig.getPlane( uiPlane ), cRefStart );
  }
  //===== loop over search positions in raster scan order =====
  for( int y = 0; y < iYMax; y++ )
  {
    for( int x = 0; x < iXMax; x++ )
    {
      if( xCheckCandidate( PrimPixelPos( x, y ),
                           rcOrig,
                           rcRef,
                           racSearchScalePars,
                           rcNorms,
                           rcOrig4x4L1Norm,
                           cLambda * t_MvMeasure::getRate( x, y ),
                           aiPseudoSAD4x4RowValues,
                           rcBestCostVal,
                           rcBestPos ) )
      {
        ruiNumEliminated++;
      }
    }
  }
  return Err::m_nOK;
}


template<DistortionMeasure t_eDistMeasure, class t_MvMeasure, class t_OrigPel, class t_RefPel>
__inline ErrVal 
FullSearch<t_eDistMeasure,t_MvMeasure,t_OrigPel,t_RefPel>::fullPelSpiralSearch( const PicBlock<const t_OrigPel> &rcOrig,
                                                                                const PicBlock<const t_RefPel> &rcRef,
                                                                                PrimPixelPos &rcBestPos,
                                                                                const Lambda cLambda,
                                                                                LambdaTimesRate &rcBestCostVal,
                                                                                UInt &ruiNumEliminated,
                                                                                UInt &ruiTotal,
                                                                                const PicBlock<const UInt> &rcNorms,
                                                                                const std::vector<SearchScaleParameters>& racSearchScalePars,
                                                                                const PicBlock<const UInt>& rcOrig4x4L1Norm )
{
  Int64 aiPseudoSAD4x4RowValues[ 4 * MAX_4X4_BLOCKS ]; // Only works for block heights up to 1024 and up to 4 planes
  ROF( racSearchScalePars.size()                            <= 4 );
  ROF( rcOrig4x4L1Norm.getStats().getPrimBlockSize().getY() <= MAX_4X4_BLOCKS );

  const int iYMax = rcRef.getStats().getPrimBlockSize().getY() - rcOrig.getStats().getPrimBlockSize().getY() + 1;
  const int iXMax = rcRef.getStats().getPrimBlockSize().getX() - rcOrig.getStats().getPrimBlockSize().getX() + 1;
  ruiTotal       += iXMax * iYMax;

  PrimPixelPos  cCurrPos        = rcBestPos;
  int           iToRightLimit   = rcBestPos.getX() + 1;
  int           iToBottomLimit  = rcBestPos.getY() + 1;
  int           iToLeftLimit    = rcBestPos.getX() - 1;
  int           iToTopLimit     = rcBestPos.getY();
  bool          bDone           = false;
  while( !bDone )
  {
    bDone = true;
    if( iToTopLimit >= 0 )
    {
      for( ; cCurrPos.getX() < iToRightLimit; cCurrPos.getX()++ )
      {
        if( xCheckCandidate( cCurrPos, rcOrig, rcRef, racSearchScalePars, rcNorms, rcOrig4x4L1Norm,
                             cLambda * t_MvMeasure::getRate( cCurrPos.getX(), cCurrPos.getY() ),
                             aiPseudoSAD4x4RowValues, rcBestCostVal, rcBestPos ) )
        {
          ruiNumEliminated++;
        }
      }
      iToTopLimit--;
      bDone = false;
    }
    else
    {
      cCurrPos.getX() = iToRightLimit;
      cCurrPos.getY() = 0;
    }

    if( iToRightLimit < iXMax )
    {
      for( ; cCurrPos.getY() < iToBottomLimit; cCurrPos.getY()++ )
      {
        if( xCheckCandidate( cCurrPos, rcOrig, rcRef, racSearchScalePars, rcNorms, rcOrig4x4L1Norm,
                             cLambda * t_MvMeasure::getRate( cCurrPos.getX(), cCurrPos.getY() ),
                             aiPseudoSAD4x4RowValues, rcBestCostVal, rcBestPos ) )
        {
          ruiNumEliminated++;
        }
      }
      iToRightLimit++;
      bDone = false;
    }
    else
    {
      cCurrPos.getX() = iXMax - 1;
      cCurrPos.getY() = iToBottomLimit;
    }

    if( iToBottomLimit < iYMax )
    {
      for( ; cCurrPos.getX() > iToLeftLimit; cCurrPos.getX()-- )
      {
        if( xCheckCandidate( cCurrPos, rcOrig, rcRef, racSearchScalePars, rcNorms, rcOrig4x4L1Norm,
                             cLambda * t_MvMeasure::getRate( cCurrPos.getX(), cCurrPos.getY() ),
                             aiPseudoSAD4x4RowValues, rcBestCostVal, rcBestPos ) )
        {
          ruiNumEliminated++;
        }
      }
      iToBottomLimit++;
      bDone = false;
    }
    else
    {
      cCurrPos.getX() = iToLeftLimit;
      cCurrPos.getY() = iYMax - 1;
    }

    if( iToLeftLimit >= 0 )
    {
      for( ; cCurrPos.getY() > iToTopLimit; cCurrPos.getY()-- )
      {
        if( xCheckCandidate( cCurrPos, rcOrig, rcRef, racSearchScalePars, rcNorms, rcOrig4x4L1Norm,
                             cLambda * t_MvMeasure::getRate( cCurrPos.getX(), cCurrPos.getY() ),
                             aiPseudoSAD4x4RowValues, rcBestCostVal, rcBestPos ) )
        {
          ruiNumEliminated++;
        }
      }
      iToLeftLimit--;
      bDone = false;
    }
    else
    {
      cCurrPos.getX() = 0;
      cCurrPos.getY() = iToTopLimit;
    }
  }
  return Err::m_nOK;
}


template<DistortionMeasure t_eDistMeasure, class t_MvMeasure, class t_OrigPel, class t_RefPel>
FORCEINLINE bool 
FullSearch<t_eDistMeasure,t_MvMeasure,t_OrigPel,t_RefPel>::searchMV( const PicBlock<const t_OrigPel> &rcOrig,
                                                                     const PicBlock<const t_RefPel> &rcRef,
                                                                     PrimPixelPos &rcBestPos,
                                                                     const Lambda cLambda,
                                                                     LambdaTimesRate &rcBestCostVal,
                                                                     const MvMeasureInit& rcMvMeasureInit,
                                                                     const UInt uiLog2RefStep )
{
  t_MvMeasure::init( rcMvMeasureInit );
  bool bFoundNewBestPos = false;

  const PrimBlockSize cSizeInRef    = ( ( rcOrig.getStats().getPrimBlockSize() - PrimBlockSize( 1, 1 ) ) << uiLog2RefStep ) + PrimBlockSize( 1, 1 );
  const int           iYMax         = rcRef .getStats().getPrimBlockSize().getY() - cSizeInRef.getY() + 1;
  const int           iXMax         = rcRef .getStats().getPrimBlockSize().getX() - cSizeInRef.getX() + 1;

  std::vector<SearchScaleParameters> acSearchScalePars;
  for( UInt uiPlane = 0; uiPlane < rcRef.getStats().getNumPlanes(); uiPlane++ )
  {
    acSearchScalePars.push_back( SearchScaleParameters( cSizeInRef, rcRef.getStats().getSubsampling()[ uiPlane ], rcOrig.getStats().getSubsampling()[ uiPlane ] ) );
  }

  for( int y = 0; y < iYMax; y++ )
  {
    for( int x = 0; x < iXMax; x++ )
    {
      const PrimPixelPos  cCurrPos( x, y );
      LambdaTimesRate     cCurrCostVal = cLambda * t_MvMeasure::getRate( x, y );
      for( UInt uiPlane = 0; uiPlane < acSearchScalePars.size(); uiPlane++ )
      {
        const SearchScaleParameters&  rcSSP = acSearchScalePars[ uiPlane ];
        Block<const int>              cRef( rcSSP.getPixelPos( cCurrPos ), rcSSP.getBlockSize(), rcRef.getPlane( uiPlane ) );
        cCurrCostVal += DistortionComputation<t_OrigPel,t_RefPel,t_eDistMeasure>::computeDist( rcOrig.getPlane( uiPlane ), cRef, 
                                                                                               uiLog2RefStep + rcSSP.getAddRefShiftX(),
                                                                                               uiLog2RefStep + rcSSP.getAddRefShiftY() );
      }
      if( cCurrCostVal < rcBestCostVal )
      {
        bFoundNewBestPos  = true;
        rcBestCostVal     = cCurrCostVal;
        rcBestPos         = cCurrPos;
      }
    }
  }
  return bFoundNewBestPos;
}


template<class t_OrigPel,DistortionMeasure t_eDistMeasure>
class FastFullSearch
  : public MotionSearchIf<t_OrigPel>
{
public:
  FastFullSearch()
    : m_uiNumEliminatedFull( 0 )
    , m_uiTotalFull( 0 ) {}

  virtual ~FastFullSearch()
  {
/*    if( m_uiTotalFull )
    {
      std::cout << std::endl << "Triangle inequality elimination ratio (Full-Pel): " << double( m_uiNumEliminatedFull ) / double( m_uiTotalFull ) << std::endl;
    }*/
  }

  virtual double getTriangleInequalityEliminationRatio() const { return m_uiTotalFull == 0 ? 0.0 : double( m_uiNumEliminatedFull ) / double( m_uiTotalFull ); }

  virtual ErrVal searchMotion( int iNumPlanesToSearch,
                               const std::vector<SearchControl> &racSearchControl,
                               const QTNavi &rcNav,
                               const SliceData &rcSliceData,
                               const PicBlock<const t_OrigPel> &rcOrigPicBlock,
                               const ReferencePlanePackage<Pel> &rcRefPlaPa,
                               const Lambda cLambda,
                               const MvMeasureInit&  rcMvMeasureInitInput,
                               MotionVector &rcBestPos,
                               const std::vector<MotionVector>& rcCandMvList );

private:
  ErrVal xSearchMvFixedPrecision( int iNumPlanesToSearch,
                                  const PicBlock<const t_OrigPel> &rcOrigPicBlock,
                                  const ReferencePlanePackage<Pel> &rcRefPlaPa,
                                  const MotionVector &rcPosVecToSearchWindow,
                                  const MvMeasureInit&  rcMvMeasureInitInput,
                                  const SearchControl& rcSearchControl,
                                  const Lambda cLambda,
                                  LambdaTimesRate &rcBestCostVal,
                                  MotionVector &rcPosVecToBestPos );

  UInt m_uiNumEliminatedFull;
  UInt m_uiTotalFull;
  FullSearch<t_eDistMeasure,MvMeasure<>,t_OrigPel,int> m_cSearch;
};



template<class t_OrigPel,DistortionMeasure t_eDistMeasure>
ErrVal FastFullSearch<t_OrigPel,t_eDistMeasure>::searchMotion( int iNumPlanesToSearch,
                                                               const std::vector<SearchControl> &racSearchControl,
                                                               const QTNavi &rcNav,
                                                               const SliceData &rcSliceData,
                                                               const PicBlock<const t_OrigPel> &rcOrigPicBlock,
                                                               const ReferencePlanePackage<Pel> &rcRefPlaPa,
                                                               const Lambda cLambda,
                                                               const MvMeasureInit&  rcMvMeasureInitInput,
                                                               MotionVector &rcBestPos,
                                                               const std::vector<MotionVector>& rcCandMvList )
{
  MotionVector cPosVecToBestPos = rcBestPos.plusPrimResPos( rcNav.getPrimPixelPos() );

  LambdaTimesRate cBestCostVal;
  cBestCostVal.setNumerator( TypeLimits<Int64>::m_iMax );

  for( UInt ui = 0; ui < racSearchControl.size(); ui++ )
  {
    const SearchControl &rcCtrl = racSearchControl[ui];
    const ReferencePlanePackage<Pel> &rcRefPlaPaForME = rcRefPlaPa;
    cPosVecToBestPos = cPosVecToBestPos.getMvWithPrecisionIncreasedTo( rcCtrl.getPrecision() );

    MotionVector cPosVecToSearchWindow( cPosVecToBestPos.getX() - rcCtrl.getSearchWindowCenter().getX(),
                                        cPosVecToBestPos.getY() - rcCtrl.getSearchWindowCenter().getY(), cPosVecToBestPos.getPrecision() );
    
    if( rcCtrl.getPrecision() == 0 && iNumPlanesToSearch > 1 ) // decrease vector accuracy if required
    {
      int iPosVecPrecision = 0;
      for( UInt uiPlane = 0; uiPlane < rcOrigPicBlock.getStats().getNumPlanes(); uiPlane++ )
      {
        const Subsampling&  rcSub   = rcOrigPicBlock.getStats().getSubsampling()[ uiPlane ];
        const UInt          uiLog2X = rcSub.getX().getDownShift();
        const UInt          uiLog2Y = rcSub.getY().getDownShift();
        ROF( 0 <= rcSub.getX().getDownShift() );
        ROF( 0 <= rcSub.getY().getDownShift() );
        iPosVecPrecision = gMin( iPosVecPrecision, -int( gMax( uiLog2X, uiLog2Y ) ) );
      }
      cPosVecToSearchWindow = cPosVecToSearchWindow.getMvWithPrecisionDecreasedTo( iPosVecPrecision ).getMvWithPrecisionIncreasedTo( rcCtrl.getPrecision() );
    }

    RNOK( xSearchMvFixedPrecision( iNumPlanesToSearch,
                                   rcOrigPicBlock,
                                   rcRefPlaPaForME,
                                   cPosVecToSearchWindow,
                                   rcMvMeasureInitInput,
                                   rcCtrl,
                                   cLambda,
                                   cBestCostVal,
                                   cPosVecToBestPos ) );
  }
  rcBestPos = cPosVecToBestPos.minusPrimResPos( rcNav.getPrimPixelPos() );
  return Err::m_nOK;
}


template<class t_OrigPel,DistortionMeasure t_eDistMeasure>
FORCEINLINE ErrVal FastFullSearch<t_OrigPel,t_eDistMeasure>::xSearchMvFixedPrecision( int iNumPlanesToSearch,
                                                                                      const PicBlock<const t_OrigPel> &rcOrigPicBlock,
                                                                                      const ReferencePlanePackage<Pel> &rcRefPlaPa,
                                                                                      const MotionVector &rcPosVecToSearchWindow,
                                                                                      const MvMeasureInit&  rcMvMeasureInitInput,
                                                                                      const SearchControl& rcSearchControl,
                                                                                      const Lambda cLambda,
                                                                                      LambdaTimesRate &rcBestCostVal,
                                                                                      MotionVector &rcPosVecToBestPos )
{
  ROF( iNumPlanesToSearch <= (int)rcOrigPicBlock.getStats().getNumPlanes() );
  const int iSearchPrecision = rcPosVecToSearchWindow.getPrecision();
  ROF( rcPosVecToBestPos.getPrecision() == iSearchPrecision );

  const PrimBlockSize cUpsBlockSize     = ( ( rcOrigPicBlock.getStats().getPrimBlockSize() - PrimBlockSize( 1, 1 ) )<< iSearchPrecision ) + PrimBlockSize( 1, 1 );
  const PrimBlockSize cSearchRegionSize = rcSearchControl.getSearchWindowSize() + cUpsBlockSize - PrimBlockSize( 1, 1 );

  std::vector<Subsampling> acSRSub;
  for( int i = 0; i < iNumPlanesToSearch; i++ )
  {
    acSRSub.push_back( iSearchPrecision == 0 ? rcOrigPicBlock.getStats().getSubsampling()[ i ] : Subsampling( ShiftScale(0), ShiftScale(0) ) );
  }
  PicBlockStats       cSearchRegionStats( cSearchRegionSize, acSRSub );
  AllocPicBlock<int>  cSearchRegion     ( cSearchRegionStats );

  if( iSearchPrecision == 0 )
  {
    ReferencePlanePackageAccessor<SetPredictionToBlock<int>,Pel> cRefPlaPaAccess( rcRefPlaPa, SetPredictionToBlock<int>() );
    RNOK( cRefPlaPaAccess.getUpsampledBlock( cSearchRegion, rcPosVecToSearchWindow ) );
  }
  else
  {
    ReferencePlanePackageAccessor<SetPredictionToBlockWithStepSize<int>,Pel> cRefPlaPaAccess( rcRefPlaPa, SetPredictionToBlockWithStepSize<int>() );
    RNOK( cRefPlaPaAccess.getUpsampledBlock( cSearchRegion, rcPosVecToSearchWindow ) );
  }

  const int           iMvPrecision  = rcMvMeasureInitInput.m_iMvScale;
  const MotionVector  cSWOrigin     = rcPosVecToSearchWindow.getMvWithPrecisionIncreasedTo( iMvPrecision );
  MvMeasureInit       cMvMeasureInit;
  cMvMeasureInit.m_iMvScale         = iMvPrecision - rcPosVecToSearchWindow.getPrecision();
  cMvMeasureInit.m_bMvCompetition   = rcMvMeasureInitInput.m_bMvCompetition;
  if( cMvMeasureInit.m_bMvCompetition )
  {
    for( UInt uiIdx = 0; uiIdx < rcMvMeasureInitInput.m_acMvCompOrigins.size(); uiIdx++ )
    {
      int iX = cSWOrigin.getX() - rcMvMeasureInitInput.m_acMvCompOrigins[uiIdx].getX();
      int iY = cSWOrigin.getY() - rcMvMeasureInitInput.m_acMvCompOrigins[uiIdx].getY();
      cMvMeasureInit.m_acMvCompOrigins.push_back( PrimPixelPos( iX, iY ) );
    }
  }
  else
  {
    cMvMeasureInit.m_iYOrigin = cSWOrigin.getY() - rcMvMeasureInitInput.m_iYOrigin;
    for( UInt uiIdx = 0; uiIdx < rcMvMeasureInitInput.m_aiYThreshold.size(); uiIdx++ )
    {
      cMvMeasureInit.m_aiYThreshold.push_back( rcMvMeasureInitInput.m_aiYThreshold[ uiIdx ] - cSWOrigin.getY() );
      cMvMeasureInit.m_aiXOrigin   .push_back( cSWOrigin.getX() - rcMvMeasureInitInput.m_aiXOrigin   [ uiIdx ] );
    }
  }

  PrimPixelPos cBestPos = rcSearchControl.getSearchWindowCenter();

  bool bBlockSizeIsMultipleOf4x4 = true;
  if( t_eDistMeasure == DIST_SAD && iSearchPrecision == 0 )
  {
    for( UInt ui = 0; ui < (UInt)iNumPlanesToSearch && bBlockSizeIsMultipleOf4x4; ui++ )
    {
      const BlockSize& rcBlkSize = rcOrigPicBlock.getPlane( ui ).getSize();
      if( ( rcBlkSize.getX() % 4 ) != 0 || ( rcBlkSize.getY() % 4 ) != 0 )
      {
        bBlockSizeIsMultipleOf4x4 = false;
      }
    }
  }

  bool bFoundNewBestPos = false;
  if( t_eDistMeasure == DIST_SAD && iSearchPrecision == 0 && bBlockSizeIsMultipleOf4x4 )
  {
    AllocPicBlock<UInt> cNormRegion( cSearchRegionStats );
    RNOK( rcRefPlaPa.getL1Norm4x4Block( cNormRegion, rcPosVecToSearchWindow ) );
    bFoundNewBestPos = true;
    RNOK( m_cSearch.searchMvFullPel( rcSearchControl.getFullPelSearchMode(),
                                     rcOrigPicBlock,
                                     cSearchRegion,
                                     cBestPos,
                                     cLambda,
                                     rcBestCostVal,
                                     m_uiNumEliminatedFull,
                                     m_uiTotalFull,
                                     cMvMeasureInit,
                                     cNormRegion ) );
  }
  else
  {
    rcBestCostVal.setNumerator( TypeLimits<Int64>::m_iMax );
    bFoundNewBestPos = m_cSearch.searchMV( rcOrigPicBlock,
                                           cSearchRegion,
                                           cBestPos,
                                           cLambda,
                                           rcBestCostVal,
                                           cMvMeasureInit,
                                           iSearchPrecision );
    ROF( bFoundNewBestPos );
  }

  if( bFoundNewBestPos )
  {
    rcPosVecToBestPos.getX() = rcPosVecToSearchWindow.getX() + cBestPos.getX();
    rcPosVecToBestPos.getY() = rcPosVecToSearchWindow.getY() + cBestPos.getY();
    rcPosVecToBestPos.getPrecision() = rcPosVecToSearchWindow.getPrecision();
  }

  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _FASTFULLSEARCH_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
