/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoEncoderLibPch.h"
#include "InterTranslationalEstimator.h"
#include "VideoCommonLib/SliceData.h"
#include "VideoCommonLib/ReferencePictureAccessor.h"
#include "VideoCommonLib/AccessUnit.h"
//#include "FastFullSearch.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

InterTranslationalEstimator::InterTranslationalEstimator( UInt uiSearchModeFullPel )
: m_cNav( NULL )
, m_cBestPredInfo( true )
, m_cCurrPredInfo( true )
, m_uiMaxNumRefBlocksPerPred( 0 )
, m_pcMotionSearchIf( 0 )
, m_uiDuration( 0 )
{
  if     ( uiSearchModeFullPel  < 2 ) m_pcMotionSearchIf = new FastFullSearch <OrigPel,DIST_SAD      >();
  else if( uiSearchModeFullPel == 2 ) m_pcMotionSearchIf = new FastLossySearch<OrigPel,DIST_SAD,false>();
  else if( uiSearchModeFullPel == 3 ) m_pcMotionSearchIf = new FastLossySearch<OrigPel,DIST_SAD,true >();
  else AOT(1);
}


ErrVal
InterTranslationalEstimator::xGetMvCandVector( PicOrderCnt cRefPoc,
                                             const QTNavi &rcNav,
                                             const QuadtreeGroup<PredInfo*> &rcTree,
                                             const MotionVector& rcMvPred,
                                             std::vector<MotionVector>& rcMvCandList )
{
  ROF( rcMvCandList.empty() );
  int                 iMvPrecision  = 0;
  const MotionVector  cMvP          = rcMvPred.getMvWithPrecisionModifiedTo( iMvPrecision );
  rcMvCandList.push_back( cMvP );
  if( cMvP.getX() != 0 || cMvP.getY() != 0 )
  {
    rcMvCandList.push_back( MotionVector( 0, 0, iMvPrecision ) );
  }
  RNOK( InterTranslational::xGetMvPredCandLstExt( cRefPoc, rcNav, rcTree, iMvPrecision, rcMvCandList ) );
  return Err::m_nOK;
}

ErrVal InterTranslationalEstimator::xSearchOneAdditionalHypothesis( const QTNavi &rcNav,
                                                                  MotionSearchIf<OrigPel> &rcMotionSearchIf,
                                                                  InterTranslational &rcPredictionCalculator,
                                                                  const PredSyntaxInterTranslational &rcPredInfoToExtend,
                                                                  const EstInfoInterTranslational &rcEstInfo,
                                                                  const CodingState &rcCodingState,
                                                                  const PlanePackageParameters &rcPlaPaPars,
                                                                  const AccessUnitParameters &rcAUPars,
                                                                  const SliceData &rcCurrSlice,
                                                                  const PicBlock<const Pel> &rcOrigPicBlock,
                                                                  std::vector<PredSyntaxInterTranslational*> &rapcPredInfo,
                                                                  bool &rbPredictionValid )
{
  ROF( rapcPredInfo.empty() );
  ROF( rcCodingState.getPlanePackagesForPrediction().size() );

  rbPredictionValid = false;
  const UInt uiNumPred = rcPredInfoToExtend.getNumMv() + 1;

  AllocPicBlock<int> cNewOriginalSignal( rcOrigPicBlock.getStats() );
  cNewOriginalSignal.assign( rcOrigPicBlock );
  cNewOriginalSignal *= ( uiNumPred * ( 1 << EXTRA_PRECISION_BITS ) );

  AutoDeletePtrVector<const PicBlock<const DPel> > apcPredictions;
  AutoDeletePtrVector<const PicBlock<DPel> > apcWorkingCopy;
  // accumulate already determined predictions with n-1 vectors
  for( UInt ui = 0; ui < rcPredInfoToExtend.getNumMv(); ui++ )
  {
    AllocPicBlock<DPel> *pcPartialPredictionSignal = new AllocPicBlock<DPel>( rcOrigPicBlock.getStats() );
    PredSyntaxInterTranslational cDummy( true );
    bool bValid = false;
    cDummy.pushBack( rcPredInfoToExtend.getMvAndRefPoc( ui ) );
    RNOK( rcPredictionCalculator.calculatePrediction( cDummy, rcNav, rcNav.getPrimPixelPos(), NULL, rcCodingState, rcCurrSlice, *pcPartialPredictionSignal, bValid ) );
    ROF( bValid );
    cNewOriginalSignal -= *pcPartialPredictionSignal;
    apcPredictions.push_back( pcPartialPredictionSignal );

    apcWorkingCopy.push_back( new AllocPicBlock<DPel>( rcOrigPicBlock.getStats() ) );
  }
  apcWorkingCopy.push_back( new AllocPicBlock<DPel>( rcOrigPicBlock.getStats() ) );

  std::vector<MotionVector> cMotionVectorVector;
  // for n = 1 to n_max
  for( UInt uiIdx = 0; uiIdx < rcCodingState.getPlanePackagesForPrediction().size(); uiIdx++ )
  {
    const PicOrderCnt                 cRefPoc    =  rcCodingState.getPlanePackagesForPrediction()[uiIdx]->getPicOrderCnt();
    const ReferencePlanePackage<Pel>& rcRefPlaPa = *rcCodingState.getPlanePackagesForPrediction()[uiIdx];

    // Get motion vector predictor
    MvMeasureInit cMvMeasureInit;
    cMvMeasureInit.m_iMvScale       = (int)rcEstInfo.getMvAccuracy();
    cMvMeasureInit.m_bMvCompetition = rcPlaPaPars.getMvCompetition();
    MotionVector cMvPredictor( 0, 0, cMvMeasureInit.m_iMvScale );
    if( cMvMeasureInit.m_bMvCompetition )
    {
      std::vector<MotionVector> acMvCand;
      RNOK( InterTranslational::getMvPredLst( cRefPoc, rcNav, rcCurrSlice.getPredInfoTree(), acMvCand, cMvMeasureInit.m_iMvScale ) );
      for( UInt uiPrdIdx = 0; uiPrdIdx < acMvCand.size(); uiPrdIdx++ )
      {
        MotionVector cMvX = acMvCand[uiPrdIdx].plusPrimResPos( rcNav.getPrimPixelPos() );
        cMvMeasureInit.m_acMvCompOrigins.push_back( PrimPixelPos( cMvX.getX(), cMvX.getY() ) );
      }
      RNOK( InterTranslational::getMvPred( cRefPoc, rcNav, rcCurrSlice.getPredInfoTree(), cMvPredictor, cMotionVectorVector) );
    }
    else
    {
      const int iYOffset  = rcNav.getPrimPixelPos().getY() << cMvMeasureInit.m_iMvScale;
      const int iXOffset  = rcNav.getPrimPixelPos().getX() << cMvMeasureInit.m_iMvScale;
      RNOK( InterTranslational::getYThresXPredLists( cRefPoc, rcNav, rcCurrSlice.getPredInfoTree(), cMvMeasureInit.m_iMvScale, cMvMeasureInit.m_iYOrigin, cMvMeasureInit.m_aiYThreshold, cMvMeasureInit.m_aiXOrigin ) );
      cMvMeasureInit.m_iYOrigin += iYOffset;
      for( UInt uiPrdIdx  = 0; uiPrdIdx < cMvMeasureInit.m_aiYThreshold.size(); uiPrdIdx++ )
      {
        cMvMeasureInit.m_aiYThreshold[ uiPrdIdx ] += iYOffset;
        cMvMeasureInit.m_aiXOrigin   [ uiPrdIdx ] += iXOffset;
      }
      RNOK( InterTranslational::getMvPredYInd( cRefPoc, rcNav, rcCurrSlice.getPredInfoTree(), cMvPredictor.getPrecision(), cMvPredictor.getY() ) );
      RNOK( InterTranslational::getMvPredXDep( cRefPoc, rcNav, rcCurrSlice.getPredInfoTree(), cMvPredictor.getPrecision(), cMvPredictor.getY(), cMvPredictor.getX() ) );
    }

    std::vector<MotionVector> acMvCandidates;
    if( rcEstInfo.getSearchModeFullPel() >= 2 )
    {
      RNOK( xGetMvCandVector( cRefPoc, rcNav, rcCurrSlice.getPredInfoTree(), cMvPredictor, acMvCandidates ) );
    }

    const Lambda cMotionLambda = Lambda( sqrt( rcEstInfo.getMotionLambda().getValueDouble() ) ) * BinaryFraction<0>( int( uiNumPred ) * ( 1 << EXTRA_PRECISION_BITS ) );
    MotionVector cLocalBestMv = cMvPredictor.getMvWithPrecisionDecreasedTo( 0 );

    std::vector<SearchControl> acSearchControl;

    const int iDeltaPoc = gAbs( int( rcAUPars.getPicOrderCnt() - cRefPoc ) );
    BlockSize cSrchRange = rcEstInfo.getSearchRange();

    cSrchRange.getX() = gMax( 64, cSrchRange.getX() - 8 * gMax( 0, 8 - iDeltaPoc ) );
    cSrchRange.getY() = gMax( 64, cSrchRange.getY() - 8 * gMax( 0, 8 - iDeltaPoc ) );

    acSearchControl.push_back( SearchControl( 0, cSrchRange, rcEstInfo.getSearchModeFullPel() ) );

    if( rcEstInfo.getMvAccuracy() != 0 )
    {
      if( rcEstInfo.getSearchModeSubPel() > 1 )
      {
        for( int iSearchAccuracy = 1; iSearchAccuracy <= (int)rcEstInfo.getMvAccuracy(); iSearchAccuracy++ )
        {
          acSearchControl.push_back( SearchControl( iSearchAccuracy, 1 ) );
        }
      }
      else if( rcEstInfo.getSearchModeSubPel() == 1 )
      {
        acSearchControl.push_back( SearchControl( rcEstInfo.getMvAccuracy(), ( 1 << rcEstInfo.getMvAccuracy() ) >> 1 ) );
      }
      else
      {
        acSearchControl.push_back( SearchControl( rcEstInfo.getMvAccuracy(), ( 1 << rcEstInfo.getMvAccuracy() ) - 1 ) );
      }
    }

    RNOK( rcMotionSearchIf.searchMotion( rcEstInfo.searchFirstPlaneOnly() ? 1 : rcOrigPicBlock.getStats().getNumPlanes(),
                                         acSearchControl,
                                         rcNav,
                                         rcCurrSlice,
                                         cNewOriginalSignal,
                                         rcRefPlaPa,
                                         cMotionLambda,
                                         cMvMeasureInit,
                                         cLocalBestMv,
                                         acMvCandidates ) );

    PredSyntaxInterTranslational *pcResult = new PredSyntaxInterTranslational( rcPredInfoToExtend );
    pcResult->pushBack( PredInfoInterTranslational::MvAndRefPoc( cLocalBestMv, cRefPoc ) );

    UInt uiNumIterations = apcWorkingCopy.size() > 1 ? rcEstInfo.getMaxNumIterations() : 0;
    if( uiNumIterations != 0 )
    {
      // copy the found prediction parts
      for( UInt ui = 0; ui < apcPredictions.size(); ui++ )
      {
        apcWorkingCopy[ui]->assign( *apcPredictions[ui] );
      }

      // create the latest found prediction part
      PredSyntaxInterTranslational cDummy( true );
      bool bValid = false;
      cDummy.pushBack( pcResult->getMvAndRefPoc( uiNumPred - 1 ) );
      RNOK( rcPredictionCalculator.calculatePrediction( cDummy, rcNav, rcNav.getPrimPixelPos(), NULL, rcCodingState, rcCurrSlice, *apcWorkingCopy.back(), bValid ) );

      // prepare a copy of the currently valid residual (scaled by uiNumPred)
      AllocPicBlock<int> cNewOriginalCopy( rcOrigPicBlock.getStats() );
      cNewOriginalCopy.assign( cNewOriginalSignal );
      cNewOriginalCopy -= *apcWorkingCopy.back();

      // Refine the motion vectors
      RNOK( xRefineMVs( rcNav,
                        rcMotionSearchIf,
                        rcPredictionCalculator,
                        rcPlaPaPars,
                        rcEstInfo,
                        cNewOriginalCopy,
                        apcWorkingCopy,
                        rcCodingState,
                        rcCurrSlice,
                        *pcResult,
                        cMotionLambda ) );
    }

    rapcPredInfo.push_back( pcResult );
    {
      PredSyntaxInterTranslational *pcZeroVec = new PredSyntaxInterTranslational( rcPredInfoToExtend );
      pcZeroVec->pushBack( PredInfoInterTranslational::MvAndRefPoc( cMvPredictor, cRefPoc ) );
      rapcPredInfo.push_back( pcZeroVec );
    }
    pcResult = NULL;
  }

  if( rcAUPars.getEntropySliceFlag() == false )
  {
    // merge mode is active. Test vectors that can be coded in merge mode (Motion Vectors of Blocks above and left).
    const QuadtreeGroup<PredInfo*> &rcTree( rcCurrSlice.getPredInfoTree() );
    const PrimPixelPos cNavPos = rcNav.getPrimPixelPos();
    const QTNavi cLeftPos = rcNav.findLeafNode( cNavPos + PrimPixelPos( -1,  0 ) );
    const QTNavi cTopPos  = rcNav.findLeafNode( cNavPos + PrimPixelPos(  0, -1 ) );
    const PredInfoInterTranslational* pcTop = ( cTopPos.isLeafNode() ? rcTree.getT( cTopPos )->getPredInfoInterTranslational() : 0 );
    const PredInfoInterTranslational* pcLeft = ( cLeftPos.isLeafNode() ? rcTree.getT( cLeftPos )->getPredInfoInterTranslational() : 0 );

    UInt uiNumberOfMVRequired = rcPredInfoToExtend.getNumMv() + 1;

    const std::vector<PredInfoInterTranslational::MvAndRefPoc>* pacTopVectors = NULL;
    const std::vector<PredInfoInterTranslational::MvAndRefPoc>* pacLeftVectors = NULL;
    if( pcTop )
    {
      pacTopVectors = &( pcTop->getMvAndRefPocVec() );
      if( pacTopVectors->size() == uiNumberOfMVRequired )
      {
        PredSyntaxInterTranslational *pcMerged = new PredSyntaxInterTranslational( true );
        for( UInt ui = 0; ui < pacTopVectors->size(); ui++ )
        {
          pcMerged->pushBack( ( *pacTopVectors )[ ui ] );
        }
        rapcPredInfo.push_back( pcMerged );
      }
    }
    if( pcLeft )
    {
      pacLeftVectors = &( pcLeft->getMvAndRefPocVec() );
      if( pacLeftVectors->size() == uiNumberOfMVRequired )
      {
        bool bSameVectors = false;
        if( pcTop )
        {
          bSameVectors = PredInfoInterTranslational::compareMvAndRefPocVec( *pacTopVectors, *pacLeftVectors );
        }
        
        if( !bSameVectors )
        {
          PredSyntaxInterTranslational *pcMerged = new PredSyntaxInterTranslational( true );
          for( UInt ui = 0; ui < pacLeftVectors->size(); ui++ )
          {
            pcMerged->pushBack( ( *pacLeftVectors )[ ui ] );
          }
          rapcPredInfo.push_back( pcMerged );
        }
      }
    }
  }

  if( rcAUPars.isDirectModeFlagPresent( rcPlaPaPars.getPlaPaId() ) && uiNumPred == rcEstInfo.getMaxNumRefBlocksPerPred() && rcEstInfo.getMaxNumRefBlocksPerPred() > 1 )
  {
    PredSyntaxInterTranslational* pcDirect  = new PredSyntaxInterTranslational( true );
    RNOK( pcDirect->setDirectModeData( rcNav, rcPlaPaPars, rcAUPars, rcCurrSlice ) );
    bool bDirectModeOk = true;
    for( UInt ui = 0; ui < pcDirect->getNumMv() && bDirectModeOk; ui++ )
    {
      bDirectModeOk = ( rcCodingState.getRefPlanePackage( pcDirect->getRefPoc( ui ) ) != 0 );
    }
    if( bDirectModeOk )
    {
      rapcPredInfo.push_back( pcDirect );
    }
    else
    {
      delete pcDirect;
    }
  }

  rbPredictionValid = true;

  return Err::m_nOK;
}


ErrVal InterTranslationalEstimator::xRefineMVs( const QTNavi &rcNav,
                                              MotionSearchIf<OrigPel> &rcMotionSearchIf,
                                              InterTranslational &rcPredictionCalculator,
                                              const PlanePackageParameters& rcPlaPaPars,
                                              const EstInfoInterTranslational &rcEstInfo,
                                              PicBlock<int> &rcResidualCurr,
                                              const std::vector<const PicBlock<DPel>*> &rapcPreds,
                                              const CodingState &rcCodingState,
                                              const SliceData &rcCurrSlice,
                                              PredInfoInterTranslational &rcPredInfo,
                                              Lambda cLambda )
{
  const UInt uiNumPred    = UInt( rapcPreds.size() );
  const int  iMvAccuracy  = int ( rcEstInfo.getMvAccuracy() );
  std::vector<SearchControl> acSearchControl;
  acSearchControl.push_back( SearchControl( 0, rcEstInfo.getIterSearchRange(), rcEstInfo.getSearchModeFullPel() ) );
  if( iMvAccuracy != 0 )
  {
    if( rcEstInfo.getSearchModeSubPel() > 1 )
    {
      for( int iSearchAccuracy = 1; iSearchAccuracy <= iMvAccuracy; iSearchAccuracy++ )
      {
        acSearchControl.push_back( SearchControl( iSearchAccuracy, 1 ) );
      }
    }
    else if( rcEstInfo.getSearchModeSubPel() == 1 )
    {
      acSearchControl.push_back( SearchControl( iMvAccuracy, ( 1 << iMvAccuracy ) >> 1 ) );
    }
    else
    {
      acSearchControl.push_back( SearchControl( iMvAccuracy, ( 1 << iMvAccuracy ) - 1 ) );
    }
  }

  for( UInt uiIter = 0; uiIter < rcEstInfo.getMaxNumIterations(); uiIter++ )
  {
    for( UInt uiPartPred = 0; uiPartPred < uiNumPred; uiPartPred++ )
    {
      rcResidualCurr += *rapcPreds[uiPartPred];
      const PicOrderCnt cRefPoc = rcPredInfo.getRefPoc( uiPartPred );
      const ReferencePlanePackage<Pel>& rcRefPlaPa  = *rcCodingState.getRefPlanePackage( cRefPoc );

      // Get motion vector predictor
      ROF( iMvAccuracy == rcPredInfo.getMv( uiPartPred ).getPrecision() );

      MvMeasureInit cMvMeasureInit;
      cMvMeasureInit.m_iMvScale       = iMvAccuracy;
      cMvMeasureInit.m_bMvCompetition = rcPlaPaPars.getMvCompetition();
      if( cMvMeasureInit.m_bMvCompetition )
      {
        std::vector<MotionVector> acMvCand;
        RNOK( InterTranslational::getMvPredLst( cRefPoc, rcNav, rcCurrSlice.getPredInfoTree(), acMvCand, cMvMeasureInit.m_iMvScale ) );
        for( UInt uiPrdIdx = 0; uiPrdIdx < acMvCand.size(); uiPrdIdx++ )
        {
          MotionVector cMvX = acMvCand[uiPrdIdx].plusPrimResPos( rcNav.getPrimPixelPos() );
          cMvMeasureInit.m_acMvCompOrigins.push_back( PrimPixelPos( cMvX.getX(), cMvX.getY() ) );
        }
      }
      else
      {
        const int iYOffset  = rcNav.getPrimPixelPos().getY() << cMvMeasureInit.m_iMvScale;
        const int iXOffset  = rcNav.getPrimPixelPos().getX() << cMvMeasureInit.m_iMvScale;
        RNOK( InterTranslational::getYThresXPredLists( cRefPoc, rcNav, rcCurrSlice.getPredInfoTree(), cMvMeasureInit.m_iMvScale, cMvMeasureInit.m_iYOrigin, cMvMeasureInit.m_aiYThreshold, cMvMeasureInit.m_aiXOrigin ) );
        cMvMeasureInit.m_iYOrigin += iYOffset;
        for( UInt uiPrdIdx  = 0; uiPrdIdx < cMvMeasureInit.m_aiYThreshold.size(); uiPrdIdx++ )
        {
          cMvMeasureInit.m_aiYThreshold[ uiPrdIdx ] += iYOffset;
          cMvMeasureInit.m_aiXOrigin   [ uiPrdIdx ] += iXOffset;
        }
      }

      // refine here
      // two-step search
      rcPredInfo.getMv( uiPartPred ) = rcPredInfo.getMv( uiPartPred ).getMvWithPrecisionDecreasedTo( 0 );
      std::vector<MotionVector> acMvCandidates;
      if( rcEstInfo.getSearchModeFullPel() >= 2 )
      {
        acMvCandidates.push_back( rcPredInfo.getMv( uiPartPred ) );
      }
      RNOK( rcMotionSearchIf.searchMotion( rcEstInfo.searchFirstPlaneOnly() ? 1 : rcResidualCurr.getStats().getNumPlanes(),
                                           acSearchControl,
                                           rcNav,
                                           rcCurrSlice,
                                           rcResidualCurr,
                                           rcRefPlaPa,
                                           cLambda,
                                           cMvMeasureInit,
                                           rcPredInfo.getMv( uiPartPred ),
                                           acMvCandidates ) );

      // update rapcPreds[uiPartPred] with the refined prediction part
      bool bValid = false;
      PredSyntaxInterTranslational cSinglePred( true );
      cSinglePred.pushBack( rcPredInfo.getMvAndRefPoc( uiPartPred ) );
      RNOK( rcPredictionCalculator.calculatePrediction( cSinglePred, rcNav, rcNav.getPrimPixelPos(), NULL, rcCodingState, rcCurrSlice, *rapcPreds[uiPartPred], bValid ) );

      rcResidualCurr -= *rapcPreds[uiPartPred];
    }
  }
  return Err::m_nOK;
}


ErrVal InterTranslationalEstimator::estimateAndCalculatePrediction( PredInfo *&rpcPredInfo,
                                                                  const QTNavi &rcNav,
                                                                  const EstInfo &rcEstInfo,
                                                                  const CodingState &rcCodingState,
                                                                  const AccessUnitParameters &rcAUPars,
                                                                  const PlanePackageParameters &rcPlaPaPars,
                                                                  const SliceData &rcCurrSlice,
                                                                  const PicBlock<const Pel> &rcOrigPicBlock,
                                                                  const PicBlock<const Pel> *pcEncPic,
                                                                  const PicBlock<DPel> &rcPredPicBlock,
                                                                  bool &rbPredictionValid )
{
  if( m_apcPredInfo.empty() )
  {
    m_cBestRDCost.setMax();
    TimeStamp cBefore, cAfter;
    cBefore.stamp();
    m_cNav = rcNav;

    const EstInfoInterTranslational &rcEstInfoInterTranslational = *rcEstInfo.getEstInfoInterTranslational();
    ROF( m_uiMaxNumRefBlocksPerPred == 0 || m_uiMaxNumRefBlocksPerPred == rcEstInfoInterTranslational.getMaxNumRefBlocksPerPred() );
    m_uiMaxNumRefBlocksPerPred = rcEstInfoInterTranslational.getMaxNumRefBlocksPerPred();

    RNOK( xSearchOneAdditionalHypothesis( rcNav,
                                          *m_pcMotionSearchIf,
                                          *this,
                                          m_cBestPredInfo,
                                          rcEstInfoInterTranslational,
                                          rcCodingState,
                                          rcPlaPaPars,
                                          rcAUPars,
                                          rcCurrSlice,
                                          rcOrigPicBlock,
                                          m_apcPredInfo,
                                          rbPredictionValid ) );
    cAfter.stamp();
    m_uiDuration += (cAfter - cBefore).getMSec();
  }
  m_cCurrPredInfo = *m_apcPredInfo.back();
  rpcPredInfo = m_apcPredInfo.back();
  m_apcPredInfo.pop_back();

  RNOK( calculatePrediction( *rpcPredInfo, rcNav, rcNav.getPrimPixelPos(), NULL, rcCodingState, rcCurrSlice, rcPredPicBlock, rbPredictionValid ) );
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
