/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _RDCOST_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _RDCOST_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/PicBlock.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

typedef BinaryFraction<10> Lambda;
typedef BinaryFraction<26> LambdaTimesRate;

class RDCost
{
public:
  RDCost( Lambda cLambda )
    : m_cRate( 0 )
    , m_uiDistortion( 0 )
    , m_uiNumSamples( 0 )
    , m_cLambda( cLambda )
    , m_bIsMax( false ) {}

  RDCost()
    : m_cRate( 0 )
    , m_uiDistortion( 0 )
    , m_uiNumSamples( 0 )
    , m_cLambda( 0 )
    , m_bIsMax( false ) {}

#if DISABLE_CHANGES_THAT_BREAK_0_89_COMPATIBILITY
  RDCost( Lambda cLambda, UInt uiNumSamples, UInt uiDistortion, BitRate cRate )
#else
  RDCost( Lambda cLambda, UInt uiNumSamples, UInt64 uiDistortion, BitRate cRate )
#endif
    : m_cRate( cRate )
    , m_uiDistortion( uiDistortion )
    , m_uiNumSamples( uiNumSamples )
    , m_cLambda( cLambda )
    , m_bIsMax( false ) {}

  ErrVal traceRDCost( std::ostream &rcOut ) const
  {
    rcOut << std::fixed << std::setprecision( 3 ) << std::setw( 10 ) << "Rate: " << m_cRate.getValueDouble() << "   Dist: " << m_uiDistortion << std::endl;
    return Err::m_nOK;
  }

  void reset() { m_cRate = 0; m_uiDistortion = 0; m_uiNumSamples = 0; m_bIsMax = false; }

  Lambda getLambda() const { return m_cLambda; }
  void addRate( BitRate cRate ) { AOT_DBG( m_cLambda == Lambda( 0 ) ); m_cRate += cRate; }
  void setMax () { m_bIsMax = true; }

  template<class T, class U>
  void addDistortion( const PicBlock<const T> &rcOrig, const PicBlock<const U> &rcEnc, UInt uiShift = 0 )
  {
    CAOT( sizeof( T ) > 2 );
    CAOT( sizeof( U ) > 2 );

    AOF( rcOrig.getStats() == rcEnc.getStats() );
    for( UInt uiPl = 0; uiPl < rcOrig.getStats().getNumPlanes(); uiPl++ )
    {
#if DISABLE_CHANGES_THAT_BREAK_0_89_COMPATIBILITY
      UInt uiSSD = 0;
#else
      UInt64 uiSSD = 0;
#endif
      const Block<const T> &rcOrigBlk = rcOrig.getPlane( uiPl );
      const Block<const U> &rcEncBlk  = rcEnc .getPlane( uiPl );
      const int iWidth = rcOrigBlk.getSize().getX();
      const int iOrigStride = rcOrigBlk.getStride();
      const int iEncStride = rcEncBlk.getStride();
      const T* pcTOrig = &rcOrigBlk.getEl( 0);
      const U* pcUEnc  = &rcEncBlk.getEl( 0);
      const T* pcTOrigEnd = pcTOrig + rcOrigBlk.getSize().getY() * iOrigStride;

      for( ; pcTOrig < pcTOrigEnd; pcTOrig += iOrigStride, pcUEnc += iEncStride )
      {
        for( int x = 0; x < iWidth; x++ )
        {
#if DISABLE_CHANGES_THAT_BREAK_0_89_COMPATIBILITY
          int iDiff = (int)pcTOrig[x]-(int)pcUEnc[x];
#else
          const Int64 iDiff = (Int64)pcTOrig[x]-(Int64)pcUEnc[x];
#endif
          uiSSD += iDiff * iDiff;
        }
      }
      m_uiDistortion += uiSSD;
      m_uiNumSamples += rcOrigBlk.getSize().getX()*rcOrigBlk.getSize().getY();
    }
    if( uiShift )
    {
      m_uiDistortion  += ( 1 << ( uiShift << 1 ) ) >> 1;
      m_uiDistortion >>=        ( uiShift << 1 );
    }
  }

  RDCost& operator+=( const RDCost& rcRDCost )
  {
    AOF_DBG( m_cLambda == rcRDCost.m_cLambda );
    AOT_DBG( m_cLambda == Lambda( 0 ) );
    m_cRate        += rcRDCost.m_cRate;
    m_uiDistortion += rcRDCost.m_uiDistortion;
    m_uiNumSamples += rcRDCost.m_uiNumSamples;
    return *this;
  }

  bool operator<( const RDCost& rcRDCost ) const
  {
    TOT( m_bIsMax && rcRDCost.m_bIsMax );
    if( m_bIsMax )
      return false;
    if( rcRDCost.m_bIsMax )
      return true;
    AOF_DBG( m_cLambda == rcRDCost.m_cLambda );
    AOT_DBG( m_cLambda == Lambda( 0 ) );
    TOF( m_uiNumSamples == rcRDCost.m_uiNumSamples );
#if DISABLE_CHANGES_THAT_BREAK_0_89_COMPATIBILITY
    // original with overflow bug
    return BinaryFraction<0>( (int)m_uiDistortion ) + m_cLambda * m_cRate <
           BinaryFraction<0>( (int)rcRDCost.m_uiDistortion ) + m_cLambda * rcRDCost.m_cRate;
#else
    // with overflow test
    const bool bLeftFits  = ( m_uiDistortion          >> (64-28) ) == 0;
    const bool bRightFits = ( rcRDCost.m_uiDistortion >> (64-28) ) == 0;
    if( bLeftFits && bRightFits )
    {
      return BinaryFraction<0>( (Int64)m_uiDistortion ) + m_cLambda * m_cRate <
             BinaryFraction<0>( (Int64)rcRDCost.m_uiDistortion ) + m_cLambda * rcRDCost.m_cRate;
    }
    TOT( m_uiDistortion          >> 63 );
    TOT( rcRDCost.m_uiDistortion >> 63 );
    const Int64 iD1MinusD2 = Int64( m_uiDistortion ) - Int64( rcRDCost.m_uiDistortion );

    bool bReturnValue;
    const BitRate cR2MinusR1 = rcRDCost.m_cRate - m_cRate;
    const bool bSignD = ( iD1MinusD2                < 0 );
    const bool bSignR = ( cR2MinusR1.getNumerator() < 0 );
    if( bSignD != bSignR )
    {
      bReturnValue = bSignD;
    }
    else
    {
      const LambdaTimesRate cValue = m_cLambda * cR2MinusR1;
      const Int64 iValueShifted = ( cValue.getNumerator() >> cValue.getAccuracy() );
      if( iD1MinusD2 < iValueShifted )
      {
        bReturnValue = true;
      }
      else if( iD1MinusD2 == iValueShifted )
      {
        bReturnValue = ( cValue.getNumerator() & ( ( Int64( 1 ) << cValue.getAccuracy() ) - 1 ) ) != 0;
      }
      else
      {
        bReturnValue = false;
      }
    }

#if 0 // check
    if( bLeftFits && bRightFits )
    {
      TOF( bReturnValue == ( BinaryFraction<0>( (Int64)m_uiDistortion ) + m_cLambda * m_cRate <
                             BinaryFraction<0>( (Int64)rcRDCost.m_uiDistortion ) + m_cLambda * rcRDCost.m_cRate ) );
    }
#endif
    return bReturnValue;
#endif
  }

  bool operator==( const RDCost &rcCost ) const
  {
    AOT( m_bIsMax );
    AOF_DBG( m_cLambda == rcCost.m_cLambda );
    AOT_DBG( m_cLambda == Lambda( 0 ) );
    AOF_DBG( m_uiNumSamples == rcCost.m_uiNumSamples );
    return m_uiDistortion == rcCost.m_uiDistortion && m_cRate == rcCost.m_cRate;
  }

  bool operator!=( const RDCost &rcCost ) const { return !(*this == rcCost); }

#if DISABLE_CHANGES_THAT_BREAK_0_89_COMPATIBILITY
  UInt            getDistortion () const { return m_uiDistortion; }
#else
  UInt64          getDistortion () const { return m_uiDistortion; }
#endif
  const BitRate&  getBitRate    () const { return m_cRate; }
  bool            isMax         () const { return m_bIsMax; }

private:
  BitRate m_cRate;
#if DISABLE_CHANGES_THAT_BREAK_0_89_COMPATIBILITY
  UInt m_uiDistortion;
#else
  UInt64 m_uiDistortion;
#endif
  UInt m_uiNumSamples;
  Lambda m_cLambda;
  bool m_bIsMax;
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _RDCOST_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
