/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoEncoderLibPch.h"
#include "RDOQ.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN


const UInt RDOQ::sm_uiQuantShift = 31;

const Int64 RDOQ::sm_aai64QStep[3][6] =
{
  { 16, 18, 20, 23, 25, 29 },
  { 13, 14, 16, 18, 20, 23 },
  { 10, 11, 13, 14, 16, 18 },
};

const int RDOQ::sm_aiCtxModelRate[2][64] =
{
  {
    655,  609,  566,  528,  492,  460,  430,  403,  378,  355,  333,  313,  294,  277,  261,  246,
      232,  218,  206,  194,  183,  173,  164,  155,  146,  138,  131,  123,  117,  110,  105,   99,
      94,   89,   84,   80,   75,   71,   68,   64,   61,   58,   55,   52,   49,   46,   44,   42,
      40,   38,   36,   34,   32,   30,   29,   27,   26,   25,   23,   22,   21,   20,   19,   18
  },
  {
    655,  705,  754,  803,  852,  902,  951, 1000, 1050, 1099, 1148, 1197, 1247, 1296, 1345, 1395,
      1444, 1493, 1542, 1592, 1641, 1690, 1739, 1789, 1838, 1887, 1937, 1986, 2035, 2084, 2134, 2183,
      2232, 2281, 2331, 2380, 2429, 2479, 2528, 2577, 2626, 2676, 2725, 2774, 2824, 2873, 2922, 2971,
      3021, 3070, 3119, 3168, 3218, 3267, 3316, 3366, 3415, 3464, 3513, 3563, 3612, 3661, 3711, 3760
    }
};


RDOQ::RDOQ( const ProbabilityAllocatorBase<CabacContextModel>& rcProbAlloc,
            const ContextArrayStartIndices&     rcCtxArrayStartIndices,
            const CtxAccess&                    rcCtxAccess,
            const AccessUnitEncoderParameters&  rcAUEstPars,
            const PlanePackageParameters&       rcPlaPaPar,
            const SliceEstimator::Parameters&   rcSliceEstPars )
: m_rcProbAlloc                  ( rcProbAlloc )
, m_uiQuantizationMode         ( rcAUEstPars.getQuantizationMode( rcPlaPaPar.getPlaPaId() ) )
, m_uiMinLog2PrimSize          ( rcPlaPaPar.getTransformMode() & ResInfo::RES_TRAFO_POW2 ? rcPlaPaPar.getLog2MinPow2TrafoBlkSize() : 0 )
, m_uiMaxLog2PrimSize          ( rcPlaPaPar.getTransformMode() & ResInfo::RES_TRAFO_POW2 ? rcPlaPaPar.getLog2MaxPow2TrafoBlkSize() : 0 )
, m_uiNumPlanes                ( rcPlaPaPar.getNumPlanes() )
, m_cPrimLambda                ( rcSliceEstPars.getLambda() )
, m_acLambda                   ( m_uiNumPlanes, rcSliceEstPars.getLambda() )
, m_aaauiScanPattern           ( rcPlaPaPar.getLog2MaxPow2TrafoBlkSize() + 1 )
, m_rcCtxArrayStartIndices     ( rcCtxArrayStartIndices )
, m_rcCtxAccess                ( rcCtxAccess )
, m_uiLog2PrimSize             ( 0 )
, m_uiLog2BlkSize              ( 0 )
, m_uiPlane                    ( 0 )
, m_uiCBFCtxInc                ( 0 )
{
  for( UInt uiPlane = 0; uiPlane < m_uiNumPlanes; uiPlane++ )
  {
    int iDeltaQP = rcPlaPaPar.getPlaneQp( uiPlane ) - rcAUEstPars.getPrimQp();
    if( iDeltaQP )
    {
      double  dLambda       = m_acLambda[ uiPlane ].getValueDouble();
      double  dScale        = pow( 2.0, double( iDeltaQP ) / 3.0 );
      dLambda              *= dScale;
      m_acLambda[ uiPlane ] = Lambda( dLambda );
    }
  }

  for( UInt uiLog2TSize = 0; uiLog2TSize < m_aaauiScanPattern.size(); uiLog2TSize++ )
  {
    m_aaauiScanPattern[ uiLog2TSize ].resize( 2 );
    TNOK( ResTrafoPow2::setScanPattern( m_aaauiScanPattern[ uiLog2TSize ][ 0 ], uiLog2TSize, false ) );
    TNOK( ResTrafoPow2::setScanPattern( m_aaauiScanPattern[ uiLog2TSize ][ 1 ], uiLog2TSize, true  ) );
  }
}

RDOQ::~RDOQ()
{
}


FORCEINLINE int RDOQ::xGetIEPRate() const
{
  return sm_aiCtxModelRate[ 0 ][ 0 ];
}


FORCEINLINE int RDOQ::xGetIRate( UShort ui16CtxNumber, UInt uiBit ) const
{
  const CabacContextModel&  rcCtx = m_rcProbAlloc.getCtxModel( ui16CtxNumber );
  UInt                      uiLPS =  rcCtx.getMps  () ^ uiBit;
  return sm_aiCtxModelRate[ uiLPS ][ rcCtx.getCabacState() ];
}


FORCEINLINE Int64 RDOQ::xGetICost( int iRate ) const
{
  return m_i64Lambda * iRate;
}

FORCEINLINE Int64 RDOQ::xGetICost( UShort ui16CtxNumber, UInt uiBit ) const
{
  return m_i64Lambda * xGetIRate( ui16CtxNumber, uiBit );
}


FORCEINLINE Int64 RDOQ::xGetICRateCost( UInt   uiAbsLevel,
                                        bool   bLastScanPos,
                                        UShort ui16CtxNumSig,
                                        UShort ui16CtxNumOne,
                                        UShort ui16CtxNumAbs ) const
{
  if( uiAbsLevel == 0 )
  {
    int iRate = 0;
    if( ! bLastScanPos )
    {
      iRate += xGetIRate( ui16CtxNumSig, 0 );
    }
    return  xGetICost( iRate );
  }
  int iRate = xGetIEPRate();
  if( ! bLastScanPos )
  {
    iRate += xGetIRate( ui16CtxNumSig, 1 );
  }
  if( uiAbsLevel == 1 )
  {
    iRate += xGetIRate( ui16CtxNumOne, 0 );
  }
  else if( uiAbsLevel < 15 )
  {
    iRate += xGetIRate( ui16CtxNumOne, 1 );
    iRate += xGetIRate( ui16CtxNumAbs, 0 );
    iRate += xGetIRate( ui16CtxNumAbs, 1 ) * (int)( uiAbsLevel - 2 );
  }
  else
  {
    uiAbsLevel -= 14;
    int iEGS    = 1;  for( UInt uiMax = 2; uiAbsLevel >= uiMax; uiMax <<= 1, iEGS += 2 );
    iRate += xGetIRate( ui16CtxNumOne, 1 );
    iRate += xGetIRate( ui16CtxNumAbs, 1 ) * 13;
    iRate += xGetIEPRate() * iEGS;
  }
  return  xGetICost( iRate );
}


FORCEINLINE int RDOQ::xGetCodedLevel( Int64&  ri64UncodedCost,
                                      Int64&  ri64CodedCost,
                                      int     iOrgCoeff,
                                      bool    bLastScanPos,
                                      UShort  ui16CtxNumSig,
                                      UShort  ui16CtxNumOne,
                                      UShort  ui16CtxNumAbs,
                                      UInt    uiQuantIdx ) const
{
  int     iSign           = ( iOrgCoeff < 0 ? -1 : 1 );
  Int64   i64AbsCoeff     = ( iOrgCoeff < 0 ? -iOrgCoeff : iOrgCoeff );
  Int64   i64QStepSize    = m_ai64QuantStepSize[ uiQuantIdx ];
  ri64UncodedCost         = i64AbsCoeff * i64AbsCoeff;
  ri64CodedCost           = ri64UncodedCost + xGetICRateCost( 0, bLastScanPos, ui16CtxNumSig, ui16CtxNumOne, ui16CtxNumAbs );
  UInt    uiBestAbsLevel  = 0;
  UInt    uiMaxAbsLevel   = UInt( ( i64AbsCoeff * m_ai64QuantFactor[ uiQuantIdx ] + ( Int64( 1 ) << ( sm_uiQuantShift - 1 ) ) ) >> sm_uiQuantShift );
  for( UInt uiAbsLevel = uiMaxAbsLevel; uiAbsLevel > 0 ; uiAbsLevel-- )
  {
    Int64 i64Delta  = i64AbsCoeff - i64QStepSize * (int)uiAbsLevel;
    Int64 i64Cost   = i64Delta * i64Delta + xGetICRateCost( uiAbsLevel, bLastScanPos, ui16CtxNumSig, ui16CtxNumOne, ui16CtxNumAbs );
    if( i64Cost < ri64CodedCost )
    {
      uiBestAbsLevel  = uiAbsLevel;
      ri64CodedCost   = i64Cost;
    }
    else if( uiAbsLevel < uiMaxAbsLevel )
    {
      uiAbsLevel      = 1;
    }
  }
  return iSign * (int)uiBestAbsLevel;
}


Int64 RDOQ::xGetUncodedCost( const Block<const int>&  rcOrgCoeff,
                             const UShort             ui16CtxNumCBF,
                             UInt&                    ruiMaxLineNum ) const
{
  Int64 i64UncodedCost  = xGetICost( ui16CtxNumCBF, 0 );
  Int64 i64QFactor      = m_ai64QuantFactor  [ 0 ];
  Int64 i64Add          = Int64(1) << ( sm_uiQuantShift - 1 );
  ruiMaxLineNum         = 0;
  for( UInt uiPosY = 0; uiPosY < UInt( rcOrgCoeff.getSize().getY() ); uiPosY++ )
  {
    for( UInt uiPosX = 0; uiPosX < UInt( rcOrgCoeff.getSize().getX() ); uiPosX++ )
    {
      int     iOrgCoeff   = rcOrgCoeff.getEl( uiPosX, uiPosY );       
      Int64   i64AbsCoeff = ( iOrgCoeff < 0 ? -iOrgCoeff : iOrgCoeff );
      i64UncodedCost     += i64AbsCoeff * i64AbsCoeff;

      UInt uiMaxAbsLevel  = UInt( ( i64AbsCoeff * i64QFactor + i64Add ) >> sm_uiQuantShift );
      if(  uiMaxAbsLevel )
      {
        UInt  uiLineNum = uiPosY + uiPosX;
        if(   uiLineNum > ruiMaxLineNum )
        {
          ruiMaxLineNum = uiLineNum;
        }
      }
    }
  }
  return i64UncodedCost;
}


Int64 RDOQ::xQuantize( const Block<const int>&  rcOrgCoeff,
                       const Block<int>&        rcLevel,
                       bool&                    rbCoded ) const
{
  UInt    uiBlkSizeM1       = ( 1 << m_uiLog2BlkSize ) - 1;
  UInt    uiMaxNumCoeff     = 1 << ( m_uiLog2BlkSize << 1 );
  UShort  ui16CtxBaseSig    = (UShort)m_rcCtxAccess.getContextOffset( m_rcCtxArrayStartIndices.m_acResCtx[m_uiPlane].m_uiStartTrafoPow2Sig  + m_uiLog2PrimSize );
  UShort  ui16CtxBaseLast   = (UShort)m_rcCtxAccess.getContextOffset( m_rcCtxArrayStartIndices.m_acResCtx[m_uiPlane].m_uiStartTrafoPow2Last + m_uiLog2PrimSize );
  UShort  ui16CtxOne        = (UShort)m_rcCtxAccess.getContextOffset( m_rcCtxArrayStartIndices.m_acResCtx[m_uiPlane].m_uiResCtxStart + RES_CTX_START_POW2_ABS_GREATER_ONE, 4 );       // just some context
  UShort  ui16CtxAbs        = (UShort)m_rcCtxAccess.getContextOffset( m_rcCtxArrayStartIndices.m_acResCtx[m_uiPlane].m_uiResCtxStart + RES_CTX_START_POW2_COEFF_LEVEL,     4 + 1  );  // just some context
  UShort  ui16CtxCbf        = (UShort)m_rcCtxAccess.getContextOffset( m_rcCtxArrayStartIndices.m_acResCtx[m_uiPlane].m_uiResCtxStart + RES_CTX_START_POW2_CBF,             m_uiCBFCtxInc );
  UInt    uiBestLastIdxP1   = 0;
  UInt    uiMaxLineNum      = 0;
  Int64   i64BestCost       = xGetUncodedCost( rcOrgCoeff, ui16CtxCbf, uiMaxLineNum );
  Int64   i64BaseCost       = i64BestCost - xGetICost( ui16CtxCbf, 0 ) + xGetICost( ui16CtxCbf, 1 );

  const std::vector<std::vector<UInt> >&  raauiScanPattern  = m_aaauiScanPattern[ m_uiLog2BlkSize ];
  UInt                                    uiDownLeft        = 1;
  UInt                                    uiNumSigTopRight  = 0;
  UInt                                    uiNumSigBotLeft   = 0;
  Int64                                   i64CodedCost      = 0;
  Int64                                   i64UncodedCost    = 0;

  for( UInt uiScanPos = 0; uiScanPos < uiMaxNumCoeff; uiScanPos++ )
  {
    UInt    uiBlkPos                = raauiScanPattern[ uiDownLeft ][ uiScanPos ];
    UInt    uiPosY                  = uiBlkPos >> m_uiLog2BlkSize;
    UInt    uiPosX                  = uiBlkPos - ( uiPosY << m_uiLog2BlkSize );
    if( uiPosY + uiPosX > uiMaxLineNum )
    {
      break;
    }
    UShort  uiCtxSig                = ui16CtxBaseSig + (UShort)ResTrafoPow2::getSigCtxInc( rcLevel, uiPosX, uiPosY, m_uiLog2BlkSize, ( uiDownLeft > 0 ) );
    int     iOrgCoeff               = rcOrgCoeff.getEl( uiPosX, uiPosY );
    bool    bLastScanPos            = ( uiScanPos == uiMaxNumCoeff - 1 );
    UInt    uiQuantIdx              = 0;
    int     iLevel                  = xGetCodedLevel( i64UncodedCost, i64CodedCost, iOrgCoeff, bLastScanPos, uiCtxSig, ui16CtxOne, ui16CtxAbs, uiQuantIdx );
    rcLevel.getEl( uiPosX, uiPosY ) = iLevel;
    i64BaseCost                    -= i64UncodedCost;
    i64BaseCost                    += i64CodedCost;
    if( iLevel )
    { //----- check for last flag -----
      UShort  uiCtxLast             = ui16CtxBaseLast + (UShort)ResTrafoPow2::getLastCtxInc( rcLevel, uiPosX, uiPosY, m_uiLog2BlkSize );
      Int64   i64CostLastZero       = xGetICost( uiCtxLast, 0 );
      Int64   i64CostLastOne        = xGetICost( uiCtxLast, 1 );
      Int64   i64CurrIsLastCost     = i64BaseCost + i64CostLastOne;
      i64BaseCost                  += i64CostLastZero;
      if( i64CurrIsLastCost < i64BestCost )
      {
        i64BestCost       = i64CurrIsLastCost;
        uiBestLastIdxP1   = uiScanPos + 1;
      }
      //----- update coeff counts -----
      if( uiPosX > uiPosY )
      {
        uiNumSigTopRight++;
      }
      else if( uiPosY > uiPosX )
      {
        uiNumSigBotLeft ++;
      }
    }
    //===== update scan direction =====
    if( ( uiDownLeft == 1 && ( uiPosX == 0 || uiPosY == uiBlkSizeM1 ) ) ||
      ( uiDownLeft == 0 && ( uiPosY == 0 || uiPosX == uiBlkSizeM1 ) )   )
    {
      uiDownLeft = ( uiNumSigTopRight >= uiNumSigBotLeft ? 1 : 0 );
    }
  }

  //===== clean uncoded coefficients =====
  {
    uiDownLeft          = 1;
    uiNumSigTopRight    = 0;
    uiNumSigBotLeft     = 0;
    for( UInt uiScanPos = 0; uiScanPos < uiMaxNumCoeff; uiScanPos++ )
    {
      UInt    uiBlkPos  = raauiScanPattern[ uiDownLeft ][ uiScanPos ];
      UInt    uiPosY    = uiBlkPos >> m_uiLog2BlkSize;
      UInt    uiPosX    = uiBlkPos - ( uiPosY << m_uiLog2BlkSize );

      if( uiScanPos < uiBestLastIdxP1 )
      {
        if( rcLevel.getEl( uiPosX, uiPosY ) )
        {
          if( uiPosX > uiPosY )
          {
            uiNumSigTopRight++;
          }
          else if( uiPosY > uiPosX )
          {
            uiNumSigBotLeft ++;
          }
        }
      }
      else
      {
        rcLevel.getEl( uiPosX, uiPosY ) = 0;
      }

      if( ( uiDownLeft == 1 && ( uiPosX == 0 || uiPosY == uiBlkSizeM1 ) ) ||
        ( uiDownLeft == 0 && ( uiPosY == 0 || uiPosX == uiBlkSizeM1 ) )   )
      {
        uiDownLeft = ( uiNumSigTopRight >= uiNumSigBotLeft ? 1 : 0 );
      }
    }
  }
  rbCoded = ( uiBestLastIdxP1 > 0 );
  return Err::m_nOK;
}


#if CHECK_RDOQ_COST
Int64 RDOQ::xGetDistortion( const Block<const int>& rcOrgCoeff, const Block<const int>& rcLevel ) const
{
  UInt          uiSize    = 1 << m_uiLog2BlkSize;
  Int64         i64Dist   = 0;
  const int*    paiCoeff  = &rcOrgCoeff.getEl( 0 );
  const int*    paiLevel  = &rcLevel   .getEl( 0 );
  UInt          uiCStride = rcOrgCoeff.getStride();
  UInt          uiLStride = rcLevel   .getStride();

  for( UInt uiY = 0, uiCnt = 0; uiY < uiSize; uiY++, paiCoeff += uiCStride, paiLevel += uiLStride )
  {
    for( UInt uiX = 0; uiX < uiSize; uiX++, uiCnt++ )
    {
      Int64 i64Delta  = (Int64)paiCoeff[ uiX ] - m_ai64QuantStepSize[ 0 ] * paiLevel[ uiX ];
      i64Dist        += i64Delta * i64Delta;
    }
  }
  return i64Dist;
}
#endif


ErrVal RDOQ::quantizePow2( int                     iQP,
                           UInt                    uiLog2PrimSize,
                           UInt                    uiPlane,
                           UInt                    uiCBFCtxInc,
                           const Block<const int>& rcOrgCoeff,
                           const Block<int>&       rcLevel,
                           bool&                   rbCoded,
                           RDCost&                 rcRdCost )
{
  ROF( rcOrgCoeff.getSize() == rcLevel.getSize() );
  ROF( uiLog2PrimSize >= m_uiMinLog2PrimSize );
  ROF( uiLog2PrimSize <= m_uiMaxLog2PrimSize );
  ROF( uiPlane <  m_uiNumPlanes );
  m_uiLog2PrimSize          = uiLog2PrimSize;
  m_uiPlane                 = uiPlane;
  m_uiLog2BlkSize           = gCeilLog2( rcOrgCoeff.getSize().getX() );
  m_uiCBFCtxInc             = uiCBFCtxInc;
  m_i64Lambda               = ( m_acLambda[ uiPlane ].getNumerator() << 8 ) >> m_acLambda[ uiPlane ].getAccuracy();
  m_ai64QuantStepSize [ 0 ] = sm_aai64QStep[ 0 ][ iQP % 6 ] << ( 4 + iQP / 6 );
  m_ai64QuantFactor   [ 0 ] = ( ( Int64( 1 ) << sm_uiQuantShift ) + ( m_ai64QuantStepSize[ 0 ] >> 1 ) ) / m_ai64QuantStepSize[ 0 ];
  Int64   i64Cost           = xQuantize( rcOrgCoeff, rcLevel, rbCoded );
//  ROF( rcRdCost.getRDCostValue().getAccuracy() == 26 ); // Kann man sich nochmal angucken.
#if CHECK_RDOQ_COST
  Int64   i64Dist           = xGetDistortion( rcOrgCoeff, rcLevel );
  UInt    uiDist            = UInt( ( Int64( 25 ) * i64Dist + Int64( 1 << 21 ) ) >> 22 );
  BitRate cRate;  cRate.setNumerator( ( Int64( 400 ) * ( i64Cost - i64Dist ) + ( m_cPrimLambda.getNumerator() >> 1 ) ) / m_cPrimLambda.getNumerator() );
  rcRdCost                 += RDCost( m_cPrimLambda, 1 << ( m_uiLog2BlkSize << 1 ), uiDist, cRate );
#else
  BitRate cRate;  cRate.setNumerator( ( Int64( 400 ) * i64Cost + ( m_cPrimLambda.getNumerator() >> 1 ) ) / m_cPrimLambda.getNumerator() );
  rcRdCost                 += RDCost( m_cPrimLambda, 1 << ( m_uiLog2BlkSize << 1 ), 0, cRate );
#endif
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
