/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoEncoderLibPch.h"
#include "ResTrafoPow2Estimator.h"
#include "SubdivisionEstimator.h"
#include <math.h>

// video namespace begin
VIDEO_NAMESPACE_BEGIN


ResTrafoPow2Estimator::ResTrafoPow2Estimator( const PlanePackageParameters&      rcPlaPaPars,
                                              const AccessUnitEncoderParameters& rcAUEstPars,
                                              SyntaxElementEncoderIf&            rcSEEncoderIf, 
                                              BitRateEstimatorIf&                rcBitRateEstIf,
                                              StateIf&                           rcStateIf, 
                                              RDOQ&                              rcRDOQ, 
                                              Lambda                             cLambda )
: ResTrafoPow2( rcPlaPaPars.getLog2MaxPow2TrafoBlkSize() )
, m_rcSEEncoderIf( rcSEEncoderIf )
, m_rcBitRateEstIf( rcBitRateEstIf )
, m_rcStateIf( rcStateIf )
, m_rcRDOQ( rcRDOQ )
, m_cLambda( cLambda )
, m_bResEstResetted( true )
, m_iMaxQP0ToCheck( MAX_QP )
, m_pcPaddedBlock( new AllocPicBlock<DPel>( PicBlockStats( PrimBlockSize( 1 << rcPlaPaPars.getLog2MaxPow2TrafoBlkSize(), 1 << rcPlaPaPars.getLog2MaxPow2TrafoBlkSize() ), rcPlaPaPars.getSubsampling() ) ) )
, m_uiMinLog2TrafoBlkSize( rcPlaPaPars.getLog2MinPow2TrafoBlkSize() )
, m_uiMaxLog2PredBlkSize( rcPlaPaPars.getLog2MaxPredBlkSize() )
, m_bFastTSizeDecision( rcAUEstPars.getFastModeDecision() )
, m_bCheckFmdThreshold( rcAUEstPars.getFastModeDecision() )
{
  m_aauiTrafoDone           .resize( rcPlaPaPars.getLog2MaxPow2TrafoBlkSize() + 1 );
  m_aauiAllBelowFmdThreshold.resize( rcPlaPaPars.getLog2MaxPow2TrafoBlkSize() + 1 );
  for( UInt uiLog2BlkSize = m_uiMinLog2TrafoBlkSize; uiLog2BlkSize <= rcPlaPaPars.getLog2MaxPow2TrafoBlkSize(); uiLog2BlkSize++ )
  {
    std::vector<UInt>&  rauiTrafoDone             = m_aauiTrafoDone           [ uiLog2BlkSize ];
    std::vector<UInt>&  rauiAllBelowFmdThreshold  = m_aauiAllBelowFmdThreshold[ uiLog2BlkSize ];
    UInt                uiNumEntries              = 1 << ( ( m_uiMaxLog2PredBlkSize - uiLog2BlkSize ) << 1 );
    rauiTrafoDone           .resize( uiNumEntries );
    rauiAllBelowFmdThreshold.resize( uiNumEntries );
    for( UInt uiIdx = 0; uiIdx < uiNumEntries; uiIdx++ )
    {
      rauiTrafoDone           [ uiIdx ] = 0;
      rauiAllBelowFmdThreshold[ uiIdx ] = 1;
    }
  }
  if( m_bCheckFmdThreshold )
  {
    const int iInvQShift      = 16;     // 8..31-sm_uiQBits
    const int aiInvQFactor[6] =         // ( 1 << ( sm_uiQBits + iInvQShift ) ) / ( 100 * sm_aiQFactor[i] )
    {
      10486, 11799, 13110, 15081, 16393, 19021
    };
    m_aiQThresholds.resize( rcPlaPaPars.getNumPlanes() );
    for( UInt uiPlaneId = 0; uiPlaneId < rcPlaPaPars.getNumPlanes(); uiPlaneId++ )
    {
      int iPQp  = gMax<int>( MIN_QP, gMin<int>( MAX_QP, rcPlaPaPars.getPlaneQp( uiPlaneId ) ) );
      int iQST  = (int)rcAUEstPars.getFmdQStepPercentThreshold();
      m_aiQThresholds[ uiPlaneId ]  = ( iQST * aiInvQFactor[ iPQp % 6 ] ) >> ( iInvQShift - ( iPQp / 6 ) - ( m_rcRDOQ.isActive() ? 4 : 0 ) );
    }
  }
}

ResTrafoPow2Estimator::~ResTrafoPow2Estimator()
{
  while( m_apcTICoeffs.size() )
  {
    delete m_apcTICoeffs.back();
    m_apcTICoeffs.pop_back();
  }
  delete m_pcPaddedBlock;
}

ErrVal ResTrafoPow2Estimator::estimateResInfo( ResInfo *&rpcResInfo,
                                               bool& rbAllBelowFmdThreshold,
                                               const PredInfo *pcPredInfo,
                                               const QTNavi &rcNav,
                                               const PlanePackageParameters &rcPlaPaPars,
                                               const AccessUnitParameters &rcAUPars,
                                               const SliceData &rcContextSlice,
                                               const ResEstInfo &rcResEstInfo,
                                               const PicBlock<const DPel> &rcOrigResidual,
                                               bool &rbResidualConversionValid )
{
  ROT( rpcResInfo );
  ROF( rcResEstInfo.getResidualMode() == ResInfo::RES_TRAFO_POW2 );
  const ResEstInfoTrafoPow2* pcEstInfo = rcResEstInfo.getResEstInfoTrafoPow2();
  ROF( pcEstInfo );

  if( ! m_bResEstResetted && pcEstInfo->getQps()[0] > m_iMaxQP0ToCheck )
  {
    rpcResInfo                = 0;
    rbResidualConversionValid = false;
    return Err::m_nOK;
  }
  
  const UInt uiNumPlanes = rcOrigResidual.getStats().getNumPlanes();
  const UInt uiLog2MaxTransformSize = gMin( rcNav.getLog2BlockSize(), pcEstInfo->getLog2MaxBlockSize() ); // maximum size is bounded by residual block size and by user defined size
  const UInt uiLog2MinTransformSize = pcEstInfo->getSubdivide() ? pcEstInfo->getLog2MinBlockSize() : uiLog2MaxTransformSize; // if residual subdivision is selected, minimum transform size is defined by the user given value. else it is the size of the residual, so no subdivision will take place except on the picture boundaries.
  QuadtreeGroupStructure* pcTreeStructure = new QuadtreeGroupStructure( rcNav.getPrimBlockSize(), uiLog2MaxTransformSize, uiLog2MinTransformSize );
  ResInfoTrafoPow2 *pcResInfoTrafoPow2 = new ResInfoTrafoPow2( uiNumPlanes, *pcTreeStructure );

  RNOK( xInit( rcNav,
               pcEstInfo,
               rcOrigResidual.getStats() ) );

  // save initial rate estimation state
  State *pcRootState = NULL;
  RNOK( m_rcStateIf.getState( pcRootState ) );

  // Grow and prune tree and attach results to leaf nodes.
  bool bNonzeroCoeffsPresent  = false;
  rbAllBelowFmdThreshold = true;
  for( QTNavi cNavRes = pcTreeStructure->getFirstQuadtreeRoot(); !cNavRes.liesOutside(); cNavRes = pcTreeStructure->getNextQuadtreeRoot( cNavRes ) )
  {
    RDCost cRDCost( m_cLambda );
    RNOK( xGrowAndPruneTree( cNavRes, rcAUPars, cRDCost, bNonzeroCoeffsPresent, rbAllBelowFmdThreshold, pcPredInfo, uiLog2MinTransformSize, pcResInfoTrafoPow2->getTree(), rcNav, rcPlaPaPars, rcContextSlice, rcOrigResidual, pcEstInfo->getQps(), pcEstInfo->getDeadzone() ) );
  }

  if( ! bNonzeroCoeffsPresent )
  {
    m_iMaxQP0ToCheck = pcResInfoTrafoPow2->getQps()[0];
    RNOK( rcContextSlice.derivePredQPs( rcNav, pcResInfoTrafoPow2->getQps() ) );
  }
  else
  {
    pcResInfoTrafoPow2->getQps() = pcEstInfo->getQps();
  }

  // return to state before residual estimation
  RNOK( pcRootState->setAndDeleteState() );

  // return filled ResInfo object
  rbResidualConversionValid = true;
  rpcResInfo = pcResInfoTrafoPow2;
  return Err::m_nOK;
}

ErrVal ResTrafoPow2Estimator::xInit( const QTNavi &rcNav,
                                     const ResEstInfoTrafoPow2 *pcEstInfo,
                                     const PicBlockStats &rcOrigResidualStats )
{
  AOF( rcNav.getPrimBlockSize() == rcOrigResidualStats.getPrimBlockSize() );
  if( m_bResEstResetted )
  {
    ROF( pcEstInfo );
    const PicBlockStats cPaddedStats( PrimBlockSize( 1 << rcNav.getLog2BlockSize(), 1 << rcNav.getLog2BlockSize() ), rcOrigResidualStats.getSubsampling() );
    const UInt uiLog2MaxTransformSize = gMin( rcNav.getLog2BlockSize(), pcEstInfo->getLog2MaxBlockSize() ); // maximum size is bounded by residual block size and by user defined size
    const UInt uiLog2MinTransformSize = pcEstInfo->getSubdivide() ? pcEstInfo->getLog2MinBlockSize() : uiLog2MaxTransformSize; // if residual subdivision is selected, minimum transform size is defined by the user given value. else it is the size of the residual, so no subdivision will take place except on the picture boundaries.
    while( m_apcTICoeffs.size() )
    {
      delete m_apcTICoeffs.back();
      m_apcTICoeffs.pop_back();
    }

    for( UInt ui = 0; ui < uiLog2MinTransformSize; ui++ )
    {
      m_apcTICoeffs.push_back( 0 );
    }
    for( UInt ui = uiLog2MinTransformSize; ui <= uiLog2MaxTransformSize; ui++ )
    {
      m_apcTICoeffs.push_back( new AllocPicBlock<int>( cPaddedStats ) );
//        m_apcTICoeffs.push_back( new AllocPicBlock<int>( rcOrigResidualStats ) );
    }
    for( UInt uiLog2BlkSize = m_uiMinLog2TrafoBlkSize; uiLog2BlkSize <= uiLog2MaxTransformSize; uiLog2BlkSize++ )
    {
      std::vector<UInt>&  rauiTrafoDone             = m_aauiTrafoDone           [ uiLog2BlkSize ];
      std::vector<UInt>&  rauiAllBelowFmdThreshold  = m_aauiAllBelowFmdThreshold[ uiLog2BlkSize ];
      for( UInt uiIdx = 0; uiIdx < (UInt)rauiTrafoDone.size(); uiIdx++ )
      {
        rauiTrafoDone           [ uiIdx ] = 0;
        rauiAllBelowFmdThreshold[ uiIdx ] = 1;
      }
    }
    m_bResEstResetted = false;
    m_iMaxQP0ToCheck  = MAX_QP;
  }
  return Err::m_nOK;
}


ErrVal ResTrafoPow2Estimator::xGrowAndPruneTree( const QTNavi &rcNav,
                                                 const AccessUnitParameters &rcAUPars,
                                                 RDCost &rcCost,
                                                 bool& rbCoded,
                                                 bool& rbAllBelowFmdThreshold,
                                                 const PredInfo *pcPredInfo,
                                                 UInt uiLog2TrafoMinSize,
                                                 QuadtreeGroup<ResInfoTrafoPow2Leaf*> &rcResTree,
                                                 const QTNavi &rcPredBlkNavi,
                                                 const PlanePackageParameters &rcPlaPaPars,
                                                 const SliceData &rcCurrSlice,
                                                 const PicBlock<const DPel> &rcOriginalResidualPredBlk,
                                                 const std::vector<int> &raiQPs,
                                                 const Ratio& rcDeadzone  )
{
  ROF( rcNav.isLeafNode() );

  // check possibility of encoding the block with or without subdivision
  const bool bEncodeWithSubdivision = rcNav.getLog2BlockSize() > uiLog2TrafoMinSize;

  RDCost cSingleCost( m_cLambda );
  RDCost cChildCost( m_cLambda );
  
  State *pcRootState = NULL;
  State *pcSingleState = NULL;

  ResInfoTrafoPow2Leaf *pcSingleResult = NULL;
  bool bSingleCoded = rbCoded;
  bool bChildCoded  = rbCoded;
  bool bAllBelowFmdThresholdSingle = true;
  bool bAllBelowFmdThresholdChilds = true;

  // if possible, encode without subdivision
  if ( bEncodeWithSubdivision )
  {
    RNOK( m_rcStateIf.getState( pcRootState ) );
  }
  // encode without subdivision
  RNOK( xEncodeSubDivisionInfo( rcNav.getLog2BlockSize(), false, cSingleCost ) );
  RNOK( xEncodeSingleNode( rcNav, rcAUPars, pcSingleResult, &cSingleCost, pcPredInfo, bSingleCoded, bAllBelowFmdThresholdSingle, rcResTree, rcPredBlkNavi, rcPlaPaPars, rcCurrSlice, rcOriginalResidualPredBlk, raiQPs, rcDeadzone ) );
  bool bBreakTree = ( m_bFastTSizeDecision && ! bSingleCoded && bAllBelowFmdThresholdSingle );

  // if possible, encode with subdivision
  if( bEncodeWithSubdivision && ! bBreakTree )
  {
    RNOK( m_rcStateIf.getState( pcSingleState ) );
    RNOK( pcRootState->setAndDeleteState() );

    bool bTest = false;
    RNOK( rcResTree.getTreeStructure().appendChildNodes( rcNav, bTest ) );
    ROF( bTest );

    RNOK( xEncodeSubDivisionInfo( rcNav.getLog2BlockSize(), true, cChildCost ) );
    for ( UInt uiChildIndex = 0; uiChildIndex < 4; uiChildIndex++ )
    {
      const QTNavi &rcChildNav = rcNav.getChild( uiChildIndex );
      if( rcChildNav.isLeafNode() )
      {
        RNOK( xGrowAndPruneTree( rcChildNav, rcAUPars, cChildCost, bChildCoded, bAllBelowFmdThresholdChilds, pcPredInfo, uiLog2TrafoMinSize, rcResTree, rcPredBlkNavi, rcPlaPaPars, rcCurrSlice, rcOriginalResidualPredBlk, raiQPs, rcDeadzone ) );
      }
    }

    if( cSingleCost < cChildCost )
    {
      RNOK( QuadtreeGroupPointerDeleter::deleteChildNodePointers( rcResTree, rcNav ) );
      RNOK( rcResTree.getTreeStructure().removeChildNodes( rcNav ) );
      rcResTree.getT( rcNav ) = pcSingleResult;
      RNOK( pcSingleState->setAndDeleteState() );
      rcCost += cSingleCost;
      rbCoded = bSingleCoded;
      rbAllBelowFmdThreshold = ( rbAllBelowFmdThreshold && bAllBelowFmdThresholdSingle );
    }
    else
    {
      delete pcSingleResult;
      rcCost += cChildCost;
      rbCoded = bChildCoded;
      rbAllBelowFmdThreshold = ( rbAllBelowFmdThreshold && bAllBelowFmdThresholdChilds );
      RNOK( pcSingleState->deleteState() );
    }
  }
  else
  {
    rcResTree.getT( rcNav ) = pcSingleResult;
    rcCost += cSingleCost;
    rbCoded = bSingleCoded;
    rbAllBelowFmdThreshold = ( rbAllBelowFmdThreshold && bAllBelowFmdThresholdSingle );
    if( bEncodeWithSubdivision )
    {
      RNOK( pcRootState->deleteState() );
    }
  }

  return Err::m_nOK;
}


__inline ErrVal ResTrafoPow2Estimator::xEncodeSubDivisionInfo( UInt uiLog2BlockSize, bool bSubdiv, RDCost &rcRDCost )
{
  BitRate cRateBefore( 0 );
  cRateBefore = m_rcBitRateEstIf.getBitRateEstimate();
  RNOK( m_rcSEEncoderIf.processResTreeSubdivisionFlag( uiLog2BlockSize, bSubdiv ) );
  rcRDCost.addRate( m_rcBitRateEstIf.getBitRateEstimate() - cRateBefore );
  return Err::m_nOK;
}


__inline ErrVal ResTrafoPow2Estimator::xEncodeSingleNode( const QTNavi &rcNav,
                                                          const AccessUnitParameters &rcAUPars,
                                                          ResInfoTrafoPow2Leaf *&rpResult,
                                                          RDCost *pcCost,
                                                          const PredInfo *pcPredInfo,
                                                          bool& rbCoded,
                                                          bool& rbAllBelowFmdThreshold,
                                                          QuadtreeGroup<ResInfoTrafoPow2Leaf*> &rcResTree,
                                                          const QTNavi &rcPredBlkNavi,
                                                          const PlanePackageParameters &rcPlaPaPars,
                                                          const SliceData &rcCurrSlice,
                                                          const PicBlock<const DPel> &rcOriginalResidualPredBlk,
                                                          const std::vector<int> &raiQPs,
                                                          const Ratio& rcDeadzone )
{
  ROT( rpResult );
  // do Quantization
  PicBlock<const DPel> cOriginalResidual;
  const PrimBlockSize cFullSize( 1 << rcNav.getLog2BlockSize(), 1 << rcNav.getLog2BlockSize() );
  if( rcNav.getPrimBlockSize() != cFullSize )
  {
    PicBlock<DPel> cOriginalResidualTemp( PrimPixelPos( 0, 0 ), cFullSize, *m_pcPaddedBlock );
    cOriginalResidualTemp.assign( 0 );
    RNOK( PicBlock<DPel>( PrimPixelPos( 0, 0 ), rcNav.getPrimBlockSize(), cOriginalResidualTemp ).assign( PicBlock<const DPel>( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), rcOriginalResidualPredBlk ) ) );
    cOriginalResidual = cOriginalResidualTemp;
  }
  else
  {
    cOriginalResidual = PicBlock<const DPel>( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), rcOriginalResidualPredBlk );
  }

  rpResult = new ResInfoTrafoPow2Leaf( cOriginalResidual.getStats() );

#if CHECK_RDOQ_COST
  RDCost  cCostRDOQ ( m_cLambda );
  RDCost  cCostAdd  ( m_cLambda );
  if( m_rcRDOQ.isActive() )
  {
    RNOK( xQuantizeRDOQ( rcNav.getPrimPixelPos(), cFullSize, rcNav.getLog2BlockSize(), cOriginalResidual, *rpResult, rbAllBelowFmdThreshold, cCostRDOQ, raiQPs ) );
  }
  else
  {
    RNOK( xQuantize( rcNav.getPrimPixelPos(), cFullSize, rcNav.getLog2BlockSize(), cOriginalResidual, *rpResult, rbAllBelowFmdThreshold, raiQPs, cDeadzone ) );
  }
  for( UInt uiPlane = 0; uiPlane < cOriginalResidual.getStats().getNumPlanes(); uiPlane++ )
  {
    rbCoded = rbCoded || rpResult->getCbf( uiPlane );
  }
  BitRate cRateBefore( 0 );
  cRateBefore = m_rcBitRateEstIf.getBitRateEstimate();
  RNOK( m_rcSEEncoderIf.processResidualLeaf( rcNav, rcPredBlkNavi, rcPlaPaPars, rcAUPars, rcCurrSlice, rcResTree, *rpResult ) );
  cCostAdd.addRate( m_rcBitRateEstIf.getBitRateEstimate() - cRateBefore );
  AllocPicBlock<DPel> cReconstructedResidual( cOriginalResidual.getStats() );
  xReconstruct( *rpResult, cReconstructedResidual, raiQPs );
  cCostAdd.addDistortion( cOriginalResidual, cReconstructedResidual, EXTRA_PRECISION_BITS );
  double  dDistRDOQ = (double)cCostRDOQ.getDistortion();
  double  dDistAdd  = (double)cCostAdd .getDistortion();
  double  dRateRDOQ = cCostRDOQ.getBitRate().getValueDouble();
  double  dRateAdd  = cCostAdd .getBitRate().getValueDouble();
  double  dCostRDOQ = dDistRDOQ + m_cLambda.getValueDouble() * dRateRDOQ;
  double  dCostAdd  = dDistAdd  + m_cLambda.getValueDouble() * dRateAdd;
  double  dPerDist  = 100.0 * ( dDistRDOQ - dDistAdd ) / dDistAdd;
  double  dPerRate  = 100.0 * ( dRateRDOQ - dRateAdd ) / dRateAdd;
  double  dPerCost  = 100.0 * ( dCostRDOQ - dCostAdd ) / dCostAdd;
  printf("\n%6.2lf%% [D:%6.2lf%%  R:%6.2lf%%]  (%8.0lf - %8.0lf)  (%8.3lf - %8.3lf)", dPerCost, dPerDist, dPerRate, dDistRDOQ, dDistAdd, dRateRDOQ, dRateAdd );
  fflush( stdout );
  *pcCost += cCostAdd;
#else
  if( m_rcRDOQ.isActive() )
  {
    RDCost cCostRDOQ( m_cLambda );
    RNOK( xQuantizeRDOQ( rcNav.getPrimPixelPos(), cFullSize, rcNav.getLog2BlockSize(), cOriginalResidual, *rpResult, rbAllBelowFmdThreshold, cCostRDOQ, raiQPs ) );
    for( UInt uiPlane = 0; uiPlane < cOriginalResidual.getStats().getNumPlanes(); uiPlane++ )
    {
      rbCoded = rbCoded || rpResult->getCbf( uiPlane );
    }
    if( m_rcRDOQ.skipIDCT() )
    {
      if( rcNav.getPrimBlockSize() != cFullSize )
      {
        // hack for padded blocks (border): adapt uiNumSamples in cCostRDOQ to PrimBlockSize
        UInt uiNumSamples = 0;
        PicBlockStats cStatsWithoutPadding( rcNav.getPrimBlockSize(), cOriginalResidual.getStats().getSubsampling() );
        for( UInt uiPlane = 0; uiPlane < cStatsWithoutPadding.getNumPlanes(); uiPlane++ )
        {
          BlockSize cPlaneSize = cStatsWithoutPadding.getPrimBlockSize() * cStatsWithoutPadding.getSubsampling()[uiPlane];
          uiNumSamples += cPlaneSize.getX() * cPlaneSize.getY();
        }
        *pcCost += RDCost( cCostRDOQ.getLambda(), uiNumSamples, cCostRDOQ.getDistortion(), cCostRDOQ.getBitRate() );
      }
      else
      {
        *pcCost += cCostRDOQ;
      }      
      return Err::m_nOK;
    }
  }
  else
  {
    RNOK( xQuantize( rcNav.getPrimPixelPos(), cFullSize, rcNav.getLog2BlockSize(), cOriginalResidual, *rpResult, rbAllBelowFmdThreshold, raiQPs, rcDeadzone ) );
    for( UInt uiPlane = 0; uiPlane < cOriginalResidual.getStats().getNumPlanes(); uiPlane++ )
    {
      rbCoded = rbCoded || rpResult->getCbf( uiPlane );
    }
  }

  // calculate Rate
  BitRate cRateBefore( 0 );
  cRateBefore = m_rcBitRateEstIf.getBitRateEstimate();
  RNOK( m_rcSEEncoderIf.processResidualLeaf( rcNav, rcPredBlkNavi, rcPlaPaPars, rcAUPars, rcCurrSlice, rcResTree, *rpResult ) );
  pcCost->addRate( m_rcBitRateEstIf.getBitRateEstimate() - cRateBefore );

  // calculate Distortion
  AllocPicBlock<DPel> cReconstructedResidual( cOriginalResidual.getStats() );
  RNOK( xReconstruct( *rpResult, cReconstructedResidual, raiQPs ) );

  if( rcNav.getPrimBlockSize() != cFullSize )
  {
    // do not calculate distortion for padded area outside the image (right or bottom right)
    PicBlock<const DPel> cOrigResWithoutPadding( PrimPixelPos( 0, 0 ), rcNav.getPrimBlockSize(), cOriginalResidual );
    PicBlock<const DPel> cRecResWithoutPadding ( PrimPixelPos( 0, 0 ), rcNav.getPrimBlockSize(), cReconstructedResidual );
    pcCost->addDistortion( cOrigResWithoutPadding, cRecResWithoutPadding, EXTRA_PRECISION_BITS );
  }
  else
  {
    pcCost->addDistortion( cOriginalResidual, cReconstructedResidual, EXTRA_PRECISION_BITS );
  }
#endif

  return Err::m_nOK;
}


ErrVal ResTrafoPow2Estimator::xReconstruct( const ResInfoTrafoPow2Leaf &rcQuantized, const PicBlock<DPel> &rcReconstruction, const std::vector<int> &raiQPs )
{
  UInt uiNumPlanes = rcReconstruction.getStats().getNumPlanes();
  for( UInt uiPlane = 0; uiPlane < uiNumPlanes; uiPlane ++ )
  {
    if ( rcQuantized.getCbf( uiPlane ) )
    {
      AllocBlock<int> cCoeffs( rcReconstruction.getPlane( uiPlane ).getSize() );
      RNOK( xFastIScale( raiQPs[uiPlane], rcQuantized.getCoeff().getPlane( uiPlane ), cCoeffs ) );
      RNOK( xFast2dIDCT( cCoeffs, rcReconstruction.getPlane( uiPlane ), false ) );
    }
    else
    {
      rcReconstruction.getPlane( uiPlane ).assign( 0 );
    }
  }
  return Err::m_nOK;
}


ErrVal ResTrafoPow2Estimator::xQuantize( PrimPixelPos cPos,
                                         PrimBlockSize cSize,
                                         UInt uiLog2BlockSize,
                                         const PicBlock<const DPel> &rcOriginal,
                                         ResInfoTrafoPow2Leaf &rcQuantized,
                                         bool& rbAllBelowFmdThreshold,
                                         const std::vector<int> &raiQPs,
                                         const Ratio& cDeadzone )
{ 
  UInt  uiUniBlkIdx   = ( ( cPos.getY() >> uiLog2BlockSize ) << ( m_uiMaxLog2PredBlkSize - uiLog2BlockSize ) ) + ( cPos.getX() >> uiLog2BlockSize );
  UInt& ruiTrafoDone  = m_aauiTrafoDone           [ uiLog2BlockSize ][ uiUniBlkIdx ];
  UInt& ruiAllBelow   = m_aauiAllBelowFmdThreshold[ uiLog2BlockSize ][ uiUniBlkIdx ];
  ROT( !ruiTrafoDone && !ruiAllBelow );

  for( UInt uiPlane = 0; uiPlane < rcOriginal.getStats().getNumPlanes(); uiPlane ++ )
  {
    const PicBlock<int>&  rcPicBlock  = *m_apcTICoeffs[ uiLog2BlockSize ];
    const Subsampling&    rcSub       = rcPicBlock.getStats().getSubsampling()[ uiPlane ];
    PixelPos              cPixPos     = cPos  * rcSub;
    BlockSize             cBlkSize    = cSize * rcSub;
    const Block<int>      cTCoeffs    ( cPixPos, cBlkSize, rcPicBlock.getPlane( uiPlane ) );
    if( ! ruiTrafoDone )
    {
      RNOK( xFast2dDCT( rcOriginal.getPlane( uiPlane ), cTCoeffs ) );
    }
    RNOK( xFastQuant( raiQPs[uiPlane], cDeadzone, cTCoeffs, rcQuantized.getCoeff().getPlane( uiPlane ), rcQuantized.getCbf( uiPlane ) ) );
    if( m_bCheckFmdThreshold && ! ruiTrafoDone )
    {
      if( rcQuantized.getCbf( uiPlane ) )
      {
        ruiAllBelow = 0;
      }
      else
      {
        RNOK( xCheckCoeffs( ruiAllBelow, cTCoeffs, m_aiQThresholds[ uiPlane ] ) );
      }
    }
  }
  ruiTrafoDone = 1;
  if( m_bCheckFmdThreshold && rbAllBelowFmdThreshold && ruiAllBelow == 0 )
  {
    rbAllBelowFmdThreshold = false;
  }
  return Err::m_nOK;
}


ErrVal ResTrafoPow2Estimator::xQuantizeRDOQ( PrimPixelPos cPos,
                                             PrimBlockSize cSize,
                                             UInt uiLog2BlockSize,
                                             const PicBlock<const DPel> &rcOriginal,
                                             ResInfoTrafoPow2Leaf &rcQuantized,
                                             bool& rbAllBelowFmdThreshold,
                                             RDCost& rcRdCost,
                                             const std::vector<int> &raiQPs )
{ 
  UInt  uiUniBlkIdx   = ( ( cPos.getY() >> uiLog2BlockSize ) << ( m_uiMaxLog2PredBlkSize - uiLog2BlockSize ) ) + ( cPos.getX() >> uiLog2BlockSize );
  UInt& ruiTrafoDone  = m_aauiTrafoDone           [ uiLog2BlockSize ][ uiUniBlkIdx ];
  UInt& ruiAllBelow   = m_aauiAllBelowFmdThreshold[ uiLog2BlockSize ][ uiUniBlkIdx ];
  ROT( !ruiTrafoDone && !ruiAllBelow );

  for( UInt uiPlane = 0; uiPlane < rcOriginal.getStats().getNumPlanes(); uiPlane ++ )
  {
    UInt  uiCBFCtx  = 0;

    const PicBlock<int>&  rcPicBlock  = *m_apcTICoeffs[ uiLog2BlockSize ];
    const Subsampling&    rcSub       = rcPicBlock.getStats().getSubsampling()[ uiPlane ];
    PixelPos              cPixPos     = cPos  * rcSub;
    BlockSize             cBlkSize    = cSize * rcSub;
    const Block<int>      cTCoeffs    ( cPixPos, cBlkSize, rcPicBlock.getPlane( uiPlane ) );
    if( ! ruiTrafoDone )
    {
      RNOK( xFast2dDCT( rcOriginal.getPlane( uiPlane ), cTCoeffs ) );
    }
    RNOK( m_rcRDOQ.quantizePow2( raiQPs[uiPlane], uiLog2BlockSize, uiPlane, uiCBFCtx, cTCoeffs, rcQuantized.getCoeff().getPlane( uiPlane ), rcQuantized.getCbf( uiPlane ), rcRdCost ) );
    if( m_bCheckFmdThreshold && ! ruiTrafoDone )
    {
      if( rcQuantized.getCbf( uiPlane ) )
      {
        ruiAllBelow = 0;
      }
      else
      {
        RNOK( xCheckCoeffs( ruiAllBelow, cTCoeffs, m_aiQThresholds[ uiPlane ] ) );
      }
    }
  }
  ruiTrafoDone = 1;
  if( m_bCheckFmdThreshold && rbAllBelowFmdThreshold && ruiAllBelow == 0 )
  {
    rbAllBelowFmdThreshold = false;
  }
  return Err::m_nOK;
}


void ResTrafoPow2Estimator::xFast1dDCT( int* paiData, UInt uiInc, UInt uiLog2TSize, UInt uiInsideShift, UInt uiFinalShift )
{
  UInt  uiCoeffAccessShift  = sm_uiLog2MaxTSize - uiLog2TSize;
  UInt  uiTSize             = 1 << uiLog2TSize;
  UInt  uiDataTSize         = uiInc << uiLog2TSize;
  UInt  uiDoubleDataTSize   = uiDataTSize << 1;
  int   iMinAuxShiftMinus1  = (int)uiInsideShift - (int)uiLog2TSize;
  int   iInsideAdd          = ( 1 << uiInsideShift ) >> 1;

  //===== generate coefficients for multiplication =====
  for( UInt uiLog2X = 1; uiLog2X <= uiLog2TSize; uiLog2X++ )
  {
    UInt  uiStep      = 1 << uiLog2X;
    UInt  uiLast      = ( uiDoubleDataTSize >> uiLog2X ) - uiInc;
    UInt  uiAuxShift  = gMax( 0, iMinAuxShiftMinus1 + (int)uiLog2X );
    int   iAuxAdd     = ( 1 << uiAuxShift ) >> 1;
    for( UInt uiIdx = uiStep >> 1, uiFIdx = 0, uiBIdx = uiLast; uiIdx < uiTSize; uiIdx += uiStep, uiFIdx += uiInc, uiBIdx -= uiInc )
    {
      m_aiAuxArray [ uiIdx  ]   = (int)paiData[ uiFIdx ] - (int)paiData[ uiBIdx ];
      m_aiAuxArray [ uiIdx  ]  += iAuxAdd;
      m_aiAuxArray [ uiIdx  ] >>= uiAuxShift;
      paiData      [ uiFIdx ]  += paiData[ uiBIdx ];                      
    }
  }

  //===== DC coefficient =====
  paiData[ 0 ]  += iInsideAdd;
  paiData[ 0 ] >>= uiInsideShift;
  paiData[ 0 ]  *= m_aiDCTCoeffArray[ 0 ];
  uiInsideShift  = gMin( uiLog2TSize - 1, uiInsideShift );
  iInsideAdd     = ( 1 << uiInsideShift ) >> 1;

  //===== AC coefficients =====
  for( UInt uiLog2X = 1; uiLog2X <= uiLog2TSize; uiLog2X++ )
  {
    UInt  uiStep      = 1      << uiLog2X;
    UInt  uiStepData  = uiInc  << uiLog2X;
    UInt  uiStartIdx  = uiStep >> 1;
    for( UInt uiIdx = uiStartIdx, uiDataIdx = uiStepData >> 1; uiIdx < uiTSize; uiIdx += uiStep, uiDataIdx += uiStepData )
    {
      UInt  uiCoeffIdx              = uiIdx << uiCoeffAccessShift;
      UInt  uiCoeffIdxAdd           = uiCoeffIdx << 1;
      paiData        [ uiDataIdx ]  = ( m_aiAuxArray[ uiStartIdx ] * m_aiDCTCoeffArray[ uiCoeffIdx ] + iInsideAdd ) >> uiInsideShift; 
      for( UInt uiAuxIdx = uiStartIdx + uiStep; uiAuxIdx < uiTSize; uiAuxIdx += uiStep )
      {
        uiCoeffIdx                 += uiCoeffIdxAdd;
        paiData      [ uiDataIdx ] += ( m_aiAuxArray[ uiAuxIdx ] * m_aiDCTCoeffArray[ uiCoeffIdx & sm_uiCoeffAccessMask ] + iInsideAdd ) >> uiInsideShift; 
      }
    }
    if( uiInsideShift )
    {
      uiInsideShift--;
      iInsideAdd >>= 1;
    }
  }

  //===== final scale =====
  if( uiFinalShift )
  {
    int iAdd = ( 1 << uiFinalShift ) >> 1;
    for( UInt uiDataIdx = 0; uiDataIdx < uiDataTSize; uiDataIdx += uiInc )
    {
      paiData[ uiDataIdx ]   += iAdd;
      paiData[ uiDataIdx ]  >>= uiFinalShift;
    }
  }
}

ErrVal ResTrafoPow2Estimator::xFast2dDCT( const Block<const DPel>& rcIn, const Block<int>& rcOut )
{
  //===== get Log2TSize =====
  UInt  uiStride    = rcOut.getStride();
  int*  paiTCoeff   = &rcOut.getEl( 0 );
  UInt  uiLog2TSize = 0; while( rcIn.getSize().getX() > ( 1 << uiLog2TSize ) ) uiLog2TSize++;
  UInt  uiTSize     = 1 << uiLog2TSize;
  int   iFullShift  =          int(   uiLog2TSize        + ( sm_uiDCTPrecision << 1 ) ) + EXTRA_PRECISION_BITS - ( m_rcRDOQ.isActive() ? 4 : 0 );
  int   iIShift1    = gMax( 0, int(   uiLog2TSize        +   sm_uiDCTPrecision        ) + EXTRA_PRECISION_BITS - 20 );
  int   iFShift1    = gMax( 0, int( ( uiLog2TSize << 1 ) +   sm_uiDCTPrecision        ) + EXTRA_PRECISION_BITS - 20 - iIShift1 );
  int   iIShift2    = gMax( 0, int( ( uiLog2TSize << 1 ) + ( sm_uiDCTPrecision << 1 ) ) + EXTRA_PRECISION_BITS - 17 - iIShift1 - iFShift1 );
  int   iFShift2    = iFullShift - iIShift1 - iFShift1 - iIShift2;
  ROF( rcIn .getSize().getX() == int( uiTSize ) );
  ROF( rcIn .getSize().getY() == int( uiTSize ) );
  ROF( rcOut.getSize().getX() == int( uiTSize ) );
  ROF( rcOut.getSize().getY() == int( uiTSize ) );
  //===== copy input data =====
  {
    int*        paiData     = paiTCoeff;
    UInt        uiInStride  = rcIn.getStride();
    const DPel* paiInData   = &rcIn.getEl( 0 );
    for( UInt uiRow = 0; uiRow < uiTSize; uiRow++, paiData += uiStride, paiInData += uiInStride )
    {
      for( UInt uiCol = 0; uiCol < uiTSize; uiCol++ )
      {
        paiData[ uiCol ] = paiInData[ uiCol ];
      }
    }
  }
  //===== vertical transform =====
  {
    int* paiData = paiTCoeff;
    for( UInt uiCol = 0; uiCol < uiTSize; uiCol++, paiData++ )
    {
      xFast1dDCT( paiData, uiStride, uiLog2TSize, (UInt)iIShift1, (UInt)iFShift1 );
    }
  }
  //===== horizontal transform =====
  {
    int* paiData = paiTCoeff;
    for( UInt uiRow = 0; uiRow < uiTSize; uiRow++, paiData += uiStride )
    {
      xFast1dDCT( paiData, 1, uiLog2TSize, (UInt)iIShift2, (UInt)iFShift2 );
    }
  }
  return Err::m_nOK;
}


ErrVal ResTrafoPow2Estimator::xFastQuant( int iQp, const Ratio &rcDeadzone, const Block<const int> &rcIn, const Block<int> &rcOut, bool &rbOutNonzero )
{
  ROT( iQp < 0 );
  UInt        uiNumCols   = rcIn .getSize().getX();
  UInt        uiNumRows   = rcIn .getSize().getY();
  UInt        uiInStride  = rcIn .getStride();
  UInt        uiOutStride = rcOut.getStride();
  const int*  paiInData   = &rcIn .getEl( 0 );
  int*        paiOutData  = &rcOut.getEl( 0 );
  int         iShift      = sm_uiQBits + ( iQp / 6 );
  int         iFactor     = sm_aiQFactor [ iQp % 6 ];
  int         iAdd        = ( ( 1 << iShift ) * rcDeadzone.getNumerator() + ( rcDeadzone.getDenominator() >> 1 ) ) / rcDeadzone.getDenominator();
  rbOutNonzero            = false;

  for( UInt uiRow = 0; uiRow < uiNumRows; uiRow++, paiInData += uiInStride, paiOutData += uiOutStride )
  {
    for( UInt uiCol = 0; uiCol < uiNumCols; uiCol++ )
    {
      int iSign           = ( paiInData[ uiCol ] < 0 ? -1 : 1 );
      paiOutData[ uiCol ] = (int)( ( paiInData[ uiCol ] * iSign * iFactor + iAdd ) >> iShift );
      if( paiOutData[ uiCol ] )
      {
        paiOutData[ uiCol ] *= iSign;
        rbOutNonzero         = true;
      }
    }
  }
  return Err::m_nOK;
}


ErrVal
ResTrafoPow2Estimator::xCheckCoeffs( UInt& ruiAllBelowFMDThreshold, const Block<const int> &rcIn, int iThreshold ) const
{
  ROFRS( ruiAllBelowFMDThreshold, Err::m_nOK );
  UInt        uiNumCols =  rcIn.getSize().getX();
  UInt        uiNumRows =  rcIn.getSize().getY();
  UInt        uiStride  =  rcIn.getStride();
  const int*  paiData   = &rcIn.getEl( 0 );

  for( UInt uiRow = 0; uiRow < uiNumRows; uiRow++, paiData += uiStride )
  {
    for( UInt uiCol = 0; uiCol < uiNumCols; uiCol++ )
    {
      if( paiData[ uiCol ] < -iThreshold || paiData[ uiCol ] > iThreshold )
      {
        ruiAllBelowFMDThreshold = 0;
        return Err::m_nOK;
      }
    }
  }
  return Err::m_nOK;
}



// video namespace end
VIDEO_NAMESPACE_END
