/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _RESETTABLESTREAM_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _RESETTABLESTREAM_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/InStreamIf.h"
#include "StateIf.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class T>
class ResettableStream
  : public StateIf
  , public InStreamIf<const T&>
{
public:

  ResettableStream( InStreamIf<const T&> &rcOutputIf )
    : m_rcOutputIf( rcOutputIf )
    , m_pcRootNode( new Node( NULL ) )
  {
    m_pcActiveNode = m_pcRootNode;
  }

  virtual ~ResettableStream() { delete m_pcRootNode; }

  ErrVal getState( State *&rpcState );

  ErrVal pushBack( const T &rcT );

private:
  ErrVal xPruneRoot();

  class Node
  {
  public:
    Node( Node *pcParent ) : m_bHasState( false ), m_pcParent( pcParent ) {}

    static ErrVal addNode( Node *&rpcActiveNode )
    {
      Node *pcNewNode = new Node( rpcActiveNode );
      if( rpcActiveNode )
      {
        rpcActiveNode->m_apcBranches.push_front( pcNewNode );
        pcNewNode->m_papcThisInParent = rpcActiveNode->m_apcBranches.begin();
      }
      rpcActiveNode = pcNewNode;
      return Err::m_nOK;
    }

    ErrVal deleteNode()
    {
      ROT( m_bHasState );
      if( m_pcParent )
      {
        m_pcParent->m_apcBranches.erase( m_papcThisInParent );
      }
      ROF( m_apcBranches.empty() || m_apcBranches.size() == 1 );
      if( m_apcBranches.size() == 1 )
      {
        m_apcBranches.front()->m_pcParent = NULL;
      }
      delete this;
      return Err::m_nOK;
    }

    bool                                 m_bHasState;
    Node                                *m_pcParent;
    typename std::list<Node*>::iterator  m_papcThisInParent;
    std::vector<T>                       m_acT;
    std::list<Node*>                     m_apcBranches;
  };

  class ResStrState
    : public State
  {
  public:
    ResStrState( Node *pcActiveNode, ResettableStream *pcStatesObject ) : m_pcActiveNode( pcActiveNode ), m_pcStatesObject( pcStatesObject ) {}

    Node *m_pcActiveNode;

  private:
    ResettableStream *m_pcStatesObject;

    ErrVal setState();
    ErrVal deleteState();
  };

  InStreamIf<const T&> &m_rcOutputIf;
  Node                 *m_pcRootNode;
  Node                 *m_pcActiveNode;
};


template<class T>
ErrVal ResettableStream<T>::getState( State *&rpcState )
{
  ROT( m_pcActiveNode->m_bHasState );
  m_pcActiveNode->m_bHasState = true;
  rpcState = new ResStrState( m_pcActiveNode, this );
  RNOK( Node::addNode( m_pcActiveNode ) );
  return Err::m_nOK;
}


template<class T>
ErrVal ResettableStream<T>::ResStrState::setState()
{
  ROT( m_pcStatesObject->m_pcActiveNode->m_bHasState );
  ROF( m_pcStatesObject->m_pcActiveNode->m_apcBranches.empty() );
  ROT( NULL == m_pcStatesObject->m_pcActiveNode );
  while( m_pcStatesObject->m_pcActiveNode && !m_pcStatesObject->m_pcActiveNode->m_bHasState && m_pcStatesObject->m_pcActiveNode->m_apcBranches.empty() )
  {
    Node *pcParent = m_pcStatesObject->m_pcActiveNode->m_pcParent;
    RNOK( m_pcStatesObject->m_pcActiveNode->deleteNode() );
    m_pcStatesObject->m_pcActiveNode = pcParent;
  }

  RNOK( m_pcStatesObject->xPruneRoot() );

  ROT( NULL == m_pcActiveNode );

  m_pcStatesObject->m_pcActiveNode = m_pcActiveNode;
  RNOK( Node::addNode( m_pcStatesObject->m_pcActiveNode ) );
  return Err::m_nOK;
}


template<class T>
ErrVal ResettableStream<T>::ResStrState::deleteState()
{
  Node *pcNode = m_pcActiveNode;

  ROF( pcNode->m_bHasState );
  pcNode->m_bHasState = false;
  while( pcNode->m_apcBranches.empty() && !pcNode->m_bHasState )
  {
    Node *pcParent = pcNode->m_pcParent;
    RNOK( pcNode->deleteNode() );
    pcNode = pcParent;
  }

  RNOK( m_pcStatesObject->xPruneRoot() );
  delete this;
  return Err::m_nOK;
}


template<class T>
ErrVal ResettableStream<T>::xPruneRoot()
{
  while( !m_pcRootNode->m_bHasState && m_pcRootNode->m_apcBranches.size() == 1 )
  {
    Node *pcChild = m_pcRootNode->m_apcBranches.front();
    if( !pcChild->m_acT.empty() )
    {
      for( UInt ui = 0; ui < UInt( pcChild->m_acT.size() ); ++ui )
      {
        m_rcOutputIf.pushBack( pcChild->m_acT[ui] );
      }
    }
    RNOK( m_pcRootNode->deleteNode() );
    m_pcRootNode = pcChild;
  }

  if( m_pcRootNode->m_apcBranches.empty() )
  {
    ROF( m_pcRootNode == m_pcActiveNode );
  }
  return Err::m_nOK;
}


template<class T>
ErrVal ResettableStream<T>::pushBack( const T &rcT )
{
  if( m_pcRootNode != m_pcActiveNode )
  {
    m_pcActiveNode->m_acT.push_back( rcT );
  }
  else
  {
    // nothing to store -> move data to output
    m_rcOutputIf.pushBack( rcT );
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _RESETTABLESTREAM_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
