/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_SEQUENCESTRUCTURE_H__268768B8_4D1D_484A_904E_586985833BAC__INCLUDED_)
#define AFX_SEQUENCESTRUCTURE_H__268768B8_4D1D_484A_904E_586985833BAC__INCLUDED_

#include <cctype>

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class FrameDescriptor
{
public:
  ErrVal initFrameDescriptor( const std::string&  rcString, UInt &ruiIncrement );
  UInt getMaxNumRefBlocksPerInterTranslationalPred() const { return m_uiMaxNumRefBlocksPerInterTranslationalPred; }
  std::vector<int> getAllowedRelativeRefPocs() const { return m_aiAllowedRelativeRefPocs; }
  UInt getSequenceStructureLayer() const { return m_uiLayer; }
  bool getStoreForRef() const { return m_bStoreForRef; }
  bool isIdr()          const { return m_bIsIDR; }

  void setSequenceStructureLayer( UInt uiNewLayer ) { m_uiLayer = uiNewLayer; }

private:
  UInt m_uiMaxNumRefBlocksPerInterTranslationalPred;
  bool m_bStoreForRef;
  bool m_bIsIDR;
  UInt m_uiLayer;
  std::vector<int> m_aiAllowedRelativeRefPocs;
};


class SequenceStructure
{
private:
  class SequencePart
  {
    DISABLE_DEFAULT_CPYCTOR_AND_ASOP( SequencePart );
  protected:
    SequencePart() : m_uiNumberOfRepetitions( 0 ) {}

  public:
    virtual ~SequencePart() {}

    virtual bool isLeafNode() const = 0;
    virtual SequencePart* getChildNode( UInt64 ui ) const { TOT( 1 ); return NULL; }
    virtual const FrameDescriptor* getFrameDescriptor( UInt64 ui ) const { TOT( 1 ); return NULL; }

    UInt64 getSize() const { TOT( isInfinitelyLong() ); return UInt64( m_uiNumberOfRepetitions ) * xGetSize(); }
    UInt64 getIncrement( UInt64 ui ) const { return ui - UInt64( xGetIdx( ui ) ) + UInt64( xGetIncrement( ui ) ); }
    bool isInfinitelyLong() const { return m_uiNumberOfRepetitions == TypeLimits<UInt>::m_iMax; }

    virtual UInt64 findIncrement( UInt64 uiIncrement, bool &rbSuccess ) const { TOT( 1 ); return 0; }

  protected:
    UInt xGetIdx( UInt64 ui ) const { return UInt( ui % xGetSize() ); }
    void xSetNumRep( UInt ui )
    {
      TOT( m_uiNumberOfRepetitions );
      TOF( ui );
      m_uiNumberOfRepetitions = ui;
    }

  private:
    virtual UInt   xGetIncrement( UInt64 ui ) const { TOT( 1 ); return 0; }
    virtual UInt64 xGetSize() const = 0;

    UInt m_uiNumberOfRepetitions;
  };


  class FrameSequencePart : public SequencePart
  {
  public:
    FrameSequencePart( const std::string& rcString );
    virtual ~FrameSequencePart() {}

    UInt64 xGetSize() const { return UInt64( m_acFrameDescriptors.size() ); }
    bool isLeafNode() const { return true; }
    const FrameDescriptor* getFrameDescriptor( UInt64 ui ) const { return &m_acFrameDescriptors[xGetIdx( ui )]; }
    UInt xGetIncrement( UInt64 ui ) const { return m_auiIncrements[xGetIdx( ui )]; }
    virtual UInt64 findIncrement( UInt64 uiIncrement, bool &rbSuccess ) const;

  private:
    ErrVal xCheck();

    std::vector<FrameDescriptor> m_acFrameDescriptors;
    std::vector<UInt> m_auiIncrements;
  };


  class GeneralSequencePart : public SequencePart
  {
  public:
    GeneralSequencePart( const std::string& rcString );
    virtual ~GeneralSequencePart() {}

    UInt64 xGetSize() const { return UInt64( m_apcSequenceParts.size() ); }
    bool isLeafNode() const { return false; }
    SequencePart* getChildNode( UInt64 ui ) const { return m_apcSequenceParts[xGetIdx( ui )]; }

  private:
    AutoDeletePtrVector<SequencePart> m_apcSequenceParts;
  };

public:
  SequenceStructure();
  virtual ~SequenceStructure();

  ErrVal init( const std::string& rcString );
  UInt getMaxAbsPocDiff( const UInt uiNumberOfFrames );

  class Iterator
  {
  private:
    class SequencePartWithPos
    {
    public:
      const SequencePart *m_pcSeqPart;
      UInt64 m_uiCurrPos;
    };

  public:
    Iterator() : m_cBasePoc( 0 ) {}
    virtual ~Iterator() {}

    Iterator( const SequenceStructure &r, PicOrderCnt cLayerChangeStartPoc, int iLayerOffset );

    FrameDescriptor getFrameDescriptor() const
    {
      if( m_iLayerOffset != 0 && getPoc() >= m_cLayerChangeStartPoc )
      {
        FrameDescriptor cModified = *xGetCurr().m_pcSeqPart->getFrameDescriptor( xGetCurr().m_uiCurrPos );
        cModified.setSequenceStructureLayer( cModified.getSequenceStructureLayer() + m_iLayerOffset );
        return cModified;
      }
      return *xGetCurr().m_pcSeqPart->getFrameDescriptor( xGetCurr().m_uiCurrPos );
    }

    PicOrderCnt getPoc() const { return m_cBasePoc + xGetCurr().m_pcSeqPart->getIncrement( xGetCurr().m_uiCurrPos ); }

    Iterator& operator++();
    Iterator& traverseByPocDiff( Int64 iPocDiff );
    Iterator getIterByPocDiff( Int64 iPocDiff ) const;

  private:
    SequencePartWithPos&       xGetCurr()       { return m_acSeqPartPath.back(); }
    const SequencePartWithPos& xGetCurr() const { return m_acSeqPartPath.back(); }

    void xGoToPreviousFrameSequencePart();
    void xGoToNextFrameSequencePart();

    void xGoToLeaf( bool bGoToRightmostLeaf );

    std::vector<SequencePartWithPos> m_acSeqPartPath;
    PicOrderCnt m_cBasePoc;
    PicOrderCnt m_cLayerChangeStartPoc;
    int m_iLayerOffset;
  };

private:
  SequencePart *m_pcSequencePart;
};


// video namespace end
VIDEO_NAMESPACE_END


#endif // !defined(AFX_SEQUENCESTRUCTURE_H__268768B8_4D1D_484A_904E_586985833BAC__INCLUDED_)
