/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoEncoderLibPch.h"
#include "SliceEstimator.h"
#include "IntraDCEstimator.h"
#include "IntraDirectionalEstimator.h"
#include "InterTranslationalEstimator.h"
#include "InterPlanePackageEstimator.h"
#include "ResTrafoPow2Estimator.h"
#include "ResTrafoSkipEstimator.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


ErrVal SliceEstimator::addIntraPars( VecPredEstInfo &racPars, const VecResEstInfo &racResEstInfo )
{
  std::vector<EstInfo* > acPars;
  acPars.push_back( new EstInfoIntraDC );

  acPars.push_back( new EstInfoIntraDirectional( PredInfoIntraDirectional::INTRA_WEST_SOUTHWEST,  EstInfoIntraDirectional::SMOOTH_PREDICTION ) );
  acPars.push_back( new EstInfoIntraDirectional( PredInfoIntraDirectional::INTRA_WEST,            EstInfoIntraDirectional::SMOOTH_PREDICTION ) );
  acPars.push_back( new EstInfoIntraDirectional( PredInfoIntraDirectional::INTRA_WEST_NORTHWEST,  EstInfoIntraDirectional::SMOOTH_PREDICTION ) );
  acPars.push_back( new EstInfoIntraDirectional( PredInfoIntraDirectional::INTRA_NORTHWEST,       EstInfoIntraDirectional::SMOOTH_PREDICTION ) );
  acPars.push_back( new EstInfoIntraDirectional( PredInfoIntraDirectional::INTRA_NORTH_NORTHWEST, EstInfoIntraDirectional::SMOOTH_PREDICTION ) );
  acPars.push_back( new EstInfoIntraDirectional( PredInfoIntraDirectional::INTRA_NORTH,           EstInfoIntraDirectional::SMOOTH_PREDICTION ) );
  acPars.push_back( new EstInfoIntraDirectional( PredInfoIntraDirectional::INTRA_NORTH_NORTHEAST, EstInfoIntraDirectional::SMOOTH_PREDICTION ) );
  acPars.push_back( new EstInfoIntraDirectional( PredInfoIntraDirectional::INTRA_NORTHEAST,       EstInfoIntraDirectional::SMOOTH_PREDICTION ) );
 
  // test best mode so far with toggled smoothing
  acPars.push_back( new EstInfoIntraDirectional( PredInfoIntraDirectional::IntraDirections( 0 ),  EstInfoIntraDirectional::REESTIMATE_BEST_WITH_TOGGLED_SMOOTHING ) );

  for( UInt ui=0; ui < acPars.size(); ui++ )
  {
    acPars[ui]->setVecResEstInfo( &racResEstInfo );
    racPars.push_back(acPars[ui]);
  }

  return Err::m_nOK;
}


ErrVal SliceEstimator::addInterPars( VecPredEstInfo &racPars,
                                     UInt uiMvAccuracy,
                                     UInt uiSearchModeFullPel,
                                     UInt uiSearchModeSubPel,
                                     bool bPrimPlaneOnlySearch,
                                     BlockSize cSearchRange,
                                     BlockSize cIterSearchRange,
                                     Lambda cLambdaMotion,
                                     const VecResEstInfo &racResEstInfo,
                                     UInt uiMaxNumRefBlocksPerInterTranslationalPred,
                                     UInt uiMaxNumIterations )
{
  EstInfo* p = new EstInfoInterTranslational( uiMvAccuracy, cSearchRange, cIterSearchRange, uiSearchModeFullPel, uiSearchModeSubPel, bPrimPlaneOnlySearch, cLambdaMotion, uiMaxNumRefBlocksPerInterTranslationalPred, uiMaxNumIterations );
  p->setVecResEstInfo( &racResEstInfo );
  racPars.push_back( p );
  return Err::m_nOK;
}

ErrVal SliceEstimator::addInterPlaPaPars( VecPredEstInfo&      racPars,
                                          const VecResEstInfo& racResEstInfo,
                                          const PlanePackage*  pcRefPlanePackage )
{
  ROF( pcRefPlanePackage );
  EstInfo* p = new EstInfoInterPlanePackage( *pcRefPlanePackage );
  p->setVecResEstInfo( &racResEstInfo );
  racPars.push_back( p );
  return Err::m_nOK;
}

ErrVal SliceEstimator::addResEstInfos( VecResEstInfo &racResEstInfo,
                                       const Ratio& cDeadzone,
                                       const std::vector<int> &raiQps,
                                       UInt uiMaxAbsQPDelta,
                                       UInt uiLog2MaxTransformBlockSize,
                                       UInt uiLog2MinTransformBlockSize,
                                       bool bTrafoPow2Subdivide,
                                       UInt uiQuantizers,
                                       bool bEnableResidualSkip )
{
  std::vector<int> cQps( raiQps );
  int iMaxAbsDQP = (int)uiMaxAbsQPDelta;
  for( int iDQP = -iMaxAbsDQP; iDQP <= iMaxAbsDQP; iDQP++ )
  {
    for( UInt uiIdx = 0; uiIdx < cQps.size(); uiIdx++ ) cQps[ uiIdx ] = gMax<int>( MIN_QP, gMin<int>( MAX_QP, raiQps[ uiIdx ] + iDQP ) );
    if( uiQuantizers & ResInfo::RES_TRAFO_POW2 )
    {
      racResEstInfo.push_back( new ResEstInfoTrafoPow2( cQps, cDeadzone, uiLog2MaxTransformBlockSize, uiLog2MinTransformBlockSize, bTrafoPow2Subdivide ) );
    }
    if( bEnableResidualSkip )
    {
      racResEstInfo.push_back( new ResEstInfoTrafoSkip( cQps ) );
    }
  }
  ROT( racResEstInfo.empty() );
  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END
