/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _SUBDIVISIONESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _SUBDIVISIONESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "SingleNodeEstimatorIf.h"
#include "VideoCommonLib/TraceCommon.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class t_EstResult, class t_ContextModel>
class SubdivisionEstimator
  : protected TraceBase
{
public:
  SubdivisionEstimator( const AccessUnitEncoderParameters& rcAUEstPars,
                        QuadtreeGroupStructure &rcTree,
                        SingleNodeEstimatorIf<t_EstResult> &rcSingleNodeEstIf,
                        ProbabilityAllocatorBase<t_ContextModel> &rcProbabilityAllocator,
                        ContextStore<t_ContextModel>* pcContextStore,
                        const ContextStore<t_ContextModel> *pcRefCtxStore,
                        Lambda cLambda,
                        StateIf *pcState,
                        TraceCommon *pcTrace = NULL )
    : m_rcTree( rcTree )
    , m_rcSingleNodeEstIf( rcSingleNodeEstIf )
    , m_rcProbabilityAllocator( rcProbabilityAllocator )
    , m_pcContextStore( pcContextStore )
    , m_pcRefCtxStore( pcRefCtxStore )
    , m_cLambda( cLambda )
    , m_bFastModeDecision( rcAUEstPars.getFastModeDecision() )
    , m_pcState( pcState )
  {
    TraceBase::setTraceCommonPtr( pcTrace );
  }

  ErrVal estimateTree();

private:
  ErrVal xProcessNode( const QTNavi &rcNav, RDCost &rcCost );

  QuadtreeGroupStructure &m_rcTree;
  SingleNodeEstimatorIf<t_EstResult> &m_rcSingleNodeEstIf;
  ProbabilityAllocatorBase<t_ContextModel> &m_rcProbabilityAllocator;
  ContextStore<t_ContextModel>* m_pcContextStore;
  const ContextStore<t_ContextModel> *m_pcRefCtxStore;
  Lambda m_cLambda;
  const bool m_bFastModeDecision;
  StateIf *m_pcState;
};


template<class t_EstResult, class t_ContextModel>
ErrVal SubdivisionEstimator<t_EstResult, t_ContextModel>::estimateTree()
{
  UInt uiIdx = 0;
  UInt uiSlice = 0;
  t_ContextModel *pacCtxStates;

  if ( m_pcRefCtxStore == NULL && m_pcContextStore != NULL )
  {
    RNOK( m_rcProbabilityAllocator.getCtxStates( pacCtxStates ) );
    RNOK( m_pcContextStore->setCtxInitStates( pacCtxStates ) );
  }

  for( QTNavi cNav = m_rcTree.getFirstQuadtreeRoot(); !cNav.liesOutside(); cNav = m_rcTree.getNextQuadtreeRoot( cNav ) )
  {
    TRACE_TO_STR( "TreeBlock Nr " );
    TRACE_TO_STR( uiIdx );
    TRACE_TO_STR( " (" );
    TRACE_TO_STR( cNav.getPrimPixelPos().getX() );
    TRACE_TO_STR( ", " );
    TRACE_TO_STR( cNav.getPrimPixelPos().getY() );
    TRACE_TO_STR( ")" );
    TRACE_TO_STR( std::endl );
    uiIdx++;

    if ( m_rcTree.getEntropySlice( cNav ) != uiSlice || cNav == m_rcTree.getFirstQuadtreeRoot() )
    {
      uiSlice = m_rcTree.getEntropySlice( cNav );

      if ( m_pcRefCtxStore != NULL && m_pcContextStore != NULL && !m_pcContextStore->isCtxStates( uiSlice ) )
      {
        RNOK( m_pcRefCtxStore->getCtxStates( uiSlice, pacCtxStates ) );
        RNOK( m_rcProbabilityAllocator.setCtxStates( pacCtxStates ) );
      }
      else if ( m_pcContextStore != NULL )
      {
        RNOK( m_pcContextStore->getCtxStates( uiSlice, pacCtxStates ) );
        RNOK( m_rcProbabilityAllocator.setCtxStates( pacCtxStates ) );
      }
    }

    RDCost cRDCost( m_cLambda );
    RNOK( xProcessNode( cNav, cRDCost ) );

    if ( m_rcTree.getNextQuadtreeRoot( cNav ).liesOutside() || m_rcTree.getEntropySlice( m_rcTree.getNextQuadtreeRoot( cNav ) ) != uiSlice )
    {
      if ( m_pcContextStore != NULL )
      {
        RNOK( m_rcProbabilityAllocator.getCtxStates( pacCtxStates ) );
        RNOK( m_pcContextStore->setCtxStates( uiSlice, pacCtxStates ) );
      }
    }
  }

  return Err::m_nOK;
}


template<class t_EstResult, class t_ContextModel>
ErrVal SubdivisionEstimator<t_EstResult, t_ContextModel>::xProcessNode( const QTNavi &rcNav, RDCost &rcCost )
{
  AOF_DBG( rcNav.isLeafNode() );

  State *pcRoot;
  RNOK( m_pcState->getState( pcRoot ) );

  RDCost cSingleCost( m_cLambda );
  bool bSubdiv = false;
  RNOK( m_rcSingleNodeEstIf.processSubdivisionInfo( rcNav, bSubdiv, &cSingleCost ) );
  t_EstResult *pSingleResult = NULL;
  t_EstResult *pSinglePreResult = NULL;
  t_EstResult *pSingleResResult = NULL;
  RNOK( m_rcSingleNodeEstIf.estimateSingleNode( rcNav, pSingleResult, pSinglePreResult, pSingleResResult, &cSingleCost ) );
  bool bBreakPredTree = false;
  if( m_bFastModeDecision )
  {
    bBreakPredTree = m_rcSingleNodeEstIf.getAllBelowFmdThreshold();
  }

  bool bHasChildNodes = false;
  if( ! bBreakPredTree )
  {
    RNOK( m_rcTree.appendChildNodes( rcNav, bHasChildNodes ) );
  }
  if( bHasChildNodes )
  {
    State *pcSingle;
    RNOK( m_pcState->getState( pcSingle ) );
    RNOK( pcRoot->setState() );

    RDCost cChildCost( m_cLambda );
    bSubdiv = true;
    RNOK( m_rcSingleNodeEstIf.processSubdivisionInfo( rcNav, bSubdiv, &cChildCost ) );

    TRACE_INDENT_INC;
    for( UInt uiIdx = 0; uiIdx < 4; uiIdx++ )
    {
      QTNavi cChildNav = rcNav.getChild( uiIdx );
      if( cChildNav.isLeafNode() )
      {
        RNOK( xProcessNode( cChildNav, cChildCost ) );
      }
    }
    TRACE_INDENT_DEC;

    if( cChildCost < cSingleCost )
    {
      RNOK( m_rcSingleNodeEstIf.releaseCurrentNode( rcNav ) );
      RNOK( pcRoot->deleteState() );
      RNOK( pcSingle->deleteState() );
      delete pSingleResult;
      delete pSinglePreResult;
      delete pSingleResResult;
      rcCost += cChildCost;
      return Err::m_nOK;
    }
    RNOK( m_rcSingleNodeEstIf.releaseChildNodes( rcNav ) );
    RNOK( pcSingle->setAndDeleteState() );
    RNOK( m_rcTree.removeChildNodes( rcNav ) );
  }
  RNOK( m_rcSingleNodeEstIf.setSingleResult( rcNav, pSingleResult, pSinglePreResult, pSingleResResult ) );
  RNOK( pcRoot->deleteState() );

  rcCost += cSingleCost;
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _SUBDIVISIONESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
