/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _SYNTAXELEMENTPROCESSINGCHAINENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _SYNTAXELEMENTPROCESSINGCHAINENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/InStreamIf.h"
#include "SyntaxElementEncoder.h"
#include "ResettableStream.h"
#include "ProbabilityAllocatorEnc.h"
#include "BitRateEstimator.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class T1, class T2>
class TwoResettableStreams
  : public BinStreamIf<T1&, true, T2&>
  , public StateMultiplier
{
public:
  TwoResettableStreams( BinStreamIf<T1&, true, T2&> &rcOut )
    : m_cSplitter( rcOut )
    , m_cResStr( m_cSplitter )
  {
    StateMultiplier::addStateIf( &m_cResStr );
  }

  ErrVal pushBack( const T1 &rcT ) { Item i = { true,  rcT,  T2() }; return m_cResStr.pushBack( i ); }
  ErrVal pushBack( const T2 &rcT ) { Item i = { false, T1(), rcT  }; return m_cResStr.pushBack( i ); }

private:
  struct Item
  {
    bool m_bIsT1;
    T1 m_T1;
    T2 m_T2;
  };

  class Splitter
    : public InStreamIf<const Item&>
  {
  public:
    Splitter( BinStreamIf<T1&, true, T2&> &rcOut ) : m_rcOut( rcOut ) {}
    ErrVal pushBack( const Item &rcT )
    {
      if( rcT.m_bIsT1 )
      {
        RNOK(( static_cast<typename BinStreamIf<T1&, true, T2&>::TypeCall&>( m_rcOut ).pushBack( rcT.m_T1 ) ));
      }
      else
      {
        RNOK(( static_cast<typename BinStreamIf<T1&, true, T2&>::BoolCall&>( m_rcOut ).pushBack( rcT.m_T2 ) ));
      }
      return Err::m_nOK;
    }

  private:
    BinStreamIf<T1&, true, T2&> &m_rcOut;
  };

  Splitter m_cSplitter;
  ResettableStream<Item> m_cResStr;
};

template<class t_ContextModel = CabacContextModel>
class SyntaxElementProcessingChainEncIf
{
protected:
  SyntaxElementProcessingChainEncIf() {}

public:
  virtual ~SyntaxElementProcessingChainEncIf() {}

  virtual ProbabilityAllocatorBase<t_ContextModel>& getProbAlloc() = 0;
  virtual SyntaxElementEncoderIf& getSyntaxElementEncoderIf() = 0;
  virtual BitRateEstimatorIf& getBitRateEstimatorIf() = 0;

};


template<bool t_bEnableBREst, class t_CabacTables = DefaultCabacTables, bool t_bEnableForwarding = true, class t_ContextModel = CabacContextModel>
class SyntaxElementProcessingChainEnc
  : public SyntaxElementProcessingChainEncIf<t_ContextModel>
{
  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( SyntaxElementProcessingChainEnc );

public:
  typedef ProbBinBitRateEstimator<t_CabacTables, ProbBinStreamEncIf, t_bEnableForwarding>                      BitRateEstimatorType;
  typedef typename TypeSelect<t_bEnableBREst, BitRateEstimatorType, BinStreamProcessor<ProbBin&, true> >::Type InStreamIfType;
  typedef ProbabilityAllocatorEnc<InStreamIfType, t_ContextModel>                                               ProbAllocator;

  //typedef ProbabilityAllocatorEnc<InStreamProcessor<const ProbBin>, t_ContextModel> ProbAllocator;

  SyntaxElementProcessingChainEnc( StateMultiplier *pcStateMultiplier,
                                   ProbBinStreamEncIf *pcProbBinOut,
                                   const ContextArrayStartIndices &rcCtxIndices,
                                   const CtxInitArray &rcCtxInit,
                                   const PlanePackageParameters&  rcPlaPaPars,
                                   int iSliceQpForInit,
                                   TraceCommon *pcTrace );
  
  BitRateEstimatorIf& getBitRateEstimatorIf() { TOF( m_pcBREst ); return *m_pcBREst; }
  SyntaxElementEncoderIf& getSyntaxElementEncoderIf() { return *m_pcSEEncoder; }

  ProbabilityAllocatorBase<t_ContextModel>& getProbAlloc() { return *m_pcProbAlloc; }

  virtual ~SyntaxElementProcessingChainEnc();

private:
  ErrVal xRecreateProbAlloc( const std::vector<CtxIni> &racCtxInit );

  TwoResettableStreams<ProbBin, bool> *m_pcResStr;
  BitRateEstimatorType *m_pcBREst;
  BREstStateIf<BitRateEstimatorType> *m_pcBREstIf;
  ProbAllocator *m_pcProbAlloc;
  SyntaxElementEncoder<ProbAllocator> *m_pcSEEncoder;
  TraceCommon *m_pcTrace;
  int m_iSliceQpForInit;
};

template<class t_Out, class t_In> class GetIfEqual               { public: static t_Out getIfEqual( t_In x  ) { AOT( 1 ); t_Out *p(NULL); return *p; } };
template<class t_Out>             class GetIfEqual<t_Out, t_Out> { public: static t_Out getIfEqual( t_Out x ) { return x; } };


template<bool t_bEnableBREst, class t_CabacTables, bool t_bEnableForwarding, class t_ContextModel>
SyntaxElementProcessingChainEnc<t_bEnableBREst, t_CabacTables, t_bEnableForwarding, t_ContextModel>::SyntaxElementProcessingChainEnc( StateMultiplier *pcStateMultiplier,
                                                                                                                                      ProbBinStreamEncIf *pcProbBinOut,
                                                                                                                                      const ContextArrayStartIndices &rcCtxIndices,
                                                                                                                                      const CtxInitArray &rcCtxInit,
                                                                                                                                      const PlanePackageParameters&  rcPlaPaPars,
                                                                                                                                      int iSliceQpForInit,
                                                                                                                                      TraceCommon *pcTrace )
: m_pcResStr( NULL )
, m_pcBREst( NULL )
, m_pcBREstIf( NULL )
, m_pcProbAlloc( NULL )
, m_pcTrace( pcTrace )
, m_iSliceQpForInit( iSliceQpForInit )
{
  if( t_bEnableBREst && NULL != pcProbBinOut )
  {
    m_pcResStr = new TwoResettableStreams<ProbBin, bool>( *pcProbBinOut );
    pcStateMultiplier->addStateIf( m_pcResStr );
    pcProbBinOut = m_pcResStr;
  }

  {
    typename BitRateEstimatorType::CtorPars cBREstCtorPars = { pcProbBinOut };
    BinStreamProcessor<ProbBin&, true>::CtorPars cDummy;

    typename InStreamIfType::CtorPars *pcCtorPars = NULL;
    if( t_bEnableBREst )
    {
      pcCtorPars = GetIfEqual<typename InStreamIfType::CtorPars*, typename BitRateEstimatorType::CtorPars*>::getIfEqual( &cBREstCtorPars );
    }
    else
    {
      pcCtorPars = GetIfEqual<typename InStreamIfType::CtorPars*, BinStreamProcessor<ProbBin&, true>::CtorPars*>::getIfEqual( &cDummy );
    }

    typename ProbAllocator::CtorPars cProbAllocPars = { UInt( rcCtxInit.getCtxInit().size() ), TRACE_PROB_BINS ? m_pcTrace : NULL, *pcCtorPars };
    m_pcSEEncoder = new SyntaxElementEncoder<ProbAllocator>( rcCtxInit.getCtxAccess(), rcCtxIndices, rcPlaPaPars, pcTrace, cProbAllocPars );
  }

  m_pcProbAlloc = m_pcSEEncoder;
  if( t_bEnableBREst )
  {
    m_pcBREst = GetIfEqual<BitRateEstimatorType*, InStreamIfType*>::getIfEqual( &m_pcSEEncoder->getProbBinProcessor() );
    m_pcBREstIf = new BREstStateIf<BitRateEstimatorType>( *m_pcBREst );
    pcStateMultiplier->addStateIf( &m_pcProbAlloc->getStateIf() );
    pcStateMultiplier->addStateIf( m_pcBREstIf );
  }
  else if( NULL != pcProbBinOut )
  {
    GetIfEqual<BinStreamProcessor<ProbBin&, true>*, InStreamIfType*>::getIfEqual( &m_pcSEEncoder->getProbBinProcessor() )->InStreamProcessor<const ProbBin&>::setInStreamIf( *pcProbBinOut );
    GetIfEqual<BinStreamProcessor<ProbBin&, true>*, InStreamIfType*>::getIfEqual( &m_pcSEEncoder->getProbBinProcessor() )->InStreamProcessor<const bool   &>::setInStreamIf( *pcProbBinOut );
  }
  TNOK( m_pcProbAlloc->initCtx( m_iSliceQpForInit, rcCtxInit.getCtxInit() ) );
}


template<bool t_bEnableBREst, class t_CabacTables, bool t_bEnableForwarding, class t_ContextModel>
SyntaxElementProcessingChainEnc<t_bEnableBREst, t_CabacTables, t_bEnableForwarding, t_ContextModel>::~SyntaxElementProcessingChainEnc()
{
  delete m_pcResStr;
  delete m_pcSEEncoder;
  delete m_pcBREstIf;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _SYNTAXELEMENTPROCESSINGCHAINENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
