/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoEncoderLibPch.h"
#include "UvlcWriter.h"
#include "VideoCommonLib/BitWriterIf.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN


__inline 
ErrVal UvlcWriter::xWriteCode( UInt uiCode, UInt uiLength )
{
  AOT_DBG(uiLength<1);
  return m_pcBitWriterIf->write( uiCode, uiLength );
}


__inline 
ErrVal UvlcWriter::xWriteFlag( UInt uiCode )
{
  return m_pcBitWriterIf->write( uiCode, 1 );
}


ErrVal UvlcWriter::xWriteUvlcCode( UInt uiVal)
{
  UInt uiLength = 1;
  UInt uiTemp = ++uiVal;

  AOF_DBG( uiTemp );

  while( 1 != uiTemp )
  {
    uiTemp >>= 1;
    uiLength += 2;
  }

  RNOK( m_pcBitWriterIf->write( uiVal, uiLength ) );

  return Err::m_nOK;
}

ErrVal UvlcWriter::xWriteUvlcCodeLong( UInt uiVal)
{
  UInt uiLength = 1;
  UInt uiTemp = ++uiVal;

  AOF_DBG( uiTemp );

  while( 1 != uiTemp )
  {
    uiTemp >>= 1;
    uiLength ++;
  }

  if( uiLength > 1 )
  {
    RNOK( m_pcBitWriterIf->write( 0, uiLength-1 ) );
  }
  RNOK( m_pcBitWriterIf->write( uiVal, uiLength ) );

  return Err::m_nOK;
}


ErrVal UvlcWriter::xWriteUvlcCodeUInt64( UInt64 uiVal)
{
  UInt uiLength = 1;
  UInt64 uiTemp = ++uiVal;

  AOF_DBG( uiTemp );

  while( 1 != uiTemp )
  {
    uiTemp >>= 1;
    uiLength ++;
  }

  if( uiLength > 1 )
  {
    for( UInt ui = 0; ui < uiLength-1; ui++ )
    {
      RNOK( m_pcBitWriterIf->write( 0, 1 ) );
    }
  }

  for( int i = uiLength - 1; i >= 0; i-- )
  {
    RNOK( m_pcBitWriterIf->write( UInt((uiVal>>i) & 1), 1 ) );
  }

  return Err::m_nOK;
}


ErrVal UvlcWriter::xWriteSvlcCode( int iVal)
{
  UInt uiVal = xConvertToUInt( iVal );
  UInt uiLength = 1;
  UInt uiTemp = ++uiVal;

  AOF_DBG( uiTemp );

  while( 1 != uiTemp )
  {
    uiTemp >>= 1;
    uiLength += 2;
  }

  RNOK( m_pcBitWriterIf->write( uiVal, uiLength ) );

  return Err::m_nOK;
}

ErrVal UvlcWriter::writeUvlc( UInt uiCode, const char* pcTraceString )
{
  xTraceString( pcTraceString, uiCode );
  RNOK( xWriteUvlcCodeLong( uiCode ) );
  return Err::m_nOK;
}

ErrVal UvlcWriter::writeUvlc64( UInt64 uiCode, const char* pcTraceString )
{
  xTraceString( pcTraceString, uiCode );
  RNOK( xWriteUvlcCodeUInt64( uiCode ) );
  return Err::m_nOK;
}


ErrVal UvlcWriter::writeSvlc( int iCode, const char* pcTraceString )
{
  xTraceString( pcTraceString, iCode );
  UInt uiCode = xConvertToUInt( iCode );
  RNOK( xWriteUvlcCodeLong( uiCode ) );
  return Err::m_nOK;
}

ErrVal UvlcWriter::writeFlag( bool bFlag, const char* pcTraceString )
{
  xTraceString( pcTraceString, bFlag );
  RNOK( xWriteFlag( bFlag ? 1:0) );
  return Err::m_nOK;
}


ErrVal UvlcWriter::writeSCode( int iCode, UInt uiLength, const char* pcTraceString )
{
  xTraceString( pcTraceString, iCode );
  AOT_DBG(uiLength<1);
  UInt uiShift = 32 - uiLength;
  UInt uiCode = ((UInt)(iCode << uiShift)) >> uiShift;
  RNOK( m_pcBitWriterIf->write( uiCode, uiLength ) );
  return Err::m_nOK;
}

ErrVal UvlcWriter::writeCode( UInt uiCode, UInt uiLength, const char* pcTraceString )
{
  xTraceString( pcTraceString, uiCode );
  AOT_DBG(uiLength<1);
  RNOK( m_pcBitWriterIf->write( uiCode, uiLength ) );
  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END
