/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoEncoderLibPch.h"
#include "VideoEncoderImpl.h"
#include "VideoEncoderLib/VideoEncoderParameter.h"
#include "VideoCodecToolsLib/BitstreamSink.h"

#include "AccessUnitEncoder.h"

#include "../VideoDecoderLib/AccessUnitDecoder.h"
#include "../VideoDecoderLib/BitReader.h"
#include "VideoCommonLib/BitWriter.h"

#include "VideoCommonLib/H264SequenceParameterSet.h"
#include "VideoCommonLib/H264PictureParameterSet.h"
#include "H264NalUnitEncoder.h"

#include "TraceEnc.h"

#include <fstream>

// video namespace begin
VIDEO_NAMESPACE_BEGIN

extern const int g_aaiLumaToChromaQp[2][52];

VideoEncoderImpl::VideoEncoderImpl()
: m_pcEncPar( NULL )
, m_pcPrecodedBitstreamParser( NULL )
, m_uiNumAUsToParse( 0 )
, m_cNextInputPoc( 0 )
, m_cDecToDispBuffer( 0 )
, m_cDecPreToDispBuffer( 0 )
, m_cDecResToDispBuffer( 0 )
, m_iCodOrder( 0 )
, m_cBitstreamSequencer( 0 )
, m_cConsoleOutputSequencer( 0 )
, m_uiFrameEncodingDurationSumMSec( 0 )
, m_uiFrameMEDurationSumMSec( 0 )
, m_pcSequenceParameterSet( 0 )
, m_pcPictureParameterSet( 0 )
, m_pcSequenceParameterSetParse( 0 )
, m_pcPictureParameterSetParse( 0 )
, m_uiIdrPicId(0) 
{
}


VideoEncoderImpl::~VideoEncoderImpl()
{
  if( m_cPrecodedBitstreamFile.is_open() )
  {
    m_cPrecodedBitstreamFile.close();
  }
  delete m_pcPrecodedBitstreamParser;
  delete m_pcSequenceParameterSet;
  delete m_pcPictureParameterSet;
  delete m_pcSequenceParameterSetParse;
  delete m_pcPictureParameterSetParse;
}


ErrVal VideoEncoderImpl::runAlternative( int argc, char **argv, bool &rbExit )
{
  rbExit = false;
  return Err::m_nOK;
}


ErrVal VideoEncoderImpl::create( VideoEncoderInstance *&rpcVideoEncoderInstance )
{
  try
  {
    rpcVideoEncoderInstance = new VideoEncoderImpl;
  }
  RETURN_ERR_ON_EXCEPTION;
  return Err::m_nOK;
}

ErrVal VideoEncoderImpl::destroy()
{
  try
  {
    delete this;
  }
  RETURN_ERR_ON_EXCEPTION;
  return Err::m_nOK;
}

ErrVal VideoEncoderImpl::setEncoderParameter( const VideoEncoderParameter* pcEncoderParameter )
{
  try
  {
    ROT( m_pcEncPar );
    m_pcEncPar = pcEncoderParameter;
    ROF( m_pcEncPar );

    RNOKS( m_cSequenceStructure.init( pcEncoderParameter->getGOPSettings().getGOPFormatString() ) );
    RNOKS( m_cRelRefPocGenerator.initDefRelRefPOCGen( m_cSequenceStructure,
                                                      m_pcEncPar->getSPSParams().getCodedPictureStoreSize(),
                                                      m_pcEncPar->getSPSParams().getNumberReferenceFrames(),
                                                      m_pcEncPar->getGOPSettings().getLayerChangeStartPoc(),
                                                      m_pcEncPar->getGOPSettings().getLayerOffset() ) );

    m_cSeqIter  = SequenceStructure::Iterator( m_cSequenceStructure, m_pcEncPar->getGOPSettings().getLayerChangeStartPoc(), m_pcEncPar->getGOPSettings().getLayerOffset() );
    m_cNextIter = SequenceStructure::Iterator( m_cSequenceStructure, m_pcEncPar->getGOPSettings().getLayerChangeStartPoc(), m_pcEncPar->getGOPSettings().getLayerOffset() );
    ++m_cNextIter;

#if ALLOW_BOOST_LIBRARY
    if( m_pcEncPar->getMultithreadingControlParams().getNumAccessUnitEncodingThreads() > 1 )
    {
      m_cAUCodingStarter.setNumThreads( m_pcEncPar->getMultithreadingControlParams().getNumAccessUnitEncodingThreads() );
    }
#endif

    ROT( pcEncoderParameter->getGeneralParams().getChromaFormatIdc() > 1 );
    m_acPSNR.resize( pcEncoderParameter->getGeneralParams().getChromaFormatIdc() == 0 ? 1 : 3 );

    m_uiNumAUsToParse = m_pcEncPar->getPrecodedPictureParsingParams().getNumFramesToParse();
    if( m_uiNumAUsToParse > 0 )
    {
      if( Err::m_nOK != m_cPrecodedBitstreamFile.open( m_pcEncPar->getPrecodedPictureParsingParams().getPrecodedBitstreamName(), LargeFile::OM_READONLY ) )
      {
        std::cerr << "Failed to open precoded bitstream file: " << m_pcEncPar->getPrecodedPictureParsingParams().getPrecodedBitstreamName() << std::endl;
        return Err::m_nERR;
      }
      m_pcPrecodedBitstreamParser = new PacketParserStartCode<LargeFile>( m_cPrecodedBitstreamFile );
    }
  }
  RETURN_ERR_ON_EXCEPTION;
  return Err::m_nOK;
}

ErrVal VideoEncoderImpl::xSetAccessUnitEncoderParameters( AccessUnitParameters &rcAUPars,
                                                          AccessUnitEncoderParameters &rcAUEstPars,
                                                          std::ostream &rcOut )
{
  //===== set subsampling vector =====
  std::vector<Subsampling> acSubsampling;
  if( m_pcEncPar->getGeneralParams().getChromaFormatIdc() == 0 )
  {
    acSubsampling.push_back( Subsampling( ShiftScale( 0 ), ShiftScale( 0 ) ) );
  }
  else if( m_pcEncPar->getGeneralParams().getChromaFormatIdc() == 1 )
  {
    acSubsampling.push_back( Subsampling( ShiftScale(  0 ), ShiftScale(  0 ) ) );
    acSubsampling.push_back( Subsampling( ShiftScale(  1 ), ShiftScale(  1 ) ) );
    acSubsampling.push_back( Subsampling( ShiftScale(  1 ), ShiftScale(  1 ) ) );
  }
  else
  {
    ROT(1);
  }
  //===== access unit parameters (and basic plane package parameters) =====
  rcAUPars = AccessUnitParameters      ( m_cSeqIter.getPoc(), m_pcEncPar->getGeneralParams().getInputPicSizeX(), m_pcEncPar->getGeneralParams().getInputPicSizeY(), 
                                         acSubsampling, m_pcEncPar->getGeneralParams().getPlaneGrouping() );

  {
    // th in this section poc and frame num are calculated 
    // to prevent from screwing up the whole encoding
    // poc calculation is done to check if the poc is 
    // reconstructable on the decoder site

    AOT( NULL == m_pcSequenceParameterSet );
    const UInt uiNumPocBits = m_pcSequenceParameterSet->getLog2MaxPicOrderCntLsb();
    const PicOrderCnt uiPocMask = (1<<uiNumPocBits)-1;

    // if the following check fails the Log2MaxPicOrderCntLsb is probably to small 
    // please check SequenceStructure::getMaxAbsPocDiff  
    ROT( rcAUPars.getPicOrderCnt() != m_cPocAndFrameNum.calculatePoc( UInt( rcAUPars.getPicOrderCnt() & uiPocMask ), uiNumPocBits, true, false )); 
    
    const UInt uiNumFrameNumBits = m_pcSequenceParameterSet->getLog2MaxFrameNum();
    const UInt64 uiFrameNumMask = (1<<uiNumFrameNumBits)-1;
    const UInt64 uiFrameNum = m_cPocAndFrameNum.getFrameNum();

    // the following check should never fail on the encoder site 
    // a similar check at the decoder potentially fails 
    // if there are gaps in frame num caused by data loss
    ROT( uiFrameNum != m_cPocAndFrameNum.calculateFrameNum( UInt( uiFrameNum & uiFrameNumMask ), uiNumFrameNumBits ))

      // store frame num for transmission
    rcAUPars.setFrameNum( uiFrameNum ); 
  }

  rcAUPars.setIdr                      ( m_cSeqIter.getFrameDescriptor().isIdr() );
  rcAUPars.setMaxDPBSize               ( m_pcEncPar->getSPSParams().getCodedPictureStoreSize() );
  rcAUPars.setRefPicFlag               ( m_cSeqIter.getFrameDescriptor().getStoreForRef() );
  rcAUPars.setMaxMvsPerBlk             ( m_cSeqIter.getFrameDescriptor().getMaxNumRefBlocksPerInterTranslationalPred() );
  rcAUPars.setDeblockingIdc            ( m_pcEncPar->getDeblockingFilterParams().getFilterIdc() );
  rcAUPars.setDeblockAlphaOffs         ( m_pcEncPar->getDeblockingFilterParams().getAlphaOffsetDiv2() );
  rcAUPars.setDeblockBetaOffs          ( m_pcEncPar->getDeblockingFilterParams().getBetaOffsetDiv2() );
  rcAUPars.setInterpolationMethod      ( m_pcEncPar->getSPSParams().getInterpolationMethod() );
  rcAUPars.setAdapInterpSwitch         ( m_pcEncPar->getSPSParams().getAdapInterpSwitch() );
  rcAUPars.setEntropySliceSizeX        ( m_pcEncPar->getEntropySliceParams().getEntropySliceSizeX() );
  rcAUPars.setEntropySliceSizeY        ( m_pcEncPar->getEntropySliceParams().getEntropySliceSizeY() );
  rcAUPars.setCtxStatesReuse           ( m_pcEncPar->getEntropySliceParams().getContextStatesReuse() );
  {
    ExtendedCabacTables aeExtendedCabacTablesMapping[2] = { DEFAULT_CABAC_TABLES, AGGREGATE_CABAC_TABLES12 };
    rcAUPars.setExtendedCabacTables( aeExtendedCabacTablesMapping[m_pcEncPar->getQuadtreeParams().getMultiCodewordArithmeticCodingMode()] );
  }
  rcAUPars.setVLCFlag                  ( m_pcEncPar->getQuadtreeParams().getVLCMode() > 0 );
  rcAUPars.setVLCMaxDelayInBits        ( m_pcEncPar->getQuadtreeParams().getVLCMaxDelay() );

  if( rcAUPars.getIdr() )
  {
    rcAUPars.setFrameNum( m_uiIdrPicId ); 
    m_uiIdrPicId++;
    m_uiIdrPicId &= 1;
  }

  if( rcAUPars.getMaxMvsPerBlk() > 0 )
  {
    rcAUPars.getRefDeltaPocList() = m_cRelRefPocGenerator.getRelRefPocs( rcAUPars.getPicOrderCnt(), rcOut );
    ROT( rcAUPars.getRefListSize() == 0 );
  }

  for( UInt uiPlaPaId = 0; uiPlaPaId < rcAUPars.getNumPlanePacks(); uiPlaPaId++ )
  {
    PlanePackageParameters& rcPlaPaPars = rcAUPars.getPlaPaPars( uiPlaPaId );
    UInt uiInterPlaPaPredMode = ( rcAUPars.getMaxMvsPerBlk() == 0 ? 0 : m_pcEncPar->getGeneralParams().getInterPlanePred( uiPlaPaId ) );
    rcPlaPaPars.setInterPlaPaPredMode         ( uiInterPlaPaPredMode );
    rcPlaPaPars.setLog2MinPredBlkSize         ( m_pcEncPar->getQuadtreeParams ().getLog2MinPredictionBlockSize( uiPlaPaId ) );
    rcPlaPaPars.setLog2MaxPredBlkSize         ( m_pcEncPar->getQuadtreeParams ().getLog2MaxPredictionBlockSize( uiPlaPaId ) );
    rcPlaPaPars.setMvAccuracy                 ( m_pcEncPar->getQuadtreeParams ().getLog2MvAccuracy            ( uiPlaPaId ) );
    rcPlaPaPars.setMvCompetition              ( m_pcEncPar->getQuadtreeParams ().getMvCompetition             ( uiPlaPaId ) > 0 );
    rcPlaPaPars.setLog2MinPow2TrafoBlkSize    ( m_pcEncPar->getQuadtreeParams ().getLog2MinTransformBlockSize ( uiPlaPaId ) );
    rcPlaPaPars.setLog2MaxPow2TrafoBlkSize    ( m_pcEncPar->getQuadtreeParams ().getLog2MaxTransformBlockSize ( uiPlaPaId ) );
    UInt uiTransformMode = ResInfo::RES_TRAFO_POW2;
    if( m_cSeqIter.getFrameDescriptor().getMaxNumRefBlocksPerInterTranslationalPred() )
    {
      uiTransformMode |= ResInfo::RES_TRAFO_SKIP;
    }
    rcPlaPaPars.setTransformMode( uiTransformMode );
    UInt  uiNumSigCtxClasses  = ( m_pcEncPar->getQuadtreeParams().getNumCtxEntriesSig () > 1 ? rcPlaPaPars.getNumPow2TrafoBlkSizes() : m_pcEncPar->getQuadtreeParams().getNumCtxEntriesSig () );
    UInt  uiNumLastCtxClasses = ( m_pcEncPar->getQuadtreeParams().getNumCtxEntriesLast() > 1 ? rcPlaPaPars.getNumPow2TrafoBlkSizes() : m_pcEncPar->getQuadtreeParams().getNumCtxEntriesLast() );
    for( UInt uiIdx = 0; uiIdx < uiNumSigCtxClasses; uiIdx++ )
    {
      rcPlaPaPars.getNumPow2SigCtxs().push_back( m_pcEncPar->getQuadtreeParams().getTrafoPow2ContextModelsSig( uiIdx ) );
    }
    for( UInt uiIdx = 0; uiIdx < uiNumLastCtxClasses; uiIdx++ )
    {
      rcPlaPaPars.getNumPow2LastCtxs().push_back( m_pcEncPar->getQuadtreeParams().getTrafoPow2ContextModelsLast( uiIdx ) );
    }
    int iLumaQp = gMinMax( m_pcEncPar->getQuantParams().getQp( m_cSeqIter.getFrameDescriptor().getSequenceStructureLayer() ), MIN_QP, MAX_QP );
    if( uiPlaPaId == 0 )
    {
      rcPlaPaPars.getPlaneQps().push_back( iLumaQp );
    }
    for( UInt uiPlaneId = ( uiPlaPaId == 0 ? 1 : 0 ); uiPlaneId < rcPlaPaPars.getNumPlanes(); uiPlaneId++ )
    {
      rcPlaPaPars.getPlaneQps().push_back( g_aaiLumaToChromaQp[0][ iLumaQp ] );
    }
  }
  //===== encoder only parameters =====
  const int                   iPrimQp             = gMinMax( m_pcEncPar->getQuantParams().getQp( m_cSeqIter.getFrameDescriptor().getSequenceStructureLayer() ), MIN_QP, MAX_QP );
  const std::pair<UInt,UInt>  puiuiDeadzoneInter  = m_pcEncPar->getQuantParams().getLayerDeadzonesInter( m_cSeqIter.getFrameDescriptor().getSequenceStructureLayer() );
  const std::pair<UInt,UInt>  puiuiDeadzoneIntra  = m_pcEncPar->getQuantParams().getLayerDeadzonesIntra( m_cSeqIter.getFrameDescriptor().getSequenceStructureLayer() );
  rcAUEstPars.setPrimQp                     ( iPrimQp );
  rcAUEstPars.setDeadzoneInter              ( Ratio( puiuiDeadzoneInter.first, puiuiDeadzoneInter.second ) );
  rcAUEstPars.setDeadzoneIntra              ( Ratio( puiuiDeadzoneIntra.first, puiuiDeadzoneIntra.second ) );
  rcAUEstPars.setDisableIntraInInter        ( m_pcEncPar->getQuadtreeParams().getDisableIntraInInter() );
  rcAUEstPars.setNumberOfReferenceFrames    ( m_pcEncPar->getSPSParams().getNumberReferenceFrames() );
  rcAUEstPars.setSearchModeFullPel          ( m_pcEncPar->getMotionVectorSearchParams().getSearchModeFullPel() );
  rcAUEstPars.setSearchModeSubPel           ( m_pcEncPar->getMotionVectorSearchParams().getSearchModeSubPel() );
  rcAUEstPars.setPrimPlaneOnlySearch        ( m_pcEncPar->getMotionVectorSearchParams().getPrimPlaneOnlySearch() );
  rcAUEstPars.setSearchRange                ( m_pcEncPar->getMotionVectorSearchParams().getSearchRange() );
  rcAUEstPars.setIterSearchRange            ( m_pcEncPar->getMotionVectorSearchParams().getIterSearchRange() );
  rcAUEstPars.setMaxIterations              ( m_pcEncPar->getMotionVectorSearchParams().getMaxIterations() );
  rcAUEstPars.setLambdaCalculationMode      ( m_pcEncPar->getQuantParams().getLambdaCalculationMode() );
  rcAUEstPars.setNextPoc                    ( m_cNextIter.getPoc() );
  rcAUEstPars.setTwoPassEncodingFlag        ( !m_pcEncPar->getGeneralParams().getDisableTwoPassEncodingFlag() );
  rcAUEstPars.getTrafoPow2SubdivideResidual ().clear();
  rcAUEstPars.getQuantizationMode           ().clear();
  rcAUEstPars.getMaxAbsQPDelta              ().clear();
  rcAUEstPars.setFastModeDecision           ( m_pcEncPar->getQuadtreeParams().getFastModeDecision(), 
                                              m_pcEncPar->getQuadtreeParams().getFMDThresholdHighQP(),
                                              m_pcEncPar->getQuadtreeParams().getFMDThresholdLowQP(),
                                              ( rcAUPars.getMaxMvsPerBlk() == 0 ), iPrimQp );
  rcAUEstPars.setDisableWienerFlag          ( m_pcEncPar->getInLoopWienerParams().getDisableFilterFlag() );
  rcAUEstPars.setWienerFilterType           ( m_pcEncPar->getInLoopWienerParams().getFilterType() );
  rcAUEstPars.setWienerMinLength            ( m_pcEncPar->getInLoopWienerParams().getMinFilterLength() );
  rcAUEstPars.setWienerMaxLength            ( m_pcEncPar->getInLoopWienerParams().getMaxFilterLength() );
  rcAUEstPars.setWienerPower2MinBlock       ( m_pcEncPar->getInLoopWienerParams().getPower2MinBlock() );
  rcAUEstPars.setWienerPower2MaxBlock       ( m_pcEncPar->getInLoopWienerParams().getPower2MaxBlcok() );

  for( UInt uiPlaPaId = 0; uiPlaPaId < rcAUPars.getNumPlanePacks(); uiPlaPaId++ )
  {
    rcAUEstPars.getTrafoPow2SubdivideResidual ().push_back( m_pcEncPar->getQuadtreeParams ().getTrafoPow2SubdivideResidual( uiPlaPaId ) );
    rcAUEstPars.getQuantizationMode           ().push_back( m_pcEncPar->getQuadtreeParams ().getQuantizationMode          ( uiPlaPaId ) );
    rcAUEstPars.getMaxAbsQPDelta              ().push_back( m_pcEncPar->getQuantParams    ().getMaxAbsDeltaQP             ( uiPlaPaId ) );
  }
  return Err::m_nOK;
}


ErrVal VideoEncoderImpl::getPSNR( std::vector<double> &radPSNR ) const
{
  radPSNR.clear();
  for( UInt ui = 0; ui < m_acPSNR.size(); ui++ )
  {
    double dSumPsnr = 0.0;
    std::map<Int64,double>::const_iterator cIter = m_acPSNR[ui].begin();
    for( UInt uiIter = 0; cIter != m_acPSNR[ui].end(); uiIter++, ++cIter )
    {
      ROF( cIter->first == uiIter );
      dSumPsnr += cIter->second;
    }
    if( m_acPSNR[ui].size() > 1 )
    {
      dSumPsnr /= double( m_acPSNR[ui].size() );
    }
    radPSNR.push_back( dSumPsnr );
  }
  return Err::m_nOK;
}

bool VideoEncoderImpl::isUncodedPictureWaiting() const
{
  bool bPicWaiting = false;
  if( m_acInputPicMap.find( m_cSeqIter.getPoc() ) != m_acInputPicMap.end() )
  {
    bPicWaiting = true;
  }
  return bPicWaiting;
}


bool VideoEncoderImpl::hasUncodedPictures() const
{
  return !m_acInputPicMap.empty();
}


ErrVal VideoEncoderImpl::xCreateNextCodingItem( AccessUnitEncodingItem *&rpcAUCodingItem, const OriginalAndReconstructedPic &rcInputPics, VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf )
{
  bool bEncodeSPSAndPPSForParsing = false;
  const PicOrderCnt cNextPocToCode = m_cSeqIter.getPoc();

  AccessUnitEncoderParameters cAUEstPars;
  AccessUnitParameters cAUPars;
  std::stringstream cOut;

  {
    if( NULL == m_pcSequenceParameterSet )
    {
      AOT( NULL != m_pcPictureParameterSet);
      RNOKS( SequenceParameterSet::create( m_pcSequenceParameterSet ));
      RNOKS( PictureParameterSet::create( m_pcPictureParameterSet ));

      RNOK( xSetAccessUnitEncoderParameters( cAUPars, cAUEstPars, cOut ) );

      // determine the number of bits to transmit valid poc values
      UInt uiLog2MaxPicOrderCntLsb = 0;
      {
        const int iMaxAbsFrameDiffRef = m_cSequenceStructure.getMaxAbsPocDiff( m_pcEncPar->getGeneralParams().getTotalInputFrames() );
        if( iMaxAbsFrameDiffRef )
        {
          uiLog2MaxPicOrderCntLsb  = gMax<int>( 4, gCeilLog2( iMaxAbsFrameDiffRef * 2 /* * 2 */) ); // frame coding only coding
        }
        else
        {
          uiLog2MaxPicOrderCntLsb = 4;
        }
      }

      RNOK( m_pcSequenceParameterSet->setAccessUnitParameters( cAUPars, uiLog2MaxPicOrderCntLsb, m_pcEncPar->getSPSParams().getLevelIdc(), m_pcEncPar->getSPSParams().getProfileIdc() ) );
      RNOK( m_pcPictureParameterSet->setAccessUnitParameters( cAUPars ) );

      if( ! m_pcPrecodedBitstreamParser )
      {
        RNOK( xWriteParameterSets( rcBitstreamPacketSinkIf ) );
      }
      else
      {
        ROT( m_pcSequenceParameterSetParse );
        ROT( m_pcPictureParameterSetParse );
        RNOKS( SequenceParameterSet::create( m_pcSequenceParameterSetParse ) );
        RNOKS( PictureParameterSet ::create( m_pcPictureParameterSetParse  ) );
        *m_pcSequenceParameterSetParse  = *m_pcSequenceParameterSet;
        *m_pcPictureParameterSetParse   = *m_pcPictureParameterSet;
        bEncodeSPSAndPPSForParsing = true;
      }
    }
    else
    {
      RNOK( xSetAccessUnitEncoderParameters( cAUPars, cAUEstPars, cOut ) );
    }
    cAUPars.setSPS( m_pcSequenceParameterSet );
    cAUPars.setPPS( m_pcPictureParameterSet );
  }

  ROF( cAUPars.getMaxDPBSize() == m_cRelRefPocGenerator.getCodedPictureStoreMaxSize() );

  UInt  uiIntraPic    = ( cAUPars.getMaxMvsPerBlk() == 0 ? 1 : 0 );
  int   iLumaQp       = cAUPars.getPlaPaPars( 0 ).getPlaneQp( 0 );
  UInt  uiChromaIdx   = 0;
  for( UInt uiPlaPaId = 0; uiPlaPaId < cAUPars.getNumPlanePacks(); uiPlaPaId++ )
  {
    PlanePackageParameters& rcPlaPaPars = cAUPars.getPlaPaPars( uiPlaPaId );
    std::vector<int>&       rcQps       = rcPlaPaPars.getPlaneQps();
    for( UInt uiPlaneId = ( uiPlaPaId == 0 ? 1 : 0 ); uiPlaneId < rcPlaPaPars.getNumPlanes(); uiPlaneId++, uiChromaIdx++ )
    {
      rcQps[uiPlaneId] = g_aaiLumaToChromaQp[uiIntraPic][gMinMax( iLumaQp + m_pcPictureParameterSet->getChromaQpIndexOffset( uiChromaIdx > 1 ? true : false ), 0, 51 )];
    }
  }

  const PicBlock<const Pel> cOrig   ( m_acInputPicMap[cNextPocToCode].m_pcOriginalPic->m_acPlanes );
  const PicBlock<Pel>       cEnc    ( m_acInputPicMap[cNextPocToCode].m_pcReconstructedPic->m_acPlanes );
  const PicBlock<Pel> *pcEncPre = m_acInputPicMap[cNextPocToCode].m_pcReconstructedPrePic ? new PicBlock<Pel>( m_acInputPicMap[cNextPocToCode].m_pcReconstructedPrePic->m_acPlanes ) : NULL;
  const PicBlock<Pel> *pcEncRes = m_acInputPicMap[cNextPocToCode].m_pcReconstructedResPic ? new PicBlock<Pel>( m_acInputPicMap[cNextPocToCode].m_pcReconstructedResPic->m_acPlanes ) : NULL;

  AccessUnit *pcAccessUnit = new AccessUnit;

  const bool bParse = m_uiNumAUsToParse > 0;
#if MULTICABAC_TRANSCODING_HACK
  UInt uiParsedPacketSize = 0;
#endif
  if( m_uiNumAUsToParse > 0 )
  {
    m_uiNumAUsToParse--;
  }

  if( m_pcPrecodedBitstreamParser )
  {
    TraceEnc* pcTraceEnc = NULL;
#if TRACE_ENC_ENABLED
    std::fstream cTraceFile;
    cTraceFile.open( "TraceEnc_PRE.txt", std::ios::out | std::ios::binary | std::ios::trunc );
    ROT( cTraceFile.fail() );
    TraceEnc cTraceEnc( cTraceFile );
    pcTraceEnc = &cTraceEnc;
#endif

    cAUPars.setPocAndFrameNum( &m_cPocAndFrameNum );
    pcAccessUnit->getAUPars().setPocAndFrameNum( &m_cPocAndFrameNum );
#if MULTICABAC_TRANSCODING_HACK
    RNOK( AccessUnitEncoder::parsePrecodedAccessUnit( *m_pcPrecodedBitstreamParser,
                                                      m_cCodedPictureStore,
                                                      bParse,
                                                      uiParsedPacketSize,
                                                      cAUPars,
                                                      pcAccessUnit,
                                                      pcTraceEnc,
                                                      m_pcSequenceParameterSetParse,
                                                      m_pcPictureParameterSetParse ) );
#else
    RNOK( AccessUnitEncoder::parsePrecodedAccessUnit( *m_pcPrecodedBitstreamParser,
                                                      m_cCodedPictureStore,
                                                      bParse,
                                                      cAUPars,
                                                      pcAccessUnit,
                                                      pcTraceEnc,
                                                      m_pcSequenceParameterSetParse,
                                                      m_pcPictureParameterSetParse ) );
#endif
    if( bParse )
    {
      const AccessUnitParameters& rcReadAUPars = pcAccessUnit->getAUPars();
      ROF( cAUPars.getPicOrderCnt()      == rcReadAUPars.getPicOrderCnt()      );
      ROF( cAUPars.getAUStats()          == rcReadAUPars.getAUStats()          );
      ROF( cAUPars.getRefPicFlag()       == rcReadAUPars.getRefPicFlag()       );
      ROF( cAUPars.getMaxMvsPerBlk()     == rcReadAUPars.getMaxMvsPerBlk()     );
      ROF( equal( cAUPars.getRefDeltaPocList().begin(), cAUPars.getRefDeltaPocList().end(), rcReadAUPars.getRefDeltaPocList().begin() ) );
      ROF( cAUPars.getDeblockingIdc()    == rcReadAUPars.getDeblockingIdc()    );
      ROF( cAUPars.getDeblockAlphaOffs() == rcReadAUPars.getDeblockAlphaOffs() );
      ROF( cAUPars.getDeblockBetaOffs()  == rcReadAUPars.getDeblockBetaOffs()  );
      cAUEstPars.setPrimQp( rcReadAUPars.getPlaPaPars( 0 ).getPlaneQp( 0 ) ); // for lambda calculation used by Wiener loop filter estimation only, can be removed if Wiener filter is not re-estimated for parsed slices
      if( bEncodeSPSAndPPSForParsing )
      {
        RNOK( xWriteParameterSets( rcBitstreamPacketSinkIf ) );
      }
    }
  }

  RNOK( m_cCodedPictureStore.markAUsAsUsed( cAUPars.getPicOrderCnt(), cAUPars.getRefDeltaPocList() ) );

  std::vector<PicOrderCnt> aiPocsToRemoveAfterCurrentPoc;
  if( cAUPars.getRefPicFlag() )
  {
    RNOK( m_cCodedPictureStore.registerAU( cAUPars.getPicOrderCnt() ) );
    RNOK( m_cRelRefPocGenerator.getPocsToRemoveFromStoreAfterCurrentPoc( aiPocsToRemoveAfterCurrentPoc ) );
  }

  rpcAUCodingItem = new AccessUnitEncodingItem( rcInputPics,
                                                cOrig,
                                                cEnc,
                                                pcEncPre,
                                                pcEncRes,
                                                m_cRelRefPocGenerator.getLastKeyPicPoc( 0 ),
                                                m_cRelRefPocGenerator.getLastKeyPicPoc( 1 ),
                                                m_cRelRefPocGenerator.getCodingPriorityIndex(),
                                                aiPocsToRemoveAfterCurrentPoc,
                                                cAUEstPars,
                                                cAUPars,
                                                pcAccessUnit,
                                                cOut.str(),
                                                bParse,
#if MULTICABAC_TRANSCODING_HACK
                                                uiParsedPacketSize,
#endif
                                                m_iCodOrder++ );

  return Err::m_nOK;
}


ErrVal VideoEncoderImpl::xReleaseAllCodedItems( VideoCodecSinkIf<const VideoCodecPicture<const Pel>*> &rcOriginalPicSinkIf,
                                                VideoCodecSinkIf<const VideoCodecPicture<Pel>*> &rcReconstructedPicSinkIf,
                                                VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedPrePicSinkIf,
                                                VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedResPicSinkIf,
                                                VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf,
                                                bool bWaitForOne )
{
  std::vector<AccessUnitEncodingItem*> apcCodedItems;
  RNOK( m_cAUCodingStarter.checkForCodedItems( apcCodedItems, bWaitForOne ) );
  for( UInt ui = 0; ui < UInt( apcCodedItems.size() ); ui++ )
  {
    RNOK( xReleaseCodingItem( apcCodedItems[ui], rcOriginalPicSinkIf, rcReconstructedPicSinkIf, pcReconstructedPrePicSinkIf, pcReconstructedResPicSinkIf, rcBitstreamPacketSinkIf ) );
  }
  return Err::m_nOK;
}

ErrVal VideoEncoderImpl::xWriteParameterSets( VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf )
{
  std::list< std::vector<UChar> >*pcSPSNalUnitList;
  pcSPSNalUnitList = new std::list< std::vector<UChar> >;
  pcSPSNalUnitList->push_back( std::vector<UChar>() );

  std::list< std::vector<UChar> >*pcPPSNalUnitList;
  pcPPSNalUnitList = new std::list< std::vector<UChar> >;
  pcPPSNalUnitList->push_back( std::vector<UChar>() );

  NalUnitEncoder cNalUnitEncoder;
  RNOK( cNalUnitEncoder.writeSPS( *m_pcSequenceParameterSet, pcSPSNalUnitList->front(), NULL ));
  RNOK( cNalUnitEncoder.writePPS( *m_pcPictureParameterSet, pcPPSNalUnitList->front(), NULL ));

  rcBitstreamPacketSinkIf.receive( pcSPSNalUnitList );
  rcBitstreamPacketSinkIf.receive( pcPPSNalUnitList );

  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
