/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoEncoderLibPch.h"

#include "VideoEncoderLib/VideoEncoderParameter.h"
#include "VideoCommonLib/LevelLimit.h"
#include "VideoCommonLib/ParameterIo.h"
#include "FormattedStringParser.h"
#include "VideoCommonLib/ReferencePicture.h"

#include <cmath>
#include <fstream>

#ifdef MSYS_WIN32
  #if( _MSC_VER >= 1400 )
    #define snprintf _snprintf_s
    #define SPRINTF sprintf_s
  #else
    #define snprintf _snprintf
    #define SPRINTF sprintf
  #endif
#else
  #define SPRINTF sprintf
#endif


#define ROTREPORT(x,t) {if(  x ) { std::cout << std::endl << t << std::endl;TOT(1);return Err::m_nInvalidParameter;}}
#define ROFREPORT(x,t) {if(!(x)) { std::cout << std::endl << t << std::endl;TOT(1);return Err::m_nInvalidParameter;}}
#define WARNINGOT(x,t) {if(  x ) { std::cout << std::endl << "WARNING: " << t << std::endl;}}


const UInt VideoEncoderParameter::GeneralParams::sm_auiNumPP[4]    = { 3, 2, 2, 1 };
const UInt VideoEncoderParameter::GeneralParams::sm_aauiNumP[4][3] = { { 1, 1, 1 }, { 2, 1, 0 }, { 1, 2, 0 }, { 3, 0, 0 } };


VideoEncoderParameter::VideoEncoderParameter()
{
}


ErrVal VideoEncoderParameter::MotionVectorSearchParams::check() const
{
  ROTREPORT( m_uiSearchModeFullPel > 3, "Unsupported value for SearchModeFullPel" );
  ROTREPORT( m_uiSearchModeSubPel  > 2, "Unsupported value for SearchModeSubPel" );
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::DeblockingFilterParams::check() const
{
  ROTREPORT( getFilterIdc() > 1,        "DisableDeblockingFilter may only be equal to 0 or 1");
  ROTREPORT(  6 < getAlphaOffsetDiv2(), "AlphaC0OffsetDiv2 exceeds supported range [-6..6]");
  ROTREPORT( -6 > getAlphaOffsetDiv2(), "AlphaC0OffsetDiv2 exceeds supported range [-6..6]");
  ROTREPORT(  6 < getBetaOffsetDiv2(),  "BetaOffsetDiv2 exceeds supported range [-6..6]");
  ROTREPORT( -6 > getBetaOffsetDiv2(),  "BetaOffsetDiv2 exceeds supported range [-6..6]");

  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::InLoopWienerFilterParams::check() const
{
  if( getUseFilter() )
  {
    ROTREPORT( (!(getMinFilterLength() & 1) || ( getMinFilterLength() < 1 )  ), "InLoopMinLength must be odd and greater 1." );
    ROTREPORT( (!(getMaxFilterLength() & 1) || (getMinFilterLength() > getMaxFilterLength())), "InLoopMaxLength must be odd and greater or equal InLoopMinLength." );
    //ROTREPORT( getMaxFilterLength() >= getMinFilterLength(), "InLoopMaxLength must be equal or greater than InLoopMinLength." );
    ROTREPORT( getFilterType() < 0 || getFilterType() > 1, "InLoopFilterType must equal 0 or 1" );
    ROTREPORT( getPower2MaxBlcok() < getPower2MinBlock(), "InLoopMaxBlock must be equal or greater than InLoopMinBlock." );
  }
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::QuadtreeParams::check( UInt uiNumPlaPa, UInt uiChromaFormatIdc, UInt uiPlaneGrouping ) const
{
  ROTREPORT( m_auiMvCompetition             .size() == 0, "No value specified for MvCompetition" );
  ROTREPORT( m_auiLog2MvAccuracy            .size() == 0, "No value specified for Log2MvAccuracy" );
  ROTREPORT( m_auiLog2MaxPredictionBlockSize.size() == 0, "No value specified for Log2MaxPredictionBlockSize" );
  ROTREPORT( m_auiLog2MinPredictionBlockSize.size() == 0, "No value specified for Log2MinPredictionBlockSize" );
  ROTREPORT( m_auiLog2MaxTransformBlockSize .size() == 0, "No value specified for Log2MaxTransformBlockSize" );
  ROTREPORT( m_auiLog2MinTransformBlockSize .size() == 0, "No value specified for Log2MinTransformBlockSize" );
  ROTREPORT( m_abTrafoPow2SubdivideResidual .size() == 0, "No value specified for TrafoPow2SubdivideResidual" );
  ROTREPORT( m_auiQuantizationMode          .size() == 0, "No value specified for QuantizationMode" );
  ROTREPORT( m_auiTrafoPow2ContextModelsSig .size() == 0, "No value specified for TrafoPow2ContextModelStringSig" );
  ROTREPORT( m_auiTrafoPow2ContextModelsLast.size() == 0, "No value specified for TrafoPow2ContextModelStringLast" );

  ROTREPORT( m_uiMultiCodewordArithmeticCoding > 1, "MultiCodewordArithmeticCoding must be smaller than 2." );

  WARNINGOT( m_auiMvCompetition             .size() < uiNumPlaPa, "MvCompetition doesn't specify values for all plane packages" );
  WARNINGOT( m_auiLog2MvAccuracy            .size() < uiNumPlaPa, "Log2MvAccuracy doesn't specify values for all plane packages" );
  WARNINGOT( m_auiLog2MaxPredictionBlockSize.size() < uiNumPlaPa, "Log2MaxPredictionBlockSize doesn't specify values for all plane packages" );
  WARNINGOT( m_auiLog2MinPredictionBlockSize.size() < uiNumPlaPa, "Log2MinPredictionBlockSize doesn't specify values for all plane packages" );
  WARNINGOT( m_auiLog2MaxTransformBlockSize .size() < uiNumPlaPa, "Log2MaxTransformBlockSize doesn't specify values for all plane packages" );
  WARNINGOT( m_auiLog2MinTransformBlockSize .size() < uiNumPlaPa, "Log2MinTransformBlockSize doesn't specify values for all plane packages" );
  WARNINGOT( m_abTrafoPow2SubdivideResidual .size() < uiNumPlaPa, "TrafoPow2SubdivideResidual doesn't specify values for all plane packages" );
  WARNINGOT( m_auiQuantizationMode          .size() < uiNumPlaPa, "QuantizationMode doesn't specify values for all plane packages" );

  for( UInt ui = 0; ui < uiNumPlaPa; ui++ )
  {
    ROTREPORT( m_auiMvCompetition             [ui] > 1,                                     "MvCompetition greater than 1 not supported" );
    WARNINGOT( m_auiLog2MvAccuracy            [ui] > 3,                                     "Motion vector accuarcy greater than 1/8-th sample" );
    ROTREPORT( m_auiLog2MaxPredictionBlockSize[ui] > 15,                                    "Prediction blocks larger than 2^15 are not supported" );
    ROTREPORT( m_auiLog2MaxTransformBlockSize [ui] > m_auiLog2MaxPredictionBlockSize[ui],   "Transform blocks larger than the maximum prediction block size are not supported" );
    ROTREPORT( m_auiLog2MinTransformBlockSize [ui] > m_auiLog2MinPredictionBlockSize[ui],   "Smallest Transform size has to be smaller or equal to smallest prediction size" );
    ROTREPORT( m_auiLog2MaxTransformBlockSize [ui] < m_auiLog2MinTransformBlockSize [ui],   "MaxTransformBlockSize must be >= MinTransformBlockSize" );
    ROTREPORT( m_auiLog2MaxPredictionBlockSize[ui] < m_auiLog2MinPredictionBlockSize[ui],   "MaxPredictionBlockSize must be >= MinPredictionBlockSize" );
    ROTREPORT( m_auiLog2MinPredictionBlockSize[ui] < 2,                                     "Smallest prediction size must not be less than 4x4" );
    ROTREPORT( m_auiLog2MinTransformBlockSize [ui] < 2,                                     "Smallest transform size must not be less than 4x4" );
    ROTREPORT( m_auiLog2MaxTransformBlockSize [ui] > 7,                                     "Maximum transform size is not supported" );
    ROTREPORT( m_auiQuantizationMode          [ui] > 2,                                     "QuantizationMode greater than 2 not supported" );
    UInt uiVecSize = m_auiLog2MaxTransformBlockSize[ui] - m_auiLog2MinTransformBlockSize[ui] + 1;
    WARNINGOT( uiVecSize > m_auiTrafoPow2ContextModelsSig .size() && 1 != m_auiTrafoPow2ContextModelsSig .size(), "Number of values in TrafoPow2ContextModelsSig  less than Log2MaxTransformBlockSize - Log2MinTransformBlockSize + 1 and greater than 1" );
    WARNINGOT( uiVecSize > m_auiTrafoPow2ContextModelsLast.size() && 1 != m_auiTrafoPow2ContextModelsLast.size(), "Number of values in TrafoPow2ContextModelsLast less than Log2MaxTransformBlockSize - Log2MinTransformBlockSize + 1 and greater than 1" );
  }
  for( UInt ui = 0; ui < m_auiTrafoPow2ContextModelsSig .size(); ui++ )
  {
    ROTREPORT( ! xCheckIfPowerOfTwoWithEvenExponent( m_auiTrafoPow2ContextModelsSig[ui] ), "Number of models for significance bits must be a power of two with even exponent" );
  }
  for( UInt ui = 0; ui < m_auiTrafoPow2ContextModelsLast.size(); ui++ )
  {
    ROTREPORT( ! xCheckIfPowerOfTwoWithEvenExponent( m_auiTrafoPow2ContextModelsLast[ui] ), "Number of models for last bits must be a power of two with even exponent" );
  }
  ROTREPORT( m_uiVLCMode > 1, "VLCMode greater than 1 is not supported" );
  return Err::m_nOK;
}


ErrVal VideoEncoderParameter::QuantParams::check( UInt uiNumPlanePacks, UInt uiNumLayers ) const
{
  ROTREPORT( m_uiLambdaCalculationMode          >  1,               "LambdaCalculationMode greater than 1 not supported" );
  ROTREPORT( m_auiMaxAbsQPDelta         .size() == 0,               "No value specified for MaxAbsQPDelta" );
  ROTREPORT( m_aiLayerQpOffsets         .size() == 0,               "No value specified for LayerQpOffsets" );
  ROTREPORT( m_apuiuiLayerDeadzonesInter.size() == 0,               "No value specified for LayerDeadzoneInter" );
  ROTREPORT( m_apuiuiLayerDeadzonesIntra.size() == 0,               "No value specified for LayerDeadzoneIntra" );

  WARNINGOT( m_iPPSQuant                        >  MAX_QP,          "PPSQuant exceeds supported range" );
  WARNINGOT( m_iPPSQuant                        <  0,               "PPSQuant exceeds supported range" );
  WARNINGOT( m_auiMaxAbsQPDelta         .size() < uiNumPlanePacks,  "MaxAbsQPDelta doesn't specify values for all plane packages" );
  WARNINGOT( m_aiLayerQpOffsets         .size() < uiNumLayers,      "LayerQpOffsets doesn't specify values for all layers" );
  WARNINGOT( m_apuiuiLayerDeadzonesInter.size() < uiNumLayers,      "LayerDeadzoneInter doesn't specify values for all layers" );
  WARNINGOT( m_apuiuiLayerDeadzonesIntra.size() < uiNumLayers,      "LayerDeadzoneIntra doesn't specify values for all layers" );
  for( UInt ui = 0; ui < m_auiMaxAbsQPDelta.size(); ui++ )
  {
    WARNINGOT( m_auiMaxAbsQPDelta[ui] > 7,                          "MaxAbsQPDelta is greater than 7" );
    WARNINGOT( m_iPPSQuant + (int)m_auiMaxAbsQPDelta[ui] > MAX_QP,  "PPSQuant + MaxAbsQPDelta exceeds supported range for one or more layers" );
    WARNINGOT( m_iPPSQuant - (int)m_auiMaxAbsQPDelta[ui] < 0,       "PPSQuant + MaxAbsQPDelta exceeds supported range for one or more layers" );
  }
  return Err::m_nOK;
}


ErrVal VideoEncoderParameter::GeneralParams::check() const
{
  ROTREPORT( 0 == m_uiTotalInputFrames,                                     "Total Number Of Frames == 0" );
  ROTREPORT( m_uiChromaFormatIdc > 1 ,                                      "Chroma Format no supported" );
  
  if( m_uiChromaFormatIdc == 0 )
  {
    ROTREPORT( m_uiInputPicSizeX % 4, "InputPicSizeX must be a multiple of 4 for 4:0:0." );
    ROTREPORT( m_uiInputPicSizeY % 4, "InputPicSizeY must be a multiple of 4 for 4:0:0." );
  }
  else
  {
    ROTREPORT( m_uiInputPicSizeX % 8, "InputPicSizeX must be a multiple of 8 for 4:2:0." );
    ROTREPORT( m_uiInputPicSizeY % 8, "InputPicSizeY must be a multiple of 8 for 4:2:0." );
  }

  ROTREPORT( m_uiPlaneGrouping > 3,                                 "Unsupported value for PlaneGrouping" );
  ROTREPORT( m_auiInterPlanePred.size() == 0,                       "No value specified for InterPlanePrediction" );
  WARNINGOT( m_auiInterPlanePred.size() < getNumberOfPlanePacks(),  "InterPlanePrediction doesn't specify values for all planes" );
  ROTREPORT( m_auiInterPlanePred[0] != 0,                           "First entry of InterPlanePrediction must be equal to 0" );
  for( UInt ui = 1; ui < m_auiInterPlanePred.size(); ui++ )
  {
    ROTREPORT( m_auiInterPlanePred[ui] > 2,                         "InterPlanePrediction must be less than or equal to 2" );
  }
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::PrecodedPictureParsingParams::check() const
{
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::MultithreadingControlParams::check() const
{
  ROTREPORT( m_uiNumAccessUnitEncodingThreads < 1, "NumAccessUnitEncodingThreads must be greater than 0." );
#if !ALLOW_BOOST_LIBRARY
  ROTREPORT( m_uiNumAccessUnitEncodingThreads > 1, "NumAccessUnitEncodingThreads must equal 1 if compiled without multithreading." );
#endif
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::EntropySliceParams::check() const
{
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::SPSParams::check() const
{
  ROTREPORT( m_uiInterpolationMethod > ReferencePicture<Pel>::NUM_UPS_METHODS-1, "Unsupported value for InterpolationMethod." );
  ROTREPORT( m_uiAdapInterpSwitch    > 1                                       , "Unsupported value for AdapInterpSwitch"    );
  return Err::m_nOK;
}

UInt VideoEncoderParameter::GOPSettings::getNumLayers() const
{
  UInt ui = 0;
  ANOK( FormattedStringParser::extractHighestLayerOfGOPString( getGOPFormatString(), ui ) );
  return ui+1;
}

ErrVal VideoEncoderParameter::GOPSettings::check()
{
  ROTREPORT( getLayerChangeStartPoc() < 0, "LayerChangeStartPoc may not be negative." );
  return Err::m_nOK;
}


ErrVal VideoEncoderParameter::create( VideoEncoderParameter*& rpcEncoderParameter )
{
  rpcEncoderParameter = new VideoEncoderParameter;
  ROT( NULL == rpcEncoderParameter );
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::destroy()
{
  delete this;
  return Err::m_nOK;
}


ErrVal VideoEncoderParameter::checkEncoderParameter()
{
  RNOK( m_cSPSParams                   .check() );
  RNOK( m_cPrecodedPictureParsingParams.check() );
  RNOK( m_cMultithreadingControlParams .check() );
  RNOK( m_cEntropySliceParams          .check() );
  RNOK( m_cGeneralParams               .check() );
  RNOK( m_cLoopFilterParams            .check() );
  RNOK( m_cInLoopWienerParams          .check() );
  RNOK( m_cMotionVectorSearchParams    .check() );
  RNOK( m_cGOPSettings                 .check() );
  UInt  uiNumPlaPacks     = m_cGeneralParams.getNumberOfPlanePacks();
  UInt  uiChromaFormatIdc = m_cGeneralParams.getChromaFormatIdc   ();
  UInt  uiPlaneGrouping   = m_cGeneralParams.getPlaneGrouping     ();
  UInt  uiNumLayers       = m_cGOPSettings  .getNumLayers         ();
  RNOK( m_cQuadtreeParams              .check( uiNumPlaPacks, uiChromaFormatIdc, uiPlaneGrouping ) );
  RNOK( m_cQuantParams                 .check( uiNumPlaPacks, uiNumLayers ) );

  ROTREPORT( m_cGeneralParams.getDisableTwoPassEncodingFlag() && ( m_cEntropySliceParams.getEntropySliceSizeX() || m_cEntropySliceParams.getEntropySliceSizeY() ), "Entropy slices must be disabled if DisableTwoPassEncoding is set to 1." );
  ROTREPORT( m_cGeneralParams.getDisableTwoPassEncodingFlag() && m_cQuadtreeParams.getMultiCodewordArithmeticCodingMode() > 0, "MultiCodewordArithmeticCoding must equal 0 if DisableTwoPassEncoding is set to 1." );
  ROTREPORT( m_cGeneralParams.getDisableTwoPassEncodingFlag() && m_cQuadtreeParams.getVLCMode() > 0, "VLCMode must equal 0 if DisableTwoPassEncoding is set to 1." );
  ROTREPORT( m_cQuadtreeParams.getVLCMaxDelay() > 0 && m_cQuadtreeParams.getVLCMaxDelay() < ( 1 << VLC_DELAY_PARAMETER_SHIFT ), "VLCMaxDelay must not be less than minimum supported value" );

  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::xFileName2ContentString( const std::string& rcFileName, std::string& rcContentString )
{
  std::ifstream cIfStream;

  // file is opened binary in order to get both CR and LF  opening the file non-binary would lead to 0d 0a to be replaced by 0a only
  cIfStream.open( rcFileName.c_str(), std::ios::in | std::ios::binary );

  if( false == cIfStream.is_open() )
  {
    std::cerr << "file could not be opened: " << rcFileName.c_str();
    return Err::m_nERR;
  }

  cIfStream.seekg( 0, std::ios::end );
  size_t uiFileSize = cIfStream.tellg();
  cIfStream.seekg( 0, std::ios::beg );

  if( 0 == uiFileSize )
  {
    std::cerr << "file is empty: " << rcFileName.c_str();
    return Err::m_nERR;
  }

  UChar* pucData = new UChar[uiFileSize+1];
  ROT( NULL == pucData );

  // read data from the stream
  cIfStream.read( (char*) pucData, uiFileSize );
  pucData[uiFileSize]=0;

  if( (UInt)cIfStream.gcount() != uiFileSize )
  {
    delete[] pucData;
    std::cerr << "file could not be read completely: " << rcFileName.c_str();
    return Err::m_nERR;
  }

  size_t uiSize = cIfStream.gcount();

  std::string cContentString;
  cContentString.reserve( uiSize );

  ROT( cContentString.capacity() < uiSize );

  cContentString.assign( (char*) pucData, 0, uiSize );

  if( cContentString.size() != uiSize)
  {
    delete[] pucData;

    std::cerr << "illegal input file. Seems not to be a text file: " << rcFileName.c_str();

    return Err::m_nERR;
  }

  delete[] pucData;

  rcContentString.append( cContentString );
  return Err::m_nOK;
}


ErrVal VideoEncoderParameter::initConfigFile( const std::string& rcConfigFileName, bool bIgnoreMissingParameters )
{
  std::string cFileContent;

  RNOKS( xFileName2ContentString( rcConfigFileName, cFileContent ) );

  ParameterRead cParameterRead( cFileContent, ! bIgnoreMissingParameters, 0 ,bIgnoreMissingParameters );
  RNOKS( xProcessConfigFileParameters( &cParameterRead ) );
  RNOK ( checkEncoderParameter() );

  return Err::m_nOK;
}


ErrVal VideoEncoderParameter::xProcessArgList( const std::string& rcCommand, const std::string& rcParameter )
{
  if( rcCommand == "-pf" ) 
  {
    RNOKS( initConfigFile( rcParameter ) );
    return Err::m_nOK;
  }
  return Err::m_nInvalidParameter;
}


ErrVal VideoEncoderParameter::initCommandLineArgs( int argc, char**argv, bool bPrintOptionsOnFail )
{
  ErrVal nRet = Err::m_nInvalidParameter;

  if( argc >= 2 )
  {
    for( int n = 1; n < argc; n++ )
    {
      std::string cCommand = argv[n++];

      if( cCommand == "-h" )
      {
        xPrintHelp();
        return Err::m_nOK;
      }
      else
      {
        if( n < argc )
        {
          std::string cParameter = argv[n];
          nRet = xProcessArgList( cCommand, cParameter );
        }
        if( Err::m_nOK != nRet )
        {
          break;
        }
      }
    }
  }

  if( Err::m_nOK != nRet && bPrintOptionsOnFail)
  {
    std::cout << "bad coding parameter " << std::endl;
    xPrintHelp();
  }
  else
  {
    RNOK( checkEncoderParameter() );
  }

  
  return nRet;
}

void VideoEncoderParameter::xPrintHelp()
{
  std::cout << std::endl;
  std::cout << " supported options:" << std::endl << std::endl;
  std::cout << "  -pf     Parameter File Name" << std::endl << std::endl;
  std::cout << "  -h      Print Option List" << std::endl;
  std::cout  << std::endl << std::endl;
}


ErrVal VideoEncoderParameter::GeneralParams::process( ParameterIoIf* pcParameterIoIf )
{
  GeneralParams& rcGP = *this;
  RNOK( pcParameterIoIf->value( "InputFile",        rcGP.m_cInputName,               true ) );
  RNOK( pcParameterIoIf->value( "InputPicSizeX",    rcGP.m_uiInputPicSizeX,          true ) );
  RNOK( pcParameterIoIf->value( "InputPicSizeY",    rcGP.m_uiInputPicSizeY,          true ) );
  RNOK( pcParameterIoIf->value( "ChromaFormatIdc",  rcGP.m_uiChromaFormatIdc,        true ) );
  RNOK( pcParameterIoIf->value( "PlaneGrouping",    rcGP.m_uiPlaneGrouping,          true ) );
  RNOK( pcParameterIoIf->value( "TotalInputFrames", (int&)rcGP.m_uiTotalInputFrames, true ) );
  RNOK( pcParameterIoIf->value( "OutputFile",       rcGP.m_cOutputName,              true ) );
  RNOK( pcParameterIoIf->value( "ResidualFile",     rcGP.m_cResidualName,            true ) );
  RNOK( pcParameterIoIf->value( "PredictionFile",   rcGP.m_cPredictionName,          true ) );
  RNOK( pcParameterIoIf->value( "BitstreamFile",    rcGP.m_cBitstreamName,           true ) );
  {
    std::string cInterPlanePred;
    RNOK( pcParameterIoIf->value( "InterPlanePrediction", cInterPlanePred, true ) );
    RNOK( xParseStringOfTs( cInterPlanePred, rcGP.m_auiInterPlanePred ) );
  }
  pcParameterIoIf->value( "DisableTwoPassEncoding", rcGP.m_bDisableTwoPassEncoding, false );
  return Err::m_nOK;
}


ErrVal VideoEncoderParameter::PrecodedPictureParsingParams::process( ParameterIoIf* pcParameterIoIf )
{
  PrecodedPictureParsingParams& rcPars = *this;
  RNOK( pcParameterIoIf->value( "PrecodedBitstream",                  rcPars.m_cPrecodedBitstreamName, true ) );
  RNOK( pcParameterIoIf->value( "NumFramesToParseFromPrecodedStream", rcPars.m_uiNumFramesToParse,     true ) );
  if( rcPars.m_uiNumFramesToParse == 0 )
  {
    rcPars.m_cPrecodedBitstreamName.clear();
  }
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::MultithreadingControlParams::process( ParameterIoIf* pcParameterIoIf )
{
  MultithreadingControlParams& rcPars = *this;
  pcParameterIoIf->value( "NumAccessUnitEncodingThreads", rcPars.m_uiNumAccessUnitEncodingThreads, false );
  if( rcPars.m_uiNumAccessUnitEncodingThreads > 1 )
  {
    RNOK( pcParameterIoIf->value( "NumPicsInMultithreadingQueue", rcPars.m_uiNumPicsInMultithreadingQueue, false ) );
  }
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::EntropySliceParams::process( ParameterIoIf* pcParameterIoIf )
{
  EntropySliceParams& rcPars = *this;
  pcParameterIoIf->value( "EntropySliceSizeX",  rcPars.m_uiEntropySliceSizeX, false );
  pcParameterIoIf->value( "EntropySliceSizeY",  rcPars.m_uiEntropySliceSizeY, false );
  pcParameterIoIf->value( "ContextStatesReuse", rcPars.m_bContextStatesReuse, false );
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::SPSParams::process( ParameterIoIf* pcParameterIoIf )
{
  SPSParams& rcSPS = *this;
  RNOK( pcParameterIoIf->value( "NumberReferenceFrames", rcSPS.m_uiNumberReferenceFrames,  true ) );
  RNOK( pcParameterIoIf->value( "CodedPictureStoreSize", rcSPS.m_uiCodedPictureStoreSize,  true ) );
  RNOK( pcParameterIoIf->value( "InterpolationMethod",   rcSPS.m_uiInterpolationMethod  ,  true ) );
//  RNOK( pcParameterIoIf->value( "ProfileIdc",            rcSPS.m_uiProfileIdc,  true ) );
//  RNOK( pcParameterIoIf->value( "LevelIdc",              rcSPS.m_uiLevelIdc,  true ) );
  pcParameterIoIf->value( "AdapInterpSwitch",      rcSPS.m_uiAdapInterpSwitch     ,  false );
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::DeblockingFilterParams::process( ParameterIoIf* pcParameterIoIf )
{
  DeblockingFilterParams& rcLF = *this;
  RNOK( pcParameterIoIf->value( "DisableDeblockingFilter",  rcLF.m_uiFilterIdc,       true ) );
  RNOK( pcParameterIoIf->value( "AlphaC0OffsetDiv2",        rcLF.m_iAlphaOffsetDiv2,  true ) );
  RNOK( pcParameterIoIf->value( "BetaOffsetDiv2",           rcLF.m_iBetaOffsetDiv2,   true ) );
  return Err::m_nOK;
}  

ErrVal VideoEncoderParameter::InLoopWienerFilterParams::process( ParameterIoIf* pcParameterIoIf)
{
  InLoopWienerFilterParams& rcLWF = *this;
  RNOK( pcParameterIoIf->value( "DisableInLoopFilter",   rcLWF.m_bDisableFilter,    true ) );
  if( !rcLWF.m_bDisableFilter )
  {
    pcParameterIoIf->value( "InLoopFilterType",      rcLWF.m_uiFilterType,      false);
    pcParameterIoIf->value( "InLoopMinLength",       rcLWF.m_uiMinFilterLength, false);
    pcParameterIoIf->value( "InLoopMaxLength",       rcLWF.m_uiMaxFilterLength, false);
    pcParameterIoIf->value( "InLoopUseBlocks",       rcLWF.m_bUseBlocks,        false);
    pcParameterIoIf->value( "InLoopMaxBlock",        rcLWF.m_uiPower2MaxBlock,  false);
    pcParameterIoIf->value( "InLoopMinBlock",        rcLWF.m_uiPower2MinBlock,  false);
  }
  return Err::m_nOK;
}



ErrVal VideoEncoderParameter::MotionVectorSearchParams::process( ParameterIoIf* pcParameterIoIf )
{
  MotionVectorSearchParams& rcME = *this;
  RNOK( pcParameterIoIf->value( "SearchModeFullPel",      rcME.m_uiSearchModeFullPel,   true ) );
  RNOK( pcParameterIoIf->value( "SearchModeSubPel",       rcME.m_uiSearchModeSubPel,    true ) );
  RNOK( pcParameterIoIf->value( "SearchPrimaryPlaneOnly", rcME.m_bSearchPrimPlaneOnly,  true ) );
  RNOK( pcParameterIoIf->value( "SearchRange",            rcME.m_uiSearchRange,         true ) );
  RNOK( pcParameterIoIf->value( "MaxIterations",          rcME.m_uiMaxIterations,       true ) );
  RNOK( pcParameterIoIf->value( "IterSearchRange",        rcME.m_uiIterSearchRange,     true ) );
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::QuadtreeParams::process( ParameterIoIf* pcParameterIoIf )
{
  QuadtreeParams& rcQtP = *this;
  {
    std::string cMvCompetition;
    RNOK( pcParameterIoIf->value( "MvCompetition", cMvCompetition, true ) );
    RNOK( xParseStringOfTs( cMvCompetition, rcQtP.m_auiMvCompetition ) );
  }
  {
    std::string cLog2MvAccuracy;
    RNOK( pcParameterIoIf->value( "Log2MvAccuracy", cLog2MvAccuracy, true ) );
    RNOK( xParseStringOfTs( cLog2MvAccuracy, rcQtP.m_auiLog2MvAccuracy ) );
  }
  {
    std::string cLog2MaxPredBlkSize;
    RNOK( pcParameterIoIf->value( "Log2MaxPredictionBlockSize", cLog2MaxPredBlkSize, true ) );
    RNOK( xParseStringOfTs( cLog2MaxPredBlkSize, rcQtP.m_auiLog2MaxPredictionBlockSize ) );
  }
  {
    std::string cLog2MinPredBlkSize;
    RNOK( pcParameterIoIf->value( "Log2MinPredictionBlockSize", cLog2MinPredBlkSize, true ) );
    RNOK( xParseStringOfTs( cLog2MinPredBlkSize, rcQtP.m_auiLog2MinPredictionBlockSize ) );
  }
  {
    std::string cLog2MaxTransBlkSize;
    RNOK( pcParameterIoIf->value( "Log2MaxTransformBlockSize", cLog2MaxTransBlkSize, true ) );
    RNOK( xParseStringOfTs( cLog2MaxTransBlkSize, rcQtP.m_auiLog2MaxTransformBlockSize ) );
  }
  {
    std::string cLog2MinTransBlkSize;
    RNOK( pcParameterIoIf->value( "Log2MinTransformBlockSize", cLog2MinTransBlkSize, true ) );
    RNOK( xParseStringOfTs( cLog2MinTransBlkSize, rcQtP.m_auiLog2MinTransformBlockSize ) );
  }
  {
    std::string cSubdivideRes;
    RNOK( pcParameterIoIf->value( "TrafoPow2SubdivideResidual", cSubdivideRes, true ) );
    RNOK( xParseStringOfTs( cSubdivideRes, rcQtP.m_abTrafoPow2SubdivideResidual ) );
  }
  {
    std::string cQuantMode;
    RNOK( pcParameterIoIf->value( "QuantizationMode", cQuantMode, true ) );
    RNOK( xParseStringOfTs( cQuantMode, rcQtP.m_auiQuantizationMode ) );
  }
  RNOK( pcParameterIoIf->value( "DisableIntraInInter",           rcQtP.m_bDisableIntraInInter,            true ) );
  RNOK( pcParameterIoIf->value( "MultiCodewordArithmeticCoding", rcQtP.m_uiMultiCodewordArithmeticCoding, true ) );
  {
    std::string acStringSig;
    RNOK( pcParameterIoIf->value("TrafoPow2ContextModelStringSig", acStringSig, true ) );
    RNOK( xParseStringOfTs( acStringSig, rcQtP.m_auiTrafoPow2ContextModelsSig ) );
  }
  {
    std::string acStringLast;
    RNOK( pcParameterIoIf->value("TrafoPow2ContextModelStringLast", acStringLast, true ) );
    RNOK( xParseStringOfTs( acStringLast, rcQtP.m_auiTrafoPow2ContextModelsLast ) );
  }
  pcParameterIoIf->value( "FastModeDecision",   rcQtP.m_bFastModeDecision,    false );
  pcParameterIoIf->value( "FMDThresholdHighQP", rcQtP.m_uiFMDThresholdHighQP, false );
  pcParameterIoIf->value( "FMDThresholdLowQP",  rcQtP.m_uiFMDThresholdLowQP,  false );
  pcParameterIoIf->value( "VLCMode",            rcQtP.m_uiVLCMode,            false );
  pcParameterIoIf->value( "VLCMaxDelay",        rcQtP.m_uiVLCMaxDelay,        false );
  return Err::m_nOK;
}

bool VideoEncoderParameter::QuadtreeParams::xCheckIfPowerOfTwoWithEvenExponent( UInt uiVal )
{
  if( uiVal == 0 )
  {
    return false;
  }
  const UInt uiExponent = gCeilLog2( uiVal );
  return( ( UInt(1) << uiExponent ) == uiVal && ( uiExponent & 1 ) == 0 );
}


template<class T>
ErrVal VideoEncoderParameter::xParseStringOfTs( const std::string &racString, std::vector<T> &ratVec )
{
  std::istringstream cStream( racString );
  while( !cStream.eof() )
  {
    T t = 0;
    cStream >> t;
    ratVec.push_back( t );
  }
  return Err::m_nOK;
}

ErrVal VideoEncoderParameter::xParseStringOfRatios( const std::string &racString, std::vector<std::pair<UInt,UInt> > &rapuiuiVec )
{
  std::string cWithoutWS = racString.substr( racString.find_first_not_of( "\t " ), racString.find_last_not_of( "\t " ) + 1 - racString.find_first_not_of( "\t " ) );
  std::istringstream cStream( cWithoutWS );
  while( !cStream.eof() )
  {
    UInt uiNumerator = 0;
    UChar ucDelimiter = 0;
    UInt uiDenominator = 0;

    cStream >> uiNumerator >> ucDelimiter >> uiDenominator;
    ROF( '/' == ucDelimiter );
    ROT( 0 == uiDenominator );
    ROT( uiNumerator > uiDenominator );
    rapuiuiVec.push_back( std::make_pair( uiNumerator, uiDenominator ) );
  }
  return Err::m_nOK;
}



ErrVal VideoEncoderParameter::QuantParams::process( ParameterIoIf* pcParameterIoIf )
{
  QuantParams& rcQP = *this;
  RNOK( pcParameterIoIf->value( "PPSQuant",       rcQP.m_iPPSQuant,               true ) );
  RNOK( pcParameterIoIf->value( "LambdaCalcMode", rcQP.m_uiLambdaCalculationMode, true ) );
  {
    std::string cMaxAbsQPDelta;
    RNOK( pcParameterIoIf->value( "MaxAbsQPDelta", cMaxAbsQPDelta, true ) );
    RNOK( xParseStringOfTs( cMaxAbsQPDelta, rcQP.m_auiMaxAbsQPDelta ) );
  }
  {
    std::string cLayerQpOffsets;
    RNOK( pcParameterIoIf->value( "LayerQpOffsets", cLayerQpOffsets, true ) );
    RNOK( xParseStringOfTs( cLayerQpOffsets, rcQP.m_aiLayerQpOffsets ) );
  }
  {
    std::string cLayerDeadzones;
    RNOK( pcParameterIoIf->value( "LayerDeadzoneInter", cLayerDeadzones, true ) );
    RNOK( xParseStringOfRatios( cLayerDeadzones, rcQP.m_apuiuiLayerDeadzonesInter ) );
  }
  {
    std::string cLayerDeadzones;
    RNOK( pcParameterIoIf->value( "LayerDeadzoneIntra", cLayerDeadzones, true ) );
    RNOK( xParseStringOfRatios( cLayerDeadzones, rcQP.m_apuiuiLayerDeadzonesIntra ) );
  }
  return Err::m_nOK;
}


ErrVal VideoEncoderParameter::GOPSettings::process( ParameterIoIf* pcParameterIoIf )
{
  GOPSettings& rcGOP = *this;
  RNOK( pcParameterIoIf->value( "GOPFormatString", rcGOP.m_cInputFormatString,   true ) );
  pcParameterIoIf->value( "LayerChangeStartPoc",   rcGOP.m_iLayerChangeStartPoc, false );
  pcParameterIoIf->value( "LayerOffset",           rcGOP.m_iLayerOffset,         false );
  return Err::m_nOK;
}


ErrVal VideoEncoderParameter::xProcessConfigFileParameters( ParameterIoIf* pcParameterIoIf )
{
  RNOK( m_cGeneralParams.               process( pcParameterIoIf ) );
  RNOK( m_cSPSParams.                   process( pcParameterIoIf ) );
  RNOK( m_cGOPSettings.                 process( pcParameterIoIf ) );
  RNOK( m_cQuantParams.                 process( pcParameterIoIf ) );
  RNOK( m_cMotionVectorSearchParams.    process( pcParameterIoIf ) );
  RNOK( m_cLoopFilterParams.            process( pcParameterIoIf ) );
  RNOK( m_cInLoopWienerParams.          process( pcParameterIoIf ) );
  RNOK( m_cQuadtreeParams.              process( pcParameterIoIf ) );
  RNOK( m_cPrecodedPictureParsingParams.process( pcParameterIoIf ) );
  RNOK( m_cMultithreadingControlParams .process( pcParameterIoIf ) );
  RNOK( m_cEntropySliceParams          .process( pcParameterIoIf ) );
  return Err::m_nOK;
}
