/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef WIENERFILTER_ENCODER_H_
#define WIENERFILTER_ENCODER_H_




#include "VideoCommonLib/WienerFilterParameter.h"
#include "RDCost.h"
#include "StateIf.h"
#include "VideoCommonLib/WienerFilter.h"
#include "VideoCommonLib/AccessUnit.h"
#include "AccessUnitEncoderParameters.h"


#include "VideoCommonLib/ContextTables.h"
#include "SingleNodeEstimatorWF.h"
#include "WienerEncoder.h"
#include "TraceEnc.h"
#include "SyntaxElementProcessingChainEnc.h"
#include "SubdivisionEstimator.h"
#include "WienerCorrEstimator.h"
#include "../VideoDecoderLib/WienerDecoderIterator.h"
#include <iostream>
#include <fstream>


// video namespace begin
VIDEO_NAMESPACE_BEGIN

class WienerResult
{
public:
  WienerResult( Lambda cLambda )
    : m_cCost( cLambda )
    , m_pcEnc( NULL )
    , m_pcState( NULL )
    , m_pcWFParam( NULL ){}


  ErrVal setEncBlock( PicBlock<Pel> *pcEncBlock ) { ROT( !m_pcEnc   == !pcEncBlock ); m_pcEnc = pcEncBlock;  return Err::m_nOK; }
  ErrVal setState( State *pcState )        { ROT( !m_pcState == !pcState );    m_pcState = pcState;   return Err::m_nOK; }
  ErrVal setWFParam( WienerFilterParameters* pWfParam) {m_pcWFParam = pWfParam; return Err::m_nOK; }

  RDCost& getRDCost()          { return m_cCost; }
  PicBlock<Pel>* getEncBlock() { return m_pcEnc; }
  State* getState()   { return m_pcState; }
  WienerFilterParameters& getWFparam() { return *m_pcWFParam; }
  const WienerFilterParameters& getWFparam() const {return *m_pcWFParam; }

private:
  RDCost m_cCost;
  PicBlock<Pel> *m_pcEnc;
  State *m_pcState;
  WienerFilterParameters* m_pcWFParam;
};


template<class t_Pel>
class WienerFilterEncoder
  : WienerFilter<t_Pel>
{
public:
  static ErrVal estimateAndEncodeInLoopFilter( AccessUnitParameters &rcAUPars,
                                               const AccessUnitEncoderParameters &rcAUEstPar,
                                               ContextStore< CabacContextModel > &rcContextStore,
                                               const PicBlock<t_Pel> &rcDecPic,
                                               const PicBlock<const t_Pel> &rcOrigPic,
                                               ProbBinStreamEncIf *pcProbBinOut,
                                               int& iQp,
                                               Lambda rcLambda,
                                               TraceEnc *pcTrace,
                                               Block<double>* &rpcInLoopFiltMatrix,
                                               bool bAdapInterpSwitch );
  static ErrVal encodeInLoopFilter(AccessUnitParameters &rcAUPars,
                                  ProbBinStreamEncIf *pcProbBinOut,
                                  int& iQp,
                                  TraceEnc *pcTrace);
  
private:
  static bool solveSLE (std::vector< std::vector<double> > rcM, std::vector<double> &rcResult);
  static ErrVal testFullFrameSep(const PicBlock<t_Pel> &rcDecPic,
                                 const PicBlock<const t_Pel> &rcOrigPic,
                                 WienerResult* &pcBestResult,
                                 BitRateEstimatorIf* pcBitRateEstIf,
                                 SyntaxElementEncoderIf* pcSEEncoderIf,
                                 StateIf *pcStateIf,
                                 Lambda &rcLambda,
                                 int iFilterType,
                                 int iFilterMinLength,
                                 int iFilterMaxLength );

};


template<class t_Pel>
ErrVal WienerFilterEncoder<t_Pel>::estimateAndEncodeInLoopFilter( AccessUnitParameters &rcAUPars,
                                                                  const AccessUnitEncoderParameters &rcAUEstPar,
                                                                  ContextStore< CabacContextModel > &rcContextStore,
                                                                  const PicBlock<t_Pel> &rcDecPic,
                                                                  const PicBlock<const t_Pel> &rcOrigPic,
                                                                  ProbBinStreamEncIf *pcProbBinOut,
                                                                  int& iQp,
                                                                  Lambda rcLambda,
                                                                  TraceEnc *pcTrace,
                                                                  Block<double>* &rpcInLoopFiltMatrix,
                                                                  bool bAdapInterpSwitch )
{
  if( !rcAUEstPar.getDisableWienerFlag() )
  {
    UInt uiNumOfPlanes = rcDecPic.getStats().getNumPlanes();
    WienerResult* pcCurrentResult = new WienerResult( rcLambda );
    WienerResult* pcBestResult = new WienerResult( rcLambda );

    // init best and current result with decoded frame
    pcBestResult->setEncBlock( new AllocPicBlock<t_Pel>( rcDecPic.getStats() ) );
    pcBestResult->getEncBlock()->assign( rcDecPic );
    pcBestResult->setWFParam( new WienerFilterParameters() );
    pcCurrentResult->setEncBlock( new AllocPicBlock<t_Pel>( rcDecPic.getStats() ) );
    pcCurrentResult->getEncBlock()->assign( rcDecPic );
    pcCurrentResult->setWFParam( new WienerFilterParameters() );

    ContextArrayStartIndices cCtxIndices ;
    CtxInitArray cCtxInit ;
    cCtxInit = ContextTables::m_cCtxWiener;
    cCtxIndices.m_uiStartRefIdx = cCtxInit.getCtxAccess().getNumCtxInitArrays();

    StateMultiplier cStateMultiplier;
    SyntaxElementProcessingChainEnc<true, DefaultCabacTables, false> cSEProcChainEst( &cStateMultiplier, NULL, cCtxIndices, cCtxInit, rcAUPars.getPlaPaPars(0), iQp, NULL );

    BitRateEstimatorIf*  pcBitRateEstIf = &cSEProcChainEst.getBitRateEstimatorIf();
    SyntaxElementEncoderIf* pcSEEncoderIf = &cSEProcChainEst.getSyntaxElementEncoderIf();

    StateIf *pcStateIfForEncoder = &cStateMultiplier;

    //store current state and bitrate
    BitRate cRootRate( 0 );
    if( pcBitRateEstIf )
    {
      cRootRate = pcBitRateEstIf->getBitRateEstimate();
    }

    State *pcRootState = NULL;
    if( pcStateIfForEncoder )
    {
      RNOK( pcStateIfForEncoder->getState( pcRootState ) );
    }

    // no filter
    pcBestResult->getWFparam().setUseFilterFlag(false);

    //transmit params
    pcSEEncoderIf->processWienerParameters( pcBestResult->getWFparam() , uiNumOfPlanes);

    pcBestResult->getRDCost().reset();
    pcBestResult->getRDCost().addRate( pcBitRateEstIf->getBitRateEstimate() - cRootRate);
    pcBestResult->getRDCost().addDistortion(rcOrigPic,rcDecPic);
    RNOK( pcRootState->setState( ));

    // full frame
    RNOK (testFullFrameSep( rcDecPic,
                            rcOrigPic,
                            pcCurrentResult,
                            pcBitRateEstIf,
                            pcSEEncoderIf,
                            pcStateIfForEncoder,
                            rcLambda,
                            rcAUEstPar.getWienerFilterType(),
                            rcAUEstPar.getWienerMinLength(),
                            rcAUEstPar.getWinerMaxLength() ) );

    // copy filter for adaptive interpolation switching
    if ( bAdapInterpSwitch && pcCurrentResult->getWFparam().getUseFilterFlag() )
    {
      const std::vector<int> aiHorizontalFiltCoeff = pcCurrentResult->getWFparam().getHorizontalCoeffOfPlane( 0 ).getQuantCoeffs();
      const std::vector<int> aiVerticalFiltCoeff = pcCurrentResult->getWFparam().getVerticalCoeffOfPlane( 0 ).getQuantCoeffs();
      const UInt uiHorizontalQuantBits = pcCurrentResult->getWFparam().getHorizontalCoeffOfPlane( 0 ).getNumberOfQuantizationBits();
      const UInt uiVerticalQuantBits = pcCurrentResult->getWFparam().getVerticalCoeffOfPlane( 0 ).getNumberOfQuantizationBits();
      const UInt uiHorizontalFilterLength = pcCurrentResult->getWFparam().getHorizontalCoeffOfPlane( 0 ).getFilterLength();
      const UInt uiVerticalFilterLength = pcCurrentResult->getWFparam().getVerticalCoeffOfPlane( 0 ).getFilterLength();
      const std::vector<UInt> auiHorizontalMap = pcCurrentResult->getWFparam().getHorizontalCoeffOfPlane( 0 ).getMappingRule();
      const std::vector<UInt> auiVerticalMap = pcCurrentResult->getWFparam().getVerticalCoeffOfPlane( 0 ).getMappingRule();
      rpcInLoopFiltMatrix = new AllocBlock<double>( BlockSize(uiHorizontalFilterLength, uiVerticalFilterLength) );

      UInt iX, iY;
      for ( iX = 0; iX < uiHorizontalFilterLength; iX++ )
      {
        for ( iY = 0; iY < uiVerticalFilterLength; iY++ )
        {
           double dHorCoeff = double ( aiHorizontalFiltCoeff[ auiHorizontalMap[iX] ] ) / ( 1 << uiHorizontalQuantBits );
           double dVerCoeff = double ( aiVerticalFiltCoeff  [   auiVerticalMap[iY] ] ) / ( 1 << uiVerticalQuantBits );
           rpcInLoopFiltMatrix->getEl( iX, iY ) = dHorCoeff * dVerCoeff;
        }
      }
    }

    if ( pcCurrentResult->getWFparam().getUseFilterFlag() )
    {
      PicBlock<t_Pel> *pcFilteredPic = new AllocPicBlock<t_Pel>( rcDecPic.getStats() );
      pcFilteredPic->assign( *pcCurrentResult->getEncBlock() );

      //transmit WFparam from filtered result
      pcSEEncoderIf->processWienerParameters(pcCurrentResult->getWFparam(), uiNumOfPlanes );

      pcCurrentResult->getRDCost().reset();
      pcCurrentResult->getRDCost().addRate( pcBitRateEstIf->getBitRateEstimate() - cRootRate );
      pcCurrentResult->getRDCost().addDistortion(rcOrigPic, *pcFilteredPic );
      RNOK( pcRootState->setState( ));

      if (pcCurrentResult->getRDCost() < pcBestResult->getRDCost() )
      {
        gSwapPtr( pcCurrentResult, pcBestResult );
        pcCurrentResult->getWFparam() = pcBestResult->getWFparam();
      }
      pcCurrentResult->getEncBlock()->assign( rcDecPic );

      // QuadTree
      pcCurrentResult->getWFparam().setUseBlocksFlag(true);
      pcCurrentResult->getWFparam().setPower2MaxBlock( rcAUEstPar.getWienerPower2MaxBlock() );
      pcCurrentResult->getWFparam().setPower2MinBlock( rcAUEstPar.getWienerPower2MinBlock() );

      pcSEEncoderIf->processWienerParameters(pcCurrentResult->getWFparam() , uiNumOfPlanes );

      BitRate cRateWithoutTree( 0 );
      cRateWithoutTree = pcBitRateEstIf->getBitRateEstimate();

      pcCurrentResult->getWFparam().initQuadtreeGroupStructure(rcDecPic.getStats().getPrimBlockSize());
      pcCurrentResult->getWFparam().setQuadTreeUsedExternFlag( true );

      SingleNodeEstimatorWF<WienerBlockPar*> cSingleNodeEstWf( &cSEProcChainEst.getSyntaxElementEncoderIf(), &cSEProcChainEst.getBitRateEstimatorIf(), pcCurrentResult->getWFparam().getQTBlockPars() , rcLambda, *pcCurrentResult->getEncBlock() , *pcFilteredPic , rcOrigPic, pcStateIfForEncoder );

      SubdivisionEstimator< SingleNodeEstimatorWF_IfType, CabacContextModel> cLeafNodeEst( rcAUEstPar, pcCurrentResult->getWFparam().getQuadtreeGroupStructure(),
                                                                                           cSingleNodeEstWf,
                                                                                           cSEProcChainEst.getProbAlloc(),
                                                                                           NULL,
                                                                                           NULL,
                                                                                           rcLambda,
                                                                                           pcStateIfForEncoder,
                                                                                           NULL );

      RNOK( cLeafNodeEst.estimateTree() );
      BitRate cRateQuadTree( 0 );
      cRateQuadTree = pcBitRateEstIf->getBitRateEstimate() - cRateWithoutTree;

      pcCurrentResult->getRDCost().reset();
      pcCurrentResult->getRDCost().addRate( pcBitRateEstIf->getBitRateEstimate() - cRootRate );
      pcCurrentResult->getRDCost().addDistortion(rcOrigPic, *pcCurrentResult->getEncBlock() );

      RNOK( pcRootState->setState( ));
      if (pcCurrentResult->getRDCost() < pcBestResult->getRDCost() )
      {
        gSwapPtr( pcCurrentResult, pcBestResult );
        pcCurrentResult->getWFparam() = pcBestResult->getWFparam();
        pcCurrentResult->getEncBlock()->assign( *pcBestResult->getEncBlock() );
      }


      PicBlock<t_Pel>* pcQuadtreeResult = new AllocPicBlock<t_Pel>( rcDecPic.getStats() );
      pcQuadtreeResult->assign( *pcCurrentResult->getEncBlock() );

      delete pcQuadtreeResult;
      delete pcFilteredPic;

    }


    if (pcBestResult->getWFparam().getUseFilterFlag() )
    {

      rcAUPars.getWienerFilterParameters() = pcBestResult->getWFparam();

      //actually encoding (settings from pcBestResult)
      rcDecPic.assign( *pcBestResult->getEncBlock() );
      encodeInLoopFilter(rcAUPars, pcProbBinOut, iQp, pcTrace) ;

    }

    if( pcRootState )
    {
      pcRootState->deleteState();
    }

    if( pcCurrentResult->getWFparam().getUseBlocksFlag() )
    {
      pcCurrentResult->getWFparam().setQuadTreeUsedExternFlag( false );
    }
    else
    {
      pcBestResult->getWFparam().setQuadTreeUsedExternFlag( false );
    }

    rcAUPars.getWienerFilterParameters().setUseFilterFlag( pcBestResult->getWFparam().getUseFilterFlag() );

    delete pcCurrentResult->getEncBlock();
    delete &pcCurrentResult->getWFparam();
    delete pcCurrentResult;

    delete pcBestResult->getEncBlock();
    delete &pcBestResult->getWFparam();
    delete pcBestResult;

  }
  else
  {
    rcAUPars.getWienerFilterParameters().setUseFilterFlag( false );
  }
  return Err::m_nOK;
}

template<class t_Pel>
ErrVal WienerFilterEncoder<t_Pel>::encodeInLoopFilter(AccessUnitParameters &rcAUPars,
                                                      ProbBinStreamEncIf *pcProbBinOut,
                                                      int& iQp,
                                                      TraceEnc *pcTrace)
{

  ContextArrayStartIndices cCtxIndices ;
  CtxInitArray cCtxInit ;
  cCtxInit = ContextTables::m_cCtxWiener;
  cCtxIndices.m_uiStartRefIdx = cCtxInit.getCtxAccess().getNumCtxInitArrays();

  //actually encoding (settings from pcBestResult)
  SyntaxElementProcessingChainEnc<false, DefaultCabacTables, true> cSEProcChainEnc( NULL,
                                                                                    pcProbBinOut,
                                                                                    cCtxIndices,
                                                                                    cCtxInit,
                                                                                    rcAUPars.getPlaPaPars(0),
                                                                                    iQp,
                                                                                    pcTrace );


   SyntaxElementEncoderIf* pcSEEncoderIfENC = &cSEProcChainEnc.getSyntaxElementEncoderIf();
   pcSEEncoderIfENC->processWienerParameters(rcAUPars.getWienerFilterParameters() , rcAUPars.getNumPlanes() );
   //use quadTree?
   if ( rcAUPars.getWienerFilterParameters().getUseBlocksFlag() )
   {
      WienerEncoder cWienerEnc( rcAUPars.getWienerFilterParameters().getQuadtreeGroupStructure(),
                                  rcAUPars.getWienerFilterParameters().getQTBlockPars(),
                                  &cSEProcChainEnc.getSyntaxElementEncoderIf(),
                                  pcTrace );
      RNOK( cWienerEnc.processTree() );
   }




  return Err::m_nOK;

}




template<class t_Pel>
ErrVal WienerFilterEncoder<t_Pel>::testFullFrameSep(const PicBlock<t_Pel> &rcDecPic,
                           const PicBlock<const t_Pel> &rcOrigPic,
                           WienerResult* &pcBestResult,
                           BitRateEstimatorIf* pcBitRateEstIf,
                           SyntaxElementEncoderIf* pcSEEncoderIf,
                           StateIf *pcStateIf,
                           Lambda &rcLambda,
                           int iFilterType,
                           int iFilterMinLength,
                           int iFilterMaxLength )
{
  UInt uiNumOfPlanes = rcDecPic.getStats().getNumPlanes();
  pcBestResult->getRDCost().reset();
  pcBestResult->getRDCost().setMax();

  //store current state and bitrate
  BitRate cRootRate( 0 );
  if( pcBitRateEstIf )
  {
    cRootRate = pcBitRateEstIf->getBitRateEstimate();
  }
  State *pcRootState = NULL;
  if( pcStateIf )
  {
    RNOK( pcStateIf->getState( pcRootState ) );
  }

  WienerResult* pcCurrentResult = new WienerResult( rcLambda );
  pcCurrentResult->setEncBlock( new AllocPicBlock<t_Pel>( rcDecPic.getStats() ) );
  pcCurrentResult->getEncBlock()->assign( rcDecPic );
  pcCurrentResult->setWFParam( new WienerFilterParameters() );
  std::vector< std::vector<double> > cHorCorr ;
  std::vector< std::vector<double> > cVerCorr ;
  std::vector<double> dHorCoeff;
  std::vector<double> dVerCoeff;

  std::vector<UInt> auiSets;
  std::vector<UInt> auiTypes;
  std::vector<UInt> auiLength;
  auiSets.push_back( 1 );
  auiSets.push_back( 2 );
  auiSets.push_back( 3 );
  auiTypes.push_back( iFilterType );
  auiTypes.push_back( iFilterType );
  auiTypes.push_back( iFilterType );
  auiLength.push_back( iFilterMaxLength );
  auiLength.push_back( iFilterMaxLength );
  auiLength.push_back( iFilterMaxLength );

  WienerCorrEstimatorBase<t_Pel> cWCorrBaseHoriz( rcDecPic , rcOrigPic , auiSets ,  auiTypes , auiLength );
  RNOK( cWCorrBaseHoriz.addCorrHorizontal() );

  WienerCorrEstimatorBase<t_Pel> cWCorrBaseVert( rcDecPic , rcOrigPic , auiSets ,  auiTypes , auiLength );
  RNOK( cWCorrBaseVert.addCorrVertical() );




  for( UInt uiPlane = 0; uiPlane < rcDecPic.getStats().getNumPlanes(); uiPlane++ )
  {
    FilterCoefficients cf_tempHorizontal;
    FilterCoefficients cf_tempVertical;

    cf_tempHorizontal.setHorizontal() ;
    cf_tempVertical.setVertical() ;

    for (UInt uiAvailCoeffs=0; uiAvailCoeffs < uiPlane ; uiAvailCoeffs++ )
    {
      pcCurrentResult->getEncBlock()->getPlane(uiAvailCoeffs).assign( pcBestResult->getEncBlock()->getPlane(uiAvailCoeffs) );
      pcCurrentResult->getWFparam().setFilterSetForPlane( uiAvailCoeffs , pcBestResult->getWFparam().getSetNumOfPlane( uiAvailCoeffs ) );
      pcCurrentResult->getWFparam().setFilterPlaneFlag( uiAvailCoeffs, pcBestResult->getWFparam().getFilterPlaneFlag( uiAvailCoeffs ));
    }

    for (UInt uiAvailSets=0; uiAvailSets < pcBestResult->getWFparam().getNewSetNumberSep() ; uiAvailSets++)
      {
        pcCurrentResult->getWFparam().getHorizontalFilterCoeff(uiAvailSets) = pcBestResult->getWFparam().getHorizontalFilterCoeff(uiAvailSets);
      }

     //check available coeffs first
     for (UInt uiAvailSet=0; uiAvailSet < pcCurrentResult->getWFparam().getNewSetNumberSep() ; uiAvailSet++ )
     {
       cf_tempHorizontal = pcBestResult->getWFparam().getHorizontalFilterCoeff( uiAvailSet);
       cf_tempVertical   = pcBestResult->getWFparam().getVerticalFilterCoeff( uiAvailSet);

       AllocBlock<t_Pel> cHorizontallyFilteredBlock(rcDecPic.getPlane( uiPlane ).getSize());

       filterBlockSeparableHor( rcDecPic.getPlane( uiPlane ), cHorizontallyFilteredBlock, cf_tempHorizontal);
       filterBlockSeparableVer( cHorizontallyFilteredBlock, pcCurrentResult->getEncBlock()->getPlane( uiPlane), cf_tempVertical);

       pcCurrentResult->getWFparam().setFilterPlaneFlag( uiPlane , true);
       pcCurrentResult->getWFparam().setFilterSetForPlane( uiPlane, uiAvailSet);

       //send filter params
       pcSEEncoderIf->processWienerParameters(pcCurrentResult->getWFparam() , uiNumOfPlanes );

       pcCurrentResult->getRDCost().reset();
       pcCurrentResult->getRDCost().addRate( pcBitRateEstIf->getBitRateEstimate() - cRootRate );
       pcCurrentResult->getRDCost().addDistortion( rcOrigPic, *pcCurrentResult->getEncBlock() );

       if ( pcCurrentResult->getRDCost() < pcBestResult->getRDCost() )
       {
         pcBestResult->getWFparam().setFilterPlaneFlag( uiPlane , true);
         pcBestResult->getWFparam().setFilterSetForPlane( uiPlane , uiAvailSet);
         pcBestResult->getEncBlock()->getPlane( uiPlane ).assign( pcCurrentResult->getEncBlock()->getPlane( uiPlane) );
         pcBestResult->getRDCost() = pcCurrentResult->getRDCost() ;
       }

       pcCurrentResult->getEncBlock()->getPlane( uiPlane ).assign( rcDecPic.getPlane( uiPlane ) );
       pcRootState->setState( );
     }


     UInt uiNewSetNumber = pcCurrentResult->getWFparam().getNewSetNumberSep();
   // get best params for current plane
     for (int iHorizontalFilterLength=iFilterMaxLength; iHorizontalFilterLength>iFilterMinLength-1;iHorizontalFilterLength-=2)
     {
       for (int iVerticalFilterLength=iFilterMaxLength; iVerticalFilterLength>iFilterMinLength-1;iVerticalFilterLength-=2)
       {
         if ( iHorizontalFilterLength == 1 && iVerticalFilterLength == 1)
             continue;

         cf_tempHorizontal.setFilterTypeSep(iFilterType, iHorizontalFilterLength);
         cf_tempVertical.setFilterTypeSep(iFilterType, iVerticalFilterLength);

         cHorCorr = cWCorrBaseHoriz.getCorrSeparable( uiPlane , iHorizontalFilterLength );

         bool bIsSingular = solveSLE(cHorCorr,dHorCoeff);
         AllocBlock<t_Pel> cHorizontallyFilteredBlock(rcDecPic.getPlane(uiPlane).getSize());
         if (! bIsSingular )
         {
           //filter horizontally
           cf_tempHorizontal.setCoefficients(dHorCoeff);
           filterBlockSeparableHor( rcDecPic.getPlane( uiPlane ), cHorizontallyFilteredBlock, cf_tempHorizontal);

           cVerCorr = cWCorrBaseVert.getCorrSeparable( uiPlane, iVerticalFilterLength);
           bIsSingular = solveSLE(cVerCorr,dVerCoeff) ;
         }
         if (!bIsSingular)
         {
           cf_tempVertical.setCoefficients(dVerCoeff);

           filterBlockSeparableVer(cHorizontallyFilteredBlock,pcCurrentResult->getEncBlock()->getPlane( uiPlane ), cf_tempVertical ) ;

           pcCurrentResult->getWFparam().setFilterPlaneFlag( uiPlane , true);
           pcCurrentResult->getWFparam().setFilterSetForPlane( uiPlane, uiNewSetNumber );
           pcCurrentResult->getWFparam().getHorizontalFilterCoeff( uiNewSetNumber ) = cf_tempHorizontal;
           pcCurrentResult->getWFparam().getVerticalFilterCoeff( uiNewSetNumber ) = cf_tempVertical;
           pcCurrentResult->getWFparam().setUseFilterFlag(true);
           //send filter params

           pcSEEncoderIf->processWienerParameters(pcCurrentResult->getWFparam() , uiNumOfPlanes );

           pcCurrentResult->getRDCost().reset();
           pcCurrentResult->getRDCost().addRate( pcBitRateEstIf->getBitRateEstimate() - cRootRate );
           pcCurrentResult->getRDCost().addDistortion( rcOrigPic, *pcCurrentResult->getEncBlock() );

           if ( pcCurrentResult->getRDCost() < pcBestResult->getRDCost() )
           {
             pcBestResult->getEncBlock()->getPlane( uiPlane ).assign( pcCurrentResult->getEncBlock()->getPlane( uiPlane)  );
             pcBestResult->getWFparam().setFilterPlaneFlag( uiPlane , true);

             pcBestResult->getWFparam().setFilterSetForPlane( uiPlane, uiNewSetNumber );
             pcBestResult->getWFparam().getHorizontalFilterCoeff( uiNewSetNumber ) = cf_tempHorizontal;
             pcBestResult->getWFparam().getVerticalFilterCoeff( uiNewSetNumber ) = cf_tempVertical;
             pcBestResult->getWFparam().setUseFilterFlag(true);
             pcBestResult->getRDCost() = pcCurrentResult->getRDCost() ;

           }

          if (pcBestResult->getWFparam().getSetNumOfPlane( uiPlane ) != uiNewSetNumber)
          {
            pcCurrentResult->getWFparam().removeLastSetSep();
          }
         }
          pcRootState->setState( );
        }
      }

  }

  if ( pcRootState )
  {
    pcRootState->deleteState();
  }
  delete pcCurrentResult->getEncBlock();
  delete &pcCurrentResult->getWFparam();
  delete pcCurrentResult;

  return Err::m_nOK;
}

template<class t_Pel>
bool WienerFilterEncoder<t_Pel>::solveSLE(std::vector< std::vector<double> > rcM, std::vector<double> &rcResult) // TODO rcM should be reference
{
  //solve SLE of Form Mx = a
  //a is last row of rcM

  int i,j,k;
  double t;
  int N = (int) rcM.size();

  rcResult.assign(N,0);

  for(k=0; k<N-1; k++)
  {
    for(i=k+1;i<N; i++)
    {
      t=rcM[i][k]/rcM[k][k];
      for(j=k+1; j<=N; j++)
      {
        rcM[i][j] -= t * rcM[k][j];
        if(i==j && fabs(rcM[i][j])<0.000001) return true;
      }
    }
  }
  for(i=N-1; i>=0; i--)
  {
    t = rcM[i][N];
    for(j=i+1; j<N; j++)
      t -= rcM[i][j] * rcM[j][N];
    rcResult[i] = t / rcM[i][i];
    rcM[i][N] = rcResult[i];
  }
  return false;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif
